;;
;; Variables
;;
(define (var-const?      var) (vector-ref var 1))
(define (var-num-lookups var) (vector-ref var 2))
(define (var-value       var)
  (vector-set! var 2 (+ 1 (var-num-lookups var)))
  (vector-ref var 0))

(define (variable . value)
 (vector (if (null? value) *unspecified* (car value))
         #t
         0))
 
;;
;; Bindings
;;

(define (binding symbol value) (cons symbol (variable value)))

(define (extend-env formals args env)
 (cond ((not (equal? (delete-duplicates formals) formals))
             (lookup-error "double binding in environment"))
       (else (cons (map binding formals args) env)) ))

(define (lookup symbol env)
 (let loop ((rest env))
  (let* ((binding (assq symbol (car rest))))
   (cond
    (binding            (var-value (cdr binding)))
    ((null? (cdr rest)) (lookup:import-sys symbol env))
    (else               (loop (cdr rest)) )) )))

(define (add-binding! env sym val)
 (set-car! env (cons (binding sym val) (car env))))

(define (lookup:import-sys symbol env)
 (let ((val-box       (cons #f *unspecified*))
       (sys-namespace (last-pair env)))
  (catch 'undefined (lambda () (set-cdr! val-box (eval symbol '()))
                               (set-car! val-box #t))
         (lambda (key . args) (begin #f)))
  (if (car val-box) ; sic!
    (begin (add-binding! sys-namespace symbol (cdr val-box)) ; TODO nlookup+1
           (cdr val-box))
    *unspecified*))) ; XXX return a variable object


;;
;; Errors ??? error objects for analyze?
;;

(define (lookup-error msg args)
  (error msg args)) ; XXX

(when #f
;;
;; Tests
(sys-load "VSI-core/repl.scm") (repl-transcript
(define env '(()))
(set! env (extend-env '(a b c) '(3 2 1) env))
(lookup 'a env)
(lookup 'define env)
env
(lookup 'how-about-sum-errr-hear? env)
)
)
