;;
;; File References
;;

(define (file-ref path line column)
  (vector 'file-ref path line column))

(define (fr-path   fr) (vector-ref fr 1))
(define (fr-line   fr) (vector-ref fr 2))
(define (fr-column fr) (vector-ref fr 3))

(define (file-ref? object)
 (and (vector? object) (eq? 'file-ref (vector-ref fr 0))))

(define (port-ref port) ; deprecation of counting ports?
  (file-ref (port-filename port)
            (+ 1 (port-line port))
            (+ 1 (port-column port)) ))

(define (fr-display fr)
  (format "File: ~a (line ~a col. ~a)"
          (fr-path fr) (fr-line fr) (fr-column fr)))

;;
;; Pins
;;

(define (file-referenced file-reference exp)
  (list 'sexpin (cons exp file-reference)))

(define (pin port exp) (file-referenced (port-ref port) exp))

(define (pin? object) (and (list? object) (eq? (car object) 'sexpin)))

(define (pin-point pin)
  (cdr (second pin)))

(define (pin-exp pin)
  (car (second pin)))

;;; TODO We need some unpin for complete datagraphs.
;;       [??? Last mental stop was pin-aware accessors for /analyzer/.]
;;      (Maybe the analyzing reader shouldn't always ...
;;       Try (readafile "/opt/VSI/VSI-core/uc.data.scm"). There are a lot
;;       of points but ... [but why did you use the /reader/ of /analyze/ then?
;;       It's only about two times as much memory objects which is
;;       standard and acceptable for the tasks /analyzer/ is meant for.]


;;
;;
;; Source Expressions (sexp)
;;  (pin aware prototypes)
;;
;;

(define (sexp-car sexp)
 (cond ((pin?  sexp) (sexp-car (pin-exp sexp)))
       ((pair? sexp) (car sexp))
       (else         (error "sexp-car wrong type argument" sexp)) ))

(define (sexp-cdr exp)
 (cond ((pin?  sexp) (sexp-cdr (pin-exp sexp)))
       ((pair? sexp) (cdr exp))
       (else         (error "sexp-cdr wrong type argument" exp))))

;;
;; Rapidd Scheme
(when #f (sys-load "VSI-core/repl.scm") (repl-transcript
))
