;; A lgorithmic
;; N otation                                 Read Graph
;; S cheme
;;
;;  ; -> comment up until the next line
;;  ( -> list terminated by )
;;    -> . next constituent is the final cdr value
;;    -> root
;;  " -> string terminated by "
;;    -> \" escaped quotation mark
;;    -> \\ escaped backslash
;;  # -> ! -> hash comment terminated by !#
;;       ( -> vector (like list) terminated by )
;;       " -> byte string terminated by "
;;       \ -> character descriptor terminated by a symbol terminator
;;       x -> hexidecimal number
;;       o -> octal number
;;       d -> decimal number
;;       e -> exponential decimal number
;;       f -> falsity
;;       t -> not falsity
;; other printables: literal symbol or numerical symbol
;; --
;; symbol terminator: white space, newline, ; ( " #
;; or unexpected banana: )
;;
;;
;;                                    Dotted Notation Corners
;;
;; Literal pair: ( CAR . CDR )
;; Literal pairs forming an improper list:
;;   ( CAR-1 . ( CAR-2 . CDR-2 )) => (CAR-1 CAR-2 . CDR-2)
;;
;; Void CAR:          ( . VALUE ) => VALUE    [sic! traditional]
;;
;;         (lambda ( . args) ...) => (lambda args ...)
;;
;; whereas
;;
;;       (lambda ( a . rest) ...) => (lambda ( a . rest) ...)
;;
;; Invalid void CDR:  ( VALUE . ) => ERROR: unexpected banana
;;


;;
;; Character Supplements

;; Line Control doesn't count as white space.
(define (newline? character)
  (and (char? character) (control-code? character)
    (case (char->byte character)
      ((#x000A #x000D) #t) ; newlines was missing intentionally???
      (else #f)) ))

;; !!! This still works on a byte level. To cope with multibyte
;; encodings a soft port aware of the encoding might do the job. In
;; addition the following predicates used ...
;; Future port stack?: Counting Port -> Encoding Port -> Byte String Port
;; but it seems better to use the correct accessors. Apart from
;; Viper development we'll need to be able to read in files.
(define (next-char port)
 (let ((char (read-char port)))
   ;(display "->next-char:") (display (represent-char-safely char)) (newline)
  (cond
   ((eof-object? char)    char)
   ((or (printable? char) (white-space? char) (newline? char)) char)
   ((control-code? char)  (next-char port)) ;; drain escape sequences?
   (else #f))))
    ;; unknown char and we could choose some replacement thingy
    ;; from c0??? null perhaps with correct printers and brass or else))))


;; ??? What about peek-char, unread-char


;;
;; Comments and White Space

(define (skip-whitespace port)
 (let ((char (peek-char port)))
  (when (and (not (eof-object? char)) (white-space? char))
    (next-char port) (skip-whitespace port))))

(define (skip-to-next-line port)  ; port-ref and bail on eof
 (let ((char (peek-char port)))
  (when (not (or (eof-object? char) (newline? char)))
    (next-char port) (skip-to-next-line port))))

(define (skip-whitespace-and-comments port)
 (skip-whitespace port)
 (let ((char (peek-char port)))
  (cond ((eof-object? char) char)
        ((newline? char)    (next-char port) (skip-whitespace-and-comments port))
        ((eq? char #\;)     (skip-to-next-line port) (skip-whitespace-and-comments port)) )))


;;
;; Primitive Reader (preliminary)
(sys-load "Analyzer/sexp.scm")

(define read-some (lambda (port)
 (display "->read-some") (display (port-ref port)) (newline)
 (skip-whitespace-and-comments port)
 (let ((char (next-char port)))
   (if (eof-object? char) char (begin ;; XXX Something needs to rethrow this! ???
 (case char
   ;;; (( #\; ) (skip-to-next-line port) (read-some port))
   (( #\( ) (pin port (read-list   port)))
   (( #\" ) (pin port (read-string port)))
   (( #\# ) (pin port (read-hashs  port)))
   (( #\) ) (error "unexpected banana" (port-ref port)))
   (else    (unread-char char port) (read-a-sym port)) ))))))

(define (read-list port)
 (display "->read-list") (newline)
 (skip-whitespace-and-comments port)
 (let ((char (read-char port)))
  (cond ((eq? char #\) ) '())
        ((eq? char #\. ) (let ((val (read-some port))) ;; XXX
                           (skip-whitespace-and-comments port)
                           (or (and (eq? (next-char port) #\) ) val)
                               (error "misplaced dot" (port-ref port)))))
        (else (unread-char char port)
              (cons (read-some port) (read-list port))))))

;; It's /read-pair/ and /lambda/ which make you realize what cheap
;; list processing hack means. The notation and function hack: fixed
;; slopes of practical means to convey some theory? Maybe that implementation
;; of environment even was a simple mistake.
;; (lambda ( . args) ...) is valid and makes this OBSOLETE.
(define (read-pair port)
 (skip-whitespace-and-comments port)
 (let ((char (peek-char port)))
   (cond ((eq? char #\) ) (next-char port) '())
         ((eq? char #\. ) (next-char port) (error "empty pair slot CAR"))
         (else (read-list port)))))

(define (sym-terminator? char)
  (or (memq char '(#\;   #\(  #\)   #\"   #\#))
      (white-space? char)
      (newline?     char)
      (eof-object?  char)))

(define (read-a-sym port)
 (let ((characters '()))
   (let loop ((char (next-char port)))
     ; (display "->read-a-sym") (display (map represent-char-safely characters)) (newline)
      (cond
       ((sym-terminator? char) (interpret-a-sym (reverse characters) char port))
       (else               (set! characters (cons char characters))
                           (loop (next-char port))) )) ))

(define (interpret-a-sym character-list char port)
  (display "->interpret-a-sym:") (write character-list) (newline)
 (unread-char char port)
 (let* ((str (list->string character-list))
        (num (string->number str)))
   (or num
       (and (>= (string-length str) 1) (string->symbol str))
       (error "mischiefed character sequence" (port-ref port)))))

(define (skip-hash-comment port) ; port-ref and bail on eof
 (let ((char (next-char port)))
  (when (not (or (eof-object? char)
                 (and (eq? char #\!) (eq? (peek-char port) #\#)
                      (next-char port)))) ;side effect
    (skip-hash-comment port))))

(define (read-string port) ;; TODO escapes
 (let ((characters '())
       (ref (port-ref port)))
  (let loop ((char (read-char port))) ;; ??? read-char or next-char
    (cond ((eof-object? char) (error "eof in string" ref))
          ((eq? char #\")  (list->string (reverse characters)))
          (else (set! characters (cons char characters))
                (loop (read-char port)))))))

(define (read-hashs port)
 (let ((char (next-char port)))
  (case char
   (( #\" ) (read-byte-string port))
   (( #\( ) (read-vector port))
   (( #\\ ) (read-hchar  port))
   (( #\x #\b #\o #\d #\e #\i ) (read-a-sym port)) ; XXX read a number
   (( #\f ) #f)
   (( #\t ) #t)
   ((#\!)   (skip-hash-comment port) (read-some port))
   (else some-error-here))))

(define (read-vector port)
  ;; XXX read-list may return some else (dotted notation)
  ;; check for proper list and bail + port ref right here
  (list->vector (read-list port)))

;; Hashed Character Notation
;;
;;   Unicode Character Dictionary
;;   (describe-dictionary UD) => ((number-of-definitions . 29151)
;;                                (number-of-nodes    .   163945))
;;
;; In addition we'll need some traditional names and / or we'll need
;; to clean up the Unicode control mess: #\space is called SPACE where
;; our regular #\newline is called LINE FEED (LF) but only in another
;; field of the database entry. Is that sane? That dictionary which
;; we're going to use might be the start of some sort of read table
;; again but why not? What about spaces in character names? The usual
;; convention, I guess. #\space -> #\-
;;
(sys-load "VSI-core/dictionary.scm")
(define sys-character-dictionary (make-dictionary)) ; or UD from encoding.scm
(insert-definition! sys-character-dictionary "newline" (integer->char #x0A))
(insert-definition! sys-character-dictionary "space"   (integer->char #x20))
;;
;; XXX TODO Bad Escape Sequence Handling in VSI-C:
;;  "#\newline"  -> # \n ewline
;;  "#\\newline" -> #\\newline
;; Fix this please! We currently have three escape methods:
;;   - character notation
;;   - string with backslash (VSI-C)
;;   - string with tilde (format)
;;
;; TODO And this one:
;;
;; (read-some (current-input-port)) => ->read-some#(file-ref standard input 5 33)
;; ->skip-whitespace
;; ->skip-whitespace
;; #\x20
;; ->read-a-sym()
;; ->interpret-a-sym
;;  * VSI-C (c_string_to_symbol): mischiefed string for symbol * Aborted (core dumped)
;; mt@nPong:~/Desktop/Analyzer$ x20
;; No command 'x20' found, did you mean: ...
;; 
;;

;; ANS Character Resolution
;;
;; Characters as introduced by #\ or represent themselves
;; (single characters) or correspond to a symbolic name in
;; the character dictionary. (That means: numerical character
;; specifications introduced by #\ aren't accepted anymore.
;; Use the mostly declaritive conversion routines instead!)
;; Any value from 7 bit ASCII (Unicode C0) can be considered
;; a character (including control codes and beware of null)
;; whereas higher code points aren't compatible with our
;; concept of character.
;;
(define (read-hchar port)
 ;; XXX #\; is a character! readasym breaks on that
 (let ((ref (port-ref port))
       (sym (read-a-sym port))) ; XXX read a string
  (when (and (number? sym) (exact? sym) (integer? sym) (<= 0 sym 9))
    (set! sym (string->symbol (number->string sym)))) ;; XXX cludge
  (cond ((symbol? sym)
           (let* ((str (symbol->string sym)))
             (if (= (string-length str) 1)
               (integer->char (char->integer (string-ref str 0))) ;; noop: byte or number plz
               (let ((def (lookup-definition sys-character-dictionary str)))
                 (if (not (eq? def 'undefined))
                     def
                     (error "unknown character: " sym port-ref))) )))
        (else
           (error "unknown character: " sym port-ref)))))


(define (read-byte-string port)
 (let ((answer-box (cons #t '()))
       (ref (port-ref port)))
  (skip-whitespace-and-comments port)
  (let loop ((last answer-box) (char (peek-char port)))
   (cond
    ((eof-object? char) (error "eof in byte string" ref))
    ((eq? char #\")     (next-char port) (list->byte-string (cdr answer-box)))
    (else (let ((iref (port-ref port)) (b   (read-a-sym port))) ;XXX readanum
      (cond
       ((and (number? b) (exact? b) (integer? b) (<= 0 b 255))
           (set-cdr! last (cons b '()))
           (skip-whitespace-and-comments port)
           (loop (cdr last) (peek-char port)))
       ((eof-object? b)
           (error "eof in byte string" ref))
       (else
           (error "mischiefed byte in byte string" iref ref)) ))) )) ))

;;
;; Articulations
(define (readafile file-name)
 (when sys-debug-interpreter
   (display "vsi-nature-load: " (current-error-port))
   (display file-name           (current-error-port))
   (newline))
 (let ((expressions (cons 'box '())))
  ;; Read in all expressions.
  (with-input-from-file file-name (lambda ()
   (let loop ((expression (read-some (current-input-port))) (last expressions))
    (when (not (eof-object? expression))
     (set-cdr! last (cons expression '()))
     (loop (read-some (current-input-port)) (cdr last)) )) ))
  (cdr expressions)))
