;;; Private Prefix: sexp

;;
;; File References
;;

(define (file-ref path line column)
  (vector 'file-ref path line column))

(define (fr-path   fr) (vector-ref fr 1))
(define (fr-line   fr) (vector-ref fr 2))
(define (fr-column fr) (vector-ref fr 3))

(define (file-ref? object)
 (and (vector? object) (eq? 'file-ref (vector-ref fr 0))))

(define (port-ref port) ; deprecation of counting ports?
  (file-ref (port-filename port)
            (+ 1 (port-line port))
            (+ 1 (port-column port)) ))

(define (fr-display fr)
  (display (format "~a (line ~3d col. ~2d)"
                   (fr-path fr) (fr-line fr) (fr-column fr))))

;;
;; Pins
;;

(define (file-referenced file-reference exp)
  (list 'sexpin (cons exp file-reference)))

(define (pin port exp) (file-referenced (port-ref port) exp))

(define (pin? object) (and (pair? object) (eq? (car object) 'sexpin)))

(define (pin-point pin)
  (cdr (second pin)))

(define (pin-exp pin)
  (car (second pin)))


;;
;;
;; Source Expressions (sexp)
;;  (pin aware prototypes)
;;
;; Source expression form acyclic directed graphs.
;;

(define (sexp-car sexp)
 (cond ((pin?  sexp) (sexp-car (pin-exp sexp)))
       ((pair? sexp) (car sexp))
       (else         (error "sexp-car wrong type argument" sexp)) ))

(define (sexp-cdr sexp)
 (cond ((pin?  sexp) (sexp-cdr (pin-exp sexp)))
       ((pair? sexp) (cdr sexp))
       (else         (error "sexp-cdr wrong type argument" sexp))))
(define (sexp-cddr sexp) (sexp-cdr (sexp-cdr sexp)))

(define (sexp-unpin sexp)
  (cond ((pin?  sexp) (pin-exp sexp))
        (else         sexp)))

(define (sexp-ref   sexp)
  (cond ((pin?  sexp) (pin-point sexp))
        (else         #f)))

(define (repin ref exp)
 (if ref
   (file-referenced ref exp)
   exp))

(define (sexp-unpin-vector! v)
 (let ((l (vector-length v)))
  (let loop ((i (- l 1)))
   (vector-set! v i (sexp-unpin (vector-ref v i)))
   (when (> i 0)
     (loop (- i 1)) )) )
 v)
(define (sexp-unpin-vector v) (sexp-unpin-vector! (vector-copy v)))

(define (sexp-unpin-tree:vector! v)
 (let ((l (vector-length v)))
  (let loop ((i (- l 1)))
   (vector-set! v i (sexp-unpin-tree! (vector-ref v i)))
   (when (> i 0)
     (loop (- i 1)) )) )
 v)

(define (sexp-unpin-tree! sexp)
 (cond ((pin?    sexp) (sexp-unpin-tree! (pin-exp sexp)))
       ((pair?   sexp) (set-car! sexp (sexp-unpin-tree! (car sexp)))
                       (set-cdr! sexp (sexp-unpin-tree! (cdr sexp)))
                       sexp)
       ((vector? sexp) (sexp-unpin-tree:vector! sexp))
       (else           sexp)))
        
(define (sexp-unpin-tree sexp)
 (cond ((pin?    sexp) (sexp-unpin-tree (pin-exp sexp)))
       ((pair?   sexp) (cons (sexp-unpin-tree (car sexp))
                             (sexp-unpin-tree (cdr sexp))))
       ((vector? sexp) (sexp-unpin-tree:vector! (vector-copy sexp)))
       (else           sexp)))

;;
;; Rapidd Scheme
(when #f (sys-load "VSI-core/repl.scm") (repl-transcript
)
(REPL)
)
