(define ct (current-time))
(define simulation-random-state (seed->random-state ct))
(display "State: ") (display ct) (newline)

(define (sim-random n)
  (random n simulation-random-state))

;;
;; TRIM trim list
;;

(define segments-end  '())

(define (add-segment)
 (let ((segment (make-vector 24 #f) ))
 (set! segments-end (cons segment segments-end)) ))

(define (saturation segment)
 (/ (let loop ((n 0) (c 0))
     (if (= n 24)
       c
       (loop (+ n 1) (if (vector-ref segment n) (+ c 1) c))))
    24))

(define (zero-saturation? segment)
 (let loop ((i 0))
  (if (= i 24)
    #t
    (if (vector-ref segment i)
      #f
      (loop (+ i 1)) )) ))

(define (free-cell? v i) (not (vector-ref v i)))
(define free-list-head '())
(define free-list-tail '())

(define (drop-free-list)
  (set! free-list-head '())
  (set! free-list-tail '()))

(define (add-free-list-segment! segment-head segment-tail)
 (set-cdr! segment-tail free-list-head)
 (set! free-list-head segment-head)
 (when (null? free-list-tail)
   (set! free-list-tail segment-tail)))

;;
;; TRIM trim list revision 3 (the register machine)
;;
;; Maybe another approach without thinking of computing registers
;; would do the job.
(define (add-freelist-segment segment)
 (let ((data (car segment))
       (free-list-segment-tail  '())
       (free-list-segment-head  '())
       (register-zero-saturation #t)
       (register-n 23))

   (define (find-free-cell segment n)
    (if (< n 0)
      #f 
      (if (free-cell? segment n)
        n
        (begin (set! register-zero-saturation #f)
               (find-free-cell segment (- n 1))) )))

  ;; Scan the segment to initialize the tail of the free list segment.
  (set! register-n (find-free-cell data register-n))
  (when register-n
    (set! free-list-segment-tail (cons #f free-list-segment-head))
    (set! free-list-segment-head free-list-segment-tail)
    (set! register-n (- register-n 1)))

  ;; Scan the segment constructing the free list segment and
  ;; keeping track of the zero saturation state.
  (let loop ((register-n (find-free-cell data register-n)))
   (when (and register-n (>= register-n 0))
    (set! free-list-segment-head (cons #f free-list-segment-head))
    (loop (find-free-cell data (- register-n 1))) ))

  ;; Deliver the result depending on wether we found a segment with
  ;; zero saturation or not.
  (if (not register-zero-saturation)
    (begin (display (length free-list-segment-head)) (newline)
           (when (not (null? free-list-segment-head))
             ; Do you know where snakes get their poison from?
             (add-free-list-segment! free-list-segment-head
                                     free-list-segment-tail))
           #f)
    (begin #t)) ))

(define (free segment) #t)

(define (trim)
 ; (when (null? segments-end) => error condition
 (drop-free-list)
 (let loop ((segment segments-end) (last   '()))
  (if (null? segment)
   'finished-1
   (if (add-freelist-segment segment)
     (begin (when (not (null? last))
              (set-cdr! last (cdr segment)))
            (when (eq? segment segments-end)
              (set! segments-end (cdr segment)))
            (free segment)
            (loop (cdr segment) last))
     (begin (loop (cdr segment) segment) )) )))

;;
;; Load Simulation
;;
(define (random-allocation! segment)
 (let ((target (sim-random 25)))
  (let loop ((allocations 0))
   (when (not (= allocations target))
    (let ((r (sim-random 24)))
     (if (vector-ref segment r)
       (loop allocations)
       (begin (vector-set! segment r #t)
              (loop (+ allocations 1))) ))) )))

(define (bubble n)
 (when (> n 0)
   (add-segment)
   (bubble (- n 1))))

;;
;; Simulation Session
;;
(sys-load "VSI-core/repl.scm")

(repl-transcript
 (define (saturation-map) (reverse (map saturation segments-end)))
 (add-segment)
 segments-end
 ;(random-allocation! (car segments-end)) zero allocation check
 (saturation (car segments-end))
 segments-end


 (bubble 10)
 (saturation-map)

 (length segments-end)
 (trim)
 (length segments-end)
 (length free-list-head)


 (bubble 5)
 (random-allocation! (car segments-end))
 (bubble 2)
 (saturation-map)

 (length segments-end)
 (trim)
 (length segments-end)
 (length free-list-head)

 (begin (trim) (trim) (trim))
 (length segments-end)
 (length free-list-head)
 (saturation-map)
)

(REPL)

;;
;; What we would need for now:
;;
;; DONE next segment pointer
;; DONE next segment getter & setter as well as the
;; DONE segment cells getter
;; 
;; DONE statistics: number of freed segments -> +-number of allocated segments
;; DONE free list reconstruction by segment (add segment free list)
