;;
;; EI: Experimental Idiosyncrisis
;;
;; Highly Experimental Forms from confusing to cheating.
;;

(sys-load "VSI-core/repl.scm")


;;
;; Part 1: Fixed Inlining
;;

;;
;; We would need to track the scope of each form
;; and probably provide some distinction between
;; literal symbols (as in a /case/ form) and 
;; symbols to be resolved as variables. Maybe
;; at some point /go-eval/ will provide this
;; functionality as this needs to considerate
;; almost every known (and unknown) form.
;;  For now this thing is as intelligent as
;; a blade or /once/. Maybe a (get-expressions expression)
;; would do the job to safe us from /let/ variable
;; shadowing, /case/ literals and byte structure
;; descriptions. Perhaps we shouldn't think of a blacklist
;; but a white list? But then we would need to know about
;; every "regular" binding? It's confusing down here
;; climbing on the other side of the cube high above the
;; ground which doesn't exist anyway - that cube.
;; Just like /go-eval/ is still out of reach ...
;;
(define (replace-symbol! expression symbol replacement)
  (define count 0)
  (define (recurse expression)
    (if (not (pair? expression))
      count ;(begin (pcyan count) count)
      (begin
       (if (and (symbol?    (car expression))
                (eq? symbol (car expression)))
         (begin (set-car! expression replacement)
                (set! count (+ 1 count)))
         (when (pair? (car expression))
           (recurse (car expression))))
       (recurse (cdr expression))) ))
  (cut! recurse
        (recurse expression)))

(define (replace-token! expression token replacement)
  (define count 0)
  (define (recurse expression)
    (if (not (pair? expression))
      count ;(begin (pcyan count) count)
      (begin
       (if (and (pair?     (car expression))
                (eq? token (car expression)))
         (begin (set-car! expression replacement)
                (set! count (+ 1 count)))
         (when (pair? (car expression))
           (recurse (car expression))))
       (recurse (cdr expression))) ))
  (cut! recurse
        (recurse expression)))

(define (pcyan . args)
  (tty-color (current-output-port) 'foreground 'cyan)
  (for-each (lambda (a) (display a) (display " ")) args)
  (newline))

(define log-warning pcyan)
(define log-info pcyan)
(define tick 0)

;;
;; Now we do this before memoization on a symbolic level which
;; feels awkward as every occurrence needs to be memoized on its
;; own and we need to make sure with /copy-tree/ that this won't
;; effect other occurrences of this inline. Maybe this is the
;; right thing but the overhead for inlining is bigger than
;; expected for this experimental form.
;;
(define (inline:transformer expression environment)
  ;(pcyan "inline:transformer: " expression)
 (when (not (and (= (length expression) 2)
                 (pair? (cdr expression))
                 (eq? (car (second expression)) 'lambda)))
   (error "mischiefed inline declaration" expression
       (false-if-exception (source-property expression 'filename))
       (false-if-exception (source-property expression 'line))
       (false-if-exception (source-property expression 'column))
          ))
 (let* ((lambda-form (second expression))
        (formals     (second lambda-form))
        (sequence    (cddr   lambda-form))
        (applicable  (eval! (copy-tree lambda-form) environment)))
  (define (inlined-syntax expression environment)
    ;(pcyan "inlined-syntax: " expression)
   (let* ((arguments (cdr expression))
          (replacement-tokens (map (lambda (a)
                                     (cons 'inline-syntax:replacement-token #f))
                                   arguments))
          (sequence  (copy-tree sequence))
          (safe-inline #t))
    (when (not (= (length formals) (length arguments)))
      (error "inliner: wrong number-of-arguments"
       (false-if-exception (source-property expression 'filename))
       (false-if-exception (source-property expression 'line))
       (false-if-exception (source-property expression 'column))
))
    ;; Yet again another "doubled stage": to avoid replacement in
    ;; replacements we first replace the symbols with replacement
    ;; tokens to then replace the replacement tokens with the
    ;; replacements, i.e. argument expressions.
    (for-each (lambda (symbol replacement-token)
                (when (eq? symbol 'inline-syntax:replacement-token)
                  (error "inline-syntax: fatal inlining error"))
                (replace-symbol! sequence symbol replacement-token))
              formals
              replacement-tokens)
    (for-each (lambda (token argument)
               (let ((count (replace-token! sequence token argument)))
                 (when (not (or (= count 1)
                                ;; XXX We are missing the right predicate here.
                                (immediate? argument)
                                (number?    argument)
                                (symbol?    argument)
                                (and (pair? argument)
                                     (eq? (car argument)
                                          'quote)) ))
                    ;; XXX We could even recurse into begin here ...
                   (set! safe-inline #f)
                   )))
              replacement-tokens
              arguments)
    (when (and sys-debug-interpreter (not safe-inline))
     ;; That's the fun of not only detecting such situations but
     ;; reporting them with markers in the expression? Error reporting
     ;; is one fine XXX.
      (log-info
       "inlined-syntax safety: using applicable instead of inlining\n"
       expression
       "\n(from file "
       (false-if-exception (source-property expression 'filename))
       " line #"
       (false-if-exception (source-property expression 'line))
       " column #"
       (false-if-exception (source-property expression 'column))
       ")"
       ))
    (if (not safe-inline)
      (begin (set-car! expression applicable)
             (set-cdr! expression arguments))
      (begin (set-car! expression 'begin)
             (set-cdr! expression sequence)))
    expression)) ; closes inlined-syntax
    ;; XXX twice problem
    ;; (if (not safe-inline)
    ;;   (cons applicable arguments)
    ;;   ;(if (= (length sequence) 1) ; no one proved this theoreme
    ;;   ;    sequence
    ;;   (cons 'begin
    ;;         sequence))) ) ;) ; closes inlined-syntax
  (procedure->memoizing-macro inlined-syntax) ))

(define inline (procedure->syntax inline:transformer))


(when #f
(repl-transcript
 (define aproc (inline (lambda (a b) (+ 1 a b))))
 (define (b c d) (+ (aproc 5 4) (aproc d c)))
 (b 7 8)
 (procedure-source b)

 (define (x) (aproc (begin (display 5) 10) 3))
 (x)
 (procedure-source x)

 (define bproc (inline (lambda (a) (+ a a))))
 (define (y) (bproc (begin (display 5) 3)))
 (y)
 (procedure-source y)

 (define cproc (inline (lambda (a b) (+ 1 a b))))
 (define (b c d) (+ (cproc 5 4) (cproc b c)))
 (not (false-if-exception (b 7 8)))
 (procedure-source b)
)

;(REPL)
) ; closes inlining session

;;
;; The resulting expression after inlining looks a little bit
;; like this:
;;
;; (lambda (c d) (+ (begin (+ 1 5 4)) (begin (+ 1 d c))))
;;
;; In some model /begin/ doesn't have any cost but we would need to
;; check back with the actual implementation: it's at least another
;; clockwork tick but doesn't really need to be.
;;
;; Usually we inline to avoid the overhead of applying a procedure.
;; Simplifying the resulting expression seems to be that simple it's
;; almost tempting. Another nice property of this experimental form
;; is the implicit "error correction": when we replace (aproc d c)
;; with (aproc b c) what will we get? (begin (+ 1 c c)) and no error
;; whatsoever. [That's what happens when you start to replace in the
;; replacements and over again for the next symbol. The "doubled
;; inling" of arguments into the inline and of the inline sequence
;; into the surrounding expression encapsulates a double replacement
;; with replacement tokens and all the bells and whistles where SCA
;; will be defined for original literals only? If you redefine /define/
;; or /case/ that's your own business and SCA won't help you? Maybe
;; to do this (half-way right) you'd need to know about a lot of
;; corners of the underlying implementation. Macros can be as confusing
;; as ... and all it does is teach you how to write your own /eval/.
;;
;; Applications
;;
;; The second tokenizer example became slower after the introduction
;; of inlines: our missing predicate didn't recognize symbols as
;; something that could be replaced safely. While it may not hold
;; for every combination of side effects and global variables we
;; just consider them safe for replacement now and gain about 7%-10%
;; only after applying the inlining technique to the fundamental
;; dictionary accessors. One couldn't be converted to an inline as
;; it was used with /filter/ where it needs to be an applicable.
;; While inlines do contain an applicable in the current experimental
;; setup this isn't yet recognized by the system and their integration
;; with the run time engine would be a manual task. All of the others
;; could be inlined (w/o any additional inline safety message or
;; functionality degration).
;;
;;
;; Why do we /inline/ the definition and not the calls? This is
;; modelled after its C counterpart where we hope to insert the
;; /inline/ keyword in one position to gain performance every-
;; where? While using the /inline/ syntax on the calls clearly
;; better matches the actual /inlining/ action and doesn't
;; impose any problem as long as the procedure source is
;; available the "impact" of evaluation hints and instructions
;; like /once/ and /inline/ shouldn't be underestimated. In
;; fact in some theory as well as for compatibility these
;; syntactical units can be seen as calls to the /identity/
;; function and that theory continues into practice where
;; inlining results in a lot of sequences introduced redundantly
;; with /begin/. In fact in some graphical representation we
;; would perhaps even prefer to add tags to the parenthese
;; instead of having to insert another /node/ into the /source
;; tree/ of expressions but that explicit notation is one
;; advantage up to a certain point even or especially because
;; it already is on the level of an analysis of what would
;; usually have been with the ambigue mathematical notation
;; or by some text needing interpretation.
;;



;;
;; Part 2: Loops like back then (loop without lexical scope)
;;
;; There are circular lists. Your program consists of lists!
;; You want a real loop?

(define (loop:syntax! expressions environment)
 (let loop ((node (cdr expressions)))
  (cond ((and (pair? node) (eq? 'loop (car node)) (null? (cdr node)))
           (set-car! node 'begin)
           (set-cdr! node (cdr expressions)))
        ((pair? node)
           (loop (car node))
           (loop (cdr node)))))
 (set-car! expressions 'begin)
 expressions)

(define loop (procedure->macro loop:syntax!))

(when #t
(repl-transcript
(define (rep n closure)
 (loop
  (set! n (- n 1))
  (closure)
  (when (> n 0) (loop))))

(global-call-delta)
;(rep (* 5 1M) (lambda () (cons 'a 'b)))  ; ~ 7.7 seconds
(repeat (* 5 1M) (lambda () (cons 'a 'b))) ; ~ 8.9 seconds
(global-call-delta)
)

;(REPL)
) ; closes real loop session

;; Real Loop Performance Paragraphs
;
; 7.7 / 8.9 is about 86% => 15% improvement (single "argument" loop)
; pairs: 5 million in our algorithm
;   rep: fuel counter ceval +5M => 10 million pairs (acons) (+15M cell reqs)
;repeat: fuel counter ceval +10M => 20 million pairs but +30M cell reqs
;        => 5M unexplained pairs: one cell req per call (closure application)
;           is missing in the fuel counters.
;
; Prospective: multi "argument" loops w/o scope might "suffer" from 
; iloc lookup performance whereas the implicit loop apply of a named let
; is as fast as a map when it comes to argument evaluation?
;
; It's probably better to go the whole way and start optimizing with
; the data graphs of the analyzer (to then continue to a "flattend"
; list of load and store instructions to then optimize again to then
; translate into machine instructions?) But we'll need good models or
; better specifications of environments, argument lists and "inner"
; calling conventions where an iloc lookup is an indirection which
; we might hit again with references onto the call stack by memory
; address arithmetic? ... instead of creating an optimized unchecked
; evaluator where /set-iloc/ would provoke a segmentation fault for a
; wrong type argument but only in the best case.
; 
;;
