;; This file is in a transitional state
;;  system routines: lib uc.scm uc-db.scm
;;  examples remain here
;;  TODO Purge staging efforts.
;;
;; SCA: General Encoding Handling Example
;;
;; Former Brass PROTOTYPE generation / Unicode Workbench
;;

(sys-load "VSI-core/repl.scm")
(sys-load "VSI-core/byte-structures.scm") ; casts

(sys-load "lib/uc-db.scm")
(sys-load "lib/uc.scm")

(define (encoding-error . args) (error "mischiefed encoding" args))
;(define encoding-error REPL-HERE)
;(define encoding-error (macro ( . exp) '(REPL-HERE)))

;;
;; UTF-8
;; 3.9 Unicode Encoding Forms (p. 125) [Please Review Surrogates]
;;
;; Some reflections:
;;  - UTF-8 probably allows for double and triple representation 
;;    for many scalars greater than 127 (and smaller than the
;;    uuu uu range) (even 65 can be represented with a two byte
;:    sequence).
;;  - the sequence of operations can of course be implemented
;;    more efficiently than in this reference implementation
;;    but there should remain a sufficiently complicated set
;;    of operations for systems w/o byte operations which can
;;    be disguised by explicit and implicit casts in C but not
;;    (yet) in SCA
;;  - that wasn't even one full day to implement this theoretical
;;    UTF-8 support thanks to the standard documents and SCA
;;
;; Aftermath:
;;  - That one day approach had errors in bit shift offsets which
;;    only showed up for higher code points; the stream decoder
;;    needed to be written from scratch if not for the bit masks.
;;  - The C1 table and other sources used didn't specify any best
;;    practices regarding specifications of 8-bit ASCII: CSI (128+ESC)
;;    is defined as an 8 bit constellation which looks like a sequence
;;    byte in UTF-8 whereas the C1 table ignores the bits and assigns
;;    it some arbitrary code point matching that numerical value of
;;    that 8-bit joke from back then.
;;
;; Encoding Errors (http://www.unicode.org/faq/utf_bom.html)
;; [This isn't specific to UTF-8 but here for convenience.]
;;
;; Q: Are there any byte sequences that are not generated by a UTF?
;; How should I interpret them?
;;
;; A: None of the UTFs can generate every arbitrary byte sequence. For
;; example, in UTF-8 every byte of the form 110xxxxx2 must be followed
;; with a byte of the form 10xxxxxx2. A sequence such as <110xxxxx2
;; 0xxxxxxx2> is illegal, and must never be generated. When faced with
;; this illegal byte sequence while transforming or interpreting, a
;; UTF-8 conformant process must treat the first byte 110xxxxx2 as an
;; illegal termination error: for example, either signaling an error,
;; filtering the byte out, or representing the byte with a marker such
;; as FFFD (REPLACEMENT CHARACTER). In the latter two cases, it will
;; continue processing at the second byte 0xxxxxxx2.
;;
;; A conformant process must not interpret illegal or ill-formed byte
;; sequences as characters, however, it may take error recovery
;; actions. No conformant process may use irregular byte sequences to
;; encode out-of-band information.

;;
;; Decoding UTF-8

(define utf-8:two-byte-mask   (byte->byte-string #b11100000))
(define utf-8:two-byte-tag    (byte->byte-string #b11000000))

(define utf-8:three-byte-mask (byte->byte-string #b11110000))
(define utf-8:three-byte-tag  (byte->byte-string #b11100000))

(define utf-8:four-byte-mask  (byte->byte-string #b11111000))
(define utf-8:four-byte-tag   (byte->byte-string #b11110000))

(define utf-8:sequence-byte-mask (byte->byte-string #b11000000))
(define utf-8:sequence-byte-tag  (byte->byte-string #b10000000))

(define (utf-8:sequence-byte? byte-string)
  (equal? (bit-and utf-8:sequence-byte-mask byte-string)
          utf-8:sequence-byte-tag))

(define (utf-8:two-byte? byte-string)
  (equal? (bit-and utf-8:two-byte-mask byte-string)
          utf-8:two-byte-tag))
(define (utf-8:three-byte? byte-string)
  (equal? (bit-and utf-8:three-byte-mask byte-string)
          utf-8:three-byte-tag))
(define (utf-8:four-byte? byte-string)
  (equal? (bit-and utf-8:four-byte-mask byte-string)
          utf-8:four-byte-tag))

(define (utf-8:sequence-bits byte-string)
 (cond
  ((utf-8:sequence-byte? byte-string)
        (bit-and (bit-not utf-8:sequence-byte-mask) byte-string))
  (else (encoding-error 'decoding-utf-8)) ))

(define from-msb! from-network-byte-order!)
(define to-msb    network-byte-order!)

(define (utf-8->usv byte-string)
 (let* ((sel  (byte-string-select byte-string 0 1))
        (byte (byte-string->number sel))
        (usv-1 (byte-string-clear! (make-byte-string 4)))
        (usv-2 (byte-string-clear! (make-byte-string 4)))
        (usv-3 (byte-string-clear! (make-byte-string 4)))
        (usv-4 (byte-string-clear! (make-byte-string 4))))
 (if (< byte 128)
     byte
 (begin ; that's one long else
 (cond

  ;; Somehow I didn't really get the point about that encoding and
  ;; that's not even about ...
  ((utf-8:two-byte? sel)
     (byte-string-set! usv-2 2
       (bit-and (bit-not utf-8:two-byte-mask) sel))
     (byte-string-set! usv-1 3
       (utf-8:sequence-bits (byte-string-select byte-string 1 1))))

  ((utf-8:three-byte? sel)
     (byte-string-set! usv-3 1
       (bit-and (bit-not utf-8:three-byte-mask) sel))
     (byte-string-set! usv-2 2
       (utf-8:sequence-bits (byte-string-select byte-string 1 1)))
     (byte-string-set! usv-1 3
       (utf-8:sequence-bits (byte-string-select byte-string 2 1))))

  ((utf-8:four-byte?  sel)
     (byte-string-set! usv-4 0
       (bit-and (bit-not utf-8:four-byte-mask) sel))
     (byte-string-set! usv-3 1
       (utf-8:sequence-bits (byte-string-select byte-string 1 1)))
     (byte-string-set! usv-2 2
       (utf-8:sequence-bits (byte-string-select byte-string 2 1)))
     (byte-string-set! usv-1 3
       (utf-8:sequence-bits (byte-string-select byte-string 3 1))))

  (else
   (encoding-error 'decoding-utf-8)))
  
 (map from-msb! (list usv-1 usv-2 usv-3 usv-4))

 ;; (begin ; debugging paragraph
 ;;   (for-each (lambda (w) (newline) (byte-bit-printer w))
 ;;             (list usv-1 usv-2 usv-3 usv-4))
 ;;   (newline)
 ;; )

 (byte-string->number
   (bit-or usv-1 (bit-shift-right usv-2 2)
                 (bit-shift-right usv-3 4)
                 (bit-shift-right usv-4 6)))
 ))))

;; simple stream decoding
(define (read-byte/throwing port)
  (let ((byte (read-byte port)))
    (if (eof-object? byte)
      (throw 'eof)
      (byte->byte-string byte))))
(define (read-char:utf-8 port)
 (let* ((read-byte read-byte/throwing)
        (sel  (read-byte port))
        (byte (byte-string->number sel))
        (usv-1 (byte-string-clear! (make-byte-string 4)))
        (usv-2 (byte-string-clear! (make-byte-string 4)))
        (usv-3 (byte-string-clear! (make-byte-string 4)))
        (usv-4 (byte-string-clear! (make-byte-string 4))))
 (if (< byte 128)
     byte
 (begin ; that's one long else
 (cond

  ;; Somehow I didn't really get the point about that encoding and
  ;; that's not even about ...
  ((utf-8:two-byte? sel)
     (byte-string-set! usv-2 2
       (bit-and (bit-not utf-8:two-byte-mask) sel))
     (byte-string-set! usv-1 3
       (utf-8:sequence-bits (read-byte port))))

  ((utf-8:three-byte? sel)
     (byte-string-set! usv-3 1
       (bit-and (bit-not utf-8:three-byte-mask) sel))
     (byte-string-set! usv-2 2
       (utf-8:sequence-bits (read-byte port)))
     (byte-string-set! usv-1 3
       (utf-8:sequence-bits (read-byte port))))

  ((utf-8:four-byte?  sel)
     (byte-string-set! usv-4 0
       (bit-and (bit-not utf-8:four-byte-mask) sel))
     (byte-string-set! usv-3 1
       (utf-8:sequence-bits (read-byte port)))
     (byte-string-set! usv-2 2
       (utf-8:sequence-bits (read-byte port)))
     (byte-string-set! usv-1 3
       (utf-8:sequence-bits (read-byte port))))

  (else
   (encoding-error 'decoding-utf-8)))
  
 (map from-msb! (list usv-1 usv-2 usv-3 usv-4))

 ;; (begin ; debugging paragraph
 ;;   (for-each (lambda (w) (newline) (byte-bit-printer w))
 ;;             (list usv-1 usv-2 usv-3 usv-4))
 ;;   (newline)
 ;; )

 (byte-string->number
   (bit-or usv-1 (bit-shift-right usv-2 2)
                 (bit-shift-right usv-3 4)
                 (bit-shift-right usv-4 6)))
 ))))



;;; Stream Reader
(define utf-8:stream-namespace
 (namespace

(define sys-read-byte read-byte)
(define (utf-8:sequence-bits byte-string) ; w/o validation; we do that here
 (bit-and (bit-not utf-8:sequence-byte-mask) byte-string))

;; Prerequisites
(define (read-byte port) ;; XXX use (read-byte-string p 1) TODO rename read-byts
 (let ((b (sys-read-byte port)))
  (cond ((eof-object? b) b)
        (else            (byte->byte-string  b)))))
(define (unread-byte b) (unread-char (integer->char b))) ;; XXX
(define null-4          #"0 0 0 0")

;; Error Handling
(define uc:error-code #xFFFD) ; replacement character
(define (bad-sequence   port sel)
  (unread-byte (byte-string->number sel))
  uc:error-code)
(define (bad-terminator port sel) uc:error-code)
; We probably should replace each byte read with one error code each
; when an underflow occurs. TODO Check back with the standards.
; BUT: the caller expects to receive one code point or EOF. Do we
; need to return /n/ replacement characters before trying to read
; the stream again? Should we unread those bytes? But EOF had
; already been signalled which means we're finished in this time
; slice?
(define (underflow  n   port)     uc:error-code)
    
;; Multibyte Decoding Stages (outsoured inlines)
(define (cp-from-bytes! usv-1 usv-2 usv-3 usv-4)
  ;; The following byte order conversion isn't needed when
  ;; you read /byte-string-set! usv n byte/ as a left shift
  ;; of the byte bits in a four byte word. Thus the final
  ;; right shifts below could be taken into account and
  ;; both left and right shifts could be replaced with the
  ;; resulting operations which would be independent of the
  ;; endianess. But we would need to check back with the
  ;; standards and test again.
  (map from-msb! (list usv-1 usv-2 usv-3 usv-4))
  (byte-string->number
   (bit-or usv-1 (bit-shift-right usv-2 2)
                 (bit-shift-right usv-3 4)
                 (bit-shift-right usv-4 6))))

(define (two-byte port sel usv-1 usv-2)
  (byte-string-set! usv-2 2 (bit-and (bit-not utf-8:two-byte-mask) sel))
  (set! sel (read-byte port))
  (cond
   ((eof-object? sel)
       (underflow 2 port))
   ((utf-8:sequence-byte? sel)
       (byte-string-set! usv-1 3 (utf-8:sequence-bits sel))
       (cp-from-bytes! usv-1 usv-2 null-4 null-4))
  (else
       (bad-sequence port sel) )))

(define (three-byte port sel usv-1 usv-2 usv-3)
  (byte-string-set! usv-3 1 (bit-and (bit-not utf-8:three-byte-mask) sel))
  (set! sel (read-byte port))
  (cond
   ((eof-object? sel)
       (underflow 2 port))
   ((utf-8:sequence-byte? sel)
       (byte-string-set! usv-2 2 (utf-8:sequence-bits sel))
       (set! sel (read-byte port))
       (cond
        ((eof-object? sel)
             (underflow 3 port))
        ((utf-8:sequence-byte? sel)
             (byte-string-set! usv-1 3 (utf-8:sequence-bits sel))
             (cp-from-bytes! usv-1 usv-2 usv-3 null-4))
        (else
             (bad-sequence port sel) )))
  (else
       (bad-sequence port sel) )))

(define (four-byte port sel usv-1 usv-2 usv-3 usv-4)
  (byte-string-set! usv-4 0 (bit-and (bit-not utf-8:four-byte-mask) sel))
  (set! sel (read-byte port))
  (cond
   ((eof-object? sel)
       (underflow 2 port))
   ((utf-8:sequence-byte? sel)
       (byte-string-set! usv-3 1 (utf-8:sequence-bits sel))
       (set! sel (read-byte port))
       (cond
        ((eof-object? sel)
             (underflow 3 port))
        ((utf-8:sequence-byte? sel)
             (byte-string-set! usv-2 2 (utf-8:sequence-bits sel))
             (set! sel (read-byte port))
             (cond
              ((eof-object? sel)
                  (underflow 4 port))
              ((utf-8:sequence-byte? sel)
                  (byte-string-set! usv-1 3 (utf-8:sequence-bits sel))
                  (cp-from-bytes! usv-1 usv-2 usv-3 usv-4))
              (else
                  (bad-sequence port sel) )))
        (else
             (bad-sequence port sel) )))
  (else
       (bad-sequence port sel) )))

;; Stream Decoding Dispatch
(define (read-char:utf-8/stream port)
 (let ((usv-1 (byte-string-clear! (make-byte-string 4)))
       (usv-2 (byte-string-clear! (make-byte-string 4)))
       (usv-3 (byte-string-clear! (make-byte-string 4)))
       (usv-4 (byte-string-clear! (make-byte-string 4)))
       (sel   (read-byte port)))

 (cond
  ((eof-object? sel) sel)
  ((utf-8:four-byte?  sel) (four-byte  port sel usv-1 usv-2 usv-3 usv-4))
  ((utf-8:three-byte? sel) (three-byte port sel usv-1 usv-2 usv-3))
  ((utf-8:two-byte?   sel) (two-byte   port sel usv-1 usv-2))
  ((< (byte-string->number sel) 128) (byte-string->number sel))
  (else                    (bad-terminator port sel)))
 ))
)) ;; Closes UTF-8 Stream Reader Namespace
(define utf-8:read-char/error-code
  (from-namespace utf-8:stream-namespace 'read-char:utf-8/stream))


;; (define (read-codepoint)
;;  (catch 'eof (lambda ()
;;                (read-char:utf-8 (current-input-port)))
;;         (lambda (key . args) 'eof)))
(define (read-codepoint)
  (utf-8:read-char/error-code (current-input-port)))
;; dirty brass
(define (uc:read-file path)
 (let ((s (cons #t '() )))
  (with-input-from-file path (lambda ()
   (let loop ((tail s) (cp (read-codepoint)))
    (cond ((eof-object? cp ) (list->brass (map integer->char (cdr s))))
          (else
             (set-cdr! tail (cons cp '()))
             (loop (cdr tail) (read-codepoint)) )) ))) ))
  
;; end of simple stream decoding

;; (define (brass:display brass)
;;   (for-each write-bytes
;;             (map char->byte-string/utf-8 (brass->list brass))))


;(define (list->brass list) (list->vector list))
;(define (brass->list bcv)  (vector->list bcv))

;; (define (brass->byte-string cp-vector)
;;   ;; XXX returns a list of byte strings one code point each
;;   (map char->byte-string/utf-8 (brass->list cp-vector)))

;; (define (byte-string->brass bytes)
;;   (list->brass (map integer->char (map utf-8->usv bytes))))

;; end of dirty brass

;;
;; Encoding UTF-8

;; Offsets of the masks from left.
(define utf-8:usv-bits-offset-first  0)
(define utf-8:usv-bits-offset-second 6)
(define utf-8:usv-bits-offset-third  (+ 6 utf-8:usv-bits-offset-second))
(define utf-8:usv-bits-offset-fourth (+ 4 utf-8:usv-bits-offset-third))

(define utf-8:usv-bits-mask-first  ; x
  (number->byte-string-4 #b00111111))
(define utf-8:usv-bits-mask-second ; y
  (bit-shift-left! (number->byte-string-4 #b00111111)
                   utf-8:usv-bits-offset-second))
(define utf-8:usv-bits-mask-third  ; z
  (bit-shift-left! (number->byte-string-4 #b00001111)
                   utf-8:usv-bits-offset-third))
(define utf-8:usv-bits-mask-fourth ; u
  (bit-shift-left! (number->byte-string-4 #b00011111)
                   utf-8:usv-bits-offset-fourth))

(define (usv->utf-8 u-scalar-v)
 (let* ((usv   (number->byte-string-4 u-scalar-v))
        (usv-1 (bit-and usv utf-8:usv-bits-mask-first))
        (usv-2 (bit-and usv utf-8:usv-bits-mask-second))
        (usv-3 (bit-and usv utf-8:usv-bits-mask-third))
        (usv-4 (bit-and usv utf-8:usv-bits-mask-fourth)))
   ;; Byte Number Selection based on Common Sense (not on any read)
  (cond ((<= 0 u-scalar-v 127)
           (byte->byte-string u-scalar-v))

        ((<= (once(expt 2 16)) u-scalar-v (once(- (expt 2 21) 1)))
           (byte-string-adjoin
            (bit-or utf-8:four-byte-tag
             (cast:4-1
              (bit-shift-right usv-4 (+ 2 utf-8:usv-bits-offset-fourth))))

            (bit-or utf-8:sequence-byte-tag
             (cast:4-1
              (bit-or (bit-shift-right usv-3 utf-8:usv-bits-offset-third)
                      (bit-and (number->byte-string-4 48) ; #b0011.0000
                       (bit-shift-right usv-4
                                       (- utf-8:usv-bits-offset-fourth 4))))))

            (bit-or utf-8:sequence-byte-tag
             (cast:4-1
              (bit-shift-right usv-2 utf-8:usv-bits-offset-second)))

            (bit-or utf-8:sequence-byte-tag
             (cast:4-1 usv-1)) ))

        ((<= (once(expt 2 11)) u-scalar-v (once(- (expt 2 16) 1)))
           (byte-string-adjoin
            (bit-or utf-8:three-byte-tag
             (cast:4-1
              (bit-shift-right usv-3 utf-8:usv-bits-offset-third)))

            (bit-or utf-8:sequence-byte-tag
             (cast:4-1
              (bit-shift-right usv-2 utf-8:usv-bits-offset-second)))

            (bit-or utf-8:sequence-byte-tag
             (cast:4-1 usv-1)) ))

        ((<= 128 u-scalar-v (once(- (expt 2 11) 1)))
           (byte-string-adjoin
            (bit-or utf-8:two-byte-tag
              (cast:4-1 (bit-shift-right usv-2 utf-8:usv-bits-offset-second)))

            (bit-or utf-8:sequence-byte-tag
              (cast:4-1 usv-1))))

        (else
           (encoding-error (list 'encoding-utf8 usv))) )))



;;
;; Example Sessions
;;
(sys-load "VSI-core/posix.constants.scm")
;(define pc/read-only (number->byte-string O_RDONLY (size-of-int)))
(define pc/mode (bit-and S_IRUSR S_IWUSR S_IRGRP S_IWGRP))

(repl-transcript
 ;(define f (open "/opt/VSI/Examples/utf-8.txt" O_RDONLY pc/mode))
 ; ^ TODO Please review binary mode and /setlocale/ wrt posix /read/
 (number->string 163 16)
 (define f (open-file "/opt/VSI/Examples/utf-8.txt" "r"))
 (define b (read-bytes 10 f))
 ;; (define b #f)
 ;; (with-input-from-file "/opt/VSI/Examples/utf-8.txt"
 ;;   ; (set! b (read-bytes 5 (port->fdes (current-input-port)))))
 ;;   (set! b (read-bytes 5)))
 b
 (utf-8->usv b)
 (usv->utf-8 (utf-8->usv b))

 (define uc:error-code #xFFFD) ; replacement character
 (define uc:mahjong-tile:joker #x1f02a)
 (define uc:playing-card:white-joker #x1f0df)
 (define s uc:playing-card:white-joker)
 (number->string  s)
 (usv->utf-8      s)
 (write-bytes (usv->utf-8 s))

 (define uc-data (uc-db-init))
 (vector-length uc-data)
 ;; (let loop ((n 0))
 ;;  (if (not (= n (vector-ref (vector-ref uc-data n) 0)))
 ;;    (begin (display n) (newline))
 ;;    (when (< n (vector-length uc-data))
 ;;      (loop (+ n 1)))))
 (uc-lookup-codepoint uc-data s)
 (uc-lookup-codepoint uc-data (utf-8->usv b))
 (uc-lookup-codepoint uc-data uc:error-code)
) ; Closes regular REPL transcript

;;
;; Extended REPL transcript XXX this file is becoming too large
;; Unicode Character Data as a Dictionary Data Type
(when #f
(repl-transcript
(sys-load "/opt/VSI/VSI-core/dictionary.scm")
(if (access? "/opt/VSI/VSI-core/uc.dictionary.scm" R_OK)
 (define UD (with-input-from-file "/opt/VSI/VSI-core/uc.dictionary.scm" read))
 (begin
(define UD (make-dictionary))
(let loop ((i (- (vector-length uc-data) 1)))
 (when (>= i 0)
  (let* ((u-entry (vector-ref uc-data i))
         (name  (vector-ref u-entry 1))
         (idx   (vector-ref u-entry 0))
         (d-entry (create-entry! (get-root UD) (string->list name) )))
    (if (definition? d-entry)
      (set-definition! d-entry (cons idx (get-definition d-entry)))
      (set-definition! d-entry (list idx))))
  (loop (- i 1))))
(describe-dictionary UD)
(with-output-to-file "/opt/VSI/VSI-core/uc.dictionary.scm" (lambda () (write UD)))
)) ; closes dictionary on demand

(define A (lookup-prefix UD "LATIN CAPITAL LETTER A"))
(define (write-line arg) (write arg) (newline))
(for-each write-line A)
(length A)
(define b (lookup-prefix UD "LATIN SMALL"))

(lookup-prefix UD "<control>")
;; Field idx#10 sometimes contains a ... (see UnicodeData File Format)
;;
;; "This is the old name as published in Unicode 1.0. This name is only
;; provided when it is significantly different from the current name
;; for the character. The value of field 10 for control characters
;; does not always match the Unicode 1.0 names. Instead, field 10
;; contains ISO 6429 names for control functions, for printing in the
;; code charts."
;;
;; Playing with relational data bases and fostering a useful encoding
;; are two distinct things, I guess. But then everything is table,
;; record and cubic, isn't it. Except for these bananas: ))()(()
;; Let's do the banana split again.

; XXX please change /scan-dictionary?/ to deliver the (list of) keys
(define c (filter (lambda (e) (pair? (cddr e))) ; short for length > 1
                  (dictionary->alist UD)))
(length c)
c
(define d (lookup-prefix UD "DIGIT"))
(write-bytes (usv->utf-8 ; Yes, they really dare call this encoding.
                 (car (lookup-definition UD "DIGIT ONE COMMA"))))
)) ; end of extended session

;;
;; Case Mapping Session / Review
;;
;; The Viper doesn't need case mappings and a case insensitive
;; "symbol scheme" doesn't work out for cases where symbols don't
;; correspond to linguistic but mathematical symbols. The distinction
;; cannot be determined by algorithms.
;; (define A (some-set))
;; (Define a 5)
;; (Some Set ~> UnSpeci-fied
;;
(when #f
(repl-transcript
 (define ucm (uc:read-case "CaseFolding.txt"))
 (length ucm)
 (define nan (filter (lambda (e) (not (number? (vector-ref e 0)))) ucm))
 (define seq (filter (lambda (e) (not (number? (vector-ref e 2)))) ucm))
 (uc-case-mapping ucm 65)
 (define a (utf-8->usv (list->byte-string ; XXX
                        (map char->integer (string->list "à")))))
 a
 (define A (uc-case-mapping ucm a))
 A
 (write-bytes (usv->utf-8 (first A)))
 (define T (uc-case-mapping ucm 73))
 T
 (write-bytes (usv->utf-8 73))
 (write-bytes (usv->utf-8 (first T)))
 (write-bytes (usv->utf-8 (second T)))
 (uc-lookup-codepoint uc-data (second T))
))


;;
;; Brass
;;

(define (brass-from-db)
 (list->brass (map (lambda (e) (integer->char (first (vector->list e))))
                   (vector->list uc-data))))

(define i 0)
(define (u-all-cp)
(with-output-to-file "U.txt" (lambda ()
 (for-each (lambda (cp)
             (set! i (+ i 1))
             (write-bytes (char->byte-string/utf-8 cp)))
             ;(when (= i 75) (newline) (set! i 0)))
           (brass->list (brass-from-db))) )))
(define (u-all-read)
 (let ((answer (cons #t '()))
       (file   (open-file "U.txt" "r")))
  (let loop ((last answer) (cp (utf-8:read-char/error-code file)))
   (cond ((eof-object? cp)
              (close file)
              (set! answer (cdr answer))
              (list->brass (map integer->char answer)))
         (else
              (set-cdr! last (cons cp '()))
              (loop (cdr last) (utf-8:read-char/error-code file)))))))

(newline)
(newline) (display "Dirty Brass") (newline)
(newline)
(repl-transcript

;; Self test enclosed
(equal? (brass->list (byte-string->brass (brass->byte-string (brass-from-db))))
        (brass->list (brass-from-db)))

;; Self test with file
(begin (u-all-cp) (equal? (brass->list (u-all-read))
                          (brass->list (brass-from-db))))

;; Visual Check
; 
; (Gedit matches Konsole and Emacs in X11)
;
; This failed on the Linux console: quite different special chars here.
;                                   - vi here just the same
;  see also /unicode_start/ (1) - you might think it's just the font
; that's off but the number of displayed characters seems to be much
; longer than what is in the text file. They do look like eight bit
; characters of some old code page with fancy frame drawing glyphs.
;
; A fresh tty on the console does display the file as well as
; key presses correctly.
; Fix this, please. Nothing to do: this is independent of VSI and seems
; to be a fixed setting of tty1 where tty2 (and presumably all others)
; do represent the characters correctly. As a reminder: if and even if
; by ten years there will be a release of the Viper Editor it will bear
; the version number C0 as it won't support anything outside of C0 if
; not to filter it out.
; Please note: the binary file / error recovery test further down
; produces visually quite different results on tty2 from what you can
; see in Konsole (brass:display f). Every output byte sequence should
; correspond to a valid encoding of a USV. Is it the replacement
; character or other non code point USVs which make some of the
; readable text of the binary file disappear? Strange error recovery
; there or encoding error here? Maybe C0 controls but outputting
; random bytes onto a terminal type writer isn't sth to rely on
; anyway.
;
(define br (uc:read-file "/opt/VSI/Examples/utf-8.txt"))
(display br)

;; Encoding Errors
; Binary test: a proper code point check (proper-brass? br . uc-version)
; on our side would be up to 30.000 x n database "lookups" where /n/ denotes
; the number of USVs in the brass. But which version of that UC stuff
; then? Some range checks perhaps? [After the last recompilation the
; binary as a brass even makes to change the colors of the terminal. Have
; fun! BTW That day no Linux console tty was able to represent any unicode!
; Maybe tomorrow again? Maybe some kind of fallback mode of some internal
; error occurred where then because of missing tty flag effects everything
; goes wrong? Or right? Right? Have fun! That security-by-i-dont-need-it-approach
; then really meant: i dont need it.]
(define f (uc:read-file "/opt/VSI/interpreter"))
 ; Ratio of replacement characters to other "misinterpretations":
 ; that's about one fourth on a 64 bit system. Like all values
 ; in the range of C0 null bytes are recognized as valid encodings
 ; of the first code point. Then there is also some text in the file.
(exact->inexact
 (/ (length (filter (lambda (e) (= uc:error-code (char->integer e)))
                    (brass->list f)))
    (brass-length f)))

)
(define (release) ; Couldn't a namespace offer this automatically?
  (set! uc-data #f)
  (set! UD      #f)
  (set! A       #f)
  (set! f       #f)
  (set! br      #f)
  ;(set! T       #f)
  (set! A       #f)
  ;(set! a       #f)
  (set! c       #f)
  (set! d       #f)
  (set! b       #f)
  ; missing: case symbols
  )
(REPL)
