(sys-load "VSI-core/namespace.scm")
(sys-load "VSI-core/dictionary.scm")

;; XXX We really should offer this as part of VSI core.
(define canonicalize-define namespace:canonicalize-define) 

(define index:debug #f)
(define (index:location file line-number) (cons file (+ 1 line-number)))

(define (add-location! index symbol location)
 (let ((def (lookup-definition index (symbol->string symbol))))
  (if (pair? def)
    ; D-Rex doubles (insert-definition! index (symbol->string symbol) (cons location def))
    (insert-definition! index (symbol->string symbol) (list location))
    (insert-definition! index (symbol->string symbol) (list location)) )))

(define (index-file! index file-name)
 (when index:debug (display "index-file!: ") (display file-name) (newline))
 (with-input-from-file file-name (lambda ()
  (let loop ((exp (read)))
   (if (eof-object? exp)
     (cut! index index)
     (begin
       (and (pair? exp) (symbol? (car exp))
            (case (car exp)
              ; XXX what about nested load in a catch and cose simili
              ((define %define-global)
                 (add-location! index (second (canonicalize-define exp))
                                (index:location
                                  file-name
                                  (port-line (current-input-port)))) )
              ((load)
                 (index-file! index (second exp)))
              ((sys-load primitive-load-path)
                 (index-file! index (%search-load-path (second exp))) )))
       (loop (read))) )) )))

(define (index->table:csv index)
 (let ((l (sort (dictionary->alist index)
            (lambda (a b) (string<? (car a) (car b)) )) ))
  (for-each (lambda (k)
             (let ((symbol (car k)))
               ; (display k) (newline) (force-output)
              (for-each (lambda (e)
                         (display symbol)
                         (display "; ")
                         (display (car e))
                         (display "; ")
                         (display (cdr e))
                         (newline))
                        (cdr k)) ))
            l)))

;;
;; There is now some first shortcoming of core dependencies:
;; to make a snapshot of the state of the system dictionary we would
;; like to use /dictionary->alist/ but to have that we would need to
;; load ihub.scm first. For now the double entries are better than any
;; solution I can think of.
(define (system-dictionary->index)
 (let ((system-dictionary (cons 'dictionary (system-dictionary-root)))
       (index (make-dictionary)))
  (for-each (lambda (e)
              (insert-definition! index (car e) (list (cdr e))))
    (map (lambda (e)
           (cons (car e) (index:location "D-Rex" -1)))
         (filter (lambda (e) ; (display e) (newline) (force-output)
                         (and (pair? (cdr e))
                              (variable? (cddr e))))
           (dictionary->alist system-dictionary)) ))
  index))

;;
;; Application 1: Interned POSIX Signal Constants
(define (signal-alist idx)
 (let ((sigs (lookup-prefix idx "SIG")))
  (map (lambda (s) (cons (string->symbol
                          (list->string
                           (map char-downcase
                            (string->list
                             (substring s 3 (string-length s)) )) ))
                         (eval (string->symbol s) '() )))
       (map car sigs))))

(sys-load "VSI-core/repl.scm")
(define (void side-effects) #t)
(repl-transcript
(define I (make-dictionary))
(void (index-file! I "Examples/index.scm"))
(with-output-to-file "Examples/index.csv" (lambda ()
 (index->table:csv I)))

(define core-index (system-dictionary->index))
(with-output-to-file "Core-Index.csv" (lambda ()
 (index->table:csv (index-file! core-index
                                "VSI-core/ihub.scm"))))

(system "less Examples/index.csv")
;(signal-alist core-index)
)
(REPL)
