;; 
;; inotify example
;;
(sys-load "VSI-core/posix.constants.scm")
(sys-load "VSI-core/byte-structures.scm")

(define control:read-bytes read-bytes)

(define NAME_MAX 255) ; XXX see cee.scm and limits.h
(define iNAME_MAX 256) 
(define inotify-event-header
  (byte-structure-description
   (wd   int)
   (mask   4) ; [sic! that was an uint32]
   (cookie 4)
   (len    4)
   ;(name   pointer) ; That's not a pointer but an array: NAME_MAX + 1
   ;(name   (+ 1 NAME_MAX)) ; XXX fails to evaluate <- that's only in
                            ; the next protocol; right now only symbols
                            ; are evaluated
   ;(name    iNAME_MAX) ; Now it's guaranteed to be big enough or
   ; they fail for us as our buffer wouldn't have ... that's datagram
   ; style, right? But we don't read in enough data to use the accessor:
   ; or we opt to include variable size arrays in byte structures or
   ; we have to allocate a template byte-string.
   ))
(define-bs-accessors 'inotify inotify-event-header)
;; The nice thing about the current computed accessors is the /getters/
;; don't check the length of their argument to match the length of the
;; described structure. With all those protocols this is important but
;; the /setters/ do check and that might not be everyone's favourite
;; flavor.
(define inotify-tail-length (+ NAME_MAX 1))
(define (inotify-name bytes)
 (c-string->deprecated-string
  (byte-string-select bytes
                      (bs-length inotify-event-header)
                      (byte-string->number (inotify:len bytes)))))

(define (inotify-read-event fd)
 (let* ((expectation (+ (bs-length inotify-event-header)
                        inotify-tail-length))
        (l (select (list fd) '() '())))
   ;(display l)
  (let ((bytes (control:read-bytes expectation fd)))
   (cond
    ((eof-object? bytes) #f)
    ((not (byte-string? bytes)) (error "read error"))
    (else
      bytes)) )))

(sys-load "VSI-core/repl.scm")
(repl-transcript
 inotify-event-header
 (byte-bit-printer IN_ALL_EVENTS)
 (define ifd (unchecked-inotify-init))
 ifd
 (define path (deprecated-string->c-string "/home/mt/Desktop"))
 (define iwd (unchecked-inotify-add-watch ifd path
                                          (bit-or IN_CREATE IN_DELETE)))
 iwd
 (when (and ifd iwd)
  (let loop ((e (inotify-read-event ifd)))
   (when e
    (newline)(newline) (display "Sth happened on my DESKTOP:")
    (let ((m (inotify:mask e)))
      (when (bit-and? m IN_CREATE) (display " File Creation"))
      (when (bit-and? m IN_DELETE) (display " File Deletion"))
      ;
      ; Those times when /delete/ meant move to trash ... but compressing
      ; files counts as /delete/. Desktop Metaphors Movies
      ;
      ; Now originally the /inotify/ support in VSI was meant to make
      ; a text editor say: "Hey, something else changed that file!" but
      ; that has been forgotten by now, I guess.
      ;
      ; Deleting several files only gave me one event? Somehow there
      ; could still be some way of ...
      ;
      )
    (newline)(newline)
    (repl-transcript
     (byte-string->number (inotify:len e))
     (write (inotify-name e))
    ;(REPL-HERE)
    ;(loop (inotify-read-event))
    ))))
)
;(REPL)
