(sys-load "VSI-core/namespace.scm")
(sys-load "VSI-core/repl.scm")

(repl-transcript
 (define e '(namespace (define x 10)))
 ;(namespace:parse e '())
 (namespace:transformer e '())
 (defined? 'x)

;; Simple Name Space
 (define n
 (namespace
  (define y 10)
  (define (x n) (+ n y 100))
  (define t 'let)))
n
 ((cdr (assq 'x (cdr n))) 3)
 (call n 'x 5)
 (from-namespace n 't)

;; General Name Space
 (define m
  (namespace
   (if system-is-big-endian
     (define (x a) 1)
     (define (x b) 2) )))
m
 (call m 'x 5)

;;
;; Name Spaces, Files, Import and Selective Value Retrieval
;;
;; Maybe an advanced MOS /call/ should be able to provide getters
;; and setters for non procedure objects but on a second thought:
;; procedure objects are data like anything else. Maybe the
;; selective retrieval (here /from-namespace/) could be defined
;; to resolve the name in the given namespace once. But then
;; further evaluations of the same expression wouldn't be able
;; to handle exchanged name spaces and that is the same as the
;; situation with /call/ and object method handlers.
;;
;; A file with some /load/ variant breaks the name space as /load/
;; and its variants are broken in several respects: our brand new
;; /namespace-load/ loads all the expression from the file into
;; a data graph before evaluating. That should become standard for
;; all /loads/. Our brand new /namespace-load/ respects the
;; current environment and that should become the standard for
;; all /load/ variants.
;;
;; Retrieval of values within nested name spaces could be a general
;; option for all retrieval routines when the symbol to be resolved
;; is a list of symbols. Let's say one name space /n/ contains a
;; binding /m/ with a name space which knows about /fact/: a
;; (from-namespace n (m fact)) would retrieve that value for us. But
;; that might not even be that useful in practice ... [XXX define
;; itsself can be considered broken, too. Without the current inner
;; defines it might as well add bindings to the environmnet like a
;; name space. With the prospective of other code transformations
;; like optimizers and machine instruction code generation the
;; practice of modifying environment frames should be kept in bounds.
;;
;; namespace->alist and alist->namespace should be provided instead
;; of a rich API to allow for selective imports, symbol modifications
;; and else. Modification of the bindings "visible" within the data
;; graph of a name space aren't respected while modification of data
;; is. That is the expected behavior. The name space itsself can as
;; well be considered immutable whereas its data graph (currently 
;; an association list) can be seen as a truthful description which
;; could result in another name space after modification with the
;; routines mentioned above.
;;
;; Future implementations of this prototype might opt for a sytem
;; dictionary in the environments to realize the name space and
;; this would allow for memoization of variable objects eliminating
;; the performance penalty if not for the look up of the memoizer
;; which can be considered O(C) with respect to the number of
;; bindings. As that approach would be similar to the old module
;; system it needs to be scrutinized thouroughly to not impose the
;; same restrictions and shortcomings.
;;
 (define (void . args) *unspecified*) ; XXX repl-transcript helper
 (define u (namespace-load "VSI-core/posix.constants.scm"))
 (void (namespace
  (repl-transcript
   (namespace-import u)
   S_IRUSR )))
 (not (false-if-exception (write S_IRUSR)))
 (write (from-namespace u 'S_IRUSR))


;;
;; Rudimentary Namespace Modifications
;;
(define l (namespace->alist u))
(define k (filter (lambda (e) (not (eq? (car e) 'S_IRUSR))) l))
(define j (map (lambda (e)
                (cons (string->symbol
                       (string-append "posix:"
                                      (symbol->string (car e)) ))
                      (cdr e)))
               k))
(define v (alist->namespace j))
(namespace-import v)
(defined? 'S_IRUSR)
(defined? 'posix:S_IRUSR)
(defined? 'posix:S_IWUSR)
posix:S_IWUSR

;;
;; Fast Call vs Call vs etc (1M)
;; -real 0m6.921s  user 0m6.912s  sys 0m0.004s (fast)-
;; real 0m6.259s  user 0m6.232s  sys 0m0.008s (call)
;; real 0m4.586s  user 0m4.572s  sys 0m0.004s (fixed)
;; real 0m2.662s  user 0m2.644s  sys 0m0.008s (p see below)
;; real 0m2.729s  user 0m2.708s  sys 0m0.012s (x see below)
;; (These numbers may shape MOS a little bit.)
;;
;; Changes after profiling ~-.5s (R1083->R1084)
;; Additional simplifications of the call variants.
;; real 0m2.375s  user 0m2.356s  sys 0m0.008s (fixed implicit apply)
;; -real 0m5.141s  user 0m5.116s  sys 0m0.012s (fast with unchecked apply)-
;; real 0m5.838s  user 0m5.816s  sys 0m0.008s (call)
;;  Both /p/ and /x/ only gained 0.3 seconds less time after the changes
;;  based on profiling R1083 which means that as of R1084 a fixed call
;;  can be considered as fast as a direct implicit function application
;;  or regular procedure call (which it is).
;;
;; /p/ is somehow always slightly faster than /x/ which is unexpected.
;; The /fast call/ variant might not be worth it.
;;
;; Check imported namespace performance, too, please!
;;
(define 1k 1000)
(define 10k 10000)
(define 1M (* 1k 1k))

(define p (from-namespace n 'x))

(define y 10)
(define (x n) (+ n y 100))

(define sum 0)
(let loop ((i 0))
  (when (< i 1M)
    ;(set! sum (+ sum (p 5)))
    ;(set! sum (+ sum (x 5)))
    (set! sum (+ sum (fixed-call n 'x i)))
    ;(set! sum (+ sum (call n 'x i)))
    (loop (+ i 1))))
sum
)
;(REPL)
