;;
;; Subservice: compute results in parallel w/o interrupting evaluation
;;           INCOMPLETE PROTOTYPES with implicit serialization
;;

;;
;; Subservice   K A P P A
;;              Experimental Transmission of Lambda Form Results (Closures)
;; 
;; This is a top level only prototype: closures embeddeded in a data 
;; graph will only be handled correctly when the serialization form
;; has been implemented in the hosting system.
;;
;; TODO system environment dependencies
;; TODO name bindings for the closure object in its own environment
;;      While circular data structures can in general not be serialized
;;      effectively its should be possible to resolve any bindings to
;;      lambda objects or closures and convert them to maybe pointing
;;      kappa objects.
;;
;; CAVEATS This severely violates the primitive coherence between outer
;;         form and theoretical concept. A kappa form by itself looks
;;         ridiculous and may impose future problems. The best match
;;         outer form is in fact a simple /let/ form: but even then
;;         the concept of /environment/ is broken as the data graph
;;         used as the enironment doesn't necessarily truthfully
;;         reflect the original textual environment. A kappa form
;;         encloses /values/ in the binding list where a closure
;;         encloses /locations/.
;;
(define (subservice:write object evaluation-port)
 (cond
  ((closure? object)
     (let ((form (procedure-source object)))
       (write (list 'kappa
                    form
                    (procedure-environment object))
              evaluation-port) ))
  (else
   (write object evaluation-port))))

(define (subservice:closure? object)
 (and (pair? object)
      (equal? (car object) 'kappa)
      (= 3 (length object))))

(define (subservice:read port)
 (let ((result (read port)))
  (cond
   ((subservice:closure? result)
      (eval (second result) (third result)))
   (else
      result)) ))
   
(define (simple-controller closure evaluation-port)
 (let loop ()
   (catch #t (lambda () (subservice:write (closure) evaluation-port))
     (lambda (key . args)
       (when (not (eq? key 'quit))
         (write (cons 'error-object-prototype
                      (cons (values key args) '()))
                evaluation-port) )) )))

(define (error-object? object)
  (and (pair? object)
       (eq? (car object) 'error-object-prototype)
       (pair? (cdr object))
       (pair? (cadr object))
       (symbol? (caadr object))))

(define (rethrow error-object)
 (apply throw (caadr error-object)
              (cadadr error-object)))

(define (when-its-ready closure)
 (let ((finished #f)
       (final-result '())
       (comports (pipe))
       (pid      (primitive-fork)))
  (setvbuf (cdr comports) _IONBF) ; XXX twice
  (if (not (= pid 0)) ; that's zero in the new subprocess
    (lambda ()
     (if finished
      (if (error-object? final-result) (rethrow final-result) final-result)
      ; XXX prepare for this: control (or controlled-select) needs to 
      ; multiplex timeouts
      (let ((res (select (list (car comports)) '() '() 0 0)))
       (if (and (pair? res) (pair? (car res))
                (eq? (caar res) (car comports)))
           
         (begin ;(REPL-HERE) ;debugging
                (let ((result (subservice:read (car comports))))
                  (set! finished         #t)
                  (set! final-result result)
                  (false-if-exception (kill pid SIGTERM))
                  (false-if-exception (waitpid pid))
                  (false-if-exception (close (car comports)))
                 (if (error-object? result)
                   (rethrow result)
                   result)))
         'busy-Try-again-later-please))))
    (begin
      ;; empty IO: we use an explicit evaluation port
      ;; (or use dup2 to exchange C stdio as well)
      (close (set-current-error-port  (%make-void-port "w")))
      (close (set-current-output-port (%make-void-port "w")))
      (close (set-current-input-port  (%make-void-port "r")))
      (simple-controller closure (cdr comports))
      (newline (cdr comports))
      (force-output)
      ;(pause)
      (primitive-exit)
      )) ))

(define (parallel:closure closure comport)
  (setvbuf comport _IONBF)
  (close (set-current-error-port  (%make-void-port "w")))
  (close (set-current-output-port (%make-void-port "w")))
  (close (set-current-input-port  (open-input-string "")))
  (simple-controller closure comport)
  (newline comport)
  ;(force-output comport)
  ;(pause)
  (primitive-exit))

;; This would be easier with a class /parallel/ ...
;; That's not experimental that is experimenting!
;; That's not subservice but it belongs to parallels as it blocks
;; evaluation waiting for the results. <- Maybe that's just the 
;; missing control part.
(define (parallel-values:transformer expression environment)
 (let ((pids '()) (all-comports '()))
  (let loop ((expressions (cdr expression)))
   (cond ((not (null? expressions))
           (let ((comports (pipe))
                 (closure (lambda () (eval (car expressions) environment)))
                 (pid -1))
            (set! all-comports (cons (car comports) all-comports))
            (set! pid (primitive-fork))
            (if (= 0 pid)
              (parallel:closure closure (cdr comports)) ; discontinued
              (set! pids (cons pid pids))))
           (loop (cdr expressions)))
         (else ;; Don't think about control flow here: think about
               ;; multidimensional cracks in the ice. ;-)
          (set! pids         (reverse! pids)) ; XXX directional pity
          (set! all-comports (reverse! all-comports))
          (let ((result '()))
            ;; Collect results
           (set! result (map read all-comports))

           ;; Wait for all parallels to finish
           (let loop ((p pids))
            (when (not (null? p))
             (waitpid (car p)) (loop (cdr p)) ))

           result)) )) ))
(define parallel-values (procedure->syntax parallel-values:transformer))


;;
;; Example Session
;;
(define (long-running-task)
 (let loop ((answer '())
            (n 1))
  (if (> n 1000)
    answer
    (begin (usleep 10000)
           (loop (cons n answer) (+ n 1))) )))

(sys-load "VSI-core/repl.scm")
(repl-transcript
;(simple-controller long-running-task)
(define x (when-its-ready long-running-task))
(define y (when-its-ready (lambda () (load "err.scm"))))

;; Serialization Tests
(define s (when-its-ready (lambda () "one string with a ; serialization test
newline in the string"))) ;; PASSED
(define u (when-its-ready (lambda () (if #f #t)))) ;; FAILED
;; pi PASSED: inaccuracy is in the range of regular rounding errors
(define p (when-its-ready (lambda () (/ pi 2)) ))
(define t 100)
(define k (when-its-ready (lambda () (let ((u t)) (lambda (n) (+ 1 n u)) )) ))
(x)
(p)
(usleep (* 100 1k))
(= (p) (/ pi 2))
(- (p) (/ pi 2))
(k)
((k) 10)
(subservice:write (k) (current-output-port))
) (REPL)

;;
;; Extended Session: Parallel Values
;; (this should be part of the parallels example but it depends on
;; some in the subservice example)
(when #f
(repl-transcript
(parallel-values (+ 1 2) (+ 3 4) (+ 5 6))
(define l (map 1+ (iota- (* 100 1k))))
(define v (map (lambda (i) (vector (/ i 2)
                                   (* 1 (+ 1 i))
                                   (- i)))
               l))
(define vl-i (lambda (i op) (lambda (v a) (op (vector-ref v i) a))))

(define v0 #(0 0 0))
(define vl-li (lambda (i op) (fold (vl-i i op) 0 v)))
(define vl-1i (lambda (i op) (fold (vl-i i op) 1 v)))
;; (define vld1 (vl-li 0))
;; (define vld2 (vl-li 1))
;; (define vld3 (vl-li 2))
;; (apply vector (parallel-values (vl-li 0 +) (vl-li 1 +) (vl-li 2 +)))
;; (apply vector (parallel-values (vl-li 0 -) (vl-li 1 -) (vl-li 2 -)))
(define (parallel-vector3-fold op vectors)
  (apply vector (parallel-values (vl-li 0 op) (vl-li 1 op) (vl-li 2 op))))
(define (parallel-vector3-1i op vectors)
  (apply vector (parallel-values (vl-1i 0 op) (vl-1i 1 op) (vl-1i 2 op))))

 ; Multiplication takes ages in one dimension - we really shouldn't
 ; multiply vectors in that way but sometimes you sit there and wait
 ; for that one parallel to finally finish.
(define m
  (when-its-ready (lambda ()
    (let ((values (vector->list (parallel-vector3-1i * v)) ))
      (parallel-values (log10 (first values))
                       (log10 (second values))
                       (log10 (third values))) )) ))
(parallel-values
 ; It's not good to map such forms uncontrolled onto OS calls. Thus for
 ; the real parallelism syntax forms we'll need some kind of control
 ; to keep track of useful parallels and useful limits.
 (parallel-vector3-fold + v)
 (parallel-vector3-fold - v)
 (map log10 (vector->list (parallel-vector3-1i / v)))
) ; closes parallel-values
) ; closes repl-transcript
(REPL)
) ; closes extension paragraph
