;; SCA: System Call and Library Function Abstractions
;; EXPERIMENTAL PROTOTYPES
;;
;; Now that's foam: nothing for nothing!
;; - Yes, you can could even build structures with foam
;; that's why we tried to keep these things flat. The
;; implementation doesn't really need them: these
;; metaphors, these name bindings ... that's only for our 
;; wellfare to push imagination not into reality but
;; into these illusionary machines called computers.
;;
;; SCA Stage 1: Argument Handling
;;
;; Byte Structures provide means to give names to parts
;; of byte strings. This is achieved by providing a
;; /byte structure description/ from which accessors
;; for byte strings with the given layout will be
;; computed and defined.
;;
;; [The format and syntax of byte structure descriptions
;;  as well as their strictness regarding evaluation are
;;  still subject to evaluation respectively still
;;  unspecified.]
;;
;; Even if the computation of the accessors is automatic
;; the development of that system is documented serving
;; as some sort of quick reference as the "lambda magic" 
;; of the final definition routines looks a bit complicated.
;;
;;   Stage 0: As an afterthought we define a way to access
;;            structure members with a functional style,
;;            i.e. a function is called with the members
;;            of the structure in their original sequence
;;            as described by the structure descriptions
;;            and means are provided to "get the changes
;;            into" byte structures again.
;;
;;   Stage 1: This true origin defines the accessors
;;            we want to have for one specific byte
;;            structure.
;;
;;   Stage 2: A generalization of stage 1 encapsulates
;;            values for offsets and sizes into an accessor
;;            closure which is really in use in the final
;;            accessor definition routines but still don't
;;            make use of the byte structure description.
;;
;;   Stage 3: The routines provide resolution of member
;;            names to offset and length values following
;;            the byte structure description as needed by
;;            the accessor generators of stage two.
;;
;;   Stage 4: We add accessor symbol creation and definition
;;            for whole byte structures.
;;
;;  [Stage 5: We still wait for the pressure to flatten the
;;            other stages: but in fact we just have to 
;;            review the other stages to realize there is
;;            only one clean accessor closure bound to its
;;            name in the system environment and each and
;;            every of them is as flat as our stage 1
;;            accessors. No name look-ups after their
;;            definition - nothing else than a simple
;;            call to a /byte-string-select/ or
;;            /byte-string-set!/.
;;
;; Even the above text is still considered a prototype.
;; The current implementation is yet to be tested.
;;                  
;; TODO Implement the prototype for the parsing of a byte
;; structure description according to the preliminary
;; specification in the Roadmap: Outer Form.
;;
;; TODO Arrays need to be introduced: final accessors
;; should allow for an index to retrieve the i'th member
;; of the array where each member has the same size in
;; bytes. Arrays of arrays are considered pointers.
;; Arrays of structures are just that: each member has
;; has a size equal to the length of the structure in
;; bytes. Maybe they will call them 'byte-array'or 
;; 'byte word array'.
;;
;; (BAP [DRAFT]: Byte Architectional Protocol ;-)
;;
;; 1) Parse the byte structure description.
;; 2) Add padding if it is an ABI byte structure description.
;; 3) Define the accessors.
;; 4) Return the parsed byte structure description.
;;
;; Further modifications of the [parsed] byte structure
;; description will not have any effect, i.e. no entity
;; should rely dynamically on that particular data graph
;; of the byte structure description after it has been
;; processed which should happen only once. If extensions
;; of the protocol will need parts or all of the byte
;; structure description after the initial processing
;; some copy mechanism (probably of the return value
;; of the byte structure description) needs to be in
;; effect.
;;  The return value of a byte structure description is
;; a data graph describing the byte structure description.
;; It is not in use anymore by the system and can be
;; modified at will. It's main purpose is to provide
;; introspection for technical reviews and else. Its
;; exact contents are unspecified and may well differ
;; from one machine architecture to the next or not.
;; The contents of the data graph are defined by the
;; internal parsing routines.


(define (sysdep-amd/intel64:alignment)
  ;; See page 12 of the System V ABI 
  ;;
  ;; BTW That 80 bit floating point thing was the extended
  ,; variant but part IEEE-754. Then the standard describes
  ;; some called ieee 754R or decimal floating point. Then
  ;; they describe packed variants as if these were only in
  ;; in use by SIMO registers (maybe the compacted format
  ;; is defined by MMX & Co where compacted and packed are
  ;; two different ... these 4 and 8 byte floating points
  ;; are described as part of FIPS?. But when it comes to
  ;; argument passing they tell us to push a 16 byte word
  ;; into two 8 byte registers. Maybe this text is the
  ;; confusion catcher.
  ;;
#! from System V Application Binary Interface / AMD64 Supplement (p. 13)

Structures and unions assume the alignment of their most strictly
aligned component. Each member is assigned to the lowest available
offset with the appropriate alignment. The size of any object is
always a multiple of the object‘s alignment.

An array uses the same alignment as its elements, except that a local
or global array variable of length at least 16 bytes or a C99
variable-length array variable always has alignment of at least 16
bytes.

Structure and union objects can require padding to meet size and
alignment constraints. The contents of any padding is undefined.

!#
 ;; We currently don't need to care about heap variable alignment
 ;; and we assume an eight byte alignment for the start of each
 ;; and every byte string and we assume the described structure
 ;; is targeted to a byte string. Maybe byte strings of sizes
 ;; smaller than eight bytes are not aligned to eight bytes but
 ;; then we won't have to align to eight bytes anyway. 
 ;;
 ;; SIGBUS: Why doesn't C offer the possiblity to distinguish
 ;; between manual and automatic padding. Maybe historical
 ;; tradition but maybe byte structures following other padding
 ;; rules (marsian padding rules but they didn't get the call ;-)
 ;; would lead a violation of alignment rules in effect on the
 ;; hosting system: SIGBUS. System scheme or SCA should be become
 ;; translatable to machine code: using the current byte copy
 ;; method shouldn't be too difficult compared to single byte
 ;; handling. [That sentence is another confusion catcher, right?]
system-type-size)

;;
;; Generalizing Examples: #1 TERMIOS
;;

(sys-load "VSI-core/byte-structures.scm")
(sys-load "VSI-core/posix.constants.scm")

(define termios
 (byte-structure-description
  (iflags   int)
  (oflags   int)
  (cflags   int)
  (lflags   int)
  (cc-line  byte)
  (cc-array NCCS) ; CC array dummy
  (padding  3)
  (ispeed   int)
  (ospeed   int)))

(define-bs-accessors 'termios termios)


(define (termio-eof)
 (let* ((fdes (port->fdes (current-output-port)))
        (ts (unchecked-tcgetattr fdes))
        (cc (termios:cc-array ts)))
   (byte-string-set! cc VEOF #"4")
   (termios:set-cc-array! ts cc)
   (unchecked-tcsetattr fdes ts)))

;;
;; Stage 0: Functional
(define (call-for-termios whatnot fdes)
  (unchecked-tcsetattr fdes
   (values->byte-structure
    (call-with-bytes termios (unchecked-tcgetattr fdes) whatnot)
    termios)))

(define (termios-dump fdes)
  (call-for-termios (lambda ( . args)
                      (for-each (lambda (member) (display member) (newline))
                                args)
                      (newline)
                      args)
                    fdes))

(define (termio-noecho fdes)
 (call-with-bytes termios (unchecked-tcgetattr fdes)
  (lambda (iflags oflags cflags lflags
           cc-line cc-array ispeed ospeed)
   (unchecked-tcsetattr fdes
    (values->byte-structure
     (values iflags oflags cflags
             (bit-and lflags (bit-not ECHO ))
             cc-line cc-array ispeed ospeed)
     termios)) )))

;;
;; You can whine for a nice API and call them whatnot but
;; if you use bytes directly you're probably just creating
;; that nice comfortable safe and whatnot API.
;;
;; SCA still means System Call Abstraction and here we try
;; to figure out the inner workings to then forget about
;; them again.
;;
(sys-load "VSI-core/repl.scm")
(repl-transcript

(define fdes (port->fdes (current-output-port)))

;;
;; Collect some default terminal settings.
(define b-original (unchecked-tcgetattr fdes))
(begin
 (%terminal-raw-mode fdes) (define b-raw (unchecked-tcgetattr fdes))
;)
;(begin
 (%terminal-cooked-mode fdes) (define b-cooked (unchecked-tcgetattr fdes))
)
; (define iflags (bit-or BRKINT IGNPAR ISTRIP ICRNL IXON))
; (define lflags (bit-or ECHO ISIG ICANON))

;; (define (sca-local-echo-off fdes)
;;   (unchecked-tcsetattr fdes
;;    (termios:set-lflags! (unchecked-tcgetattr fdes)
;;                         (bit-and (bit-not ECHO) )))

;; (define (sca-local-echo-on)
   ; that's too easy to be
;;   #t)

;;
;; Make a copy to keep our defaults clean.
(define b-noecho (byte-string-copy b-cooked))

;; (define noecho-lflag (bit-and (byte-string-select b-noecho 12 4)
;;                               (bit-not ECHO)))
;; (byte-string-set! b-noecho 12 noecho-lflag)

(termios:set-lflags! b-noecho (bit-and (bit-not ECHO)
                                       (termios:lflags b-noecho)))

; p-set is currently DEFUNCT
;(termios:set-lflags! b-noecho (p-set (termios:lflags b-noecho) ECHO))

(if (unchecked-tcsetattr fdes b-noecho)
    (begin 
           (newline) (newline)
           (display "Try to type some letters and press enter. ")
           (display "KEYBOARD ECHO is OFF!> ")
           (read))
           ; (star-reader))
    'tc-error)


(%terminal-cooked-mode fdes)
;(termios-dump fdes)
;(define l (byte-structure->list termios b-noecho))
)


;(termio-eof)
;(display "Try EOF ...")
;(read)

(display "Please note: the REPL tries to restore the terminal operation mode\n")
(display "for each prompt, i.e. after each expression. Use /begin/ to test\n")
(display "terminal settings from within the REPL.\n")
;(display "Try EOF ...")
;(begin (termio-eof) (read))
;(newline)
(REPL)
