
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

(define display-class (class            (extension viper-object-class) ; some GIO/X root class?
  (slots connection riddle rr root-window)
  (slots xdetail screens visuals root-visual formats default-format)
  (slots xext-keyboard xext-big-requests xext-shared-memory xext-render) ; required extensions
  (slots keyboard pointer-device)

;;
;; What a mess! This an adaption of the original SCA/X steep hack
;; to VOS. This class has four parts:
;;      slots   (see above)
;;      methods (see below)
;;      definitions
;;        structs: formats
;;        inner classes: riddle, response reader
;;        namespace: xcons (X Connection Setup)
;;                   xserv (X Server Extensions, Fonts, Display Routines)
;;

;; DISPLAY CLASS PART II: METHODS

(method initialize ()
 (let* ((setup    (call Xcons 'X-connection))
        (socket   (assq-ref setup 'connection))
        (response (assq-ref setup 'bytes)))
  (this (set! 'xdetail setup))
  (this (set! 'connection (X-make-transport socket)))
  (this (set! 'riddle (make-object sca/x-riddle-class 
                        (assq-ref setup 'rid-base)
                        (assq-ref setup 'rid-mask))))
  (this (set! 'rr (make-object sca/x-response-reader-class this)))

  (this (set! 'screens (map (lambda (screen-detail) (make-object screen-class this screen-detail))
                            (assq-ref setup 'screens)) ))
  (this (set! 'root-window ((this riddle) (lookup (assq-ref setup 'root-window)))))
  (this (set! 'root-visual ((this riddle) (lookup (assq-ref (first (assq-ref setup 'screens)) 'root-visual))))) ;; XXX patch
  (when (not (this root-window)) (error "GIO initialization failure: no default root window"))

  ;; Pixmap Formats
  (this (set! 'formats (map alist->format (assq-ref setup 'formats))))
  (this (set! 'default-format (let ((res (filter (lambda (f) (= (slot-ref f 'depth) 24)) ;; XXX some pixmap format ...
                                                 (this formats))))
                                (and (not (null? res)) (first res)))))
 ) ; end of traditional X connection setup

 (this (set! 'pointer-device (make-object pointer-device-class this)))

 ;; Initialize Required Extensions 
 (this (set! 'xext-keyboard      (make-object xext-keyboard-class      this))) ; XXX this initializes the keyboard slot
 (this (set! 'xext-big-requests  (make-object xext-big-requests-class  this))) ; currently not required
 (this (set! 'xext-shared-memory (make-object xext-shared-memory-class this)))
 (this (set! 'xext-render        (make-object xext-render-class        this)))
)

(method destroy ()
 #f
 ;; YES! Release'em all!
)

(method describe ()
 (apply brass-append "Display\n"
                     (map (lambda (d) (d (describe))) (this screens)))
)

(method bell () #f ;; TODO all the bells & whistles with strings & brass
)

(method ungrab-pointer (time)
 (fixed-call xserv 'X-ungrab-pointer (this connection) time)
)

(method X-extensions ()
 (let ((con (this connection)))
  (call xserv 'X-list-extensions con)
))

(method X-query-extension (name)
 (call xserv 'X-query-extension (this connection) name)
)

(method list-fonts (pattern)
 (let ((con (this connection)))
   (call xserv 'X-list-fonts con pattern)
))


;; DISPLAY CLASS PART III: DEFINITIONS (structs & inner classes)

;; This should describe pixmap and bitmap formats: especially
;; /scanline-pad/ should be used somehow.
(define format-layout (struct depth bits-per-pixel scanline-pad))
(define (alist->format setup-alist)
 (let ((f (make-struct format-layout)))
   (@slot-set! f 'depth          (assq-ref setup-alist 'depth))
   (@slot-set! f 'bits-per-pixel (assq-ref setup-alist 'bits-per-pixel))
   (@slot-set! f 'scanline-pad   (assq-ref setup-alist 'scanline-pad))
   f
))

(define sca/x-response-reader-class (class  (extension viper-object-class)
 (slots event-vector error-vector)

 (method initialize (display)
  (this (set! 'event-vector (make-vector 256 #f)))

                           ;; ??? xevent codes into common?
  (vector-set! (this event-vector)  0 default-error-handler)
  (vector-set! (this event-vector)  2 default-key-event-handler)
  (vector-set! (this event-vector)  3 default-key-event-handler)
  (vector-set! (this event-vector)  4 default-button-event-handler)
  (vector-set! (this event-vector)  5 default-button-event-handler)
  (vector-set! (this event-vector)  6 default-motion-event-handler)
  (vector-set! (this event-vector)  7 default-menter-event-handler)
  (vector-set! (this event-vector)  8 default-mleave-event-handler)
  (vector-set! (this event-vector) 12 default-expose-event-handler)
  ;(vector-set! (this event-vector) 13 default-graphics-exposure-event-handler)
  (vector-set! (this event-vector) 17 default-destroy-event-handler)
  (vector-set! (this event-vector) 22 default-configure-event-handler)

  (this (set! 'error-vector
              (vector-project! (make-vector 256 'unknown-error-code) 0 constant-error-vector)))

  (sys-control (add-read-handler (slot-ref (display connection) 'channel)
                                 (sca/x-channel-handler display)))
 )
 (method install-handler (idx handler)
  (when (vector-ref (this event-vector) idx) (error "GIO sca/x response reader install handler collision"))
  (vector-set! (this event-vector) idx handler)
 )

(define (sca/x-channel-handler display)
 (let* ((transport  (display connection))
        (con        (slot-ref transport 'channel)))
(lambda ()
 (let ((fbw (read-bytes 32 con)))
  (or (and (byte-string? fbw)
           (if (>= (byte-string-length fbw) 32) ; always at least 32 bytes!
               (X:dispatch display fbw)
               'X-control-would-need-to-fill-a-buffer-but..)) ; shouldn't happen
      (and (or (not fbw) (eof-object? fbw))
           (begin (sys-control (remove-read-handler con))
                  (close con)
                  (display (destroy))
                  (X-log-debug "X eos")))
      (X-log-debug "X-error:bad-message-from-channel " fbw) )))
))

(define (sca/x-queue-handler display)  ;;; SCA/X transport layer event queue
 (let ((transport (display connection)))
  (let loop ((fbw (X-next-qevent! transport)))
   (when (not (null? fbw))
    (X:queue-dispatch display fbw)
    (loop (X-next-qevent! transport))
  )) ; there's somehow a guarantee to see events only but but another request because of an event ...
))

(define (X:queue-dispatch display fbw) ; almost like X:dispatch
  (let* ((code    (byte-string->number (bit-and #"127" (byte-string-select fbw 0 1))))
         (handler (vector-ref ((display rr) event-vector) code)))
    (if handler
      (handler display fbw)
      (X:log-debug "VSI SCA/X: unhandled event: " fbw) )))

;; Maybe this should be a method since this is an inner class anyway.
(define (X:dispatch display fbw)
  (let* ((code    (byte-string->number (bit-and #"127" (byte-string-select fbw 0 1))))
         (handler (vector-ref ((display rr) event-vector) code)))
    (if handler
      (begin (handler display fbw) (sca/x-queue-handler display))
      (X:log-debug "VSI SCA/X: unhandled event: " fbw) )))

(define (default-error-handler display fbw)
 (let ((err-code     (byte-string->number (byte-string-select fbw 1 1)))
       (error-vector ((display rr) error-vector)))
  (X:log-debug "VSI SCA/X: unhandled error: " (vector-ref error-vector err-code) fbw)
  (error (format "Window System Error: ~A" (vector-ref error-vector err-code)))
))

;; These destroy the events.
;; The listeners could create selective copies if needed for further
;; processing.
(define make-event-handler (macro (class method-name)
 `(lambda (display fbw)
   (let* (;(riddle (display riddle))
          (ev  (make-object ,class display fbw))
          (win (ev event-window)))
     (cond (win  (win (,method-name ev)))
           (else (X-log-debug "GIO ignoring event with unresolved event window id ")))
     (ev (destroy)) )) ))

(define       default-key-event-handler (make-event-handler key-event-class              key-event))
(define    default-button-event-handler (make-event-handler button-event-class        button-event))
(define    default-motion-event-handler (make-event-handler motion-event-class        motion-event))
(define    default-menter-event-handler (make-event-handler motion-enter-event-class   enter-event))
(define    default-mleave-event-handler (make-event-handler motion-leave-event-class   leave-event))
(define    default-expose-event-handler (make-event-handler expose-event-class        expose-event))
(define   default-destroy-event-handler (make-event-handler destroy-event-class      destroy-event))
(define default-configure-event-handler (make-event-handler configure-event-class  configure-event))

;(define default-graphics-exposure-event-handler ;; it's called drawable where the macro has event-window <- use the same swap as the structure notify events
;                    (make-event-handler graphics-exposure-event-class graphics-exposure-event))

;; Symbols for error codes
(define constant-error-vector #(undefined
                       Request  Value    Window Pixmap   ; 4
                       Atom     Cursor   Font   Match    ; 8
                       Drawable Access   Alloc  Colormap ; 12
                       GContext IDChoice Name   Length   ; 16
                       Implementation))

)) ; closes class sca/x-response-reader


(define sca/x-riddle-class (class  (extension viper-object-class)
  (slots tree depth offset base seed)

;; NB For advanced memory statistic analysis we might want to
;; include the number of pairs used in the tree as they may
;; look like a memory leak but are the expected behavior.

(define x:riddle
  (bit-and (bit-not! (bit-reverse! (number->byte-string-4 7)))))

(method initialize (base mask)
 (let* ((bits (call xriddle 'X:initialize-bits mask)))
  (this (set! 'tree   (cons #f #f )))
  (this (set! 'offset (second bits)))
  (this (set! 'base   base))
  (this (set! 'depth  (call sca/x-common 'ones-4! (byte-string-copy mask))))
  (this (set! 'seed   (number->byte-string-4 0)))
  ))

(method lookup (id)
 (fixed-call xriddle 'X:riddle-lookup (this depth) (this tree)
             (bit-shift-right id (this offset))))

(method remove-id (id)
  (or (fixed-call xriddle 'X:riddle-register!
            (this tree)
            (this depth)
            (bit-shift-right id (this offset))
            #f)
      (error "VSI SCA/X: resource id mismatch: id not found")))

(method register-id (id object)
  (or (not (fixed-call xriddle 'X:riddle-register!
                       (this tree)
                       (this depth)
                       (bit-shift-right id (this offset))
                       object))
      (error "VSI SCA/X: resource id collision")))

(method register-object (object)
 (let ((bits (this seed)))
  (set! bits (fixed-call xriddle 'X:riddle-find-next!
                         (this tree)
                         (this depth)
                         (bit-shift-right bits (this offset))
                         object))
  (when (null? bits)
    (error "VSI SCA/X: free resource id underflow: clean the tree, please!"))
  (set! bits (bit-and (fixed-call xriddle 'bit-path->byte-string-4 bits)))
  (this (set! 'seed bits))
  (bit-and x:riddle (bit-or (this base) (bit-shift-left bits (this offset))))
))

(define xriddle (namespace
  (namespace-import sca/x-common)

(define (X:initialize-bits mask)
 (let loop ((bits (byte-string-copy mask)) (offset 0))
  (if (bit-and? one-4 bits)
    (values bits offset)
    (loop (bit-shift-right! bits 1) (+ offset 1)) )))

(define (X:riddle-register! tree riddle-depth id value)
 (let loop ((node  tree)
            (bits  (byte-string-copy id))
            (depth 0))
  (if (= depth riddle-depth)
    (let ((terminator (car node)))
       (set-car! node value)
       terminator)
    (let* ((one (bit-and? one-4 bits))
           (get-branch  (if one     cdr      car))
           (set-branch! (if one set-cdr! set-car!))
           (branch (get-branch node)))
     (when (not branch)
       (set! branch (cons #f #f))
       (set-branch! node branch))
     (loop branch (bit-shift-right! bits 1) (+ 1 depth)) )) ))


(define (X:riddle-find-next! tree riddle-depth seed value)
 (let recurse ((node  tree)
               (bits  seed)
               (depth 0)
               (bit-path   '() ))
  (if (= depth riddle-depth)
    (let ((terminator (car node)))
       ;; Terminators which aren't falsity are collisions. We just
       ;; return falsity to then try other bit paths.
       (and (not terminator)
            (begin (set-car! node value) bit-path)))
    (let* ((one         (bit-and? one-4 bits))
           (get-branch  (if one     cdr      car))
           (set-branch! (if one set-cdr! set-car!))
           (branch      (get-branch node)))
     (when (not branch)
       (set! branch (cons #f #f))
       (set-branch! node branch))
     (or (recurse branch
                  (bit-shift-right bits 1)
                  (+ 1 depth)
                  (cons one bit-path))
         (begin (set! one         (not one))
                (set! get-branch  (if one     cdr      car))
                (set! set-branch! (if one set-cdr! set-car!))
                (set! branch      (get-branch node))
                (when (not branch)
                  (set! branch (cons #f #f))
                  (set-branch! node branch))
                (recurse branch
                         (bit-shift-right bits 1)
                         (+ 1 depth)
                         (cons one bit-path)))
         )) )))

(define (bit-path->byte-string-4 bit-path)
 (let loop ((b (byte-string-clear! (make-byte-string 4)))
            (r bit-path))
  (cond ((null? r) b)
        ((car   r) (loop (bit-or (bit-shift-left! b 1) one-4) (cdr r)))
        (else      (loop         (bit-shift-left! b 1)        (cdr r))) )))


 (define (X:riddle-lookup riddle-depth tree id)
  (let loop ((node  tree) (depth 0) (bits  (byte-string-copy id)))
   (and node
        (if (= depth riddle-depth)
            (car node)
            (loop (if (bit-and? one-4 bits)
                      (cdr node)
                      (car node))
                  (1+ depth)
                  (bit-shift-right! bits 1)) )) ))
))
)) ; closes sca/x-riddle-class


;; DISPLAY CLASS PART III: DEFINITIONS (xcons namespace)

(define Xcons (namespace (namespace-import sca/x-common)
;;
;; X Connection Setup & Authentication
;;

(define X-conn-client-prefix-layout
 (byte-structure-description
  (byte-order 1)
  (padding    1)
  (major-version 2)
  (minor-version 2)
  (nbytes-auth-proto  2)
  (nbytes-auth-string 2)
  (padding    2)))
(define-bs-accessors 'xconc X-conn-client-prefix-layout)

;; (define xau-file
;;   (with-input-from-file "/home/mt/.Xauthority"
;;     (lambda () (read-bytes 16000 (port->fdes (current-input-port))))))
(define Xau-type    first)
(define Xau-address second)
(define Xau-number  third)
(define Xau-data    fourth)
(define Xau-entries (x-authentications))
(define x-auth-method-string "MIT-MAGIC-COOKIE-1")
(define x-auth-method-bytes (deprecated-string->c-string "MIT-MAGIC-COOKIE-1"))
(define x-auth-data (Xau-data (first Xau-entries))) ; XXX explicit

(define (pc n v) (number->byte-string v n))
(define X-conn-client-prefix
  (make-byte-string (bs-length X-conn-client-prefix-layout)))

(define x X-conn-client-prefix)
(xconc:set-byte-order!         x (pc 1 (if system-is-big-endian #o102 #o154)))
(xconc:set-major-version!      x (pc 2 11))
(xconc:set-minor-version!      x (pc 2 0))
(xconc:set-nbytes-auth-proto!  x (pc 2
    (- (byte-string-length x-auth-method-bytes) 1))) ; w/o trailing NULL!!!
(xconc:set-nbytes-auth-string! x (pc 2
    (byte-string-length x-auth-data) )) ; 128 bits for mcookie
;; XXX invalid mcookie key: reversing the byte order doesn't help

 ;; (values->byte-structure
 ;;  (list
 ;;   (pc 1 (if system-is-big-endian #o102 #o154))
 ;;   (pc 2 11)
 ;;   (pc 2 0)
 ;;   (pc 2 (- (byte-string-length x-auth-method-bytes) 1)) ; w/o trailing NULL!!!
 ;;   (pc 2 (byte-string-length x-auth-data)) ; 128 bits for mcookie
 ;;  ) X-conn-client-prefix-layout ))

;; Lock Picking (Noobie Level)
;; (or how long does it take to open a window on my own computer)
;;
;; There is some interpretable description of padding in the standards
;; when it comes to C strings: first of all one has to know every C string
;; ends with a zero byte. Second one has to realize the padding byte after
;; a string in the protocol specification refers to that implicit zero byte
;; which we have to add explicitely as Scheme knows about the lengths of
;; its strings but they expect us to use the number of bytes before that zero
;; byte token when it comes to the number of bytes fields. Third: break that
;; rule again and put the final padding byte in front of the authentication
;; data string which isn't expected to be a C string but a sequence of bytes.
;;  How is that possible? Does the authentication violate the protocol
;; specifications? The protocol specifications declares these padding bytes
;; to depend on the fields containing the number of bytes of the strings.
;; (see p. 123 of the standard)
(define X-conn-client-helo
  (byte-string-adjoin X-conn-client-prefix
                      x-auth-method-bytes
                      #"0" ; manual mcookie pre-padding (reverse engineered)!!! 
                      x-auth-data
                      ))

;; According to the standard /success/ is one of Authenticate,
;; Failure or Success but I couldn't find all those constants and
;; as Xproto.h states it:
;; > ...
;; > The protocol also defines a case of success == Authenticate, but
;; > that doesn't seem to have ever been implemented by the X Consortium.
;; > ...
;; Then there is a definition of "Success" in the header files and CEE tells
;; us this:
;;   (lookup-definition X "Success") => 0
;; While reality tells us that:
;;   (xcons:length-reason r) => #"0"
;;   (xcons:success r)       => #"1"
;; when there isn't any reason to not read the server response of a
;; successful connection setup the success field of the response prefix
;; is /byte-string-true?/ and that is what we will branch on.
;;
;; Now the encoding part of the protocol specifications (pp. 123) is a
;; bit more concise on things: Failure 0, Success 1, Authenticate 2 and
;; this is what we will branch on.
;;
(define X-conn-setup-prefix-layout
 (byte-structure-description
  (success       1) ; branch on this
  (length-reason 1)
  (major-version 2)
  (minor-version 2)
  (length        2))) ; Is this with the prefix or w/o???
(define-bs-accessors 'xcons X-conn-setup-prefix-layout)

(define X-conn-setup-layout
 (byte-structure-description
  (release   4)
  (rid-base  4)
  (rid-mask  4)
  (motion-buffer-size 4)
  (nbytes-vendor      2)
  (max-request-size   2)
  (num-roots   1)
  (num-formats 1)
  (image-byte-order 1)
  (bitmap-bit-order 1)
  (bitmap-scanline-unit 1)
  (bitmap-scanline-pad  1)
  (min-key-code     1)
  (max-key-code     1)
  (padding          4)
))
(define-bs-accessors 'xcons-info X-conn-setup-layout)

(define X-format-layout
 (byte-structure-description
  (depth          1)
  (bits-per-pixel 1)
  (scanline-pad   1)
  (padding        5)))

(define-bs-accessors 'xformat X-format-layout)

(define X-screen-layout
 (byte-structure-description
  (root-window      4)
  (default-colormap 4)
  (white-pixel      4)
  (black-pixel      4)
  (current-input-masks 4)
  (width-in-pixels  2)
  (height-in-pixels 2)
  (width-in-millimeters  2)
  (height-in-millimeters 2)
  (min-installed-maps 2)
  (max-installed-maps 2)
  (root-visual      4)
  (backing-stores   1)
  (save-unders      1)
  (root-depth       1)
  (number-of-depths 1)))
(define-bs-accessors 'xscreen X-screen-layout)

(define X-depth-layout
 (byte-structure-description
  (depth   1)
  (padding 1)
  (number-of-visuals 2)
  (padding 4)))
(define-bs-accessors 'xdepth X-depth-layout)

(define X-visual-layout
 (byte-structure-description
  (visual-id  4)
  (class      1)
  (bits-per-rgb-value 1)
  (colormap-entries   2)
  (red-mask   4)
  (green-mask 4)
  (blue-mask  4)
  (padding    4)))
(define-bs-accessors 'xvisual X-visual-layout)


(define (xcons:setup-format r)
 (values (bs-length X-format-layout)
  (list (cons 'depth (byte-string->number (xformat:depth r)))
        (cons 'bits-per-pixel (byte-string->number (xformat:bits-per-pixel r)))
        (cons 'scanline-pad (byte-string->number (xformat:scanline-pad r))))))

(define (xcons:setup-visual r)
 (values (bs-length X-visual-layout)
 (list (cons 'visual-id (xvisual:visual-id r)) ; use as the key please
       (cons 'class     (xvisual:class     r)) ; Convert please.
       (cons 'channel-bits     (byte-string->number
                                (xvisual:bits-per-rgb-value r)))
       (cons 'colormap-entries (byte-string->number
                                (xvisual:colormap-entries r)))
       (cons 'red-mask   (xvisual:red-mask   r))
       (cons 'green-mask (xvisual:green-mask r))
       (cons 'blue-mask  (xvisual:blue-mask  r)))))

(define (xcons:setup-depth r)
 (let ((num-visuals (byte-string->number (xdepth:number-of-visuals r)))
       (visuals-offset (bs-length X-depth-layout))
       (size+visuals '()))
  (set! size+visuals (X:read-out-list num-visuals xcons:setup-visual
                                      (bs-length X-visual-layout)
                                      r visuals-offset))
  (values (+ (bs-length X-depth-layout)
             (first size+visuals))
  (list (cons 'depth   (byte-string->number (xdepth:depth r)))
        (cons 'visuals (second size+visuals)) ))))

(define (xcons:setup-screen r)
 (let ((num-depths (byte-string->number (xscreen:number-of-depths r)))
       (depths-offset (bs-length X-screen-layout))
       (size+depths '()))
 (set! size+depths (X:read-out-list num-depths xcons:setup-depth
                                    (bs-length X-depth-layout)
                                    r depths-offset))
 (values (+ (bs-length X-screen-layout)
            (first size+depths))
   ;; XXX The root window is the key to find the correct visuals for
   ;; new top level windows? We only support true color for now.
   ;; !!! We definitely need to put the screens as root windows into
   ;; our resource index. What presumptions can be made about these
   ;; IDs delivered by the server? 32 bits, 29 bits or bit mask bits???
  (list (cons 'root-window (xscreen:root-window r))
        (cons 'default-colormap (xscreen:default-colormap r))
        (cons 'white-pixel (xscreen:white-pixel r))
        (cons 'black-pixel (xscreen:black-pixel r))
        (cons 'current-input-masks (xscreen:current-input-masks r))
        (cons 'pixel-width  (byte-string->number (xscreen:width-in-pixels  r)))
        (cons 'pixel-height (byte-string->number (xscreen:height-in-pixels r)))
        (cons 'millimeter-width  (byte-string->number (xscreen:width-in-millimeters  r)))
        (cons 'millimeter-height (byte-string->number (xscreen:height-in-millimeters r)))
        (cons 'root-visual (xscreen:root-visual r))
        (cons 'backing-stores (xscreen:backing-stores r))
        (cons 'save-unders (byte-string-true? (xscreen:save-unders r)))
        (cons 'root-depth  (xscreen:root-depth r))
        (cons 'depths (second size+depths))))))

(define (xcons:setup socket r)
 (let ((vendor-offset (bs-length X-conn-setup-layout))
         (format-offset #f)
         (num-formats   #f)
         (  root-offset #f)
         (num-screens   #f))

    (set! num-formats (byte-string->number (xcons-info:num-formats r)))
    (set! format-offset 
      (+ vendor-offset (byte-string->number (xcons-info:nbytes-vendor r))))

    (set! num-screens (byte-string->number (xcons-info:num-roots r)))
    (set! root-offset ; w/o vendor padding
      (+ format-offset (* num-formats (bs-length X-format-layout) )))

    ;; How do we get unique tokens when we can't just use the memory
    ;; address of our virtual address space? Here is the answer: (p.14)
    ;; (That's the parallel object IPC/ID trick from back then. Many
    ;;  more to come.)

   (values
    (cons 'connection socket)
    (cons 'vendor
     (c-string->deprecated-string
      (byte-string-select r vendor-offset
         (byte-string->number (xcons-info:nbytes-vendor r)))))
    (cons 'root-window ;; XXX steep hack to get a window and nothing else
          (byte-string-select r root-offset 4))
    (cons 'release (byte-string->number (xcons-info:release r)))

    (cons 'rid-base (xcons-info:rid-base r))
    (cons 'rid-mask (xcons-info:rid-mask r))

    (cons 'min-key-code (byte-string->number (xcons-info:min-key-code r)))
    (cons 'max-key-code (byte-string->number (xcons-info:max-key-code r)))

    ;; Looks like they're maxing out here. We wait for XII to make
    ;; requests bigger 64k.
    (cons 'motion-buffer-size
      (byte-string->number (xcons-info:motion-buffer-size r)))
    (cons 'max-request-size
      (byte-string->number (xcons-info:max-request-size r)))

    ;; Protocol Encoding: 0 means MSB first, 1 means LSB first (pp. 123)
    (cons 'image-byte-order
      (if (byte-string-true? (xcons-info:image-byte-order r))
        'MSB-first 'LSB-first))
    (cons 'bitmap-bit-order
      (if (byte-string-true? (xcons-info:bitmap-bit-order r))
        'MSB-first 'LSB-first))
    (cons 'bitmap-scanline-unit (byte-string->number (xcons-info:bitmap-scanline-unit r)))
    (cons 'bitmap-scanline-pad  (byte-string->number (xcons-info:bitmap-scanline-pad  r)))
    
    (cons 'formats (second (X:read-out-list num-formats xcons:setup-format
                                            (bs-length X-format-layout)
                                            r format-offset)))
    (cons 'screens (second (X:read-out-list num-screens xcons:setup-screen
                                            (bs-length X-screen-layout)
                                            r root-offset)))
    (cons 'bytes  r) ;; XXX debugging
 )))

(define (X-connection)
 (let ((s (socket AF_UNIX SOCK_STREAM 0))
       (f "/tmp/.X11-unix/X0"))
  (connect s AF_UNIX f)
  (write-bytes X-conn-client-helo (port->fdes s))
  (let ((r (read-bytes 8 s)))
    (if (not (byte-string-true? (xcons:success r)))
      (begin
        (set! r (read-bytes (+ (byte-string->number (xcons:length-reason r))
                               (byte-string->number (xcons:length  r)))
                            (port->fdes s))) ;; XXX port returns falsity
        (error 'X (c-string->deprecated-string r)))
      (begin
        (set! r (read-bytes (* 4 (+ (byte-string->number (xcons:length r))))
                            s))
        (xcons:setup s r)))
  )))
)) ; closes namespace xcons


(define xserv (namespace (namespace-import sca/x-common)

;; X Server Extensions (p. 152)
(define X-rp-list-extensions
 (byte-structure-description
  (opcode  1)
  (padding 1)
  (request-length 2)))
(define-bs-accessors 'xrp-list-extensions X-rp-list-extensions)
(define X-list-extensions:template
 (let ((t (byte-string-clear!
           (make-byte-string (bs-length X-rp-list-extensions))) ))
  (xrp-list-extensions:set-opcode!         t (byte->byte-string 99))
  (xrp-list-extensions:set-request-length! t (number->byte-string-2 1))
  t))

(define (split-strings byte-string offset)
 (cond
  ((= offset (byte-string-length byte-string)) '() )
  (else (let ((l (byte-string->number
                  (byte-string-select byte-string offset 1))))
   (cons ;(c-string->deprecated-string
         (byte-string->brass ;; XXX official encoding is latin-1 ...
             (byte-string-select byte-string (+ offset 1) l))
         (split-strings byte-string (+ offset 1 l)))))))

(define (X-list-extensions connection)
 (X-request connection X-list-extensions:template (lambda (con fbw)
  (let* ((r fbw)
         (opcode       (byte-string->number (byte-string-select r 0 1)))
         (num-strings  (byte-string->number (byte-string-select r 1 1)))
         (seq-number   (byte-string->number (byte-string-select r 2 2)))
         (reply-length (* 4 (byte-string->number (byte-string-select r 4 4))))
         (tail         (X-receive reply-length con)))
    ;; TODO use read-out-list
    (split-strings tail 0) ))
))


;; X Server Fonts (see also font.class.scm)
  ; ListFonts TODO ListFontsWithInfo
  ; SetFontPath GetFontPath
 (define X-rp-list-fonts
  (byte-structure-description
   (opcode  1)
   (padding 1)
   (request-length 2)
   (max-names      2)
   (length-of-pattern 2)))
 (define-bs-accessors 'xrp-list-fonts X-rp-list-fonts)
(define X-rr-list-fonts
  (byte-structure-description
   (reply   1)
   (padding 1)
   (sequence-number 2)
   (reply-length    4)
   (num-strings     2)
   (padding 22)))
 (define-bs-accessors 'xrr-list-fonts X-rr-list-fonts)
 (define X-list-fonts:template
  (let ((t (byte-string-clear! (make-byte-string (bs-length X-rp-list-fonts)))))
    (xrp-list-fonts:set-opcode!    t (byte->byte-string 49))
    (xrp-list-fonts:set-max-names! t (number->byte-string-2 65535))
    t))
 (define (X-list-fonts connection pattern)
  (let* ((pl (brass-length pattern)) ;; TODO verify C0
         (rb X-list-fonts:template)
         (bp (brass->string-8 pattern))
         (rl (+ (byte-string-length rb) (byte-string-length bp)))
         (pad (byte-string-clear! (make-byte-string (- 4 (modulo rl 4))))))
    ;(REPL-HERE)
   (xrp-list-fonts:set-length-of-pattern! rb (number->byte-string-2 pl))
   (xrp-list-fonts:set-request-length!    rb (number->byte-string-2 (/ (+ rl (byte-string-length pad)) 4)))
   (X-request connection (byte-string-adjoin rb bp pad) (lambda (con fbw)
     (let* ((r fbw)
            (seq-number   (byte-string->number      (xrr-list-fonts:sequence-number r)))
            (reply-length (* 4 (byte-string->number (xrr-list-fonts:reply-length r))))
            (num-strings  (byte-string->number      (xrr-list-fonts:num-strings r)))
            (tail         (X-receive reply-length con)))
    ;; TODO use read-out-list with reactivad element-size to avoid copying overhead
    (split-strings tail 0) )))
 ))

(define X-rp-ungrab-pointer
  (byte-structure-description
   (opcode   1)
   (padding  1)
   (request-length 2)
   (time     4)))
(define-bs-accessors 'xrp-ungrab-pointer X-rp-ungrab-pointer)
(define X-ungrab-pointer:template
 (let ((template (make-byte-string (bs-length X-rp-ungrab-pointer))))
  (byte-string-clear! template)
  (xrp-ungrab-pointer:set-opcode!         template (byte->byte-string    27))
  (xrp-ungrab-pointer:set-request-length! template (number->byte-string-2 2))
  template))
(define (X-ungrab-pointer connection time)
 (let ((rb X-ungrab-pointer:template))
  (xrp-ungrab-pointer:set-time! rb (number->byte-string-4 time))
  (X-request connection rb no-response)))
   
)) ; closes namespace xserv
)) ; closes display-class
