
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; DISPLAY These should be inner classes of the display-class

(define screen-class (class (extension display-object-class)
 (slots xdetail)
 (slots root-window)
 (slots depths default-colormap)
 (slots fallback-picture-format) ; render extension

 (method initialize (display xdetail)
  (this (set! 'display display))
  (this (set! 'fallback-picture-format #f))
  (this (set! 'xdetail xdetail))
  (let* ((rid    (assq-ref xdetail 'root-window))
         (depths (copy-tree (assq-ref xdetail 'depths)))
         (colormap-id (assq-ref xdetail 'default-colormap))
         (root   ((display riddle) (lookup rid))))
   (cond
    (root (X-log-warning "SCA/X root window of screen already mapped") ;; shouldn't happen
          (this (set! 'root-window root)))
    (else (this (set! 'root-window (make-object window-class display rid))) ))

   ;; Reflect the data graph of all depths and visuals for a screen.
   ;; If this doesn't work throw it away and write it anew.
   (this (set! 'depths
     (map (lambda (d)
            (cons (cons 'depth (assq-ref d 'depth))
                  (map (lambda (vdetail)
                        (let ((vid (assq-ref vdetail 'visual-id)))
                          ;; Or we already know the rid or we have to
                          ;; create a new object for that rid.
                          (or ((display riddle) (lookup vid))
                              (make-object visual-class display vdetail))))
                       (assq-ref d 'visuals))))
        depths)))
   (this (set! 'default-colormap (make-object colormap-class display colormap-id)))
 ))
 (method describe ()
  (let ((indention "  ")
        (newline   "\n"))
   (apply brass-append
   (list "Screen " newline
         indention (format "Fallback Picture Format: ~A" (and (this fallback-picture-format) ((this fallback-picture-format) (describe))))
         newline
         (apply brass-append (map (lambda (l) (brass-append indention l newline))   ;; Maybe this recursion and its indention are better done by another recursing introspection printer?
                                  (flatten! (map describe-depth (this depths)) )))
         indention ((this default-colormap) (describe))
)) ))

 (define (describe-depth d)
   (let ((dd (format "Depth: ~A " (cdar d))))
     (map (lambda (visual) (brass-append dd (visual (describe)))) (cdr d))
 ))
)) ; closes screen class

