
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; DISPLAY These should be inner classes of the display-class

(define visual-class  (class (extension display-resource-class)
 (slots xdetail)
 (slots type channel-bits colormap-entries)
 (slots red-mask green-mask blue-mask)
 (slots color->pixel)
 (slots picture-format) ; X render extension

 (method initialize (display xdetail)
  (this (set! 'display display))
  (this (set! 'picture-format #f))
  (this (set! 'xdetail xdetail))
  (this (set! 'channel-bits (assq-ref xdetail 'channel-bits)))
  (this (set! 'colormap-entries (assq-ref xdetail 'colormap-entries)))
  (this (set! 'red-mask     (assq-ref xdetail 'red-mask)))
  (this (set! 'green-mask   (assq-ref xdetail 'green-mask)))
  (this (set! 'blue-mask    (assq-ref xdetail 'blue-mask)))
  (this (set! 'type   (decode-visual-class (assq-ref xdetail 'class))))

  (let ((vid (assq-ref xdetail 'visual-id)))
   (this (set! 'rid vid)) ((display riddle) (register-id vid this)))

  ;; XXX this should be subclassed for every supported visual type?
  (cond
   ((and (eq? (this type) 'true-color))
         ;(= channel-bits 8))
      (this (set! 'color->pixel (compute:color->pixel (this red-mask) (this green-mask) (this blue-mask)) )))
   (else
      (this (set! 'color->pixel (lambda arguments
              (error (format "GIO SCA/X visual-class color->pixel unsupported visual type ~A" (this type))) )))))
      
 ) ; closes method initialize

;; TODO review Xft for a list of common visual encodings
 (define (bgr? r g b)
   (and (equal? r #"0 0 255 0")
        (equal? g #"0 255 0 0")
        (equal? b #"255 0 0 0")))
 (define (compute:color->pixel r g b)
  (cond ((bgr? r g b) (lambda (color) (byte-string (max 0 (min 255 (floor (* -255 (color:blue  color)))))
                                                   (max 0 (min 255 (floor (* -255 (color:green color)))))
                                                   (max 0 (min 255 (floor (* -255 (color:red   color)))))
                                                   0)))
        (else         (lambda (color) (error "GIO SCA/X visual-class color->pixel unsupported bit masks")))))


 (method describe ()
  (let ((nl "\n")) (brass-append
   (format "Visual ~A (Type ~A)" (this (get-sid)) (this type))                      nl
   (format " R~A       channel-bits ~A " (this red-mask) (this channel-bits))       nl
   (format " G~A   colormap-entries ~A"  (this green-mask) (this colormap-entries)) nl
   (format " B~A" (this blue-mask))                                                 nl
   (format " Picture Format: ~A" (and (this picture-format) ((this picture-format) (describe)))) nl
 )))

 (method color->pixel (color) ((this color->pixel) color))
 (method get-color->pixel:converter () (this color->pixel))

 ; Delegation 
 (method color->picsel (color alfa)     ((this picture-format) (color->picsel color alfa)))
 (method get-color->picsel:converter () ((this picture-format) (get-color->picsel:converter color->picsel)))

 (define (decode-visual-class bs)
  (case (byte-string->number bs)
   ((0) 'static-gray)
   ((1) 'gray-scale)
   ((2) 'static-color)
   ((3) 'pseudo-color)
   ((4) 'true-color)
   ((5) 'direct-color)
   (else (X-log-debug (format "SCA/X Unsupported Visual Class: ~A" bs))
         'unsupported)))

)) ; closes visual class
