
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; DISPLAY These should be inner classes of the display-class
;; X Extension

;; TOC
;; xext render class
;;   inner class: picture format class
;; xrender namespace


(define xext-render-class (class                              (extension x-extension-class)
  (slots major-version minor-version formats)
(method initialize (display) (super display "RENDER")
 (let ((ver (call xrender 'query-version (display connection) (this major-opcode)))
       (riddle (display riddle))
       (rr  (display rr)))
  (this (set! 'major-version (first  ver)))
  (this (set! 'minor-version (second ver)))
  (X-log-debug "X Render extension first error: " (this first-error))
  (let ((formats (call xrender 'query-pict-formats (display connection) (this major-opcode))))
   (this (set! 'formats (map (lambda (fb) (make-object picture-format-class fb display))
                             (assq-ref formats 'picture-formats-bytes))))
   (for-each ; screen
       (lambda (ps ds)
         (ds (set! 'fallback-picture-format (riddle (lookup (assq-ref ps 'fallback-picture-format-rid))) ))
         (for-each ; depth
             (lambda (d) (for-each ; visual picture format extension
                             (lambda (vp) (let ((v (riddle (lookup (assq-ref vp 'visual-rid)) ))
                                                (p (riddle (lookup (assq-ref vp 'picture-format-rid))) ))
                                            (when (v picture-format) (log-warning "VSI GIO xext render class init: visual-picture format collision"))
                                            (v (set! 'picture-format p)) ))
                             (assq-ref d 'picture-visuals)))
             (assq-ref ps 'depths)))
       (assq-ref formats 'picture-screens)
       (display screens))
     ; subpixel hint   -> to not be used if possible
   )
))

(define picture-format-class (class                         (extension display-resource-class)
 (slots type depth direct-format colormap)
 (slots color->picsel)

 (define direct-format-layout (struct red green blue red-mask green-mask blue-mask alfa alfa-mask))
 (define (decode-direct-format bytes)
   (struct-from-values direct-format-layout (byte-string->number (fixed-call xrender 'direct-format:red        bytes))
                                            (byte-string->number (fixed-call xrender 'direct-format:green      bytes))
                                            (byte-string->number (fixed-call xrender 'direct-format:blue       bytes))
                                            #! identity !#       (fixed-call xrender 'direct-format:red-mask   bytes)
                                            #! identity !#       (fixed-call xrender 'direct-format:green-mask bytes)
                                            #! identity !#       (fixed-call xrender 'direct-format:blue-mask  bytes)
                                            (byte-string->number (fixed-call xrender 'direct-format:alfa       bytes))
                                            #! identity !#       (fixed-call xrender 'direct-format:alfa-mask  bytes)))
 
 (method initialize (bytes display) ; no call to super here; this needs to map a rid
  ;; init display object
   (this (set! 'display display))
  ;; init display resource
   (this (set! 'rid  (fixed-call xrender 'pict-form-info:id bytes)))
   ((display riddle) (register-id (this rid) this))

  ;; picture-format
  (this (set! 'type          (case (byte-string->number  (fixed-call xrender 'pict-form-info:type          bytes))
                                   ((0) 'indexed) ((1) 'direct) (else  (byte-string->number  (fixed-call xrender 'pict-form-info:type  bytes))) )))
  (this (set! 'depth         (byte-string->number  (fixed-call xrender 'pict-form-info:depth         bytes)) ))
  (this (set! 'direct-format (decode-direct-format (fixed-call xrender 'pict-form-info:direct-format bytes)) ))
  (this (set! 'colormap      ((display riddle) (lookup (fixed-call xrender 'pict-form-info:colormap  bytes))) ))

  (when (and (not (this colormap))
             (byte-string-true? (fixed-call xrender 'pict-form-info:colormap  bytes)))
    (X-log-warn "VSI GIO picture format class init: colormap not found "
              (fixed-call xrender 'pict-form-info:colormap  bytes)))

  (when (not (eq? 'direct (this type)))
    (X-log-info "VSI GIO picture format class init: unsupported type (non direct)"))


  (let ((pf this))
  (cond
   ((and pf (eq? (pf type) 'direct))
      (this (set! 'color->picsel (compute:color->picsel (slot-ref (pf direct-format) 'red-mask)
                                                        (slot-ref (pf direct-format) 'green-mask)
                                                        (slot-ref (pf direct-format) 'blue-mask)
                                                        (slot-ref (pf direct-format) 'alfa-mask))) ))
   (else
      (this (set! 'color->picsel (lambda arguments
              (error (format "GIO SCA/X visual-class color->picsel unsupported picture format type")) )) ))
   ))
 ) ; closes method initialize

 (method color->picsel (color alfa) ((this color->picsel) color alfa))
 (method get-color->picsel:converter () (this color->picsel))

 (method describe ()
  (format (brass-append "* Picture Format (type ~A; depth ~A)\n"
                        "   Direct Format: ~A\n")
          (this type) (this depth) (this direct-format))
 )


 (define (byte-rgba? r g b a) (equal? #"255 0" r g b))
 (define (compute:color->picsel r g b a) ; x render pixel format
  (cond ((byte-rgba? r g b a) (lambda (color alfa)
                             (byte-string (max 0 (min 255 (floor (* -255 (color:blue  color)))))
                                          0
                                          (max 0 (min 255 (floor (* -255 (color:green color)))))
                                          0
                                          (max 0 (min 255 (floor (* -255 (color:red   color)))))
                                          0
                                          (max 0 (min 255 (floor (* 255 alfa))))
                                          0)))
        (else         (lambda arguments (error "GIO SCA/X visual-class color->picsel unsupported bit masks")))))

)) ; closes picture format class
)) ; closes xext render class
  

(define xrender (namespace (namespace-import sca/x-common)

(define rp-query-version
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (major-version  4)
  (minor-version  4)))
(define-bs-accessors 'rp-query-version rp-query-version)
(define rr-query-version
 (byte-structure-description
  (reply           1)
  (padding         1)
  (sequence-number 2)
  (request-length  4)
  (major-version   4)
  (minor-version   4)
  (padding        16)))
(define-bs-accessors 'rr-query-version rr-query-version)
(define query-version:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length rp-query-version)))))
  (rp-query-version:set-minor-opcode!   template (byte->byte-string     0))
  (rp-query-version:set-request-length! template (number->byte-string-2 3))
  template))
(define (query-version connection major-opcode)
 (let ((rb query-version:template))
  (rp-query-version:set-major-opcode!  rb (byte->byte-string major-opcode))
  (rp-query-version:set-major-version! rb (number->byte-string-4        0))
  (rp-query-version:set-minor-version! rb (number->byte-string-4       11))
  (X-request connection rb (lambda (con fbw)
   (values (byte-string->number (rr-query-version:major-version fbw))
           (byte-string->number (rr-query-version:minor-version fbw)) )))
))

(define direct-format-layout
 (byte-structure-description
  (red   2) (red-mask   2)
  (green 2) (green-mask 2)
  (blue  2) (blue-mask  2)
  (alfa  2) (alfa-mask  2)))
(define-bs-accessors 'direct-format direct-format-layout)

(define pict-form-info-layout
 (byte-structure-description
  (id             4)
  (type           1)
  (depth          1)
  (padding        2)
  (direct-format 16) ; bs-length direct-format-layout
  (colormap       4)))
(define-bs-accessors 'pict-form-info pict-form-info-layout)
(define (decode-pict-form-info bytes)
 (values (once (bs-length pict-form-info-layout))
         (byte-string-select bytes 0 (once (bs-length pict-form-info-layout))) ;; bytes go to pict form class init
))

(define pict-visual-layout
 (byte-structure-description
  (visual-id   4)
  (pict-format 4)))
(define-bs-accessors 'pict-visual pict-visual-layout)
(define (decode-pict-visual bytes)
 (values 8 (list (cons 'visual-rid         (pict-visual:visual-id bytes))
                 (cons 'picture-format-rid (pict-visual:pict-format bytes)) )))

(define pict-depth-layout
 (byte-structure-description
  (depth            1)
  (padding          1)
  (num-pict-visuals 2)
  (padding          4)))
(define-bs-accessors 'pict-depth pict-depth-layout)
(define (decode-pict-depth bytes)
 (let* ((len   (byte-string->number (pict-depth:num-pict-visuals bytes)))
        (nb+pv (X:read-out-list len decode-pict-visual #f
                                bytes (once (bs-length pict-depth-layout))) ))
  (values (+ (once (bs-length pict-depth-layout)) (first nb+pv))
          (list (cons 'depth (byte-string->number (pict-depth:depth bytes)))
                (cons 'picture-visuals (second nb+pv)) ))
))

(define pict-screen-layout
 (byte-structure-description
  (num-depths           4)
  (fallback-pict-format 4)))
(define-bs-accessors 'pict-screen pict-screen-layout)
(define (decode-pict-screen bytes)
 (let* ((len   (byte-string->number (pict-screen:num-depths bytes)))
        (nb+dl (X:read-out-list len decode-pict-depth #f
                                bytes (once (bs-length pict-screen-layout))) ))
   (values (+ (once (bs-length pict-screen-layout)) (first nb+dl))
           (list (cons 'fallback-picture-format-rid (pict-screen:fallback-pict-format bytes))
                 (cons 'depths (second nb+dl)) ))
))

(define subpixel-codes '(
 (subpixel-unknown        . 0)
 (subpixel-horizontal-rgb . 1)
 (subpixel-horizontal-bgr . 2)
 (subpixel-vertical-rgb   . 3)
 (subpixel-vertical-bgr   . 4)
 (subpixel-none           . 5)))
;; XXX TODO resolve local symbol validity / numerical code associations
(define subpixel-codes-i (map (lambda (a) (cons (cdr a) (car a))) subpixel-codes))

(define rp-query-pict-formats
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)))
(define-bs-accessors 'rp-query-pict-formats rp-query-pict-formats)
(define rr-query-pict-formats
 (byte-structure-description
  (reply           1)
  (padding         1)
  (sequence-number 2)
  (reply-length    4)
  (num-formats     4)
  (num-screens     4)
  (num-depths      4)
  (num-visuals     4)
  (num-subpixels   4)
  (padding         4)))
(define-bs-accessors 'rr-query-pict-formats rr-query-pict-formats)
(define query-pict:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length rp-query-pict-formats)))))
  (rp-query-pict-formats:set-minor-opcode!   t (byte->byte-string     1))
  (rp-query-pict-formats:set-request-length! t (number->byte-string-2 1))
  t))
(define (query-pict-formats connection major-opcode)
 (let ((rb query-pict:template))
  (rp-query-pict-formats:set-major-opcode! rb (byte->byte-string major-opcode))
  (X-request connection rb (lambda (con fbw)
   (let* ((reply-length  (* 4 (byte-string->number (rr-query-pict-formats:reply-length  fbw))))
          (num-formats   (byte-string->number      (rr-query-pict-formats:num-formats   fbw)))
          (num-screens   (byte-string->number      (rr-query-pict-formats:num-screens   fbw)))
          (num-depths    (byte-string->number      (rr-query-pict-formats:num-depths    fbw)))
          (num-visuals   (byte-string->number      (rr-query-pict-formats:num-visuals   fbw)))
          (num-subpixels (byte-string->number      (rr-query-pict-formats:num-subpixels fbw)))
          (tail         (X-receive reply-length connection))
          (n+fs          (X:read-out-list num-formats decode-pict-form-info #f tail 0))
          (n+scrs        (X:read-out-list num-screens decode-pict-screen    #f tail (first n+fs)))
          ;(subpixel      (assq-ref subpixel-codes-i (byte-string->number (byte-string-select tail (first n+scrs) 4))) ))
          (subpixel      (byte-string->number (byte-string-select tail (+ (first n+fs) (first n+scrs)) 4)) ))
     (list (cons 'picture-formats-bytes (second n+fs))
           (cons 'picture-screens       (second n+scrs))
           (cons 'subpixel-hint         subpixel)) )))
))

(define rp-create-picture
 (byte-structure-description
  (major-opcode     1)
  (minor-opcode     1)
  (request-length   2)
  (picture-rid      4)
  (drawable-rid     4)
  (pict-format-rid  4)
  (value-mask       4)))
(define-bs-accessors 'rp-create-picture rp-create-picture)
(define create-picture:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length rp-create-picture)))))
  (rp-create-picture:set-minor-opcode! t (byte->byte-string 4))
  t))
(define (create-picture connection major-opcode picture-rid drawable-rid pict-format-rid)
 (let ((rb create-picture:template))
  (rp-create-picture:set-major-opcode!    rb (byte->byte-string major-opcode))
  (rp-create-picture:set-picture-rid!     rb picture-rid)
  (rp-create-picture:set-drawable-rid!    rb drawable-rid)
  (rp-create-picture:set-pict-format-rid! rb pict-format-rid)
  (rp-create-picture:set-value-mask!      rb #"0 0 0 0")
  (rp-create-picture:set-request-length!  rb (number->byte-string-2 (once (/ (bs-length rp-create-picture) 4))))
  (X-request connection rb no-response)
))

(define direct-picsel-size (* 2 4))

(define rp-create-solid-fill
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (picture-rid    4)))
  ;(picsel         8))) ; appended by the function request mapper
(define-bs-accessors 'rp-create-solid-fill rp-create-solid-fill)
(define create-solid-fill:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length rp-create-solid-fill))) ))
  (rp-create-solid-fill:set-minor-opcode!   t (byte->byte-string    33))
  (rp-create-solid-fill:set-request-length! t (number->byte-string-2 4))
  t))
(define (create-solid-fill connection major-opcode picture-rid picsel)
 (let ((rb create-solid-fill:template))
  (rp-create-solid-fill:set-major-opcode! rb (byte->byte-string major-opcode))
  (rp-create-solid-fill:set-picture-rid!  rb picture-rid)
  (X-request connection (byte-string-adjoin rb picsel) no-response)
))
  
;;
;; Picture Properties

(define (pc bitpos) (bit-shift-left (number->byte-string-4 1) bitpos))
(define change-picture:repeat         (pc  0))

(define change-picture:alfa-map       (pc  1))
(define change-picture:alfa-x-origin  (pc  2))
(define change-picture:alfa-y-origin  (pc  3))

(define change-picture:clip-x-origin  (pc  4))
(define change-picture:clip-y-origin  (pc  5))
(define change-picture:clip-mask      (pc  6))

; (define change-picture:graphics-exposure (pc 7)) " are never generated by this extension " 
(define change-picture:subwindow-mode (pc  8))

(define change-picture:poliedge       (pc  9))
(define change-picture:polimode       (pc 10))
; (define change-picture:dither         (pc 11)) ; " is ignored "
(define change-picture:component-alfa (pc 12))

(define none #"0 0 0 0")

(define rp-change-picture
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (picture-rid    4)
  (value-mask     4)))
(define-bs-accessors 'rp-change-picture rp-change-picture)
(define change-picture:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length rp-change-picture))) ))
  (rp-change-picture:set-minor-opcode! t (byte->byte-string 5))
  t))

(define change-picture-alfa (lambda (connection major-opcode picture-rid
                                     alfa-map-rid alfa-x-origin alfa-y-origin)
 (let ((rb change-picture:template))
  (rp-change-picture:set-major-opcode! rb (byte->byte-string major-opcode))
  (rp-change-picture:set-value-mask!   rb (bit-or change-picture:alfa-map
                                                  change-picture:alfa-x-origin
                                                  change-picture:alfa-y-origin))
  (rp-change-picture:set-request-length! rb
    (once (number->byte-string-2 (/ (+ (bs-length rp-change-picture) 4 2 2)
                                    4)) ))
  (X-request (byte-string-adjoin rb alfa-map-rid
                                    (number+sign->byte-string-2 alfa-x-origin)
                                    (number+sign->byte-string-2 alfa-y-origin))
             no-response)
)))

;; While libxrender/src/Picture.c uses /unsigned long/ to process the
;; attributes there's some additional magic Data32 which seems to keep
;; things in the bounds of the 32 bit age. Whatever encoding the modi
;; as 4 byte quantities works.
;; The specification as well as the header files don't seem to
;; specify the actual encoding when it comes to byte quantities.
;; Welcome to the age of the C obfuscation layer on our historical
;; journey of graphical input output.

(define (encode-poliedge ed)
 (number->byte-string-4
 (case ed ((sharp) 0) ((smooth) 1)
       (else (error "GIO SCA/X xrender encode-poliedge: unknown edge style")))
))
(define (encode-polimode md)
 (number->byte-string-4
 (case md ((precise) 0) ((imprecise) 1)
       (else (error "GIO SCA/X xrender encode-polimode: unknown mode")))
))
(define (encode-repeat rm)
 (number->byte-string-4
 (case rm ((none) 0) ((normal) 1) ((pad) 2) ((reflect) 3)
       (else (error "GIO SCA/X xrender encode-repeat: unknown repetition mode")))
))
       

(define change-picture-draw-mode (lambda (connection major-opcode picture-rid
                                                     poliedge polimode repeat)
 (let* ((rb  change-picture:template)
        (rl  (+ (bs-length rp-change-picture) 4 4 4))
        (pad (X-pad rl)))
  (rp-change-picture:set-major-opcode! rb (byte->byte-string major-opcode))
  (rp-change-picture:set-picture-rid!  rb picture-rid)
  (rp-change-picture:set-value-mask!   rb (bit-or change-picture:repeat
                                                  change-picture:poliedge
                                                  change-picture:polimode))
  (rp-change-picture:set-request-length! rb
    (number->byte-string-2 (/ (+ rl (byte-string-length pad))
                              4)))
  (X-request connection
             (byte-string-adjoin rb (encode-repeat   repeat)
                                    (encode-poliedge poliedge)
                                    (encode-polimode polimode)
                                    pad)
             no-response)
)))


(define rp-triangles
 (byte-structure-description
  (major-opcode    1)
  (minor-opcode    1)
  (core-request-length 2)
  (request-length  4)  ; big request
  (grafix-opcode   1)
  (padding         3)
  (src-rid         4)
  (dst-rid         4)
  (mask-format-rid 4)
  (src-x           2)
  (src-y           2)))
(define-bs-accessors 'rp-triangles rp-triangles)
(define triangles:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length rp-triangles)))))
  (rp-triangles:set-minor-opcode! t (byte->byte-string 11))
  t))

;; Map a scalar from 0 to 1 onto the 16 bit fixed point ranges
(define (encode-normalized-scalar s)
 (let* ((s1  (* s (once (- (expt 2 16) 1)) ))
        (int (inexact->exact (truncate s1)))
        (fra (inexact->exact (truncate (* (- s1 int) (once (- (expt 2 16) 1)))) )))
  (bit-or (bit-shift-left (number->byte-string-4 int) 16)
          (number->byte-string-4 fra)) ))

(define (encode-point p)
 (byte-string-adjoin (encode-normalized-scalar (vector-ref p 0))
                     (encode-normalized-scalar (vector-ref p 1))))
 
(define (triangles connection major-opcode grafix-opcode source-rid dest-rid mask-format-rid src-x src-y . points)
 (let ((rb  triangles:template)
       (esc (map encode-point points)))
  (rp-triangles:set-major-opcode!    rb (byte->byte-string major-opcode))
  (rp-triangles:set-grafix-opcode!   rb (byte->byte-string grafix-opcode))
  (rp-triangles:set-src-rid!         rb source-rid)
  (rp-triangles:set-dst-rid!         rb dest-rid)
  (rp-triangles:set-mask-format-rid! rb mask-format-rid)
  (rp-triangles:set-src-x!           rb (number+sign->byte-string-2 src-x))
  (rp-triangles:set-src-y!           rb (number+sign->byte-string-2 src-y))
  (set! rb (apply byte-string-adjoin rb esc))
  (byte-string-set! rb (once (first (bs:resolve rp-triangles 'request-length))) ; we revert to the inerts of the current byte structure implementation because regular setters assume the length of the byte strings equals the length of their byte structure and we don't want to lose that check.
                    (number->byte-string-4 (/ (byte-string-length rb) 4)))
  (X-request connection rb no-response)
))

(define (encode-rectangle r) ; Xlib.h XRectangle from list notation: (x y width height)
 (byte-string-adjoin (number+sign->byte-string-2 (first  r))
                     (number+sign->byte-string-2 (second r))
                     (number->byte-string-2      (third  r))
                     (number->byte-string-2      (fourth r))))

(define rp-set-picture-clip-rectangles
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (picture-rid    4)))
(define-bs-accessors 'rp-set-picture-clip-rectangles rp-set-picture-clip-rectangles)
(define set-picture-clip-rectangles:template
 (let ((t (byte-string-clear! (make-byte-string (bs-length rp-set-picture-clip-rectangles))) ))
  (rp-set-picture-clip-rectangles:set-minor-opcode! t (byte->byte-string 6))
  t))
(define (set-picture-clip-rectangles connection major-opcode picture-rid . rectangles)
 (let ((rb  set-picture-clip-rectangles:template)
       (ers (map encode-rectangle rectangles) ))
  (rp-set-picture-clip-rectangles:set-major-opcode! rb (byte->byte-string major-opcode))
  (rp-set-picture-clip-rectangles:set-picture-rid!  rb picture-rid)
  (set! rb (apply byte-string-adjoin rb ers))
  (byte-string-set! rb (once (first (bs:resolve rp-set-picture-clip-rectangles 'request-length))) ; we revert to the inerts of the current byte structure implementation because regular setters assume the length of the byte strings equals the length of their byte structure and we don't want to lose that check.
                    (number->byte-string-4 (/ (byte-string-length rb) 4)))
  (X-request connection rb no-response)
))

)) ; closes xrender namespace
