
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; DISPLAY These should be inner classes of the display-class
;; X Extension


;; The security extension isn't usually needed.
(define xext-security-class (class     (extension x-extension-class)

(method initialize (display auth-name-bytes auth-data-bytes) (super display "SECURITY")
  (call xsec 'Xsec-query-version (display connection) (this major-opcode))
  (call xsec 'Xsec-generate-authorization (display connection) (this major-opcode)
                                          auth-name-bytes auth-data-bytes)
)

(define xsec (namespace (namespace-import sca/x-common)
(define Xsec-rp-query-version
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (major-version  2)
  (minor-version  2)))
(define-bs-accessors 'xsec-rp-query-version Xsec-rp-query-version)
(define Xsec-query-version:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length Xsec-rp-query-version)))))
  (xsec-rp-query-version:set-minor-opcode!   template (byte->byte-string     0))
  (xsec-rp-query-version:set-request-length! template (number->byte-string-2 2))
  template))
(define Xsec-rr-query-version
 (byte-structure-description
  (type            1)
  (padding         1)
  (sequence-number 2)
  (reply-length    4)
  (major-version   2)
  (minor-version   2)
  (padding        20)))
(define-bs-accessors 'xsec-rr-query-version Xsec-rr-query-version)
(define (Xsec-query-version connection major-opcode)
 (let ((rb Xsec-query-version:template))
   (xsec-rp-query-version:set-major-opcode!  rb (byte->byte-string     major-opcode))
   (xsec-rp-query-version:set-major-version! rb (number->byte-string-2 1))
   (xsec-rp-query-version:set-minor-version! rb (number->byte-string-2 0))
   (X-request connection rb (lambda (con fbw) fbw))))

(define pc number->byte-string-4)
(define xsec-timeout     (pc 0))
(define xsec-trust-level (pc 2))
(define xsec-group       (pc 4))
(define xsec-event-mask  (pc 8))

(define xsec-trusted     (pc 0))
(define xsec-untrusted   (pc 1))

(define Xsec-rp-generate-authorization
 (byte-structure-description
  (major-opcode      1)
  (minor-opcode      1)
  (request-length    2)
  (nbytes-auth-proto 2)
  (nbytes-auth-data  2)
  (value-mask        4)))
(define-bs-accessors 'xsec-rp-generate-authorization Xsec-rp-generate-authorization)
(define Xsec-rr-generate-authorization
 (byte-structure-description
  (reply            1)
  (padding          1)
  (sequence-number  2)
  (reply-length     4)
  (auth-id          4)
  (data-length      2)
  (padding         12)))
(define-bs-accessors 'xsec-rr-generate-authorization Xsec-rr-generate-authorization)
(define Xsec-generate-authorization:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length Xsec-rp-generate-authorization)))))
  (xsec-rp-generate-authorization:set-minor-opcode! template (byte->byte-string 1))
  template))
(define (Xsec-generate-authorization connection major-opcode auth-name auth-data)
 (let* ((rb        Xsec-generate-authorization:template)
        (auth-name (brass->c-string "MIT-MAGIC-COOKIE-1"))
        (pad1      (X-pad (+ (byte-string-length auth-name) (byte-string-length auth-data)))))
   (xsec-rp-generate-authorization:set-major-opcode! rb (byte->byte-string major-opcode))
   (xsec-rp-generate-authorization:set-value-mask!   rb (bit-or xsec-timeout
                                                                xsec-trust-level))
                                                                ;xsec-event-mask))
   (xsec-rp-generate-authorization:set-nbytes-auth-proto! rb (number->byte-string-2 (byte-string-length auth-name)))
   (xsec-rp-generate-authorization:set-nbytes-auth-data!  rb (number->byte-string-2 (byte-string-length auth-data)))
   (xsec-rp-generate-authorization:set-request-length!    rb (number->byte-string-2 (+ 3
                                                               (/ (apply + (map byte-string-length
                                                                                (list rb pad1 auth-name auth-data)))
                                                                  4))))
   
   (X-request connection (byte-string-adjoin rb auth-name auth-data pad1
                                             #"0 0 0 0"     ; no timeout
                                             xsec-trusted
                                             #"255 255 255 255") ; event-mask
              (lambda (con fbw) (X-receive (* 4 (byte-string->number (xsec-rr-generate-authorization:reply-length fbw)))
                                           con) #t))
))
; TODO
)) ; closes xsec namespace
)) ; closes xext security class

