
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;; DISPLAY These should be inner classes of the display-class

;; SHARED MEMORY Data Flow / IPC / Transport Layer
;; Depends on x-extensions/xext-shared-memory.class


(define x-shared-memory-class (class (extension display-resource-class)
 (slots shm-id shm-mapping)

(method initialize (display num-bytes) (super display)
 (let ((shs (sys-control (create-shared-segment num-bytes))))
   (this (set! 'shm-id      shs))
   (this (set! 'shm-mapping (sys-control (shared-mapping shs num-bytes))))
   (cond
   ((not (byte-string-mapping? (this shm-mapping)))
     (log-error "GIO x-shared-memory-class initialize: shared memory allocation failed"))
   (else
     (fixed-call xshm 'Xshm-attach (display connection)
                                   ((display xext-shared-memory) major-opcode)
                                   (this rid)
                                   (this shm-id)) ))
))

(method describe ()
  (format (brass-append "* X Shared Memory Segment \n"
                        "  Number of Bytes: ~A \n"
                        "              RID: ~A \n"
                        "               ID: ~A \n"
                        "          Address: ~A \n")
                        (byte-string-length  (this shm-mapping))
                        (this rid)
                        (this shm-id)
                        (byte-string-address (this shm-mapping)))
)

(method destroy ()
 (let ((display (this display)))
  (fixed-call xshm 'Xshm-detach (display connection) ((display xext-shared-memory) major-opcode) (this rid))
  (super)
  (sys-control (unshare-mapping (this shm-mapping)))
  ;; TODO error handling
  (sys-control (destroy-shared-segment (this shm-id)))
))

)) ; closes x shared memory class



(define xshm (namespace (namespace-import sca/x-common)
(define Xshm-rp-create-segment ; most idiotic request ever: the non-sharable shared whatnot LOL
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (shm-rid        4)
  (size           4)
  (read-only      1)
  (padding        3)))
(define-bs-accessors 'xshm-rp-create-segment Xshm-rp-create-segment)
(define Xshm-create-segment:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length Xshm-rp-create-segment)))))
  (xshm-rp-create-segment:set-minor-opcode!   template (byte->byte-string     7))
  (xshm-rp-create-segment:set-request-length! template (number->byte-string-2 4))
  template))
(define (Xshm-create-segment connection major-opcode shm-rid size read-only)
 (let ((rb Xshm-create-segment:template))
  (xshm-rp-create-segment:set-major-opcode! rb (byte->byte-string major-opcode))
  (xshm-rp-create-segment:set-shm-rid!      rb shm-rid)
  (xshm-rp-create-segment:set-size!         rb (number->byte-string-4 size))
  (xshm-rp-create-segment:set-read-only!    rb (if read-only #"1" #"0"))
  (X-request connection rb no-response)))

(define Xshm-rp-attach
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (xshm-id        4)
  (shm-id         4)
  (read-only      1)
  (padding        3)))
(define-bs-accessors 'xshm-rp-attach Xshm-rp-attach)
(define Xshm-attach:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length Xshm-rp-attach)))))
   (xshm-rp-attach:set-minor-opcode!   template (byte->byte-string     1))
   (xshm-rp-attach:set-request-length! template (number->byte-string-2 4))
   template))
(define (Xshm-attach connection major-opcode shm-rid shm-id)
 (let ((rb Xshm-attach:template))
  (xshm-rp-attach:set-major-opcode! rb (byte->byte-string major-opcode))
  (xshm-rp-attach:set-shm-id!       rb shm-id)
  (xshm-rp-attach:set-xshm-id!      rb shm-rid)
  (xshm-rp-attach:set-read-only!    rb #"0")
  (X-request connection rb no-response)))

(define Xshm-rp-detach
 (byte-structure-description
  (major-opcode   1)
  (minor-opcode   1)
  (request-length 2)
  (xshm-id        4)))
(define-bs-accessors 'xshm-rp-detach Xshm-rp-detach)
(define Xshm-detach:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length Xshm-rp-detach)))))
  (xshm-rp-detach:set-minor-opcode!   template (byte->byte-string     2))
  (xshm-rp-detach:set-request-length! template (number->byte-string-2 2))
  template))
(define (Xshm-detach connection major-opcode shm-rid)
 (let ((rb Xshm-detach:template))
  (xshm-rp-detach:set-major-opcode! rb (byte->byte-string major-opcode))
  (xshm-rp-detach:set-xshm-id!      rb shm-rid)
  (X-request connection rb no-response)))

(define Xshm-rp-create-pixmap
 (byte-structure-description
  (major-opcode     1)
  (minor-opcode     1)
  (request-length   2)
  (pixmap           4)
  (drawable         4)
  (width            2)
  (height           2)
  (depth            1)
  (padding          3)
  (shm-id           4)
  (offset           4)))
(define-bs-accessors 'xshm-rp-create-pixmap Xshm-rp-create-pixmap)
(define Xshm-create-pixmap:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length Xshm-rp-create-pixmap)))))
  (xshm-rp-create-pixmap:set-minor-opcode!   template (byte->byte-string      5))
  (xshm-rp-create-pixmap:set-request-length! template (number->byte-string-2  7))
  template))
(define (Xshm-create-pixmap connection major-opcode pixmap drawable width height depth shm-id offset)
 (let ((rb Xshm-create-pixmap:template))
  (xshm-rp-create-pixmap:set-major-opcode! rb (byte->byte-string major-opcode))
  (xshm-rp-create-pixmap:set-pixmap!       rb pixmap)
  (xshm-rp-create-pixmap:set-drawable!     rb drawable)
  (xshm-rp-create-pixmap:set-width!        rb (number+sign->byte-string-2 width))
  (xshm-rp-create-pixmap:set-height!       rb (number+sign->byte-string-2 height))
  (xshm-rp-create-pixmap:set-depth!        rb (byte->byte-string depth))
  (xshm-rp-create-pixmap:set-shm-id!       rb shm-id)
  (xshm-rp-create-pixmap:set-offset!       rb (number->byte-string-4 offset))
  (X-request connection rb no-response)
))

(define Xshm-rp-put-image
 (byte-structure-description
  (major-opcode     1)
  (minor-opcode     1)
  (request-length   2)
  (drawable         4)
  (gcontext         4)
  (total-width      2)
  (total-height     2)
  (src-x            2)
  (src-y            2)
  (src-width        2)
  (src-height       2)
  (dst-x            2)
  (dst-y            2)
  (depth            1)
  (format           1)
  (send-event       1)
  (padding          1)
  (shm-rid          4)
  (offset           4)))
(define-bs-accessors 'xshm-rp-put-image Xshm-rp-put-image)
(define Xshm-put-image:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length Xshm-rp-put-image)))))
  (xshm-rp-put-image:set-minor-opcode!   template (byte->byte-string      3))
  (xshm-rp-put-image:set-request-length! template (number->byte-string-2 10))
  template))
(define (Xshm-put-image connection major-opcode drawable gcontext total-width total-height
                                                src-x src-y src-width src-height dst-x dst-y
                                                depth format-id do-send-event shm-rid offset)
  (let ((rb Xshm-put-image:template))
   (xshm-rp-put-image:set-major-opcode! rb (byte->byte-string major-opcode))
   (xshm-rp-put-image:set-drawable!     rb drawable)
   (xshm-rp-put-image:set-gcontext!     rb gcontext)
   (xshm-rp-put-image:set-total-width!  rb (number->byte-string-2      total-width))
   (xshm-rp-put-image:set-total-height! rb (number->byte-string-2      total-height))
   (xshm-rp-put-image:set-src-x!        rb (number->byte-string-2      src-x))
   (xshm-rp-put-image:set-src-y!        rb (number->byte-string-2      src-y))
   (xshm-rp-put-image:set-src-width!    rb (number->byte-string-2      src-width))
   (xshm-rp-put-image:set-src-height!   rb (number->byte-string-2      src-height))
   (xshm-rp-put-image:set-dst-x!        rb (number+sign->byte-string-2 dst-x))
   (xshm-rp-put-image:set-dst-y!        rb (number+sign->byte-string-2 dst-y))
   (xshm-rp-put-image:set-depth!        rb (byte->byte-string depth))
   (xshm-rp-put-image:set-format!       rb (byte->byte-string format-id))
   (xshm-rp-put-image:set-send-event!   rb (if do-send-event #"1" #"0"))
   (xshm-rp-put-image:set-shm-rid!      rb shm-rid)
   (xshm-rp-put-image:set-offset!       rb (number->byte-string-4 offset))
   (X-request connection rb no-response) ; TODO review wrt send-event / completion-event
))
)) ; closes xshm namespace


;;
;; References
;; 1) MIT-SHM(The MIT Shared Memory Extension): How the shared memory extension works
;;    /usr/share/doc/x11proto-xext-dev/shm.html
;; 2) libXext - library for common extensions to the X11 protocol (1.3.3)
;; 3) missing in libXext: /usr/include/X11/shmproto.h
;;
