
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;
;          +---------------+
;          |  Taco Shells  |
;          |    Topping    |
;          |  Sour Cream   |
;          +---------------+
;
  
(define menu-class (class  (extension window-class)
 (slots items) ; Have a nice D-List to support keyboard up down ...
 (slots last-y)

 (define margin 6)

 (method initialize (parent x y width height) (super parent x y width height 4)
   (this (set! 'items '()))
   (this (set! 'last-y margin))

   (let ((p ((this display) pointer-device)))
    (p (add-button-listener       this))
    (p (add-motion-enter-listener this))
    (p (add-motion-leave-listener this))
   )
   (this (activate-events expose-event-class mel-event-class))
 )

 (method add-item (menu-item)
  (when (not (is-a? menu-item menu-item-class)) (error "GIO menu add-item wrong-typ-arg"))
  ;; (when (not (eq? this (menu-item parent)))     (error "GIO menu add-item mischiefed item: wrong parent"))
  (this (set! 'items (append! (this items) (list menu-item)))) ;; doubles add-child
  (this (set! 'last-y (+ (this last-y)
                         (menu-item height)
                         0
                         ;(menu-item border-width) )))
                         (* 2 (menu-item border-width)) )))
 )

 (method adjust-dimensions ()
  (let ((widths (map (lambda (c) (c width)) (this items)))) ;; XXX use the width of the line metrics + margin -> accessor in menu-item
   (when (not (null? widths)) (this (set-width (apply max widths))))
   (this (set-height (+ margin margin (this last-y))))
))

 ;;; When the pointer enters a child window that provokes a leave event.
 ;;; Grabbing makes the server send enter/leave events ??? which can
 ;;; provoke a feedback loop. Moving the pointer onto another menu item
 ;;; stops drains that feedback loop. Maybe a regrab (or ChangeGrab???)
 ;;; with handling of implicit enter/leave events ...
 ;;;
 ;;; => ungrab pointer in the menu items
 ;; (method enter-event (event) (X-log-debug (event (describe)))
 ;;  (this             (grab-pointer (event time))))
 ;; (method leave-event (event) (X-log-debug (event (describe)))
 ;;  ((this display) (ungrab-pointer (event time))))
 (method button-event (event)
  (cond ; pointer number importance? we will only provide action requests (like scrolls) by means of the widgets but would the button number make a difference here? do scroll button events grab, too. All those mysteries out there ...                                                                         
  ((eq? 'down (event type))     ((this display) (ungrab-pointer (event time)) ))
 )) 

 (method expose-event (event)
  ; (X-log-debug "menu expose")
  ;; !!! actually every window should handle exposures on their own.
  ;;     The events arrive here only for regions which aren't obscured
  ;;     by the items. The count always seems to be zero (when the
  ;;     regions aren't disjoint).
  (when (= (event count) 0) (this (draw)))
 )

 (method draw ()
  (this (clear))
  ; (for-each (lambda (m) (m (draw))) (this children)) doesn't work well when obscured by items
 )

 (method show () (super)
  (for-each (lambda (m) (m (show))) (this children))
 )
))


(define menu-item-class (class                        (extension window-class)
 (slots action label)
 (slots baseline indention reg-gc highlight-gc)
 (slots is-selected)

 (method initialize (menu label action)
  (when (not (is-a? menu menu-class)) (error "GIO menu-item initialize wrong-typ-arg"))
  ; (this (set! 'exposures 0)) ; XXX debugging; see below
  (this (set! 'action action))
  (this (set! 'label label))
  (let* ((border-width 0)
         (dim (gio-font (line-metrics label)))
         (h-space (round (/ (- (menu width) (first dim)) 2)))
         ;0  1  2  3  4  5  6 
         ;__ __ __ __ __ __ __   integer steps 
         ;  .  .  .  .  .  .     exact halfs 
         ;   .  .  .  .  .  .    biased (regular) round  
         ;.     :     :     .    current round implementation (with overall error reduction) is problematic when it's about maintaining geometric patterns? 
                               ; Perhaps more complex examples might show the odd/even rounding
                               ; approach works as well for rasterization as the biased rounding. 
         ;(dim (gio-font (line-metrics "M")))
         ;(h-space (floor (* 6 (* 1/13 (first dim)))))
         (ascent  (gio-font ascent))
         (v-space (* 1/2 ascent 1/2))
         (v-space-b (floor   v-space))
         (v-space-u (ceiling v-space))
         (v-space (+ v-space-u v-space-b)))
    (super menu 0 (menu last-y)
           (- (menu width) (* 2 border-width)) (+ ascent v-space)
           border-width)
    (this (set! 'indention h-space))
    (this (set! 'baseline (+ ascent v-space-u))))
  (let ((gc1 (make-object graphics-context-class this))
        (gc2 (make-object graphics-context-class this)))
    (gc1 (set-font gio-font))
    (gc1 (set-color 'foreground white))
    (gc1 (set-color 'background vsiblue))
    (this (set! 'reg-gc gc1))
    (gc2 (set-font gio-font))
    (gc2 (set-color 'foreground white))
    (gc2 (set-color 'background black))
    ;(gc2 (set-color 'foreground white))
    ;(gc2 (set-color 'background black))
    (this (set! 'highlight-gc gc2))
  )
  (this (set! 'is-selected #f))
  (menu (add-item this))

  (let ((p ((this display) pointer-device)))
   (p (add-button-listener       this))
   (p (add-motion-enter-listener this))
   (p (add-motion-leave-listener this))
  )
  (this (activate-events expose-event-class))
 )

 (method enter-event  (event)  (this (set! 'is-selected #t)) (this (draw)))
 (method leave-event  (event)  (this (set! 'is-selected #f)) (this (draw)))
 (method button-event (event)
  (cond
  ((and (this is-selected)
        (= 1 (event button))
        (eq? 'up (event type))) ((this action)))
  ;; The following flat menu support kills the pop-up interaction model if
  ;; there isn't any notion of the difference somewhere: if the parent is 
  ;; a pop-up the pointer grab - but then ...
  ;; Every window grabs by default? Isn't there any method against that?
  ;; The following also is redundant with respect to is-selected but it's
  ;; part of the X quirks where is-selected is part of our interaction model.
  ((eq? 'down (event type))     ((this display) (ungrab-pointer (event time))))
 ))

 ;; (slots exposures) ; this should become standard? current simple classes don't handle tiled redraws and
                      ; the delivered event count seems to be different from zero only for those regional
                      ; redraws.
 (method expose-event (event)
  ;(this (set! 'exposures (+ (this exposures) 1)))
  (when (= (event count) 0)
    (this (draw))))

 (method draw ()
  ; (this (clear))
  (let ((gc (if (this is-selected) (this highlight-gc) (this reg-gc))))  ;; put the gc into the menu???
   (gc (image-text-8 (this indention) (this baseline) (this label)))
 ))
)) ; closes menu item class

(define font-menu-item-class (class (extension menu-item-class)
 (method initialize (font . args) (apply super args)
   ((this reg-gc)       (set-font font))
   ((this highlight-gc) (set-font font))
 )
))

