(sys-load "VSI-core/repl.scm")
(sys-load "lib/hashtable.scm")
(sys-load "lib/cee.scm")

;;
;; Auxiliaries
;;

;; CEE override and adaptions
(define value-table (make-hashtable (* 2 1024)))
(define (cpp-definition-symbol+value! dictionary)
 (let loop ((char  (read-char))
            (current-token '()))
   (if (eof-object? char)
     (throw 'unexpected-eof)
     (cond  ; ANS top down sequence conditional
        ((is-unicode-c0-printable? (char->byte char))
         ;; in token
           (loop (read-char)
                 (cons char current-token) ))

        ((not (null? current-token))
         ;; end of token
          (catch 'eof skip-whitespace
                 (lambda () (throw 'unexpected-eof)))
          (let ((value (cpp-definition-value)))
           ; See keysymdef.h: only the first of several mnenomics for a
           ;                  keysym should be used (the others are deprecated)
           (when (not (and (number? value) 
                           (hashtable-lookup value-table value)))
            (let ((sym (reverse! current-token)))
             (when (number? value) (hashtable-insert! value-table value sym))
             (set-definition!
               (create-entry! (get-root dictionary) sym)
              value)) ))
           #t ); finish

        (else
           #f) )) ))

;; Debugging
(define (log-debug . args)
 (for-each display args) (newline))
(define debug-on-throw (macro expressions
 `(catch #t (lambda () ,@expressions)
            (lambda (key . args) (REPL-HERE))) ))

;; Characters like back then.
(define char-upcase   char-upcase/c0)
(define char-downcase char-downcase/c0)
(define char-upcase?   char-upcase/c0?)
(define char-downcase? char-locase/c0?)

;; KEYSYM Symbol Selection & Transformations
(define (xkl-clean xkl) (filter (lambda (e) (number? (cdr e))) xkl))

(define (xkl-remove-a-prefix def)
 (let* ((sym (first def))
        (idx (brass-index sym #\_ )))
   (cons (subrass sym (+ 1 idx) (brass-length sym))
         (cdr def)) ))
(define (xkl-remove-prefix xkl) (map xkl-remove-a-prefix xkl))


;; CPP Header File Extraction
;; Extract some keysyms from the header files.
(repl-transcript
(define XK (make-dictionary))
;; We really should signal collisions here. This is actually needed by GIO.
(cpp-definitions! XK "/usr/include/X11/keysymdef.h")
(cpp-definitions! XK "/usr/include/X11/XF86keysym.h") ; PC keysyms
(describe-dictionary XK)
(define XKL (dictionary->alist XK))


(define keysyms (xkl-remove-prefix (xkl-clean XKL)))
) ; closes repl transcript

;;
;; What we really want to extract here is:
;;  - function keysyms of the core protocol
;;  - xf86 vendor keysyms 
;;
;; Latin-1 and Unicode will be used to compute USVs for chars directly.
;;
;; The function keysyms will then be used to deliver some symbol in the
;; key event where the mapping to the former protocol expected by the
;; controllers of the Viper (current terminal key event protocol) needs
;; to be done as a second step.
;;

(define vendor-offset #x10000000)
(define vendor-last   #x1FFFFFFF)
(define vendor-maximum (+ 1 (- vendor-last vendor-offset)))

(define latin1-offset #x20)
(define latin1-last   #x7E)
(define latin1-maximum (+ 1 (- latin1-last latin1-offset)))

(define unicode-offset #x01000100)
(define unicode-last   #x0110FFFF)
(define unicode-maximum (+ 1 (- unicode-last unicode-offset)))

;(define  latin1-vector (make-vector  latin1-maximum #f))
;(define unicode-vector (make-vector unicode-maximum #f))

;(define keysym-table (make-hashtable (floor (* 3/4 (length keysyms) ))))
(define keysym-table (make-hashtable (length keysyms) ))

(define vendor-min +inf.0)
(define vendor-max -inf.0)
(define unknown '())

(define   latin1-count 0)
(define  unicode-count 0)
(define   vendor-count 0)
(define     core-count 0)
(define collisions '())
(define (my-table-insert keysym-table n s)
  (let ((e (hashtable-lookup keysym-table n)))
    ; TODO
    ; Some revision of the actual collisions suggested we should use
    ; the first candidate, i.e. the one already in the table.
    ; Examples: (Thorn . THORN) (F11 . L1) (F12 . L2)
    ; but then there are other exampels like (Prior . Page_Up) which
    ; in conjunction with the above suggest we should review this issue
    ; again (in particular with respect to the bars and vchars which
    ; might be based on hashtables and code descriptions as well).
#! from keysymdef.h
Auxiliary functions; note the duplicate definitions for left and right
function keys;  Sun keyboards and a few other manufacturers have such
function key groups on the left and/or right sides of the keyboard.
 We've not found
!# ; .. no good reason for this but all this means: that dozen "function"
   ; keys we'll need in the Viper and else will be corrected manually.
   ; The rest of the function keys is then delivered as some random
   ; symbol in the *value* field of the key event? Maybe we could even
   ; deliver a list of symbols in that field.
   ; Do we need the rgb-3270 definitions or should we cancel that
   ; second prefix from our keysym list?
   ; Deliver the symbol list in the value field. Too much processing of
   ; of the original definitions only complicates further corrections
   ; based on use later on.
#! from keysymdef.h
 * Where several mnemonic names are defined for the same keysym in this
 * file, all but the first one listed should be considered deprecated.
!# ; That doesn't help much I'm afraid. No one calls that key "Prior".
   ; Have a nice dozen!
; Igrave issue: instead of some unicode thingy we see a latin-1 code
; for "ì" (i with accent downwards) which is decoded to the symbol
; Igrave found in keysymdef.h where that code point is considered part
; of latin 1 while our ranges here (from the core documents) and in
; keyboard.class.scm decode-keysym don't "allow" us to treat it as
; such.
; Rename this to the dead_abovering issue: the latin1 range has been
; extended up to #xff (in the keyboard class) but the "dead above ring"
; counts as a function key?
    (when e (set! collisions (cons (cons e s) collisions)))
    (hashtable-insert! keysym-table n s)))
(define (project-keysym p)
 (let* ((n (cdr p))
        (b (number->byte-string-4 n))
        (s (brass->symbol (car p))))
  (cond
  ((<=  latin1-offset n  latin1-last) (set! latin1-count (+ 1 latin1-count)))
  ((<=  vendor-offset n  vendor-last)
     (when (> n vendor-max) (set! vendor-max n))
     (when (< n vendor-min) (set! vendor-min n))
     (my-table-insert keysym-table n s)
     (set! vendor-count (+ 1 vendor-count)))
  ((<= unicode-offset n unicode-last) (set! unicode-count (+ 1 unicode-count)))
  (else (let* ((nbs (byte-string->list (network-byte-order b)))
               (one (first nbs))
               (two (second nbs))
               (three (third nbs))
               (four  (fourth nbs)))
   (cond
    ((= 0 one two) (set! core-count (+ 1 core-count))
                   (my-table-insert keysym-table n s))
    (else          (set! unknown (cons p unknown)) )) ))
)))

(define (save) 
 (with-output-to-file "sca-x-keysyms.data.scm" (lambda ()
   (write keysym-table))))

(when (defined? 'keysyms) (repl-transcript
vendor-maximum
latin1-maximum
unicode-maximum
(length keysyms)
(for-each project-keysym keysyms)
vendor-max
vendor-min
(- vendor-max vendor-min)
(length unknown)
(describe-hashtable keysym-table)
(define al (hashtable->alist keysym-table))
(list core-count vendor-count unicode-count latin1-count)
(hashtable-length keysym-table)
(+ core-count vendor-count)
(length collisions)

(when #f ; performance paragraph
 (define l (length keysyms))
 (define v (list->vector keysyms))
 (repeat (* 1M) (lambda () (hashtable-lookup keysym-table (cdr (vector-ref v (random l))))))
 ;; (define res '())
 ;; (repeat (* 500 1k) (lambda ()
 ;;   (set! res (cons (hashtable-lookup keysym-table (cdr (vector-ref v (random l))))
 ;;                   res))))
)

)) ; closes conditional repl transcript
(REPL)

(when #f 
;; Print a not well aligned table of hashtable descriptions.
(define start        (*   53 2))
(define  end  (quotient 2279 2))

(let loop ((KTS start))
 (when (<= KTS end)
   (set! keysym-table (make-hashtable KTS))
   (for-each project-keysym keysyms)
   (log-debug (format "~4D: " KTS) (describe-hashtable keysym-table))
   (loop (+ KTS 1))))
)
