
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

(define keyboard-class (class              (extension display-object-class)
 (slots xopcode xdetail) ; Xkb extension slots
 (slots mapping key-types lookup-modifier-state)

(define (keysym-map->array-4 keysym-map) ; SCA/X put this into function-request mapper
 (let ((group-width (byte-string->number (assq-ref keysym-map 'width)))
       (keysyms     (assq-ref keysym-map 'keysyms))
       (group-info  (assq-ref keysym-map 'group-info))
       (num-groups  0))
 (let ((group-v (make-vector 4 #f)))
  (let group-loop ((gid 0)) (when (< gid 4)
    (let ((gv (make-vector group-width #f)))
     (vector-set! group-v gid gv)
     (when (not (null? keysyms)) (set! num-groups (+ num-groups 1)))
     (let w-loop ((wid 0)) (when (and (< wid group-width)
                                      (not (null? keysyms)))
        (vector-set! gv wid (car keysyms))
        (set! keysyms (cdr keysyms))
        (w-loop (+ wid 1)) )))
    (group-loop (+ gid 1)) ))
  (values group-info num-groups group-width group-v (assq-ref keysym-map 'kid1)
                                                    (assq-ref keysym-map 'kid2)
                                                    (assq-ref keysym-map 'kid3)
                                                    (assq-ref keysym-map 'kid4))
)))

(method initialize (display xext-keyboard) (super display)
 (this (set! 'xopcode (xext-keyboard major-opcode)))

 ;; That's our race condition setup of the display class initialization routine can make this
 ;; activate events only long after /get-state/ has been called here. If we turn this around
 ;; we might end up without state or with a stale state overwriting one from a notification.
 (let ((state (call xekb 'Xkb-get-state ((this display) connection) (this xopcode))))
  (this (set! 'lookup-modifier-state (fixed-call xekb 'xkbrr-get-state:lookup-mods state))))
 
 (this (update-mapping))
)



(method sid () "default keyboard")

(method make-event-handler ()
 (lambda (display fbw)
  (let ((type (byte-string->number (fixed-call xekb 'xkbev-state-change:xkb-type fbw))))
  (cond
   ((= type (once (from-namespace xekb 'XkbStateNotify)))
      ;(X-log-debug "GIO keyboard event handler: new modifier lookup state")
      (this (set! 'lookup-modifier-state (fixed-call xekb 'xkbev-state-change:lookup-mods fbw))))
   ((= type (once (from-namespace xekb 'XkbMapNotify))) ;; No such events???
      (X-log-debug "GIO keyboard event handler: new keyboard mapping")
      (this (update-mapping)))
   ((= type (once (from-namespace xekb 'XkbNewKeyboardNotify)))
      ; That's about nine such events for a keyboard layout change:
      ; the additional events are queued for our first synchronous request
      ; with the function mapper but maybe it works good enough like this.
      (X-log-debug "GIO keyboard event handler: new keyboard")
      (this (update-mapping)))
   (else
      (X-log-debug "GIO keyboard event handler: unsupported xekb event type " type))
  ))
))

(method activate-events ()
 ;; XXX That way (in which) select-events currently works a scroll action provokes a state change
 ;;     notification as well as many other events on the display.
 (fixed-call xekb 'Xkb-select-events ((this display) connection) (this xopcode) 'mapping 'lookup-state)
 (let ((state (call xekb 'Xkb-get-state ((this display) connection) (this xopcode))))
   ;; race: clearing pending events isn't a good idea.
   ;; Maybe we should better provoke an update notification.
  (this (set! 'lookup-modifier-state (fixed-call xekb 'xkbrr-get-state:lookup-mods state))))
)

;; Key Event Listeners
(method add-key-listener (listener)  ;; expects method /key-event/
 (when (not (is-a? listener window-class))
  (error (brass-append "GIO keyboard-class add-key-listener:"
                       " argument needs to be an area of interest aka window")))
 (listener (activate-events key-event-class))
)
(method remove-key-listener (listener)
 (when (not (is-a? listener window-class))
  (error (brass-append "GIO keyboard-class remove-key-listener:"
                       " argument needs to be an area of interest aka window")))
 (listener (deactivate-events key-event-class))
)

(method keysym-map (key-code) ; XXX debugging
 (let ((kc (byte-string->number key-code)))
  (and (< kc (vector-length (this mapping)))
       (vector-ref (this mapping) (byte-string->number key-code)))))


(method update-mapping ()
 (let* ((mapping      (call xekb 'Xkb-get-map   ((this display) connection) (this xopcode)))
        (min-key-code (assq-ref mapping 'min-key-code))
        (max-key-code (assq-ref mapping 'max-key-code))
        (keysym-maps  (assq-ref mapping 'keysym-map))
        (keysym4      (map keysym-map->array-4 keysym-maps))
        (keysymv      (list->vector keysym4))
        (v            (make-vector (+ 1 max-key-code) #f))

        (total-types  (assq-ref mapping 'total-types))
        (type-vector  (make-vector total-types #f))
        (types        (assq-ref mapping 'keytypes))
        (first-type   (assq-ref mapping 'first-type)))

   (vector-project! v min-key-code keysymv)
   (this (set! 'mapping v))

   (cond ((and (> (length types) 0) (<= (+ first-type (length types)) total-types))
            (vector-project! type-vector first-type (list->vector types)))
         (else
            (X-log-error "GIO SCA/X keyboard get-map key type inconsistency")))
   (this (set! 'key-types type-vector))
))

;; Evaluate a Key Code
;; key-mapping |+| key-code => keysym-map  |+| state => group  |+| key-type |+| shift-level => value
;;
   ;; How to compute the shift level from the
   ;;   lookup-modifiers-state
   ;;        0 for no modifier
   ;;        1 for shift
   ;;        2 for capslock
   ;;        4 for ctrl
   ;;        8 for alt (left alt)
   ;;       64 for super (r+l)
   ;;      128 for ralt (= iso shift lvl crap)
   ;;      129 for ralt+shift
   ;;
   ;; BTW The bit flags for these keys are defined in /usr/include/X11/X.h
   ;;     where they're called masks again. ShiftMask LockMask ControlMask
   ;      and ModnMasks
   ;;
   ;; We'll need the following:
   ;;  0 -> 0
   ;;  1 -> 1
   ;;  8 -> 2 [current practice in X seems to be 128 -> 2]
   ;;       the sum of these is then the shift level (at least for alphabetics)
   ;; the rest should count as "preserved" where that word only appears in the
   ;; documentation.
   ;; Since we know we DIY and it was probably meant like that.
   ;; So we add this nice key event evaluation layer which is already called
   ;; as such.
   ;; XEKB in fact dictates some data mangling process to retrieve the shift level
   ;; from the keyboard map key type entry map whatnot: but when you know that
   ;; ctrl is ctrl and shift is shift it'll still be hard to get to some original
   ;; representative character for a key. That keyboard model of X with latched
   ;; and whatnot and some thousand keys bound to symbols and USVs. UC broken concepts.
   ;; GIO shouldn't export that crap and override where need arises.
;; There are currently five return values:
;;  1) the value of the key event
;;  2) the modifiers-lookup-state used (bit field)
;;  3) a list of symbols corresponding to known modifiers active during this key event
;;  4) the computed shift level
;;  5) the actual group (vector) used
;;  6) the group info byte
;;  7) the corresponding key type
(method evaluate-kc (state key-code)
 (let* ((mapping (this mapping))
        (modloks (this lookup-modifier-state))
        (group   (fixed-call xekb 'kb-state-group state))
        (kc      (byte-string->number key-code))
        (kv      (and (< kc (vector-length mapping))
                      (vector-ref (this mapping) (byte-string->number key-code))))
        (group-width #f)
        (num-groups #f)
        (shift-M #f)
        (kits    #f)
        (kit     #f)
        (key-type #f)
        (group-info #f)
        (gv      #f)
        (ms      '()))
   (set! kits       (and kv (list->vector (cddddr kv)))) ; one /d/ for every try on xkb up until next to regular operation ;-)
   (set! group-info (and kv (first kv)))
   (set! num-groups (and kv (second kv)))
   (set! group-width (and kv (third kv)))
   (set! kv         (and kv (fourth kv)))

   ;(X-log-debug "evaluate-kc modloks " modloks " group-info " group-info)
   ;(set! gv (and (vector? kv) (< group (vector-length kv)) (vector-ref kv group)))
   ;(X-log-debug "num-groups " num-groups)
   (set! gv (and (vector? kv) (cond ((< group num-groups) (vector-ref kv group))
                                    (else (case group-info ((wrap)     (vector-ref kv (modulo group num-groups)))
                                                           ((clamp)    (vector-ref kv (- num-groups 1)))
                                                           ((redirect) (vector-ref kv 0)) ; ??? There's a call to out-of-range-group-number in xlib. What does this do??? XKBBind.c ll. 294 dunno it's used but not even part of xlib perhaps the server sources ...
                                                           (else       #f))) )))

   ;; Bye bye xekb data mangling: @sum point itsnaf
   (when (bit-and? modloks #"1")   (set! ms (cons 'shift    ms)))
   (when (bit-and? modloks #"2")   (set! ms (cons 'capslock ms)))
   (when (bit-and? modloks #"4")   (set! ms (cons 'control  ms)))
   (when (bit-and? modloks #"8")   (set! ms (cons 'alt      ms)))
   (when (bit-and? modloks #"64")  (set! ms (cons 'super    ms)))
   (when (bit-and? modloks #"128") (set! ms (cons 'altgr    ms)))

   (set! key-type (begin
                       (set! kit (and kits (< group (vector-length kits))
                                   (vector-ref kits group)))
                       (and kit (begin (set! kit (byte-string->number kit)) #t)
                            (< kit (vector-length (this key-types)))
                            (vector-ref (this key-types) kit))))

   ;; This adheres to the Xlib functionality (see XKBBind.c ll. 316)
   ;; Especially the bit magic with /type-mask/ and /modifier-lookup-state/ seems to be
   ;; otherwise undocumented if not by the original source code.
   (set! shift-M 0)
   (when key-type (let* ((entries (assq-ref key-type 'entries))
                         (actives (filter (lambda (e) (assq-ref e 'active)) entries))
                         (tmask   (assq-ref key-type 'mask))
                         (entry   #f))
                    (when actives (set! entry (assoc (bit-and tmask modloks) actives)))
                    (when entry   (set! shift-M (assq-ref entry 'level)) )))

   ;; TODO evaluate /preserve/
   ;; TODO streamline data structure layout of key types ??? maybe that's not worth it

   (values (and gv (< shift-M (vector-length gv)) (vector-ref gv shift-M))
           modloks
           ms
           shift-M
           gv
           (cons 'group-info group-info)
           (cons 'type (cons kit key-type)))
))

(method x () ; XXX
 (call xekb 'Xkb-get-map ((this display) connection) (this xopcode))
)



(define xekb (namespace (namespace-import sca/x-common)

 ;; KEYSYM table: X keysym to ANS symbol mappings (See also the file GIO/gio-keysymv.scm)
 (sys-load "lib/hashtable.scm")
 (define keysym-table (let ((path (%search-load-path "GIO/sca-x-keysyms.data.scm")))
   (when (not path) (error "SCA/X file not found GIO/sca-x-keysyms.data.scm"))
   (with-input-from-file path read)))
 (define (x-keysym->symbol keysym) (hashtable-lookup keysym-table keysym))

 ;; See also Computing a State Field from an XKB State
 (define kb-state-group:mask (bit-shift-left (number->byte-string-2 #b11) 13))
 (define (kb-state-group  kb-state)
  (byte-string->number (bit-shift-right (bit-and kb-state kb-state-group:mask) 13)))
 ;(define kb-state:keymask-mask (number->byte-string-2 127))
 ;(define (kb-state:keymask kb-state) (bit-and kb-state kb-state:keymask-mask))
 ;; The lookup state is transmitted seperately as part of the extension.

(define pc number->byte-string-2)
(define one-2 (number->byte-string-2 1))
;; Values for device-spec
(define xkb-use-core-kbd (pc #x0100))  ;; XKB.h l. 317
(define xkb-use-core-ptr (pc #x0200))

(define Xkb-rp-get-state
 (byte-structure-description
  (opcode         1)
  (minor-opcode   1)
  (request-length 2)
  (device-spec    2)
  (padding        2)))
(define-bs-accessors 'xkbrp-get-state Xkb-rp-get-state)
(define Xkb-rr-get-state
 (byte-structure-description
  (type           1)
  (device-id      1)
  (sequence-number 2)
  (reply-length    4)
  (mods            1)
  (base-mods       1)
  (latched-mods    1)
  (locked-mods     1)
  (group           1)
  (locked-group    1)
  (base-group      2)
  (latched-group   2)
  (compat-state    1)
  (grab-mods       1)
  (compat-grab-mods   1)
  (lookup-mods     1)
  (compat-lookup-mods 1)
  (padding         1)
  (ptr-btn-state   2)
  (padding         6)))
(define-bs-accessors 'xkbrr-get-state Xkb-rr-get-state)
(define Xkb-get-state:template
 (let ((template (make-byte-string (bs-length Xkb-rp-get-state))))
  (byte-string-clear! template)
  (xkbrp-get-state:set-minor-opcode!   template (byte->byte-string     4))
  (xkbrp-get-state:set-request-length! template (number->byte-string-2 2))
  template))
(define (Xkb-get-state connection major-opcode)
 (let ((rb Xkb-get-state:template))
  (xkbrp-get-state:set-opcode!      rb (byte->byte-string major-opcode))
  (xkbrp-get-state:set-device-spec! rb xkb-use-core-kbd)
  (X-request connection rb (lambda (con fbw)
   ; return the bytes for now
   fbw
))))


;; These denote the flags but end in "Mask" in the original.
(define	XkbKeyTypesMask		(bit-shift-left one-2 0))
(define	XkbKeySymsMask		(bit-shift-left one-2 1))
(define	XkbModifierMapMask	(bit-shift-left one-2 2))
(define	XkbExplicitComponentsMask (bit-shift-left one-2 3))
(define XkbKeyActionsMask	(bit-shift-left one-2 4))
(define	XkbKeyBehaviorsMask	(bit-shift-left one-2 5))
(define	XkbVirtualModsMask	(bit-shift-left one-2 6))
(define	XkbVirtualModMapMask	(bit-shift-left one-2 7))

(define	XkbAllClientInfoMask	(bit-or XkbKeyTypesMask XkbKeySymsMask XkbModifierMapMask))
(define	XkbAllServerInfoMask	(bit-or XkbExplicitComponentsMask XkbKeyActionsMask
                                        XkbKeyBehaviorsMask XkbVirtualModsMask XkbVirtualModMapMask))
(define	XkbAllMapComponentsMask	(bit-or XkbAllClientInfoMask XkbAllServerInfoMask))

(define Xkb-rp-get-map
 (byte-structure-description
  (opcode          1)
  (minor-opcode    1)
  (request-length  2)
  (device-spec     2)
  (full            2)
  (partial         2)
  (first-type         1) (number-of-types         1)
  (first-keysym       1) (number-of-keysyms       1)
  (first-keyact       1) (number-of-keyacts       1)
  (first-keybehavior  1) (number-of-keybehaviors  1)
  (virtual-mods       2)
  (first-keyexplicit  1) (number-of-keyexplicit   1)
  (first-modmap-key   1) (number-of-modmap-keys   1)
  (first-v-modmap-key 1) (number-of-v-modmap-keys 1)
  (padding         2)))
(define-bs-accessors 'xrp-get-map Xkb-rp-get-map)
(define Xkb-rr-get-map
 (byte-structure-description
  (type            1)
  (device-id       1)
  (sequence-number 2)
  (reply-length    4)
  (padding         2)
  (min-key-code    1)
  (max-key-code    1)
  (present         2)
  (first-type         1) (number-of-types         1) (total-types         1)
  (first-keysym       1) (total-syms 2)              (number-of-keysyms 1)      ; sic!
  (first-keyact       1) (total-acts 2)              (number-of-keyacts 1)      ; sic!
  (first-keybehavior  1) (number-of-keybehaviors  1) (total-keybehaviors  1)
  (first-keyexplicit  1) (number-of-keyexplicit   1) (total-keyexplicit   1)
  (first-modmap-key   1) (number-of-modmap-keys   1) (total-modmap-keys   1)
  (first-v-modmap-key 1) (number-of-v-modmap-keys 1) (total-v-modmap-keys 1)
  (padding         1)
  (virtual-mods    2)))
(define-bs-accessors 'xrr-get-map Xkb-rr-get-map)
(define Xkb-get-map:template
 (let* ((length   (bs-length Xkb-rp-get-map))
        (template (make-byte-string length)))
   (byte-string-clear! template)
   (xrp-get-map:set-minor-opcode!    template (byte->byte-string     8))
   (xrp-get-map:set-request-length!  template (number->byte-string-2 (/ length 4)))
   template))
(define unicode-keysym-offset #x01000000)
(define function-keysym-mask  (number->byte-string-4 #xFFFF0000))
(define lbyte (number->byte-string-4 #xFF))

;; (decode-keysym keysym-bytes)
;; Returns a /character/ where applicable
;; or a /symbol/ for function keys (derived from the system header files)
;; or falsity for unsupported keysyms.
(define (decode-keysym keysym-bytes)
 (values 4 ; read out list protocol / calling convention
 (let* ((keysym-bytes (byte-string-select keysym-bytes 0 4))
        (num (byte-string->number keysym-bytes)))
 (cond
 ((<= #x01000100 num #x0110FFFF) (integer->char (- num unicode-keysym-offset))) ; Unicode Keysym
                                 ;; (ISO/IEC 8859-1 = Unicode U+0020..U+00FF)
 ;;((<= #x0020     num #x007E)     (integer->char num))                          ; Latin-1 Keysym XXX we're missing some here but that range doesn't coincide with the sections in keysymdef.h (Igrave issue)
 ((<= #x0020     num #x00FF)     (integer->char num))                          ; Latin-1 Keysym XXX (dead_abovering issue; see gio-keysymv.scm)
 ((= num 0)                      'xkeysym:no-symbol)
 ((= num #x0FFFFFF)              'xkeysym:void-symbol)
 ((byte-string-false? (bit-and function-keysym-mask keysym-bytes))             : Function Keysym
  ;; The above predicate isn't distinct from the special value no-symbol. Top down sequence here.
  ;; [That no-symbol is usually contained in the header files and thus in our keysym table.]
  ;; (let ((byte3 (byte-string->number (bit-and lbyte (bit-shift-right keysym-bytes 16))))
  ;;       (byte4 (byte-string->number (bit-and lbyte (bit-shift-right keysym-bytes 24)))))
                                 (x-keysym->symbol num))
  ;;))
 ((<= #x10000000 num #x1FFFFFFF) (x-keysym->symbol num))                        ; Vendor Keysym
 (else                           #f) ; Some legacy keysyms based on code pages where ...
))))

;; See also xkbSymMapWireDesc in XKBproto.i
(define keysymmap-layout (byte-structure-description
  (kt-index-1 1) ; Groups 1 to 4
  (kt-index-2 1)
  (kt-index-3 1)
  (kt-index-4 1)
  (group-info 1)
  (width      1)
  (number-of-syms 2)
))  
(define-bs-accessors 'xkbe:ksm keysymmap-layout)
(define (decode-keysymmap byte-string)
 (let ((numsyms (byte-string->number (xkbe:ksm:number-of-syms byte-string))))
 (values (+ (bs-length keysymmap-layout) (* numsyms 4))
 (list (cons 'kid1 (xkbe:ksm:kt-index-1 byte-string))
       (cons 'kid2 (xkbe:ksm:kt-index-2 byte-string))
       (cons 'kid3 (xkbe:ksm:kt-index-3 byte-string))
       (cons 'kid4 (xkbe:ksm:kt-index-4 byte-string))
       (cons 'group-info (decode-group-info (xkbe:ksm:group-info byte-string)))
       (cons 'width (xkbe:ksm:width byte-string))
       (cons 'number-of-syms (byte-string->number (xkbe:ksm:number-of-syms byte-string)))
       (cons 'keysyms (second (X:read-out-list numsyms decode-keysym 'element-size
                                               byte-string (bs-length keysymmap-layout))))
))))

(define	XkbWrapIntoRange	0) ;; #x00)
(define	XkbClampIntoRange	1) ;; #x40)
(define	XkbRedirectIntoRange	2) ;; #x80)
;; XKB.h: " ... bits 6 & 7 specify out of range action ... " but we only see 0, 1 and 2 as the byte value
(define (decode-group-info group-info-byte)
 (case (byte-string->number group-info-byte)
  ((0) 'wrap) ((1) 'clamp) ((2) 'redirect) (else group-info-byte)))

;; See also xkbKeyTypeWireDesc in XKBproto.i
(define keytype-layout (byte-structure-description
 (mask          1) (real-mods     1) (virtual-mods  2) (num-levels    1)
 (n-map-entries 1) (preserve      1) (padding       1)
))
(define-bs-accessors 'keytype keytype-layout)
;; See also xkbMapEntryWireDesc in XKBproto.i
(define kt-mapentry-layout (byte-structure-description
 (active       1) (mask         1) (level        1) (real-mods    1)
 (virtual-mods 2) (padding      2)
))
(define-bs-accessors 'kt-mapentry kt-mapentry-layout)
;; See also xkbModsWireDesc in XKBproto.i
(define kb-moddef-layout (byte-structure-description
 (mask 1) (real-mods 1) (virtual-mods 2)
))
(define-bs-accessors 'kb-moddef kb-moddef-layout)
(define (pc n) (bit-shift-left #"1" n))
(define	XkbOneLevelMask		(pc 0))
(define	XkbTwoLevelMask		(pc 1))
(define	XkbAlphabeticMask	(pc 2))
(define	XkbKeypadMask		(pc 3))
(define	XkbAllRequiredTypes	(byte->byte-string #xf))

(define (decode-keymap-entry-rl bytes)
 (values (once (bs-length kt-mapentry-layout))
 (list (kt-mapentry:mask bytes) ;; <- the Key as is
       (cons 'active (byte-string-true? (kt-mapentry:active bytes)))
       (cons 'level        (byte-string->number (kt-mapentry:level bytes)))
       (cons 'real-mods    (kt-mapentry:real-mods bytes))
       (cons 'virtual-mods (kt-mapentry:virtual-mods bytes)) )))
(define (decode-moddef-rl bytes)
 (values (once (bs-length kb-moddef-layout))
  (list (cons 'mask         (kb-moddef:mask bytes))
        (cons 'real-mods    (kb-moddef:real-mods bytes))
        (cons 'virtual-mods (kb-moddef:virtual-mods bytes)) )))

(define (decode-keytype-rl byte-string)
 (let* ((num-map-entries (byte-string->number (keytype:n-map-entries byte-string)))
        (preserve        (byte-string-true? (keytype:preserve byte-string)))
        (entries         (X:read-out-list num-map-entries decode-keymap-entry-rl 'element-size 
                                          byte-string (once (bs-length keytype-layout))))
        (moddef          (and preserve (X:read-out-list num-map-entries decode-moddef-rl 'element-size
                                                        byte-string (first entries)))))
  (values (+ (once (bs-length keytype-layout))
             (* num-map-entries (once (bs-length kt-mapentry-layout)))
             (* num-map-entries (if preserve (once (bs-length kb-moddef-layout)) 0)))
          (list (cons 'mask      (keytype:mask      byte-string))
                (cons 'real-mods (keytype:real-mods byte-string))
                (cons 'virtual-mods (keytype:virtual-mods byte-string))
                (cons 'num-levels (byte-string->number (keytype:num-levels byte-string)))
                (cons 'entries   (second entries))
                (cons 'preserve  (and moddef (second moddef))) ))
))

(define (Xkb-get-map connection major-opcode)
 (let ((rb Xkb-get-map:template))
  (xrp-get-map:set-opcode!       rb (byte->byte-string major-opcode))
  (xrp-get-map:set-device-spec!  rb xkb-use-core-kbd)
  (xrp-get-map:set-full!         rb (bit-or XkbKeyTypesMask XkbKeySymsMask))
  ;(xrp-get-map:set-full!         rb XkbKeySymsMask)
  ;(xrp-get-map:set-full!         rb XkbKeyTypesMask)
  ;(xrp-get-map:set-full!         rb XkbAllMapComponentsMask)
  (X-request connection rb (lambda (con fbw)
   (let* ((reply-length (* 4 (byte-string->number (xrr-get-map:reply-length    fbw))))
          (seq-number   (byte-string->number      (xrr-get-map:sequence-number fbw)))
          (tail         (X-receive reply-length connection))
          (number-of-keytypes (byte-string->number (xrr-get-map:number-of-types   fbw)))
          (number-of-keysyms  (byte-string->number (xrr-get-map:number-of-keysyms fbw)))
          (types (X:read-out-list number-of-keytypes decode-keytype-rl 'element-size tail 8)) ;; XXX mysterious prepadding of 8
          (syms  (X:read-out-list number-of-keysyms decode-keysymmap 'element-size tail (+ 8 (first types)))))
   (list
   (cons 'device-id                          (xrr-get-map:device-id    fbw))
   (cons 'min-key-code  (byte-string->number (xrr-get-map:min-key-code fbw)))
   (cons 'max-key-code  (byte-string->number (xrr-get-map:max-key-code fbw)))
   (cons 'first-type    (byte-string->number (xrr-get-map:first-type   fbw)))
   (cons 'total-types   (byte-string->number (xrr-get-map:total-types  fbw)))
   (cons 'first-keysym  (byte-string->number (xrr-get-map:first-keysym fbw)))
   (cons 'total-syms    (byte-string->number (xrr-get-map:total-syms   fbw)))
   (cons 'present                            (xrr-get-map:present      fbw))
   (cons 'keytypes   (second types))
   (cons 'keysym-map (second syms))
    ))
   ))
))


;; XKB has a minor event code so it can use one X event code for 
;; multiple purposes.  
;; - reported in the xkbType field of all XKB events.
;; - XkbSelectEventDetails: Indicates the event for which event details 
;; are being changed
(define	XkbNewKeyboardNotify		 0)
(define XkbMapNotify			 1)
(define	XkbStateNotify			 2)
(define XkbControlsNotify		 3)
(define	XkbIndicatorStateNotify		 4)
(define	XkbIndicatorMapNotify		 5)
(define	XkbNamesNotify			 6)
(define XkbCompatMapNotify		 7)
(define	XkbBellNotify			 8)
(define	XkbActionMessage		 9)
(define	XkbAccessXNotify		10)
(define	XkbExtensionDeviceNotify	11)


;; Event Mask:
;; - XkbSelectEvents:  Specifies event interest.
(define one-2 (number->byte-string-2 1))
(define (pcs offset) (bit-shift-left one-2 offset))
(define	XkbNewKeyboardNotifyMask	(pcs  0))
(define XkbMapNotifyMask		(pcs  1))
(define	XkbStateNotifyMask		(pcs  2))
(define XkbControlsNotifyMask		(pcs  3))
(define	XkbIndicatorStateNotifyMask	(pcs  4))
(define	XkbIndicatorMapNotifyMask	(pcs  5))
(define	XkbNamesNotifyMask		(pcs  6))
(define XkbCompatMapNotifyMask		(pcs  7))
(define	XkbBellNotifyMask		(pcs  8))
(define	XkbActionMessageMask		(pcs  9))
(define	XkbAccessXNotifyMask		(pcs 10))
(define	XkbExtensionDeviceNotifyMask	(pcs 11))
(define	XkbAllEventsMask		(number->byte-string-2 #xFFF))

(define Xkb-rp-select-events
 (byte-structure-description
  (opcode         1)
  (minor-opcode   1)
  (request-length 2)
  (device-spec    2)
  (affect-which   2)
  (clear          2)
  (select-all     2)
  (affect-map     2)
  (map            2)))
(define-bs-accessors 'xkbrp-select-events Xkb-rp-select-events)
(define Xkb-rp-select-events:template
 (let ((template (make-byte-string (bs-length Xkb-rp-select-events))))
  (byte-string-clear! template)
  (xkbrp-select-events:set-minor-opcode! template (byte->byte-string 1))
  template))
(define (Xkb-select-events connection major-opcode . event-types)
 (let ((rb Xkb-rp-select-events:template))
   (xkbrp-select-events:set-opcode!         rb (byte->byte-string major-opcode))
   (xkbrp-select-events:set-device-spec!    rb xkb-use-core-kbd)
   (xkbrp-select-events:set-request-length! rb (number->byte-string-2 (/ (bs-length Xkb-rp-select-events)
                                                                         4)))
   (when (not (and (pair? event-types) (equal? event-types '(mapping lookup-state))))
     (error "SCA/X preliminary limitation Xkb-select-events unsupported event types"))

   (xkbrp-select-events:set-select-all!     rb (bit-or XkbStateNotifyMask XkbNewKeyboardNotifyMask))
   (xkbrp-select-events:set-affect-which!   rb (bit-or XkbStateNotifyMask XkbNewKeyboardNotifyMask))
   ;; (xkbrp-select-events:set-affect-which!   rb (bit-or XkbStateNotifyMask XkbMapNotifyMask))
   ;; (xkbrp-select-events:set-map!            rb (bit-or XkbKeyTypesMask XkbKeySymsMask))
   ;; (xkbrp-select-events:set-affect-map!     rb (bit-or XkbKeyTypesMask XkbKeySymsMask))
   ; NB There's some tail consisting of some related to /affectWhich/ and
   ;    mutual exlusive with /selectAll/ and /clear/
   (X-request connection rb no-response)
))

(define Xkb-ev-state-change
 (byte-structure-description
  (type     1)
  (xkb-type 1)
  (sequence-number 2)
  (time            4)
  (device-id       1)
  (mods            1)
  (base-mods       1)
  (latched-mods    1)
  (locked-mods     1)
  (group           1)
  (base-group      2)
  (latched-group   2)
  (locked-group    1)
  (compat-state    1)
  (grab-mods       1)
  (compat-grab-mods   1)
  (lookup-mods        1)
  (compat-lookup-mods 1)
  (ptr-btn-state      2)
  (changed            2)
  (keycode            1)
  (event-type         1)
  (request-major      1)
  (request-minor      1)))
(define-bs-accessors 'xkbev-state-change Xkb-ev-state-change)

(define Xkb-ev-map
 (byte-structure-description
  (type            1)
  (xkb-type        1)
  (sequence-number 2)
  (time            4)
  (device-id       1)
  (ptr-btn-actions 1)
  (changed         2)
  (min-key-code      1) (max-key-code           1)
  (first-type        1) (number-of-types        1)
  (first-keysym      1) (number-of-keysyms      1)
  (first-keyact      1) (number-of-keyacts      1)
  (first-keybehavior 1) (number-of-keybehaviors 1)
  (first-keyexplicit 1) (number-of-keyexplicit  1)
  (first-modmapkey   1) (number-of-modmapkeys   1)
  (first-vmodmapkey  1) (number-of-vmodmapkeys 1)
  (virtual-mods      2) (padding 2) ))
(define-bs-accessors 'xkbev-map Xkb-ev-map)
  
)) ; closes xekb namespace
)) ; closes keyboard class
