
   ;;   ;;   ;;     Graphical        Viper System Interface
 ;;     ;; ;;  ;;   Input                    SCA/X
 ;;  ;; ;; ;;  ;;   Output            Viper Object System
   ;;   ;;   ;;                  

;;
;; Z Pixmap: the number of scanlines is the /height/ and the number of pixels
;;           in each scanline is the width.
;; (Scanline bits-per-pixel ... (padding bits-per-pixel=1 (|+| width bitmap-scanline-pad)))
;;
;; scanline-pad, bits-per-pixel are found via the /format/ of the /depth/
;;
;; (Maybe doesn't apply to z pixmap formats bitmap-scanline-unit from the Display)
;;
(sys-load "lib/matrices.scm")

(define pixmap-class (class                    (extension display-resource-class)
 (slots drawable format-id width height)
 (slots x-shared-mem)
 ;(slots shm-id shm-mapping)

(method initialize (drawable width height) (let ((display (drawable display)))
 (super display)
 (this (set! 'format-id  ((display xext-shared-memory) pixmap-format-id))) ;; should always be two: ZPixmap
 (this (set! 'width    width))
 (this (set! 'height   height))
 (this (set! 'x-shared-mem (make-object x-shared-memory-class display (* width height 4)))) ; 4 bitmap-scanline-unit? of the display
   
 (this (set! 'drawable drawable))
 (fixed-call xshm 'Xshm-create-pixmap (display connection)
                                      ((display xext-shared-memory) major-opcode)
                                      (this rid)
                                      (drawable rid)
                                      width height
                                      24 ;(@slot-ref (this format) 'depth)
                                      ((this x-shared-mem) rid)
                                      0)
))

(method set-pixels-from-color-matrix (color-matrix converter)
 (let* ((w (matrix:number-of-columns color-matrix))
        (h (matrix:number-of-rows    color-matrix))
        (m ((this x-shared-mem) shm-mapping))
        (off 0))
   (when (not (and (= w (this width)) (= h (this height))))
     (error "pixmap class set pixels from color matrix: dimensional mismatch"))
   (matrix-for-each (lambda (C) (let ((b (converter C))) 
                                  (byte-string-set! m off b)
                                  (set! off (+ (byte-string-length b) off))))
                    color-matrix)
))

(method get-pixel-mapping () ((this x-shared-mem) shm-mapping))

(method set-pixels (pixels)
  (byte-string-set! ((this x-shared-mem) shm-mapping) 0 pixels)
)

(method destroy ()
 ((this x-shared-mem) (destroy))
 (fixed-call xpm 'X-free-pixmap ((this display) connection) (this rid))
 (super) ; presumption: not a recipient
)

(method set-pixels-from (source src-x src-y)
 (let ((res (fixed-call xpm 'X-get-image
                            ((this display) connection)
                            (source rid)
                            src-x src-y (this width) (this height)) ))
  (X-log-debug res)
  ;; XXX this returns a byte string of length zero?
  (this (set-pixels (assq-ref res 'pixels )))
))

(define xpm (namespace (namespace-import sca/x-common)
(define X-rp-create-pixmap
 (byte-structure-description
  (opcode   1)
  (depth    1)
  (request-length 2)
  (pid      4)
  (drawable 4)
  (width    2)
  (height   2) ))
(define-bs-accessors 'xrp-create-pixmap X-rp-create-pixmap)
(define X-create-pixmap:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-create-pixmap))) ))
   (xrp-create-pixmap:set-opcode!         template (byte->byte-string    53))
   (xrp-create-pixmap:set-request-length! template (number->byte-string-2 4))
   template))
(define (X-create-pixmap connection pid drawable width height)
  (let ((rb X-create-pixmap:template))
    (xrp-create-pixmap:set-depth!    rb (byte->byte-string 8))
    (xrp-create-pixmap:set-pid!      rb pid)
    (xrp-create-pixmap:set-drawable! rb drawable)
    (xrp-create-pixmap:set-width!    rb (number->byte-string-2 width))
    (xrp-create-pixmap:set-height!   rb (number->byte-string-2 height))
    (X-request connection rb no-response)
))

(define X-rp-free-pixmap 
 (byte-structure-description
  (opcode   1)
  (padding  1)
  (request-length 2)
  (pid      4)))
(define-bs-accessors 'xrp-free-pixmap X-rp-free-pixmap)
(define X-free-pixmap:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-free-pixmap))) ))
   (xrp-free-pixmap:set-opcode!         template (byte->byte-string    54))
   (xrp-free-pixmap:set-request-length! template (number->byte-string-2 2))
   template))
(define (X-free-pixmap connection pid)
  (let ((rb X-free-pixmap:template))
    (xrp-free-pixmap:set-pid! rb pid)
    (X-request connection rb no-response)
))

;; X-put-image needs a GC and is thus in the namespace xdraw of the graphics context class

(define X-rp-get-image
 (byte-structure-description
  (opcode     1)
  (format     1)
  (request-length 2)
  (drawable   4)
  (x          2)
  (y          2)
  (width      2)
  (height     2)
  (plane-mask 4)))
(define-bs-accessors 'xrp-get-image X-rp-get-image)
(define X-rr-get-image
 (byte-structure-description
  (reply           1)
  (depth           1)
  (sequence-number 2)
  (reply-length    4)
  (visual-id       4)
  (padding        20)))
(define-bs-accessors 'xrr-get-image X-rr-get-image)
(define X-get-image:template
 (let ((template (byte-string-clear! (make-byte-string (bs-length X-rp-get-image))) ))
   (xrp-get-image:set-opcode!         template (byte->byte-string    73))
   (xrp-get-image:set-request-length! template (number->byte-string-2 5))
   template))
(define (X-get-image connection drawable x y width height)
  (let ((rb X-get-image:template))
    (xrp-get-image:set-format!   rb (byte->byte-string 2)) ; 0 bitmap 1 XYPixmap 2 Zpixmap
    (xrp-get-image:set-drawable! rb drawable)
    (xrp-get-image:set-x!        rb (number+sign->byte-string-2 x))
    (xrp-get-image:set-y!        rb (number+sign->byte-string-2 y))
    (xrp-get-image:set-width!    rb (number->byte-string-2  width))
    (xrp-get-image:set-height!   rb (number->byte-string-2 height))
    (X-request connection rb (lambda (con fbw)
     (let* ((reply-length (* (byte-string->number (xrr-get-image:reply-length fbw)) 4))
            (tail         (X-receive reply-length con)))
       (list (cons 'depth  (byte-string->number (xrr-get-image:depth fbw)))
             (cons 'visual                      (xrr-get-image:visual-id fbw))
             (cons 'pixels tail)) )))
))

)) ; closes xpm namespace
)) ; closes pixmap class
