;;
;; <<< Sort of what vanished here again.
;;     It's called classoid, a lisp struct in src/code/class.lisp
;;


;; Some completely unrelated "goodies" from meta-vmdef.lisp

;;; Define the symbol NAME to be a Virtual OPeration in the compiler.
;;; If specified, INHERITS is the name of a VOP that we default
;;; unspecified information from. Each SPEC is a list beginning with a
;;; keyword indicating the interpretation of the other forms in the
;;; SPEC:
;;;
;;; :ARGS {(Name {Key Value}*)}*
;;; :RESULTS {(Name {Key Value}*)}*
;;;     The Args and Results are specifications of the operand TNs passed
;;;     to the VOP. If there is an inherited VOP, any unspecified options
;;;     are defaulted from the inherited argument (or result) of the same
;;;     name. The following operand options are defined:
;;;
;;;     :SCs (SC*)
;;;         :SCs specifies good SCs for this operand. Other SCs will
;;;         be penalized according to move costs. A load TN will be
;
;; (def!macro define-vop ((name &optional inherits) &body specs)
;;   (declare (type symbol name))
;;   ;; Parse the syntax into a VOP-PARSE structure, and then expand into
;;   ;; code that creates the appropriate VOP-INFO structure at load time.
;;   ;; We implement inheritance by copying the VOP-PARSE structure for
;;   ;; the inherited structure.

;;
;; In Scheme the determining constituent is the first constituent of an
;; expression (determinant constituents) where we currently have:
;;  primary primitives, syntax objects, closures and secondary primitives
;; From an object orientated point of view the second constituent seems
;; to be the determining constituent: (vector-ref v n) and (car p) could
;; be read as messages to the objects /v/ and /p/. The messages /vector-ref/
;; and /car/ are only valid for certain classes of objects: vectors and
;; pairs in this case. Thus a system based on object orientation would
;; retrieve the actual methods from the second constituent where in Scheme
;; (close port) already does it a little bit like that depending on the
;; port type but only within /close/. What about a function call:
;; (fun a b c) where /fun/ is probably a closure object defined by the
;; user. In the canonical form which (we would need to transform to beforehand)
;; a function call is nothing but an implicit apply: (apply fun a b c)
;; and the class of closures would provide that method /apply/. After
;; that change of point of view we can optimize back into every routine
;; and every spot which applies needs to branch for every optimization
;; depending on the number of arguments (original SCM practice in the
;; interpreter).
;;
;; Now let's add (Viper) Objects as determining first constituents.
;; How should we do this without crippling the runtime before we know
;; what we want to have? Every object is an instance of a class and
;; in our prototype every object is a syntax object. That setup allows
;; to model and design our object system without the need to change
;; three optimized apply branches about three times.
;;


;;;;;; Here it is:
;;;
;;; OBJECT SYNTAX
;;;
;;; (make-object class . arguments) => object
;;; (object) => storage-of-object
;;; (object (method . arguments)) => (method object . arguments)
;;; (object slot-name) => slot-value
;;;

(define (method-ref class method) (slot-ref (@slot-ref class 'class-methods) method))
(define make-object:transformer (lambda (exp env)
 ;; Is this MOP?
 (let* ((class      (eval! (second exp) env))
        (class-env  (slot-ref class 'environment))
        (storage    (make-vector (vector-length (slot-ref class 'object-layout))))
        (inits      (map (lambda (e) (eval! e env)) (cddr exp)))
        (initialize (method-ref class 'initialize))) ; Viper Objects
   (vector-set! storage 0 (slot-ref class 'object-layout))
   (let ((this #f)) 
     (set! this
   (procedure->syntax (lambda (exp env)
    ;;
    ;; !!! Be careful here to not memoize storage directly.
    ;;
    ;; Or we quote which makes /ceval/ hop or this becomes a
    ;; syntax object which needs to evaluate in addition to
    ;; the memoizing. TODO try the syntax variant
    ;; This is a prototype - not the optimized last thing.
    ;;  What we can presume here is the following: every
    ;; further evaluation of the position where this syntax
    ;; object had been called will feature a compatible class, i.e.
    ;; name slot bindings are equal as they will be upwards the
    ;; hierarchy. (both for methods and object slots)
    ;;  What we can't presume is to find the same method for
    ;; the same name only the same slot index of the method.
    ;; Therefore in this prototype we can prepare the next
    ;; evaluation of this node but we can't fix storage or
    ;; methods.
    ;;  After a day of crafting this prototype we have forgotten
    ;; about the implicit fix call? Or did we run into situations
    ;; where this doesn't work as expected. This first day of the
    ;; new Viper Objects Prototype we shouldn't have memoized at
    ;; all.
    ;;
    (cond ((null? (cdr exp))           storage)
          ((and (pair? (second exp))
                (= (length exp) 2))    (let* ((call        (second exp))
                                              (method-name (first  call))
                                              (arguments   (cdr    call))
                                              (obj         this)  ; unmemoizable
                                              ;;(obj         (eval! (car exp) env))
                                       ;; We could only memoize one (car exp) as it might be an expression to
                                       ;; be evaluated on every pass through.
                                              (new-exp     `((method-ref (class-of (,obj)) (quote ,method-name))
                                                             ,obj ,@arguments) ))
                                        (case method-name
                                          ((destroy) (and (eval! (cons (method-ref class method-name)
                                                                       (cons obj arguments)) env)
                                                          ;; ??? void storage if the method doesn't return false or unconditionally ???
                                                          (begin (vector-fill! storage *unspecified*)
                                                                 (vector-set! storage 0 'destroyed-object)))
                                                     (set! this #f)
                                                     *unspecified*)
                                          (else

                                        ;(when (not (symbol? obj)) (display obj) (newline) (REPL-HERE))
                                        ;(set-car! exp (car new-exp))
                                        ;(set-cdr! exp (cdr new-exp))
                                        ;(eval! exp env) ))
                                        (eval! (cons (method-ref class method-name)
                                                     (cons obj arguments)) env)))))
          ((and (symbol? (second exp))
                (= (length exp) 2))     (let* ((slot-name (second exp))
                                              ;(new-exp `(slot-ref ,storage (quote ,slot-name)))) ; unmemoizable
                                              ;; All objects in this text position (where we memoize) need
                                              ;; to have the same indices for the same slot names.
                                              ;(mem-exp `(slot-ref (,(car exp)) (quote ,slot-name))))
                                              (mem-exp `(slot-ref (,(car exp)) (quote ,slot-name))))
                                         (set-car! exp (car mem-exp))
                                         (set-cdr! exp (cdr mem-exp))
                                         (eval! mem-exp env)))
          (else (error "mischiefed object expression" exp)))))
   ) ; closes object syntax this
     (and initialize
          ; Reminder: /make-object/ was called twice: macros should
          ; substitute not duplicate expressions or you'll get twice
          ; as many objects and one w/o the possiblity to release:
          ; ([make-object a-class] (release)) =>
          ;   ((method-ref (class-of [make-object a-class] 'release [make-object a-class])))
          ; Some' based on the analyzed call graph might employ a more intelligent
          ; substitution pattern here.
          (apply initialize this inits))

     (when #f ;; Debugging Paragraph
       (let storage-check ((i (- (vector-length storage) 1)))
         (when (eq? (vector-ref storage i) *unspecified*)
           (display (format "VO MOP: uninitialized slot: ~a #~a"
                            (vector-ref (slot-ref class 'object-layout) i)
                            i)
                    (current-error-port))
           (newline))
        (when (not (= i 0)) (storage-check (- i 1))))
     )
     this
     ;(cut! this this)
)) ))
(define make-object (procedure->syntax make-object:transformer))



;;;
;;; CLASS NAMESPACE
;;;

(define (class-namespace expression environment)
 (if (null? expression)
   environment
   (namespace-environment:transformer expression environment)))



;;;
;;; CLASSOIDS and CLASS SYNTAX FORMS
;;;
;;; DATA GRAPH from the object up to the classoid
;;;
;;; (prototype-syntax) => storage-of-object -> object-layout <-> class
;;;
;;; Classes are considered constant and therefore we don't handle the
;;; rings.
;;;
;;; (class expressions)
;;; where expressions are constituted of the following:
;;; Mandatory Class Extension: (extension super-class)
;;; Object Storage Descriptor: (slots  slot-names)
;;;         Method Definition: (method name formals body)
;;;        Regular Definition: (define regular-define-syntax)
;;;
;;; Methods can access the object they were called on by using object syntax
;;; with the reserved identifier /this/ which is implicitely bound to the
;;; current object. Direct slot access to object storage is possible with
;;; /slot-ref/ and /slot-set!/ using /(this)/ to access the storage directly.
;;;
;;; See also Object Syntax on how to access storage and call methods.
;;;
;;; Regular definitions extend the environment of the definition of the 
;;; class and are evaluated when the definition of the class is parsed.
;;; They can be used from within method definitions but they can't access
;;; the current object denoted by /this/ if not when given as an argument.
;;; Thus the environment of a class serves to provide bindings for class wide
;;; definitions and each binding only exists once per class.
;;;
;;; NB Right now there are three distinct namespaces:
;;;    class environment, method names, slot names
;;;
;;;
;;; Class Extensions (aka inheritance)
;;;
;;; Object Syntax Extension /super/
;;;  (super . arguments)
;;; Calls the overridden method of the super class. Only valid within the override of a method.
;;;
;;; - Root classes like the viper-object-class use the classoid itsself as the extension class.
;;; - Methods & Storage are stacked and retain the name-index bindings of the extension class.
;;; - The class environment remains class private: every class has its own environment
;;;   (reflecting the actual textual environment, ie. its valid bindings)
;;; - Behavior for advanced jokes like an inner class with the surrounding class as the
;;;   extension class remains unspecified and should provoke a fatal error.
;;;
;;;      class-methods has then the length of the super class-methods plus
;;;                    the number of the methods in the class definition
;;;                    which aren't overrides.
;;;      environment   is class private for each and every class. Thus the
;;;                    environments are completely distinct in our concept
;;;                    as they need to represent the actual textual environment
;;;                    of the class definition.
;;;      object-layout is the object layout of the super class extended
;;;                    with the declared slots in the /slots/ statements.
;;;                    In the case of a name collision an error is signalled.
;;;
;;;      There can only be one extension class as to enforce an artificial
;;;      hierarchy onto the class graph: only that way we can stack storage
;;;      and maintain performance. The /extension/ statement is mandatory
;;;      and part of the class definition syntax as the second constituent. [<- can currently occur at any position in the class]
;;;      The root class is called /viper-object-class/ and provides the method /set!/
;;;      on slots as well as a noop of the mandatory method /destroy/ which [<- not yet]
;;;      is expected by the current implementation of /this/.
;;       -> TODO encapsulate structure magic (see class transformer)
;;;
;;; TODO Respect /load/ statements in the class definition.
;;;      Maybe a (define m (namespace-load file)) with namespace import
;;;      could do the job?
;;;
;;; TODO interfaces (and interface slot offset transposition?)
;;;
;;;      Favourite typo of our unchecked class definition expressions is /slot a/
;;;      instead of /slots a/ where this should remain ??? 

(define classoid (struct environment extension class-methods object-layout))

(define (class-of object-storage) (vector-ref (vector-ref object-storage 0) 0))

(define method-name    second)
(define method-formals third)
(define method-body    cdddr)

(define (method->lambda exp) `(lambda ,(cons 'this (third exp)) ,@(cdddr exp)))
(define (method->lambda/override super-class exp)
 `(lambda ,(cons 'this (method-formals exp))
          ,`(let* ((super (lambda arguments
                            (apply ,(method-ref super-class (method-name exp)) ; memoize the applicable (classes are considered constant)
                                   this arguments)))
                   (this this))  ;; protect /super/ against changes to the binding of /this/
              ,@(method-body exp)) ))


(define (sold-methods! super-class super-method-names super-methods method-expressions class-env)
 (let ((method-names (map method-name method-expressions))
       (methods      method-expressions))
  (when (not (and (= (length super-method-names) (length super-methods))
                  (= (length method-names)       (length (delete-duplicates method-names)))
                  (= (length super-method-names) (length (delete-duplicates super-method-names))) ))
    (error "viper objects: class coherency check failed"))
 
  ;; Resolve overrides, ie. methods which also exist in the super class.
  (let loop ((rest super-method-names) (sidx 0))
   (when (not (null? rest))
    (let* ((name (car rest))
           (cidx (list-index method-names name)))
     (when cidx (let ((m (list-ref methods cidx)))                                                                    ; my screen ends here  -->
                  (list-set! super-methods sidx (eval (method->lambda/override super-class m) class-env)) ; NB list-set! returns some crap
                  (set! methods      (delete! m methods))
                  (set! method-names (delete! name method-names)) ))
     (loop (cdr rest) (+ sidx 1)) )))
  (values (append! super-method-names method-names)
          (append! super-methods      (map (lambda (m) (eval m class-env)) (map method->lambda methods)) ))
))

(define class:transformer (lambda (exp env)
 (let ((extension (filter (lambda (e) (and (pair? e) (eq? (car e) 'extension))) exp))
       (super-class   #f)
       (super-method-names '())  ;; <- Lists of 
       (super-methods      '())  ;;
       (super-slots        '())) ;;
  (cond ((and (= (length extension) 1)
              (= (length (first extension)) 2))
              (set! super-class (eval! (second (first extension)) env))
              (when (not (eq? super-class classoid))
                (let ((sm (slot-ref super-class 'class-methods))
                      (so (slot-ref super-class 'object-layout)))
                  (set! super-method-names (values-from-struct (vector-ref sm 0)))
                  (set! super-methods      (values-from-struct sm))
                  (set! super-slots        (values-from-struct so)) )))
        (else (error "mischiefed class extension" extension)))

 (let ((slots   (apply append (map cdr (filter (lambda (e) (and (pair? e) (eq? (car e) 'slots))) exp))))
       (methods (filter (lambda (e) (and (pair? e) (eq? (car e) 'method))) exp))
       (definitions (filter (lambda (e) (and (pair? e) (eq? (car e) 'define))) exp)))
   (when (not (= (length (append super-slots slots)) (length (delete-duplicates (append super-slots slots)))))
     ;; TODO sth between explicit and informative?
     (error "viper objects: slot name collision"))
  (let* ((class-env           (class-namespace (cons 'begin definitions) env))
                              ;;     or just provide a struct-join to encapsulate our nice list processing here
         (object-layout       (eval `(struct ,@(append super-slots slots)) class-env)) ; ??? class-env? see below
         (method-lists        (sold-methods! super-class
                                             super-method-names super-methods
                                             methods
                                             class-env))
         (class-method-layout (eval `(struct ,@(first method-lists)) class-env)) ; ??? Was there any need to evaluate this in the /class-env/?
         (class-methods       (apply struct-from-values class-method-layout (second method-lists))))
    (let ((class (struct-from-values classoid class-env super-class class-methods object-layout)))
      (vector-set! object-layout 0 class)
      class ))) )))

(define class (procedure->syntax class:transformer))

(define viper-object-class (class               (extension classoid)
 (method set! (slot value) (slot-set! (this) slot value))
))

(define viper-object? macro?)  ; XXX preliminary predicate
(when (not (defined? 'sys-struct:is-a?)) ; double load protection
  (define sys-struct:is-a? is-a?))
(define (class? object) (sys-struct:is-a? object classoid))

(define (is-a? object type)
 (cond
  ((and (viper-object? object) (class? type)) ; prototype order
    (has-class? (class-of (object)) type))
  (else
    (sys-struct:is-a? object type)) ))

;; Is /class-2/ the same as /class-1/ or one of its super classes?
(define (has-class? class-1 class-2)
  (let loop ((class class-1))
    (and (class? class)  ;; OPT we could presume to find classoids here until we hit the classoid
         (or (eq? class-2 class)
             (loop (@slot-ref class 'extension)) )) ))

;; Maybe a new tag system would allow to branch on the tag instead of using predicates?
;; Perhaps the class of a vector isn't a memory object but the bits of its tag might be
;; taken for its type id (same for every primitive data type)?

(define (describe-class class)
 (when (not (is-a? class classoid)) (error "describe-class: wrong type argument"))
 (let ((nl "\n"))
  (brass-append ;; TODO We would actually need a reverse lookup (of a value in the system dictionary)
                ;;      to present the list of super classes adequately. Some copy & paste of the
                ;;      interested outsourced inlines and a check to not search for /undefined/.
   (format "  Slots: ~A" (cdr (vector->list (slot-ref class 'object-layout))))
   ;; Maybe we should actually show the prototypes with arguments.
   nl (format "Methods: ~A" (cdr (vector->list (vector-ref (slot-ref class 'class-methods) 0))))
   nl)))


;;
;; Testing Paragraphs
;;
(when #f
(sys-load "VOS/vo.scm")
(sys-load "VSI-core/repl.scm")
(repl-transcript
(define test-class (class (extension viper-object-class)
 (slots a)

 (method initialize ( . args)
   (when (not (null? args)) (this (set! 'a (car args)))))

 (method p (b)
   (this (set! 'a (+ (this a) b)))
   (this a))

))
(define o (make-object test-class 4))
(o (p 5))
(o)
(o (set! 'a 10))
(o (p 10))
(o (p 10))

(define t2-class (class (extension test-class)
 (method initialize () (this (set! 'a 100)))
 (method p (b) (+ (super 50) 11))
))
(define q (make-object t2-class))
(q (p 5))
;(q)
(q (set! 'a 10))
(q (p 10))
(o (p 10))
(q (p 10))
;; ^^^ the worst numbers for a test case to check

(method->lambda/override 'sc '(method myth args body))
)
(REPL)
)

;;; PS introspection
;;
;; Examining a data structure defined in a namespace within a class:
;;
;; (eval '(REPL-HERE) (slot-ref keyboard-class 'environment))
;;  (define x (from-namespace xekb 'Xkb-rr-get-map))
;;  (length x) => 31
;;;
