(sys-load "VSI-core/byte-structures.scm")

(define control-class (class (extension viper-object-class)

 ;; Control Class

 (slots read-handlers write-handlers) ;; fs-select
 (slots shared-segments)              ;; shm

 (method initialize ()
  (slot-set! (this) 'shared-segments '())
  (slot-set! (this) 'read-handlers   '())
  (slot-set! (this) 'write-handlers  '())
 )


;; FS-SELECT Serializing Concurrent Controlling Input Output

 (method roll ()
  (let ((res #f)) (let roll ()
   (set! res (io-select (map handler-chan (this read-handlers))
                        (map handler-chan (this write-handlers))
                        '()))
   (when (not (null? res))
     (call-handlers res (this read-handlers)
                        (this write-handlers) ))
   (set! res #f)
   (when (not (and (null? (this read-handlers))
                   (null? (this write-handlers))))
     (roll))
 )))

 (method add-read-handler (channel callback)
   (slot-set! (this) 'read-handlers (acons channel callback (this read-handlers))))

 (method remove-read-handler (channel)
   (slot-set! (this) 'read-handlers (assq-remove! (this read-handlers) channel)))


 (define make-handler cons)
 (define handler-chan car)
 (define handler-cont cdr)

 (define (call-handlers select-result read-handlers write-handlers)
  (let ((read-handlers  (filter identity
                          (map (lambda (s)
                                 (handler-for-channel s read-handlers))
                               (first select-result)) ))
        (write-handlers (filter identity
                          (map (lambda (s)
                                 (handler-for-channel s write-handlers))
                               (second select-result)) )))
    (for-each call-handler read-handlers)
    (for-each call-handler write-handlers) ))

 (define (handler-for-channel channel handlers)
  (let ((res (assq channel handlers)))
    (and res (handler-cont res))))

 (define (call-handler continuation) (continuation))


;; PRELIMINARY
;; SHM Shared Memory IPC facilities (uses SCA POSIX JTEs)
;; TODO shmctl and id descriptor (shmid_ds)

;; System Call Abstraction  Portable Operating System Interface  Jump Table Entries
;; TODO POSIX error signalization, conversion and handling
(define null-void (byte-string-clear! (make-byte-string (system-type-size 'pointer))))
(define null-4    (byte-string-clear! (make-byte-string 4)))
(define perror-bytes (number+sign->byte-string-4 -1))
(define (perror? byte-string) (equal? byte-string perror-bytes))
(define (pand byte-string) (and (not (perror? byte-string))
                                byte-string))

(method create-shared-segment (num-bytes)
  (let* ((id (pand (posix-shmget IPC_PRIVATE
                                 (number->byte-string-8 num-bytes)
                                 (bit-or IPC_CREAT (number->byte-string-4 #o777))))))
   (this (set! 'shared-segments (cons id (this shared-segments))))
   id
))

(method shared-mapping (id num-bytes)
 (let ((addr (posix-shmat id null-void null-4)))
   (and (byte-string-true? addr)
        (byte-string-root-mapping addr num-bytes))))

(method unshare-mapping (byte-string-mapping)
 (posix-shmdt (byte-string-address byte-string-mapping)))

;; Pay attention to remove all ids when applicable:
;; These leak into the current session / control group and
;; into the runtime kernel control structure data of the
;; user account if not removed correctly. Restarting the
;; complete system after mischiefed stress tests is in
;; fact easier than fiddling around with /ipcs/ and /ipcrm/.
;; Do shared segments with an attached count of zero really
;; persist after the end of the process if not removed by it?
;; Seems like.
;; Out of segment situations due to rlimit / cgroup can make the
;; X11 desktop session hang. That's system design, isn't it? LOL
(method destroy-shared-segment (id)
 (or (and (pand (posix-shmctl id IPC_RMID null-void))
          (begin (this (set! 'shared-segments (delete! id (this shared-segments))))
                 #t))
     (error "destroy-shared-mapping provoked an error")))

)) : closes control class
