;; See also Examples/parallels.scm and
;; POSIX Processes in Documentation/Reference.txt.org 

 ;; System Dependencies
;; XXX classes should allow loads into their namespace / environment
(sys-load "VSI-core/posix.sigset.scm")

;; This should be a singleton for the current process!!!
(define parallel-class (class (extension viper-object-class)
 (slots pid signal-handlers)

 (method initialize ()
   (slot-set! (this) 'pid (getpid))
   (slot-set! (this) 'signal-handlers '()))

 (method signal (signal)
   (raise (this pid) signal)) ;; <- This means we need to keep
 ;; track of forks and similar goodies.

 ;;
 ;; SIGNALS
 ;;

 ;; System Dependencies
 ;; XXX classes should allow loads into their namespace / environment
 (sys-load "VSI-core/posix.sigset.scm") ;; XXX ignored
 (define SIG_BLOCK   (number->byte-string-4 0))
 (define SIG_UNBLOCK (number->byte-string-4 1))
 (define SIG_SETMASK (number->byte-string-4 2))

 (define SIG_IGN (number->byte-string 1 (size-of-pointer)))
 (define SIG_DFL (number->byte-string 0 (size-of-pointer)))

 ;; XXX check for an installed signal handler and exchange
 ;; atomically?
 (method add-signal-handler (signal callback)
  (let* ((sigset (sigset-add (asigset) signal))
         (oss    (asigset))
         (chan   (signalfd sigset)))
    (slot-set! (this) 'signal-handlers
      (acons signal chan (this signal-handlers)))
    ;(sys-control (add-read-handler chan callback))
    (sys-control (add-read-handler chan (lambda ()
     (read-bytes 128 chan) ;;XXX size of a sigset
     (callback) )))

    ;; First block the signal ...
    (sigprocmask/unchecked SIG_BLOCK
                           (byte-string-address sigset) 
                           (byte-string-address oss))
 ))

 (method remove-signal-handler (signal)
  (let* ((sigset (sigset-add (asigset) signal))
         (oss    (asigset))
         (oh     #f)
         (chan   (assoc-ref (this signal-handlers) signal)))
  (cond ((not chan)  #f) ;; ??? log-warning
   (else
   ;; ... then ignore and unblock the signal to ...
   ;; ??? should we do this in every roll?
   (set! oh (signal/unchecked (number->byte-string-4 signal) SIG_IGN))

   (sys-control (remove-read-handler chan))
   (slot-set! (this) 'signal-handlers
     (assoc-remove! (this signal-handlers) signal))
   (close chan)

   (sigprocmask/unchecked SIG_UNBLOCK
                          (byte-string-address sigset)
                          (byte-string-address oss))
   ;; ... then restore the previous signal handler.
   (signal/unchecked (number->byte-string-4 signal) oh)
    ))
 ))


 ;;
 ;; PARALLELS (VM)
 ;;

(define parallels-files  '())
(define parallels-pids   '())
(define parallels-output '())

(define (parallel:clear-control-data)
  ;; TODO Oops! Did you close those comports in the subprocess?
  (set! parallels-output '())
  (set! parallels-pids   '())
  (set! parallels-files  '())
)

(define (parallel:subprocess closure iport oport eport error-handler)
  ;; Prepare the error handler.
  ;; XXX error reporting is a mess: /read/ errors are output
  ;; to stderr (sometimes) and there isn't any concept of
  ;; exceptions or exception nesting. CLR3 error objects,
  ;; please! But clr2/a is enough work for now ...
  ;; Then again the /catch/ syntax is nothing but an awful XXX
  (cond
   ((and (pair? error-handler) (procedure? error-handler))
     (set! error-handler (car error-handler)))
   (else
     (set! error-handler (lambda (key . args)
                           (display key)
                           (display ": UNCATCHED THROW for ")
                           (write file) (newline)) )))

  ;; Prepare subprocess input output path.
  ;; XXX Yes, stderr is gone and at some point - after the unification
  ;; of file descriptors and ports and further advances of SCA ...
  ;; (or use dup2 to exchange C stdio as well)
  (let ((laste (set-current-error-port  eport))
        (lasto (set-current-output-port oport))
        (lasti (set-current-input-port  iport)))
    ;; TODO
    (begin (quote a))
    #f)

  ;; Clear control structures of previous parallels.
  (for-each close parallels-output)
  (parallel:clear-control-data)

  ;; Now we can do would we should do. BEWARE of any throws!
  ;; (Yes, I have seen this thing throw away the parent process
  ;;  w/o this catch. How nice, how nice: no oddity that doesn't
  ;;  exist. But with all those nice models around /load/ and
  ;;  /catch/ which could be called nothing but XXX hackery.)
  (catch #t closure (or error-handler (lambda (key . args) (primitive-exit 1))))

  ;; Put an end to this.
  (primitive-exit 0))

(define (parallel:add-process pid)
  (set! parallels-pids   (cons pid parallels-pids)))

;; TODO

(method parallel args ; (closure iport oport eport error-handler)
 (let ((pid (primitive-fork)))
  (if (= 0 pid)
    (begin (slot-set! (this) 'pid (getpid))
           (apply parallel:subprocess args)) ; discontinued exit XXX control?
    (begin (parallel:add-process pid)
           pid))))

;; TODO

;; (define (parallel file . error-handler) ;;; XXX
;;  (let* ((comports (pipe))
;;         (pid      (primitive-fork)))
;;   (setvbuf (cdr comports) _IONBF) ; XXX twice
;;   (if (= 0 pid)
;;    (parallel:subprocess file comports error-handler) ; discontinued branch
;;    (parallel:add-process pid comports file) )))

(define (cancel-all-parallels)
  (for-each (lambda (parallel-id)
              (false-if-exception (kill parallel-id SIGKILL)))
            parallels-pids)
  (parallel:collect-all/blocking)
  (for-each (lambda (port) (false-if-exception (close port)))
            parallels-output)
  (parallel:clear-control-data))

(define (parallel:collect-all/blocking)
  (map (lambda (pid) (waitpid pid )) ; WNOHANG))
       parallels-pids))

)) ; closes parallel class
