#include "include/config.h"

#include "include/_scm.h"

#include "include/alist.h"
#include "include/brass.h"
#include "include/eval.h"
#include "include/procs.h"
#include "include/gsubr.h"
#include "include/root.h"
#include "include/vectors.h"

#include "include/validate.h"
#include "include/procprop.h"


unsigned long global_counter_procprop_fuel = 0;

SCM_GLOBAL_SYMBOL (scm_sym_system_procedure, "system-procedure");
SCM_GLOBAL_SYMBOL (scm_sym_arity, "arity");

SCM
scm_i_procedure_arity (SCM proc)
{
  reference_parameter (proc);

  int a = 0, o = 0, r = 0;
  if (SCM_IMP (proc)) {
    dereference_parameter (proc);
    return SCM_BOOL_F;
  }
 loop:
  switch (SCM_TYP7 (proc))
    {
    case scm_tc7_subr_1o:
      o = 1;
    case scm_tc7_subr_0:
      break;
    case scm_tc7_subr_2o:
      o = 1;
    case scm_tc7_subr_1:
    case scm_tc7_dsubr:
    case scm_tc7_cxr:
      a += 1;
      break;
    case scm_tc7_subr_2:
      a += 2;
      break;
    case scm_tc7_subr_3:
      a += 3;
      break;
    case scm_tc7_asubr:
    case scm_tc7_rpsubr:
    case scm_tc7_lsubr:
      r = 1;
      break;
    case scm_tc7_lsubr_2:
      a += 2;
      r = 1;
      break;
    case scm_tcs_closures:
      proc = rc_X (SCM_CLOSURE_FORMALS (rc_Fly (proc)));
      if (scm_is_null (proc))
	break;
      while (scm_is_pair (proc))
	{
	  ++a;
	  proc = FLY_CDR (proc);
	}
      if (!scm_is_null (proc))
	r = 1;
      break;
    case scm_tc7_subr_n:
      {
        int type = SCM_GSUBR_NARY (proc);
        a += SCM_GSUBR_REQ (type);
        o = SCM_GSUBR_OPT (type);
        r = SCM_GSUBR_REST (type);
      }
      break;
    default:
      dereference_parameter (proc);
      return SCM_BOOL_F;
    }
  dereference_parameter (proc);
  return scm_list_3 (scm_from_int (a), scm_from_int (o), scm_from_bool(r));
}

/* XXX - instead of using a stand-in value for everything except
   closures, we should find other ways to store the procedure
   properties for those other kinds of procedures.  For example, subrs
   have their own property slot, which contains a dummy closure at present.
*/

/* static SCM */
/* scm_stand_in_scm_proc (SCM proc) */
/* { */
/*   reference_parameter (proc); */

/*   SCM handle, answer; */
/*   handle = scm_hashq_get_handle (scm_stand_in_procs, proc); */
/*   reference_parameter (handle); */

/*   if (scm_is_false (handle)) { */
/*       answer = scm_closure (scm_list_2 (SCM_EOL, SCM_BOOL_F), SCM_EOL); */
/*       reference_parameter (answer); */
/*       scm_hashq_set_x (scm_stand_in_procs, proc, answer); */

/*   } else { */
/*       answer = SCM_CDR (handle); */
/*       reference_parameter (answer); */
/*   } */

/*   reference_count_return (answer); */
/*   dereference_parameter (answer); */
/*   dereference_parameter (handle); */
/*   dereference_parameter (proc); */
/*   return (answer); */
/* } */

SCM_DEFINE (scm_procedure_properties, "procedure-properties", 1, 0, 0, 
           (SCM proc),
	    "Return @var{obj}'s property list.")
#define FUNC_NAME s_scm_procedure_properties
{
  SCM_VALIDATE_PROC (1, proc);
  global_counter_procprop_fuel++; 

  reference_parameter (proc);
  SCM tmp = (SCM_CLOSUREP (proc) ? proc : scm_stand_in_scm_proc (proc));
  reference_parameter (tmp); dereference_parameter (proc);
  proc = tmp;  // Reference Count exchange

  tmp = scm_acons (scm_sym_arity, scm_i_procedure_arity (proc),
                   SCM_PROCPROPS (proc));
  reference_count_return (tmp);
  dereference_parameter (proc);
  return (tmp);
}
#undef FUNC_NAME

SCM_DEFINE (scm_set_procedure_properties_x, "set-procedure-properties!", 2, 0, 0,
           (SCM proc, SCM new_val),
	    "Set @var{obj}'s property list to @var{alist}.")
#define FUNC_NAME s_scm_set_procedure_properties_x
{
  reference_parameter (proc);
  reference_parameter (new_val);

  if (!SCM_CLOSUREP (proc))
    proc = rc_X (scm_stand_in_scm_proc (rc_Fly (proc)));
  SCM_VALIDATE_CLOSURE (1, proc);
  SCM_SETPROCPROPS (proc, new_val);

  dereference_parameter (proc);
  dereference_parameter (new_val);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_procedure_property, "procedure-property", 2, 0, 0,
           (SCM p, SCM k),
	    "Return the property of @var{obj} with name @var{key}.")
#define FUNC_NAME s_scm_procedure_property
{
  reference_parameter (p);
  reference_parameter (k);

  if (scm_is_eq (k, scm_sym_arity))
    {
      SCM arity = scm_i_procedure_arity (p);
      reference_parameter (arity);
      SCM_ASSERT (scm_is_true (arity), p, SCM_ARG1, FUNC_NAME);

      reference_count_return (arity);
      dereference_parameter (p);
      dereference_parameter (k);
      return arity;
    }
  SCM_VALIDATE_PROC (1, p);

  SCM tmp;
  SCM assoc;
  tmp = SCM_CLOSUREP (p) ? p : scm_stand_in_scm_proc (p);
  reference_parameter (tmp);  // Reference Count exchange: counts for p
  dereference_parameter (p);
  p = tmp;

  assoc = scm_sloppy_assq (k, SCM_PROCPROPS (p));
  reference_parameter (assoc);
  tmp = (SCM_NIMP (assoc) ? SCM_CDR (assoc) : SCM_BOOL_F);

  reference_count_return (tmp);
  dereference_parameter (assoc);
  dereference_parameter (p);
  dereference_parameter (k);
  return tmp;
}
#undef FUNC_NAME

SCM_DEFINE (scm_set_procedure_property_x, "set-procedure-property!", 3, 0, 0,
           (SCM p, SCM k, SCM v),
	    "In @var{obj}'s property list, set the property named @var{key} to\n"
	    "@var{value}.")
#define FUNC_NAME s_scm_set_procedure_property_x
{
  reference_parameter (p);
  reference_parameter (k);
  reference_parameter (v);

  if (!SCM_CLOSUREP (p))
    p = rc_X (scm_stand_in_scm_proc (rc_Fly (p)));

  SCM_VALIDATE_CLOSURE (1, p);
  if (scm_is_eq (k, scm_sym_arity))
    {
      SCM_MISC_ERROR ("arity is a read-only property", SCM_EOL);
    }

  SCM assoc = scm_sloppy_assq (k, SCM_PROCPROPS (p));
  reference_parameter (assoc);

  if (SCM_NIMP (assoc)) {
    SCM_SETCDR (assoc, v);
  } else {
    global_counter_procprop_fuel++; 
    SCM_SETPROCPROPS (p, scm_acons (k, v, SCM_PROCPROPS (p)));
  }

  dereference_parameter (assoc);
  dereference_parameter (p);
  dereference_parameter (k);
  dereference_parameter (v);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME




void
x_inclusion_procprop ()
{
#include "include/procprop.x"
}
