// Strings & Brass
// TODO (replace unknown buffer magicals) <- io path revision
// Originally the SCM part of VSI aka VSI-C - still before the times
// of the ugly cow - had been written by a hobbit. They're still 
// in search of help against the evil ring spell: please help'em
// and cut any rings you create.

#include "include/config.h"

#include "include/_scm.h"

#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "include/eval.h"
#include "include/ports.h"
#include "include/read.h"
#include "include/root.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/validate.h"

#include "include/strports.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif


scm_t_tag scm_tc16_strport;


static int
stfill_buffer (SCM port)  // ??? is this a misnomer
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  
  if (pt->read_pos >= pt->read_end)
    return EOF;
  else
    return (*pt->read_pos);
}

/* change the size of a port's string to new_size.  this doesn't
   change read_buf_size.  */
static void 
st_resize_port (scm_t_port *pt, off_t new_size)
{
  SCM old_stream = SCM_PACK (pt->stream);
  reference_parameter (old_stream);
  const char *src = BYTE_STRING_DATA (old_stream);
  SCM new_stream = scm_c_make_byte_string (new_size);
  char *dst = BYTE_STRING_DATA (new_stream);
  reference_parameter (new_stream);
  unsigned long int old_size = BYTE_STRING_LENGTH (old_stream);
  unsigned long int min_size = min (old_size, new_size);
  unsigned long int i;

  off_t index = pt->write_pos - pt->write_buf;

  pt->write_buf_size = new_size;

  memcpy (dst, src, min_size);
  i = min_size;
  /* for (i = 0; i != min_size; ++i) // XXX what a fast loop here */
  /*   dst[i] = src[i]; */


  /* reset buffer. */
  {
    pt->stream = SCM_UNPACK (new_stream);
    heap_reference (pt, new_stream);
    heap_dereference (pt, old_stream);
    pt->read_buf = pt->write_buf = (unsigned char *)dst;
    pt->read_pos = pt->write_pos = pt->write_buf + index;
    pt->write_end = pt->write_buf + pt->write_buf_size;
    pt->read_end = pt->read_buf + pt->read_buf_size;
  }
  dereference_parameter (old_stream);
  dereference_parameter (new_stream);
  return;
}

/* amount by which write_buf is expanded.  */
#define SCM_WRITE_BLOCK 4000000

/* ensure that write_pos < write_end by enlarging the buffer when
   necessary.  update read_buf to account for written chars.

   The buffer is enlarged by 1.5 times, plus SCM_WRITE_BLOCK.  Adding just a
   fixed amount is no good, because there's a block copy for each increment,
   and that copying would take quadratic time.  In the past it was found to
   be very slow just adding 80 bytes each time (eg. about 10 seconds for
   writing a 100kbyte string). [LOL Exactly! Escalation has been deprecated
   nonetheless. Go search your desktop in the swap area! LOL] */

static void
st_flush (SCM port)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);

  if (pt->write_pos == pt->write_end)
    {
      st_resize_port (pt, pt->write_buf_size + SCM_WRITE_BLOCK);
    }
  pt->read_pos = pt->write_pos;
  if (pt->read_pos > pt->read_end)
    {
      pt->read_end = (unsigned char *) pt->read_pos;
      pt->read_buf_size = pt->read_end - pt->read_buf;
    }
  pt->rw_active = SCM_PORT_NEITHER;
}

static void
st_write (SCM port, const void *data, size_t size)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  const char *input = (char *) data;

  while (size > 0)
    {
      int space = pt->write_end - pt->write_pos;
      int write_len = (size > space) ? space : size;
      
      memcpy ((char *) pt->write_pos, input, write_len);
      pt->write_pos += write_len;
      size -= write_len;
      input += write_len;
      if (write_len == space)
	st_flush (port);
    }
}

static void
st_end_input (SCM port, int offset)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  
  if (pt->read_pos - pt->read_buf < offset)
    scm_misc_error ("st_end_input", "negative position", SCM_EOL);

  pt->write_pos = (unsigned char *) (pt->read_pos = pt->read_pos - offset);
  pt->rw_active = SCM_PORT_NEITHER;
}

static off_t
st_seek (SCM port, off_t offset, int whence)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  off_t target;

  if (pt->rw_active == SCM_PORT_READ && offset == 0 && whence == SEEK_CUR)
    /* special case to avoid disturbing the unread-char buffer.  */
    {
      if (pt->read_buf == pt->putback_buf)
	{
	  target = pt->saved_read_pos - pt->saved_read_buf
	    - (pt->read_end - pt->read_pos);
	}
      else
	{
	  target = pt->read_pos - pt->read_buf;
	}
    }
  else
    /* all other cases.  */
    {
      if (pt->rw_active == SCM_PORT_WRITE)
	st_flush (port);
  
      if (pt->rw_active == SCM_PORT_READ)
	scm_end_input (port);

      switch (whence)
	{
	case SEEK_CUR:
	  target = pt->read_pos - pt->read_buf + offset;
	  break;
	case SEEK_END:
	  target = pt->read_end - pt->read_buf + offset;
	  break;
	default: /* SEEK_SET */
	  target = offset;
	  break;
	}

      if (target < 0)
	scm_misc_error ("st_seek", "negative offset", SCM_EOL);
  
      if (target >= pt->write_buf_size)
	{
	  if (!(SCM_CELL_WORD_0 (port) & SCM_WRTNG))
	    {
	      if (target > pt->write_buf_size)
		{
		  scm_misc_error ("st_seek", 
				  "seek past end of read-only strport",
				  SCM_EOL);
		}
	    }
	  else
	    {
	      st_resize_port (pt, target + (target == pt->write_buf_size
					    ? SCM_WRITE_BLOCK
					    : 0));
	    }
	}
      pt->read_pos = pt->write_pos = pt->read_buf + target;
      if (pt->read_pos > pt->read_end)
	{
	  pt->read_end = (unsigned char *) pt->read_pos;
	  pt->read_buf_size = pt->read_end - pt->read_buf;
	}
    }
  return target;
}

static void
st_truncate (SCM port, off_t length)
{
  scm_t_port *pt = SCM_PTAB_ENTRY (port);

  if (length > pt->write_buf_size)
    st_resize_port (pt, length);

  pt->read_buf_size = length;
  pt->read_end = pt->read_buf + length;
  if (pt->read_pos > pt->read_end)
    pt->read_pos = pt->read_end;
  
  if (pt->write_pos > pt->read_end)
    pt->write_pos = pt->read_end;
}

SCM 
scm_mkstrport (SCM str_arg, long modes, const char *caller)
{
  reference_parameter (str_arg);

  SCM str;
  SCM z;
  scm_t_port *pt;
  size_t str_len, c_pos;

  SCM_ASSERT (BYTE_STRING_P (str_arg), str_arg, SCM_ARG1, caller);

  str_len = BYTE_STRING_LENGTH (str_arg);
  c_pos = 0;

  if (!((modes & SCM_WRTNG) || (modes & SCM_RDNG))) {
    scm_misc_error ("scm_mkstrport",
                    "port must be or readable or writable not both",
                    SCM_EOL);
  }

  // if (modes & SCM_WRTNG)
  // We always copy.
  str = scm_byte_string_copy (str_arg);
  //else
  //  str = scm_c_substring (str, 0, str_len);
  reference_parameter (str);

  z = scm_new_port_table_entry (scm_tc16_strport);
  pt = SCM_PTAB_ENTRY(z);
  SCM_SETSTREAM (z, SCM_UNPACK (str));
  SCM_SET_CELL_WORD_0(z, modes); // flags
  reference (z, str);

  pt->write_buf = pt->read_buf = (unsigned char *) BYTE_STRING_DATA (str);
  pt->read_pos = pt->write_pos = pt->read_buf + c_pos;
  pt->write_buf_size = pt->read_buf_size = str_len;
  pt->write_end = pt->read_end = pt->read_buf + pt->read_buf_size;

  pt->rw_random = 1;

  /* ensure write_pos is writable. */
  if ((modes & SCM_WRTNG) && pt->write_pos == pt->write_end)
    st_flush (z);

  dereference_parameter (str);
  dereference_parameter (str_arg);
  return z;
}

/* create a new string from a string port's buffer.  */
SCM scm_strport_to_string (SCM port)
{
  reference_parameter (port);

  scm_t_port *pt = SCM_PTAB_ENTRY (port);
  SCM str;
  char *dst;
  
  if (pt->rw_active == SCM_PORT_WRITE)
    st_flush (port);

  str = scm_make_byte_string_with_copy (pt->read_buf, pt->read_buf_size);
  // str = scm_i_make_string (pt->read_buf_size, &dst);
  // memcpy (dst, (char *) pt->read_buf, pt->read_buf_size);

  dereference_parameter (port);
  return str;
}

SCM_DEFINE (scm_object_to_string, "object->string", 1, 1, 0,
	    (SCM obj, SCM printer),
	    "Return a Scheme string obtained by printing @var{obj}.\n"
	    "Printing function can be specified by the optional second\n"
	    "argument @var{printer} (default: @code{write}).")
#define FUNC_NAME s_scm_object_to_string
{
  reference_parameter (obj);
  reference_parameter (printer);

  SCM str, port;

  if (!SCM_UNBNDP (printer))
    SCM_VALIDATE_PROC (2, printer);

  str = scm_c_make_byte_string (0);  // RC pass through
  port = scm_mkstrport (str, SCM_OPN | SCM_WRTNG, FUNC_NAME);

  if (SCM_UNBNDP (printer))
    scm_write (obj, port);
  else
    rc_Void (scm_call_2 (printer, obj, port));

  SCM answer = scm_strport_to_string (port);
  reference_parameter (answer);
  scm_close_port (port);
  dereference_parameter (obj);
  dereference_parameter (printer);
  return (rc_return_stack_reference (answer));
}
#undef FUNC_NAME

SCM_DEFINE (scm_call_with_output_string, "call-with-output-string", 1, 0, 0, 
           (SCM proc), "")
#define FUNC_NAME s_scm_call_with_output_string
{
  reference_parameter (proc);

  SCM p;

  p = scm_mkstrport (scm_c_make_byte_string (0),
		     SCM_OPN | SCM_WRTNG,
                     FUNC_NAME);
  reference_parameter (p);
  scm_call_1 (proc, p);

  SCM answer = scm_get_output_string (p);
  reference_count_return (answer);

  scm_close_port (p);
  dereference_parameter (p);
  dereference_parameter (proc);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_call_with_input_string, "call-with-input-string", 2, 0, 0,
           (SCM string, SCM proc),
	    "Calls the one-argument procedure @var{proc} with a newly\n"
	    "created input port from which @var{string}'s contents may be\n"
	    "read.  The value yielded by the @var{proc} is returned.")
#define FUNC_NAME s_scm_call_with_input_string
{
  reference_parameter (string);
  reference_parameter (proc);

  SCM p = scm_mkstrport(string, SCM_OPN | SCM_RDNG, FUNC_NAME);
  reference_parameter (p);

  SCM answer = scm_call_1 (proc, p);
  reference_count_return (answer);

  scm_close_port (p);
  dereference_parameter (p);
  dereference_parameter (string);
  dereference_parameter (proc);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE (scm_open_input_string, "open-input-string", 1, 0, 0,
	    (SCM str),
	    "Take a string and return an input port that delivers characters\n"
	    "from the string. The port can be closed by\n"
	    "@code{close-input-port}, though its storage will be reclaimed\n"
	    "by the garbage collector if it becomes inaccessible.")
#define FUNC_NAME s_scm_open_input_string
{
  SCM p = scm_mkstrport(str, SCM_OPN | SCM_RDNG, FUNC_NAME);
  return p;
}
#undef FUNC_NAME

SCM_DEFINE (scm_open_output_string, "open-output-string", 0, 0, 0, 
	    (void),
	    "Return an output port that will accumulate characters for\n"
	    "retrieval by @code{get-output-string}. The port can be closed\n"
	    "by the procedure @code{close-output-port}, though its storage\n"
	    "will be reclaimed by the garbage collector if it becomes\n"
	    "inaccessible.")
#define FUNC_NAME s_scm_open_output_string
{
  SCM p;

  p = scm_mkstrport (scm_c_make_byte_string (0),
		     SCM_OPN | SCM_WRTNG,
                     FUNC_NAME);
  return p;
}
#undef FUNC_NAME

SCM_DEFINE (scm_get_output_string, "get-output-string", 1, 0, 0, 
	    (SCM port),
	    "Given an output port created by @code{open-output-string},\n"
	    "return a string consisting of the characters that have been\n"
	    "output to the port so far.")
#define FUNC_NAME s_scm_get_output_string
{
  SCM_VALIDATE_OPOUTSTRPORT (1, port);
  return scm_strport_to_string (port);
}
#undef FUNC_NAME


/* Given a null-terminated string EXPR containing a Scheme expression
   read it, and return it as an SCM value. */
SCM
scm_c_read_string (const char *expr)
{
  SCM port = scm_mkstrport (scm_from_sen (expr, strlen (expr)),
			    SCM_OPN | SCM_RDNG,
			    "scm_c_read_string");
  reference_parameter (port);

  /* Read expressions from that port; ignore the values.  */
  SCM form = scm_read (port);

  scm_close_port (port);
  dereference_parameter (port);
  return (form);
}

static void
strport_release (SCM string_port)
{
  dereference (string_port, SCM_PACK (SCM_STREAM (string_port)));
  SCM_SETSTREAM (string_port, SCM_UNPACK (SCM_UNSPECIFIED));
  return;
}

static scm_t_tag
scm_make_stptob ()
{
  scm_t_tag tc = scm_make_port_type ("string", stfill_buffer, st_write);

  scm_set_port_release     (tc, strport_release);
  scm_set_port_end_input   (tc, st_end_input);
  scm_set_port_flush       (tc, st_flush);
  scm_set_port_seek        (tc, st_seek);
  scm_set_port_truncate    (tc, st_truncate);

  return tc;
}

void
scm_init_strports ()
{
  scm_tc16_strport = scm_make_stptob ();
}

void
x_inclusion_strports ()
{
#include "include/strports.x"
}
