#include "include/config.h"
#include "include/ihub.h"
#include "include/scmconfig.h"
#include "include/libpath.h"
#include "include/debug.h"
#include <unistd.h>


/*
 * Debugging: System Information
 */

static void byte_bit_printer (const unsigned long byte)
{
  unsigned long mask;
  for (mask = (1 << 7); mask >= 1; mask = (mask >> 1)) {
    if (mask & (1 << 6)) fprintf (stderr, " ");
    if (mask & (1 << 2)) fprintf (stderr, " ");
    fprintf (stderr, "%d", ((mask & byte) ? 1 : 0));
  }
}

static void second_byte_bit_printer (const unsigned long thing)
{
  unsigned long mask;
  unsigned long byte;
  byte = thing >> 8;

  for (mask = (1 << 7); mask >= 1; mask = (mask >> 1)) {
    if (mask & (1 << 3)) fprintf (stderr, " ");
    fprintf (stderr, "%d", ((mask & byte) ? 1 : 0));
  }
}

void debug_print_smob_info ()
{
  long i; 
  fprintf (stderr, "\nPS Viper System Information (interpreter-main.c)");
  fprintf (stderr, "\n * Registered SMOB Types *\n");

  fprintf (stderr, " Number of registered SMOBS: %ld\n", scm_numsmob);
  
  for (i = 0; i < scm_numsmob; i++)
    {
      fprintf (stderr, "  SMOB #%ld: %s (cell: %d)\n",
               i, scm_smobs[i].name,
               scm_smobs[i].cell_size);
    }
  fprintf (stderr, "- - - end of vsi system information post scriptum - - -\n");
}

#include <signal.h>
void debug_print_system_info (char **arg1)
{
  fprintf (stderr, "VSI: Viper System Information (interpreter-main.c)\n\n");
  fprintf (stderr, " * Internal Data Type Sizes in Bytes *\n");

  fprintf (stderr, " SCM: %lu", sizeof (SCM));
  fprintf (stderr, ", scm_t_bits: %lu", sizeof (scm_t_bits));
  fprintf (stderr, ", pointers: %lu", sizeof (void *));
  fprintf (stderr, ", SCM_EOL: %lu\n", sizeof (SCM_EOL));

  fprintf (stderr, " int: %lu", sizeof (int));
  fprintf (stderr, ", short: %lu", sizeof (short));
  fprintf (stderr, ", long: %lu", sizeof (long));
  fprintf (stderr, ", long long: %lu\n", sizeof (long long));
  fprintf (stderr, ", sigset_t: %lu\n", sizeof (sigset_t));

  fprintf (stderr, "\n * Data Type Bit Masks *\n");
  fprintf (stderr, "                   8 tc7  tc3\n");
  fprintf (stderr, "     scm_tc3_cons: "); byte_bit_printer (scm_tc3_cons); fprintf (stderr, "\n");
  fprintf (stderr, "  scm_tc3_closure: "); byte_bit_printer (scm_tc3_closure); fprintf (stderr, "\n");

  fprintf (stderr, "\n");
  fprintf (stderr, "    scm_tc3_imm24: "); byte_bit_printer (scm_tc3_imm24); fprintf (stderr, "\n");
  fprintf (stderr, "     scm_tc8_flag: "); byte_bit_printer (scm_tc8_flag); fprintf (stderr, "\n");
  fprintf (stderr, "     scm_tc8_char: "); byte_bit_printer (scm_tc8_char); fprintf (stderr, "\n");
  fprintf (stderr, "     scm_tc8_isym: "); byte_bit_printer (scm_tc8_isym); fprintf (stderr, "\n");
  fprintf (stderr, "     scm_tc8_iloc: "); byte_bit_printer (scm_tc8_iloc); fprintf (stderr, "\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) SCM_BOOL_F); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) SCM_BOOL_F); fprintf (stderr, " (SCM_BOOL_F)\n");
  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) SCM_BOOL_T); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) SCM_BOOL_T); fprintf (stderr, " (SCM_BOOL_T)\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) SCM_UNDEFINED); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) SCM_UNDEFINED); fprintf (stderr, " (SCM_UNDEFINED)\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) SCM_EOF_VAL); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) SCM_EOF_VAL); fprintf (stderr, " (SCM_EOF_VAL)\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) SCM_EOL); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) SCM_EOL); fprintf (stderr, " (SCM_EOL)\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) SCM_UNSPECIFIED); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) SCM_UNSPECIFIED); fprintf (stderr, " (SCM_UNSPECIFIED)\n");


  fprintf (stderr, "\n");
  fprintf (stderr, "    scm_tc3_tc7_1: "); byte_bit_printer (scm_tc3_tc7_1); fprintf (stderr, "\n");
  fprintf (stderr, "    scm_tc3_tc7_2: "); byte_bit_printer (scm_tc3_tc7_2); fprintf (stderr, "\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "   scm_tc7_subr_n: "); byte_bit_printer (scm_tc7_subr_n); fprintf (stderr, "\n");
  fprintf (stderr, "    scm_tc7_lsubr: "); byte_bit_printer (scm_tc7_lsubr); fprintf (stderr, "\n");
  fprintf (stderr, "    scm_tc7_asubr: "); byte_bit_printer (scm_tc7_asubr); fprintf (stderr, "\n");
  fprintf (stderr, "   scm_tc7_symbol: "); byte_bit_printer (scm_tc7_symbol); fprintf (stderr, "\n");

  fprintf (stderr, " scm_tc7_variable: "); byte_bit_printer (scm_tc7_variable); fprintf (stderr, "\n");
  fprintf (stderr, "   scm_tc7_vector: "); byte_bit_printer (scm_tc7_vector); fprintf (stderr, "\n");
  fprintf (stderr, "    scm_tc7_brass: "); byte_bit_printer (scm_tc7_brass); fprintf (stderr, "\n");
  fprintf (stderr, "scm_tc7_byte_string: "); byte_bit_printer (scm_tc7_byte_string); fprintf (stderr, "\n");
  fprintf (stderr, "     scm_tc7_port: "); byte_bit_printer (scm_tc7_port); fprintf (stderr, "\n");
  fprintf (stderr, "\n");
  fprintf (stderr, "   scm_tc7_number: "); byte_bit_printer (scm_tc7_number); fprintf (stderr, "\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) scm_tc16_real); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) scm_tc16_real); fprintf (stderr, " (scm_tc16_real)\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) scm_tc16_fraction); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) scm_tc16_fraction); fprintf (stderr, " (scm_tc16_fraction)\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) scm_tc16_complex); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) scm_tc16_complex); fprintf (stderr, " (scm_tc16_complex)\n");

  fprintf (stderr, "     ... ");  second_byte_bit_printer ((unsigned long) scm_tc16_big); fprintf (stderr, " ");
  byte_bit_printer ((unsigned long) scm_tc16_big); fprintf (stderr, " (scm_tc16_big)\n");


  fprintf (stderr, "tc16:scm_tc7_smob: "); byte_bit_printer (scm_tc7_smob); fprintf (stderr, "\n");

  fprintf (stderr, "\n * Default Resource Folders *\n");
  /* fprintf (stderr, " current: %s (relative load is enabled by default)\n", */
  /*           get_current_dir_name ()); */
  fprintf (stderr, " package: %s\n", SCM_PKGDATA_DIR);
  fprintf (stderr, " library: %s\n", SCM_LIBRARY_DIR);
  fprintf (stderr, "    site: %s\n", SCM_SITE_DIR);
  fprintf (stderr, " initial load files: init-9.scm\n");

  fprintf (stderr, "\n- - - end of vsi system information - - -\n\n");
  return;
}
