#include "include/config.h"

#include <stdio.h>
#include "include/_scm.h"
#include "include/procprop.h"
#include "include/root.h"

#include "include/gsubr.h"

/*
 * gsubr.c
 * Provide `gsubrs' -- subrs taking a prescribed number of required, optional,
 * and rest arguments.
 */

/* #define GSUBR_TEST */

SCM_GLOBAL_SYMBOL (scm_sym_name, "name");

#include "include/procs.h"
static SCM
create_gsubr (int define, const char *name,
	      int req, int opt, int rst, SCM (*fcn)())
{
  SCM subr;

  switch (SCM_GSUBR_MAKTYPE (req, opt, rst))
    {
    case SCM_GSUBR_MAKTYPE(0, 0, 0):
      subr = scm_c_make_subr (name, scm_tc7_subr_0, fcn);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(1, 0, 0):
      subr = scm_c_make_subr (name, scm_tc7_subr_1, fcn);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(0, 1, 0):
      subr = scm_c_make_subr (name, scm_tc7_subr_1o, fcn);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(1, 1, 0):
      subr = scm_c_make_subr (name, scm_tc7_subr_2o, fcn);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(2, 0, 0):
      subr = scm_c_make_subr (name, scm_tc7_subr_2, fcn);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(3, 0, 0):
      subr = scm_c_make_subr (name, scm_tc7_subr_3, fcn);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(0, 0, 1):
      subr = scm_c_make_subr (name, scm_tc7_lsubr, fcn);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(2, 0, 1):
      subr = scm_c_make_subr (name, scm_tc7_lsubr_2, fcn);
    create_subr:
      if (define)
	scm_define (SCM_SUBR_ENTRY(subr).name, subr);
      return subr;
    default:
      {
        // fprintf (stderr, "create_gsubr: %s\n", name);
	SCM subr = scm_c_make_subr (name, scm_tc7_subr_n, fcn);
	SCM sym = SCM_SUBR_ENTRY(subr).name;
	if (SCM_GSUBR_MAX < req + opt + rst)
	  {
            fprintf (stderr,
                     "ERROR in scm_c_make_gsubr: too many args (%d) to %s\n",
                     req + opt + rst, name);
	    exit (1);
	  }
        SCM_SET_GSUBR_NARY (subr, SCM_GSUBR_MAKTYPE (req, opt, rst));
        scm_set_procedure_property_x (subr, scm_sym_name, sym);
	if (define)
	  scm_define (sym, subr);
      return subr;
      }
    }
}

SCM
scm_c_make_gsubr (const char *name, int req, int opt, int rst, SCM (*fcn)())
{
  return create_gsubr (0, name, req, opt, rst, fcn);
}

SCM
scm_c_define_gsubr (const char *name, int req, int opt, int rst, SCM (*fcn)())
{
  return create_gsubr (1, name, req, opt, rst, fcn);
}

static SCM
create_gsubr_with_generic (int define,
			   const char *name,
			   int req,
			   int opt,
			   int rst,
			   SCM (*fcn)(),
			   SCM *gf)
{
  SCM subr;

  switch (SCM_GSUBR_MAKTYPE(req, opt, rst))
    {
    case SCM_GSUBR_MAKTYPE(0, 0, 0):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_subr_0, fcn, gf);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(1, 0, 0):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_subr_1, fcn, gf);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(0, 1, 0):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_subr_1o, fcn, gf);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(1, 1, 0):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_subr_2o, fcn, gf);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(2, 0, 0):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_subr_2, fcn, gf);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(3, 0, 0):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_subr_3, fcn, gf);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(0, 0, 1):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_lsubr, fcn, gf);
      goto create_subr;
    case SCM_GSUBR_MAKTYPE(2, 0, 1):
      subr = scm_c_make_subr_with_generic (name, scm_tc7_lsubr_2, fcn, gf);
    // ??? do we need to add n-ary scm_tc7_subr_n here ???
    create_subr:
      if (define)
	scm_define (SCM_SUBR_ENTRY(subr).name, subr);
      return subr;
    default:
      ;
    }
  scm_misc_error ("scm_c_make_gsubr_with_generic",
		  "can't make primitive-generic with this arity",
		  SCM_EOL);
  return SCM_BOOL_F; /* never reached */
}

SCM
scm_c_make_gsubr_with_generic (const char *name,
			       int req,
			       int opt,
			       int rst,
			       SCM (*fcn)(),
			       SCM *gf)
{
  return create_gsubr_with_generic (0, name, req, opt, rst, fcn, gf);
}

SCM
scm_c_define_gsubr_with_generic (const char *name,
				 int req,
				 int opt,
				 int rst,
				 SCM (*fcn)(),
				 SCM *gf)
{
  return create_gsubr_with_generic (1, name, req, opt, rst, fcn, gf);
}


SCM
scm_i_gsubr_apply (SCM proc, SCM args)
#define FUNC_NAME "scm_i_gsubr_apply"
{
  reference_parameter (proc);
  reference_parameter (args);

  SCM (*fcn)() = SCM_SUBRF (proc);
  SCM v[SCM_GSUBR_MAX];
  int typ = SCM_GSUBR_NARY (proc);
  long i, n = SCM_GSUBR_REQ (typ) + SCM_GSUBR_OPT (typ) + SCM_GSUBR_REST (typ);

#if 0
  if (n > SCM_GSUBR_MAX)
    scm_misc_error (FUNC_NAME,
		    "Function ~S has illegal arity ~S.",
		    scm_list_2 (proc, scm_from_int (n)));
#endif
  for (i = 0; i < SCM_GSUBR_REQ (typ); i++) {
    if (scm_is_null (args))
      scm_wrong_num_args (SCM_SNAME (proc));
    v[i] = SCM_CAR(args);
    args = FLY_CDR(args);
  }
  for (; i < SCM_GSUBR_REQ (typ) + SCM_GSUBR_OPT (typ); i++) {
    if (SCM_NIMP (args)) {
      v[i] = SCM_CAR (args);
      args = FLY_CDR(args);
    }
    else
      v[i] = SCM_UNDEFINED;
  }
  if (SCM_GSUBR_REST(typ)) {
    v[i] = args;
    reference_count_return (args);
  }
  else if (!scm_is_null (args))
    scm_wrong_num_args (SCM_SNAME (proc));

  // The arguments are referenced as return values by now.
  dereference_parameter (args);

  SCM answer = SCM_UNDEFINED;
  switch (n) {
  case 2: answer = (*fcn)(v[0], v[1]); break;
  case 3: answer = (*fcn)(v[0], v[1], v[2]); break;
  case 4: answer = (*fcn)(v[0], v[1], v[2], v[3]); break;
  case 5: answer = (*fcn)(v[0], v[1], v[2], v[3], v[4]); break;
  case 6: answer = (*fcn)(v[0], v[1], v[2], v[3], v[4], v[5]); break;
  case 7:
    answer = (*fcn)(v[0], v[1], v[2], v[3], v[4], v[5], v[6]); break;
  case 8:
    answer = (*fcn)(v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7]); break;
  case 9:
    answer = (*fcn)(v[0], v[1], v[2], v[3], v[4], v[5], v[6], v[7], v[8]);
    break;
  case 10:
    answer = (*fcn)(v[0], v[1], v[2], v[3], v[4], \
                    v[5], v[6], v[7], v[8], v[9]);
    break;
  }

  reference_count_return (answer);
  dereference_parameter (proc);
  return answer;
}
#undef FUNC_NAME


#ifdef GSUBR_TEST
/* A silly example, taking 2 required args, 1 optional, and
   a scm_list of rest args
   */
SCM
gsubr_21l(SCM req1, SCM req2, SCM opt, SCM rst)
{
  reference_parameter (req1);
  reference_parameter (req2);
  reference_parameter (opt);
  reference_parameter (rst);

  scm_puts ("gsubr-2-1-l:\n req1: ", scm_cur_outp);
  scm_display(req1, scm_cur_outp);
  scm_puts ("\n req2: ", scm_cur_outp);
  scm_display(req2, scm_cur_outp);
  scm_puts ("\n opt: ", scm_cur_outp);
  scm_display(opt, scm_cur_outp);
  scm_puts ("\n rest: ", scm_cur_outp);
  scm_display(rst, scm_cur_outp);
  scm_newline(scm_cur_outp);

  dereference_parameter (req1);
  dereference_parameter (req2);
  dereference_parameter (opt);
  dereference_parameter (rst);
  return SCM_UNSPECIFIED;
}
#endif


void
x_inclusion_gsubr()
{
#ifdef GSUBR_TEST
  /* example */
  rc_Void (scm_c_define_gsubr ("gsubr-2-1-l", 2, 1, 1, gsubr_21l));
#endif

#include "include/gsubr.x"
}
