/* classes: h_files */

#ifndef SCM_EVAL_H
#define SCM_EVAL_H
#include "include/__scm.h"


/* {Ilocs}
 *
 * Ilocs are relative pointers into local environment structures.
 * 
 */
#define SCM_ILOCP(n)		(SCM_ITAG8(n)==scm_tc8_iloc)

/*
 * Environments
 */
// CEVAL: Save proc in old_proc for every goto upwards.
// applicator: tail call to CEVAL; needs to let fly the env though
// scm_foreach: seems to be fast anyway but doesn't allow for reuse if
//              not with a complete rewrite
// All the trampoline calls are like that.
//
// [Reusing environment frames]
// This is only an idea and there was some case where this wouldn't have
// worked out but I forgot.
//
#define SCM_REUSE_ENV_P (old_proc, new_proc, env) \
 ((old_proc == new_proc) && (env != SCM_EOL) && (TRIM_REFCOUNT (env) == 2))


/* {Evaluator}
 */

typedef SCM (*scm_t_trampoline_0) (SCM proc);
typedef SCM (*scm_t_trampoline_1) (SCM proc, SCM arg1);
typedef SCM (*scm_t_trampoline_2) (SCM proc, SCM arg1, SCM arg2);



unsigned long global_counter_ceval_fuel;
SCM ceval_fuel (SCM a, SCM b, SCM c);
#define SCM_EXTEND_ENV ceval_fuel

/*fixme* This should probably be removed throught the code. */

#define SCM_TOP_LEVEL_LOOKUP_CLOSURE (scm_current_module_lookup_closure())



SCM scm_sym_and;
SCM scm_sym_begin;
SCM scm_sym_case;
SCM scm_sym_cond;
SCM scm_sym_define;
SCM scm_sym_if;
SCM scm_sym_lambda;
SCM scm_sym_let;
SCM scm_sym_letstar;
SCM scm_sym_letrec;
SCM scm_sym_quote;
SCM scm_sym_quasiquote;
SCM scm_sym_unquote;
SCM scm_sym_uq_splicing;

SCM scm_sym_atapply;
// SCM scm_sym_atcall_cc;
// SCM scm_sym_at_call_with_values;
SCM scm_sym_arrow;
SCM scm_sym_else;
SCM scm_sym_apply;
SCM scm_sym_set_x;
SCM scm_sym_args;



SCM * scm_ilookup (SCM iloc, SCM env);
SCM * scm_lookupcar (SCM vloc, SCM genv, int check);
SCM scm_eval_car (SCM pair, SCM env);
SCM scm_eval_body (SCM code, SCM env);
SCM scm_eval_args (SCM i, SCM env, SCM proc);
SCM scm_m_quote (SCM xorig, SCM env);
SCM scm_m_begin (SCM xorig, SCM env);
SCM scm_m_if (SCM xorig, SCM env);
SCM scm_m_set_x (SCM xorig, SCM env);
SCM scm_m_vref (SCM xorig, SCM env); // ???
SCM scm_m_vset (SCM xorig, SCM env); // ???
SCM scm_m_and (SCM xorig, SCM env);
SCM scm_m_or (SCM xorig, SCM env);
SCM scm_m_case (SCM xorig, SCM env);
SCM scm_m_cond (SCM xorig, SCM env);
SCM scm_m_lambda (SCM xorig, SCM env);
SCM scm_m_letstar (SCM xorig, SCM env);
SCM scm_m_quasiquote (SCM xorig, SCM env);
SCM scm_m_generalized_set_x (SCM xorig, SCM env); // ???
SCM scm_m_define (SCM x, SCM env);
SCM scm_m_letrec (SCM xorig, SCM env);
SCM scm_m_let (SCM xorig, SCM env);
SCM scm_m_apply (SCM xorig, SCM env);
SCM scm_m_atbind (SCM xorig, SCM env); // ???
SCM scm_m_atdispatch (SCM xorig, SCM env); // ???
int scm_badargsp (SCM formals, SCM args);
SCM scm_call_0 (SCM proc);
SCM scm_call_1 (SCM proc, SCM arg1);
SCM scm_call_2 (SCM proc, SCM arg1, SCM arg2);
SCM scm_call_3 (SCM proc, SCM arg1, SCM arg2, SCM arg3);
SCM scm_call_4 (SCM proc, SCM arg1, SCM arg2, SCM arg3, SCM arg4);
SCM scm_apply_0 (SCM proc, SCM args);
SCM scm_apply_1 (SCM proc, SCM arg1, SCM args);
SCM scm_apply_2 (SCM proc, SCM arg1, SCM arg2, SCM args);
SCM scm_apply_3 (SCM proc, SCM arg1, SCM arg2, SCM arg3, SCM args);
SCM scm_i_call_closure_0 (SCM proc);
scm_t_trampoline_0 scm_trampoline_0 (SCM proc);
scm_t_trampoline_1 scm_trampoline_1 (SCM proc);
scm_t_trampoline_2 scm_trampoline_2 (SCM proc);
SCM scm_nconc2last (SCM lst);
SCM scm_apply (SCM proc, SCM arg1, SCM args);
SCM scm_dapply (SCM proc, SCM arg1, SCM args);
SCM scm_map (SCM proc, SCM arg1, SCM args);
SCM scm_for_each (SCM proc, SCM arg1, SCM args);
SCM scm_closure (SCM formals, SCM body, SCM env);
SCM scm_cons_source (SCM xorig, SCM x, SCM y);
SCM scm_copy_tree (SCM obj);
SCM scm_i_eval_x (SCM exp, SCM env);
SCM scm_i_eval (SCM exp, SCM env);
SCM scm_primitive_eval (SCM exp);
SCM scm_primitive_eval_x (SCM exp);
SCM scm_eval (SCM exp, SCM module);
SCM scm_eval_x (SCM exp, SCM module);
SCM sys_eval (SCM exp);

void scm_i_print_iloc (SCM /*iloc*/, SCM /*port*/);
void scm_i_print_isym (SCM /*isym*/, SCM /*port*/);
SCM scm_i_unmemocopy_expr (SCM expr, SCM env);
SCM scm_i_unmemocopy_body (SCM forms, SCM env);
void scm_init_eval (void);
void x_inclusion_eval (void);


#endif  /* SCM_EVAL_H */
