/* classes: h_files */

#ifndef SCM_PRINT_H
#define SCM_PRINT_H
#include "include/__scm.h"


/* State information passed around during printing.
 */

#define RESET_PRINT_STATE(pstate) \
do { \
  pstate->list_offset = 0; \
  pstate->top = 0; \
} while (0)

#define SCM_WRITINGP(pstate) ((pstate)->writingp)
#define SCM_SET_WRITINGP(pstate, x) { (pstate)->writingp = (x); }

#define SCM_COERCE_OUTPORT(p)  (p) // XXX transitional

// cubic #define SCM_PRINT_STATE_LAYOUT "sruwuwuwuwuwpwuwuwurprpw"
typedef struct scm_print_state {
  int writingp;
  int fancyp;
  unsigned long level;
  unsigned long length;
  unsigned long list_offset;
  unsigned long top;		/* Top of reference stack */
  unsigned long ceiling;	/* Max size of reference stack */
  SCM ref_vect;	 	        /* Stack of references used during
				   circular reference detection */
} scm_print_state;

// SCM scm_print_state_vtable;

SCM scm_make_print_state (void);
void scm_free_print_state (SCM print_state);
SCM scm_i_port_with_print_state (SCM port, SCM print_state);
void scm_intprint (scm_t_intmax n, int radix, SCM port);
void scm_uintprint (scm_t_uintmax n, int radix, SCM port);
void scm_ipruk (char *hdr, SCM ptr, SCM port);
void scm_iprlist (char *hdr, SCM exp, int tlr, SCM port, scm_print_state *pstate);
void scm_print_symbol_name (const char *str, size_t len, SCM port);
void scm_prin1 (SCM exp, SCM port, int writingp);
void scm_iprin1 (SCM exp, SCM port, scm_print_state *pstate);
SCM scm_write (SCM obj, SCM port);
SCM scm_display (SCM obj, SCM port);
SCM scm_simple_format (SCM port, SCM message, SCM args);
SCM scm_newline (SCM port);
SCM scm_write_char (SCM chr, SCM port);
int scm_valid_oport_value_p (SCM val);
void scm_init_print (void);
void x_inclusion_print (void);

#ifdef GUILE_DEBUG
SCM scm_current_pstate (void);
#endif 

#endif  /* SCM_PRINT_H */
