/* classes: h_files */

#ifndef SCM_PROCS_H
#define SCM_PROCS_H
#include "include/__scm.h"




/* Subrs 
 */

typedef struct
{
  SCM handle;			/* link back to procedure object */
  SCM name;
  SCM *generic;			/* 0 if no generic support
				 * *generic == 0 until first method
				 */
  SCM properties;		/* procedure properties */
  int gary;                     /* gsubroutine n-arity */
} scm_t_subr_entry;

#define SCM_SUBRNUM(subr) (SCM_CELL_WORD_0 (subr))
#define SCM_SET_SUBRNUM(subr, num) \
        SCM_SET_CELL_WORD_0 (subr, num)
#define SCM_SUBR_ENTRY(x)       (scm_subr_table[SCM_SUBRNUM (x)])
#define SCM_SNAME(x)            (SCM_SUBR_ENTRY (x).name)
#define SCM_SUBR_PROPS(x)       (SCM_SUBR_ENTRY (x).properties)
#define SCM_SUBR_GENERIC(x)     (SCM_SUBR_ENTRY (x).generic)
#define SCM_GSUBR_NARY(x)       (SCM_SUBR_ENTRY (x).gary)
#define SCM_SET_GSUBR_NARY(x,y) ((&SCM_SUBR_ENTRY (x))->gary = (y))
#define SCM_SUBRF(x)            ((SCM (*)()) SCM_CELL_WORD_1 (x))
#define SCM_SET_SUBRF(x, v)     (SCM_SET_CELL_WORD_1 ((x), (v)))
#define SCM_DSUBRF(x)           ((double (*)()) SCM_CELL_WORD_1 (x))

/* Closures
 */
#define SCM_CLOSUREP(x) (!SCM_IMP(x) && (SCM_TYP3 (x) == scm_tc3_closure))

// SCM  scm_i_get_code          (SCM closure);
SCM  scm_i_get_procprops     (SCM closure);
SCM  scm_i_get_environment   (SCM closure);
void scm_i_set_code_x        (SCM closure, SCM formals, SCM body);
void scm_i_set_procprops_x   (SCM closure, SCM properties);
void scm_i_set_environment_x (SCM closure, SCM environment);

SCM  scm_i_closure_formals   (SCM closure);
SCM  scm_i_closure_body      (SCM closure);

#define TMP_PROCPROPS(x) (SCM_PACK (SCM_CELL_WORD_0 (x)))
#define TMP_SETPROCPROPS(x, e) \
  (SCM_SET_CELL_WORD_0 (x, SCM_UNPACK (e)))

// #define SCM_CODE            scm_i_get_code
#define SCM_CLOSURE_FORMALS scm_i_closure_formals
#define SCM_CLOSURE_BODY    scm_i_closure_body
#define SCM_PROCPROPS       scm_i_get_procprops
#define SCM_SETPROCPROPS    scm_i_set_procprops_x
#define SCM_SETCODE         scm_i_set_code_x
#define SCM_ENV             scm_i_get_environment
#define SCM_SETENV          scm_i_set_environment_x

// #define TMP_CODE(x)         (TMP_CAR (SCM_CLOSCAR (x)))
// #define TMP_PROCPROPS(x)    (TMP_CDR (SCM_CLOSCAR (x)))
#define TMP_CLOSURE_FORMALS(x) (SCM_CELL_OBJECT_1 (x))
#define TMP_CLOSURE_BODY(x)    (SCM_CELL_OBJECT_2 (x))
#define TMP_ENV(x)             (SCM_CELL_OBJECT_3 (x))

#define SCM_TOP_LEVEL(ENV)  (scm_is_null (ENV) || (scm_is_true (scm_procedure_p (SCM_CAR (ENV)))))


scm_t_subr_entry *scm_subr_table;
/* long scm_subr_table_size; */
/* long scm_subr_table_room; */
SCM scm_stand_in_scm_proc (SCM subr);



void scm_mark_subr_table (void);
void scm_free_subr_entry (SCM subr);
SCM scm_c_make_subr (const char *name, long type, SCM (*fcn)());
SCM scm_c_make_subr_with_generic (const char *name, long type,
					  SCM (*fcn)(), SCM *gf);
SCM scm_c_define_subr (const char *name, long type, SCM (*fcn)());
SCM scm_c_define_subr_with_generic (const char *name, long type,
					    SCM (*fcn)(), SCM *gf);
SCM scm_procedure_p (SCM obj);
SCM scm_closure_p (SCM obj);
SCM scm_thunk_p (SCM obj);
int scm_subr_p (SCM obj);
SCM scm_procedure (SCM proc);
void x_inclusion_procs (void);

#endif  /* SCM_PROCS_H */
