/* classes: h_files */

#ifndef SCM_SMOB_H
#define SCM_SMOB_H
#include "include/__scm.h"
#include "include/print.h"


/* This is the internal representation of a smob type */

typedef struct scm_smob_descriptor
{
  char const *name;
  size_t size;
  void (*release) (SCM);
  SCM (*equalp) (SCM, SCM);
  int (*print) (SCM exp, SCM port, scm_print_state *pstate);
  int cell_size;
  long number_of_objects;
} scm_smob_descriptor;



#define SCM_NEWSMOB(z, tc, data) \
do { \
  scm_smobs[SCM_TC2SMOBNUM (tc)].number_of_objects += 1; \
  z = scm_cell ((tc), 0, (scm_t_bits) (data)); \
} while (0)

#define SCM_RETURN_NEWSMOB(tc, data) \
  do { \
   SCM __SCM_smob_answer;                          \
   SCM_NEWSMOB (__SCM_smob_answer, (tc), (data));  \
   return __SCM_smob_answer;                       \
  } while (0)

#define SCM_NEWSMOB2(z, tc, data1, data2) \
do { \
  scm_smobs[SCM_TC2SMOBNUM (tc)].number_of_objects += 1;      \
  z = scm_double_cell ((tc), 0, (scm_t_bits)(data1), (scm_t_bits)(data2), 0); \
} while (0)

#define SCM_RETURN_NEWSMOB2(tc, data1, data2) \
  do { SCM __SCM_smob_answer; \
       SCM_NEWSMOB2 (__SCM_smob_answer, (tc), (data1), (data2)); \
       return __SCM_smob_answer; \
  } while (0)

#define SCM_NEWSMOB3(z, tc, data1, data2, data3) \
do { \
  scm_smobs[SCM_TC2SMOBNUM (tc)].number_of_objects += 1;      \
  z = scm_double_cell ((tc), 0, (scm_t_bits)(data1), \
                       (scm_t_bits)(data2), (scm_t_bits)(data3)); \
} while (0)

#define SCM_RETURN_NEWSMOB3(tc, data1, data2, data3) \
  do { SCM __SCM_smob_answer; \
       SCM_NEWSMOB3 (__SCM_smob_answer, (tc), (data1), (data2), (data3)); \
       return __SCM_smob_answer; \
  } while (0)


// TODO flags w/o shift; w0 is free
#define SCM_SMOB_FLAGS(x)               (SCM_CELL_WORD_0 (x) >> 16)
#define SCM_SMOB_DATA(x)		(SCM_CELL_WORD_1 (x))
#define SCM_SMOB_DATA_2(x)		(SCM_CELL_WORD_2 (x))
#define SCM_SMOB_DATA_3(x)		(SCM_CELL_WORD_3 (x))
#define SCM_SET_SMOB_DATA(x, data)	(SCM_SET_CELL_WORD_1 ((x), (data)))
#define SCM_SET_SMOB_DATA_2(x, data)	(SCM_SET_CELL_WORD_2 ((x), (data)))
#define SCM_SET_SMOB_DATA_3(x, data)	(SCM_SET_CELL_WORD_3 ((x), (data)))
#define SCM_SET_SMOB_FLAGS(x, data)     (scm_i_set_smob_flags((x),(data)<<16))

#define SCM_SMOB_OBJECT(x)		(SCM_CELL_OBJECT_1 (x))
#define SCM_SMOB_OBJECT_2(x)	        (SCM_CELL_OBJECT_2 (x))
#define SCM_SMOB_OBJECT_3(x)	        (SCM_CELL_OBJECT_3 (x))
#define SCM_SET_SMOB_OBJECT(x,obj)	(SCM_SET_CELL_OBJECT_1 ((x), (obj)))
#define SCM_SET_SMOB_OBJECT_2(x,obj)    (SCM_SET_CELL_OBJECT_2 ((x), (obj)))
#define SCM_SET_SMOB_OBJECT_3(x,obj)    (SCM_SET_CELL_OBJECT_3 ((x), (obj)))
#define SCM_SMOB_OBJECT_LOC(x)		(SCM_CELL_OBJECT_LOC ((x), 1))
#define SCM_SMOB_OBJECT_2_LOC(x)	(SCM_CELL_OBJECT_LOC ((x), 2))
#define SCM_SMOB_OBJECT_3_LOC(x)	(SCM_CELL_OBJECT_LOC ((x), 3))

#define SCM_TC2SMOBNUM(x)		(0x0ff & ((x) >> 8))
#define SCM_SMOBNUM(x)			(SCM_TC2SMOBNUM (SCM_CELL_TYPE (x)))
/* SCM_SMOBNAME can be 0 if name is missing */
#define SCM_SMOBNAME(smobnum)		(scm_smobs[smobnum].name)
#define SCM_SMOB_PREDICATE(tag, obj)	SCM_TYP16_PREDICATE (tag, obj)
#define SCM_SMOB_DESCRIPTOR(x)		(scm_smobs[SCM_SMOBNUM (x)])

long scm_numsmob;
extern scm_smob_descriptor scm_smobs[];

void scm_i_set_smob_flags (SCM x, scm_t_bits data);


int generic_smob_release (SCM smob);
int scm_smob_print (SCM exp, SCM port, scm_print_state *pstate);

/* The following set of functions is the standard way to create new
 * SMOB types.
 *
 * Create a type tag using `scm_make_smob_type', accept default values
 * for mark, free, print and/or equalp functions, or set your own
 * values using `scm_set_smob_...'.
 */

scm_t_tag scm_make_smob_type (char const *name, size_t size);

void scm_set_smob_release (scm_t_tag tc, int cell_size, void (*release) (SCM));
void scm_set_smob_print (scm_t_tag tc,
				 int (*print) (SCM, SCM, scm_print_state*));
void scm_set_smob_equalp (scm_t_tag tc, SCM (*equalp) (SCM, SCM));

void scm_assert_smob_type (scm_t_tag tag, SCM val);

/* Function for creating smobs */

SCM scm_make_smob (scm_t_tag tc);

void scm_smob_init (void);

#endif  /* SCM_SMOB_H */
