/* classes: h_files */

#ifndef SCM_GC_H
#define SCM_GC_H
#include "include/__scm.h"


/*
 * TRIM
 */

void *global_stack_base;

// the 8th bit protection stubs (XXX see core-struct.c)
#define SCM_GC_MARK(x) 0
#define SCM_GC_MARK_P(x) 0
#define SCM_CLEAR_GC_MARK(x)
#define SCM_SET_GC_MARK(x)

/* End of the transitional setup */

#define FREE_CELL_P(x) (SCM_TYP16_PREDICATE(scm_tc_free_cell, (x)))

// #define SCM2PTR(x) ((scm_t_cell *) (SCM_UNPACK (x)))
// #define SCM2PTR(x) ((void*)x)
// #define PTR2SCM(x) ((SCM) (x)))

typedef struct scm_t_cell
{
  scm_t_refcount word_m1a;
  scm_t_tag      word_m1b;
  SCM  word_0;
  SCM  word_1;
} scm_t_cell;
// See also scm_t_tag

void *global_memory_backup_segment;
void scm_i_memory_backup_ignition ();

/* Low level cell data accessing macros.  These macros should only be used
 * from within code which is considered part of memory management, since
 * they will never check the cells they are applied to - not even if VSI-C
 * is compiled in debug mode [??? does this still work].  In particular
 * these macros will even work for free cells, which should never be
 * encountered by user code.  */

/* TRIM low level debugging */
void trim_i_debug_purple (SCM object, int ticks);

SCM global_trim_abort_on_access;
void debug_trim_check_tree (SCM data_graph);
void debug_trim_assert_list (SCM list);
void debug_trim_print_list (SCM list);
void debug_trim_print_rc (SCM object);

SCM trim_debug_cell_abort (SCM object, int n);
void trim_debug_set_cell_abort (SCM object, int n, long value);
void trim_reset_allocation ();
void trim_reset_double_allocation ();

// cell_object accessors cast to scm
// cell_word accessors cast to scm_t_bits
#define SCM_TRIM_CELL_OBJECT(x, n) (((SCM *)SCM2PTR (x))[n])
#define SCM_TRIM_CELL_WORD(x, n) \
  (SCM_UNPACK (SCM_TRIM_CELL_OBJECT ((x), (n))))
// #define TRIM_CELL_WORD_M1a(x) (((scm_t*)(x))->reference_count)
// #define TRIM_CELL_WORD_M1b(x) (((scm_t*)(x))->tag)
#define TRIM_CELL_WORD_M1a(x) SCM_TRIM_CELL_WORD ((x), 0)
#define TRIM_CELL_WORD_M1b(x) SCM_TRIM_CELL_WORD ((x), 1)
#define TRIM_CELL_WORD_0(x) SCM_TRIM_CELL_WORD ((x), 2)
#define TRIM_CELL_WORD_1(x) SCM_TRIM_CELL_WORD ((x), 3)
#define TRIM_CELL_WORD_2(x) SCM_TRIM_CELL_WORD ((x), 4)
#define TRIM_CELL_WORD_3(x) SCM_TRIM_CELL_WORD ((x), 5)

#define SCM_TRIM_SET_CELL_OBJECT(x, n, v) ((((SCM *)SCM2PTR (x))[n]) = (v))
#define SCM_TRIM_SET_CELL_WORD(x, n, v)  \
  (SCM_TRIM_SET_CELL_OBJECT ((x), (n), SCM_PACK (v)))
// #define TRIM_SET_CELL_WORD_M1(x, v) (trim_debug_set_cell_abort ((x), 0, (v)))
// #define TRIM_SET_CELL_WORD_M1a(x, v) \
//   ((((scm_t *)SCM2PTR (x))->reference_count)=(v))
// #define TRIM_SET_CELL_WORD_M1b(x, v) \
//  ((((scm_t *)SCM2PTR (x))->tag)=(v))
#define TRIM_SET_CELL_WORD_M1a(x, v) SCM_TRIM_SET_CELL_WORD ((x), 0, (v))
#define TRIM_SET_CELL_WORD_M1b(x, v) SCM_TRIM_SET_CELL_WORD ((x), 1, (v))
#define TRIM_SET_CELL_WORD_0(x, v) SCM_TRIM_SET_CELL_WORD ((x), 2, (v))
#define TRIM_SET_CELL_WORD_1(x, v) SCM_TRIM_SET_CELL_WORD ((x), 3, (v))
#define TRIM_SET_CELL_WORD_2(x, v) SCM_TRIM_SET_CELL_WORD ((x), 4, (v))
#define TRIM_SET_CELL_WORD_3(x, v) SCM_TRIM_SET_CELL_WORD ((x), 5, (v))

#define SCM_TRIM_CELL_TYPE(x) (TRIM_CELL_WORD_M1a (x))

#define SCM_FREE_CELL_CDR(x) (SCM_TRIM_CELL_OBJECT ((x), 3))
#define SCM_SET_FREE_CELL_CDR(x, v) (SCM_TRIM_SET_CELL_OBJECT ((x), 3, (v)))

/*
 * Compatibility Accessors for the old GC code
 */
#define SCM_GC_CELL_OBJECT(x, n) (SCM_TRIM_CELL_OBJECT ((x), ((n)+2)))
#define SCM_GC_CELL_WORD(x, n) \
  (SCM_UNPACK (SCM_TRIM_CELL_OBJECT ((x), ((n)+2))))

#define SCM_GC_SET_CELL_OBJECT(x, n, v) \
  (SCM_TRIM_SET_CELL_OBJECT ((x), (n+2), SCM_PACK (v)))
#define SCM_GC_SET_CELL_WORD(x, n, v)  \
  (SCM_TRIM_SET_CELL_OBJECT ((x), (n+2), SCM_PACK (v)))

#define SCM_GC_CELL_TYPE(x) SCM_CELL_TYPE(x)


/* Except for the garbage collector, no part of VSI-C should ever run over a
 * free cell.  Thus, if VSI-C is compiled in debug mode the SCM_CELL_* and
 * SCM_SET_CELL_* macros below report an error if they are applied to a free
 * cell.  Some other plausibility checks are also performed.  However, if
 * VSI-C is not compiled in debug mode, there won't be any time penalty at all
 * when using these macros.  */

#if (SCM_DEBUG_CELL_ACCESSES == 1)
#  define SCM_VALIDATE_CELL(cell, expr) (scm_assert_cell_valid (cell), (expr))
#else
#  define SCM_VALIDATE_CELL(cell, expr) (expr)
#endif

#define SCM_CELL_WORD(x, n) \
  SCM_VALIDATE_CELL ((x), SCM_TRIM_CELL_WORD ((x), ((n)+2)))
#define SCM_CELL_WORD_0(x) SCM_CELL_WORD ((x), 0)
#define SCM_CELL_WORD_1(x) SCM_CELL_WORD ((x), 1)
#define SCM_CELL_WORD_2(x) SCM_CELL_WORD ((x), 2)
#define SCM_CELL_WORD_3(x) SCM_CELL_WORD ((x), 3)

#define SCM_CELL_OBJECT(x, n) \
  SCM_VALIDATE_CELL ((x), SCM_TRIM_CELL_OBJECT ((x), ((n)+2)))
#define SCM_CELL_OBJECT_0(x) SCM_CELL_OBJECT ((x), 0)
#define SCM_CELL_OBJECT_1(x) SCM_CELL_OBJECT ((x), 1)
#define SCM_CELL_OBJECT_2(x) SCM_CELL_OBJECT ((x), 2)
#define SCM_CELL_OBJECT_3(x) SCM_CELL_OBJECT ((x), 3)

#define SCM_SET_CELL_WORD(x, n, v) \
  SCM_VALIDATE_CELL ((x), SCM_TRIM_SET_CELL_WORD ((x), ((n)+2), (v)))
#define SCM_SET_CELL_WORD_0(x, v) SCM_SET_CELL_WORD ((x), 0, (v))
#define SCM_SET_CELL_WORD_1(x, v) SCM_SET_CELL_WORD ((x), 1, (v))
#define SCM_SET_CELL_WORD_2(x, v) SCM_SET_CELL_WORD ((x), 2, (v))
#define SCM_SET_CELL_WORD_3(x, v) SCM_SET_CELL_WORD ((x), 3, (v))

#define SCM_SET_CELL_OBJECT(x, n, v) \
  SCM_VALIDATE_CELL ((x), SCM_TRIM_SET_CELL_OBJECT ((x), ((n)+2), (v)))
#define SCM_SET_CELL_OBJECT_0(x, v) SCM_SET_CELL_OBJECT ((x), 0, (v))
#define SCM_SET_CELL_OBJECT_1(x, v) SCM_SET_CELL_OBJECT ((x), 1, (v))
#define SCM_SET_CELL_OBJECT_2(x, v) SCM_SET_CELL_OBJECT ((x), 2, (v))
#define SCM_SET_CELL_OBJECT_3(x, v) SCM_SET_CELL_OBJECT ((x), 3, (v))

/* The next one starts counting from cell word zero onwards leaving
 * the reference count at offset minus one. [GC->TRIM Transitional Setup] */
#define SCM_CELL_OBJECT_LOC(x, n) \
  (SCM_VALIDATE_CELL((x), &SCM_TRIM_CELL_OBJECT ((x), (n+2))))
#define SCM_CARLOC(x)             (SCM_CELL_OBJECT_LOC ((x), 0))
#define SCM_CDRLOC(x)             (SCM_CELL_OBJECT_LOC ((x), 1))

#define SCM_CELL_TYPE(x)        ((const scm_t_tag) TRIM_CELL_WORD_M1a (x))
#define SCM_SET_CELL_TYPE(x, t) TRIM_SET_CELL_WORD_M1a ((x), (t))

#define TRIM_REFCOUNT(x)        ((scm_t_refcount) TRIM_CELL_WORD_M1b (x))
#define TRIM_SET_REFCOUNT(x, n) (TRIM_SET_CELL_WORD_M1b (x, n))

/* End of Direct Accessors */


/*
 * VSI/TRIM Private
 */
void trim_trim_allocation ();
void trim_trim_double_allocation ();
void trim_trim_phalanx ();
SCM phalanx_saturation_maps ();
SCM trim_saturation_map ();
SCM trim_double_saturation_map ();

// Release by Data Type
void rc_release (SCM object);
void rc_release_reinsert (SCM cell);
void rc_release_reinsert_double (SCM double_cell);
void trim_release_pair    (SCM pair);
void trim_release_closure (SCM closure);
void trim_release_vector  (SCM vector);
void trim_release_brass    (SCM brass);
void trim_release_string    (SCM string);
void trim_release_stringbuf (SCM stringbuf);
void trim_release_byte_string (SCM byte_string);
void trim_release_num_big      (SCM num_big);
void trim_release_num_fraction (SCM num_fraction);
void trim_release_num_complex  (SCM num_complex);

void trim_release_variable  (SCM variable);
void trim_release_winder    (SCM winder);
// void trim_release_dictionary (SCM dictionary);
// void trim_release_port   (SCM port);


/* VSI/TRIM Reference Counting API */
SCM one_cell_please ();

void reference_count_plus1  (SCM object);
void reference_count_minus1 (SCM object);

void rc_declare_data_graph_root (SCM object);

void reference   (SCM referrent, SCM referred);
void dereference (SCM referrent, SCM referred);
void multiple_reference (SCM referrent, long n, SCM referred);

void reference_count_parameter_up (SCM object);
SCM  reference_count_return       (SCM object);

SCM  rc_return_stack_reference (SCM object); // XXX that name is misleading

void dereference_return_value      (SCM object);
SCM rc_let_parameter_fly           (SCM object);
SCM rc_count_parameter_up_and_pass (SCM object);

// Don't use RC_RETURN
#define RC_RETURN(x)             return (reference_count_return ((x)))
#define reference_parameter(x)   reference_count_parameter_up ((x))
#define dereference_parameter(x) reference_count_minus1((x))

#define rc_consumed(referrent, referred) \
  { reference (referrent, referred);     \
    dereference_return_value (referred); \
  }
#define rc_Fly  rc_let_parameter_fly
#define rc_X    rc_count_parameter_up_and_pass
#define rc_Void dereference_return_value

#define rc_consume_here    rc_X
#define rc_CH   rc_consume_here
#define rc_parameter_consumption(x) ((void)(x))   // empty, declaritive

#define heap_reference(r,x)   (reference_parameter (x))
#define heap_dereference(r,x) (dereference_parameter (x))

SCM scm_i_car (SCM pair);
SCM scm_i_cdr (SCM pair);

/* #define FLY_CDR(x) (rc_X (SCM_CDR (rc_Fly ((x))))) */
/* #define FLY_CAR(x) (rc_X (SCM_CAR (rc_Fly ((x))))) */

/* End of TRIM Private */

/* XXX This is like gambling: does each thread need its own freelist?
 *     I'm guessing no right now */
SCM *global_cell_freelist;
SCM *global_double_cell_freelist;

#define TRIM_EOL      (scm_t_bits) NULL
#define TRIM_EOL_P(x) ((SCM2PTR(x)) == NULL)


#define scm_free(mem) (free(mem))
void *scm_malloc (size_t size);
void *scm_calloc (size_t size);
void *scm_realloc (void *mem, size_t size);
char *scm_strdup (const char *str);
char *scm_strndup (const char *str, size_t n);


char *scm_gc_strdup (const char *str, const char *what);
char *scm_gc_strndup (const char *str, size_t n, const char *what);


/* Compatibility Stubs - Purge, please */
#define scm_gc_malloc(size, id_string) (scm_malloc(size))
#define scm_gc_calloc(size, id_string) (scm_calloc(size))
#define scm_gc_realloc(mem, old_size, new_size, id_string) \
  (scm_realloc(mem, new_size))
#define scm_gc_free(mem, size, id_string) (free(mem))

#define scm_gc_register_collectable_memory(mem, size, id_string)
#define scm_gc_unregister_collectable_memory(mem, size, id_string)

SCM scm_permanent_object (SCM obj);
void scm_storage_prehistory (void);
void x_inclusion_storage ();
void scm_init_storage (void);
void init_phalanx (void);

#endif  /* SCM_GC_H */
