#include "include/config.h"

#include "include/__scm.h"
#include "include/ihub.h"
#include "include/byte-strings.h"


/*
 * Ceval Statistics
 */
#include "include/alist.h"
unsigned long global_counter_ceval_fuel = 0;
inline SCM ceval_fuel (SCM a, SCM b, SCM c) {
  global_counter_ceval_fuel++;
  return scm_i_cons (scm_i_cons (a, b), c);
}

long global_number_of_closures = 0;
long global_last_number_of_closures    = 0;

SCM_DEFINE (scm_global_closure_count, "global-closure-count", 0, 0, 0,
            (),
            "Returns the total number of closures.")
{
  return scm_from_long (global_number_of_closures);
}

SCM_DEFINE (scm_global_closure_count_delta,
            "global-closure-count-delta", 0, 0, 0, (), "")
{
  long delta;

  delta = global_number_of_closures - global_last_number_of_closures;
  global_last_number_of_closures = global_number_of_closures;

  return scm_from_long (delta);
}


#include <assert.h>
#include "include/_scm.h"
#include "include/alist.h"
#include "include/debug.h"
#include "include/eq.h"
#include "include/feature.h"
#include "include/list.h"
#include "include/macros.h"
#include "include/nature.h"
#include "include/ports.h"
#include "include/print.h"
#include "include/procprop.h"
#include "include/root.h"
#include "include/brass.h"
#include "include/throw.h"
#include "include/validate.h"
#include "include/vectors.h"

#include "include/eval.h"

static SCM scm_listofnull = 0;


static SCM unmemoize_exprs (SCM expr, SCM env);
static SCM canonicalize_define (SCM expr);
static SCM *scm_lookupcar1 (SCM vloc, SCM genv, int check);
static SCM unmemoize_builtin_macro (SCM expr, SCM env);
static SCM eval_letrec_inits (SCM env, SCM init_forms, SCM last);



/* {Syntax Errors}
 *
 * This section defines the message strings for the syntax errors that can be
 * detected during memoization and the functions and macros that shall be
 * called by the memoizer code to signal syntax errors.  */


/* Syntax errors that can be detected during memoization: */

/* Circular or improper lists do not form valid scheme expressions.  If a
 * circular list or an improper list is detected in a place where a scheme
 * expression is expected, a 'Bad expression' error is signalled.  */
static const char s_bad_expression[] = "Bad expression";

/* If a form is detected that holds a different number of expressions than are
 * required in that context, a 'Missing or extra expression' error is
 * signalled.  */
static const char s_expression[] = "Missing or extra expression in";

/* If a form is detected that holds less expressions than are required in that
 * context, a 'Missing expression' error is signalled.  */
static const char s_missing_expression[] = "Missing expression in";

/* If a form is detected that holds more expressions than are allowed in that
 * context, an 'Extra expression' error is signalled.  */
static const char s_extra_expression[] = "Extra expression in";

/* The empty combination '()' is not allowed as an expression in scheme.  If
 * it is detected in a place where an expression is expected, an 'Illegal
 * empty combination' error is signalled.  Note: If you encounter this error
 * message, it is very likely that you intended to denote the empty list.  To
 * do so, you need to quote the empty list like (quote ()) or '().  */
static const char s_empty_combination[] = "Illegal empty combination";

/* A body may hold an arbitrary number of internal defines, followed by a
 * non-empty sequence of expressions.  If a body with an empty sequence of
 * expressions is detected, a 'Missing body expression' error is signalled.
 */
static const char s_missing_body_expression[] = "Missing body expression in";

/* A body may hold an arbitrary number of internal defines, followed by a
 * non-empty sequence of expressions.  Each the definitions and the
 * expressions may be grouped arbitraryly with begin, but it is not allowed to
 * mix definitions and expressions.  If a define form in a body mixes
 * definitions and expressions, a 'Mixed definitions and expressions' error is
 * signalled.  */
static const char s_mixed_body_forms[] = "Mixed definitions and expressions in";
/* Definitions are only allowed on the top level and at the start of a body.
 * If a definition is detected anywhere else, a 'Bad define placement' error
 * is signalled.  */
static const char s_bad_define[] = " the old error message lied to you ";

/* Case or cond expressions must have at least one clause.  If a case or cond
 * expression without any clauses is detected, a 'Missing clauses' error is
 * signalled.  */
static const char s_missing_clauses[] = "Missing clauses";

/* If there is an 'else' clause in a case or a cond statement, it must be the
 * last clause.  If after the 'else' case clause further clauses are detected,
 * a 'Misplaced else clause' error is signalled.  */
static const char s_misplaced_else_clause[] = "Misplaced else clause";

/* If a case clause is detected that is not in the format
 *   (<label(s)> <expression1> <expression2> ...)
 * a 'Bad case clause' error is signalled.  */
static const char s_bad_case_clause[] = "Bad case clause";

/* If a case clause is detected where the <label(s)> element is neither a
 * proper list nor (in case of the last clause) the syntactic keyword 'else',
 * a 'Bad case labels' error is signalled.  Note: If you encounter this error
 * for an else-clause which seems to be syntactically correct, check if 'else'
 * is really a syntactic keyword in that context.  If 'else' is bound in the
 * local or global environment, it is not considered a syntactic keyword, but
 * will be treated as any other variable.  */
static const char s_bad_case_labels[] = "Bad case labels";

/* In a case statement all labels have to be distinct.  If in a case statement
 * a label occurs more than once, a 'Duplicate case label' error is
 * signalled.  */
static const char s_duplicate_case_label[] = "Duplicate case label";

/* If a cond clause is detected that is not in one of the formats
 *   (<test> <expression1> ...) or (else <expression1> <expression2> ...)
 * a 'Bad cond clause' error is signalled.  */
static const char s_bad_cond_clause[] = "Bad cond clause";

/* If a cond clause is detected that uses the alternate '=>' form, but does
 * not hold a recipient element for the test result, a 'Missing recipient'
 * error is signalled.  */
static const char s_missing_recipient[] = "Missing recipient in";

/* If in a position where a variable name is required some other object is
 * detected, a 'Bad variable' error is signalled.  */
static const char s_bad_variable[] = "Bad variable";

/* Bindings for forms like 'let' and 'do' have to be given in a proper,
 * possibly empty list.  If any other object is detected in a place where a
 * list of bindings was required, a 'Bad bindings' error is signalled.  */
static const char s_bad_bindings[] = "Bad bindings";

/* Depending on the syntactic context, a binding has to be in the format
 * (<variable> <expression>) or (<variable> <expression1> <expression2>).
 * If anything else is detected in a place where a binding was expected, a
 * 'Bad binding' error is signalled.  */
static const char s_bad_binding[] = "Bad binding";

/* Some syntactic forms don't allow variable names to appear more than once in
 * a list of bindings.  If such a situation is nevertheless detected, a
 * 'Duplicate binding' error is signalled.  */
static const char s_duplicate_binding[] = "Duplicate binding";

/* If the exit form of a 'do' expression is not in the format
 *   (<test> <expression> ...)
 * a 'Bad exit clause' error is signalled.  */
static const char s_bad_exit_clause[] = "Bad exit clause";

/* The formal function arguments of a lambda expression have to be either a
 * single symbol or a non-cyclic list.  For anything else a 'Bad formals'
 * error is signalled.  */
static const char s_bad_formals[] = "Bad formals";

/* If in a lambda expression something else than a symbol is detected at a
 * place where a formal function argument is required, a 'Bad formal' error is
 * signalled.  */
static const char s_bad_formal[] = "Bad formal";

/* If in the arguments list of a lambda expression an argument name occurs
 * more than once, a 'Duplicate formal' error is signalled.  */
static const char s_duplicate_formal[] = "Duplicate formal";

/* If the evaluation of an unquote-splicing expression gives something else
 * than a proper list, a 'Non-list result for unquote-splicing' error is
 * signalled.  */
static const char s_splicing[] = "Non-list result for unquote-splicing";

/* If something else than an exact integer is detected as the argument for
 * @slot-ref and @slot-set!, a 'Bad slot number' error is signalled.  */
static const char s_bad_slot_number[] = "Bad slot number";


/* Signal a syntax error.  We distinguish between the form that caused the
 * error and the enclosing expression.  The error message will print out as
 * shown in the following pattern.  The file name and line number are only
 * given when they can be determined from the erroneous form or from the
 * enclosing expression.
 *
 * <filename>: In procedure memoization:
 * <filename>: In file <name>, line <nr>: <error-message> in <expression>.  */

SCM_SYMBOL (syntax_error_key, "syntax-error");

/* The prototype is needed to indicate that the function does not return.  */
static void
syntax_error (const char* const, const SCM, const SCM) SCM_NORETURN;

static void 
syntax_error (const char* const msg, const SCM form, const SCM expr)
{
  reference_parameter (form);
  reference_parameter (expr);

  SCM msg_string = scm_from_cen (msg);
  SCM filename = SCM_BOOL_F;
  SCM linenr = SCM_BOOL_F;
  const char *format;
  SCM args;

  reference_parameter (msg_string);

  if (!SCM_UNBNDP (expr))
    {
      if (scm_is_true (filename))
	{
	  format = "In file ~S, line ~S: ~A ~S in expression ~S.";
	  args = scm_list_5 (filename, linenr, msg_string, form, expr);
	}
      else if (scm_is_true (linenr))
	{
	  format = "In line ~S: ~A ~S in expression ~S.";
	  args = scm_list_4 (linenr, msg_string, form, expr);
	}
      else
	{
	  format = "~A ~S in expression ~S.";
	  args = scm_list_3 (msg_string, form, expr);
	}
    }
  else
    {
      if (scm_is_true (filename))
	{
	  format = "In file ~S, line ~S: ~A ~S.";
	  args = scm_list_4 (filename, linenr, msg_string, form);
	}
      else if (scm_is_true (linenr))
	{
	  format = "In line ~S: ~A ~S.";
	  args = scm_list_3 (linenr, msg_string, form);
	}
      else
	{
	  format = "~A ~S.";
	  args = scm_list_2 (msg_string, form);
	}
    }

  reference_parameter (args);
  scm_error (syntax_error_key, "memoization", format, args, SCM_BOOL_F);
}


/* Shortcut macros to simplify syntax error handling. */
#define ASSERT_SYNTAX(cond, message, form)		\
  { if (SCM_UNLIKELY (!(cond)))			\
      syntax_error (message, form, SCM_UNDEFINED); }
#define ASSERT_SYNTAX_2(cond, message, form, expr)	\
  { if (SCM_UNLIKELY (!(cond)))			\
      syntax_error (message, form, expr); }



/* {Ilocs}
 *
 * Ilocs are memoized references to variables in local environment frames.
 * They are represented as three values:  The relative offset of the
 * environment frame, the number of the binding within that frame, and a
 * boolean value indicating whether the binding is the last binding in the
 * frame.
 *
 * Frame numbers have 11 bits, relative offsets have 12 bits.
 */

#define SCM_ILOC00		SCM_MAKE_ITAG8(0L, scm_tc8_iloc)
#define SCM_IFRINC		(0x00000100L)
#define SCM_ICDR		(0x00080000L)
#define SCM_IDINC		(0x00100000L)
#define SCM_IFRAME(n) 		((long)((SCM_ICDR-SCM_IFRINC)>>8) \
				 & (SCM_UNPACK (n) >> 8))
#define SCM_IDIST(n) 		(SCM_UNPACK (n) >> 20)
#define SCM_ICDRP(n) 		(SCM_ICDR & SCM_UNPACK (n))
#define SCM_IDSTMSK		(-SCM_IDINC)
#define SCM_IFRAMEMAX           ((1<<11)-1)
#define SCM_IDISTMAX            ((1<<12)-1)
/* #define SCM_MAKE_ILOC(frame_nr, binding_nr, last_p) \ */
/*   SCM_PACK ( \ */
/*    ((frame_nr) << 8) \ */
/*    + ((binding_nr) << 20) \ */
/*    + ((last_p) ? SCM_ICDR : 0) \ */
/*    + scm_tc8_iloc ) */

SCM SCM_MAKE_ILOC(int frame_nr, int binding_nr, int last_p);
SCM inline
SCM_MAKE_ILOC(frame_nr, binding_nr, last_p)
{
  if ((frame_nr > SCM_IFRAMEMAX) || (binding_nr > SCM_IDISTMAX)) {
    fprintf (stderr, " * VSI-C: iloc maximum exceeded * ");
    abort ();
  }
  return SCM_PACK ( \
     ((frame_nr) << 8) \
     + ((binding_nr) << 20) \
     + ((last_p) ? SCM_ICDR : 0) \
     + scm_tc8_iloc );
}

void
scm_i_print_iloc (SCM iloc, SCM port)
{
  scm_puts ("#@", port);
  scm_intprint ((long) SCM_IFRAME (iloc), 10, port);
  scm_putc (SCM_ICDRP (iloc) ? '-' : '+', port);
  scm_intprint ((long) SCM_IDIST (iloc), 10, port);
}

#if (SCM_DEBUG_DEBUGGING_SUPPORT == 1)

SCM scm_dbg_make_iloc (SCM frame, SCM binding, SCM cdrp);

SCM_DEFINE (scm_dbg_make_iloc, "dbg-make-iloc", 3, 0, 0,
            (SCM frame, SCM binding, SCM cdrp),
	    "Return a new iloc with frame offset @var{frame}, binding\n"
	    "offset @var{binding} and the cdr flag @var{cdrp}.")
#define FUNC_NAME s_scm_dbg_make_iloc
{
  return SCM_MAKE_ILOC ((scm_t_bits) scm_to_unsigned_integer (frame, 0, SCM_IFRAMEMAX),
			(scm_t_bits) scm_to_unsigned_integer (binding, 0, SCM_IDISTMAX),
			scm_is_true (cdrp));
}
#undef FUNC_NAME

SCM scm_dbg_iloc_p (SCM obj);

SCM_DEFINE (scm_dbg_iloc_p, "dbg-iloc?", 1, 0, 0, 
          (SCM obj),
	    "Return @code{#t} if @var{obj} is an iloc.")
#define FUNC_NAME s_scm_dbg_iloc_p
{
  return scm_from_bool (SCM_ILOCP (obj));
}
#undef FUNC_NAME

#endif



/* {Evaluator byte codes (isyms)}
 */

#define ISYMNUM(n) 		(SCM_ITAG8_DATA (n))

/* This table must agree with the list of SCM_IM_ constants in tags.h */
static const char *const isymnames[] =
{
  "#@and",
  "#@begin",
  "#@case",
  "#@cond",
  "#@do",
  "#@if",
  "#@lambda",
  "#@let",
  "#@let*",
  "#@letrec",
  "#@or",
  "#@quote",
  "#@set!",
  "#@define",
  "#@apply",
  "#@call-with-current-continuation",  // unused
  "#@dispatch",
  "#@slot-ref",  // unused
  "#@slot-set!", // unused
  "#@delay",  // unused
  "#@future", // unused
  "#@call-with-values", // unused
  "#@else",
  "#@arrow",
  "#@nil-cond" //, unused
  // "#@bind"  // unused
};

void
scm_i_print_isym (SCM isym, SCM port)
{
  const size_t isymnum = ISYMNUM (isym);
  if (isymnum < (sizeof isymnames / sizeof (char *)))
    scm_puts (isymnames[isymnum], port);
  else
    scm_ipruk ("isym", isym, port);
}



/* The function lookup_symbol is used during memoization: Lookup the symbol in
 * the environment.  If there is no binding for the symbol, SCM_UNDEFINED is
 * returned.  If the symbol is a global variable, the variable object to which
 * the symbol is bound is returned.  Finally, if the symbol is a local
 * variable the corresponding iloc object is returned.  */

/* A helper function for lookup_symbol: Try to find the symbol in the top
 * level environment frame.  The function returns SCM_UNDEFINED if the symbol
 * is unbound and it returns a variable object if the symbol is a global
 * variable.  */
static inline SCM
lookup_global_symbol (const SCM symbol, const SCM top_level)
{
  return scm_sym2var (symbol, top_level, SCM_BOOL_F);
}

static SCM
lookup_symbol (const SCM symbol, const SCM env)
{
  SCM frame_idx;
  unsigned int frame_nr;

  for (frame_idx = env, frame_nr = 0;
       !scm_is_null (frame_idx);
       frame_idx = TMP_CDR (frame_idx), ++frame_nr)
    {
      const SCM frame = TMP_CAR (frame_idx);
      if (scm_is_pair (frame))
	{
	  /* frame holds a local environment frame */
	  SCM symbol_idx;
	  unsigned int symbol_nr;

	  for (symbol_idx = TMP_CAR (frame), symbol_nr = 0;
	       scm_is_pair (symbol_idx);
	       symbol_idx = TMP_CDR (symbol_idx), ++symbol_nr)
	    {
	      if (scm_is_eq (TMP_CAR (symbol_idx), symbol))
		/* found the symbol, therefore return the iloc */
		return SCM_MAKE_ILOC (frame_nr, symbol_nr, 0);
	    }
	  if (scm_is_eq (symbol_idx, symbol))
	    /* found the symbol as the last element of the current frame */
	    return SCM_MAKE_ILOC (frame_nr, symbol_nr, 1);
	}
      else
	{
	  /* no more local environment frames */
	  return lookup_global_symbol (symbol, frame);
	}
    }

  return lookup_global_symbol (symbol, SCM_BOOL_F);
}

/* Lookup a given local variable in an environment.  The local variable is
 * given as an iloc, that is a triple <frame, binding, last?>, where frame
 * indicates the relative number of the environment frame (counting upwards
 * from the innermost environment frame), binding indicates the number of the
 * binding within the frame, and last? (which is extracted from the iloc using
 * the macro SCM_ICDRP) indicates whether the binding forms the binding at the
 * very end of the improper list of bindings.  */
inline SCM *
scm_ilookup (SCM iloc, SCM env)
{
  if (0) {
    scm_display (SCM_BOOL_T, scm_current_error_port ());
    scm_display (env, scm_current_error_port ());
    scm_display (SCM_BOOL_T, scm_current_error_port ());
    scm_newline (scm_current_error_port ());
  }


  unsigned int frame_nr = SCM_IFRAME (iloc);
  unsigned int binding_nr = SCM_IDIST (iloc);
  SCM frames = env;
  SCM bindings;
 
  for (; 0 != frame_nr; --frame_nr)
    frames = TMP_CDR (frames);

  bindings = TMP_CAR (frames);
  for (; 0 != binding_nr; --binding_nr)
    bindings = TMP_CDR (bindings);

  if (SCM_ICDRP (iloc))
    return SCM_CDRLOC (bindings);
  return SCM_CARLOC (TMP_CDR (bindings));
}


SCM_SYMBOL (scm_unbound_variable_key, "undefined"); // see also eval.x

static void error_unbound_variable (SCM symbol) SCM_NORETURN;
static void error_defined_variable (SCM symbol) SCM_NORETURN;

/* Call this for variables that are unfound.
 */
static void
error_unbound_variable (SCM symbol)
{
  scm_error (scm_unbound_variable_key, NULL,
	     "Unbound variable: ~S",
	     scm_list_1 (symbol), SCM_BOOL_F);
}

/* Call this for variables that are found but contain SCM_UNDEFINED.
 */
static void
error_defined_variable (SCM symbol)
{
  /* We use the 'unbound-variable' key here as well, since it
     basically is the same kind of error, with a slight variation in
     the displayed message.
  */
  scm_error (scm_unbound_variable_key, NULL,
	     "Variable used before given a value: ~S",
	     scm_list_1 (symbol), SCM_BOOL_F);
}


/* SCM_LOOKUPCAR1 is what SCM_LOOKUPCAR used to be but is allowed to
   return NULL to indicate a failed lookup due to some race conditions
   between threads.  This only happens when VLOC is the first cell of
   a special form that will eventually be memoized (like `let', etc.)
   In that case the whole lookup is bogus and the caller has to
   reconsider the complete special form.

   SCM_LOOKUPCAR is still there, of course.  It just calls
   SCM_LOOKUPCAR1 and aborts on receiving NULL.  So SCM_LOOKUPCAR
   should only be called when it is known that VLOC is not the first
   pair of a special form.  Otherwise, use SCM_LOOKUPCAR1 and check
   for NULL.  I think I've found the only places where this
   applies. */

static SCM undef_object = SCM_UNDEFINED;

// returns NULL or the location of a value
static SCM *
scm_lookupcar1 (SCM vloc, SCM env, int check)
{
  reference_parameter (vloc);
  reference_parameter (env);

  register SCM *al, fl, var = TMP_CAR (vloc);
  register SCM iloc = SCM_ILOC00;
  for (; SCM_NIMP (env); env = FLY_CDR (env))
    {
      if (!scm_is_pair (TMP_CAR (env)))
	break;
      al = SCM_CARLOC (env);
      for (fl = TMP_CAR (*al); SCM_NIMP (fl); fl = TMP_CDR (fl))
	{
	  if (!scm_is_pair (fl))
	    {
	      if (scm_is_eq (fl, var))
	      {
		if (!scm_is_eq (TMP_CAR (vloc), var))
		  goto race;

                /* Memoizing into the expression form. */
		SCM_SETCAR (vloc, SCM_PACK (SCM_UNPACK (iloc) + SCM_ICDR));
                reference_count_return (TMP_CDR (*al));
                dereference_parameter (vloc);
                dereference_parameter (env);
		return SCM_CDRLOC (*al);
	      }
	      else
		break;
	    }
	  al = SCM_CDRLOC (*al);
	  if (scm_is_eq (TMP_CAR (fl), var))
	    {
	      if (SCM_UNBNDP (TMP_CAR (*al)))
		error_defined_variable (var);
	      if (!scm_is_eq (TMP_CAR (vloc), var))
		goto race;
	      SCM_SETCAR (vloc, iloc); /* Memoizing into the expression form. */
              reference_count_return (TMP_CAR (*al));
              dereference_parameter (vloc);
              dereference_parameter (env);
	      return SCM_CARLOC (*al);
	    }
	  iloc = SCM_PACK (SCM_UNPACK (iloc) + SCM_IDINC);
	}
      iloc = SCM_PACK ((~SCM_IDSTMSK) & (SCM_UNPACK(iloc) + SCM_IFRINC));
    }
  {
    SCM top_thunk, real_var;
    if (SCM_NIMP (env))
      {
	top_thunk = TMP_CAR (env); /* env now refers to a XXX
				      top level env thunk */
	env = FLY_CDR (env);
      }
    else
      top_thunk = SCM_BOOL_F;
    /* XXX we want this top-level lookup only to happen for chained
       environments */
    reference_parameter (var);
    real_var = scm_sym2var (rc_Fly (var), top_thunk, SCM_BOOL_F);
    reference_parameter (real_var);

    if (SCM_UNDEFINED_P (real_var))
      goto errout;

    if (!scm_is_null (env) || SCM_UNBNDP (TMP_VARIABLE_REF (real_var)))
      {
      errout:
	if (check)
	  {
	    if (scm_is_null (env))
              error_unbound_variable (var);
	    else
	      scm_misc_error (NULL, "Damaged environment: ~S",
			      scm_list_1 (var));
	  }
	else 
	  {
            dereference_parameter (real_var);
            dereference_parameter (vloc);
            dereference_parameter (env);
	    /* A variable could not be found, but we shall
	       not throw an error. */
	    return &undef_object;
	  }
      }

    if (!scm_is_eq (TMP_CAR (vloc), var)) // XXX shouldn't happen
      {
	/* In effect by now, someone must have done our job or messed it up
	   completely. [original comment] */
      race:
	var = TMP_CAR (vloc);
        dereference_parameter (real_var);
        dereference_parameter (vloc);
	if (SCM_VARIABLEP (var)) {
          reference_count_return (*(SCM_VARIABLE_LOC (var)));
          dereference_parameter (env);
	  return SCM_VARIABLE_LOC (var);
        }
	if (SCM_ILOCP (var)) {
	  SCM *answer = scm_ilookup (var, env);
          reference_count_return (*answer);
          dereference_parameter (env);
          return (answer);
        }
	/* We can't cope with anything else than variables and ilocs.  When
	   a special form has been memoized (i.e. `let' into `#@let') we
	   return NULL and expect the calling function to do the right
	   thing.  For the evaluator, this means going back and redoing
	   the dispatch on the car of the form. */
        dereference_parameter (real_var);
        dereference_parameter (vloc);
        dereference_parameter (env);
	return NULL;
      }

    SCM_SETCAR (vloc, real_var); /* Memoizing into the expression form. */
    reference_count_return (*(SCM_VARIABLE_LOC (real_var)));
    dereference_parameter (real_var);
    dereference_parameter (vloc);
    dereference_parameter (env);
    return SCM_VARIABLE_LOC (real_var);
    /* Come on, be honest! It's never been about the race it's just
       been about drawing that little star in front of all those
       lookupcar calls. */
  }
}

SCM *
scm_lookupcar (SCM vloc, SCM genv, int check)
{
  SCM *loc = scm_lookupcar1 (vloc, genv, check);
  if (loc == NULL) {
    fprintf (stderr, " * VSI scm_lookupcar: FATAL ERROR * ");
    abort ();
  }
  return loc;
}



/*
 * Memoizing / Unmemoizing
 */


/* During execution, look up a symbol in the top level of the given local
 * environment and return the corresponding variable object.  If no binding
 * for the symbol can be found, an 'Unbound variable' error is signalled.  */
static SCM
lazy_memoize_variable (const SCM symbol, const SCM environment)
{ // RC: pass through
  const SCM top_level = scm_env_top_level (environment);
  const SCM variable = scm_sym2var (symbol, top_level, SCM_BOOL_F);

  if (SCM_UNDEFINED_P (variable))
    error_unbound_variable (symbol);
  else
    return variable;
}



static int
literal_p (const SCM symbol, const SCM env)
{
  const SCM variable = lookup_symbol (symbol, env);
  if (SCM_UNBNDP (variable))
    return 1;
  if (SCM_VARIABLEP (variable) && SCM_MACROP (TMP_VARIABLE_REF (variable)))
    return 1;
  else
    return 0;
}


/* Return true if the expression is self-quoting in the memoized code.  Thus,
 * some other objects (like e. g. vectors) are reported as self-quoting, which
 * according to R5RS would need to be quoted.  */
static int
is_self_quoting_p (const SCM expr)
{
  if (scm_is_pair (expr))
    return 0;
  else if (scm_is_symbol (expr))
    return 0;
  else if (scm_is_null (expr))
    return 0;
  else return 1;
}


SCM_SYMBOL (sym_three_question_marks, "???");

static SCM
unmemoize_expression (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  if (SCM_ILOCP (expr))
    {
      SCM frame_idx;
      unsigned long int frame_nr;
      SCM symbol_idx;
      unsigned long int symbol_nr;

      for (frame_idx = env, frame_nr = SCM_IFRAME (expr);
           frame_nr != 0; 
           frame_idx = TMP_CDR (frame_idx), --frame_nr)
        ;
      for (symbol_idx = TMP_CAAR (frame_idx), symbol_nr = SCM_IDIST (expr);
           symbol_nr != 0;
           symbol_idx = TMP_CDR (symbol_idx), --symbol_nr)
        ;
      reference_count_return (symbol_idx);
      dereference_parameter (expr);
      dereference_parameter (env);
      return SCM_ICDRP (expr) ? symbol_idx : SCM_CAR (symbol_idx);
    }
  else if (SCM_VARIABLEP (expr))
    {
      // If we implement namespaces as dictionaries we'll need to
      // take the environment into account to find all dictionaries
      // where that variable could have come from.
      const SCM sym = nature_reverse_lookup (SCM_BOOL_F, expr);
      dereference_parameter (expr);
      dereference_parameter (env);
      if (scm_is_true (sym)) {
        return sym;
      } else {
        dereference_return_value (sym);
        return sym_three_question_marks;
      }
    }
  else if (scm_is_simple_vector (expr))
    {
      dereference_parameter (env);
      return scm_list_2 (scm_sym_quote, rc_Fly (expr));
    }
  else if (!scm_is_pair (expr))
    {
      dereference_parameter (env);
      return rc_return_stack_reference (expr);
    }
  else if (SCM_ISYMP (TMP_CAR (expr)))
    {
      return unmemoize_builtin_macro (rc_Fly (expr), rc_Fly (env));
    }
  else
    {
      return unmemoize_exprs (rc_Fly (expr), rc_Fly (env));
    }
}


static SCM
unmemoize_exprs (const SCM exprs, const SCM env)
{
  reference_parameter (exprs);
  reference_parameter (env);

  SCM r_result = SCM_EOL;  // no initial immediate reference
  SCM expr_idx = exprs;    // list referenced
  SCM um_expr;

  /* Note that due to the current lazy memoizer we may find partially memoized
   * code during execution.  In such code we have to expect improper lists of
   * expressions: On the one hand, for such code syntax checks have not yet
   * fully been performed, on the other hand, there may be even legal code
   * like '(a . b) appear as an improper list of expressions as long as the
   * quote expression is still in its unmemoized form.  For this reason, the
   * following code handles improper lists of expressions until memoization
   * and execution have been completely separated.  */
  for (; scm_is_pair (expr_idx); expr_idx = TMP_CDR (expr_idx))
    {
      const SCM expr = TMP_CAR (expr_idx);  // list referenced

      /* In partially memoized code, lists of expressions that stem from a
       * body form may start with an ISYM if the body itself has not yet been
       * memoized.  This isym is just an internal marker to indicate that the
       * body still needs to be memoized.  An isym may occur at the very
       * beginning of the body or after one or more comment strings.  It is
       * dropped during unmemoization.  */
      if (!SCM_ISYMP (expr))
        {
          um_expr = unmemoize_expression (expr, env);
          r_result = rc_X (scm_cons (um_expr, rc_Fly (r_result)));
        }
    }

  // reference check point: exprs, env, r_result

  um_expr = unmemoize_expression (expr_idx, env);  // no reference: pass through
  if (!scm_is_null (r_result))
    {
      const SCM result = scm_reverse_x (r_result, SCM_UNDEFINED);
      SCM_SETCDR (r_result, um_expr);
      reference_count_return (result);
      dereference_parameter (exprs);
      dereference_parameter (env);
      dereference_parameter (r_result);
      return result;
    }
  else
    {
      dereference_parameter (exprs);
      dereference_parameter (env);
      dereference_parameter (r_result);
      return um_expr;
    }
}


/* Rewrite the body (which is given as the list of expressions forming the
 * body) into its internal form.  The internal form of a body (<expr> ...) is
 * just the body itself, but prefixed with an ISYM that denotes to what kind
 * of outer construct this body belongs: (<ISYM> <expr> ...).  A lambda body
 * starts with SCM_IM_LAMBDA, for example, a body of a let starts with
 * SCM_IM_LET, etc.
 *
 * It is assumed that the calling expression has already made sure that the
 * body is a proper list.  */
static SCM
m_body (SCM op, SCM exprs)
{
  /* Don't add another ISYM if one is present already. */
  if (SCM_ISYMP (TMP_CAR (exprs))) {
    reference_count_return (exprs);
    dereference_return_value (op);
    return exprs;
  } else {
    return scm_cons (op, exprs);
  }
}


/* The function m_expand_body memoizes a proper list of expressions forming a
 * body.  This function takes care of dealing with internal defines and
 * transforming them into an equivalent letrec expression.  The list of
 * expressions is rewritten in place.  */ 

/* This is a helper function for m_expand_body.  If the argument expression is
 * a symbol that denotes a syntactic keyword, the corresponding macro object
 * is returned, in all other cases the function returns SCM_UNDEFINED.  */ 
static SCM
try_macro_lookup (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  if (scm_is_symbol (expr)) {
    const SCM variable = lookup_symbol (expr, env);  // rc pass through
    if (SCM_VARIABLEP (variable)) {
      const SCM value = SCM_VARIABLE_REF (variable);
      if (SCM_MACROP (value)) {
        dereference_parameter (expr);
        dereference_parameter (env);
        return value;
      } else {
        dereference_return_value (value);
      }
    } else {
      dereference_return_value (variable);  // bogus?
    }
  }

  dereference_parameter (expr);
  dereference_parameter (env);

  return SCM_UNDEFINED;
}

/* This is a helper function for m_expand_body.  It expands user macros,
 * because for the correct translation of a body we need to know whether they
 * expand to a definition. */ 
static SCM
expand_user_macros (SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  while (scm_is_pair (expr)) {
    const SCM car_expr = TMP_CAR (expr);
    const SCM new_car = expand_user_macros (car_expr, env);
    const SCM value = try_macro_lookup (new_car, env);
    reference_parameter (value);

    if (SCM_MACROP (value) && SCM_MACRO_TYPE (value) == 2) {
      /* User macros transform code into code.  */
      expr = rc_X (scm_call_2 (SCM_MACRO_CODE (value), rc_Fly (expr), env));
      dereference_parameter (value);
      /* We need to reiterate on the transformed code.  */

    } else { /* No user macro: return.  */
      SCM_SETCAR (expr, new_car);
      reference_count_return (expr);
      dereference_parameter (expr);
      dereference_parameter (value);
      dereference_parameter (env);
      return expr;
    }
  }

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return expr;
}

/* This is a helper function for m_expand_body.  It determines if a given form
 * represents an application of a given built-in macro.  The built-in macro to
 * check for is identified by its syntactic keyword.  The form is an
 * application of the given macro if looking up the car of the form in the
 * given environment actually returns the built-in macro.  */
static int
is_system_macro_p (const SCM syntactic_keyword, const SCM form, const SCM env)
{
  reference_parameter (syntactic_keyword);
  reference_parameter (form);
  reference_parameter (env);

  if (scm_is_pair (form))
    {
      const SCM car_form = TMP_CAR (form);
      const SCM value = try_macro_lookup (car_form, env);
      if (SCM_BUILTIN_MACRO_P (value))
        {
          reference_parameter (value);
          const SCM macro_name = scm_macro_name (value);
          dereference_parameter (value);

          int answer = scm_is_eq (macro_name, syntactic_keyword);
          dereference_return_value (macro_name);

          dereference_parameter (syntactic_keyword);
          dereference_parameter (form);
          dereference_parameter (env);
          return answer;
        }
    }

  dereference_parameter (syntactic_keyword);
  dereference_parameter (form);
  dereference_parameter (env);
  return 0;
}

static void
m_expand_body (const SCM forms, const SCM env)
{
  reference_parameter (forms);
  reference_parameter (env);

  /* The first body form can be skipped since it is known to be the ISYM that
   * was prepended to the body by m_body.  */
  SCM cdr_forms = TMP_CDR (forms); reference_parameter (cdr_forms);
  SCM form_idx = cdr_forms;        reference_parameter (form_idx);
  SCM definitions = SCM_EOL;
  SCM sequence = SCM_EOL;

  /* According to R5RS, the list of body forms consists of two parts: a number
   * (maybe zero) of definitions, followed by a non-empty sequence of
   * expressions.  Each the definitions and the expressions may be grouped
   * arbitrarily with begin, but it is not allowed to mix definitions and
   * expressions.  The task of the following loop therefore is to split the
   * list of body forms into the list of definitions and the sequence of
   * expressions. XXX This allows mixing which is a good thing but relies
   * on lazy memoization. Using /begin/ to group definitions is deprecated.
   * We should memoize the closure instead of a lambda form. */
  while (!scm_is_null (form_idx)) {
    const SCM form = TMP_CAR (form_idx);
    const SCM new_form = expand_user_macros (form, env);
    reference_parameter (new_form);

    if (is_system_macro_p (scm_sym_define, new_form, env)) {
      definitions = rc_X (scm_cons (new_form, rc_Fly (definitions)));
      form_idx = FLY_CDR (form_idx);

    } else if (is_system_macro_p (scm_sym_begin, new_form, env)) {
      /* We have encountered a group of forms.  This has to be either a
       * (possibly empty) group of (possibly further grouped) definitions,
       * or a non-empty group of (possibly further grouped)
       * expressions.  */
      const SCM grouped_forms = TMP_CDR (new_form);
      unsigned int found_definition = 0;
      unsigned int found_expression = 0;
      SCM grouped_form_idx = grouped_forms;
      while (!found_expression && !scm_is_null (grouped_form_idx)) {
        const SCM inner_form = TMP_CAR (grouped_form_idx);
        reference_parameter (inner_form);
        const SCM new_inner_form = expand_user_macros (inner_form, env);
        reference_parameter (new_inner_form);
        dereference_parameter (inner_form);

        if (is_system_macro_p (scm_sym_define, new_inner_form, env)) {
          found_definition = 1;
          definitions = rc_X (scm_cons (new_inner_form, rc_Fly (definitions)));
          grouped_form_idx = TMP_CDR (grouped_form_idx);

        } else if (is_system_macro_p (scm_sym_begin, new_inner_form, env)) {
          const SCM inner_group = TMP_CDR (new_inner_form);
          grouped_form_idx \
            = scm_append (scm_list_2 (inner_group,
                                      TMP_CDR (grouped_form_idx)));

        } else {
          /* The group marks the start of the expressions of the body.
           * We have to make sure that within the same group we have
          * not encountered a definition before.  */
          ASSERT_SYNTAX (!found_definition, s_mixed_body_forms, form);
          found_expression = 1;
          grouped_form_idx = SCM_EOL;
        }
        dereference_parameter (new_inner_form);
      }

      /* We have finished processing the group.  If we have not yet
       * encountered an expression we continue processing the forms of the
       * body to collect further definition forms.  Otherwise, the group
       * marks the start of the sequence of expressions of the body.  */
      if (!found_expression) {
        form_idx = FLY_CDR (form_idx);
      } else {
        sequence = form_idx;
        reference_parameter (sequence);
        dereference_parameter (form_idx);
        form_idx = SCM_EOL;
      }

    } else {
      /* We have detected a form which is no definition.  This marks the
       * start of the sequence of expressions of the body.  */
      sequence = form_idx;
      reference_parameter (sequence);
      dereference_parameter (form_idx);
      form_idx = SCM_EOL;
    }
    dereference_parameter (new_form);
  } // closes while

  ASSERT_SYNTAX (scm_is_pair (sequence), s_missing_body_expression, cdr_forms);
  dereference_parameter (cdr_forms);

  // RC checkpoint: definitions (or im), sequence (or im), forms, env

  if (!scm_is_null (definitions)) {
    SCM definition_idx;
    SCM letrec_tail;
    SCM letrec_expression;
    SCM new_letrec_expression;

    SCM bindings = SCM_EOL;
    for (definition_idx = definitions;
         !scm_is_null (definition_idx);
         definition_idx = TMP_CDR (definition_idx)) {

      const SCM definition = TMP_CAR (definition_idx);
      const SCM canonical_definition = canonicalize_define (definition);
      const SCM binding = TMP_CDR (canonical_definition);
      bindings = scm_cons (binding, bindings);
    };

    letrec_tail = scm_cons (bindings, sequence);
    letrec_expression = scm_cons_source (forms, scm_sym_letrec, letrec_tail);
    new_letrec_expression = scm_m_letrec (letrec_expression, env);
    SCM_SETCAR (forms, new_letrec_expression);
    SCM_SETCDR (forms, SCM_EOL);

  } else {
    SCM_SETCAR (forms, TMP_CAR (sequence));
    SCM_SETCDR (forms, TMP_CDR (sequence));
  }

  dereference_parameter (definitions);
  dereference_parameter (sequence);
  dereference_parameter (forms);
  dereference_parameter (env);
  return;
}

static SCM
macroexp (SCM x, SCM env)
{
  reference_parameter (x);
  reference_parameter (env);

  SCM res, proc, orig_sym;

  /* Don't bother to produce error messages here.  We get them when we
     eventually execute the code for real. */

 macro_tail:
  orig_sym = TMP_CAR (x);
  if (!scm_is_symbol (orig_sym)) {
    reference_count_return (x);
    dereference_parameter (x);
    dereference_parameter (env);
    return (x);
  }

  {
    SCM *proc_ptr = scm_lookupcar1 (x, env, 0);
    if (proc_ptr == NULL)
      {
	/* We have lost the race. */
	goto macro_tail;
      }
    proc = *proc_ptr;
  }
  reference_parameter (proc);
  
  /* Only handle memoizing macros.  `Acros' and `macros' are really
     special forms and should not be evaluated here. */

  if (!SCM_MACROP (proc)
      || (SCM_MACRO_TYPE (proc) != 2 && !SCM_BUILTIN_MACRO_P (proc))) {
    reference_count_return (x);
    dereference_parameter (x);
    dereference_parameter (env);
    dereference_parameter (proc);
    return (x);
  }

  SCM_SETCAR (x, orig_sym);  /* Undo memoizing effect of lookupcar */
  res = scm_call_2 (SCM_MACRO_CODE (proc), x, env);
  reference_parameter (res);

  if (scm_ilength (res) <= 0) {
    /* Result of expansion is not a list.  */
    dereference_parameter (x);
    dereference_parameter (env);
    dereference_parameter (proc);
    return (scm_list_2 (SCM_IM_BEGIN, rc_Fly (res)));
  } else {
    /* macroexp() is now only used in one place -
       scm_m_generalized_set_x - whereas all other macro expansion
       happens through expand_user_macros.
    */

    SCM_SETCAR (x, TMP_CAR (res));
    SCM_SETCDR (x, TMP_CDR (res));
    dereference_parameter (res);

    dereference_parameter (proc);
    goto macro_tail;
  }
}

/* Start of the memoizers for the standard R5RS builtin macros.  */


SCM_SYNTAX (s_and, "and", scm_i_makbimacro, scm_m_and);
SCM_GLOBAL_SYMBOL (scm_sym_and, s_and);

SCM
scm_m_and (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  const long length = scm_ilength (cdr_expr);

  ASSERT_SYNTAX (length >= 0, s_bad_expression, expr);

  if (length == 0)
    {
      /* Special case:  (and) is replaced by #t. */
      dereference_parameter (expr);
      dereference_parameter (env);
      return SCM_BOOL_T;
    }
  else
    {
      SCM_SETCAR (expr, SCM_IM_AND);
      reference_count_return (expr);
      dereference_parameter (expr);
      dereference_parameter (env);
      return expr;
    }
}

static SCM
unmemoize_and (const SCM expr, const SCM env)
{
  return scm_cons (scm_sym_and, unmemoize_exprs (SCM_CDR (expr), env));
}


SCM_SYNTAX (s_begin, "begin", scm_i_makbimacro, scm_m_begin);
SCM_GLOBAL_SYMBOL (scm_sym_begin, s_begin);

SCM
scm_m_begin (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);

  SCM_SETCAR (expr, SCM_IM_BEGIN);

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return (expr);
}

static SCM
unmemoize_begin (const SCM expr, const SCM env)
{
  return scm_cons (scm_sym_begin, unmemoize_exprs (SCM_CDR (expr), env));
}


SCM_SYNTAX (s_case, "case", scm_i_makbimacro, scm_m_case);
SCM_GLOBAL_SYMBOL (scm_sym_case, s_case);
SCM_GLOBAL_SYMBOL (scm_sym_else, "else");

SCM
scm_m_case (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM clauses;
  SCM all_labels = SCM_EOL;

  /* Check, whether 'else is a literal, i. e. not bound to a value. */
  const int else_literal_p = literal_p (scm_sym_else, env);

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 2, s_missing_clauses, expr);

  clauses = TMP_CDR (cdr_expr);
  while (!scm_is_null (clauses))
    {
      SCM labels;

      const SCM clause = TMP_CAR (clauses);
      ASSERT_SYNTAX_2 (scm_ilength (clause) >= 2, 
		       s_bad_case_clause, clause, expr);

      labels = TMP_CAR (clause);
      if (scm_is_pair (labels))
        {
          ASSERT_SYNTAX_2 (scm_ilength (labels) >= 0,
                           s_bad_case_labels, labels, expr);
          all_labels = rc_X (scm_append (scm_list_2 \
                                         (labels, rc_Fly (all_labels))));
        }
      else if (scm_is_null (labels))
        {
          /* The list of labels is empty.  According to R5RS this is allowed.
           * It means that the sequence of expressions will never be executed.
           * Therefore, as an optimization, we could remove the whole
           * clause.  */
        }
      else
        {
          ASSERT_SYNTAX_2 (scm_is_eq (labels, scm_sym_else) && else_literal_p,
                           s_bad_case_labels, labels, expr);
          ASSERT_SYNTAX_2 (scm_is_null (TMP_CDR (clauses)),
                           s_misplaced_else_clause, clause, expr);
        }

      /* build the new clause */
      if (scm_is_eq (labels, scm_sym_else))
        SCM_SETCAR (clause, SCM_IM_ELSE);

      clauses = TMP_CDR (clauses);
    }

  /* Check whether all case labels are distinct. */
  for (; !scm_is_null (all_labels); all_labels = FLY_CDR (all_labels))
    {
      SCM tmp = SCM_UNSPECIFIED;
      const SCM label = TMP_CAR (all_labels);
      ASSERT_SYNTAX_2 \
        (scm_is_false ((tmp=scm_c_memq (label, TMP_CDR (all_labels)))),
         s_duplicate_case_label, label, expr);
      dereference_return_value (tmp);
    }
  dereference_parameter (all_labels);

  SCM_SETCAR (expr, SCM_IM_CASE);

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return (expr);
}

static SCM
unmemoize_case (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM um_key_expr = unmemoize_expression (TMP_CADR (expr), env);
  SCM um_clauses = SCM_EOL;
  SCM clause_idx;

  for (clause_idx = TMP_CDDR (expr);
       !scm_is_null (clause_idx);
       clause_idx = TMP_CDR (clause_idx))
    {
      const SCM clause = TMP_CAR (clause_idx);
      const SCM labels = TMP_CAR (clause);
      const SCM exprs = TMP_CDR (clause);

      const SCM um_exprs = unmemoize_exprs (exprs, env);
      const SCM um_labels = (scm_is_eq (labels, SCM_IM_ELSE))
        ? scm_sym_else
        : scm_i_finite_list_copy (labels);
      const SCM um_clause = scm_cons (um_labels, um_exprs);

      um_clauses = scm_cons (um_clause, um_clauses);
    }
  um_clauses = scm_reverse_x (um_clauses, SCM_UNDEFINED);

  dereference_parameter (expr);
  dereference_parameter (env);
  return scm_cons2 (scm_sym_case, um_key_expr, um_clauses);
}


SCM_SYNTAX (s_cond, "cond", scm_i_makbimacro, scm_m_cond);
SCM_GLOBAL_SYMBOL (scm_sym_cond, s_cond);
SCM_GLOBAL_SYMBOL (scm_sym_arrow, "=>");

SCM
scm_m_cond (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  /* Check, whether 'else or '=> is a literal, i. e. not bound to a value. */
  const int else_literal_p = literal_p (scm_sym_else, env);
  const int arrow_literal_p = literal_p (scm_sym_arrow, env);

  const SCM clauses = TMP_CDR (expr);
  SCM clause_idx;

  ASSERT_SYNTAX (scm_ilength (clauses) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (clauses) >= 1, s_missing_clauses, expr);

  for (clause_idx = clauses;
       !scm_is_null (clause_idx);
       clause_idx = TMP_CDR (clause_idx))
    {
      SCM test;

      const SCM clause = TMP_CAR (clause_idx);
      const long length = scm_ilength (clause);
      ASSERT_SYNTAX_2 (length >= 1, s_bad_cond_clause, clause, expr);

      test = TMP_CAR (clause);
      if (scm_is_eq (test, scm_sym_else) && else_literal_p)
	{
	  const int last_clause_p = scm_is_null (TMP_CDR (clause_idx));
          ASSERT_SYNTAX_2 (length >= 2,
                           s_bad_cond_clause, clause, expr);
          ASSERT_SYNTAX_2 (last_clause_p,
                           s_misplaced_else_clause, clause, expr);
          SCM_SETCAR (clause, SCM_IM_ELSE);
	}
      else if (length >= 2
               && scm_is_eq (TMP_CADR (clause), scm_sym_arrow)
               && arrow_literal_p)
        {
          ASSERT_SYNTAX_2 (length > 2, s_missing_recipient, clause, expr);
          ASSERT_SYNTAX_2 (length == 3, s_extra_expression, clause, expr);
          SCM_SETCAR (TMP_CDR (clause), SCM_IM_ARROW);
	}
      /* SRFI 61 extended cond */
      else if (length >= 3
	       && scm_is_eq (TMP_CADDR (clause), scm_sym_arrow)
	       && arrow_literal_p)
	{
	  ASSERT_SYNTAX_2 (length > 3, s_missing_recipient, clause, expr);
	  ASSERT_SYNTAX_2 (length == 4, s_extra_expression, clause, expr);
	  SCM_SETCAR (TMP_CDDR (clause), SCM_IM_ARROW);
	}
    }

  SCM_SETCAR (expr, SCM_IM_COND);

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return (expr);
}

static SCM
unmemoize_cond (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM um_clauses = SCM_EOL;
  SCM clause_idx;

  for (clause_idx = TMP_CDR (expr);
       !scm_is_null (clause_idx);
       clause_idx = TMP_CDR (clause_idx))
    {
      const SCM clause = TMP_CAR (clause_idx);
      const SCM sequence = TMP_CDR (clause);
      const SCM test = TMP_CAR (clause);
      SCM um_test;
      SCM um_sequence;
      SCM um_clause;

      if (scm_is_eq (test, SCM_IM_ELSE))
        um_test = scm_sym_else;
      else
        um_test = unmemoize_expression (test, env);

      if (!scm_is_null (sequence) && scm_is_eq (TMP_CAR (sequence),
					      SCM_IM_ARROW))
        {
          const SCM target = TMP_CADR (sequence);
          const SCM um_target = unmemoize_expression (target, env);
          um_sequence = scm_list_2 (scm_sym_arrow, um_target);
        }
      else
        {
          um_sequence = unmemoize_exprs (sequence, env);
        }

      um_clause = scm_cons (um_test, um_sequence);
      um_clauses = scm_cons (um_clause, um_clauses);
    }
  um_clauses = scm_reverse_x (um_clauses, SCM_UNDEFINED);

  dereference_parameter (expr);
  dereference_parameter (env);
  return scm_cons (scm_sym_cond, um_clauses);
}


SCM_SYNTAX (s_define, "define", scm_i_makbimacro, scm_m_define);
SCM_GLOBAL_SYMBOL (scm_sym_define, s_define);

static SCM
canonicalize_define (const SCM expr)
{
  reference_parameter (expr);

  SCM body;
  SCM variable;

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 2, s_missing_expression, expr);

  body = TMP_CDR (cdr_expr);
  variable = TMP_CAR (cdr_expr);
  if (scm_is_pair (variable)) {
    const SCM formals = TMP_CDR (variable);
    const SCM tail = scm_cons (formals, body);

    /* Add source properties to the lambda expression:  */
    const SCM lambda = scm_cons_source (variable, scm_sym_lambda, tail);

    body = scm_list_1 (lambda);
    variable = TMP_CAR (variable);
  }
  ASSERT_SYNTAX_2 (scm_is_symbol (variable), s_bad_variable, variable, expr);
  ASSERT_SYNTAX (scm_ilength (body) == 1, s_expression, expr);

  SCM_SETCAR (cdr_expr, variable);
  SCM_SETCDR (cdr_expr, body);

  return rc_return_stack_reference (expr);
}

SCM
scm_m_define (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  // XXX x-debugging ASSERT_SYNTAX (SCM_TOP_LEVEL (env), s_bad_define, expr);

  {
    const SCM canonical_definition = canonicalize_define (expr);
    reference_parameter (canonical_definition);  // bogus
    const SCM cdr_canonical_definition = TMP_CDR (canonical_definition);
    const SCM variable = TMP_CAR (cdr_canonical_definition);
    const SCM value = scm_eval_car (TMP_CDR (cdr_canonical_definition), env);
    reference_parameter (value);
    const SCM location \
      = scm_sym2var (variable, scm_env_top_level (env), SCM_BOOL_T);
    reference_parameter (location);

    /* debugging paragraph */
    if (global_debug_interpreter) {
      SCM sym = variable;
      reference_parameter (sym);
      SCM var = location;
      reference_parameter (var);
      SCM val = TMP_VARIABLE_REF (var);
      reference_parameter (val);

      if (val != SCM_UNDEFINED) {
        SCM bym = scm_to_sen (scm_symbol_to_brass (sym));
        reference_parameter (bym);
        fprintf (stderr, "define: system variable redefinition: %s\n",
                         BYTE_STRING_DATA (bym));
        dereference_parameter (bym);
      }
      dereference_parameter (sym);
      dereference_parameter (var);
      dereference_parameter (val);
    } /* end of debugging paragraph */

    {
      SCM tmp = value;

      while (SCM_MACROP (tmp))
        tmp = SCM_MACRO_CODE (tmp);
      if (scm_is_true (scm_procedure_p (tmp))
          /* Only the first definition determines the name. */
          && sharp_scm_is_false (scm_procedure_property (tmp, scm_sym_name)))
        scm_set_procedure_property_x (tmp, scm_sym_name, variable);
    }

    SCM_VARIABLE_SET (location, value);

    dereference_parameter (location);
    dereference_parameter (canonical_definition);
    dereference_parameter (value);
    dereference_parameter (expr);
    dereference_parameter (env);
    return SCM_UNSPECIFIED;
  }
}


SCM_SYNTAX (s_if, "if", scm_i_makbimacro, scm_m_if);
SCM_GLOBAL_SYMBOL (scm_sym_if, s_if);

SCM
scm_m_if (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  const long length = scm_ilength (cdr_expr);
  ASSERT_SYNTAX (length == 2 || length == 3, s_expression, expr);
  SCM_SETCAR (expr, SCM_IM_IF);

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return expr;
}

static SCM
unmemoize_if (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  const SCM um_condition = unmemoize_expression (TMP_CAR (cdr_expr), env);
  reference_parameter (um_condition);

  const SCM cddr_expr = TMP_CDR (cdr_expr);
  const SCM um_then = unmemoize_expression (TMP_CAR (cddr_expr), env);
  reference_parameter (um_then);

  const SCM cdddr_expr = TMP_CDR (cddr_expr);
  if (scm_is_null (cdddr_expr)) {
    dereference_parameter (expr);
    dereference_parameter (env);
    return scm_list_3 (scm_sym_if, rc_Fly (um_condition),
                                   rc_Fly (um_then));
  } else {
    const SCM um_else = unmemoize_expression (TMP_CAR (cdddr_expr), env);
    reference_parameter (um_else);
    dereference_parameter (expr);
    dereference_parameter (env);
    return scm_list_4 (scm_sym_if, rc_Fly (um_condition),
                                   rc_Fly (um_then),
                                   rc_Fly (um_else));
  }
}


SCM_SYNTAX (s_lambda, "lambda", scm_i_makbimacro, scm_m_lambda);
SCM_GLOBAL_SYMBOL (scm_sym_lambda, s_lambda);

/* A helper function for memoize_lambda to support checking for duplicate
 * formal arguments: Return true if OBJ is `eq?' to one of the elements of
 * LIST or to the cdr of the last cons.  Therefore, LIST may have any of the
 * forms that a formal argument can have:
 *   <rest>, (<arg1> ...), (<arg1> ...  .  <rest>) */
static int
c_improper_memq (SCM obj, SCM list)
{
  for (; scm_is_pair (list); list = TMP_CDR (list))
    {
      if (scm_is_eq (TMP_CAR (list), obj))
        return 1;
    }
  return scm_is_eq (list, obj);
}

SCM
scm_m_lambda (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM formals;
  SCM formals_idx;
  SCM cddr_expr;
  int documentation;
  SCM body;
  SCM new_body;

  const SCM cdr_expr = TMP_CDR (expr);
  const long length = scm_ilength (cdr_expr);
  ASSERT_SYNTAX (length >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (length >= 2, s_missing_expression, expr);

  /* Before iterating the list of formal arguments, make sure the formals
   * actually are given as either a symbol or a non-cyclic list.  */
  formals = TMP_CAR (cdr_expr);
  if (scm_is_pair (formals)) {
    // debug_trim_assert_list (formals);
  } else {
    ASSERT_SYNTAX_2 (scm_is_symbol (formals) || scm_is_null (formals),
                     s_bad_formals, formals, expr);
  }

  /* Now iterate the list of formal arguments to check if all formals are
   * symbols, and that there are no duplicates.  */
  formals_idx = formals;
  while (scm_is_pair (formals_idx))
    {
      const SCM formal = TMP_CAR (formals_idx);
      const SCM next_idx = TMP_CDR (formals_idx);
      ASSERT_SYNTAX_2 (scm_is_symbol (formal), s_bad_formal, formal, expr);
      ASSERT_SYNTAX_2 (!c_improper_memq (formal, next_idx),
                       s_duplicate_formal, formal, expr);
      formals_idx = next_idx;
    }
  ASSERT_SYNTAX_2 (scm_is_null (formals_idx) || scm_is_symbol (formals_idx),
                   s_bad_formal, formals_idx, expr);

  cddr_expr = TMP_CDR (cdr_expr);
  // documentation = (length >= 3 && scm_is_string (TMP_CAR (cddr_expr))); // XXX
  documentation = 0;
  body = documentation ? TMP_CDR (cddr_expr) : cddr_expr;
  new_body = m_body (SCM_IM_LAMBDA, body);
  reference_parameter (new_body);

  SCM_SETCAR (expr, SCM_IM_LAMBDA);
  if (documentation)
    SCM_SETCDR (cddr_expr, new_body);
  else
    SCM_SETCDR (cdr_expr, new_body);

  reference_count_return (expr);
  dereference_parameter (new_body);
  dereference_parameter (expr);
  dereference_parameter (env);
  return expr;
}

static SCM
unmemoize_lambda (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM formals = TMP_CADR (expr);
  const SCM body = TMP_CDDR (expr);

  const SCM new_env = SCM_EXTEND_ENV (formals, SCM_EOL, env);
  reference_parameter (new_env);

  const SCM um_formals = scm_i_finite_list_copy (formals);
  reference_parameter (um_formals);

  const SCM um_body = unmemoize_exprs (body, new_env);
  reference_parameter (um_body);

  dereference_parameter (expr);
  dereference_parameter (env);
  dereference_parameter (new_env);
  return scm_cons2 (scm_sym_lambda, rc_Fly (um_formals),
                                    rc_Fly (um_body));
}


/* Check if the format of the bindings is ((<symbol> <init-form>) ...).  */
static void
check_bindings (const SCM bindings, const SCM expr)
{
  SCM binding_idx;

  ASSERT_SYNTAX_2 (scm_ilength (bindings) >= 0,
                   s_bad_bindings, bindings, expr);

  binding_idx = bindings;
  for (; !scm_is_null (binding_idx); binding_idx = TMP_CDR (binding_idx))
    {
      SCM name;         /* const */

      const SCM binding = TMP_CAR (binding_idx);
      ASSERT_SYNTAX_2 (scm_ilength (binding) == 2,
                       s_bad_binding, binding, expr);

      name = TMP_CAR (binding);
      ASSERT_SYNTAX_2 (scm_is_symbol (name), s_bad_variable, name, expr);
    }
}


/* The bindings, which must have the format ((v1 i1) (v2 i2) ... (vn in)), are
 * transformed to the lists (vn ... v2 v1) and (i1 i2 ... in).  That is, the
 * variables are returned in a list with their order reversed, and the init
 * forms are returned in a list in the same order as they are given in the
 * bindings.  If a duplicate variable name is detected, an error is
 * signalled.  */
static void
transform_bindings (
  const SCM bindings, const SCM expr,
  SCM *const rvarptr, SCM *const initptr )
{
  SCM rvariables = SCM_EOL;
  SCM rinits = SCM_EOL;
  SCM binding_idx = bindings;
  for (; !scm_is_null (binding_idx); binding_idx = TMP_CDR (binding_idx))
    {
      const SCM binding = TMP_CAR (binding_idx);
      const SCM cdr_binding = TMP_CDR (binding);
      const SCM name = TMP_CAR (binding);
      SCM tmp;
      ASSERT_SYNTAX_2 (scm_is_false ((tmp=scm_c_memq (name, rvariables))),
                       s_duplicate_binding, name, expr);
      dereference_return_value (tmp);
      rvariables = rc_X (scm_cons (name, rc_Fly (rvariables)));
      rinits     = rc_X (scm_cons (TMP_CAR (cdr_binding), rc_Fly (rinits)));
    }
  *rvarptr = rc_Fly (rvariables);
  *initptr = scm_reverse_x (rc_Fly (rinits), SCM_UNDEFINED);
}


SCM_SYNTAX(s_let, "let", scm_i_makbimacro, scm_m_let);
SCM_GLOBAL_SYMBOL(scm_sym_let, s_let);

/* This function is a helper function for memoize_let.  It transforms
 * (let name ((var init) ...) body ...) into
 * ((letrec ((name (lambda (var ...) body ...))) name) init ...)
 * and memoizes the expression.  It is assumed that the caller has checked
 * that name is a symbol and that there are bindings and a body.  */
static SCM
memoize_named_let (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM rvariables;  // RC: pass through value
  SCM variables;
  SCM inits;

  const SCM cdr_expr  = TMP_CDR (expr);
  const SCM name      = TMP_CAR (cdr_expr);
  const SCM cddr_expr = TMP_CDR (cdr_expr);
  const SCM bindings  = TMP_CAR (cddr_expr);
  check_bindings (bindings, expr);

  transform_bindings (bindings, expr, &rvariables, &inits);
  reference_parameter (inits);
  variables = scm_reverse_x (rvariables, SCM_UNDEFINED);
  reference_parameter (variables);

  {
    const SCM let_body = TMP_CDR (cddr_expr);
    const SCM lambda_body = m_body (SCM_IM_LET, let_body);
    const SCM lambda_tail = scm_cons (variables, lambda_body);
    const SCM lambda_form = scm_cons_source (expr, scm_sym_lambda, lambda_tail);

    const SCM rvar = scm_list_1 (name);
    const SCM init = scm_list_1 (lambda_form);
    const SCM body = m_body (SCM_IM_LET, scm_list_1 (name));
    const SCM letrec_tail = scm_cons (rvar, scm_cons (init, body));
    const SCM letrec_form = scm_cons_source (expr, SCM_IM_LETREC, letrec_tail);

    dereference_parameter (env);
    dereference_parameter (variables);
    return scm_cons_source (rc_Fly (expr), letrec_form, rc_Fly (inits));
  }
}

SCM_DEFINE (scm_debug_cyan, "debug-cyan", 1, 0, 0, 
            (SCM x),
	    "Temporary debugging dummy routine (cyan).")
#define FUNC_NAME s_scm_debug_cyan
{
  m_expand_body (x, SCM_EOL);
  return (x);
}
#undef FUNC_NAME

/* (let ((v1 i1) (v2 i2) ...) body) with variables v1 .. vn and initializers
 * i1 .. in is transformed to (#@let (vn ... v2 v1) (i1 i2 ...) body).  */
SCM
scm_m_let (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM bindings;

  const SCM cdr_expr = TMP_CDR (expr);
  const long length = scm_ilength (cdr_expr);
  ASSERT_SYNTAX (length >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (length >= 2, s_missing_expression, expr);

  bindings = TMP_CAR (cdr_expr);
  if (scm_is_symbol (bindings))
    {
      ASSERT_SYNTAX (length >= 3, s_missing_expression, expr);
      return memoize_named_let (expr, env);
    }

  check_bindings (bindings, expr); // lax wrt bindings

  if (scm_is_null (bindings) || scm_is_null (TMP_CDR (bindings)))
    {
      /* Special case: no bindings or single binding => let* is faster. */
      const SCM body = m_body (SCM_IM_LET, TMP_CDR (cdr_expr));
      SCM new_expr = scm_cons2 (TMP_CAR (expr), bindings, body);

      // ??? should we: SCM_SETCDR (expr, body);
      reference_count_return (new_expr);
      dereference_parameter (expr);
      return scm_m_letstar (new_expr, rc_Fly (env));
    }
  else
    {
      /* plain let */
      SCM rvariables;
      SCM inits;
      transform_bindings (bindings, expr, &rvariables, &inits);

      const SCM new_body = m_body (SCM_IM_LET, TMP_CDR (cdr_expr));
      const SCM new_tail = scm_cons2 (rvariables, inits, new_body);
      SCM_SETCAR (expr, SCM_IM_LET);
      SCM_SETCDR (expr, new_tail);

      reference_count_return (expr);
      dereference_parameter (expr);
      dereference_parameter (env);
      return expr;
    }
}

static SCM
build_binding_list (SCM rnames, SCM rinits)
{
  SCM bindings = SCM_EOL;
  while (!scm_is_null (rnames))
    {
      const SCM binding = scm_list_2 (TMP_CAR (rnames), TMP_CAR (rinits));
      bindings = rc_X (scm_cons (binding, rc_Fly (bindings)));
      rnames = TMP_CDR (rnames);
      rinits = TMP_CDR (rinits);
    }
  return rc_return_stack_reference (bindings);
}

static SCM
unmemoize_let (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  const SCM um_rnames = TMP_CAR (cdr_expr);
  reference_parameter (um_rnames);

  const SCM extended_env = SCM_EXTEND_ENV (um_rnames, SCM_EOL, env);
  reference_parameter (extended_env);

  const SCM cddr_expr = TMP_CDR (cdr_expr);
  const SCM um_inits = unmemoize_exprs (TMP_CAR (cddr_expr), env);
  reference_parameter (um_inits);

  const SCM um_rinits = scm_reverse_x (um_inits, SCM_UNDEFINED);
  reference_parameter (um_rinits);

  const SCM um_bindings = build_binding_list (um_rnames, um_rinits);
  reference_parameter (um_bindings);

  const SCM um_body = unmemoize_exprs (TMP_CDR (cddr_expr), extended_env);
  reference_parameter (um_body);

  dereference_parameter (um_rnames);
  dereference_parameter (extended_env);
  dereference_parameter (um_inits);
  dereference_parameter (um_rinits);
  dereference_parameter (expr);
  dereference_parameter (env);
  return scm_cons2 (scm_sym_let, rc_Fly (um_bindings), rc_Fly (um_body));
}


SCM_SYNTAX(s_letrec, "letrec", scm_i_makbimacro, scm_m_letrec);
SCM_GLOBAL_SYMBOL(scm_sym_letrec, s_letrec);

SCM 
scm_m_letrec (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);
  SCM bindings;

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 2, s_missing_expression, expr);

  bindings = TMP_CAR (cdr_expr);
  if (scm_is_null (bindings))
    {
      /* no bindings, let* is executed faster */
      SCM body = m_body (SCM_IM_LETREC, TMP_CDR (cdr_expr));
      SCM answer = \
        scm_m_letstar (scm_cons2 (TMP_CAR (expr), SCM_EOL, body), env);
      reference_count_return (answer);
      dereference_parameter (expr);
      dereference_parameter (env);
      return answer;
    }
  else
    {
      reference_parameter (bindings);

      SCM rvariables;
      SCM inits;
      SCM new_body;

      check_bindings (bindings, expr);
      transform_bindings (bindings, expr, &rvariables, &inits);
      reference_parameter (rvariables);
      reference_parameter (inits);

      new_body = m_body (SCM_IM_LETREC, SCM_CDR (cdr_expr));
      SCM answer = \
        scm_cons2 (SCM_IM_LETREC, rvariables, scm_cons (inits, new_body));

      reference_count_return (answer);
      dereference_parameter (expr);
      dereference_parameter (env);
      dereference_parameter (bindings);
      dereference_parameter (rvariables);
      dereference_parameter (inits);
      return (answer);
    }
}

static SCM
unmemoize_letrec (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  const SCM um_rnames = TMP_CAR (cdr_expr);
  reference_parameter (um_rnames);

  const SCM extended_env = SCM_EXTEND_ENV (um_rnames, SCM_EOL, env);
  reference_parameter (extended_env);

  const SCM cddr_expr = TMP_CDR (cdr_expr);
  const SCM um_inits = unmemoize_exprs (TMP_CAR (cddr_expr), extended_env);
  reference_parameter (um_inits);

  const SCM um_rinits = scm_reverse_x (um_inits, SCM_UNDEFINED);
  const SCM um_bindings = build_binding_list (um_rnames, um_rinits);

  const SCM um_body = unmemoize_exprs (SCM_CDR (cddr_expr), extended_env);
  SCM answer = scm_cons2 (scm_sym_letrec, um_bindings, um_body);

  reference_count_return (answer);
  dereference_parameter (um_inits);
  dereference_parameter (um_rnames);
  dereference_parameter (extended_env);
  dereference_parameter (expr);
  dereference_parameter (env);
  return (answer);
}



SCM_SYNTAX (s_letstar, "let*", scm_i_makbimacro, scm_m_letstar);
SCM_GLOBAL_SYMBOL (scm_sym_letstar, s_letstar);

/* (let* ((v1 i1) (v2 i2) ...) body) with variables v1 .. vn and initializers
 * i1 .. in is transformed into the form (#@let* (v1 i1 v2 i2 ...) body).  */
SCM
scm_m_letstar (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM binding_idx;
  SCM new_body;

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 2, s_missing_expression, expr);

  binding_idx = TMP_CAR (cdr_expr);
  check_bindings (binding_idx, expr);

  /* Transform ((v1 i1) (v2 i2) ...) into (v1 i1 v2 i2 ...).  The
   * transformation is done in place.  At the beginning of one iteration of
   * the loop the variable binding_idx holds the form
   *   P1:( (vn . P2:(in . ())) . P3:( (vn+1 in+1) ... ) ),
   * where P1, P2 and P3 indicate the pairs, that are relevant for the
   * transformation.  P1 and P2 are modified in the loop, P3 remains
   * untouched.  After the execution of the loop, P1 will hold
   *   P1:( vn . P2:(in . P3:( (vn+1 in+1) ... )) )
   * and binding_idx will hold P3.  */
  while (!scm_is_null (binding_idx))
    {
      const SCM cdr_binding_idx = TMP_CDR (binding_idx);  /* remember P3 */
      const SCM binding = TMP_CAR (binding_idx);
      const SCM name = TMP_CAR (binding);
      const SCM cdr_binding = TMP_CDR (binding);

      reference_parameter (cdr_binding);
      reference_parameter (cdr_binding_idx);
      reference_parameter (name);
      SCM_SETCDR (cdr_binding, cdr_binding_idx);        /* update P2 */
      SCM_SETCAR (binding_idx, name);                   /* update P1 */
      SCM_SETCDR (binding_idx, cdr_binding);            /* update P1 */
      dereference_parameter (cdr_binding);
      dereference_parameter (cdr_binding_idx);
      dereference_parameter (name);

      binding_idx = cdr_binding_idx;                    /* continue with P3 */
    }

  new_body = m_body (SCM_IM_LETSTAR, TMP_CDR (cdr_expr));
  SCM_SETCAR (expr, SCM_IM_LETSTAR);
  /* the bindings have been changed in place */
  SCM_SETCDR (cdr_expr, new_body);

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return expr;
}

static SCM
unmemoize_letstar (const SCM expr, const SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  const SCM body = TMP_CDR (cdr_expr);
  SCM bindings = TMP_CAR (cdr_expr);
  SCM um_bindings = SCM_EOL;
  SCM extended_env = env; reference_parameter (extended_env);
  SCM um_body;

  while (!scm_is_null (bindings))
    {
      const SCM variable = TMP_CAR (bindings);
      const SCM init = TMP_CADR (bindings);
      const SCM um_init = unmemoize_expression (init, extended_env);

      um_bindings = \
        rc_X (scm_cons (scm_list_2 (variable, um_init), rc_Fly (um_bindings)));

      extended_env = rc_X (SCM_EXTEND_ENV (variable, SCM_BOOL_F,
                                           rc_Fly (extended_env)));
      bindings = TMP_CDDR (bindings);
    }
  um_bindings = rc_X (scm_reverse_x (rc_Fly (um_bindings), SCM_UNDEFINED));

  um_body = unmemoize_exprs (body, extended_env);

  SCM answer = scm_cons2 (scm_sym_letstar, um_bindings, um_body);
  reference_count_return (answer);
  dereference_parameter (expr);
  dereference_parameter (env);
  dereference_parameter (extended_env);
  dereference_parameter (um_bindings);
  return (answer);
}


SCM_SYNTAX (s_or, "or", scm_i_makbimacro, scm_m_or);
SCM_GLOBAL_SYMBOL (scm_sym_or, s_or);

SCM
scm_m_or (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  const long length = scm_ilength (cdr_expr);

  ASSERT_SYNTAX (length >= 0, s_bad_expression, expr);

  if (length == 0)
    {
      /* Special case:  (or) is replaced by #f. */
      dereference_parameter (expr);
      dereference_parameter (env);
      return SCM_BOOL_F;
    }
  else
    {
      SCM_SETCAR (expr, SCM_IM_OR);
      reference_count_return (expr);
      dereference_parameter (expr);
      dereference_parameter (env);
      return expr;
    }
}

static SCM
unmemoize_or (const SCM expr, const SCM env)
{
  return scm_cons (scm_sym_or, unmemoize_exprs (SCM_CDR (expr), env));
}


SCM_SYNTAX (s_quasiquote, "quasiquote", scm_makacro, scm_m_quasiquote);
SCM_GLOBAL_SYMBOL (scm_sym_quasiquote, s_quasiquote);
SCM_GLOBAL_SYMBOL (scm_sym_unquote, "unquote");
SCM_GLOBAL_SYMBOL (scm_sym_uq_splicing, "unquote-splicing");

/* Internal function to handle a quasiquotation:  'form' is the parameter in
 * the call (quasiquotation form), 'env' is the environment where unquoted
 * expressions will be evaluated, and 'depth' is the current quasiquotation
 * nesting level and is known to be greater than zero.  */
static SCM 
iqq (SCM form, SCM env, unsigned long int depth)
{
  // debugging tick fprintf (stderr, "§");
  if (scm_is_pair (form))
    {
      const SCM tmp = TMP_CAR (form);
      if (scm_is_eq (tmp, scm_sym_quasiquote))
	{
	  const SCM args = TMP_CDR (form);
	  ASSERT_SYNTAX (scm_ilength (args) == 1, s_expression, form);
	  return scm_list_2 (tmp, iqq (TMP_CAR (args), env, depth + 1));
	}
      else if (scm_is_eq (tmp, scm_sym_unquote))
	{
	  const SCM args = TMP_CDR (form);
	  ASSERT_SYNTAX (scm_ilength (args) == 1, s_expression, form);
	  if (depth - 1 == 0)
	    return scm_eval_car (args, env);
	  else
	    return scm_list_2 (tmp, iqq (TMP_CAR (args), env, depth - 1));
	}
      else if (scm_is_pair (tmp)
	       && scm_is_eq (TMP_CAR (tmp), scm_sym_uq_splicing))
	{
	  const SCM args = TMP_CDR (tmp);
	  ASSERT_SYNTAX (scm_ilength (args) == 1, s_expression, form);
	  if (depth - 1 == 0)
	    {
	      const SCM list = scm_eval_car (args, env);
	      const SCM rest = TMP_CDR (form);
	      ASSERT_SYNTAX_2 (scm_ilength (list) >= 0,
			       s_splicing, list, form);
	      return scm_append (scm_list_2 (list, iqq (rest, env, depth)));
	    }
	  else
	    return scm_cons (iqq (TMP_CAR (form), env, depth - 1),
			     iqq (TMP_CDR (form), env, depth));
	}
      else
	return scm_cons (iqq (TMP_CAR (form), env, depth),
			 iqq (TMP_CDR (form), env, depth));
    }
  else if (scm_is_vector (form))
    return scm_vector (iqq (scm_vector_to_list (form), env, depth));
  else
    return form;
}

SCM 
scm_m_quasiquote (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) == 1, s_expression, expr);

  SCM answer = iqq (TMP_CAR (cdr_expr), env, 1);
  reference_count_return (answer);
  dereference_parameter (expr);
  dereference_parameter (env);
  return (answer);
}


SCM_SYNTAX (s_quote, "quote", scm_i_makbimacro, scm_m_quote);
SCM_GLOBAL_SYMBOL (scm_sym_quote, s_quote);

SCM
scm_m_quote (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM quotee;

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) == 1, s_expression, expr);
  quotee = TMP_CAR (cdr_expr);

  if (is_self_quoting_p (quotee)) {
    reference_count_return (quotee);
    dereference_parameter (expr);
    dereference_parameter (env);
    return quotee;
  }

  SCM_SETCAR (expr, SCM_IM_QUOTE);
  SCM_SETCDR (expr, quotee);  // RC: quotee is consumed here

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return expr;
}

static SCM
unmemoize_quote (const SCM expr, const SCM env)
{
  return scm_list_2 (scm_sym_quote, SCM_CDR (expr));
}


static const char s_set_x[] = "set!";
SCM_GLOBAL_SYMBOL (scm_sym_set_x, s_set_x); /* this is one reversed define - rename ??? */

SCM
scm_m_set_x (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM variable;
  SCM new_variable;

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) == 2, s_expression, expr);
  variable = TMP_CAR (cdr_expr); reference_parameter (variable);

  /* Memoize the variable form. */
  ASSERT_SYNTAX_2 (scm_is_symbol (variable), s_bad_variable, variable, expr);
  new_variable = lookup_symbol (variable, env);
  /* Leave the memoization of unbound symbols to lazy memoization: */
  if (SCM_UNBNDP (new_variable))
    new_variable = variable;

  SCM_SETCAR (expr, SCM_IM_SET_X);
  SCM_SETCAR (cdr_expr, new_variable);

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  dereference_parameter (variable);
  return (expr);
}

static SCM
unmemoize_set_x (const SCM expr, const SCM env)
{
  return scm_cons (scm_sym_set_x, unmemoize_exprs (SCM_CDR (expr), env));
}


/* Start of the memoizers for non-R5RS builtin macros.  */


SCM_SYNTAX (s_atapply, "@apply", scm_i_makbimacro, scm_m_apply);
SCM_GLOBAL_SYMBOL (scm_sym_atapply, s_atapply);
SCM_GLOBAL_SYMBOL (scm_sym_apply, s_atapply + 1);

SCM 
scm_m_apply (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) == 2, s_missing_expression, expr);

  SCM_SETCAR (expr, SCM_IM_APPLY);

  reference_count_return (expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return (expr);
}

static SCM
unmemoize_apply (const SCM expr, const SCM env)
{
  return scm_list_2 (scm_sym_atapply, unmemoize_exprs (SCM_CDR (expr), env));
}



SCM_SYNTAX (s_gset_x, "set!", scm_i_makbimacro, scm_m_generalized_set_x);
SCM_SYMBOL (scm_sym_setter, "setter");

SCM 
scm_m_generalized_set_x (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  SCM target, exp_target;

  const SCM cdr_expr = TMP_CDR (expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) >= 0, s_bad_expression, expr);
  ASSERT_SYNTAX (scm_ilength (cdr_expr) == 2, s_expression, expr);

  target = TMP_CAR (cdr_expr);
  if (!scm_is_pair (target))
    {
      return scm_m_set_x (rc_Fly (expr), rc_Fly (env));
    }
  else // Quoted symbols. Don't really know about that /begin/ ...
    {
      /* Macroexpanding the target might return things of the form
	 (begin <atom>).  In that case, <atom> must be a symbol or a
	 variable and we memoize to (set! <atom> ...).
      */
      exp_target = macroexp (target, env);
      if (scm_is_eq (TMP_CAR (exp_target), SCM_IM_BEGIN)
	  && !scm_is_null (TMP_CDR (exp_target))
	  && scm_is_null (TMP_CDDR (exp_target))) {
        exp_target= SCM_CADR (exp_target);
        ASSERT_SYNTAX_2 (scm_is_symbol (exp_target)
                         || SCM_VARIABLEP (exp_target),
                         s_bad_variable, exp_target, expr);
        SCM answer = \
          scm_cons (SCM_IM_SET_X, scm_cons (exp_target,
                                            SCM_CDR (cdr_expr)));
        reference_parameter (answer);
        dereference_parameter (expr);
        dereference_parameter (env);
        return (rc_return_stack_reference (answer));


      } else if (scm_is_eq (TMP_CAR (exp_target), SCM_IM_QUOTE)) {

        exp_target = SCM_CDR (exp_target); // see SCM_IM_QUOTE
        SCM new_exp = scm_cons (TMP_CAR (expr),
                                scm_cons (exp_target,
                                          SCM_CDR (cdr_expr)));
        reference_parameter (new_exp);
        dereference_parameter (expr);
        return (scm_m_set_x (rc_Fly (new_exp), rc_Fly (env)));

      } else { // Throw an error.
        ASSERT_SYNTAX_2 (0, s_bad_variable, exp_target, expr);
      }
    }
}


static SCM
unmemoize_builtin_macro (const SCM expr, const SCM env)
{
  switch (ISYMNUM (TMP_CAR (expr)))
    {
    case (ISYMNUM (SCM_IM_AND)):
      return unmemoize_and (expr, env);

    case (ISYMNUM (SCM_IM_BEGIN)):
      return unmemoize_begin (expr, env);

    case (ISYMNUM (SCM_IM_CASE)):
      return unmemoize_case (expr, env);

    case (ISYMNUM (SCM_IM_COND)):
      return unmemoize_cond (expr, env);

    case (ISYMNUM (SCM_IM_IF)):
      return unmemoize_if (expr, env);

    case (ISYMNUM (SCM_IM_LAMBDA)):
      return unmemoize_lambda (expr, env);

    case (ISYMNUM (SCM_IM_LET)):
      return unmemoize_let (expr, env);

    case (ISYMNUM (SCM_IM_LETREC)):
      return unmemoize_letrec (expr, env);

    case (ISYMNUM (SCM_IM_LETSTAR)):
      return unmemoize_letstar (expr, env);

    case (ISYMNUM (SCM_IM_OR)):
      return unmemoize_or (expr, env);

    case (ISYMNUM (SCM_IM_QUOTE)):
      return unmemoize_quote (expr, env);

    case (ISYMNUM (SCM_IM_SET_X)):
      return unmemoize_set_x (expr, env);

    case (ISYMNUM (SCM_IM_APPLY)):
      return unmemoize_apply (expr, env);

    default:
      return unmemoize_exprs (expr, env);  /* FIXME */
    }
}


/* scm_i_unmemocopy_expr and scm_i_unmemocopy_body take a memoized expression
 * respectively a memoized body together with its environment and rewrite it
 * to its original form.  Thus, these functions are the inversion of the
 * rewrite rules above.  The procedure is not optimized for speed.  It's used
 * in scm_i_unmemoize_expr, scm_procedure_source, macro_print and scm_iprin1.
 *
 * Unmemoizing is not a reliable process.  You cannot in general expect to get
 * the original source back.
 *
 */

SCM
scm_i_unmemocopy_expr (SCM expr, SCM env)
{
  reference_parameter (expr);
  reference_parameter (env);

  const SCM um_expr = unmemoize_expression (expr, env);
  reference_parameter (um_expr);

  reference_count_return (um_expr);
  dereference_parameter (um_expr);
  dereference_parameter (expr);
  dereference_parameter (env);
  return um_expr;
}

SCM
scm_i_unmemocopy_body (SCM forms, SCM env)
{
  reference_parameter (forms);
  reference_parameter (env);

  const SCM um_forms = unmemoize_exprs (forms, env);
  reference_parameter (um_forms);

  reference_count_return (um_forms);
  dereference_parameter (forms);
  dereference_parameter (env);
  return (um_forms);
}


/*****************************************************************************/
/*****************************************************************************/
/*                 The definitions for execution start here.                 */
/*****************************************************************************/
/*****************************************************************************/

SCM_GLOBAL_SYMBOL (scm_sym_enter_frame, "enter-frame");
SCM_GLOBAL_SYMBOL (scm_sym_apply_frame, "apply-frame");
SCM_GLOBAL_SYMBOL (scm_sym_exit_frame, "exit-frame");
SCM_GLOBAL_SYMBOL (scm_sym_trace, "trace");
SCM_SYMBOL (sym_instead, "instead");

/* A function object to implement "apply" for non-closure functions.  */
static SCM f_apply;
/* An endless list consisting of #<undefined> objects:  */
static SCM undefineds;


int
scm_badargsp (SCM formals, SCM args)
{
  while (!scm_is_null (formals))
    {
      if (!scm_is_pair (formals)) 
        return 0;
      if (scm_is_null (args)) 
        return 1;
      formals = TMP_CDR (formals);
      args = TMP_CDR (args);
    }
  return !scm_is_null (args) ? 1 : 0;
}



/* The evaluator contains a plethora of EVAL symbols.  This is an attempt at
 * explanation.
 *
 * The following macros should be used in code which is read twice (where the
 * choice of evaluator is hard soldered):
 *
 *   CEVAL is the symbol used within one evaluator to call itself.
 *   Originally, it is defined to ceval, but is redefined to deval during the
 *   second pass.
 *  
 *   SCM_I_EVALIM is used when it is known that the expression is an
 *   immediate.  (This macro never calls an evaluator.)
 *
 *   EVAL evaluates an expression that is expected to have its symbols already
 *   memoized.  Expressions that are not of the form '(<form> <form> ...)' are
 *   evaluated inline without calling an evaluator.
 *
 *   EVALCAR evaluates the car of an expression 'X:(Y:<form> <form> ...)',
 *   potentially replacing a symbol at the position Y:<form> by its memoized
 *   variable.  If Y:<form> is not of the form '(<form> <form> ...)', the
 *   evaluation is performed inline without calling an evaluator.
 *  
 * The following macros should be used in code which is read once
 * (where the choice of evaluator is dynamic):
 *
 *   SCM_I_XEVAL corresponds to EVAL, but uses ceval *or* deval depending on the
 *   debugging mode.
 *  
 *   SCM_I_XEVALCAR corresponds to EVALCAR, but uses ceval *or* deval depending
 *   on the debugging mode.
 *
 * The main motivation for keeping this plethora is efficiency
 * together with maintainability (=> locality of code).
 */

static SCM ceval (SCM x, SCM env);
#define CEVAL ceval


#define SCM_I_EVALIM2(x) \
  ((scm_is_eq ((x), SCM_EOL) \
    ? syntax_error (s_empty_combination, (x), SCM_UNDEFINED), 0 \
    : 0), \
   (x))

#define SCM_I_EVALIM(x, env) (SCM_ILOCP (x) \
                            ? *scm_ilookup ((x), (env)) \
			    : SCM_I_EVALIM2(x))

#define SCM_I_XEVAL(x, env) \
  (SCM_IMP (x) \
   ? SCM_I_EVALIM2 (x) \
   : (SCM_VARIABLEP (x) \
      ? TMP_VARIABLE_REF (x) \
      : (scm_is_pair (x) \
         ? (ceval ((x), (env))) \
         : (x))))

#define SCM_I_XEVALCAR(x, env) \
  (SCM_IMP (TMP_CAR (x)) \
   ? SCM_I_EVALIM (TMP_CAR (x), (env)) \
   : (SCM_VARIABLEP (TMP_CAR (x)) \
      ? TMP_VARIABLE_REF (TMP_CAR (x)) \
      : (scm_is_pair (TMP_CAR (x)) \
         ? (ceval (TMP_CAR (x), (env))) \
         : (!scm_is_symbol (TMP_CAR (x)) \
            ? TMP_CAR (x) \
            : *scm_lookupcar ((x), (env), 1)))))

#define EVAL(x, env) \
  (SCM_IMP (x) \
   ? SCM_I_EVALIM ((x), (env)) \
   : (SCM_VARIABLEP (x) \
      ? TMP_VARIABLE_REF (x) \
      : (scm_is_pair (x) \
         ? CEVAL ((x), (env)) \
         : (x))))

#define EVALCAR(x, env) \
  (SCM_IMP (TMP_CAR (x)) \
   ? SCM_I_EVALIM (TMP_CAR (x), (env)) \
   : (SCM_VARIABLEP (TMP_CAR (x)) \
      ? TMP_VARIABLE_REF (TMP_CAR (x)) \
      : (scm_is_pair (TMP_CAR (x)) \
         ? CEVAL (TMP_CAR (x), (env)) \
         : (!scm_is_symbol (TMP_CAR (x)) \
            ? TMP_CAR (x) \
            :  *scm_lookupcar ((x), (env), 1)))))


SCM
scm_eval_car (SCM pair, SCM env)
{
  reference_parameter (pair);
  reference_parameter (env);

  SCM answer = SCM_I_XEVALCAR (pair, env);

  reference_count_return (answer);
  dereference_parameter (pair);
  dereference_parameter (env);
  return (answer);
}


SCM 
scm_eval_args (SCM argument_list, SCM env, SCM proc)
{
  reference_parameter (argument_list);
  reference_parameter (env);
  reference_parameter (proc);

  SCM answer_box = scm_i_cons (SCM_BOOL_F, SCM_EOL);
  reference_parameter (answer_box);
  SCM last = answer_box;
  SCM l = argument_list;

  while (scm_is_pair (l)) {
    SCM answer = EVALCAR (l, env);
    SCM_SETCDR (last, scm_list_1 (answer));

    last = TMP_CDR (last);
    l = TMP_CDR (l);
  }

  if (!scm_is_null (l))
    scm_wrong_num_args (proc);

  answer_box = fly_cdr (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (argument_list);
  dereference_parameter (env);
  dereference_parameter (proc);
  return (answer_box);
}


SCM
scm_eval_body (SCM expression, SCM env)
{
  reference_parameter (expression);
  reference_parameter (env);

  SCM code = expression; reference_parameter (code);
  SCM next = SCM_UNSPECIFIED;

 again:
  dereference_parameter (next);
  next = TMP_CDR (code);
  reference_parameter (next);

  while (!scm_is_null (next)) {
    if (SCM_IMP     (TMP_CAR (code))) {
    if (SCM_ISYMP   (TMP_CAR (code))) {
      if (SCM_ISYMP (TMP_CAR (code))) {  // XXX
          m_expand_body (code, env);
        }
        goto again;
    }
    } else {
      rc_Void (SCM_I_XEVAL (TMP_CAR (code), env));
    }
    dereference_parameter (code);
    code = next;
    reference_parameter (code);
    dereference_parameter (next);
    next = TMP_CDR (code);
    reference_parameter (code);
  }

  SCM answer = SCM_I_XEVALCAR (code, env);
  reference_count_return (answer);
  dereference_parameter (next);
  dereference_parameter (code);
  dereference_parameter (expression);
  dereference_parameter (env);
  return answer;
}

static SCM
eval_letrec_inits (SCM env, SCM init_forms, SCM last)
{
  SCM argv[10];
  int i = 0, imax = sizeof (argv) / sizeof (SCM);

  while (!scm_is_null (init_forms))
    {
      if (imax == i)
	{
	  last = eval_letrec_inits (env, init_forms, last);
	  break;
	}
      argv[i++] = EVALCAR (init_forms, env);
      init_forms = TMP_CDR (init_forms);
    }

  for (i--; i >= 0; i--)
    {
      SCM_SETCDR (last, scm_list_1 (argv[i])); // list referenced
      last = TMP_CDR (last);
    }
  return last;
}



#define SCM_APPLY scm_apply
#define PREP_APPLY(proc, args)
#define ENTER_APPLY
// Counting Lambda Revision 2/a postambel
#define RETURN(cevals_return_value) \
{ proc = (cevals_return_value);  \
  reference_count_return (proc);    \
  dereference_parameter (x);        \
  dereference_parameter (env);      \
  return (proc); }

#ifdef STACK_CHECKING
#ifndef NO_CEVAL_STACK_CHECKING
#define EVAL_STACK_CHECKING
#endif
#endif



/* Update the toplevel environment frame ENV so that it refers to the
 * current module. Purge ??? */
#define XXX_UPDATE_TOPLEVEL_ENV(env) \
  do { \
    SCM p = scm_current_module_lookup_closure (); \
    if (p != SCM_CAR (env)) \
      env = scm_top_level_env (p); \
  } while (0)
#define UPDATE_TOPLEVEL_ENV(env)        \
{ dereference_parameter (env);          \
  env = scm_top_level_env (SCM_BOOL_F); \
  reference_parameter (env); }

#define SCM_VALIDATE_NON_EMPTY_COMBINATION(x) \
  ASSERT_SYNTAX (!scm_is_eq ((x), SCM_EOL), s_empty_combination, x)


/* This is the evaluator.  Like any real monster ...
 *
 * x is known to be a pair.  Since x is a single expression, it is necessarily
 * in a tail position.  If x is just a call to another function like in the
 * expression (foo exp1 exp2 ...), the realization of that call therefore
 * _must_not_ increase stack usage (the evaluation of exp1, exp2 etc.,
 * however, may do so).  This is realized by making extensive use of 'goto'
 * statements within the evaluator: The gotos replace recursive calls to
 * CEVAL, thus re-using the same stack frame that CEVAL was already using.
 * If, however, x represents some form that requires to evaluate a sequence of
 * expressions like (begin exp1 exp2 ...), then recursive calls to CEVAL are
 * performed for all but the last expression of that sequence.  */

static SCM
CEVAL (SCM x, SCM env)
{
  SCM proc, arg1;

  reference_parameter (x);
  reference_parameter (env);

loop:
dispatch:

  if (0) { /* Recompilation Debugging Switch */
    scm_display (SCM_BOOL_F, scm_current_error_port ());
    scm_display (SCM_BOOL_T, scm_current_error_port ());
    scm_display (SCM_BOOL_F, scm_current_error_port ());
    scm_display (x, scm_current_error_port ());
    scm_display (SCM_BOOL_F, scm_current_error_port ());
    scm_display (env, scm_current_error_port ());
    scm_display (SCM_BOOL_F, scm_current_error_port ());
    scm_display (SCM_BOOL_T, scm_current_error_port ());
    scm_display (SCM_BOOL_F, scm_current_error_port ());
    scm_newline (scm_current_error_port ());
  }

  if (SCM_ISYMP (TMP_CAR (x)))
    {
      switch (ISYMNUM (TMP_CAR (x)))
        {
        case (ISYMNUM (SCM_IM_AND)):
          x = FLY_CDR (x);
          while (!scm_is_null (TMP_CDR (x)))
            {
              SCM tmp_test_result = EVALCAR (x, env);
              if (scm_is_false (tmp_test_result)) {
                dereference_return_value (tmp_test_result);
                RETURN (SCM_BOOL_F);
              } else {
                dereference_return_value (tmp_test_result);
                x = FLY_CDR (x);
              }
            }
          PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
          goto carloop;

        case (ISYMNUM (SCM_IM_BEGIN)):
          x = FLY_CDR (x);
          if (scm_is_null (x))
            RETURN (SCM_UNSPECIFIED);

          PREP_APPLY (SCM_UNDEFINED, SCM_EOL);

        begin:
          /* If we are on toplevel with a lookup closure, we need to
             sync with the current module. XXX Well, this rather seems
             to "repair" "malformed" environments to avoid the
             "Damaged Environment" branch. */
          if (scm_is_pair (env) && !scm_is_pair (TMP_CAR (env)))
            {
              UPDATE_TOPLEVEL_ENV (env);
              while (!scm_is_null (TMP_CDR (x)))
                {
                  rc_Void (EVALCAR (x, env));
                  UPDATE_TOPLEVEL_ENV (env);
                  x = FLY_CDR (x);
                }
              goto carloop;
            }
          else
            goto nontoplevel_begin;

        nontoplevel_begin:
          while (!scm_is_null (TMP_CDR (x)))
            {
              const SCM form = TMP_CAR (x);
              if (SCM_IMP (form)) {
                if (SCM_ISYMP (form)) {
                  // if (SCM_ISYMP (TMP_CAR (x)))
                  m_expand_body (x, env);
                  goto nontoplevel_begin;
                } else {
                  SCM_VALIDATE_NON_EMPTY_COMBINATION (form);
                }

              } else {
                rc_Void (EVAL (form, env)); // form: data graph referenced
              }
              x = FLY_CDR (x);
            }

        carloop:
          {
            /* scm_eval last form in list */
            const SCM last_form = TMP_CAR (x);

            if (scm_is_pair (last_form)) {
              reference_parameter (last_form); // counts for x
              dereference_parameter (x); 
              x = last_form;
              goto loop;

            } else if (SCM_IMP (last_form)) {
              RETURN (SCM_I_EVALIM (last_form, env));

            } else if (SCM_VARIABLEP (last_form)) {
              RETURN (TMP_VARIABLE_REF (last_form));

            } else if (scm_is_symbol (last_form)) {
              RETURN (*scm_lookupcar (x, env, 1));

            } else {
              RETURN (last_form);
            }
          }


        case (ISYMNUM (SCM_IM_CASE)):
          x = FLY_CDR (x);
          {
            const SCM key = EVALCAR (x, env);
            reference_parameter (key);

            x = FLY_CDR (x);
            while (!scm_is_null (x)) {
              const SCM clause = TMP_CAR (x);

              SCM labels = TMP_CAR (clause);

              if (scm_is_eq (labels, SCM_IM_ELSE)) {
                reference_count_return (clause);
                dereference_parameter (x);
                x = SCM_CDR (clause);
                reference_parameter (x);

                dereference_parameter (key);
                PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
                goto begin;  // Exit Point
              } 

              while (!scm_is_null (labels)) {
                const SCM label = SCM_CAR (labels);
                reference_parameter (label);

                  if (scm_is_eq (label, key)
                      || scm_is_true (scm_eqv_p (label, key))) {
                    reference_count_return (clause);
                    dereference_parameter (x);
                    x = SCM_CDR (clause);
                    reference_parameter (x);
                      
                    dereference_parameter (key);
                    dereference_parameter (label);
                    PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
                    goto begin;  // Exit Point
                  }
                    
                labels = TMP_CDR (labels);
                dereference_parameter (label);
              } // closes label block

              x = FLY_CDR (x);
            } // closes clause, labels block
            dereference_parameter (key);
          }
          RETURN (SCM_UNSPECIFIED);


        case (ISYMNUM (SCM_IM_COND)):
          x = FLY_CDR (x);

          while (!scm_is_null (x)) {
            const SCM clause = TMP_CAR (x);
            reference_parameter (clause);

            // else clause (and it's not just a synonym for #t ;-)
            if (scm_is_eq (TMP_CAR (clause), SCM_IM_ELSE)) {
              dereference_parameter (x);
              x = TMP_CDR (clause);
              reference_parameter (x);
              dereference_parameter (clause);
              PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
              goto begin;  // Exit Point

            // conditional clauses
            } else {
              arg1 = EVALCAR (clause, env);
              reference_parameter (arg1);

              /* SRFI 61 extended cond !!! PURGE THIS PLEASE */
              if (!scm_is_null (TMP_CDR (clause))
                  && !scm_is_null (TMP_CDDR (clause))
                  && scm_is_eq (TMP_CADDR (clause), SCM_IM_ARROW)) {
                SCM xx, guard_result;
                /* XXX values purge */
                /* if (SCM_VALUESP (arg1)) */
                /*   arg1 = rc_X (scm_struct_ref (rc_Fly (arg1), SCM_INUM0)); */
                /* else */
                  arg1 = rc_X (scm_list_1 (rc_Fly (arg1)));
                xx = TMP_CDR (clause);
                proc = EVALCAR (xx, env);
                guard_result = rc_X (SCM_APPLY (proc, arg1, SCM_EOL));
                if (scm_is_true (guard_result)) {
                  proc = TMP_CDDR (xx);
                  proc = rc_X (EVALCAR (proc, env));
                  dereference_parameter (guard_result);
                  dereference_parameter (clause);

                  reference_parameter (proc);
                  reference_parameter (arg1);
                  PREP_APPLY (proc, arg1);
                  goto apply_proc;  // Exit Point
                }
                dereference_parameter (guard_result);

              /* regular clause with arrow */
              } else if (scm_is_true (arg1)) {
                dereference_parameter (x);
                x = TMP_CDR (clause);
                reference_parameter (x);
                if (scm_is_null (x)) {
                  dereference_parameter (clause);
                  reference_count_return (arg1);
                  dereference_parameter (arg1);
                  RETURN (arg1);
                } else if (!scm_is_eq (TMP_CAR (x), SCM_IM_ARROW)) {
                  PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
                  dereference_parameter (clause);
                  dereference_parameter (arg1);
                  goto begin;  // Exit Point
                } else {
                  dereference_parameter (clause);
                  proc = TMP_CDR (x);
                  proc = rc_X (EVALCAR (proc, env));
                  PREP_APPLY (proc, scm_list_1 (arg1));
                  ENTER_APPLY;
                  goto evap1;  // Exit Point
                }
              }  // closes regular clause with arrow
              x = FLY_CDR (x);
              dereference_parameter (arg1);
            }  // closes conditional clauses
            dereference_parameter (clause);
          }
          RETURN (SCM_UNSPECIFIED);


        case (ISYMNUM (SCM_IM_IF)):
          x = FLY_CDR (x);
          {
            SCM test_result = EVALCAR (x, env);
            reference_parameter (test_result);
            x = FLY_CDR (x);  /* then expression */
            if (scm_is_false (test_result))
              {
                x = FLY_CDR (x);  /* else expression */
                if (scm_is_null (x)) {
                  dereference_parameter (test_result);
                  RETURN (SCM_UNSPECIFIED);
                }
              }
            dereference_parameter (test_result);
          }
          PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
          goto carloop;


        case (ISYMNUM (SCM_IM_LET)):
          x = FLY_CDR (x);
          {
            SCM init_forms = TMP_CADR (x);
            SCM init_values = SCM_EOL;
            do
              {
                init_values = scm_cons (EVALCAR (init_forms, env), init_values);
                init_forms = TMP_CDR (init_forms);
              }
            while (!scm_is_null (init_forms));
            env = rc_X (SCM_EXTEND_ENV (TMP_CAR (x), init_values,
                                        rc_Fly (env)));
          }
          proc = x;
          x = TMP_CDDR (x);
          reference_parameter (x);
          dereference_parameter (proc);

          PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
          goto nontoplevel_begin;


        case (ISYMNUM (SCM_IM_LETREC)):
          x = FLY_CDR (x);
          env = rc_X (SCM_EXTEND_ENV (TMP_CAR (x), undefineds, rc_Fly (env)));
          x = FLY_CDR (x);
          {
            SCM init_forms = TMP_CAR (x);
	    SCM init_values = scm_list_1 (SCM_BOOL_T);
            reference_parameter (init_values);
	    rc_Void (eval_letrec_inits (env, init_forms, init_values));
            SCM_SETCDR (TMP_CAR (env), TMP_CDR (init_values));
            dereference_parameter (init_values);
          }
          x = FLY_CDR (x);
          PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
          goto nontoplevel_begin;


        case (ISYMNUM (SCM_IM_LETSTAR)):
          x = FLY_CDR (x);
          {
            SCM bindings = TMP_CAR (x);
            if (!scm_is_null (bindings)) {
              do {
                SCM name = TMP_CAR (bindings);
                SCM init = TMP_CDR (bindings);

                // Manual let-fly/receive because of the EVALCAR/env param
                proc = env;
                env = SCM_EXTEND_ENV (name, EVALCAR (init, env), env);
                reference_parameter (env);
                dereference_parameter (proc);

                bindings = TMP_CDR (init);
              }
              while (!scm_is_null (bindings));
            }
          }
          x = FLY_CDR (x);
          PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
          goto nontoplevel_begin;


        case (ISYMNUM (SCM_IM_OR)):
          x = FLY_CDR (x);
          while (!scm_is_null (TMP_CDR (x)))
            {
              SCM val = EVALCAR (x, env);
              if (scm_is_true (val)) {
                RETURN (val);
              } else {
                dereference_return_value (val);
                x = FLY_CDR (x);
              }
            }
          PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
          goto carloop;


        case (ISYMNUM (SCM_IM_LAMBDA)):
          RETURN (scm_closure (TMP_CADR (x), TMP_CDDR (x), env));


        case (ISYMNUM (SCM_IM_QUOTE)):
          RETURN (TMP_CDR (x));


        case (ISYMNUM (SCM_IM_SET_X)):
          x = FLY_CDR (x);
          {
            SCM *location;
            SCM tmp;
            SCM variable = TMP_CAR (x);

            if (SCM_ILOCP (variable)) {
              SCM answer = EVALCAR (TMP_CDR (x), env);
              reference (variable, answer);
              dereference_return_value (answer);

              location = scm_ilookup (variable, env);
              dereference (variable, *location);
              *location = answer;

            } else if (SCM_VARIABLEP (variable)) {
              SCM_VARIABLE_SET (variable, EVALCAR (TMP_CDR (x), env));

            } else {
              /* (scm_is_symbol (variable)) is known to be true */
              variable = lazy_memoize_variable (variable, env);
              SCM_SETCAR (x, variable);
              SCM_VARIABLE_SET (variable, EVALCAR (TMP_CDR (x), env));
            }
            x = FLY_CDR (x);
          }
          RETURN (SCM_UNSPECIFIED);


	case (ISYMNUM (SCM_IM_APPLY)):
          /* Evaluate the procedure to be applied.  */
	  x = FLY_CDR (x);
	  proc = EVALCAR (x, env);
          reference_parameter (proc);
          PREP_APPLY (proc, SCM_EOL);

          /* Evaluate the argument holding the list of arguments */
          x = FLY_CDR (x);
          arg1 = EVALCAR (x, env);
          reference_parameter (arg1);

        apply_proc:
          /* Go here to tail-apply a procedure.  PROC is the procedure and
           * ARG1 is the list of arguments. PREP_APPLY must have been called
           * before jumping to apply_proc. Both /proc/ and /arg1/ need to be
           * declared stack referenced before the goto.*/
	  if (SCM_CLOSUREP (proc))
	    {
              SCM formals = TMP_CLOSURE_FORMALS (proc);
              if (SCM_UNLIKELY (scm_badargsp (formals, arg1)))
                scm_wrong_num_args (proc);
              ENTER_APPLY;
              /* Copy argument list */
              if (scm_is_null (arg1)) {
                dereference_parameter (env);
                env = SCM_EXTEND_ENV (formals, SCM_EOL, TMP_ENV (proc));
                reference_parameter (env);

              } else {
                SCM args = scm_list_1 (TMP_CAR (arg1));
                reference_parameter (args);

                SCM tail = args;
                arg1 = FLY_CDR (arg1);
                while (!scm_is_null (arg1)) {
                  SCM new_tail = scm_list_1 (TMP_CAR (arg1));
                  SCM_SETCDR (tail, new_tail);
                  tail = new_tail;
                  arg1 = FLY_CDR (arg1);
                }
                dereference_parameter (env);
                env = SCM_EXTEND_ENV (formals, args, TMP_ENV (proc));
                reference_parameter (env);
                dereference_parameter (args);
              }
              dereference_parameter (x);
              x = TMP_CLOSURE_BODY (proc);
              reference_parameter (x);
              dereference_parameter (proc);
              dereference_parameter (arg1);
              goto nontoplevel_begin;
	    }
	  else
	    {
              ENTER_APPLY;
              RETURN (SCM_APPLY (rc_Fly (proc), rc_Fly (arg1), SCM_EOL));
	    }


	default:
	  break;
	}
    }
  else /* Do you remember the matching if clause? if (SCM_ISYMP (TMP_CAR (x))) ... */
    {
      if (SCM_VARIABLEP (TMP_CAR (x))) {
        proc = TMP_VARIABLE_REF (TMP_CAR (x));
        reference_parameter (proc);

      } else if (SCM_ILOCP (TMP_CAR (x))) {
        proc = *scm_ilookup (TMP_CAR (x), env);
        reference_parameter (proc);

      } else if (scm_is_pair (TMP_CAR (x))) {
	proc = CEVAL (TMP_CAR (x), env);
        reference_parameter (proc);

      } else if (scm_is_symbol (TMP_CAR (x))) {
        SCM orig_sym = TMP_CAR (x);
        {
          SCM *location = scm_lookupcar1 (x, env, 1);
          if (location == NULL)
            {
              /* "we have lost the race, start again."
               * I fail to see any sense in that comment if not
               * to prepare an endless loop??? */
              goto dispatch;
            }
          proc = *location;
          reference_parameter (proc);
        }

        if (SCM_MACROP (proc)) {
          SCM_SETCAR (x, orig_sym);  /* Undo memoizing effect of
                                        lookupcar */

        handle_a_macro: /* inputs: x, env, proc */
          arg1 = SCM_APPLY (SCM_MACRO_CODE (proc), x,
                            scm_cons (env, scm_listofnull));
          reference_parameter (arg1);
          switch (SCM_MACRO_TYPE (proc))
            {
            case 3:
            case 2:
              if (!scm_is_pair (arg1))
                arg1 = rc_X (scm_list_2 (SCM_IM_BEGIN, rc_Fly (arg1)));

              assert (!scm_is_eq (x, TMP_CAR (arg1))
                      && !scm_is_eq (x, TMP_CDR (arg1)));

              SCM_SETCAR (x, SCM_CAR (arg1));
              SCM_SETCDR (x, SCM_CDR (arg1));
              dereference_parameter (arg1); arg1 = SCM_UNSPECIFIED;
              dereference_parameter (proc); proc = SCM_UNSPECIFIED;
              PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
              goto loop;

            case 1:
              dereference_parameter (x);
              x = arg1;
              reference_parameter (x);
              if (SCM_NIMP (x)) {
                dereference_parameter (proc);
                dereference_parameter (arg1);
                PREP_APPLY (SCM_UNDEFINED, SCM_EOL);
                goto loop;
              } else {
                reference_count_return (arg1);
                dereference_parameter (proc);
                dereference_parameter (arg1);
                RETURN (arg1);
              }

            case 0:
              reference_count_return (arg1);
              dereference_parameter (proc);
              dereference_parameter (arg1);
              RETURN (arg1);
            }
          } // closes macro block

      } else {   // this matches whatever this is: proc is referenced
        proc = TMP_CAR (x);
        reference_parameter (proc);
      }

      if (SCM_MACROP (proc))
	goto handle_a_macro;

    } // closes pre-apply dispatch block

  /* When reaching this part of the code, the following is granted: Variable x
   * holds the first pair of an expression of the form (<function> arg ...).
   * Variable proc holds the object that resulted from the evaluation of
   * <function>.  In the following, the arguments (if any) will be evaluated,
   * and proc will be applied to them.  If proc does not really hold a
   * function object, this will be signalled as an error on the scheme
   * level.  If the number of arguments does not match the number of arguments
   * that are allowed to be passed to proc, also an error on the scheme level
   * will be signalled.  */
  PREP_APPLY (proc, SCM_EOL);
  if (scm_is_null (TMP_CDR (x))) {
    ENTER_APPLY;
  evap0:
    SCM_ASRTGO (!SCM_IMP (proc), badfun);
    switch (SCM_TYP3 (proc)) {
      case scm_tc3_closure:
        {
          // data graph referenced by proc (equals closure here)
          const SCM formals = TMP_CLOSURE_FORMALS (proc);
          if (SCM_UNLIKELY (scm_is_pair (formals)))
            goto wrongnumargs;

          dereference_parameter (x);
          dereference_parameter (env);
          x = TMP_CLOSURE_BODY (proc);
          env = SCM_EXTEND_ENV (formals, SCM_EOL, TMP_ENV (proc));
          reference_parameter (x);
          reference_parameter (env);
          dereference_parameter (proc);
          goto nontoplevel_begin;  // Exit Point
        }

      case scm_tc3_tc7_1:
      case scm_tc3_tc7_2:
    switch (SCM_TYP7 (proc))
      {				/* no arguments given */
        SCM answer;
      case scm_tc7_subr_0:
	answer = SCM_SUBRF (proc) ();
        dereference_parameter (proc);
	RETURN (answer);
      case scm_tc7_subr_1o:
	answer = SCM_SUBRF (proc) (SCM_UNDEFINED);
        dereference_parameter (proc);
	RETURN (answer);
      case scm_tc7_lsubr:
	answer = SCM_SUBRF (proc) (SCM_EOL);
        dereference_parameter (proc);
	RETURN (answer);
      case scm_tc7_rpsubr:
        dereference_parameter (proc);
	RETURN (SCM_BOOL_T);
      case scm_tc7_asubr:
	answer = SCM_SUBRF (proc) (SCM_UNDEFINED, SCM_UNDEFINED);
        dereference_parameter (proc);
	RETURN (answer);
      case scm_tc7_subr_n: // XXX looks like a diversion from apply
        answer = scm_i_gsubr_apply (proc, SCM_EOL);
        dereference_parameter (proc);
        RETURN (answer);
      case scm_tc7_subr_1:
      case scm_tc7_subr_2:
      case scm_tc7_subr_2o:
      case scm_tc7_dsubr:
      case scm_tc7_cxr:
      case scm_tc7_subr_3:
      case scm_tc7_lsubr_2:
      wrongnumargs:
	scm_wrong_num_args (proc);
      }
    default:
    badfun:
      scm_misc_error (NULL, "Wrong type to apply: ~S", scm_list_1 (proc));
    }
  }

  /* must handle macros by here <- ??? wha'? */
  x = FLY_CDR (x);
  if (SCM_LIKELY (scm_is_pair (x)))
    arg1 = EVALCAR (x, env);
  else
    scm_wrong_num_args (proc);
  reference_parameter (arg1);
  x = FLY_CDR (x);

  { // proc and arg1 should be set and referenced by now
    SCM arg2 = SCM_UNSPECIFIED;
    if (scm_is_null (x)) {
      ENTER_APPLY;
    evap1: /* inputs: proc, arg1 */
      SCM_ASRTGO (!SCM_IMP (proc), badfun);

      switch (SCM_TYP3 (proc)) {
        case scm_tc3_closure:
        {
          const SCM formals = TMP_CLOSURE_FORMALS (proc); // data graph referenced
          if (SCM_UNLIKELY (scm_is_null (formals)
                            || (scm_is_pair (formals) &&
                                scm_is_pair (TMP_CDR (formals)))))
            goto wrongnumargs;
          dereference_parameter (x);
          dereference_parameter (env);
          x = TMP_CLOSURE_BODY (proc);
          env = SCM_EXTEND_ENV (formals,
                                scm_list_1 (arg1),
                                TMP_ENV (proc));
          reference_parameter (x);
          reference_parameter (env);
          dereference_parameter (proc);
          dereference_parameter (arg1);
          goto nontoplevel_begin;
        }

        case scm_tc3_tc7_1:
        case scm_tc3_tc7_2:
      switch (SCM_TYP7 (proc)) { /* have one argument in arg1 */
        SCM answer;
      case scm_tc7_subr_2o:
        answer = SCM_SUBRF (proc) (arg1, SCM_UNDEFINED);
        dereference_parameter (proc);
        dereference_parameter (arg1);
        RETURN (answer);
      case scm_tc7_subr_1:
      case scm_tc7_subr_1o:
        answer = SCM_SUBRF (proc) (arg1);
        dereference_parameter (proc);
        dereference_parameter (arg1);
        RETURN (answer);
      case scm_tc7_dsubr:
        if (SCM_I_INUMP (arg1))
          {
            answer = scm_from_double (SCM_DSUBRF (proc) ((double) SCM_I_INUM (arg1)));
            dereference_parameter (proc);
            dereference_parameter (arg1);  // bogus
            RETURN (answer);
          }
        else if (SCM_REALP (arg1))
          {
            answer = scm_from_double (SCM_DSUBRF (proc) (SCM_REAL_VALUE (arg1)));
            dereference_parameter (proc);
            dereference_parameter (arg1);
            RETURN (answer);
          }
        else if (SCM_BIGP (arg1))
          {
            answer = scm_from_double (SCM_DSUBRF (proc) (scm_i_big2dbl (arg1)));
            dereference_parameter (proc);
            dereference_parameter (arg1);
            RETURN (answer);
          }
        else if (SCM_FRACTIONP (arg1))
          {
            answer = scm_from_double (SCM_DSUBRF (proc) (scm_i_fraction2double (arg1)));
            dereference_parameter (proc);
            dereference_parameter (arg1);
            RETURN (answer);
          }
        SCM_WTA_DISPATCH_1 (*SCM_SUBR_GENERIC (proc), arg1,
                            SCM_ARG1,
                            BYTE_STRING_DATA (scm_to_sen (scm_symbol_to_brass (SCM_SNAME (proc))))); // CLR3 policy
      case scm_tc7_cxr:
        answer = scm_i_chase_pairs (arg1, (scm_t_bits) SCM_SUBRF (proc));
        dereference_parameter (proc);
        dereference_parameter (arg1);
        RETURN (answer);
      case scm_tc7_rpsubr:
        dereference_parameter (proc);
        dereference_parameter (arg1);
        RETURN (SCM_BOOL_T);
      case scm_tc7_asubr:
        answer = SCM_SUBRF (proc) (arg1, SCM_UNDEFINED);
        dereference_parameter (proc);
        dereference_parameter (arg1);
        RETURN (answer);
      case scm_tc7_lsubr:
        answer = SCM_SUBRF (proc) (scm_list_1 (arg1));
        dereference_parameter (proc);
        dereference_parameter (arg1);
        RETURN (answer);
      case scm_tc7_subr_n:
        answer = scm_i_gsubr_apply (proc, scm_list_1 (arg1));
        dereference_parameter (proc);
        dereference_parameter (arg1);
        RETURN (answer);
      case scm_tc7_subr_2:
      case scm_tc7_subr_0:
      case scm_tc7_subr_3:
      case scm_tc7_lsubr_2:
        scm_wrong_num_args (proc);
      }
      default:
        goto badfun;
      }
    }
    if (SCM_LIKELY (scm_is_pair (x)))
      arg2 = EVALCAR (x, env);
    else
      scm_wrong_num_args (proc);

    reference_parameter (arg2);

    {				/* have two or more arguments */
      x = FLY_CDR (x);
      if (scm_is_null (x)) {
	ENTER_APPLY;
      evap2:
        SCM_ASRTGO (!SCM_IMP (proc), badfun);

        switch (SCM_TYP3 (proc)) {
        case scm_tc3_closure:
          {
            const SCM formals = TMP_CLOSURE_FORMALS (proc);
            if (SCM_UNLIKELY
                (scm_is_null (formals)
                 || (scm_is_pair (formals)
                     && (scm_is_null (TMP_CDR (formals))
                         || (scm_is_pair (TMP_CDR (formals))
                             && scm_is_pair (TMP_CDDR (formals)))))))
              goto wrongnumargs;
            dereference_parameter (env);
            env = SCM_EXTEND_ENV (formals,
                                  scm_list_2 (arg1, arg2),
                                  TMP_ENV (proc));
            reference_parameter (env);
            dereference_parameter (x);
            x = TMP_CLOSURE_BODY (proc);
            reference_parameter (x);

            dereference_parameter (proc);
            dereference_parameter (arg1);
            dereference_parameter (arg2);
            goto nontoplevel_begin;
          }

        case scm_tc3_tc7_1:
        case scm_tc3_tc7_2:
	switch (SCM_TYP7 (proc)) {  /* have two arguments */
          SCM answer;
        case scm_tc7_subr_2:
        case scm_tc7_subr_2o:
          answer = SCM_SUBRF (proc) (arg1, arg2);
          dereference_parameter (arg2);
          dereference_parameter (arg1);
          dereference_parameter (proc);
          RETURN (answer);
        case scm_tc7_lsubr:
          answer = SCM_SUBRF (proc) (scm_list_2 (arg1, arg2));
          dereference_parameter (arg2);
          dereference_parameter (arg1);
          dereference_parameter (proc);
          RETURN (answer);
        case scm_tc7_lsubr_2:
          answer = SCM_SUBRF (proc) (arg1, arg2, SCM_EOL);
          dereference_parameter (arg2);
          dereference_parameter (arg1);
          dereference_parameter (proc);
          RETURN (answer);
        case scm_tc7_rpsubr:
        case scm_tc7_asubr:
          answer = SCM_SUBRF (proc) (arg1, arg2);
          dereference_parameter (arg2);
          dereference_parameter (arg1);
          dereference_parameter (proc);
          RETURN (answer);
        case scm_tc7_subr_n:
          answer = scm_i_gsubr_apply (proc, scm_list_2 (arg1, arg2));
          dereference_parameter (proc);
          dereference_parameter (arg1);
          dereference_parameter (arg2);
          RETURN (answer);
        case scm_tc7_subr_0:
        case scm_tc7_dsubr:
        case scm_tc7_cxr:
        case scm_tc7_subr_1o:
        case scm_tc7_subr_1:
        case scm_tc7_subr_3:
          scm_wrong_num_args (proc);
        }
        default:
          goto badfun;
        }
      }
      if (SCM_UNLIKELY (!scm_is_pair (x)))
	scm_wrong_num_args (proc);
      ENTER_APPLY;
    evap3:
      SCM_ASRTGO (!SCM_IMP (proc), badfun);

      switch (SCM_TYP3 (proc)) {
      case scm_tc3_closure:
	  {
	    const SCM formals = TMP_CLOSURE_FORMALS (proc);
	    if (scm_is_null (formals)
		|| (scm_is_pair (formals)
		    && (scm_is_null (TMP_CDR (formals))
			|| (scm_is_pair (TMP_CDR (formals))
			    && scm_badargsp (TMP_CDDR (formals), x)))))
	      goto wrongnumargs;

            env = rc_X (SCM_EXTEND_ENV (formals,
                                  scm_cons2 (arg1,
                                             arg2,
                                             scm_eval_args (x, rc_Fly (env),
                                                            proc)),
                                  TMP_ENV (proc)));
            
            dereference_parameter (x);
            x = TMP_CLOSURE_BODY (proc);
            reference_parameter (x);

            dereference_parameter (proc);
            dereference_parameter (arg1);
            dereference_parameter (arg2);
            goto nontoplevel_begin;
	  }
          
      case scm_tc3_tc7_1:
      case scm_tc3_tc7_2:
      switch (SCM_TYP7 (proc))
	{			/* have 3 or more arguments */
          SCM answer;
	case scm_tc7_subr_3:
	  if (SCM_UNLIKELY (!scm_is_null (TMP_CDR (x)))) {
	    scm_wrong_num_args (proc);
	  } else {
	    answer = SCM_SUBRF (proc) (arg1, arg2, EVALCAR (x, env));
            dereference_parameter (proc);
            dereference_parameter (arg1);
            dereference_parameter (arg2);
            RETURN (answer);
          }
	case scm_tc7_asubr:
	  arg1 = rc_X (SCM_SUBRF (proc) (rc_Fly (arg1), arg2));
	  do
	    {
	      arg1 = rc_X (SCM_SUBRF(proc)(rc_Fly (arg1), EVALCAR(x, env)));
	      x = FLY_CDR(x);
	    }
	  while (!scm_is_null (x));
          rc_return_stack_reference (arg1);
          dereference_parameter (proc);
          dereference_parameter (arg2);
	  RETURN (arg1);

	case scm_tc7_rpsubr:
	  if (scm_is_false (SCM_SUBRF (proc) (arg1, arg2))) {
            dereference_parameter (proc);
            dereference_parameter (arg1);
            dereference_parameter (arg2);
	    RETURN (SCM_BOOL_F);
          }
	  do {
            dereference_parameter (arg1);
            arg1 = EVALCAR (x, env);
            reference_parameter (arg1);
            answer = SCM_SUBRF (proc) (arg2, arg1);
            if (scm_is_false (answer)) {
              dereference_return_value (answer);
              dereference_parameter (proc);
              dereference_parameter (arg1);
              dereference_parameter (arg2);
              RETURN (SCM_BOOL_F);
            }
            dereference_parameter (arg2);
            arg2 = arg1;
            reference_parameter (arg2);
            x = FLY_CDR (x);
          } while (!scm_is_null (x));

          dereference_parameter (proc);
          dereference_parameter (arg1);
          dereference_parameter (arg2);
	  RETURN (SCM_BOOL_T);

	case scm_tc7_lsubr_2:
	  answer = SCM_SUBRF (proc) (arg1, arg2, scm_eval_args (x, env, proc));
          dereference_parameter (proc);
          dereference_parameter (arg1);
          dereference_parameter (arg2);
          RETURN (answer);

	case scm_tc7_lsubr:
	  answer = SCM_SUBRF (proc) (scm_cons2 (arg1,
                                                arg2,
                                                scm_eval_args (x, env, proc)));
          dereference_parameter (proc);
          dereference_parameter (arg1);
          dereference_parameter (arg2);
          RETURN (answer);

	case scm_tc7_subr_n:
          answer = \
            scm_i_gsubr_apply (proc, scm_cons2 (arg1, arg2, \
                                                scm_eval_args (x, env, proc)));
          dereference_parameter (proc);
          dereference_parameter (arg1);
          dereference_parameter (arg2);
          RETURN (answer);

	case scm_tc7_subr_0:
	case scm_tc7_subr_2:
	case scm_tc7_subr_1o:
	case scm_tc7_subr_2o:
	case scm_tc7_dsubr:
	case scm_tc7_cxr:
	case scm_tc7_subr_1:
	  scm_wrong_num_args (proc);
	}
	default:
	  goto badfun;
      }
    }
  }
}


/* Simple procedure calls
 */

SCM
scm_call_0 (SCM proc)
{
  return scm_apply (proc, SCM_EOL, SCM_EOL);
}

SCM
scm_call_1 (SCM proc, SCM arg1)
{
  return scm_apply (proc, arg1, scm_listofnull);
}

SCM
scm_call_2 (SCM proc, SCM arg1, SCM arg2)
{
  return scm_apply (proc, arg1, scm_cons (arg2, scm_listofnull));
}

SCM
scm_call_3 (SCM proc, SCM arg1, SCM arg2, SCM arg3)
{
  return scm_apply (proc, arg1, scm_cons2 (arg2, arg3, scm_listofnull));
}

SCM
scm_call_4 (SCM proc, SCM arg1, SCM arg2, SCM arg3, SCM arg4)
{
  return scm_apply (proc, arg1, scm_cons2 (arg2, arg3,
					   scm_cons (arg4, scm_listofnull)));
}

/* Simple procedure applies
 */

SCM
scm_apply_0 (SCM proc, SCM args)
{
  return scm_apply (proc, args, SCM_EOL);
}

SCM
scm_apply_1 (SCM proc, SCM arg1, SCM args)
{
  return scm_apply (proc, scm_cons (arg1, args), SCM_EOL);
}

SCM
scm_apply_2 (SCM proc, SCM arg1, SCM arg2, SCM args)
{
  return scm_apply (proc, scm_cons2 (arg1, arg2, args), SCM_EOL);
}

SCM
scm_apply_3 (SCM proc, SCM arg1, SCM arg2, SCM arg3, SCM args)
{
  return scm_apply (proc, scm_cons (arg1, scm_cons2 (arg2, arg3, args)),
		    SCM_EOL);
}

/* This code processes the arguments to apply:

   (apply PROC ARG1 ... ARGS)

   Given a list (ARG1 ... ARGS), this function conses the ARG1
   ... arguments onto the front of ARGS, and returns the resulting
   list.  Note that ARGS is a list; thus, the argument to this
   function is a list whose last element is a list.

   Apply calls this function, and applies PROC to the elements of the
   result.  apply:nconc2last takes care of building the list of
   arguments, given (ARG1 ... ARGS).

   Rather than do new consing, apply:nconc2last destroys its argument.
   On that topic, this code came into my care with the following
   beautifully cryptic comment on that topic: "This will only screw
   you if you do (scm_apply scm_apply '( ... ))"  If you know what
   they're referring to, send me a patch to this comment.

   See below for some try on that patch. */
SCM_DEFINE (scm_nconc2last, "apply:nconc2last", 1, 0, 0,
	    (SCM lst),
	    "Given a list (@var{arg1} @dots{} @var{args}), this function\n"
	    "conses the @var{arg1} @dots{} arguments onto the front of\n"
	    "@var{args}, and returns the resulting list. Note that\n"
	    "@var{args} is a list; thus, the argument to this function is\n"
	    "a list whose last element is a list.\n"
	    "Note: Rather than do new consing, @code{apply:nconc2last}\n"
	    "destroys its argument, so use with care.")
#define FUNC_NAME s_scm_nconc2last
{
  reference_parameter (lst);

  SCM last;
  SCM_VALIDATE_NONEMPTYLIST (1, lst);
  SCM answer_box = scm_i_cons (SCM_BOOL_F, lst);
  reference_parameter (answer_box);
  last = lst;
  SCM next_to_last = answer_box;

  // find the last argument: a list
  while (!scm_is_null (TMP_CDR (last))) {
    next_to_last = last;
    last = TMP_CDR (last);
  }

  /* BENCH This call to scm_ilength degrades performance. Perhaps there
   * is some possibility to move this check elsewhere. This function is
   * the top caller of scm_ilength. */
  SCM_ASSERT (scm_ilength (TMP_CAR (last)) >= 0, lst, SCM_ARG1, FUNC_NAME);

  // "bend" the last list argument to be our tail cutting one pair
  SCM_SETCDR (next_to_last, SCM_CAR (last));

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (lst);
  return answer_box;
}
#undef FUNC_NAME


/*
 * Redefine RETURN for apply and dapply
 */

#undef RETURN
// Counting Lambda Revision 2/a
#define RETURN(applies_return_value)       \
{ SCM tmp_answer = (applies_return_value); \
  reference_count_return (tmp_answer);     \
  dereference_parameter (proc);            \
  dereference_parameter (arg1);            \
  dereference_parameter (args);            \
  return (tmp_answer); }


/* Apply a function to a list of arguments.

   The applicator [quick find keyword]

   This function is exported to the Scheme level as taking two
   required arguments and a tail argument, as if it were:
	(proc arg1 . args)

   where the last argument needs to be a list - I guess.
*/

SCM 
SCM_APPLY (SCM proc, SCM arg1, SCM args)
{
  reference_parameter (proc);
  reference_parameter (arg1);
  reference_parameter (args);

  SCM_ASRTGO (SCM_NIMP (proc), badproc);

  if (scm_is_null (args)) {
    if (scm_is_null (arg1)) {
      arg1 = SCM_UNDEFINED;
      
    } else {
      dereference_parameter (args);
      args = TMP_CDR (arg1);
      reference_parameter (args);
      arg1 = FLY_CAR (arg1);

    }

  } else {
    args = rc_X (scm_nconc2last (rc_Fly (args)));
  }
tail:
  switch (SCM_TYP3 (proc)) {
  case scm_tc3_closure: {
      {
      SCM tmp = SCM_UNBNDP (arg1) ? SCM_EOL : scm_cons (arg1, args);
      reference_parameter (tmp);  // RC exchange: counts for arg1
      dereference_parameter (arg1);
      arg1 = tmp; tmp = SCM_UNSPECIFIED;
      }
      {
      SCM tmp = SCM_UNSPECIFIED;
      if (SCM_UNLIKELY (scm_badargsp (TMP_CLOSURE_FORMALS (proc), arg1)))
	scm_wrong_num_args (proc);
      
      /* Copy argument list */
      if (SCM_IMP (arg1)) {
        dereference_parameter (args);
	args = arg1;
        reference_parameter (args);

      } else {
        dereference_parameter (args);  // precious cdr minus one = 2.0
        args = scm_cons (TMP_CAR (arg1), SCM_UNSPECIFIED);
        reference_parameter (args);
        SCM tl   = args;  // list referenced
        SCM tm   = arg1;  // list referenced
        for (tm = TMP_CDR (tm); scm_is_pair (tm); tm = TMP_CDR (tm))
          {
            SCM_SETCDR (tl, scm_cons (TMP_CAR (tm), SCM_UNSPECIFIED));
            tl = TMP_CDR (tl);
          }
        SCM_SETCDR (tl, tm);
        tl = SCM_UNSPECIFIED;
        tm = SCM_UNSPECIFIED;
      }
      // dereference_parameter (arg1); counting armor in RETURN does this
      
      /* Prepare the evaluation of the closure. */
      args = rc_X (SCM_EXTEND_ENV (TMP_CLOSURE_FORMALS (proc),
                                   rc_Fly (args),
                                   TMP_ENV (proc)));
      proc = rc_X (SCM_CLOSURE_BODY (rc_Fly (proc)));

      /* Evaluation of the closure body. */
      SCM next = SCM_UNSPECIFIED;
    again:
      next = TMP_CDR (proc);
      reference_parameter (next);  // RC exchange: counts for arg1

      while (!scm_is_null (next)) {
        if (SCM_IMP (TMP_CAR (proc))) {
          if (SCM_ISYMP (TMP_CAR (proc))) {
            /* Hey, I have seen this before! */
            /* We start to memoize the body when we have
               seen a memoized immediate symbol for a primary
               primitive? That's lazy memoization? */
            if (SCM_ISYMP (TMP_CAR (proc)))
              m_expand_body (proc, args);
            goto again;
          } else {
            SCM_VALIDATE_NON_EMPTY_COMBINATION (TMP_CAR (proc));
          }

        } else {
          rc_Void (EVAL (TMP_CAR (proc), args));
        }

        tmp  = next;
        reference_parameter (tmp);  // RC Exchange: counts for proc
        dereference_parameter (proc);
        proc = tmp;
        next = FLY_CDR (next);
      }
      } // closes apply/eval closure block
      /* Return the evaluation of the last form. */
      // !!! Make this a (pass through) tail call: use rc_Fly and /return/
      RETURN (EVALCAR (proc, args));
    }

  case scm_tc3_tc7_1:
  case scm_tc3_tc7_2:
  switch (SCM_TYP7 (proc))
    {
      SCM answer = SCM_UNSPECIFIED;
    case scm_tc7_subr_2o:
      if (SCM_UNLIKELY (SCM_UNBNDP (arg1)))
	scm_wrong_num_args (proc);
      if (scm_is_null (args))
        args = SCM_UNDEFINED;  // RC: no IM references
      else
        {
          if (SCM_UNLIKELY (!scm_is_null (TMP_CDR (args))))
            scm_wrong_num_args (proc);
          args = FLY_CAR (args);
        }
      RETURN (SCM_SUBRF (proc) (arg1, args));
    case scm_tc7_subr_2:
      if (SCM_UNLIKELY (scm_is_null (args) || !scm_is_null (TMP_CDR (args))))
	scm_wrong_num_args (proc);
      args = FLY_CAR (args);
      RETURN (SCM_SUBRF (proc) (arg1, args));
    case scm_tc7_subr_0:
      if (SCM_UNLIKELY (!SCM_UNBNDP (arg1))) {
	scm_wrong_num_args (proc);
      } else {
	RETURN (SCM_SUBRF (proc) ());
      }
    case scm_tc7_subr_1:
      if (SCM_UNLIKELY (SCM_UNBNDP (arg1)))
	scm_wrong_num_args (proc);
    case scm_tc7_subr_1o:
      if (SCM_UNLIKELY (!scm_is_null (args))) {
	scm_wrong_num_args (proc);
      } else {
	RETURN (SCM_SUBRF (proc) (arg1));
      }
    case scm_tc7_dsubr:
      if (SCM_UNLIKELY (SCM_UNBNDP (arg1) || !scm_is_null (args)))
	scm_wrong_num_args (proc);
      if (SCM_I_INUMP (arg1))
        {
          RETURN (scm_from_double (SCM_DSUBRF (proc) ((double) SCM_I_INUM (arg1))));
        }
      else if (SCM_REALP (arg1))
        {
          RETURN (scm_from_double (SCM_DSUBRF (proc) (SCM_REAL_VALUE (arg1))));
        }
      else if (SCM_BIGP (arg1))
	{
	  RETURN (scm_from_double (SCM_DSUBRF (proc) (scm_i_big2dbl (arg1))));
	}
      else if (SCM_FRACTIONP (arg1))
	{
	  RETURN (scm_from_double (SCM_DSUBRF (proc) (scm_i_fraction2double (arg1))));
	}
      SCM_WTA_DISPATCH_1 (*SCM_SUBR_GENERIC (proc), arg1,
                          SCM_ARG1, BYTE_STRING_DATA (scm_to_sen (scm_symbol_to_brass (SCM_SNAME (proc))))); // CLR3 policy
    case scm_tc7_cxr:
      if (SCM_UNLIKELY (SCM_UNBNDP (arg1) || !scm_is_null (args)))
	scm_wrong_num_args (proc);
      RETURN (scm_i_chase_pairs (arg1, (scm_t_bits) SCM_SUBRF (proc)));
    case scm_tc7_subr_3:
      if (SCM_UNLIKELY (scm_is_null (args)
			|| scm_is_null (TMP_CDR (args))
			|| !scm_is_null (TMP_CDDR (args))))
        {
          scm_wrong_num_args (proc);
        }
      else
        {
          RETURN (SCM_SUBRF (proc) (arg1, TMP_CAR (args), TMP_CADR (args)));
        }
    case scm_tc7_lsubr:
      RETURN (SCM_SUBRF (proc) (SCM_UNBNDP (arg1) ? SCM_EOL : scm_cons (arg1, args)));
    case scm_tc7_lsubr_2:
      if (SCM_UNLIKELY (!scm_is_pair (args))) {
	scm_wrong_num_args (proc);
      } else {
	RETURN (SCM_SUBRF (proc) (arg1, TMP_CAR (args), TMP_CDR (args)));
      }
    case scm_tc7_asubr:
      if (scm_is_null (args)) {
	RETURN (SCM_SUBRF (proc) (arg1, SCM_UNDEFINED));
      }
      while (SCM_NIMP (args))
	{
	  SCM_ASSERT (scm_is_pair (args), args, SCM_ARG2, "apply");
	  arg1 = rc_X (SCM_SUBRF (proc) (rc_Fly (arg1), TMP_CAR (args)));
	  args = FLY_CDR (args);
	}
      RETURN (arg1);
    case scm_tc7_rpsubr:
      if (scm_is_null (args)) {
	RETURN (SCM_BOOL_T);
      }
      while (SCM_NIMP (args))
	{
	  SCM_ASSERT (scm_is_pair (args), args, SCM_ARG2, "apply");
          SCM tmp = SCM_SUBRF (proc) (arg1, SCM_CAR (args));
	  if (scm_is_false (tmp)) {
            dereference_return_value (tmp);
	    RETURN (SCM_BOOL_F);
          }
          dereference_return_value (tmp);
          dereference_parameter (arg1);
	  arg1 = TMP_CAR (args);
          reference_parameter (arg1);
	  args = FLY_CDR (args);
	}
      RETURN (SCM_BOOL_T);

    case scm_tc7_subr_n:
      if (!SCM_UNBNDP (arg1)) {
        RETURN (scm_i_gsubr_apply (proc, scm_cons (arg1, args)));
      } else {
        goto badproc;  // ??? Shouldn't we apply? Check this, please.
      }
    }
    default:
    badproc:
      scm_wrong_type_arg ("apply", SCM_ARG1, proc);
  }
}



/* Trampolines
 *  
 * Trampolines make it possible to move procedure application dispatch
 * outside inner loops.  The motivation was clean implementation of
 * efficient replacements of R5RS primitives in SRFI-1.
 *
 * The semantics is clear: scm_trampoline_N returns an optimized
 * version of scm_call_N (or NULL if the procedure isn't applicable
 * on N args).
 *
 * Applying the optimization to map and for-each increased efficiency
 * noticeably.  For example, (map abs ls) is now 8 times faster than
 * before.
 */

static SCM
call_subr0_0 (SCM proc)
{
  return SCM_SUBRF (proc) ();
}

static SCM
call_subr1o_0 (SCM proc)
{
  return SCM_SUBRF (proc) (SCM_UNDEFINED);
}

static SCM
call_lsubr_0 (SCM proc)
{
  return SCM_SUBRF (proc) (SCM_EOL);
}

SCM 
scm_i_call_closure_0 (SCM proc)
{
  const SCM env = SCM_EXTEND_ENV (TMP_CLOSURE_FORMALS (proc),
                                  SCM_EOL,
                                  TMP_ENV (proc));
  const SCM result = scm_eval_body (TMP_CLOSURE_BODY (proc), env);
  return result;
}

static SCM
call_subr_n_0 (SCM proc)
{
  return scm_i_gsubr_apply (proc, SCM_EOL);
}

scm_t_trampoline_0
scm_trampoline_0 (SCM proc)
{
  reference_parameter (proc);
  scm_t_trampoline_0 trampoline;

  if (SCM_IMP (proc)) {
    dereference_parameter (proc);
    return NULL;
  }

  switch (SCM_TYP3 (proc)) {
    case scm_tc3_closure:
      {
	SCM formals = TMP_CLOSURE_FORMALS (proc);
	if (scm_is_null (formals) || !scm_is_pair (formals)) {
	  trampoline = scm_i_call_closure_0;
	} else {
          dereference_parameter (proc);
	  return NULL;
        }
        goto success;
      }

    case scm_tc3_tc7_1:
    case scm_tc3_tc7_2:
  switch (SCM_TYP7 (proc))
    {
    case scm_tc7_subr_0:
      trampoline = call_subr0_0;
      goto success;
    case scm_tc7_subr_1o:
      trampoline = call_subr1o_0;
      goto success;
    case scm_tc7_lsubr:
      trampoline = call_lsubr_0;
      goto success;
    case scm_tc7_subr_n: // XXX looks like ceval but apply ...
      trampoline = call_subr_n_0;
      goto success;
    case scm_tc7_asubr:
    case scm_tc7_rpsubr:
      trampoline = scm_call_0;
      goto success;
    }
  default:
    dereference_parameter (proc);
    return NULL; /* not applicable on zero arguments */
  }

  success:
  /* We only reach this point if a valid trampoline was determined.  */
  dereference_parameter (proc);

  return trampoline;
}

static SCM
call_subr1_1 (SCM proc, SCM arg1)
{
  return SCM_SUBRF (proc) (arg1);
}

static SCM
call_subr2o_1 (SCM proc, SCM arg1)
{
  return SCM_SUBRF (proc) (arg1, SCM_UNDEFINED);
}

static SCM
call_lsubr_1 (SCM proc, SCM arg1)
{
  return SCM_SUBRF (proc) (scm_list_1 (arg1));
}

static SCM
call_dsubr_1 (SCM proc, SCM arg1)
{
  reference_parameter (proc);
  reference_parameter (arg1);

  SCM args = SCM_UNDEFINED; // dummy for RETURN

  if (SCM_I_INUMP (arg1))
    {
      RETURN (scm_from_double (SCM_DSUBRF (proc) ((double) SCM_I_INUM (arg1))));
    }
  else if (SCM_REALP (arg1))
    {
      RETURN (scm_from_double (SCM_DSUBRF (proc) (SCM_REAL_VALUE (arg1))));
    }
  else if (SCM_BIGP (arg1))
    {
      RETURN (scm_from_double (SCM_DSUBRF (proc) (scm_i_big2dbl (arg1))));
    }
  else if (SCM_FRACTIONP (arg1))
    {
      RETURN (scm_from_double (SCM_DSUBRF (proc) (scm_i_fraction2double (arg1))));
    }
  SCM_WTA_DISPATCH_1 (*SCM_SUBR_GENERIC (proc), arg1,
		      SCM_ARG1, BYTE_STRING_DATA (scm_to_sen (scm_symbol_to_brass (SCM_SNAME (proc)))));  // CLR3 policy
}

static SCM
call_cxr_1 (SCM proc, SCM arg1)
{
  return scm_i_chase_pairs (arg1, (scm_t_bits) SCM_SUBRF (proc));
}

static SCM 
call_closure_1 (SCM proc, SCM arg1)
{
  const SCM env = SCM_EXTEND_ENV (TMP_CLOSURE_FORMALS (proc),
                                  scm_list_1 (arg1),
                                  TMP_ENV (proc));
  const SCM result = scm_eval_body (TMP_CLOSURE_BODY (proc), env);
  return result;
}

static SCM
call_subr_n_1 (SCM proc, SCM arg1)
{
  return scm_i_gsubr_apply (proc, scm_list_1 (arg1));
}

scm_t_trampoline_1
scm_trampoline_1 (SCM proc)
{
  reference_parameter (proc);

  scm_t_trampoline_1 trampoline;

  if (SCM_IMP (proc)) {
    dereference_parameter (proc);
    return NULL;
  }

  switch (SCM_TYP3 (proc)) {
    case scm_tc3_closure:
      {
	SCM formals = TMP_CLOSURE_FORMALS (proc);
	if (!scm_is_null (formals)
	    && (!scm_is_pair (formals) || !scm_is_pair (TMP_CDR (formals)))) {
	  trampoline = call_closure_1;
	} else {
          dereference_parameter (proc);
	  return NULL;
        }
        goto success;
      }
    case scm_tc3_tc7_1:
    case scm_tc3_tc7_2:
  switch (SCM_TYP7 (proc))
    {
    case scm_tc7_subr_1:
    case scm_tc7_subr_1o:
      trampoline = call_subr1_1;
      goto success;
    case scm_tc7_subr_2o:
      trampoline = call_subr2o_1;
      goto success;
    case scm_tc7_lsubr:
      trampoline = call_lsubr_1;
      goto success;
    case scm_tc7_dsubr:
      trampoline = call_dsubr_1;
      goto success;
    case scm_tc7_cxr:
      trampoline = call_cxr_1;
      goto success;
    case scm_tc7_subr_n:
      trampoline = call_subr_n_1;
      goto success;
    case scm_tc7_asubr:
    case scm_tc7_rpsubr:
      trampoline = scm_call_1;
      goto success;
    }
  default:
    dereference_parameter (proc);
    return NULL; /* not applicable on one arg */
  }

  success:
  /* We only reach this point if a valid trampoline was determined.  */
  dereference_parameter (proc);

  return trampoline;
}

static SCM
call_subr2_2 (SCM proc, SCM arg1, SCM arg2)
{
  return SCM_SUBRF (proc) (arg1, arg2);
}

static SCM
call_lsubr2_2 (SCM proc, SCM arg1, SCM arg2)
{
  return SCM_SUBRF (proc) (arg1, arg2, SCM_EOL);
}

static SCM
call_lsubr_2 (SCM proc, SCM arg1, SCM arg2)
{
  return SCM_SUBRF (proc) (scm_list_2 (arg1, arg2));
}

static SCM 
call_closure_2 (SCM proc, SCM arg1, SCM arg2)
{
  const SCM env = SCM_EXTEND_ENV (TMP_CLOSURE_FORMALS (proc),
                                  scm_list_2 (arg1, arg2),
                                  TMP_ENV (proc));
  const SCM result = scm_eval_body (TMP_CLOSURE_BODY (proc), env);
  return result;
}

static SCM
call_subr_n_2 (SCM proc, SCM arg1, SCM arg2)
{
  return scm_i_gsubr_apply (proc, scm_list_2 (arg1, arg2));
}

scm_t_trampoline_2
scm_trampoline_2 (SCM proc)
{
  reference_parameter (proc);

  scm_t_trampoline_2 trampoline;

  if (SCM_IMP (proc)) {
    dereference_parameter (proc);
    return NULL;
  }

  switch (SCM_TYP3 (proc)) {
  case scm_tc3_closure:
      {
	SCM formals = TMP_CLOSURE_FORMALS (proc);
	if (!scm_is_null (formals)
	    && (!scm_is_pair (formals)
		|| (!scm_is_null (TMP_CDR (formals))
		    && (!scm_is_pair (TMP_CDR (formals))
			|| !scm_is_pair (TMP_CDDR (formals))))))
        {
	  trampoline = call_closure_2;
	} else {
          dereference_parameter (proc);
	  return NULL;
        }
      goto success;
      }
  case scm_tc3_tc7_1:
  case scm_tc3_tc7_2:
  switch (SCM_TYP7 (proc))
    {
    case scm_tc7_subr_2:
    case scm_tc7_subr_2o:
    case scm_tc7_rpsubr:
    case scm_tc7_asubr:
      trampoline = call_subr2_2;
      goto success;
    case scm_tc7_lsubr_2:
      trampoline = call_lsubr2_2;
      goto success;
    case scm_tc7_lsubr:
      trampoline = call_lsubr_2;
      goto success;
    case scm_tc7_subr_n:
      trampoline = call_subr_n_2;
      goto success;
    }
  default:
    dereference_parameter (proc);
    return NULL; /* not applicable on two args */
  }
  success:
  /* We only reach this point if a valid trampoline was determined.  */
  dereference_parameter (proc);

  return trampoline;
}

/* Typechecking for multi-argument MAP and FOR-EACH.

   Verify that each element of the vector ARGV, except for the first,
   is a proper list whose length is LEN.  Attribute errors to WHO,
   and claim that the i'th element of ARGV is WHO's i+2'th argument.  */
static inline void
check_map_args (SCM argv,
		long len,
		SCM gf,
		SCM proc,
		SCM args,
		const char *who)
{
  long i;

  for (i = SCM_SIMPLE_VECTOR_LENGTH (argv) - 1; i >= 1; i--)
    {
      SCM elt = TMP_SIMPLE_VECTOR_REF (argv, i);
      long elt_len = scm_ilength (elt);

      if (elt_len < 0) {
        scm_wrong_type_arg (who, i + 2, elt);
      }

      if (elt_len != len)
	scm_out_of_range_pos (who, elt, scm_from_long (i + 2));
    }
}


SCM_GPROC (s_map, "map", 2, 0, 1, scm_map, g_map);

/* Note: Currently, scm_map applies PROC to the argument list(s)
   sequentially, starting with the first element(s).  This is used in
   evalext.c where the Scheme procedure `map-in-order', which guarantees
   sequential behaviour, is implemented using scm_map.  If the
   behaviour changes, we need to update `map-in-order'.
*/

SCM 
scm_map (SCM proc, SCM arg1, SCM args)
#define FUNC_NAME s_map
{
  reference_parameter (proc);
  reference_parameter (arg1);
  reference_parameter (args);

  long i, len;

  SCM answer_box = scm_i_cons (SCM_BOOL_F, SCM_EOL);
  reference_parameter (answer_box);
  SCM last = answer_box;
  SCM tmp = SCM_UNSPECIFIED;

  len = scm_ilength (arg1);
  SCM_GASSERTn (len >= 0, g_map,
                (tmp=scm_cons2 (proc, arg1, args)),
                SCM_ARG2, s_map);
  dereference_return_value (tmp);

  SCM_VALIDATE_REST_ARGUMENT (args);
  if (scm_is_null (args)) {
    scm_t_trampoline_1 call = scm_trampoline_1 (proc);

    SCM_GASSERT2 (call, g_map, proc, arg1, SCM_ARG1, s_map);

    while (SCM_NIMP (arg1)) {
      SCM_SETCDR (last, scm_list_1 (call (proc, TMP_CAR (arg1))));
      last = TMP_CDR (last);
      arg1 = FLY_CDR (arg1);
    }

    answer_box = fly_cdr (answer_box);
    reference_count_return (answer_box);
    dereference_parameter (answer_box);
    dereference_parameter (proc);
    dereference_parameter (arg1);
    dereference_parameter (args);
    return (answer_box);
  }

  if (scm_is_null (TMP_CDR (args))) {
    SCM arg2 = TMP_CAR (args);

    int len2 = scm_ilength (arg2);
    scm_t_trampoline_2 call = scm_trampoline_2 (proc);

    SCM_GASSERTn (call,
                  g_map, (tmp=scm_cons2 (proc, arg1, args)),
                  SCM_ARG1, s_map);
    dereference_return_value (tmp);
    SCM_GASSERTn (len2 >= 0,
                  g_map,
                  (tmp=scm_cons2 (proc, arg1, args)),
                  SCM_ARG3, s_map);
    dereference_return_value (tmp);

    if (len2 != len)
      SCM_OUT_OF_RANGE (3, arg2);

    while (SCM_NIMP (arg1)) {
      SCM_SETCDR (last, scm_list_1 (call (proc, TMP_CAR (arg1), \
                                                TMP_CAR (arg2))));
      last = TMP_CDR (last);
      arg1 = FLY_CDR (arg1);
      arg2 = TMP_CDR (arg2);
    }

    answer_box = fly_cdr (answer_box);
    reference_count_return (answer_box);
    dereference_parameter (answer_box);
    dereference_parameter (proc);
    dereference_parameter (arg1);
    dereference_parameter (args);
    return (answer_box);
  }


  arg1 = rc_X (scm_cons (rc_Fly (arg1), args));
  dereference_parameter (args);
  args = scm_vector (arg1);
  reference_parameter (args);

  check_map_args (args, len, g_map, proc, arg1, s_map);

  while (1) {
    dereference_parameter (arg1);
    arg1 = SCM_EOL;
    for (i = SCM_SIMPLE_VECTOR_LENGTH (args) - 1; i >= 0; i--) {
      SCM elt = TMP_SIMPLE_VECTOR_REF (args, i);
      if (SCM_IMP (elt)) {
        answer_box = fly_cdr (answer_box);
        reference_count_return (answer_box);
        dereference_parameter (answer_box);
        dereference_parameter (proc);
        dereference_parameter (arg1);
        dereference_parameter (args);
        return (answer_box);
      }
      arg1 = rc_X (scm_cons (TMP_CAR (elt), rc_Fly (arg1)));
      SCM_SIMPLE_VECTOR_SET (args, i, TMP_CDR (elt));
    }
    SCM_SETCDR (last, scm_list_1 (scm_apply (proc, arg1, SCM_EOL)));
    last = TMP_CDR (last);
  }
  /* never reached */
}
#undef FUNC_NAME

SCM_GPROC (s_for_each, "for-each", 2, 0, 1, scm_for_each, g_for_each);

SCM 
scm_for_each (SCM proc, SCM arg1, SCM args)
#define FUNC_NAME s_for_each
{
  reference_parameter (proc);
  reference_parameter (arg1);
  reference_parameter (args);

  long i, len;
  len = scm_ilength (arg1);
  SCM tmp = SCM_UNSPECIFIED;

  SCM_GASSERTn (len >= 0, g_for_each, (tmp=scm_cons2 (proc, arg1, args)),
		SCM_ARG2, s_for_each);
  dereference_return_value (tmp);

  SCM_VALIDATE_REST_ARGUMENT (args);
  if (scm_is_null (args)) {
    scm_t_trampoline_1 call = scm_trampoline_1 (proc);
    SCM_GASSERT2 (call, g_for_each, proc, arg1, SCM_ARG1, s_for_each);

    while (SCM_NIMP (arg1)) {
      rc_Void (call (proc, TMP_CAR (arg1)));
      arg1 = FLY_CDR (arg1);
    }

    dereference_parameter (proc);
    dereference_parameter (arg1);
    dereference_parameter (args);
    return SCM_UNSPECIFIED;
  }

  if (scm_is_null (TMP_CDR (args))) {
    SCM arg2 = TMP_CAR (args);
    int len2 = scm_ilength (arg2);
    scm_t_trampoline_2 call = scm_trampoline_2 (proc);

    SCM_GASSERTn (call, g_for_each,
                  (tmp=scm_cons2 (proc, arg1, args)),
                  SCM_ARG1, s_for_each);
    dereference_return_value (tmp);
    SCM_GASSERTn (len2 >= 0, g_for_each,
                  (tmp=scm_cons2 (proc, arg1, args)),
                  SCM_ARG3, s_for_each);
    dereference_return_value (tmp);

    if (len2 != len)
      SCM_OUT_OF_RANGE (3, arg2);
    while (SCM_NIMP (arg1)) {
      rc_Void (call (proc, TMP_CAR (arg1), TMP_CAR (arg2)));
      arg1 = FLY_CDR (arg1);
      arg2 = TMP_CDR (arg2);
    }

    dereference_parameter (proc);
    dereference_parameter (arg1);
    dereference_parameter (args);
    return SCM_UNSPECIFIED;
  }


  arg1 = rc_X (scm_cons (rc_Fly (arg1), args));
  dereference_parameter (args);
  args = scm_vector (arg1);
  reference_parameter (args);

  check_map_args (args, len, g_for_each, proc, arg1, s_for_each);
  while (1) {
    dereference_parameter (arg1);
    arg1 = SCM_EOL;

    for (i = SCM_SIMPLE_VECTOR_LENGTH (args) - 1; i >= 0; i--) {
      SCM elt = TMP_SIMPLE_VECTOR_REF (args, i);
      reference_parameter (elt);

      if (SCM_IMP (elt)) {
        dereference_parameter (elt);  // bogus
        dereference_parameter (proc);
        dereference_parameter (arg1);
        dereference_parameter (args);
        return SCM_UNSPECIFIED;
      }
      arg1 = rc_X (scm_cons (TMP_CAR (elt), rc_Fly (arg1)));
      SCM_SIMPLE_VECTOR_SET (args, i, TMP_CDR (elt));

      dereference_parameter (elt);
    }

    rc_Void (scm_apply (proc, arg1, SCM_EOL));
  }
  /* never reached */
}
#undef FUNC_NAME


SCM 
scm_closure (SCM formals, SCM body, SCM env)
{
  reference_parameter (formals);
  reference_parameter (body);
  reference_parameter (env);
  SCM z;

  // z = scm_double_cell (SCM_UNPACK (SCM_EOL) + scm_tc3_closure,
  z = scm_double_cell (scm_tc3_closure,
                       SCM_EOL,
                       SCM_UNPACK (formals),
                       SCM_UNPACK (body),
                       SCM_UNPACK (env));

  reference (z, formals);
  reference (z, body);
  reference (z, env);
  global_number_of_closures++;

  /* SCM closcar = scm_i_cons (code, SCM_EOL); */
  /* z = scm_cell (SCM_UNPACK (closcar) + scm_tc3_closure, (scm_t_bits) env); */
  /* reference (z, closcar); dereference_return_value (closcar); */
  /* reference (z, env); */

  reference_count_return (z); // bogus count
  dereference_parameter (formals);
  dereference_parameter (body);
  dereference_parameter (env);
  return z;
}


SCM_DEFINE (scm_cons_source, "cons-source", 3, 0, 0, 
            (SCM xorig, SCM x, SCM y),
	    "Create and return a new pair whose car and cdr are @var{x} and @var{y}.\n"
	    "Any source properties associated with @var{xorig} are also associated\n"
	    "with the new pair.")
#define FUNC_NAME s_scm_cons_source
{
  reference_parameter (xorig);
  
  SCM z;
  z = scm_cons (x, y);
  reference_parameter (z);

  reference_count_return (z);
  dereference_parameter (z);
  dereference_parameter (xorig);
  return (z);
}
#undef FUNC_NAME


/* The function scm_copy_tree is used to copy an expression tree to allow the
 * memoizer to modify the expression during memoization.  scm_copy_tree
 * creates deep copies of pairs and vectors, but not of any other data types,
 * since only pairs and vectors will be parsed by the memoizer.
 *
 * To avoid infinite recursion due to cyclic structures, the hare-and-tortoise
 * pattern is used to detect cycles.  In fact, the pattern is used in two
 * dimensions, vertical (indicated in the code by the variable names 'hare'
 * and 'tortoise') and horizontal ('rabbit' and 'turtle').  In both
 * dimensions, the hare/rabbit will take two steps when the tortoise/turtle
 * takes one.
 *
 * The vertical dimension corresponds to recursive calls to function
 * copy_tree: This happens when descending into vector elements, into cars of
 * lists an    REPLACE THIS CRAP PLEASE   ist.  In this dimension, the
 * tortoise follows the hare by using the processor stack: Every stack frame
 * will hold an instance of struct t_trace.  These instances are connected in
 * a way that represents the trace of the hare, which thus can be followed by
 * the tortoise.  The tortoise will always point to struct t_trace instances
 * relating to SCM objects that have already been copied.  Thus, a cycle is
 * detected if the tortoise and the hare point to the same object,
 *
 * The horizontal dimension is within one execution of copy_tree, when the
 * function cdr's along the pairs of a list.  This is the standard
 * hare-and-tortoise implementation, found several times in guile (pre-VSI).  */

struct t_trace {
  struct t_trace *trace; /* These pointers form a trace along the stack. */
  SCM obj;               /* The object handled at the respective stack frame.*/
};

static SCM
copy_tree (struct t_trace *const hare,
           struct t_trace *tortoise,
           unsigned int tortoise_delay)
{
  if (!scm_is_pair (hare->obj)) { //&& !scm_is_simple_vector (hare->obj)) {
    return hare->obj;

  } else { /* Prepare the trace along the stack.  */
    struct t_trace new_hare;
    hare->trace = &new_hare;

    /* The tortoise will make its step after the delay has elapsed.  Note
     * that in contrast to the typical hare-and-tortoise pattern, the step
     * of the tortoise happens before the hare takes its steps.  This is, in
     * principle, no problem, except for the start of the algorithm: Then,
     * it has to be made sure that the hare actually gets its advantage of
     * two steps.  */
    if (tortoise_delay == 0) {
      tortoise_delay = 1;
      tortoise = tortoise->trace;
      ASSERT_SYNTAX (!scm_is_eq (hare->obj, tortoise->obj),
                     s_bad_expression, hare->obj);
    } else {
      --tortoise_delay;
    }

    /* if (scm_is_simple_vector (hare->obj)) { */
    /*   size_t length = SCM_SIMPLE_VECTOR_LENGTH (hare->obj); */
    /*   SCM new_vector = scm_c_make_vector (length, SCM_UNSPECIFIED); */
    /*   reference_parameter (new_vector); */

    /*   /\* Each vector element is copied by recursing into copy_tree, having */
    /*    * the tortoise follow the hare into the depths of the stack.  *\/ */
    /*   unsigned long int i; */
    /*   for (i = 0; i < length; ++i) { */
    /*     SCM new_element; */
    /*     new_hare.obj = SCM_SIMPLE_VECTOR_REF (hare->obj, i); */
    /*     // data graph referenced heap_reference (new_hare, new_hare.obj); */
    /*     new_element = copy_tree (&new_hare, tortoise, tortoise_delay); */
    /*     SCM_SIMPLE_VECTOR_SET (new_vector, i, new_element); */
    /*   } */

    /*   reference_count_return (new_vector); */
    /*   dereference_parameter (new_vector); */
    /*   heap_dereference (new_hare, new_hare.obj); */
    /*   return new_vector; */
    /* } */

    // if (scm_is_pair (hare->obj)) {
      SCM result;
      SCM tail;

      SCM rabbit = hare->obj;
      SCM turtle = hare->obj;
      reference_parameter (turtle);

      SCM copy;

      /* The first pair of the list is treated specially, in order to
       * preserve a potential source code position.  */
      result = tail = scm_cons_source (rabbit, SCM_EOL, SCM_EOL);
      reference_parameter (result);
      reference_parameter (tail);
      new_hare.obj = TMP_CAR (rabbit);
      copy = copy_tree (&new_hare, tortoise, tortoise_delay);
      SCM_SETCAR (tail, copy);

      // RC checkpoint: turtle, result, tail, rabbit (list referenced)
      //      consumed: copy

      /* The remaining pairs of the list are copied by, horizontally,
       * having the turtle follow the rabbit, and, vertically, having the
       * tortoise follow the hare into the depths of the stack.  */
      rabbit = TMP_CDR (rabbit);
      while (scm_is_pair (rabbit)) {
        new_hare.obj = TMP_CAR (rabbit);
        copy = copy_tree (&new_hare, tortoise, tortoise_delay);
        SCM_SETCDR (tail, scm_cons (copy, SCM_UNDEFINED));
        tail = FLY_CDR (tail);

        rabbit = TMP_CDR (rabbit);
        if (scm_is_pair (rabbit)) {
          new_hare.obj = TMP_CAR (rabbit);
          copy = copy_tree (&new_hare, tortoise, tortoise_delay);
          SCM_SETCDR (tail, scm_cons (copy, SCM_UNDEFINED));
          tail = FLY_CDR (tail);
          rabbit = TMP_CDR (rabbit);

          turtle = FLY_CDR (turtle);
          ASSERT_SYNTAX (!scm_is_eq (rabbit, turtle),
                         s_bad_expression, rabbit);
        }
      }

      /* We have to recurse into copy_tree again for the last cdr, in
       * order to handle the situation that it holds a vector.  */
      new_hare.obj = rabbit; heap_reference (new_hare, rabbit);
      copy = copy_tree (&new_hare, tortoise, tortoise_delay);
      SCM_SETCDR (tail, copy);

      reference_count_return (result);
      dereference_parameter (result);
      dereference_parameter (turtle);
      dereference_parameter (tail);
      heap_dereference (new_hare, rabbit);
      return result;
    // }
  }
}

SCM_DEFINE (scm_copy_tree, "copy-tree", 1, 0, 0, 
            (SCM obj),
	    "Recursively copy the data tree that is bound to @var{obj}, and return a\n"
	    "the new data structure.  @code{copy-tree} recurses down the\n"
	    "contents of both pairs and vectors (since both cons cells and vector\n"
	    "cells may point to arbitrary objects), and stops recursing when it hits\n"
	    "any other object.")
#define FUNC_NAME s_scm_copy_tree
{
  reference_parameter (obj);

  /* Prepare the trace along the stack.  */
  struct t_trace trace;
  trace.obj = obj;

  /* In function copy_tree, if the tortoise makes its step, it will do this
   * before the hare has the chance to move.  Thus, we have to make sure that
   * the very first step of the tortoise will not happen after the hare has
   * really made two steps.  This is achieved by passing '2' as the initial
   * delay for the tortoise.  NOTE: Since cycles are unlikely, giving the hare
   * a bigger advantage may improve performance slightly.  */
  SCM answer = copy_tree (&trace, &trace, 2);
  dereference_parameter (obj);
  return answer;
}
#undef FUNC_NAME


SCM 
scm_i_eval_x (SCM exp, SCM env)
{
  reference_parameter (exp);
  reference_parameter (env);

  SCM answer = SCM_UNSPECIFIED;
  if (scm_is_symbol (exp))
    answer = *scm_lookupcar (scm_cons (exp, SCM_UNDEFINED), env, 1);
  else
    answer = SCM_I_XEVAL (exp, env);

  reference_count_return (answer);
  dereference_parameter (exp);
  dereference_parameter (env);
  return (answer);
}

SCM 
scm_i_eval (SCM exp, SCM env)
{
  exp = scm_copy_tree (exp);
  reference_parameter (exp);
  reference_parameter (env);

  SCM answer = SCM_UNSPECIFIED;
  if (scm_is_symbol (exp))
    answer = *scm_lookupcar (scm_cons (exp, SCM_UNDEFINED), env, 1);
  else
    answer = SCM_I_XEVAL (exp, env);

  reference_count_return (answer);
  dereference_parameter (exp);
  dereference_parameter (env);
  return (answer);
}

SCM
sys_eval (SCM exp)
{
  SCM env;
  env = scm_top_level_env (SCM_BOOL_F);
  return scm_i_eval_x (exp, env);
}

/* XXX scm_primitive_eval_x */
SCM
scm_primitive_eval_x (SCM exp)
{
  SCM env;
  /* SCM transformer = scm_current_module_transformer (); */
  /* if (SCM_NIMP (transformer)) */
  /*   exp = scm_call_1 (transformer, exp); */
  env = scm_top_level_env (SCM_BOOL_F);
  return scm_i_eval_x (exp, env);
}


SCM_DEFINE (scm_eval_x, "eval!", 2, 0, 0, (SCM exp, SCM env), "")
#define FUNC_NAME s_scm_eval
{ // RC transparent
  SCM res;

  res = scm_i_eval_x (exp, env);

  return res;
}
#undef FUNC_NAME


SCM_DEFINE (scm_immediate_p, "immediate?", 1, 0, 0, (SCM object), "")
#define FUNC_NAME s_scm_immediate_p
{
  reference_parameter (object);

  SCM res = scm_from_bool (SCM_IMP (object));

  dereference_parameter (object);
  return res;
}
#undef FUNC_NAME


void 
scm_init_eval ()
{
  undefineds = scm_list_1 (SCM_UNDEFINED);
  scm_permanent_object (undefineds);
  SCM_SETCDR (undefineds, undefineds);

  scm_listofnull = scm_list_1 (SCM_EOL);
  rc_declare_data_graph_root (scm_listofnull);

  f_apply = scm_c_define_subr ("apply", scm_tc7_lsubr_2, scm_apply);
  scm_permanent_object (f_apply);

}

void x_inclusion_eval ()
{
  scm_add_feature ("delay");
#include "include/eval.x"
}
