#include "include/config.h"

#include "include/_scm.h"
#include "include/eval.h"

#include "include/validate.h"
#include "include/evalext.h"

SCM_DEFINE (scm_defined_p, "defined?", 1, 1, 0,
            (SCM sym, SCM env),
	    "Return @code{#t} if @var{sym} is defined in the lexical "
	    "environment @var{env}.  When @var{env} is not specified, "
	    "look in the top-level environment as defined by the "
	    "current module.")
#define FUNC_NAME s_scm_defined_p
{
  reference_parameter (sym);
  reference_parameter (env);

  SCM var;

  SCM_VALIDATE_SYMBOL (1, sym);

  if (SCM_UNBNDP (env)) {
    var = scm_sym2var (sym, SCM_BOOL_F, SCM_BOOL_F);
    reference_parameter (var);

  } else {
    SCM frames = env;
    register SCM b;
    for (; SCM_NIMP (frames); frames = TMP_CDR (frames))
      {
        SCM_ASSERT (scm_is_pair (frames), env, SCM_ARG2, FUNC_NAME);
        b = TMP_CAR (frames);
        // XXX module system purge lookup closure rest
        if (scm_is_true (scm_procedure_p (b)))
          break;
        SCM_ASSERT (scm_is_pair (b), env, SCM_ARG2, FUNC_NAME);
        for (b = TMP_CAR (b); SCM_NIMP (b); b = TMP_CDR (b)) {
          if (!scm_is_pair (b)) {
            if (scm_is_eq (b, sym)) {
              dereference_parameter (sym);
              dereference_parameter (env);
              return SCM_BOOL_T;
            } else {
              break;
            }
          }
          if (scm_is_eq (TMP_CAR (b), sym))
            return SCM_BOOL_T;
        }
      }
    var = scm_sym2var (sym,
                       SCM_NIMP (frames) ? SCM_CAR (frames) : SCM_BOOL_F,
                       SCM_BOOL_F);
    reference_parameter (var);
  }
	      
  SCM answer = (SCM_UNDEFINED_P (var) || SCM_UNBNDP (TMP_VARIABLE_REF (var))
                ? SCM_BOOL_F
                : SCM_BOOL_T);
  dereference_parameter (sym);
  dereference_parameter (env);
  dereference_parameter (var);
  return (answer);
}
#undef FUNC_NAME


SCM_REGISTER_PROC (s_map_in_order, "map-in-order", 2, 0, 1, scm_map);


void 
scm_init_evalext ()
{
#include "include/evalext.x"
}
