/*
 * Nature: System Environment
 */

/* includes have been copy&paste work */
#include "include/config.h"


#include "include/_scm.h"

#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/eval.h"
#include "include/smob.h"
#include "include/procprop.h"
#include "include/vectors.h"
#include "include/variable.h"
#include "include/ports.h" /* XXX debugging */
#include "include/validate.h"

#include "include/dictionary.h"

#include <string.h>  // strlen

static SCM global_system_dictionary = 0;

#include "include/init.h" // XXX debugging

/* scm_sym2var
 *
 */
SCM 
scm_sym2var (SCM sym, SCM proc, SCM definep)
#define FUNC_NAME "scm_sym2var"
{
  reference_parameter (sym);
  reference_parameter (proc);  // TODO "environment" / system environment link
  reference_parameter (definep);

  SCM answer = SCM_UNSPECIFIED;

  if (scm_is_false (definep)) {
    answer = lookup_system_variable (sym);
  } else {
    answer = create_system_variable (sym);
  }

  /* if (scm_is_true (answer) && !SCM_VARIABLEP (answer)) */
  /*   SCM_MISC_ERROR ("~S is not bound to a variable", scm_list_1 (sym)); */

  reference_count_return (answer);
  dereference_parameter (sym);
  dereference_parameter (proc);  // TODO "environment" / system environment link
  dereference_parameter (definep);
  return (answer);
}
#undef FUNC_NAME

/* Interpreter Initialization /define/ */
SCM
scm_define (SCM sym, SCM value)
{
  reference_parameter (sym);
  reference_parameter (value);

  SCM var = scm_sym2var (sym, SCM_BOOL_T, SCM_BOOL_T);
  reference_parameter (var);
  SCM val = SCM_VARIABLE_REF (var);
  reference_parameter (val);

  if (global_debug_interpreter && val != SCM_UNDEFINED) {  // XXX
    SCM bym = scm_to_sen (scm_symbol_to_brass (sym));
    reference_parameter (bym);
    fprintf (stderr, "scm_define: system variable redefinition: %s - ",
                     BYTE_STRING_DATA (bym));
    dereference_parameter (bym);

    { // XXX debugging
      scm_puts ("scm_define: ", scm_current_error_port ());
      scm_write (var, scm_current_error_port ());
      scm_newline (scm_current_error_port ());
    }
  }

  SCM_VARIABLE_SET (var, value);

  reference_count_return (var);
  dereference_parameter (var);
  dereference_parameter (val);
  dereference_parameter (value);
  dereference_parameter (sym);
  return var;
}

SCM
scm_c_define (char *name, SCM value)
{
  return scm_define (c_string_to_symbol (name), value);
}

SCM_DEFINE (scm_define_global, "%define-global", 2, 0, 0,
            (SCM sym, SCM value), "Internal Catch All.")
{
  SCM var = scm_sym2var (sym, SCM_BOOL_T, SCM_BOOL_T);
  reference_parameter (var);
  SCM_VARIABLE_SET (var, value);
  return rc_return_stack_reference (var);
}


SCM
scm_lookup (SCM sym)
{
  reference_parameter (sym);

  SCM var = 
    scm_sym2var (sym, SCM_BOOL_F, SCM_BOOL_F);
  if (scm_is_false (var))
    scm_misc_error ("scm_lookup", "unbound variable: ~S", scm_list_1 (sym));

  dereference_parameter (sym);
  return var;
}

/* Documentational: Former Custom Environments */

SCM
scm_top_level_env (SCM thunk)
{
  if (SCM_IMP (thunk)) {
    dereference_parameter (thunk);
    return SCM_EOL;
  } else {
    return scm_cons (thunk, SCM_EOL);
  }
}

SCM
scm_env_top_level (SCM env)
{
  reference_parameter (env);

  while (scm_is_pair (env)) {
    SCM car_env = TMP_CAR (env);
    reference_parameter (car_env);

    if (!scm_is_pair (car_env) \
        && scm_is_true (scm_procedure_p (car_env))) {
      reference_count_return (car_env);
      dereference_parameter (car_env);
      dereference_parameter (env);
      return car_env;
    }
    env = FLY_CDR (env);
    dereference_parameter (car_env);
  }
  dereference_parameter (env);
  return SCM_BOOL_F;
}

/*
 * Dinosaur Interface
 */
#define SYSTEM_DEFINITION scm_cons
#define SYSDEF_SYMBOL     TMP_CAR
#define SYSDEF_VARIABLE   TMP_CDR

SCM
c_string_to_symbol (const char *string)
{
  //                            XXX source code encoding
  return scm_i_brass_to_symbol (utf8_to_brass (string, strlen (string)));
}

static SCM
scm_ii_make_symbol (SCM br)
{ // This follows brass constructors!!!
  size_t c_len = BR_LENGTH (br);
  size_t num_slots = BR_NSLOTS (c_len);

  SCM sym = scm_tagged_array (scm_tc7_symbol, num_slots);
  SCM_SET_CELL_WORD_0 (sym, c_len);
  memcpy (BR_CHARS (sym), BR_CHARS (br), c_len * sizeof (uint32_t));

  global_number_of_symbols_created++;
  return sym;
}

SCM
scm_i_brass_to_symbol (SCM br)
{
  reference_parameter (br);
  if (BR_LENGTH (br) < 1) {
    fprintf (stderr, \
      " * VSI-C (c_brass_to_symbol): mischiefed brass for symbol * ");
    abort ();
  }

  // While this started out as an unfriendly way of reintroducing the
  // dinosaurs on a second thought we really should check the brass
  // to be a valid symbol just like the /reader/ does (or should do).

  SCM entry = scm_i_create_entry_by_usv_x (global_system_dictionary,
                                           BR_CHARS (br), BR_LENGTH (br));
  SCM definition = GET_DEFINITION (entry);

  if (SCM_UNDEFINED_P (definition)) {
    SCM symbol = scm_ii_make_symbol (br);
    SET_DEFINITION_X (entry, SYSTEM_DEFINITION (symbol, SCM_UNDEFINED));
    definition = GET_DEFINITION (entry);
  }

  dereference_parameter (br);
  RC_RETURN (SYSDEF_SYMBOL (definition));  // be lax here ???
}

SCM
create_system_variable (SCM symbol)
{ // RC: data graph referenced
  long length  = BR_LENGTH (symbol);
  usv_t *chars = BR_CHARS (symbol);

  if (length < 1) {
    fprintf (stderr, \
      " * VSI-C (create_system_variable): mischiefed brass for symbol * ");
    abort ();
  }

  SCM entry = scm_i_create_entry_by_usv_x (global_system_dictionary,
                                           chars, length);
  SCM definition = GET_DEFINITION (entry);

  if (SCM_UNDEFINED_P (definition) \
      || SCM_UNDEFINED_P (SYSDEF_VARIABLE (definition))) {
    // XXX transitional
    SCM var = scm_make_variable (SCM_UNDEFINED);
    SET_DEFINITION_X (entry, SYSTEM_DEFINITION (symbol, var));
    definition = GET_DEFINITION (entry);

  } else if (global_debug_interpreter) {
    if (!(scm_is_eq (symbol, SYSDEF_SYMBOL (definition)))) {
      fprintf (stderr, "Symbol Token Uniqueness violated.\n");
    }
  }

  RC_RETURN (SYSDEF_VARIABLE (definition));  // be lax here ???
}

// Return the /variable/ associated with /symbol/ in the system
// dictionary or /undefined/ if no such thing is found.
SCM
lookup_system_variable (SCM symbol)
// the plan was to replace sym2var with two procedures:
// a define and a lookup where sym2var might remain a compatibility
// wrapper
{ // TODO RC: don't assume "globality" of argument symbol token
  long   length       = BR_LENGTH (symbol);
  usv_t *symbol_chars = BR_CHARS  (symbol);

  SCM entry = scm_i_get_entry_by_usv (global_system_dictionary,
                                      symbol_chars, length);
  if (!SCM_UNDEFINED_P (entry)) {
    SCM definition = GET_DEFINITION (entry);

    if (!SCM_UNDEFINED_P (definition))
      RC_RETURN (SYSDEF_VARIABLE (definition));
    // There are two possible error conditions here:
    // First of all someone got a symbol token we don't know of.
    // Secondly the symbol in the definition we found isn't the
    // same memory object as the argument /symbol/. We ignore
    // both error conditions and strive for correctness.
  } 
  return SCM_UNDEFINED;
}

static void
nature_scan_recurse (SCM answer_box, SCM variable, SCM entry)
{ // RC: lax, answer_box needs to be referenced
  SCM definition = GET_DEFINITION (entry);
  // ??? Maybe we should compare with the content of variables, too.
  //     - unmemoize_expression needs this to operate on variables
  //     - procedure-name would need a comparision by value, i.e.
  //       content of the variable object (in addition to a complete
  //       scan of the environment
  if (!SCM_UNDEFINED_P (definition) \
      && scm_is_eq (variable, SYSDEF_VARIABLE (definition))) {
    SCM_SETCDR (answer_box, scm_cons (definition, answer_box));
  }
  SCM tmp = GET_SUCCESSORS (entry);
  while (scm_is_pair (tmp)) {
    nature_scan_recurse (answer_box, variable, TMP_CAR (tmp));
    tmp = TMP_CDR (tmp);
  }
  return;
}

SCM
nature_reverse_lookup (SCM module, SCM variable)
#define FUNC_NAME "reverse-lookup"
{ // XXX argument module is unused but we should consider the environment!
  reference_parameter (module);
  reference_parameter (variable);

  SCM answer_box = scm_cons (SCM_BOOL_F, SCM_EOL);
  reference_parameter (answer_box);
  nature_scan_recurse (answer_box, variable, global_system_dictionary);

  answer_box = FLY_CDR (answer_box);
  SCM answer = SCM_BOOL_F;
  if (!scm_is_null (answer_box)) {
    // Every definition has a symbol or we could check & abort.
    // Selection Strategy: return the last match which happens to be
    // the head of the list in the /answer_box/.
    answer = SYSDEF_SYMBOL (TMP_CAR (answer_box));
    if (!scm_is_symbol (answer)) {
      answer = SCM_BOOL_F;
      if (global_debug_interpreter) {
        fprintf (stderr,
                 "VSI: nature_reverse_lookup: definition w/o symbol\n");
      }
    }
    reference_parameter (answer);
  }

  dereference_parameter (module);
  dereference_parameter (variable);
  dereference_parameter (answer_box);
  return (rc_return_stack_reference (answer));
}
#undef FUNC_NAME

SCM_DEFINE (scm_system_dictionary_root,
            "system-dictionary-root", 0, 0, 0,
            (), "Internal.")
{
  RC_RETURN (global_system_dictionary);
}


/*
 * Default SCM Interface (Interpreter Initialization)
 */
void
scm_init_nature ()
{
  global_system_dictionary = MAKE_ENTRY (SCM_BOOL_F, SCM_UNDEFINED, SCM_EOL);
  rc_declare_data_graph_root (global_system_dictionary);
}

void
x_inclusion_nature ()
{
  #include "include/nature.x"
}
