// Strings & Brass TODO simple_format
// Compatibilities and Printer EoL 5210
#include "include/config.h"

#include <errno.h>

#include "include/_scm.h"
#include "include/chars.h"
#include "include/smob.h"
#include "include/eval.h"
#include "include/macros.h"
#include "include/procprop.h"
#include "include/read.h"
#include "include/alist.h"
#include "include/ports.h"
#include "include/root.h"
#include "include/brass.h"
#include "include/strports.h"
#include "include/vectors.h"
#include "include/numbers.h"
#include "include/byte-strings.h"

#include "include/validate.h"
#include "include/print.h"

/* {Names of immediate symbols}
 * 
 * This table must agree with the declarations in scm.h: {Immediate Symbols}.
 */

/* This table must agree with the list of flags in tags.h.  */
static const char *iflagnames[] =
{
  "#f",
  "#t",
  "#<undefined>",
  "#<eof>",
  "()",
  "#<unspecified>",
  "#<unused>"
};

SCM_SYMBOL (sym_reader, "reader");


/* {Printing of Scheme Objects}
 */

/* Detection of circular references.
 *
 * Due to other constraints in the implementation, this code has bad
 * time complexity (O (depth * N)), The printer code can be
 * rewritten to be O(N).
 */
#define PUSH_REF(pstate, obj)			\
do						\
{						\
  PSTATE_STACK_SET (pstate, pstate->top, obj);	\
  pstate->top++;				\
  if (pstate->top == pstate->ceiling)		\
    grow_ref_stack (pstate);			\
} while(0)

#define ENTER_NESTED_DATA(pstate, obj, label)			\
do								\
{								\
  register unsigned long i;					\
  for (i = 0; i < pstate->top; ++i)				\
    if (scm_is_eq (TMP_PSTACK_REF (pstate, i), (obj)))  	\
      goto label;						\
  if (pstate->fancyp)						\
    {								\
      if (pstate->top - pstate->list_offset >= pstate->level)	\
	{							\
	  scm_putc ('#', port);					\
	  return;						\
	}							\
    }								\
  PUSH_REF(pstate, obj);					\
} while(0)

#define EXIT_NESTED_DATA(pstate)				\
do								\
{								\
  --pstate->top;						\
  PSTATE_STACK_SET (pstate, pstate->top, SCM_UNDEFINED);	\
}								\
while (0)

#define PSTATE_SIZE 3L

static void
grow_ref_stack (scm_print_state *pstate)
{
  SCM old_vect = pstate->ref_vect;
  size_t old_size = SCM_SIMPLE_VECTOR_LENGTH (old_vect);
  size_t new_size = 2 * pstate->ceiling;
  SCM new_vect = scm_c_make_vector (new_size, SCM_UNDEFINED);
  reference_parameter (new_vect);
  unsigned long int i;

  for (i = 0; i != old_size; ++i)
    SCM_SIMPLE_VECTOR_SET (new_vect, i, TMP_SIMPLE_VECTOR_REF (old_vect, i));

  heap_reference (pstate, new_vect);
  pstate->ref_vect = new_vect;
  heap_dereference (pstate, old_vect);

  pstate->ceiling = new_size;

  dereference_parameter (new_vect);
}

#define TMP_PSTACK_REF(p,i)     TMP_SIMPLE_VECTOR_REF((p)->ref_vect, (i))
#define PSTATE_STACK_REF(p,i)   SCM_SIMPLE_VECTOR_REF((p)->ref_vect, (i))
#define PSTATE_STACK_SET(p,i,v) SCM_SIMPLE_VECTOR_SET((p)->ref_vect, (i), (v))

static void
print_circref (SCM port, scm_print_state *pstate, SCM ref)
{
  reference_parameter (port);

  register long i;
  long self = pstate->top - 1;
  i = pstate->top - 1;
  if (scm_is_pair (TMP_PSTACK_REF (pstate, i)))
    {
      while (i > 0)
	{
	  if (!scm_is_pair (TMP_PSTACK_REF (pstate, i-1))
	      || !scm_is_eq (TMP_CDR (TMP_PSTACK_REF (pstate, i-1)), 
			     TMP_CDR (TMP_PSTACK_REF (pstate, i))))
	    break;
	  --i;
	}
      self = i;
    }
  for (i = pstate->top - 1; 1; --i)
    if (scm_is_eq (TMP_PSTACK_REF(pstate, i), ref))
      break;
  scm_putc ('#', port);
  scm_intprint (i - self, 10, port);
  scm_putc ('#', port);

  dereference_parameter (port);
}

/* Print the name of a symbol. */

void
scm_print_symbol_name (const char *str, size_t len, SCM port)
{ // UC This is weird!
  reference_parameter (port);

  /* This points to the first character that has not yet been written to the
   * port. */
  size_t pos = 0;
  /* This points to the character we're currently looking at. */
  size_t end;
  /* If the name contains weird characters, we'll escape them with
   * backslashes and set this flag; it indicates that we should surround the
   * name with "#{" and "}#". */
  int weird = 0;
  /* Backslashes are not sufficient to make a name weird, but if a name is
   * weird because of other characters, backslahes need to be escaped too.
   * The first time we see a backslash, we set maybe_weird, and mw_pos points
   * to the backslash.  Then if the name turns out to be weird, we re-process
   * everything starting from mw_pos.
   * We could instead make backslashes always weird.  This is not necessary
   * to ensure that the output is (read)-able, but it would make this code
   * simpler and faster. */
  int maybe_weird = 0;
  size_t mw_pos = 0;

  SCM tmp = SCM_UNDEFINED;
  if (len == 0 || str[0] == '\'' || str[0] == '`' || str[0] == ','
      // || quote_keywordish_symbol (str, len)
      || (str[0] == '.' && len == 1)
      || scm_is_true ((tmp=scm_c_locale_stringn_to_number (str, len, 10))))
    {
      scm_lfwrite ("#{", 2, port);
      weird = 1;
    }
  dereference_return_value (tmp);

  for (end = pos; end < len; ++end)
    switch (str[end])
      {
#ifdef BRACKETS_AS_PARENS
      case '[':
      case ']':
#endif
      case '(':
      case ')':
      case '"':
      case ';':
      case '#':
      case PCM_WHITE_SPACES:
      case PCM_LINE_INCREMENTORS:
      weird_handler:
	if (maybe_weird)
	  {
	    end = mw_pos;
	    maybe_weird = 0;
	  }
	if (!weird)
	  {
	    scm_lfwrite ("#{", 2, port);
	    weird = 1;
	  }
	if (pos < end)
	  scm_lfwrite (str + pos, end - pos, port);
	{
	  char buf[2];
	  buf[0] = '\\';
	  buf[1] = str[end];
	  scm_lfwrite (buf, 2, port);
	}
	pos = end + 1;
	break;
      case '\\':
	if (weird)
	  goto weird_handler;
	if (!maybe_weird)
	  {
	    maybe_weird = 1;
	    mw_pos = pos;
	  }
	break;
      default:
	break;
      }
  if (pos < end)
    scm_lfwrite (str + pos, end - pos, port);
  if (weird)
    scm_lfwrite ("}#", 2, port);

  dereference_parameter (port);
}

/* Print generally.  Handles both write and display according to PSTATE.
 */
SCM_GPROC(s_write, "write", 1, 1, 0, scm_write, g_write);
SCM_GPROC(s_display, "display", 1, 1, 0, scm_display, g_display);

static void iprin1 (SCM exp, SCM port, scm_print_state *pstate);

void 
scm_iprin1 (SCM exp, SCM port, scm_print_state *pstate)
{
  reference_parameter (exp);
  reference_parameter (port);

  switch (SCM_ITAG3 (exp))
    {
    case scm_tc3_closure:
    case scm_tc3_tc7_1:
    case scm_tc3_tc7_2:
      /* These tc3 tags should never occur in an immediate value.  They are
       * only used in cell types of non-immediates, i. e. the value returned
       * by SCM_CELL_TYPE (exp) can use these tags.
       */
      scm_ipruk ("immediate", exp, port);
      break;
    case scm_tc3_int_1:
    case scm_tc3_int_2:
      scm_intprint (SCM_I_INUM (exp), 10, port);
      break;
    case scm_tc3_imm24:
      if (SCM_CHARP (exp))
	{
	  long i = SCM_CHAR (exp);

	  if (SCM_WRITINGP (pstate))
	    {
	      scm_puts ("#\\", port);
	      if ((i >= 0) && (i <= ' ') && scm_charnames[i])
		scm_puts (scm_charnames[i], port);
#ifndef EBCDIC
	      else if (i == '\177')
		scm_puts (scm_charnames[scm_n_charnames - 1], port);
#endif
	      else if (i < 0 || i > '\177')
		scm_intprint (i, 8, port);
	      else
		scm_putc (i, port);   // sic! UC C0 only else numericals
	    }
	  else
            {
              scm_write_char (exp, port);
            }
	}
      else if (SCM_IFLAGP (exp)
	       && ((size_t) SCM_IFLAGNUM (exp) < (sizeof iflagnames / sizeof (char *))))
        {
          scm_puts (iflagnames [SCM_IFLAGNUM (exp)], port);
        }
      else if (SCM_ISYMP (exp))
        {
          scm_i_print_isym (exp, port);
        }
      else if (SCM_ILOCP (exp))
	{
          scm_i_print_iloc (exp, port);
	}
      else
	{
	  /* unknown immediate value */
	  scm_ipruk ("immediate", exp, port);
	}
      break;
    case scm_tc3_cons:
      switch (SCM_TYP7 (exp))
	{
	case scm_tcs_cons_imcar:
	case scm_tcs_cons_nimcar:
	  ENTER_NESTED_DATA (pstate, exp, circref);
	  scm_iprlist ("(", exp, ')', port, pstate);
	  EXIT_NESTED_DATA (pstate);
	  break;
	circref:
	  print_circref (port, pstate, exp);
	  break;
	case scm_tcs_closures: {
          SCM tmp1 = SCM_UNSPECIFIED;
          SCM tmp2 = SCM_UNSPECIFIED;
          SCM formals = TMP_CLOSURE_FORMALS (exp);

          scm_puts ("#<procedure", port);
          scm_putc (' ', port);
          scm_iprin1 (scm_procedure_name (exp), port, pstate);
          scm_putc (' ', port);

          scm_iprin1 (formals, port, pstate);

          scm_putc ('>', port);

          dereference_return_value (tmp1);
          dereference_return_value (tmp2);
        }
	  break;
	case scm_tc7_number:
          switch SCM_TYP16 (exp) {
          case scm_tc16_big:
            scm_bigprint (exp, port, pstate);
            break;
          case scm_tc16_real:
            scm_print_real (exp, port, pstate);
            break;
          case scm_tc16_complex:
            scm_print_complex (exp, port, pstate);
            break;
          case scm_tc16_fraction:
            scm_i_print_fraction (exp, port, pstate);
            break;
          }
	  break;
        case scm_tc7_byte_string:
          scm_print_byte_string (exp, port, pstate);
          break;
	case scm_tc7_brass:
	  if (SCM_WRITINGP (pstate))
	    {
	      size_t i, j, len;
	      usv_t *data;

	      scm_putc ('"', port);
	      len = BR_LENGTH (exp);
	      data = BR_CHARS (exp);
	      for (i = 0, j = 0; i < len; ++i)
		{
		  usv_t ch = data[i];
		  if ((ch < 32 && ch != '\n') || (127 <= ch && ch < 148))
		    {
		      static char const hex[]="0123456789abcdef";
		      char buf[4];

		      scm_lfwrite_usv (data + j, i-j, port);
		      buf[0] = '\\';
		      buf[1] = 'x';
		      buf[2] =  hex [ch / 16];
		      buf[3] = hex [ch % 16];
                      // UC range check: ctrls in common
		      scm_lfwrite (buf, 4, port);
                      
		      data = BR_CHARS (exp);
		      j = i+1;
		    }
		  else if (ch == '"' || ch == '\\')
		    {
		      scm_lfwrite_usv (data + j, i-j, port);
		      scm_putc ('\\', port);
		      data = BR_CHARS (exp);
		      j = i;
		    }
		}
	      scm_lfwrite_usv (data + j, i-j, port);
	      // scm_lfwrite (data+j, i-j, port);
	      scm_putc ('"', port);
	    }
	  else
            {
              scm_lfwrite_br (exp, port);
            }
	  break;
	case scm_tc7_symbol:
          {
            scm_lfwrite_usv (BR_CHARS (exp), BR_LENGTH (exp), port);
            /* SCM bs = scm_to_pen (scm_symbol_to_brass (exp)); */
            /* reference_parameter (bs); */
            /* scm_print_symbol_name (BYTE_STRING_DATA (exp), */
            /*                        BYTE_STRING_LENGTH (exp), */
            /*                        port); */
            /* dereference_parameter (bs); */
          }
	  break;

	case scm_tc7_variable:
	  scm_i_variable_print (exp, port, pstate);
	  break;

	case scm_tc7_vector:
	  ENTER_NESTED_DATA (pstate, exp, circref);
	  scm_puts ("#(", port);
	  {
	    register long i;
	    long last = SCM_SIMPLE_VECTOR_LENGTH (exp) - 1;
	    int cutp = 0;
	    if (pstate->fancyp
		&& SCM_SIMPLE_VECTOR_LENGTH (exp) > pstate->length)
	      {
		last = pstate->length - 1;
		cutp = 1;
	      }
	    for (i = 0; i < last; ++i)
	      {
		/* CHECK_INTS; */
		scm_iprin1 (SCM_SIMPLE_VECTOR_REF (exp, i), port, pstate);
		scm_putc (' ', port);
	      }
	    if (i == last)
	      {
		/* CHECK_INTS; */
		scm_iprin1 (SCM_SIMPLE_VECTOR_REF (exp, i), port, pstate);
	      }
	    if (cutp)
	      scm_puts (" ...", port);
	    scm_putc (')', port);
	  }
	  EXIT_NESTED_DATA (pstate);
	  break;
	case scm_tcs_subrs:
	  scm_puts (SCM_SUBR_GENERIC (exp)
		    ? "#<primitive-generic "
		    : "#<primitive-procedure ",
		    port);
          SCM bame = scm_to_sen (scm_symbol_to_brass (SCM_SNAME (exp)));
          reference_parameter (bame);
	  scm_puts (BYTE_STRING_DATA (bame), port);
          dereference_parameter (bame);
	  scm_putc ('>', port);
	  break;
	case scm_tc7_port:
	  {
	    long i = SCM_PTOBNUM (exp);
	    if (i < scm_numptob
		&& scm_ptobs[i].print
		&& (scm_ptobs[i].print) (exp, port, pstate))
	      break;
	    goto punk;
	  }
	case scm_tc7_smob:
	  ENTER_NESTED_DATA (pstate, exp, circref);
	  SCM_SMOB_DESCRIPTOR (exp).print (exp, port, pstate);
	  EXIT_NESTED_DATA (pstate);
	  break;
	default:
	punk:
	  scm_ipruk ("type", exp, port);
	}
    }
  dereference_parameter (exp);
  dereference_parameter (port);
}

void 
scm_prin1 (SCM exp, SCM port, int writingp)
{
  reference_parameter (exp);
  reference_parameter (port);

  scm_print_state pstate;

  pstate.ceiling  = PSTATE_SIZE;
  pstate.writingp = writingp;
  pstate.fancyp   = 0;

  pstate.level       = 0;
  pstate.length      = 0;
  pstate.list_offset = 0;
  pstate.top         = 0;

  // The right thing would be to use a "stack list" but what I
  // found was smob -> scm struct -> c struct -> vector with an
  // escalating growth algorithm.
  // Changing the vector to a stack list could be "simulated" by
  // by keeping head & tail and changing the PUSH_STACK and
  // ENTER_NESTED_DATA routines. Unchecked variants might be
  // interesting, too. Be brave and declare the structure static
  // and there won't be any need to reinitialize the vector all
  // the time but then it keeps referenced everything it saw last.
  pstate.ref_vect = scm_c_make_vector (PSTATE_SIZE, SCM_UNDEFINED);
  heap_reference (pstate, pstate.ref_vect);

  scm_iprin1 (exp, port, &pstate);

  heap_dereference (pstate, pstate.ref_vect);

  dereference_parameter (exp);
  dereference_parameter (port);
  return;
}


/* Print an integer.
 */

void 
scm_intprint (scm_t_intmax n, int radix, SCM port)
{
  char num_buf[SCM_INTBUFLEN];
  scm_lfwrite (num_buf, scm_iint2str (n, radix, num_buf), port);
}

void 
scm_uintprint (scm_t_uintmax n, int radix, SCM port)
{
  char num_buf[SCM_INTBUFLEN];
  scm_lfwrite (num_buf, scm_iuint2str (n, radix, num_buf), port);
}

/* Print an object of unrecognized type.
 */

void 
scm_ipruk (char *hdr, SCM ptr, SCM port)
{
  scm_puts ("#<unknown-", port);
  scm_puts (hdr, port);
  /* XXX GC->TRIM: missing scm_in_heap_p */
  /* if (scm_in_heap_p (ptr)) */
  /*   { */
  /*     scm_puts (" (0x", port); */
  /*     scm_uintprint (SCM_CELL_WORD_0 (ptr), 16, port); */
  /*     scm_puts (" . 0x", port); */
  /*     scm_uintprint (SCM_CELL_WORD_1 (ptr), 16, port); */
  /*     scm_puts (") @", port); */
  /*   } */
  scm_puts (" 0x", port);
  scm_uintprint (SCM_UNPACK (ptr), 16, port);
  scm_putc ('>', port);
}


/* Print a list.
 */
void 
scm_iprlist (char *hdr, SCM exp, int tlr, SCM port, scm_print_state *pstate)
{
  register SCM hare, tortoise;
  long floor = pstate->top - 2;
  scm_puts (hdr, port);
  /* CHECK_INTS; */
  if (pstate->fancyp)
    goto fancy_printing;
  
  /* Run a hare and tortoise so that total time complexity will be
     O(depth * N) instead of O(N^2). */
  hare = TMP_CDR (exp);
  tortoise = exp;
  while (scm_is_pair (hare))
    {
      if (scm_is_eq (hare, tortoise))
	goto fancy_printing;
      hare = TMP_CDR (hare);
      if (!scm_is_pair (hare))
	break;
      hare = TMP_CDR (hare);
      tortoise = TMP_CDR (tortoise);
    }
  
  /* No cdr cycles intrinsic to this list */
  scm_iprin1 (TMP_CAR (exp), port, pstate);
  for (exp = TMP_CDR (exp); scm_is_pair (exp); exp = TMP_CDR (exp))
    {
      register long i;

      for (i = floor; i >= 0; --i)
	if (scm_is_eq (TMP_PSTACK_REF(pstate, i), exp))
	  goto circref;
      PUSH_REF (pstate, exp);
      scm_putc (' ', port);
      /* CHECK_INTS; */
      scm_iprin1 (TMP_CAR (exp), port, pstate);
    }
  if (!scm_is_null (exp))
    {
      scm_puts (" . ", port);
      scm_iprin1 (exp, port, pstate);
    }

end:
  scm_putc (tlr, port);
  pstate->top = floor + 2;
  return;
  
fancy_printing:
  {
    long n = pstate->length;
    
    scm_iprin1 (TMP_CAR (exp), port, pstate);
    exp = TMP_CDR (exp); --n;
    for (; scm_is_pair (exp); exp = SCM_CDR (exp))
      {
	register unsigned long i;

	for (i = 0; i < pstate->top; ++i)
	  if (scm_is_eq (TMP_PSTACK_REF(pstate, i), exp))
	    goto fancy_circref;
	if (pstate->fancyp)
	  {
	    if (n == 0)
	      {
		scm_puts (" ...", port);
		goto skip_tail;
	      }
	    else
	      --n;
	  }
	PUSH_REF(pstate, exp);
	++pstate->list_offset;
	scm_putc (' ', port);
	/* CHECK_INTS; */
	scm_iprin1 (TMP_CAR (exp), port, pstate);
      }
  }
  if (!scm_is_null (exp))
    {
      scm_puts (" . ", port);
      scm_iprin1 (exp, port, pstate);
    }
skip_tail:
  pstate->list_offset -= pstate->top - floor - 2;
  goto end;

fancy_circref:
  pstate->list_offset -= pstate->top - floor - 2;
  
circref:
  scm_puts (" . ", port);
  print_circref (port, pstate, exp);
  goto end;
}



int
scm_valid_oport_value_p	(SCM val)
{
  return (SCM_OPOUTPORTP (val));
}

/* SCM_GPROC(s_write, "write", 1, 1, 0, scm_write, g_write); */

SCM 
scm_write (SCM obj, SCM port)
{
  reference_parameter (obj);
  if (SCM_UNBNDP (port)) {
    port = scm_current_output_port ();
  }
  reference_parameter (port);

  SCM_ASSERT (scm_valid_oport_value_p (port), port, SCM_ARG2, s_write);

  scm_prin1 (obj, port, 1);
#if 0
#ifdef HAVE_PIPE
# ifdef EPIPE
  if (EPIPE == errno)
    scm_close_port (port);
# endif
#endif
#endif

  dereference_parameter (obj);
  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}


/* SCM_GPROC(s_display, "display", 1, 1, 0, scm_display, g_display); */

SCM 
scm_display (SCM obj, SCM port)
{
  reference_parameter (obj);
  if (SCM_UNBNDP (port))
    port = scm_current_output_port ();
  reference_parameter (port);

  SCM_ASSERT (scm_valid_oport_value_p (port), port, SCM_ARG2, s_display);

  scm_prin1 (obj, port, 0);
#if 0
#ifdef HAVE_PIPE
# ifdef EPIPE
  if (EPIPE == errno)
    scm_close_port (port);
# endif
#endif
#endif

  dereference_parameter (obj);
  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}


SCM_DEFINE (scm_simple_format, "simple-format", 2, 0, 1,
            (SCM destination, SCM message, SCM args),
	    "Write @var{message} to @var{destination}, defaulting to\n"
	    "the current output port.\n"
	    "@var{message} can contain @code{~A} (was @code{%s}) and\n"
	    "@code{~S} (was @code{%S}) escapes.  When printed,\n"
	    "the escapes are replaced with corresponding members of\n"
	    "@var{ARGS}:\n"
	    "@code{~A} formats using @code{display} and @code{~S} formats\n"
	    "using @code{write}.\n"
	    "If @var{destination} is @code{#t}, then use the current output\n"
	    "port, if @var{destination} is @code{#f}, then return a string\n"
	    "containing the formatted text. Does not add a trailing newline.")
#define FUNC_NAME s_scm_simple_format
{
  reference_parameter (destination);
  reference_parameter (message);
  reference_parameter (args);

  SCM port, answer = SCM_UNSPECIFIED;
  int fReturnString = 0;
  int writingp;

  if (scm_is_eq (destination, SCM_BOOL_T)) {
    dereference_parameter (destination); // RC bogus
    destination = port = scm_current_output_port ();
    reference_parameter (port);
    reference_parameter (destination);

  } else if (scm_is_false (destination)) {
    fReturnString = 1;
    port = scm_mkstrport (scm_c_make_byte_string (0),
                          SCM_OPN | SCM_WRTNG,
                          FUNC_NAME);
    destination = port;
    reference_parameter (port);
    reference_parameter (destination);
  } else {
    SCM_VALIDATE_OPORT_VALUE (1, destination);
    port = destination;
    reference_parameter (port);
  }
  SCM_VALIDATE_BRASS (2, message);
  SCM_VALIDATE_REST_ARGUMENT (args);

  // RC checkpoint: destination, message, args, port

  size_t idx = 0;
  size_t start = 0;
  size_t end = start + BR_LENGTH (message);
  for (idx = start; idx != end; idx++) {
    usv_t *p = &((BR_CHARS (message))[idx]);
    if (*p == '~') {
      if (idx + 1 == end)
        break;
      
      idx++; p++;
      switch (*p) {
      case 'A': case 'a':
        writingp = 0;
        break;
      case 'S': case 's':
        writingp = 1;
        break;
      case '~':
        scm_lfwrite_usv (&((BR_CHARS (message))[start]), idx - start, port);
        start = idx + 1;
        continue;
      case '%':
        scm_lfwrite_usv (&((BR_CHARS (message))[start]), idx - start - 1, port);
        scm_newline (port);
        start = idx + 1;
        continue;
      default:
        if (global_debug_interpreter) {
          fprintf (stderr, "VSI-C: error in format: unsupported option\n");
        }
        /* SCM_MISC_ERROR ("FORMAT: Unsupported format option ~~~A - use (format) instead", */
        /*                 scm_list_1 (SCM_MAKE_CHAR (*p))); */
	    
      }

      if (!scm_is_pair (args)) {
        if (global_debug_interpreter) {
          fprintf (stderr, "VSI-C: error in format: missing argument\n");
        }
        /* SCM_MISC_ERROR ("FORMAT: Missing argument for ~~~A", */
        /*                 scm_list_1 (SCM_MAKE_CHAR (*p))); */
      }
			  		
      scm_lfwrite_usv (&((BR_CHARS (message))[start]), idx - start - 1, port);
      /* we pass destination here */
      scm_prin1 (TMP_CAR (args), destination, writingp);
      args = FLY_CDR (args);
      start = idx + 1;
    }
  }
  scm_lfwrite_usv (&((BR_CHARS (message))[start]), idx - start, port);

  if (!scm_is_eq (args, SCM_EOL)) {
    if (global_debug_interpreter) {
      fprintf (stderr, "VSI-C: error in format: superfluous arguments\n");
    }
    /* SCM_MISC_ERROR ("FORMAT: ~A superfluous arguments", */
    /*     	    scm_list_1 (scm_length (args))); */
  }

  if (fReturnString) {
    SCM b = scm_strport_to_string (destination);
    reference_parameter (b);
    answer = scm_from_sen (BYTE_STRING_DATA (b), BYTE_STRING_LENGTH (b));
    dereference_parameter (b);
  }
  reference_parameter (answer);

  dereference_parameter (destination);
  dereference_parameter (args);
  dereference_parameter (port);
  dereference_parameter (message);
  return (rc_return_stack_reference (answer));
}
#undef FUNC_NAME


SCM_DEFINE (scm_newline, "newline", 0, 1, 0, 
            (SCM port),
	    "Send a newline to @var{port}.\n"
	    "If @var{port} is omitted, send to the current output port.")
#define FUNC_NAME s_scm_newline
{
  if (SCM_UNBNDP (port))
    port = scm_current_output_port ();
  reference_parameter (port);

  SCM_VALIDATE_OPORT_VALUE (1, port);

  scm_putc ('\n', port);
  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_write_char, "write-char", 1, 1, 0,
            (SCM chr, SCM port),
	    "Send character @var{chr} to @var{port}.")
#define FUNC_NAME s_scm_write_char
{
  if (SCM_UNBNDP (port))
    port = scm_current_output_port ();
  reference_parameter (port);

  SCM_VALIDATE_CHAR (1, chr);
  SCM_VALIDATE_OPORT_VALUE (2, port);

  SCM bs = utf8_usv_to_bs (SCM_CHAR (chr)); // XXX port encoding
  reference_parameter (bs);
  scm_lfwrite (BYTE_STRING_DATA (bs), BYTE_STRING_LENGTH (bs), port);
  dereference_parameter (bs);

  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


void
scm_init_print ()
{
  /* scm_init_opts (scm_print_options, scm_print_opts, SCM_N_PRINT_OPTIONS); */

  /* scm_print_options (scm_list_4 (scm_from_locale_symbol ("highlight-prefix"), */
  /*       			 scm_from_locale_string ("*"), */
  /*       			 scm_from_locale_symbol ("highlight-suffix"), */
  /*       			 scm_from_locale_string ("*"))); */

  /* scm_print_opts[SCM_PRINT_KEYWORD_STYLE_I].val = SCM_UNPACK (sym_reader); */
}

void
x_inclusion_print ()
{
#include "include/print.x"
}
