// Strings & Brass (+ Unicode C0 compatibilities)

/* See stime.c for comments on why _POSIX_C_SOURCE is not always defined. */
#define _LARGEFILE64_SOURCE      /* ask for stat64 etc */
#ifdef __hpux
#define _POSIX_C_SOURCE 199506L  /* for readdir_r */
#endif
#if defined(__INTERIX) && !defined(_REENTRANT)
# define _REENTRANT   /* ask Interix for readdir_r prototype */
#endif

#include "include/config.h"

/* This blob per the Autoconf manual (under "Particular Functions"),
   updated to match that of Gnulib.  !!! These blurbs aren't blobs but
   all over the place. XXX */
#ifndef alloca
# if HAVE_ALLOCA_H
#  include <alloca.h>
# elif defined __GNUC__
#  define alloca __builtin_alloca
# elif defined _AIX
#  define alloca __alloca
# elif defined _MSC_VER
#  include <malloc.h>
#  define alloca _alloca
# else
#  include <stddef.h>
#  ifdef  __cplusplus
extern "C"
#  endif
void *alloca (size_t);
# endif
#endif

#include <stdio.h>
#include <errno.h>

#include "include/_scm.h"
#include "include/smob.h"
#include "include/feature.h"
#include "include/fports.h"
#include "include/iselect.h"
#include "include/brass.h"
#include "include/vectors.h"
#include "include/dynwind.h"

#include "include/validate.h"
#include "include/filesys.h"
#include "include/byte-strings.h"


#ifdef HAVE_IO_H
#include <io.h>
#endif

#ifdef HAVE_DIRECT_H
#include <direct.h>
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef LIBC_H_WITH_UNISTD_H
#include <libc.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif


#if HAVE_DIRENT_H
# include <dirent.h>
# define NAMLEN(dirent) strlen((dirent)->d_name)
#else
# define dirent direct
# define NAMLEN(dirent) (dirent)->d_namlen
# if HAVE_SYS_NDIR_H
#  include <sys/ndir.h>
# endif
# if HAVE_SYS_DIR_H
#  include <sys/dir.h>
# endif
# if HAVE_NDIR_H
#  include <ndir.h>
# endif
#endif

/* Ultrix has S_IFSOCK, but no S_ISSOCK.  Ipe!  */
#if defined (S_IFSOCK) && ! defined (S_ISSOCK)
#define S_ISSOCK(mode) (((mode) & S_IFMT) == S_IFSOCK)
#endif

/* dirfd() returns the file descriptor underlying a "DIR*" directory stream.
   Found on MacOS X for instance.  The following definition is for Solaris
   10, it's probably not right elsewhere, but that's ok, it shouldn't be
   used elsewhere.  Crib note: If we need more then gnulib has a dirfd.m4
   figuring out how to get the fd (dirfd function, dirfd macro, dd_fd field,
   or d_fd field).  */
#ifndef dirfd
#define dirfd(dirstream) ((dirstream)->dd_fd)
#endif



/* Two helper macros for an often used pattern */

#define STRING_SYSCALL(str,bstr,code)        \
  do {                                       \
    int eno;                                 \
    SCM unhbat = scm_to_sen (str);           \
    reference_parameter (unhbat);            \
    char *bstr = BYTE_STRING_DATA (unhbat);  \
    SCM_SYSCALL (code);                      \
    eno = errno; errno = eno;                \
    dereference_parameter (unhbat);          \
  } while (0)

#define STRING2_SYSCALL(str1,bstr1,str2,bstr2,code)  \
  do {                                               \
    int eno;                                         \
    SCM unhba1 = scm_to_sen (str1);                  \
    SCM unhba2 = scm_to_sen (str2);                  \
    reference_parameter (unhba1);                    \
    reference_parameter (unhba2);                    \
    char *bstr1, *bstr2;                             \
    bstr1 = BYTE_STRING_DATA (unhba1);               \
    bstr2 = BYTE_STRING_DATA (unhba2);               \
    SCM_SYSCALL (code);                              \
    eno = errno;  errno = eno;                       \
    dereference_parameter (unhba1);                  \
    dereference_parameter (unhba2);                  \
  } while (0)



/* {Permissions}
 */

#ifdef HAVE_CHOWN
SCM_DEFINE (scm_chown, "chown", 3, 0, 0, (SCM object, SCM owner, SCM group), "")
#define FUNC_NAME s_scm_chown
{
  reference_parameter (object);
  reference_parameter (owner);
  reference_parameter (group);

  int rv;

#ifdef HAVE_FCHOWN
  if (scm_is_integer (object) || (SCM_OPFPORTP (object)))
    {
      int fdes = (SCM_OPFPORTP (object)?
		  SCM_FPORT_FDES (object) : scm_to_int (object));

      SCM_SYSCALL (rv = fchown (fdes, scm_to_int (owner), scm_to_int (group)));
    }
  else
#endif
    {
      STRING_SYSCALL (object, c_object,
		      rv = chown (c_object,
				  scm_to_int (owner), scm_to_int (group)));
    }

  dereference_parameter (object);
  dereference_parameter (owner);
  dereference_parameter (group);
  if (rv == -1)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME
#endif /* HAVE_CHOWN */


SCM_DEFINE (scm_chmod, "chmod", 2, 0, 0, (SCM object, SCM mode), "")
#define FUNC_NAME s_scm_chmod
{
  reference_parameter (object);
  reference_parameter (mode);

  int rv;
  int fdes;

  SCM_VALIDATE_BYTE_STRING_AS (2, mode, mode_t);

  if (scm_is_integer (object) || SCM_OPFPORTP (object)) {
    if (scm_is_integer (object))
      fdes = scm_to_int (object);
    else
      fdes = SCM_FPORT_FDES (object);
    SCM_SYSCALL (rv = fchmod (fdes, *(mode_t*)BYTE_STRING_DATA (mode)));

  } else {
    STRING_SYSCALL (object, c_object,
                    rv = chmod (c_object, *(mode_t*)BYTE_STRING_DATA (mode)));
  }

  dereference_parameter (object);
  dereference_parameter (mode);
  if (rv == -1)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_umask, "umask", 0, 1, 0, (SCM mode), "")
#define FUNC_NAME s_scm_umask
{
  reference_parameter (mode);

  mode_t mask;
  if (SCM_UNBNDP (mode)) {
    mask = umask (0);
    umask (mask);

  } else {
    SCM_VALIDATE_BYTE_STRING_AS (1, mode, mode_t);
    mask = umask (*(mode_t*)BYTE_STRING_DATA (mode));
  }

  dereference_parameter (mode);
  return scm_make_byte_string_with_copy (&mask, sizeof (mask));
}
#undef FUNC_NAME



SCM_DEFINE (scm_open_fdes, "open-fdes", 2, 1, 0, 
            (SCM path, SCM flags, SCM mode),
	    "Similar to @code{open} but return a file descriptor instead of\n"
	    "a port.")
#define FUNC_NAME s_scm_open_fdes
{
  reference_parameter (path);
  reference_parameter (flags);
  reference_parameter (mode);

  int fd;
  int iflags;
  mode_t imode;

  SCM_VALIDATE_BYTE_STRING_AS (2, flags, int);
  SCM_VALIDATE_BYTE_STRING_AS (3, mode,  mode_t); // XXX is optional
  
  iflags =    *(int*)BYTE_STRING_DATA (flags);
  imode  = *(mode_t*)BYTE_STRING_DATA (mode);
  STRING_SYSCALL (path, c_path, fd = open_or_open64 (c_path, iflags, imode)); // implicit type check of path as br

  reference_parameter (path);
  reference_parameter (flags);
  reference_parameter (mode);
  if (fd == -1)
    SCM_SYSERROR;
  return scm_from_int (fd);
}
#undef FUNC_NAME

SCM_DEFINE (scm_open, "open", 2, 1, 0, 
            (SCM path, SCM flags, SCM mode),
	    "Open the file named by @var{path} for reading and/or writing.\n"
	    "@var{flags} is an integer specifying how the file should be opened.\n"
	    "@var{mode} is an integer specifying the permission bits of the file, if\n"
	    "it needs to be created, before the umask is applied.  The default is 666\n"
	    "(Unix itself has no default).\n\n"
	    "@var{flags} can be constructed by combining variables using @code{logior}.\n"
	    "Basic flags are:\n\n"
	    "@defvar O_RDONLY\n"
	    "Open the file read-only.\n"
	    "@end defvar\n"
	    "@defvar O_WRONLY\n"
	    "Open the file write-only.\n"
	    "@end defvar\n"
	    "@defvar O_RDWR\n"
	    "Open the file read/write.\n"
	    "@end defvar\n"
	    "@defvar O_APPEND\n"
	    "Append to the file instead of truncating.\n"
	    "@end defvar\n"
	    "@defvar O_CREAT\n"
	    "Create the file if it does not already exist.\n"
	    "@end defvar\n\n"
	    "See the Unix documentation of the @code{open} system call\n"
	    "for additional flags.")
#define FUNC_NAME s_scm_open
{
  reference_parameter (path);
  reference_parameter (flags);
  reference_parameter (mode);

  SCM newpt;
  char *port_mode;
  int fd;
  int iflags;

  SCM_VALIDATE_BYTE_STRING_AS (2, flags, int);
  fd = scm_to_int (scm_open_fdes (path, flags, mode));

  iflags = *(int*)BYTE_STRING_DATA (flags);
  if (iflags & O_RDWR) {
    if (iflags & O_APPEND)
      port_mode = "a+";
    else if (iflags & O_CREAT)
      port_mode = "w+";
    else
      port_mode = "r+";
  } else {
    if (iflags & O_APPEND)
      port_mode = "a";
    else if (iflags & O_WRONLY)
      port_mode = "w";
    else
      port_mode = "r";
  }
  newpt = scm_fdes_to_port (fd, port_mode, path);

  dereference_parameter (path);
  dereference_parameter (flags);
  dereference_parameter (mode);
  return newpt;
}
#undef FUNC_NAME

SCM_DEFINE (scm_close, "close", 1, 0, 0, (SCM fd_or_port), "")
#define FUNC_NAME s_scm_close
{
  reference_parameter (fd_or_port);

  int rv;
  int fd;

  if (SCM_PORTP (fd_or_port)) {
    return scm_close_port (rc_Fly (fd_or_port));
  }
  fd = scm_to_int (fd_or_port);
  dereference_parameter (fd_or_port);

  SCM_SYSCALL (rv = close (fd));
  /* following scsh, closing an already closed file descriptor is
     not an error. LOL */
  if (rv < 0 && errno != EBADF)
    SCM_SYSERROR;
  return scm_from_bool (rv >= 0);
}
#undef FUNC_NAME

SCM_DEFINE (scm_close_fdes, "close-fdes", 1, 0, 0, 
            (SCM fd),
	    "A simple wrapper for the @code{close} system call.\n"
	    "Close file descriptor @var{fd}, which must be an integer.\n"
	    "Unlike close (@pxref{Ports and File Descriptors, close}),\n"
	    "the file descriptor will be closed even if a port is using it.\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_close_fdes
{
  reference_parameter (fd);

  int c_fd;
  int rv;

  c_fd = scm_to_int (fd);
  SCM_SYSCALL (rv = close (c_fd));

  dereference_parameter (fd);
  if (rv < 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


/* {Files}
 */

SCM_SYMBOL (scm_sym_regular, "regular");
SCM_SYMBOL (scm_sym_directory, "directory");
#ifdef S_ISLNK
SCM_SYMBOL (scm_sym_symlink, "symlink");
#endif
SCM_SYMBOL (scm_sym_block_special, "block-special");
SCM_SYMBOL (scm_sym_char_special, "char-special");
SCM_SYMBOL (scm_sym_fifo, "fifo");
SCM_SYMBOL (scm_sym_sock, "socket");
SCM_SYMBOL (scm_sym_unknown, "unknown");

static SCM 
scm_stat2scm (struct stat_or_stat64 *stat_temp)
{ // If we opt for SCA here BEWARE the exact layout of the structure
  // seems to change considerably from system architecture to system.
  // This would definitely need to be a compile time dependency for SCA.
  // See also stat(2).
  SCM ans = scm_c_make_vector (15, SCM_UNSPECIFIED);
  reference_parameter (ans);
  
  SCM_SIMPLE_VECTOR_SET(ans, 0, scm_from_ulong (stat_temp->st_dev));
  SCM_SIMPLE_VECTOR_SET(ans, 1, scm_from_ino_t_or_ino64_t (stat_temp->st_ino));
  SCM_SIMPLE_VECTOR_SET(ans, 2, \
    scm_make_byte_string_with_copy (&(stat_temp->st_mode),
                                    sizeof (stat_temp->st_mode)));
  SCM_SIMPLE_VECTOR_SET(ans, 3, scm_from_ulong (stat_temp->st_nlink));
  SCM_SIMPLE_VECTOR_SET(ans, 4, scm_from_ulong (stat_temp->st_uid));
  SCM_SIMPLE_VECTOR_SET(ans, 5, scm_from_ulong (stat_temp->st_gid));
#ifdef HAVE_STRUCT_STAT_ST_RDEV
  SCM_SIMPLE_VECTOR_SET(ans, 6, scm_from_ulong (stat_temp->st_rdev));
#else
  SCM_SIMPLE_VECTOR_SET(ans, 6, SCM_BOOL_F);
#endif
  SCM_SIMPLE_VECTOR_SET(ans, 7, scm_from_off_t_or_off64_t (stat_temp->st_size));
  SCM_SIMPLE_VECTOR_SET(ans, 8, scm_from_ulong (stat_temp->st_atime));
  SCM_SIMPLE_VECTOR_SET(ans, 9, scm_from_ulong (stat_temp->st_mtime));
  SCM_SIMPLE_VECTOR_SET(ans, 10, scm_from_ulong (stat_temp->st_ctime));
#ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
  SCM_SIMPLE_VECTOR_SET(ans, 11, scm_from_ulong (stat_temp->st_blksize));
#else
  SCM_SIMPLE_VECTOR_SET(ans, 11, scm_from_ulong (4096L));
#endif
#ifdef HAVE_STRUCT_STAT_ST_BLOCKS
  SCM_SIMPLE_VECTOR_SET(ans, 12, scm_from_blkcnt_t_or_blkcnt64_t (stat_temp->st_blocks));
#else
  SCM_SIMPLE_VECTOR_SET(ans, 12, SCM_BOOL_F);
#endif
  {
    int mode = stat_temp->st_mode;
    
    if (S_ISREG (mode))
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_regular);
    else if (S_ISDIR (mode))
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_directory);
#ifdef S_ISLNK
    /* systems without symlinks probably don't have S_ISLNK */
    else if (S_ISLNK (mode))
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_symlink);
#endif
    else if (S_ISBLK (mode))
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_block_special);
    else if (S_ISCHR (mode))
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_char_special);
    else if (S_ISFIFO (mode))
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_fifo);
#ifdef S_ISSOCK
    else if (S_ISSOCK (mode))
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_sock);
#endif
    else
      SCM_SIMPLE_VECTOR_SET(ans, 13, scm_sym_unknown);

    mode &= ~S_IFMT;
    SCM_SIMPLE_VECTOR_SET(ans, 14, \
      scm_make_byte_string_with_copy (&mode, sizeof (mode)));
  }  

  return (rc_return_stack_reference (ans));
}

SCM_DEFINE (scm_stat, "stat", 1, 0, 0, (SCM object), "See stat(2)." )
#define FUNC_NAME s_scm_stat
{
  reference_parameter (object);

  int rv;
  int fdes;
  struct stat_or_stat64 stat_temp;

  if (scm_is_integer (object)) {
    SCM_SYSCALL (rv = fstat_or_fstat64 (scm_to_int (object), &stat_temp));

  } else if (SCM_BRASSP (object)) {
    SCM bile = scm_to_sen (object);
    reference_parameter (bile);
    SCM_SYSCALL (rv = stat_or_stat64 (BYTE_STRING_DATA (bile), &stat_temp));
    dereference_parameter (bile);

  } else {
    SCM_VALIDATE_OPFPORT (1, object);
    fdes = SCM_FPORT_FDES (object);
    SCM_SYSCALL (rv = fstat_or_fstat64 (fdes, &stat_temp));
  }

  dereference_parameter (object);
  if (rv == -1) {
    int en = errno;

    SCM_SYSERROR_MSG ("~A: ~S",
                      scm_list_2 (scm_strerror (scm_from_int (en)),
                                  object),
                      en);
  }
  return scm_stat2scm (&stat_temp);
}
#undef FUNC_NAME


/* {Modifying Directories}
 */

#ifdef HAVE_LINK
SCM_DEFINE (scm_link, "link", 2, 0, 0,
            (SCM oldpath, SCM newpath),
	    "Creates a new name @var{newpath} in the file system for the\n"
	    "file named by @var{oldpath}.  If @var{oldpath} is a symbolic\n"
	    "link, the link may or may not be followed depending on the\n"
	    "system.")
#define FUNC_NAME s_scm_link
{
  reference_parameter (oldpath);
  reference_parameter (newpath);

  int val;

  STRING2_SYSCALL (oldpath, c_oldpath,
		   newpath, c_newpath,
		   val = link (c_oldpath, c_newpath));

  dereference_parameter (oldpath);
  dereference_parameter (newpath);
  if (val != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME
#endif /* HAVE_LINK */

#ifdef HAVE_RENAME
#define my_rename rename
#else
static int
my_rename (const char *oldname, const char *newname)
{
  int rv;

  SCM_SYSCALL (rv = link (oldname, newname));
  if (rv == 0)
    {
      SCM_SYSCALL (rv = unlink (oldname));
      if (rv != 0)
	/* unlink failed.  remove new name */
	SCM_SYSCALL (unlink (newname)); 
    }
  return rv;
}
#endif

SCM_DEFINE (scm_rename, "rename-file", 2, 0, 0,
            (SCM oldname, SCM newname),
	    "Renames the file specified by @var{oldname} to @var{newname}.\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_rename
{
  reference_parameter (oldname);
  reference_parameter (newname);

  int rv;

  STRING2_SYSCALL (oldname, c_oldname,
		   newname, c_newname,
		   rv = my_rename (c_oldname, c_newname));

  dereference_parameter (oldname);
  dereference_parameter (newname);
  if (rv != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE (scm_delete_file, "delete-file", 1, 0, 0, 
           (SCM str),
	    "Deletes (or \"unlinks\") the file specified by @var{path}.")
#define FUNC_NAME s_scm_delete_file
{
  reference_parameter (str);

  int ans;
  STRING_SYSCALL (str, c_str, ans = unlink (c_str));

  dereference_parameter (str);
  if (ans != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

#ifdef HAVE_MKDIR
SCM_DEFINE (scm_mkdir, "mkdir", 1, 1, 0, (SCM path, SCM mode), "")
#define FUNC_NAME s_scm_mkdir
{
  reference_parameter (path);
  reference_parameter (mode);

  int rv;
  mode_t mask;

  if (SCM_UNBNDP (mode)) {
    mask = umask (0);
    umask (mask);
    STRING_SYSCALL (path, c_path, rv = mkdir (c_path, 0777 ^ mask));
  } else {
    SCM_VALIDATE_BYTE_STRING_AS (2, mode, mode_t);
    STRING_SYSCALL (path, c_path, \
      rv = mkdir (c_path, *(mode_t*)BYTE_STRING_DATA (mode)));
  }

  dereference_parameter (path);
  dereference_parameter (mode);
  if (rv != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME
#endif /* HAVE_MKDIR */

#ifdef HAVE_RMDIR
SCM_DEFINE (scm_rmdir, "rmdir", 1, 0, 0, 
            (SCM path),
	    "Remove the existing directory named by @var{path}.  The directory must\n"
	    "be empty for this to succeed.  The return value is unspecified.")
#define FUNC_NAME s_scm_rmdir
{
  reference_parameter (path);

  int val;

  STRING_SYSCALL (path, c_path, val = rmdir (c_path));

  dereference_parameter (path);
  if (val != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME
#endif



/* {Examining Directories}
 */

scm_t_tag scm_tc16_dir;


SCM_DEFINE (scm_directory_stream_p, "directory-stream?", 1, 0, 0, 
            (SCM obj),
	    "Return a boolean indicating whether @var{object} is a directory\n"
	    "stream as returned by @code{opendir}.")
#define FUNC_NAME s_scm_directory_stream_p
{
  reference_parameter (obj);
  SCM answer = scm_from_bool (SCM_DIRP (obj));
  dereference_parameter (obj);
  return answer;
}
#undef FUNC_NAME


SCM_DEFINE (scm_opendir, "opendir", 1, 0, 0, 
            (SCM dirname),
	    "Open the directory specified by @var{path} and return a directory\n"
	    "stream.")
#define FUNC_NAME s_scm_opendir
{
  reference_parameter (dirname);
  DIR *ds;
  STRING_SYSCALL (dirname, c_dirname, ds = opendir (c_dirname));

  dereference_parameter (dirname);
  if (ds == NULL)
    SCM_SYSERROR;

  SCM cell;
  SCM_NEWSMOB (cell, scm_tc16_dir, ds);
  SCM_SET_CELL_WORD_0 (cell, SCM_DIR_FLAG_OPEN);
  return cell;
}
#undef FUNC_NAME



SCM_DEFINE (scm_readdir, "readdir", 1, 0, 0, 
            (SCM port),
	    "Return (as a string) the next directory entry from the directory stream\n"
	    "@var{stream}.  If there is no remaining entry to be read then the\n"
	    "end of file object is returned.")
#define FUNC_NAME s_scm_readdir
{
  reference_parameter (port);

  struct dirent_or_dirent64 *rdent;

  SCM_VALIDATE_DIR (1, port);
  if (!SCM_DIR_OPEN_P (port))
    SCM_MISC_ERROR ("Directory ~S is not open.", scm_list_1 (port));

#if HAVE_READDIR_R
  {
    struct dirent_or_dirent64 de; /* just for sizeof */
    DIR    *ds = (DIR *) SCM_CELL_WORD_1 (port);
    size_t namlen;
#ifdef NAME_MAX
    char   buf [SCM_MAX (sizeof (de),
                         sizeof (de) - sizeof (de.d_name) + NAME_MAX + 1)];
#else
    char   *buf;
    long   name_max = fpathconf (dirfd (ds), _PC_NAME_MAX);
    if (name_max == -1)
      SCM_SYSERROR;
    buf = alloca (SCM_MAX (sizeof (de),
                           sizeof (de) - sizeof (de.d_name) + name_max + 1));
#endif

    errno = 0;
    SCM_SYSCALL (readdir_r_or_readdir64_r (ds, (struct dirent_or_dirent64 *) buf, &rdent));

    dereference_parameter (port);
    if (errno != 0)
      SCM_SYSERROR;
    if (! rdent)
      return SCM_EOF_VAL;

    namlen = NAMLEN (rdent);

    return (rdent ? scm_from_sen (rdent->d_name, NAMLEN (rdent))
            : SCM_EOF_VAL);
  }
#else
  {
    SCM ret;

    errno = 0;
    SCM_SYSCALL (rdent = readdir_or_readdir64 ((DIR *) SCM_CELL_WORD_1 (port)));
    if (errno != 0)
      SCM_SYSERROR;

    dereference_parameter (port);
    ret = (rdent ? scm_from_sen (rdent->d_name, NAMLEN (rdent))
           : SCM_EOF_VAL);

    return ret;
  }
#endif
}
#undef FUNC_NAME


SCM_DEFINE (scm_rewinddir, "rewinddir", 1, 0, 0, 
            (SCM port),
	    "Reset the directory port @var{stream} so that the next call to\n"
	    "@code{readdir} will return the first directory entry.")
#define FUNC_NAME s_scm_rewinddir
{
  reference_parameter (port);

  SCM_VALIDATE_DIR (1, port);
  if (!SCM_DIR_OPEN_P (port))
    SCM_MISC_ERROR ("Directory ~S is not open.", scm_list_1 (port));

  rewinddir ((DIR *) SCM_CELL_WORD_1 (port));

  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


SCM_DEFINE (scm_closedir, "closedir", 1, 0, 0, 
            (SCM port),
	    "Close the directory stream @var{stream}.\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_closedir
{
  reference_parameter (port);

  SCM_VALIDATE_DIR (1, port);

  if (SCM_DIR_OPEN_P (port))
    {
      int sts;

      SCM_SYSCALL (sts = closedir ((DIR *) SCM_CELL_WORD_1 (port)));
      if (sts != 0)
	SCM_SYSERROR;

      SCM_SET_CELL_WORD_0 (port, 0);
    }

  dereference_parameter (port);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


static int 
scm_dir_print (SCM exp, SCM port, scm_print_state *pstate SCM_UNUSED)
{
  scm_puts ("#<", port);
  if (!SCM_DIR_OPEN_P (exp))
    scm_puts ("closed: ", port);
  scm_puts ("directory stream ", port);
  scm_uintprint (SCM_CELL_WORD_1 (exp), 16, port);
  scm_putc ('>', port);
  return 1;
}


static void
scm_dir_release (SCM p)
{
  if (SCM_DIR_OPEN_P (p))
    closedir ((DIR *) SCM_CELL_WORD_1 (p));
  return;
}


/* {Navigating Directories}
 */


SCM_DEFINE (scm_chdir, "chdir", 1, 0, 0, 
            (SCM str),
	    "Change the current working directory to @var{path}.\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_chdir
{
  reference_parameter (str);
  int ans;

  STRING_SYSCALL (str, c_str, ans = chdir (c_str));

  dereference_parameter (str);
  if (ans != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

#ifdef HAVE_GETCWD
SCM_DEFINE (scm_getcwd, "getcwd", 0, 0, 0,
            (),
	    "Return the name of the current working directory.")
#define FUNC_NAME s_scm_getcwd
{
  char *rv;
  size_t size = 100;
  char *wd;
  SCM result;

  wd = scm_malloc (size);
  while ((rv = getcwd (wd, size)) == 0 && errno == ERANGE)
    {
      free (wd);
      size *= 2;
      wd = scm_malloc (size);
    }
  if (rv == 0)
    {
      int save_errno = errno;
      free (wd);
      errno = save_errno;
      SCM_SYSERROR;
    }
  result = scm_from_sen (wd, strlen (wd));
  free (wd);
  return result;
}
#undef FUNC_NAME
#endif /* HAVE_GETCWD */



#ifdef HAVE_SELECT

/* check that element is a port or file descriptor.  if it's a port
   and its buffer is ready for use, add it to the ports_ready list.
   otherwise add its file descriptor to *set.  the type of list can be
   determined from pos: SCM_ARG1 for reads, SCM_ARG2 for writes,
   SCM_ARG3 for excepts.  */
static int
set_element (SELECT_TYPE *set, SCM *ports_ready, SCM element, int pos)
{ // RC: out of protocol
  //     The argument /*ports_ready/ needs to be correctly referenced.
  //     It will conditionally be exchanged transparently, i.e. with
  //     correct referencing.
  reference_parameter (element);
  int fd;

  if (scm_is_integer (element)) {
    fd = scm_to_int (element);

  } else {
    int use_buf = 0;

    SCM_ASSERT (SCM_OPFPORTP (element), element, pos, "select");

    if (pos == SCM_ARG1) {
      /* check whether port has buffered input.  */
      scm_t_port *pt = SCM_PTAB_ENTRY (element);
      
      if (pt->read_pos < pt->read_end)
        use_buf = 1;

    } else if (pos == SCM_ARG2) {
      /* check whether port's output buffer has room.  */
      scm_t_port *pt = SCM_PTAB_ENTRY (element);

      /* > 1 since writing the last byte in the buffer causes flush.  */
      if (pt->write_end - pt->write_pos > 1)
        use_buf = 1;
    }
    fd = use_buf ? -1 : SCM_FPORT_FDES (element);
  }
  if (fd == -1)
    *ports_ready = rc_X (scm_cons (element, rc_Fly (*ports_ready)));
  else
    FD_SET (fd, set);
  return fd;
}

/* check list_or_vec, a list or vector of ports or file descriptors,
   adding each member to either the ports_ready list (if it's a port
   with a usable buffer) or to *set.  the kind of list_or_vec can be
   determined from pos: SCM_ARG1 for reads, SCM_ARG2 for writes,
   SCM_ARG3 for excepts.  */
static int
fill_select_type (SELECT_TYPE *set, SCM *ports_ready, SCM list_or_vec, int pos)
{
  int max_fd = 0;

  if (scm_is_simple_vector (list_or_vec)) {
    int i = SCM_SIMPLE_VECTOR_LENGTH (list_or_vec);
      
    while (--i >= 0) {
      int fd = set_element (set, ports_ready,
                            TMP_SIMPLE_VECTOR_REF (list_or_vec, i), pos);

      if (fd > max_fd)
        max_fd = fd;
    }

  } else {
    while (!scm_is_null (list_or_vec)) {
      int fd = set_element (set, ports_ready, TMP_CAR (list_or_vec), pos);

      if (fd > max_fd)
        max_fd = fd;
      list_or_vec = TMP_CDR (list_or_vec);
    }
  }

  return max_fd;
}

/* if element (a file descriptor or port) appears in *set, cons it to
   list.  return list.  */
static SCM
get_element (SELECT_TYPE *set, SCM element, SCM list)
{
  int fd;

  if (scm_is_integer (element)) {
    fd = scm_to_int (element);

  } else {
    fd = SCM_FPORT_FDES (element);
  }

  if (FD_ISSET (fd, set))
    list = scm_cons (element, list);
  return list;
}

/* construct component of scm_select return value.
   set: pointer to set of file descriptors found by select to be ready
   ports_ready: ports ready due to buffering
   list_or_vec: original list/vector handed to scm_select.
   the return value is a list/vector of ready ports/file descriptors. 
   works by finding the objects in list which correspond to members of
   *set and appending them to ports_ready.  result is converted to a
   vector if list_or_vec is a vector.  */
static SCM 
retrieve_select_type (SELECT_TYPE *set, SCM ports_ready, SCM list_or_vec)
{
  SCM answer_list = ports_ready;

  if (scm_is_simple_vector (list_or_vec)) {
    int i = SCM_SIMPLE_VECTOR_LENGTH (list_or_vec);

    while (--i >= 0) {
      answer_list = get_element (set,
                                 TMP_SIMPLE_VECTOR_REF (list_or_vec, i),
                                 answer_list);
    }
    return scm_vector (answer_list);

  } else {
    /* list_or_vec must be a list.  */
    while (!scm_is_null (list_or_vec)) {
      answer_list = get_element (set, TMP_CAR (list_or_vec), answer_list);
      list_or_vec = TMP_CDR (list_or_vec);
    }
    return answer_list;
  }
}

/* Static helper functions above refer to s_scm_select directly as s_select */
SCM_DEFINE (scm_select, "select", 3, 2, 0, 
            (SCM reads, SCM writes, SCM excepts, SCM secs, SCM usecs), "")
#define FUNC_NAME s_scm_select
{ // TODO Substitute this sofa' with a bare bone sca_fs_select or sca_p_select !!!
  reference_parameter (reads);
  reference_parameter (writes);
  reference_parameter (excepts);
  reference_parameter (secs);
  reference_parameter (usecs);

  struct timeval timeout;
  struct timeval * time_ptr;
  SELECT_TYPE read_set;
  SELECT_TYPE write_set;
  SELECT_TYPE except_set;
  int read_count;
  int write_count;
  int except_count;
  /* these lists accumulate ports which are ready due to buffering. XXX XXX XXX
     their file descriptors don't need to be added to the select sets.  */
  SCM read_ports_ready = SCM_EOL;
  SCM write_ports_ready = SCM_EOL;
  int max_fd;

  if (scm_is_simple_vector (reads)) {
      read_count = SCM_SIMPLE_VECTOR_LENGTH (reads);

  } else {
      read_count = scm_ilength (reads);
      SCM_ASSERT (read_count >= 0, reads, SCM_ARG1, FUNC_NAME);
  }

  if (scm_is_simple_vector (writes)) {
    write_count = SCM_SIMPLE_VECTOR_LENGTH (writes);

  } else {
    write_count = scm_ilength (writes);
    SCM_ASSERT (write_count >= 0, writes, SCM_ARG2, FUNC_NAME);
  }

  if (scm_is_simple_vector (excepts)) {
    except_count = SCM_SIMPLE_VECTOR_LENGTH (excepts);
  } else {
    except_count = scm_ilength (excepts);
    SCM_ASSERT (except_count >= 0, excepts, SCM_ARG3, FUNC_NAME);
  }

  FD_ZERO (&read_set);
  FD_ZERO (&write_set);
  FD_ZERO (&except_set);

  // RC read_ports_ready is an immediate
  max_fd = fill_select_type (&read_set, &read_ports_ready, reads, SCM_ARG1);
  // RC: read_ports_ready is referenced by now

  {
    int write_max = fill_select_type (&write_set, &write_ports_ready, 
				      writes, SCM_ARG2);
    // RC: write_ports_ready is referenced by now

    int except_max = fill_select_type (&except_set, NULL,
				       excepts, SCM_ARG3);

    if (write_max > max_fd)
      max_fd = write_max;
    if (except_max > max_fd)
      max_fd = except_max;
  }

  /* if there's a port with a ready buffer, don't block, just
     check for ready file descriptors.  */
  if (!scm_is_null (read_ports_ready) || !scm_is_null (write_ports_ready)) {
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    time_ptr = &timeout;

  } else if (SCM_UNBNDP (secs) || scm_is_false (secs)) {
    time_ptr = 0;

  } else {
    if (scm_is_unsigned_integer (secs, 0, ULONG_MAX)) {
      timeout.tv_sec = scm_to_ulong (secs);
      if (SCM_UNBNDP (usecs))
        timeout.tv_usec = 0;
      else
        timeout.tv_usec = scm_to_long (usecs);

    } else {
      double fl = scm_to_double (secs);

      if (!SCM_UNBNDP (usecs))
        SCM_WRONG_TYPE_ARG (4, secs);
      if (fl > LONG_MAX)
        SCM_OUT_OF_RANGE (4, secs);
      timeout.tv_sec = (long) fl;
      timeout.tv_usec = (long) ((fl - timeout.tv_sec) * 1000000);
    }
    time_ptr = &timeout;
  }

  {
    int rv = scm_std_select (max_fd + 1,
			     &read_set, &write_set, &except_set,
			     time_ptr);
    if (rv < 0)
      SCM_SYSERROR;
  }
  SCM answer = \
    scm_list_3 (retrieve_select_type (&read_set, read_ports_ready, reads),
                retrieve_select_type (&write_set, write_ports_ready, writes),
                retrieve_select_type (&except_set, SCM_EOL, excepts));
  dereference_parameter (read_ports_ready);
  dereference_parameter (write_ports_ready);
  dereference_parameter (reads);
  dereference_parameter (writes);
  dereference_parameter (excepts);
  dereference_parameter (secs);
  dereference_parameter (usecs);
  return (answer);
}
#undef FUNC_NAME
#endif /* HAVE_SELECT */



#ifdef HAVE_FCNTL
// Just call this our fcntl jump table. ;-)
// We'll try to get away with two stubs here:
// one for ints as the value and one for pointers to something
// (hopefully a correct byte structure). The easiest way to
// craft the pointer is to use /byte-string-address/. Creating
// the correct byte structure description is left as an exercise
// to Scheme level. TODO
// Both variants work for fcntls w/o additional arguments. An
// empty byte string equals a call w/o additional arguments.
// I wasn't able to find any use with more than one optional
// argument by scanning the documentation. In the long run we
// hope to have the means to "address" va_args more directly.
SCM_DEFINE (scm_fcntl_ptr, "unchecked-fcntl-pointer", 2, 1, 0,
            (SCM object, SCM cmd, SCM value), "")
#define FUNC_NAME s_scm_fcntl_ptr
{ // XXX SCA accomodation for va_args of fcntl
  // Use /byte-string-address/ of the byte structure as the
  // argument here.
  reference_parameter (object);
  reference_parameter (cmd);
  reference_parameter (value);

  int rv;
  int fdes;
  int ivalue;

  if (SCM_OPFPORTP (object))
    fdes = SCM_FPORT_FDES (object);
  else
    fdes = scm_to_int (object);


  if (SCM_UNBNDP (value)) {
    ivalue = 0;
    SCM_SYSCALL (rv = fcntl (fdes, scm_to_int (cmd)));
  } else {
    if (!BYTE_STRING_P (value)) {
      scm_wrong_type_arg (FUNC_NAME, 3, value);
    }

    // Just call this our jump table. ;-)
    if (0 == BYTE_STRING_LENGTH (value)) {
      SCM_SYSCALL (rv = fcntl (fdes, scm_to_int (cmd)));

    } else if (sizeof (void*) == BYTE_STRING_LENGTH (value)) {
      // Hey, Glory!, we cast a pointer to a pointer.
      SCM_SYSCALL (rv = fcntl (fdes, scm_to_int (cmd),
                               *(char**)BYTE_STRING_DATA (value))); // sic!

    } else {
      scm_out_of_range (FUNC_NAME, value);
    }
  }

  dereference_parameter (object);
  dereference_parameter (cmd);
  dereference_parameter (value);
  if (rv == -1)
    SCM_SYSERROR;
  return scm_from_int (rv);
}
#undef FUNC_NAME

SCM_DEFINE (scm_fcntl_int, "unchecked-fcntl-int", 2, 1, 0,
            (SCM object, SCM cmd, SCM value), "")
#define FUNC_NAME s_scm_fcntl_int
{ // XXX SCA accomodation for va_args of fcntl
  // /value/ is a byte string holding an /int/
  reference_parameter (object);
  reference_parameter (cmd);
  reference_parameter (value);

  int rv;
  int fdes;
  int ivalue;

  if (SCM_OPFPORTP (object))
    fdes = SCM_FPORT_FDES (object);
  else
    fdes = scm_to_int (object);


  if (SCM_UNBNDP (value)) {
    ivalue = 0;
    SCM_SYSCALL (rv = fcntl (fdes, scm_to_int (cmd)));
  } else {
    if (!BYTE_STRING_P (value)) {
      scm_wrong_type_arg (FUNC_NAME, 3, value);
    }

    // Just call this our jump table. ;-)
    if (0 == BYTE_STRING_LENGTH (value)) {
      SCM_SYSCALL (rv = fcntl (fdes, scm_to_int (cmd)));

    } else if (sizeof (int) == BYTE_STRING_LENGTH (value)) {
      SCM_SYSCALL (rv = fcntl (fdes, scm_to_int (cmd),
                               *(int*)BYTE_STRING_DATA (value)));

    } else {
      scm_out_of_range (FUNC_NAME, value);
    }
  }

  dereference_parameter (object);
  dereference_parameter (cmd);
  dereference_parameter (value);
  if (rv == -1)
    SCM_SYSERROR;
  return scm_from_int (rv);
}
#undef FUNC_NAME
#endif /* HAVE_FCNTL */

SCM_DEFINE (scm_fsync, "fsync", 1, 0, 0, (SCM object), "")
#define FUNC_NAME s_scm_fsync
{
  reference_parameter (object);

  int fdes;

  if (SCM_OPFPORTP (object))
    {
      scm_flush (object);
      fdes = SCM_FPORT_FDES (object);
    }
  else
    fdes = scm_to_int (object);

  dereference_parameter (object);
  if (fsync (fdes) == -1)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

#ifdef HAVE_SYMLINK
SCM_DEFINE (scm_symlink, "symlink", 2, 0, 0, (SCM oldpath, SCM newpath), "")
#define FUNC_NAME s_scm_symlink
{
  reference_parameter (oldpath);
  reference_parameter (newpath);
  int val;

  STRING2_SYSCALL (oldpath, c_oldpath,
		   newpath, c_newpath,
		   val = symlink (c_oldpath, c_newpath));

  dereference_parameter (oldpath);
  dereference_parameter (newpath);
  if (val != 0)
    SCM_SYSERROR;
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME
#endif /* HAVE_SYMLINK */

#ifdef HAVE_READLINK
SCM_DEFINE (scm_readlink, "readlink", 1, 0, 0, 
            (SCM path),
	    "Return the value of the symbolic link named by @var{path} (a\n"
	    "string), i.e., the file that the link points to.")
#define FUNC_NAME s_scm_readlink
{
  reference_parameter (path);

  int rv;
  int size = 255;
  char *buf;
  SCM result;
  char *c_path;
  
  SCM b_path = scm_to_sen (path);
  reference_parameter (b_path);
  c_path = BYTE_STRING_DATA (b_path);
  buf = scm_malloc (size);

  while ((rv = readlink (c_path, buf, size)) == size) {
    free (buf);
    size += size;
    buf = scm_malloc (size);
  }

  if (rv == -1) {
    int save_errno = errno;
    free (buf);
    errno = save_errno;
    SCM_SYSERROR;
  }
  result = scm_from_sen (buf, rv);
  free (buf);

  dereference_parameter (b_path);
  dereference_parameter (path);
  return result;
}
#undef FUNC_NAME
#endif /* HAVE_READLINK */

#ifdef HAVE_LSTAT
SCM_DEFINE (scm_lstat, "lstat", 1, 0, 0, 
            (SCM str),
	    "Similar to @code{stat}, but does not follow symbolic links, i.e.,\n"
	    "it will return information about a symbolic link itself, not the\n"
	    "file it points to.  @var{path} must be a string.")
#define FUNC_NAME s_scm_lstat
{
  reference_parameter (str);

  int rv;
  struct stat_or_stat64 stat_temp;

  STRING_SYSCALL (str, c_str, rv = lstat_or_lstat64 (c_str, &stat_temp));
  if (rv != 0)
    {
      int en = errno;

      SCM_SYSERROR_MSG ("~A: ~S",
			scm_list_2 (scm_strerror (scm_from_int (en)), str),
			en);
    }
  dereference_parameter (str);
  return scm_stat2scm (&stat_temp);
}
#undef FUNC_NAME
#endif /* HAVE_LSTAT */

SCM_DEFINE (scm_copy_file, "copy-file", 2, 0, 0,
            (SCM oldfile, SCM newfile), "")
#define FUNC_NAME s_scm_copy_file
{
  reference_parameter (oldfile);
  reference_parameter (newfile);

  char *c_oldfile, *c_newfile;
  SCM b_oldfile, b_newfile;
  int oldfd, newfd;
  int n, rv;
  char buf[BUFSIZ];
  struct stat_or_stat64 oldstat;

  b_oldfile = scm_to_sen (oldfile); reference_parameter (b_oldfile);
  b_newfile = scm_to_sen (newfile); reference_parameter (b_newfile);
  c_oldfile = BYTE_STRING_DATA (b_oldfile);
  c_newfile = BYTE_STRING_DATA (b_newfile);

  oldfd = open_or_open64 (c_oldfile, O_RDONLY);
  if (oldfd == -1)
    SCM_SYSERROR;

  SCM_SYSCALL (rv = fstat_or_fstat64 (oldfd, &oldstat));
  if (rv == -1)
    goto err_close_oldfd;

  /* use POSIX flags instead of 07777?.  */
  newfd = open_or_open64 (c_newfile, O_WRONLY | O_CREAT | O_TRUNC,
                          oldstat.st_mode & 07777);
  if (newfd == -1)
    {
    err_close_oldfd:
      close (oldfd);
      SCM_SYSERROR;
    }

  while ((n = read (oldfd, buf, sizeof buf)) > 0)
    if (write (newfd, buf, n) != n)
      {
	close (oldfd);
	close (newfd);
	SCM_SYSERROR;
      }
  close (oldfd);
  if (close (newfd) == -1)
    SCM_SYSERROR;

  dereference_parameter (b_oldfile);
  dereference_parameter (b_newfile);
  dereference_parameter (oldfile);
  dereference_parameter (newfile);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME


/* Filename manipulation */

SCM_DEFINE (scm_dirname, "dirname", 1, 0, 0, 
            (SCM filename),
	    "Return the directory name component of the file name\n"
	    "@var{filename}. If @var{filename} does not contain a directory\n"
	    "component, @code{.} is returned.")
#define FUNC_NAME s_scm_dirname
{
  reference_parameter (filename);

  usv_t *s;
  long int i;
  size_t len;
  SCM answer;

  SCM_VALIDATE_BRASS (1, filename);

  s   = BR_CHARS (filename);
  len = BR_LENGTH (filename);

  i = len - 1;
  while (i >= 0 && s[i] == '/') --i;
  while (i >= 0 && s[i] != '/') --i;
  while (i >= 0 && s[i] == '/') --i;
  if (i < 0) {
    if (len > 0 && s[0] == '/')
      answer = scm_c_make_brass_with_copy (s, 1);
    else
      answer = scm_from_cen (".");

  } else {
    answer = scm_c_make_brass_with_copy (s, i + 1);
  }

  dereference_parameter (filename);
  return answer;
}
#undef FUNC_NAME

SCM_DEFINE (scm_basename, "basename", 1, 1, 0, 
            (SCM filename, SCM suffix),
	    "Return the base name of the file name @var{filename}. The\n"
	    "base name is the file name without any directory components.\n"
	    "If @var{suffix} is provided, and is equal to the end of\n"
	    "@var{basename}, it is removed also.")
#define FUNC_NAME s_scm_basename
{
  reference_parameter (filename);
  reference_parameter (suffix);

  usv_t *f = 0;
  usv_t *s = 0;
  int i, j, len, end;
  SCM answer;

  SCM_VALIDATE_BRASS (1, filename);
  f   = BR_CHARS (filename);
  len = BR_LENGTH (filename);

  if (SCM_UNBNDP (suffix)) {
    j = -1;  // turns off access to /s/
  } else {
    SCM_VALIDATE_BRASS (2, suffix);
    s = BR_CHARS (suffix);
    j = BR_LENGTH (suffix) - 1;
  }
  i = len - 1;
  while (i >= 0 && f[i] == '/') --i;
  end = i;
  while (i >= 0 && j >= 0 && f[i] == s[j]) --i, --j;
  if (j == -1)
    end = i;
  while (i >= 0 && f[i] != '/') --i;

  if (i == end) {
    if (len > 0 && f[0] == '/')
      answer = scm_c_make_brass_with_copy (f, 1);
    else
      answer = scm_from_cen (".");
  } else {
    answer = scm_c_make_brass_with_copy (f + i+1, end+1);
  }

  dereference_parameter (filename);
  dereference_parameter (suffix);
  return (answer);
}
#undef FUNC_NAME





void
scm_init_filesys ()
{
  scm_tc16_dir = scm_make_smob_type ("directory", 0);
  scm_set_smob_release (scm_tc16_dir, 2, scm_dir_release);
  scm_set_smob_print (scm_tc16_dir, scm_dir_print);
}

void
x_inclusion_filesys ()
{ 
#ifdef O_RDONLY
  scm_c_define ("O_RDONLY", scm_from_int (O_RDONLY));
#endif 	       
#ifdef O_WRONLY
  scm_c_define ("O_WRONLY", scm_from_int (O_WRONLY));
#endif 	       
#ifdef O_RDWR
  scm_c_define ("O_RDWR", scm_from_int (O_RDWR));
#endif 	       
#ifdef O_CREAT
  scm_c_define ("O_CREAT", scm_from_int (O_CREAT));
#endif 	       
#ifdef O_EXCL  
  scm_c_define ("O_EXCL", scm_from_int (O_EXCL));
#endif 	       
#ifdef O_NOCTTY
  scm_c_define ("O_NOCTTY", scm_from_int (O_NOCTTY));
#endif 	       
#ifdef O_TRUNC 
  scm_c_define ("O_TRUNC", scm_from_int (O_TRUNC));
#endif 	       
#ifdef O_APPEND
  scm_c_define ("O_APPEND", scm_from_int (O_APPEND));
#endif 	       
#ifdef O_NONBLOCK
  scm_c_define ("O_NONBLOCK", scm_from_int (O_NONBLOCK));
#endif 	       
#ifdef O_NDELAY
  scm_c_define ("O_NDELAY", scm_from_int (O_NDELAY));
#endif 	       
#ifdef O_SYNC  
  scm_c_define ("O_SYNC", scm_from_int (O_SYNC));
#endif 
#ifdef O_LARGEFILE  
  scm_c_define ("O_LARGEFILE", scm_from_int (O_LARGEFILE));
#endif 

#ifdef F_DUPFD  
  scm_c_define ("F_DUPFD", scm_from_int (F_DUPFD));
#endif 
#ifdef F_GETFD  
  scm_c_define ("F_GETFD", scm_from_int (F_GETFD));
#endif 
#ifdef F_SETFD  
  scm_c_define ("F_SETFD", scm_from_int (F_SETFD));
#endif 
#ifdef F_GETFL  
  scm_c_define ("F_GETFL", scm_from_int (F_GETFL));
#endif 
#ifdef F_SETFL  
  scm_c_define ("F_SETFL", scm_from_int (F_SETFL));
#endif 
#ifdef F_GETOWN  
  scm_c_define ("F_GETOWN", scm_from_int (F_GETOWN));
#endif 
#ifdef F_SETOWN  
  scm_c_define ("F_SETOWN", scm_from_int (F_SETOWN));
#endif 
#ifdef FD_CLOEXEC  
  scm_c_define ("FD_CLOEXEC", scm_from_int (FD_CLOEXEC));
#endif

#include "include/filesys.x"
}
