// Strings & Brass

#include "include/config.h"

#include <errno.h>
#include <signal.h>  /* for SIG constants */
#include <stdlib.h>  /* for getenv */

#include "include/_scm.h"

#include "include/signals.h"
#include "include/brass.h"
#include "include/byte-strings.h"

#include "include/validate.h"
#include "include/simpos.h"
#include "include/dynwind.h"

#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif

#include "include/posix.h"


extern int system(); // yes, we surely have one of those???


#ifdef HAVE_SYSTEM
SCM_DEFINE (scm_system, "system", 0, 1, 0, (SCM cmd), "")
#define FUNC_NAME s_scm_system
{
  reference_parameter (cmd);
  int rv, eno;
  
  if (SCM_UNBNDP (cmd))
    {
      rv = system (NULL);
      return scm_from_bool(rv);
    }  
  SCM_VALIDATE_BRASS (1, cmd);
  errno = 0;
  SCM b_cmd = scm_to_sen (cmd);
  reference_parameter (b_cmd);

  rv = system (BYTE_STRING_DATA (b_cmd));
  eno = errno; errno = eno;

  dereference_parameter (b_cmd);
  dereference_parameter (cmd);
  if (rv == -1 || (rv == 127 && errno != 0))
    SCM_SYSERROR;
  return scm_from_int (rv);
}
#undef FUNC_NAME
#endif /* HAVE_SYSTEM */


SCM_DEFINE (scm_getenv, "getenv", 1, 0, 0, (SCM nam), "")
#define FUNC_NAME s_scm_getenv
{
  reference_parameter (nam);
  char *val;
  SCM bam = scm_to_sen (nam);
  reference_parameter (bam);

  val = getenv (BYTE_STRING_DATA (bam));

  dereference_parameter (bam);
  dereference_parameter (nam);
  return val ? scm_from_sen (val, strlen (val)) : SCM_BOOL_F;
}
#undef FUNC_NAME

SCM_DEFINE (scm_primitive_exit, "primitive-exit", 0, 1, 0, (SCM status), "")
#define FUNC_NAME s_scm_primitive_exit
{
  reference_parameter (status);
  int cstatus = 0;
  if (!SCM_UNBNDP (status))
    cstatus = scm_to_int (status);
  exit (cstatus); // RC: no dereferencing on exit
}
#undef FUNC_NAME

SCM_DEFINE (scm_primitive__exit, "primitive-_exit", 0, 1, 0, (SCM status), "")
#define FUNC_NAME s_scm_primitive__exit
{
  // RC OK
  int cstatus = 0;
  if (!SCM_UNBNDP (status))
    cstatus = scm_to_int (status);
  _exit (cstatus);
}
#undef FUNC_NAME



void
scm_init_simpos ()
{
#include "include/simpos.x"
}
