// Strings & Brass
// TODO: Byte String Flags

#include "include/config.h"

#include <errno.h>
#include <gmp.h>

#include "include/_scm.h"
#include "include/feature.h"
#include "include/fports.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/vectors.h"
#include "include/dynwind.h"

#include "include/validate.h"
#include "include/socket.h"

#include "include/iselect.h"

#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <sys/types.h>
#ifdef HAVE_WINSOCK2_H
#include <winsock2.h>
#else
#include <sys/socket.h>
#ifdef HAVE_UNIX_DOMAIN_SOCKETS
#include <sys/un.h>
#endif
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#endif

#if defined (HAVE_UNIX_DOMAIN_SOCKETS) && !defined (SUN_LEN)
#define SUN_LEN(ptr) ((size_t) (((struct sockaddr_un *) 0)->sun_path) \
		      + strlen ((ptr)->sun_path))
#endif

/* The largest possible socket address.  Wrapping it in a union guarantees
   that the compiler will make it suitably aligned.  */
typedef union
{
  struct sockaddr     sockaddr;
  struct sockaddr_in  sockaddr_in;

#ifdef HAVE_UNIX_DOMAIN_SOCKETS
  struct sockaddr_un  sockaddr_un;
#endif
#ifdef HAVE_IPV6
  struct sockaddr_in6 sockaddr_in6;
#endif
} scm_t_max_sockaddr;


/* Maximum size of a socket address.  */
#define MAX_ADDR_SIZE   (sizeof (scm_t_max_sockaddr))


#ifndef HAVE_INET_ATON
/* for our definition in inet_aton.c, not usually needed.  */
extern int inet_aton ();
#endif

/* Byte Strings and IPv4 Addresses
 *  While IPv4 addresses may originally have been specified with "pure"
 * numerical values we do a slight historical revision and see: in the
 * 70s everything was just a number as the C domain wasn't recognized
 * as such. To comply with manuals, user expectations and else we
 * represent addresses as sequences of bytes in host or network byte
 * order as specified by the actual system routines.
 *
 * Support for the misguided IPv6 has been dropped.
 */

SCM_DEFINE (scm_inet_aton, "inet-aton", 1, 0, 0, 
            (SCM address),
	    "Convert an IPv4 Internet address from printable string\n"
	    "(dotted decimal notation) to a byte string.")
#define FUNC_NAME s_scm_inet_aton
{
  reference_parameter (address);
  SCM b_address;

  struct in_addr soka;
  char *c_address;
  int rv;

  b_address = scm_to_sen (address);
  reference_parameter (b_address);
  c_address = BYTE_STRING_DATA (b_address);
  rv = inet_aton (c_address, &soka);

  dereference_parameter (address);
  dereference_parameter (b_address);
  if (rv == 0)
    SCM_MISC_ERROR ("bad address", SCM_EOL);
  return scm_make_byte_string_with_copy (&(soka.s_addr), 4);
}
#undef FUNC_NAME


SCM_DEFINE (scm_inet_ntoa, "inet-ntoa", 1, 0, 0, 
            (SCM inetid),
	    "Convert an IPv4 Internet address to a printable\n"
	    "(dotted decimal notation) string.")
#define FUNC_NAME s_scm_inet_ntoa
{
  reference_parameter (inetid);

  if (!(BYTE_STRING_P (inetid) && (BYTE_STRING_LENGTH (inetid) == 4))) {
    scm_wrong_type_arg (FUNC_NAME, 1, inetid);
  }
  struct in_addr addr;
  char *s;
  SCM answer;
  addr.s_addr = *((uint32_t*) BYTE_STRING_DATA (inetid));
  s = inet_ntoa (addr);
  answer = scm_from_sen (s, strlen (s));

  dereference_parameter (inetid);
  return answer;
}
#undef FUNC_NAME

#ifdef HAVE_INET_NETOF
SCM_DEFINE (scm_inet_netof, "inet-netof", 1, 0, 0,   // deprecated
            (SCM address),
	    "Return the network number part of the given IPv4\n"
	    "Internet address.")
#define FUNC_NAME s_scm_inet_netof
{
  reference_parameter (address);
  if (!(BYTE_STRING_P (address) && (BYTE_STRING_LENGTH (address) == 4))) {
    scm_wrong_type_arg (FUNC_NAME, 1, address);
  }

  struct in_addr addr;
  addr.s_addr = *((uint32_t*) BYTE_STRING_DATA (address));

  dereference_parameter (address);
  unsigned int res = inet_netof (addr);
  return scm_make_byte_string_with_copy (&res, 4);
}
#undef FUNC_NAME
#endif

// Deprecation
/* #ifdef HAVE_INET_LNAOF */
/* SCM_DEFINE (scm_lnaof, "inet-lnaof", 1, 0, 0,  */
/*             (SCM address), */
/* 	    "Return the local-address-with-network part of the given\n" */
/* 	    "IPv4 Internet address, using the obsolete class A/B/C system.\n" */
/* 	    "E.g.,\n\n" */
/* 	    "@lisp\n" */
/* 	    "(inet-lnaof 2130706433) @result{} 1\n" */
/* 	    "@end lisp") */
/* #define FUNC_NAME s_scm_lnaof */
/* { */
/*   reference_parameter (address); */

/*   struct in_addr addr; */
/*   addr.s_addr = htonl (SCM_NUM2ULONG (1, address)); */

/*   dereference_parameter (address); */
/*   return scm_from_ulong (inet_lnaof (addr)); */
/* } */
/* #undef FUNC_NAME */
/* #endif */

// Deprication
/* #ifdef HAVE_INET_MAKEADDR */
/* SCM_DEFINE (scm_inet_makeaddr, "inet-makeaddr", 2, 0, 0, */
/*             (SCM net, SCM lna), */
/* 	    "Make an IPv4 Internet address by combining the network number\n" */
/* 	    "@var{net} with the local-address-within-network number\n" */
/* 	    "@var{lna}.  E.g.,\n\n" */
/* 	    "@lisp\n" */
/* 	    "(inet-makeaddr 127 1) @result{} 2130706433\n" */
/* 	    "@end lisp") */
/* #define FUNC_NAME s_scm_inet_makeaddr */
/* { */
/*   reference_parameter (net); */
/*   reference_parameter (lna); */

/*   struct in_addr addr; */
/*   unsigned long netnum; */
/*   unsigned long lnanum; */

/*   netnum = SCM_NUM2ULONG (1, net); */
/*   lnanum = SCM_NUM2ULONG (2, lna); */
/*   addr = inet_makeaddr (netnum, lnanum); */

/*   dereference_parameter (net); */
/*   dereference_parameter (lna); */
/*   return scm_from_ulong (ntohl (addr.s_addr)); */
/* } */
/* #undef FUNC_NAME */
/* #endif */


SCM_SYMBOL (sym_socket, "socket");

#define SCM_SOCK_FD_TO_PORT(fd) scm_fdes_to_port (fd, "r+0", sym_socket)

SCM_DEFINE (scm_socket, "socket", 3, 0, 0,
            (SCM family, SCM style, SCM proto),
	    "Return a new socket port of the type specified by @var{family},\n"
	    "@var{style} and @var{proto}.  All three parameters are\n"
	    "integers.  Supported values for @var{family} are\n"
	    "@code{AF_UNIX}, @code{AF_INET} and -@code{AF_INET6}-.\n"
	    "Typical values for @var{style} are @code{SOCK_STREAM},\n"
	    "@code{SOCK_DGRAM} and @code{SOCK_RAW}.\n\n"
	    "@var{proto} can be obtained from a protocol name using\n"
	    "@code{getprotobyname}.  A value of zero specifies the default\n"
	    "protocol, which is usually right.\n\n"
	    "A single socket port cannot by used for communication until it\n"
	    "has been connected to another socket.")
#define FUNC_NAME s_scm_socket
{
  reference_parameter (family);
  reference_parameter (style);
  reference_parameter (proto);

  int fd;

  fd = socket (scm_to_int (family),
	       scm_to_int (style),
	       scm_to_int (proto));
  if (fd == -1)
    SCM_SYSERROR;

  dereference_parameter (family);
  dereference_parameter (style);
  dereference_parameter (proto);
  return SCM_SOCK_FD_TO_PORT (fd);
}
#undef FUNC_NAME

#ifdef HAVE_SOCKETPAIR
SCM_DEFINE (scm_socketpair, "socketpair", 3, 0, 0,
            (SCM family, SCM style, SCM proto),
	    "Return a pair of connected (but unnamed) socket ports of the\n"
	    "type specified by @var{family}, @var{style} and @var{proto}.\n"
	    "Many systems support only socket pairs of the @code{AF_UNIX}\n"
	    "family.  Zero is likely to be the only meaningful value for\n"
	    "@var{proto}.")
#define FUNC_NAME s_scm_socketpair
{
  reference_parameter (family);
  reference_parameter (style);
  reference_parameter (proto);

  int fam;
  int fd[2];

  fam = scm_to_int (family);

  if (socketpair (fam, scm_to_int (style), scm_to_int (proto), fd) == -1)
    SCM_SYSERROR;

  dereference_parameter (family);
  dereference_parameter (style);
  dereference_parameter (proto);
  return scm_cons (SCM_SOCK_FD_TO_PORT (fd[0]), SCM_SOCK_FD_TO_PORT (fd[1]));
}
#undef FUNC_NAME
#endif

/* Possible results for `getsockopt ()'.  Wrapping it into a union guarantees
   suitable alignment.  */
typedef union
{
#ifdef HAVE_STRUCT_LINGER
  struct linger linger;
#endif
  size_t size;
  int    integer;
} scm_t_getsockopt_result;

SCM_DEFINE (scm_getsockopt, "getsockopt", 3, 0, 0,  // TODO Review socket options please
            (SCM sock_argument, SCM level, SCM optname), "")
#define FUNC_NAME s_scm_getsockopt
{
  reference_parameter (sock_argument);
  reference_parameter (level);
  reference_parameter (optname);

  int fd;
  /* size of optval is the largest supported option.  */
  scm_t_getsockopt_result optval;
  socklen_t optlen = sizeof (optval);
  int ilevel;
  int ioptname;

  SCM sock = SCM_COERCE_OUTPORT (sock_argument);
  reference_parameter (sock);
  SCM_VALIDATE_OPFPORT (1, sock);
  ilevel = scm_to_int (level);
  ioptname = scm_to_int (optname);

  fd = SCM_FPORT_FDES (sock);
  if (getsockopt (fd, ilevel, ioptname, (void *) &optval, &optlen) == -1)
    SCM_SYSERROR;

  dereference_parameter (sock);
  dereference_parameter (sock_argument);
  dereference_parameter (level);
  dereference_parameter (optname);
  if (ilevel == SOL_SOCKET)
    {
#ifdef SO_LINGER
      if (ioptname == SO_LINGER)
	{
#ifdef HAVE_STRUCT_LINGER
	  struct linger *ling = (struct linger *) &optval;

	  return scm_cons (scm_from_long (ling->l_onoff),
			   scm_from_long (ling->l_linger));
#else
	  return scm_cons (scm_from_long (*(int *) &optval),
			   scm_from_int (0));
#endif
	}
      else
#endif
	if (0
#ifdef SO_SNDBUF
	    || ioptname == SO_SNDBUF
#endif
#ifdef SO_RCVBUF
	    || ioptname == SO_RCVBUF
#endif
	    )
	  {
	    return scm_from_size_t (*(size_t *) &optval);
	  }
    }
  return scm_from_int (*(int *) &optval);
}
#undef FUNC_NAME

SCM_DEFINE (scm_setsockopt, "setsockopt", 4, 0, 0, // degration TODO Review socket options please
            (SCM sock_argument, SCM level, SCM optname, SCM value), "")
#define FUNC_NAME s_scm_setsockopt
{
  reference_parameter (sock_argument);
  reference_parameter (level);
  reference_parameter (optname);
  reference_parameter (value);

  int fd;

  int opt_int;
#ifdef HAVE_STRUCT_LINGER
  struct linger opt_linger;
#endif

#if HAVE_STRUCT_IP_MREQ
  struct ip_mreq opt_mreq;
#endif

  const void *optval = NULL;
  socklen_t optlen = 0;

  int ilevel, ioptname;

  SCM sock = SCM_COERCE_OUTPORT (sock_argument);
  reference_parameter (sock);

  SCM_VALIDATE_OPFPORT (1, sock);
  ilevel = scm_to_int (level);
  ioptname = scm_to_int (optname);

  fd = SCM_FPORT_FDES (sock);
  
  if (ilevel == SOL_SOCKET)
    {
#ifdef SO_LINGER
      if (ioptname == SO_LINGER)
	{
#ifdef HAVE_STRUCT_LINGER
	  SCM_ASSERT (scm_is_pair (value), value, SCM_ARG4, FUNC_NAME);
	  opt_linger.l_onoff = scm_to_int (TMP_CAR (value));
	  opt_linger.l_linger = scm_to_int (TMP_CDR (value));
	  optlen = sizeof (struct linger);
	  optval = &opt_linger;
#else
	  SCM_ASSERT (scm_is_pair (value), value, SCM_ARG4, FUNC_NAME);
	  opt_int = scm_to_int (TMP_CAR (value));
	  /* timeout is ignored, but may as well validate it.  */
	  scm_to_int (TMP_CDR (value));
	  optlen = sizeof (int);
	  optval = &opt_int;
#endif
	}
      else
#endif
	if (0
#ifdef SO_SNDBUF
	    || ioptname == SO_SNDBUF
#endif
#ifdef SO_RCVBUF
	    || ioptname == SO_RCVBUF
#endif
	    )
	  {
	    opt_int = scm_to_int (value);
	    optlen = sizeof (size_t);
	    optval = &opt_int;
	  }
    }

#if HAVE_STRUCT_IP_MREQ
  if (ilevel == IPPROTO_IP &&
      (ioptname == IP_ADD_MEMBERSHIP || ioptname == IP_DROP_MEMBERSHIP))
    {
      /* Fourth argument must be a pair of addresses. */
      SCM_ASSERT (scm_is_pair (value), value, SCM_ARG4, FUNC_NAME);
      opt_mreq.imr_multiaddr.s_addr = htonl (scm_to_ulong (TMP_CAR (value)));
      opt_mreq.imr_interface.s_addr = htonl (scm_to_ulong (TMP_CDR (value)));
      optlen = sizeof (opt_mreq);
      optval = &opt_mreq;
    }
#endif

  if (optval == NULL)
    {
      /* Most options take an int.  */
      opt_int = scm_to_int (value);
      optlen = sizeof (int);
      optval = &opt_int;
    }

  if (setsockopt (fd, ilevel, ioptname, optval, optlen) == -1)
    SCM_SYSERROR;


  dereference_parameter (sock);
  dereference_parameter (sock_argument);
  dereference_parameter (level);
  dereference_parameter (optname);
  dereference_parameter (value);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_shutdown, "shutdown", 2, 0, 0,
          (SCM sock, SCM how),
	    "Sockets can be closed simply by using @code{close-port}. The\n"
	    "@code{shutdown} procedure allows reception or transmission on a\n"
	    "connection to be shut down individually, according to the parameter\n"
	    "@var{how}:\n\n"
	    "@table @asis\n"
	    "@item 0\n"
	    "Stop receiving data for this socket.  If further data arrives,  reject it.\n"
	    "@item 1\n"
	    "Stop trying to transmit data from this socket.  Discard any\n"
	    "data waiting to be sent.  Stop looking for acknowledgement of\n"
	    "data already sent; don't retransmit it if it is lost.\n"
	    "@item 2\n"
	    "Stop both reception and transmission.\n"
	    "@end table\n\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_shutdown
{
  reference_parameter (sock);
  reference_parameter (how);

  int fd;
  sock = SCM_COERCE_OUTPORT (sock);
  SCM_VALIDATE_OPFPORT (1, sock);
  fd = SCM_FPORT_FDES (sock);
  if (shutdown (fd, scm_to_signed_integer (how, 0, 2)) == -1)
    SCM_SYSERROR;

  dereference_parameter (sock);
  dereference_parameter (how);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

/* convert fam/address/args into a sockaddr of the appropriate type.
   args is modified by removing the arguments actually used.
   which_arg and proc are used when reporting errors:
   which_arg is the position of address in the original argument list.
   proc is the name of the original procedure.
   size returns the size of the structure allocated.  */

static struct sockaddr *
scm_fill_sockaddr (int fam, SCM address, SCM *args, int which_arg,
		   const char *proc, size_t *size)
#define FUNC_NAME proc
{
  reference_parameter (address);
  reference_parameter (*args);

  switch (fam)
    {
    case AF_INET:
      {
	struct sockaddr_in *soka;
	unsigned long addr;
	int port;

        if (!(BYTE_STRING_P (address) && (BYTE_STRING_LENGTH (address) == 4))) {
          scm_wrong_type_arg (FUNC_NAME, 2, address);
        }
	SCM_VALIDATE_CONS (which_arg + 1, *args);
	port = scm_to_int (TMP_CAR (*args));
	*args = FLY_CDR (*args);
	soka = (struct sockaddr_in *) scm_malloc (sizeof (struct sockaddr_in));

#if HAVE_STRUCT_SOCKADDR_SIN_LEN
	soka->sin_len = sizeof (struct sockaddr_in);
#endif
	soka->sin_family = AF_INET;
	soka->sin_addr.s_addr = *((uint32_t*) (BYTE_STRING_DATA (address)));
	soka->sin_port = port;
	*size = sizeof (struct sockaddr_in);

        dereference_parameter (address);
        dereference_parameter (*args);
	return (struct sockaddr *) soka;
      }
#ifdef HAVE_UNIX_DOMAIN_SOCKETS
    case AF_UNIX:
      {
	struct sockaddr_un *soka;
	int addr_size;
	char *c_address;

	SCM b_address = scm_to_sen (address);
        reference_parameter (b_address);
	c_address = BYTE_STRING_DATA (b_address);

	/* the static buffer size in sockaddr_un seems to be arbitrary
	   and not necessarily a hard limit.  e.g., the glibc manual
	   suggests it may be possible to declare it size 0.  let's
	   ignore it.  if the O/S doesn't like the size it will cause
	   connect/bind etc., to fail.  sun_path is always the last
	   member of the structure.  */
	addr_size = sizeof (struct sockaddr_un)
	  + max (0, strlen (c_address) + 1 - (sizeof soka->sun_path));
	soka = (struct sockaddr_un *) scm_malloc (addr_size);
	memset (soka, 0, addr_size);  /* for sun_len: see sin_len above. */
	soka->sun_family = AF_UNIX;
	strcpy (soka->sun_path, c_address);
	*size = SUN_LEN (soka);

        dereference_parameter (b_address);
        dereference_parameter (address);
        dereference_parameter (*args);
	return (struct sockaddr *) soka;
      }
#endif
    default:
      scm_out_of_range (proc, scm_from_int (fam));
    }
}
#undef FUNC_NAME

SCM_DEFINE (scm_connect, "connect", 2, 1, 1,
            (SCM sock_argument, SCM fam_or_sockaddr, SCM address, SCM args),
	    "Initiate a connection from a socket using a specified address\n"
	    "family to the address\n"
	    "specified by @var{address} and possibly @var{args}.\n"
	    "The format required for @var{address}\n"
	    "and @var{args} depends on the family of the socket.\n\n"
	    "For a socket of family @code{AF_UNIX},\n"
	    "only @var{address} is specified and must be a string with the\n"
	    "filename where the socket is to be created.\n\n"
	    "For a socket of family @code{AF_INET},\n"
	    "@var{address} must be an integer IPv4 host address and\n"
	    "@var{args} must be a single integer port number.\n\n"
	    "For a socket of family @code{AF_INET6},\n"
	    "@var{address} must be an integer IPv6 host address and\n"
	    "@var{args} may be up to three integers:\n"
	    "port [flowinfo] [scope_id],\n"
	    "where flowinfo and scope_id default to zero.\n\n"
	    "Alternatively, the second argument can be a socket address object "
	    "as returned by @code{make-socket-address}, in which case the "
	    "no additional arguments should be passed.\n\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_connect
{
  reference_parameter (sock_argument);
  reference_parameter (fam_or_sockaddr);
  reference_parameter (address);
  // RC: args pass through to /scm_fill_sockaddr/

  int fd;
  struct sockaddr *soka;
  size_t size;

  SCM sock = SCM_COERCE_OUTPORT (sock_argument);
  reference_parameter (sock);
  SCM_VALIDATE_OPFPORT (1, sock);
  fd = SCM_FPORT_FDES (sock);

  if (address == SCM_UNDEFINED) {
    /* No third argument was passed to FAM_OR_SOCKADDR must actually be a
       `socket address' object.  */
    soka = scm_to_sockaddr (fam_or_sockaddr, &size);
    dereference_return_value (args);

  } else {
    soka = scm_fill_sockaddr (scm_to_int (fam_or_sockaddr), address,
			      &args, 3, FUNC_NAME, &size);
  }

  if (connect (fd, soka, size) == -1)
    {
      int save_errno = errno;

      free (soka);
      errno = save_errno;
      // SCM_SYSERROR; Perhaps we should export errno to Scheme???
      dereference_parameter (sock);
      dereference_parameter (sock_argument);
      dereference_parameter (fam_or_sockaddr);
      dereference_parameter (address);
      return SCM_BOOL_F;
    }
  free (soka);

  dereference_parameter (sock);
  dereference_parameter (sock_argument);
  dereference_parameter (fam_or_sockaddr);
  dereference_parameter (address);
  return SCM_BOOL_T;
}
#undef FUNC_NAME

SCM_DEFINE (scm_bind, "bind", 2, 1, 1,
            (SCM sock_argument, SCM fam_or_sockaddr, SCM address, SCM args),
	    "Assign an address to the socket port @var{sock}.\n"
	    "Generally this only needs to be done for server sockets,\n"
	    "so they know where to look for incoming connections.  A socket\n"
	    "without an address will be assigned one automatically when it\n"
	    "starts communicating.\n\n"
	    "The format of @var{address} and @var{args} depends\n"
	    "on the family of the socket.\n\n"
	    "For a socket of family @code{AF_UNIX}, only @var{address}\n"
	    "is specified and must be a string with the filename where\n"
	    "the socket is to be created.\n\n"
	    "For a socket of family @code{AF_INET}, @var{address}\n"
	    "must be an integer IPv4 address and @var{args}\n"
	    "must be a single integer port number.\n\n"
	    "The values of the following variables can also be used for\n"
	    "@var{address}:\n\n"
	    "@defvar INADDR_ANY\n"
	    "Allow connections from any address.\n"
	    "@end defvar\n\n"
	    "@defvar INADDR_LOOPBACK\n"
	    "The address of the local host using the loopback device.\n"
	    "@end defvar\n\n"
	    "@defvar INADDR_BROADCAST\n"
	    "The broadcast address on the local network.\n"
	    "@end defvar\n\n"
	    "@defvar INADDR_NONE\n"
	    "No address.\n"
	    "@end defvar\n\n"
	    "For a socket of family @code{AF_INET6}, @var{address}\n"
	    "must be an integer IPv6 address and @var{args}\n"
	    "may be up to three integers:\n"
	    "port [flowinfo] [scope_id],\n"
	    "where flowinfo and scope_id default to zero.\n\n"
	    "Alternatively, the second argument can be a socket address object "
	    "as returned by @code{make-socket-address}, in which case the "
	    "no additional arguments should be passed.\n\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_bind
{
  reference_parameter (sock_argument);
  reference_parameter (fam_or_sockaddr);
  reference_parameter (address);
  // RC: args pass through to /scm_fill_sockaddr/

  struct sockaddr *soka;
  size_t size;
  int fd;

  SCM sock = SCM_COERCE_OUTPORT (sock_argument);
  reference_parameter (sock);
  SCM_VALIDATE_OPFPORT (1, sock);
  fd = SCM_FPORT_FDES (sock);

  if (address == SCM_UNDEFINED) {
    /* No third argument was passed to FAM_OR_SOCKADDR must actually be a
       `socket address' object.  */
    soka = scm_to_sockaddr (fam_or_sockaddr, &size);
    dereference_return_value (args);

  } else {
    soka = scm_fill_sockaddr (scm_to_int (fam_or_sockaddr), address,
			      &args, 3, FUNC_NAME, &size);
  }

  if (bind (fd, soka, size) == -1)
  {
    int save_errno = errno;

    free (soka);
    errno = save_errno;
    SCM_SYSERROR;
  }
  free (soka);

  dereference_parameter (sock);
  dereference_parameter (sock_argument);
  dereference_parameter (fam_or_sockaddr);
  dereference_parameter (address);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

SCM_DEFINE (scm_listen, "listen", 2, 0, 0,
            (SCM sock_argument, SCM backlog),
	    "Enable @var{sock} to accept connection\n"
	    "requests.  @var{backlog} is an integer specifying\n"
	    "the maximum length of the queue for pending connections.\n"
	    "If the queue fills, new clients will fail to connect until\n"
	    "the server calls @code{accept} to accept a connection from\n"
	    "the queue.\n\n"
	    "The return value is unspecified.")
#define FUNC_NAME s_scm_listen
{
  reference_parameter (sock_argument);
  reference_parameter (backlog);

  int fd;
  SCM sock = SCM_COERCE_OUTPORT (sock_argument);
  reference_parameter (sock);
  SCM_VALIDATE_OPFPORT (1, sock);

  fd = SCM_FPORT_FDES (sock);
  if (listen (fd, scm_to_int (backlog)) == -1)
    SCM_SYSERROR;

  dereference_parameter (sock);
  dereference_parameter (sock_argument);
  dereference_parameter (backlog);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

/* Put the components of a sockaddr into a new SCM vector.  */
static SCM_C_INLINE_KEYWORD SCM
_scm_from_sockaddr (const scm_t_max_sockaddr *address, unsigned addr_size,
		    const char *proc)
{
  SCM result = SCM_EOL;
  short int fam = ((struct sockaddr *) address)->sa_family;

  switch (fam)
    {
    case AF_INET:
      {
	const struct sockaddr_in *nad = (struct sockaddr_in *) address;

	result = scm_c_make_vector (3, SCM_UNSPECIFIED);
        reference_parameter (result);

	SCM_SIMPLE_VECTOR_SET(result, 0,
			      scm_from_short (fam));
	SCM_SIMPLE_VECTOR_SET(result, 1,
			      scm_make_byte_string_with_copy (&(nad->sin_addr.s_addr), 4));
	SCM_SIMPLE_VECTOR_SET(result, 2,
			      scm_from_ushort (nad->sin_port));
      }
      break;
#ifdef HAVE_UNIX_DOMAIN_SOCKETS
    case AF_UNIX:
      {
	const struct sockaddr_un *nad = (struct sockaddr_un *) address;

	result = scm_c_make_vector (2, SCM_UNSPECIFIED);
        reference_parameter (result);

	SCM_SIMPLE_VECTOR_SET(result, 0, scm_from_short (fam));
	/* When addr_size is not enough to cover sun_path, do not try
	   to access it. */
	if (addr_size <= offsetof (struct sockaddr_un, sun_path))
	  SCM_SIMPLE_VECTOR_SET(result, 1, SCM_BOOL_F);
	else
	  SCM_SIMPLE_VECTOR_SET(result, 1, scm_from_sen (nad->sun_path, strlen (nad->sun_path)));
      }
      break;
#endif
    default:
      result = SCM_UNSPECIFIED;
      scm_misc_error (proc, "unrecognised address family: ~A",
		      scm_list_1 (scm_from_int (fam)));

    }
  return (rc_return_stack_reference (result));
}

/* The publicly-visible function.  Return a Scheme object representing
   ADDRESS, an address of ADDR_SIZE bytes.  */
SCM
scm_from_sockaddr (const struct sockaddr *address, unsigned addr_size)
{
  return (_scm_from_sockaddr ((scm_t_max_sockaddr *) address,
			      addr_size, "scm_from_sockaddr"));
}

/* Convert ADDRESS, an address object returned by either
   `scm_from_sockaddr ()' or `scm_make_socket_address ()', into its C
   representation.  On success, a non-NULL pointer is returned and
   ADDRESS_SIZE is updated to the actual size (in bytes) of the returned
   address.  The result must eventually be freed using `free ()'.  */
struct sockaddr *
scm_to_sockaddr (SCM address, size_t *address_size)
#define FUNC_NAME "scm_to_sockaddr"
{
  reference_parameter (address);

  short int family;
  struct sockaddr *c_address = NULL;

  SCM_VALIDATE_VECTOR (1, address);

  *address_size = 0;
  family = scm_to_short (TMP_SIMPLE_VECTOR_REF (address, 0));

  switch (family) {
  case AF_INET:
    {
      if (SCM_SIMPLE_VECTOR_LENGTH (address) != 3) {
        scm_misc_error (FUNC_NAME,
                        "invalid inet address representation: ~A",
                        scm_list_1 (address));
      } else {
        SCM addr = SCM_SIMPLE_VECTOR_REF (address, 1);
        SCM port = SCM_SIMPLE_VECTOR_REF (address, 2);
        if (!(BYTE_STRING_P (addr) && (BYTE_STRING_LENGTH (addr) == 4))) {
          scm_wrong_type_arg (FUNC_NAME, 1, address);
        }

        struct sockaddr_in c_inet;
        
        c_inet.sin_addr.s_addr = *((uint32_t*) BYTE_STRING_DATA (addr));
        c_inet.sin_port = scm_to_ushort (port);
        c_inet.sin_family = AF_INET;

        *address_size = sizeof (c_inet);
        c_address = scm_malloc (sizeof (c_inet));
        memcpy (c_address, &c_inet, sizeof (c_inet));
      }

      break;
    }


#ifdef HAVE_UNIX_DOMAIN_SOCKETS
  case AF_UNIX:
    {
      if (SCM_SIMPLE_VECTOR_LENGTH (address) != 2) {
	  scm_misc_error (FUNC_NAME, "invalid unix address representation: ~A",
			  scm_list_1 (address));
      } else {
        SCM br_path;  // RC: data graph referenced
        size_t path_len = 0;

        br_path = TMP_SIMPLE_VECTOR_REF (address, 1);
        if ((!SCM_BRASSP (br_path)) && (br_path != SCM_BOOL_F)) {
          scm_misc_error (FUNC_NAME, "invalid unix address "
                          "path: ~A", scm_list_1 (br_path));
        } else {
          SCM path = SCM_UNSPECIFIED;
          struct sockaddr_un c_unix;

          if (br_path == SCM_BOOL_F) {
            path = SCM_BOOL_F;
            path_len = 0;
          } else {
            path = scm_to_sen (br_path);
            reference_parameter (path);
            path_len = BYTE_STRING_LENGTH (path);
          }

#ifdef UNIX_PATH_MAX
          if (path_len > UNIX_PATH_MAX)
#else
          if (path_len > sizeof (c_unix.sun_path))
#endif
            scm_misc_error (FUNC_NAME, "unix address path "
                            "too long: ~A", scm_list_1 (path));
          else {
            if (path_len) {
              memcpy (c_unix.sun_path, BYTE_STRING_DATA (path), path_len);
                      
              /* Sanity check.  */
              if (strlen (c_unix.sun_path) != path_len)
                scm_misc_error (FUNC_NAME, "unix address path "
                                "contains nul characters: ~A",
                                scm_list_1 (path));
            } else {
              c_unix.sun_path[0] = '\0';
            }
            dereference_parameter (path);

            c_unix.sun_family = AF_UNIX;

            *address_size = SUN_LEN (&c_unix);
            c_address = scm_malloc (sizeof (c_unix));
            memcpy (c_address, &c_unix, sizeof (c_unix));
          }
        }
      }
      
      break;
    }
#endif

  default:
    scm_misc_error (FUNC_NAME, "unrecognised address family: ~A",
                    scm_list_1 (scm_from_ushort (family)));
  }

  dereference_parameter (address);
  return c_address;
}
#undef FUNC_NAME


/* Return a newly-allocated `sockaddr' structure that reflects ADDRESS, being
   an address of family FAMILY, with the family-specific parameters ARGS (see
   the description of `connect' for details).  The returned structure may be
   freed using `free ()'.  */
struct sockaddr *
scm_c_make_socket_address (SCM family, SCM address, SCM args,
			   size_t *address_size)
{
  struct sockaddr *soka;

  soka = scm_fill_sockaddr (scm_to_ushort (family), address, &args, 1,
			    "scm_c_make_socket_address", address_size);

  return soka;
}

SCM_DEFINE (scm_make_socket_address, "make-socket-address", 2, 0, 1,
	    (SCM family, SCM address, SCM args),
	    "Return a Scheme address object that reflects @var{address}, "
	    "being an address of family @var{family}, with the "
	    "family-specific parameters @var{args} (see the description of "
	    "@code{connect} for details).")
#define FUNC_NAME s_scm_make_socket_address
{
  reference_parameter (family);
  reference_parameter (address);
  reference_parameter (args);

  SCM result = SCM_BOOL_F;
  struct sockaddr *c_address;
  size_t c_address_size;

  c_address = scm_c_make_socket_address (family, address, args,
					 &c_address_size);
  if (c_address != NULL)
    {
      result = scm_from_sockaddr (c_address, c_address_size);
      free (c_address);
    }

  reference_count_return (result);
  dereference_parameter (family);
  dereference_parameter (address);
  dereference_parameter (args);
  return result;
}
#undef FUNC_NAME


SCM_DEFINE (scm_accept, "accept", 1, 0, 0, 
            (SCM sock),
	    "Accept a connection on a bound, listening socket.\n"
	    "If there\n"
	    "are no pending connections in the queue, wait until\n"
	    "one is available unless the non-blocking option has been\n"
	    "set on the socket.\n\n"
	    "The return value is a\n"
	    "pair in which the @emph{car} is a new socket port for the\n"
	    "connection and\n"
	    "the @emph{cdr} is an object with address information about the\n"
	    "client which initiated the connection.\n\n"
	    "@var{sock} does not become part of the\n"
	    "connection and will continue to accept new requests.")
#define FUNC_NAME s_scm_accept
{
  reference_parameter (sock);

  int fd, selected;
  int newfd;
  SCM address;
  SCM newsock;
  SELECT_TYPE readfds, exceptfds;
  socklen_t addr_size = MAX_ADDR_SIZE;
  scm_t_max_sockaddr addr;

  SCM_VALIDATE_OPFPORT (1, sock);
  fd = SCM_FPORT_FDES (sock);

  FD_ZERO (&readfds);
  FD_ZERO (&exceptfds);
  FD_SET (fd, &readfds);
  FD_SET (fd, &exceptfds);

  // former sleep pipe control TODO purge please
  selected = scm_std_select (fd + 1, &readfds, NULL, &exceptfds, NULL);
  if (selected < 0)
    SCM_SYSERROR;

  newfd = accept (fd, (struct sockaddr *) &addr, &addr_size);
  if (newfd == -1)
    SCM_SYSERROR;
  newsock = SCM_SOCK_FD_TO_PORT (newfd);
  address = _scm_from_sockaddr (&addr, addr_size,
				FUNC_NAME);

  reference_count_return (newsock);
  reference_count_return (address);
  dereference_parameter (sock);
  return scm_cons (newsock, address);
}
#undef FUNC_NAME

SCM_DEFINE (scm_getsockname, "getsockname", 1, 0, 0, 
            (SCM sock),
	    "Return the address of @var{sock}, in the same form as the\n"
	    "object returned by @code{accept}.  On many systems the address\n"
	    "of a socket in the @code{AF_FILE} namespace cannot be read.")
#define FUNC_NAME s_scm_getsockname
{
  reference_parameter (sock);

  int fd;
  socklen_t addr_size = MAX_ADDR_SIZE;
  scm_t_max_sockaddr addr;

  SCM_VALIDATE_OPFPORT (1, sock);
  fd = SCM_FPORT_FDES (sock);
  if (getsockname (fd, (struct sockaddr *) &addr, &addr_size) == -1)
    SCM_SYSERROR;

  dereference_parameter (sock);
  return _scm_from_sockaddr (&addr, addr_size, FUNC_NAME);
}
#undef FUNC_NAME

SCM_DEFINE (scm_getpeername, "getpeername", 1, 0, 0, 
            (SCM sock),
	    "Return the address that @var{sock}\n"
	    "is connected to, in the same form as the object returned by\n"
	    "@code{accept}.  On many systems the address of a socket in the\n"
	    "@code{AF_FILE} namespace cannot be read.")
#define FUNC_NAME s_scm_getpeername
{
  reference_parameter (sock);

  int fd;
  socklen_t addr_size = MAX_ADDR_SIZE;
  scm_t_max_sockaddr addr;

  SCM_VALIDATE_OPFPORT (1, sock);
  fd = SCM_FPORT_FDES (sock);
  if (getpeername (fd, (struct sockaddr *) &addr, &addr_size) == -1)
    SCM_SYSERROR;

  dereference_parameter (sock);
  RC_RETURN (_scm_from_sockaddr (&addr, addr_size, FUNC_NAME));
}
#undef FUNC_NAME

SCM_DEFINE (scm_recv, "recv!", 2, 1, 0,
            (SCM sock, SCM buf, SCM flags),
	    "Receive data from a socket port.\n"
	    "@var{sock} must already\n"
	    "be bound to the address from which data is to be received.\n"
	    "@var{buf} is a string into which\n"
	    "the data will be written.  The size of @var{buf} limits\n"
	    "the amount of\n"
	    "data which can be received: in the case of packet\n"
	    "protocols, if a packet larger than this limit is encountered\n"
	    "then some data\n"
	    "will be irrevocably lost.\n\n"
	    "The optional @var{flags} argument is a value or\n"
	    "bitwise OR of MSG_OOB, MSG_PEEK, MSG_DONTROUTE etc.\n\n"
	    "The value returned is the number of bytes read from the\n"
	    "socket.\n\n"
	    "Note that the data is read directly from the socket file\n"
	    "descriptor:\n"
	    "any unread buffered port data is ignored.")
#define FUNC_NAME s_scm_recv
{
  reference_parameter (sock);
  reference_parameter (buf);
  reference_parameter (flags);

  int rv;
  int fd;
  int flg;
  char *dest;
  size_t len;

  SCM_VALIDATE_OPFPORT (1, sock);
  if (!(BYTE_STRING_P (buf))) {
    scm_wrong_type_arg (FUNC_NAME, 2, buf);
  }
  if (SCM_UNBNDP (flags))
    flg = 0;
  else
    flg = scm_to_int (flags);
  fd = SCM_FPORT_FDES (sock);

  len =  BYTE_STRING_LENGTH (buf);
  dest = BYTE_STRING_DATA (buf);
  SCM_SYSCALL (rv = recv (fd, dest, len, flg));

  if (rv == -1)
    SCM_SYSERROR;

  dereference_parameter (sock);
  dereference_parameter (buf);
  dereference_parameter (flags);
  return scm_from_int (rv);
}
#undef FUNC_NAME

SCM_DEFINE (scm_send, "send", 2, 1, 0,  // XXX func name gen
            (SCM sock, SCM message, SCM flags),
	    "Transmit the string @var{message} on a socket port @var{sock}.\n"
	    "@var{sock} must already be bound to a destination address.  The\n"
	    "value returned is the number of bytes transmitted --\n"
	    "it's possible for\n"
	    "this to be less than the length of @var{message}\n"
	    "if the socket is\n"
	    "set to be non-blocking.  The optional @var{flags} argument\n"
	    "is a value or\n"
	    "bitwise OR of MSG_OOB, MSG_PEEK, MSG_DONTROUTE etc.\n\n"
	    "Note that the data is written directly to the socket\n"
	    "file descriptor:\n"
	    "any unflushed buffered port data is ignored.")
#define FUNC_NAME s_scm_send
{
  reference_parameter (sock);
  reference_parameter (message);
  reference_parameter (flags);

  int rv;
  int fd;
  int flg;
  const char *src;
  size_t len;

  SCM_VALIDATE_OPFPORT (1, sock);
  if (!(BYTE_STRING_P (message))) {
    scm_wrong_type_arg (FUNC_NAME, 2, message);
  }

  if (SCM_UNBNDP (flags))
    flg = 0;
  else
    flg = scm_to_int (flags);
  fd = SCM_FPORT_FDES (sock);

  len = BYTE_STRING_LENGTH (message);
  src = BYTE_STRING_DATA (message);
  SCM_SYSCALL (rv = send (fd, src, len, flg));

  if (rv == -1)
    SCM_SYSERROR;

  dereference_parameter (sock);
  dereference_parameter (message);
  dereference_parameter (flags);
  return scm_from_int (rv);
}
#undef FUNC_NAME

SCM_DEFINE (scm_recvfrom, "recvfrom!", 2, 1, 0,
            (SCM sock, SCM str, SCM flags),
	    "Receive data from socket port @var{sock} (which must be already\n"
	    "bound), returning the originating address as well as the data.\n"
	    "This is usually for use on datagram sockets, but can be used on\n"
	    "stream-oriented sockets too.\n"
	    "\n"
	    "The data received is stored in the given @var{str}, using\n"
	    "either the whole string or just the region between the optional\n"
	    "@var{start} and @var{end} positions.  The size of @var{str}\n"
	    "limits the amount of data which can be received.  For datagram\n"
	    "protocols, if a packet larger than this is received then excess\n"
	    "bytes are irrevocably lost.\n"
	    "\n"
	    "The return value is a pair.  The @code{car} is the number of\n"
	    "bytes read.  The @code{cdr} is a socket address object which is\n"
	    "where the data come from, or @code{#f} if the origin is\n"
	    "unknown.\n"
	    "\n"
	    "The optional @var{flags} argument is a or bitwise OR\n"
	    "(@code{logior}) of @code{MSG_OOB}, @code{MSG_PEEK},\n"
	    "@code{MSG_DONTROUTE} etc.\n"
	    "\n"
	    "Data is read directly from the socket file descriptor, any\n"
	    "buffered port data is ignored.\n"
	    "\n"
	    "On a GNU/Linux system @code{recvfrom!} is not multi-threading,\n"
	    "all threads stop while a @code{recvfrom!} call is in progress.\n"
	    "An application may need to use @code{select}, @code{O_NONBLOCK}\n"
	    "or @code{MSG_DONTWAIT} to avoid this.")
#define FUNC_NAME s_scm_recvfrom
{
  reference_parameter (sock);
  reference_parameter (str);
  reference_parameter (flags);

  int rv;
  int fd;
  int flg;
  char *buf;
  size_t len;
  SCM address;
  socklen_t addr_size = MAX_ADDR_SIZE;
  scm_t_max_sockaddr addr;

  SCM_VALIDATE_OPFPORT (1, sock);
  fd = SCM_FPORT_FDES (sock);
  
  if (!(BYTE_STRING_P (str))) {
    scm_wrong_type_arg (FUNC_NAME, 2, str);
  }

  if (SCM_UNBNDP (flags))
    flg = 0;
  else
    SCM_VALIDATE_ULONG_COPY (3, flags, flg);

  /* recvfrom will not necessarily return an address.  usually nothing
     is returned for stream sockets.  */
  buf = BYTE_STRING_DATA (str);
  len = BYTE_STRING_LENGTH (str);
  ((struct sockaddr *) &addr)->sa_family = AF_UNSPEC;
  SCM_SYSCALL (rv = recvfrom (fd, buf, len, flg,
			      (struct sockaddr *) &addr, &addr_size));

  if (rv == -1)
    SCM_SYSERROR;
  if (((struct sockaddr *) &addr)->sa_family != AF_UNSPEC)
    address = _scm_from_sockaddr (&addr, addr_size, FUNC_NAME);
  else
    address = SCM_BOOL_F;

  dereference_parameter (sock);
  dereference_parameter (str);
  dereference_parameter (flags);
  return scm_cons (scm_from_int (rv), address);
}
#undef FUNC_NAME

SCM_DEFINE (scm_sendto, "sendto", 3, 1, 1,
            (SCM sock_argument, SCM message, SCM fam_or_sockaddr,
             SCM address, SCM arguments_flags),
	    "Transmit the string @var{message} on the socket port\n"
	    "@var{sock}.  The\n"
	    "destination address is specified using the @var{fam},\n"
	    "@var{address} and\n"
	    "@var{args_and_flags} arguments, or just a socket address object "
	    "returned by @code{make-socket-address}, in a similar way to the\n"
	    "@code{connect} procedure.  @var{args_and_flags} contains\n"
	    "the usual connection arguments optionally followed by\n"
	    "a flags argument, which is a value or\n"
	    "bitwise OR of MSG_OOB, MSG_PEEK, MSG_DONTROUTE etc.\n\n"
	    "The value returned is the number of bytes transmitted --\n"
	    "it's possible for\n"
	    "this to be less than the length of @var{message} if the\n"
	    "socket is\n"
	    "set to be non-blocking.\n"
	    "Note that the data is written directly to the socket\n"
	    "file descriptor:\n"
	    "any unflushed buffered port data is ignored.")
#define FUNC_NAME s_scm_sendto
{
  reference_parameter (sock_argument);
  reference_parameter (message);
  reference_parameter (fam_or_sockaddr);
  reference_parameter (address);
  reference_parameter (arguments_flags);

  SCM args_and_flags = arguments_flags;
  int rv;
  int fd;
  int flg;
  struct sockaddr *soka;
  size_t size;

  SCM sock = SCM_COERCE_OUTPORT (sock_argument);
  reference_parameter (sock);
  SCM_VALIDATE_FPORT (1, sock);
  if (!(BYTE_STRING_P (message))) {
    scm_wrong_type_arg (FUNC_NAME, 2, message);
  }

  fd = SCM_FPORT_FDES (sock);

  if (!scm_is_number (fam_or_sockaddr)) {
    /* FAM_OR_SOCKADDR must actually be a `socket address' object.  This
       means that the following arguments, i.e. ADDRESS and those listed in
       ARGS_AND_FLAGS, are the `MSG_' flags.  */
    soka = scm_to_sockaddr (fam_or_sockaddr, &size);
    if (address != SCM_UNDEFINED)
      args_and_flags = scm_cons (address, args_and_flags);

  } else {
    soka = scm_fill_sockaddr (scm_to_int (fam_or_sockaddr), address,
			      &args_and_flags, 3, FUNC_NAME, &size);
  }
  reference_parameter (args_and_flags);

  if (scm_is_null (args_and_flags)) {
    flg = 0;
  } else {
    SCM_VALIDATE_CONS (5, args_and_flags);
    flg = SCM_NUM2ULONG (5, TMP_CAR (args_and_flags));
  }

  SCM_SYSCALL (rv = sendto (fd,
			    BYTE_STRING_DATA (message),
			    BYTE_STRING_LENGTH (message),
			    flg, soka, size));

  dereference_parameter (args_and_flags);
  dereference_parameter (sock_argument);
  dereference_parameter (sock);
  dereference_parameter (message);
  dereference_parameter (fam_or_sockaddr);
  dereference_parameter (address);
  dereference_parameter (arguments_flags);

  if (rv == -1) {
    int save_errno = errno;
    free (soka);
    errno = save_errno;
    SCM_SYSERROR;
  }
  free (soka);
  return scm_from_int (rv);
}
#undef FUNC_NAME



void
scm_init_socket ()
{
  /* protocol families.  */
#ifdef AF_UNSPEC
  scm_c_define ("AF_UNSPEC", scm_from_int (AF_UNSPEC));
#endif
#ifdef AF_UNIX
  scm_c_define ("AF_UNIX", scm_from_int (AF_UNIX));
#endif
#ifdef AF_INET
  scm_c_define ("AF_INET", scm_from_int (AF_INET));
#endif

#ifdef PF_UNSPEC
  scm_c_define ("PF_UNSPEC", scm_from_int (PF_UNSPEC));
#endif
#ifdef PF_UNIX
  scm_c_define ("PF_UNIX", scm_from_int (PF_UNIX));
#endif
#ifdef PF_INET
  scm_c_define ("PF_INET", scm_from_int (PF_INET));
#endif

  /* standard addresses.  */
#ifdef INADDR_ANY
  scm_c_define ("INADDR_ANY", scm_from_ulong (INADDR_ANY));
#endif
#ifdef INADDR_BROADCAST
  scm_c_define ("INADDR_BROADCAST", scm_from_ulong (INADDR_BROADCAST));
#endif
#ifdef INADDR_NONE
  scm_c_define ("INADDR_NONE", scm_from_ulong (INADDR_NONE));
#endif
#ifdef INADDR_LOOPBACK
  scm_c_define ("INADDR_LOOPBACK", scm_from_ulong (INADDR_LOOPBACK));
#endif

  /* socket types.

     SOCK_PACKET is deliberately omitted, the GNU/Linux socket(2) and
     packet(7) advise that it's obsolete and strongly deprecated.  */

#ifdef SOCK_STREAM
  scm_c_define ("SOCK_STREAM", scm_from_int (SOCK_STREAM));
#endif
#ifdef SOCK_DGRAM
  scm_c_define ("SOCK_DGRAM", scm_from_int (SOCK_DGRAM));
#endif
#ifdef SOCK_SEQPACKET
  scm_c_define ("SOCK_SEQPACKET", scm_from_int (SOCK_SEQPACKET));
#endif
#ifdef SOCK_RAW
  scm_c_define ("SOCK_RAW", scm_from_int (SOCK_RAW));
#endif
#ifdef SOCK_RDM
  scm_c_define ("SOCK_RDM", scm_from_int (SOCK_RDM));
#endif

  /* setsockopt level.

     SOL_IP, SOL_TCP and SOL_UDP are defined on gnu/linux, but not on for
     instance NetBSD.  We define IPPROTOs because that's what the posix spec
     shows in its example at

     http://www.opengroup.org/onlinepubs/007904975/functions/getsockopt.html
  */
#ifdef SOL_SOCKET
  scm_c_define ("SOL_SOCKET", scm_from_int (SOL_SOCKET));
#endif
#ifdef IPPROTO_IP
  scm_c_define ("IPPROTO_IP", scm_from_int (IPPROTO_IP));
#endif
#ifdef IPPROTO_TCP
  scm_c_define ("IPPROTO_TCP", scm_from_int (IPPROTO_TCP));
#endif
#ifdef IPPROTO_UDP
  scm_c_define ("IPPROTO_UDP", scm_from_int (IPPROTO_UDP));
#endif

  /* setsockopt names.  */
#ifdef SO_DEBUG
  scm_c_define ("SO_DEBUG", scm_from_int (SO_DEBUG));
#endif
#ifdef SO_REUSEADDR
  scm_c_define ("SO_REUSEADDR", scm_from_int (SO_REUSEADDR));
#endif
#ifdef SO_STYLE
  scm_c_define ("SO_STYLE", scm_from_int (SO_STYLE));
#endif
#ifdef SO_TYPE
  scm_c_define ("SO_TYPE", scm_from_int (SO_TYPE));
#endif
#ifdef SO_ERROR
  scm_c_define ("SO_ERROR", scm_from_int (SO_ERROR));
#endif
#ifdef SO_DONTROUTE
  scm_c_define ("SO_DONTROUTE", scm_from_int (SO_DONTROUTE));
#endif
#ifdef SO_BROADCAST
  scm_c_define ("SO_BROADCAST", scm_from_int (SO_BROADCAST));
#endif
#ifdef SO_SNDBUF
  scm_c_define ("SO_SNDBUF", scm_from_int (SO_SNDBUF));
#endif
#ifdef SO_RCVBUF
  scm_c_define ("SO_RCVBUF", scm_from_int (SO_RCVBUF));
#endif
#ifdef SO_KEEPALIVE
  scm_c_define ("SO_KEEPALIVE", scm_from_int (SO_KEEPALIVE));
#endif
#ifdef SO_OOBINLINE
  scm_c_define ("SO_OOBINLINE", scm_from_int (SO_OOBINLINE));
#endif
#ifdef SO_NO_CHECK
  scm_c_define ("SO_NO_CHECK", scm_from_int (SO_NO_CHECK));
#endif
#ifdef SO_PRIORITY
  scm_c_define ("SO_PRIORITY", scm_from_int (SO_PRIORITY));
#endif
#ifdef SO_LINGER
  scm_c_define ("SO_LINGER", scm_from_int (SO_LINGER));
#endif

  /* recv/send options.  */
#ifdef MSG_DONTWAIT
  scm_c_define ("MSG_DONTWAIT", scm_from_int (MSG_DONTWAIT));
#endif
#ifdef MSG_OOB
  scm_c_define ("MSG_OOB", scm_from_int (MSG_OOB));
#endif
#ifdef MSG_PEEK
  scm_c_define ("MSG_PEEK", scm_from_int (MSG_PEEK));
#endif
#ifdef MSG_DONTROUTE
  scm_c_define ("MSG_DONTROUTE", scm_from_int (MSG_DONTROUTE));
#endif

#ifdef IP_ADD_MEMBERSHIP
  scm_c_define ("IP_ADD_MEMBERSHIP", scm_from_int (IP_ADD_MEMBERSHIP));
  scm_c_define ("IP_DROP_MEMBERSHIP", scm_from_int (IP_DROP_MEMBERSHIP));
#endif

  scm_add_feature ("socket");

#include "include/socket.x"
}
