#include "include/config.h"

#include <ctype.h>
#include <limits.h>
#include "include/_scm.h"
#include "include/validate.h"

#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/chars.h"



#ifdef _DCC
# define ASCII
#else
# if (('\n'=='\025') && (' '=='\100') && ('a'=='\201') && ('A'=='\301'))
#  define EBCDIC
# endif /*  (('\n'=='\025') && (' '=='\100') && ('a'=='\201') && ('A'=='\301')) */
# if (('\n'=='\012') && (' '=='\040') && ('a'=='\141') && ('A'=='\101'))
#  define ASCII
# endif /*  (('\n'=='\012') && (' '=='\040') && ('a'=='\141') && ('A'=='\101')) */
#endif /* def _DCC */


#ifdef EBCDIC
char *const scm_charnames[] =
{
  "nul", "soh", "stx", "etx", "pf", "ht", "lc", "del",
   0   , 0   , "smm", "vt", "ff", "cr", "so", "si",
  "dle", "dc1", "dc2", "dc3", "res", "nl", "bs", "il",
  "can", "em", "cc", 0   , "ifs", "igs", "irs", "ius",
   "ds", "sos", "fs", 0   , "byp", "lf", "eob", "pre",
   0   , 0   , "sm", 0   , 0   , "enq", "ack", "bel",
   0   , 0   , "syn", 0   , "pn", "rs", "uc", "eot",
   0   , 0   , 0   , 0   , "dc4", "nak", 0   , "sub",
   "space", scm_s_newline, "tab", "backspace", "return", "page", "null"};

const char scm_charnums[] =
"\000\001\002\003\004\005\006\007\
\010\011\012\013\014\015\016\017\
\020\021\022\023\024\025\026\027\
\030\031\032\033\034\035\036\037\
\040\041\042\043\044\045\046\047\
\050\051\052\053\054\055\056\057\
\060\061\062\063\064\065\066\067\
\070\071\072\073\074\075\076\077\
 \n\t\b\r\f\0";
#endif /* def EBCDIC */
#ifdef ASCII
const char *scm_charnames[] =
{
  "nul","soh","stx","etx","eot","enq","ack","bel",
   "bs", "ht", "newline", "vt", "np", "cr", "so", "si",
  "dle","dc1","dc2","dc3","dc4","nak","syn","etb",
  "can", "em","sub","esc", "fs", "gs", "rs", "us",
  "space", "sp", "nl", "tab", "backspace", "return", "page", "null", "del"};
const char scm_charnums[] =
"\000\001\002\003\004\005\006\007\
\010\011\012\013\014\015\016\017\
\020\021\022\023\024\025\026\027\
\030\031\032\033\034\035\036\037\
  \n\t\b\r\f\0\177";
#endif /* def ASCII */

int scm_n_charnames = sizeof (scm_charnames) / sizeof (char *);



SCM_DEFINE (scm_char_p, "char?", 1, 0, 0, 
            (SCM x),
	    "Return @code{#t} iff @var{x} is a character, else @code{#f}.")
#define FUNC_NAME s_scm_char_p
{
  reference_parameter (x);
  SCM answer = scm_from_bool (SCM_CHARP(x));
  dereference_parameter (x);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE1 (scm_char_eq_p, "char=?", scm_tc7_rpsubr,
             (SCM x, SCM y),
	     "Return @code{#t} iff @var{x} is the same character as @var{y}, else @code{#f}.")
#define FUNC_NAME s_scm_char_eq_p
{
  reference_parameter (x);
  reference_parameter (y);
  SCM_VALIDATE_CHAR (1, x);
  SCM_VALIDATE_CHAR (2, y);

  SCM answer= scm_from_bool (scm_is_eq (x, y));
  dereference_parameter (x);
  dereference_parameter (y);
  return (answer);
}
#undef FUNC_NAME


SCM_DEFINE1 (scm_char_less_p, "char<?", scm_tc7_rpsubr, 
             (SCM x, SCM y),
	     "Return @code{#t} iff @var{x} is less than @var{y} in the ASCII sequence,\n"
	     "else @code{#f}.")
#define FUNC_NAME s_scm_char_less_p
{
  reference_parameter (x);
  reference_parameter (y);
  SCM_VALIDATE_CHAR (1, x);
  SCM_VALIDATE_CHAR (2, y);

  SCM answer = scm_from_bool (SCM_CHAR(x) < SCM_CHAR(y));
  dereference_parameter (x);
  dereference_parameter (y);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE1 (scm_char_leq_p, "char<=?", scm_tc7_rpsubr,
             (SCM x, SCM y),
	     "Return @code{#t} iff @var{x} is less than or equal to @var{y} in the\n"
	     "ASCII sequence, else @code{#f}.")
#define FUNC_NAME s_scm_char_leq_p
{
  reference_parameter (x);
  reference_parameter (y);
  SCM_VALIDATE_CHAR (1, x);
  SCM_VALIDATE_CHAR (2, y);

  SCM answer = scm_from_bool (SCM_CHAR(x) <= SCM_CHAR(y));
  dereference_parameter (x);
  dereference_parameter (y);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE1 (scm_char_gr_p, "char>?", scm_tc7_rpsubr,
             (SCM x, SCM y),
	     "Return @code{#t} iff @var{x} is greater than @var{y} in the ASCII\n"
	     "sequence, else @code{#f}.")
#define FUNC_NAME s_scm_char_gr_p
{
  reference_parameter (x);
  reference_parameter (y);
  SCM_VALIDATE_CHAR (1, x);
  SCM_VALIDATE_CHAR (2, y);

  SCM answer = scm_from_bool (SCM_CHAR(x) > SCM_CHAR(y));
  dereference_parameter (x);
  dereference_parameter (y);
  return (answer);
}
#undef FUNC_NAME

SCM_DEFINE1 (scm_char_geq_p, "char>=?", scm_tc7_rpsubr,
             (SCM x, SCM y),
	     "Return @code{#t} iff @var{x} is greater than or equal to @var{y} in the\n"
	     "ASCII sequence, else @code{#f}.")
#define FUNC_NAME s_scm_char_geq_p
{
  reference_parameter (x);
  reference_parameter (y);
  SCM_VALIDATE_CHAR (1, x);
  SCM_VALIDATE_CHAR (2, y);

  SCM answer = scm_from_bool (SCM_CHAR(x) >= SCM_CHAR(y));
  dereference_parameter (x);
  dereference_parameter (y);
  return (answer);
}
#undef FUNC_NAME

// See Viper characters.scm and the roadmap: strings & brass
/* SCM_DEFINE (scm_char_alphabetic_p, "char-alphabetic?", 1, 0, 0, */
/*            (SCM chr), */
/* 	    "Return @code{#t} iff @var{chr} is alphabetic, else @code{#f}.\n") */
/* #define FUNC_NAME s_scm_char_alphabetic_p */
/* { */
/*   return scm_char_set_contains_p (scm_char_set_letter, chr); */
/* } */
/* #undef FUNC_NAME */

/* SCM_DEFINE (scm_char_numeric_p, "char-numeric?", 1, 0, 0,  */
/*            (SCM chr), */
/* 	    "Return @code{#t} iff @var{chr} is numeric, else @code{#f}.\n") */
/* #define FUNC_NAME s_scm_char_numeric_p */
/* { */
/*   return scm_char_set_contains_p (scm_char_set_digit, chr); */
/* } */
/* #undef FUNC_NAME */

/* SCM_DEFINE (scm_char_whitespace_p, "char-whitespace?", 1, 0, 0,  */
/*            (SCM chr), */
/* 	    "Return @code{#t} iff @var{chr} is whitespace, else @code{#f}.\n") */
/* #define FUNC_NAME s_scm_char_whitespace_p */
/* { */
/*   return scm_char_set_contains_p (scm_char_set_whitespace, chr); */
/* } */
/* #undef FUNC_NAME */



/* SCM_DEFINE (scm_char_upper_case_p, "char-upper-case?", 1, 0, 0,  */
/*            (SCM chr), */
/* 	    "Return @code{#t} iff @var{chr} is uppercase, else @code{#f}.\n") */
/* #define FUNC_NAME s_scm_char_upper_case_p */
/* { */
/*   return scm_char_set_contains_p (scm_char_set_upper_case, chr); */
/* } */
/* #undef FUNC_NAME */


/* SCM_DEFINE (scm_char_lower_case_p, "char-lower-case?", 1, 0, 0,  */
/*            (SCM chr), */
/* 	    "Return @code{#t} iff @var{chr} is lowercase, else @code{#f}.\n") */
/* #define FUNC_NAME s_scm_char_lower_case_p */
/* { */
/*   return scm_char_set_contains_p (scm_char_set_lower_case, chr); */
/* } */
/* #undef FUNC_NAME */



/* SCM_DEFINE (scm_char_is_both_p, "char-is-both?", 1, 0, 0,  */
/*             (SCM chr), */
/* 	    "Return @code{#t} iff @var{chr} is either uppercase or lowercase, else @code{#f}.\n") */
/* #define FUNC_NAME s_scm_char_is_both_p */
/* { */
/*   if (scm_is_true (scm_char_set_contains_p (scm_char_set_lower_case, chr))) */
/*     return SCM_BOOL_T; */
/*   return scm_char_set_contains_p (scm_char_set_upper_case, chr); */
/* } */
/* #undef FUNC_NAME */




SCM_DEFINE (scm_char_to_integer, "char->integer", 1, 0, 0, 
            (SCM chr),
	    "Return the number corresponding to ordinal position of @var{chr} in the\n"
	    "ASCII sequence.")
#define FUNC_NAME s_scm_char_to_integer
{
  reference_parameter (chr);
  SCM_VALIDATE_CHAR (1, chr);
  SCM answer = scm_from_ulong (SCM_CHAR(chr));
  dereference_parameter (chr);
  return (answer);
}
#undef FUNC_NAME



SCM_DEFINE (scm_integer_to_char, "integer->char", 1, 0, 0, 
           (SCM n), "")
#define FUNC_NAME s_scm_integer_to_char
{
  reference_parameter (n);
  uint32_t cn = scm_to_uint32 (n);
  SCM answer = SCM_MAKE_UC_CHAR (cn);
  dereference_parameter (n);
  return (answer);
}
#undef FUNC_NAME

// (usv->utf-8 (char->integer c))
SCM_DEFINE (scm_char_to_bs_utf8, "char->byte-string/utf-8", 1, 0, 0, 
            (SCM chr), "")
#define FUNC_NAME s_scm_char_to_bs_utf8
{
  reference_parameter (chr);
  SCM_VALIDATE_CHAR (1, chr);
  SCM answer = utf8_usv_to_bs (SCM_CHAR(chr));
  dereference_parameter (chr);
  return (answer);
}
#undef FUNC_NAME


int
scm_c_upcase (unsigned int c)
{
  if (c <= UCHAR_MAX)
    return toupper (c);
  else
    return c;
}


int
scm_c_downcase (unsigned int c)
{
  if (c <= UCHAR_MAX)
    return tolower (c);
  else
    return c;
}

SCM
scm_charname_lookup (SCM brass)
{
  reference_parameter (brass);
  SCM strbuf = scm_brass_to_c0_byte_string (brass);
  reference_parameter (strbuf);
  SCM ans = SCM_BOOL_F;
  if (!scm_is_false (strbuf)) {
    char *charname = BYTE_STRING_DATA (strbuf);
    size_t charname_len = BYTE_STRING_LENGTH (strbuf);
    int c;
    for (c = 0; c < scm_n_charnames; c++) {
      if (scm_charnames[c]
          && (!bcmp (scm_charnames[c], charname, charname_len))) {
        usv_t t = scm_charnums[c];
        ans = SCM_MAKE_CHAR (t);
        break;
      }
    }
  }
  dereference_parameter (brass);
  dereference_parameter (strbuf);
  return ans;
}



void
scm_init_chars ()
{
#include "include/chars.x"
}
