#include "include/_scm.h"
#include "include/validate.h"
#include "include/trim.h" // ??? isn't this included somewhere
#include "include/chars.h"

#include "include/dictionary.h"

/*
 * Entry
 */
SCM_DEFINE (scm_make_dict_entry, "make-dict-entry", 3, 0, 0,
            (SCM key, SCM definition, SCM successors),
            "Creates a new dictionary entry object.")
#define FUNC_NAME s_scm_make_dict_entry
{
  // RC consuming
  return scm_i_cons (key, scm_i_cons (definition, successors));
}
#undef FUNC_NAME


/* internal unchecked dangerous lax and direct dinosaur claws */

static void
scm_add_successor_x (SCM entry, SCM successor_entry)
{
  SCM_SETCDR (TMP_CDR (entry),
              scm_i_cons (successor_entry,
                          GET_SUCCESSORS (entry)));
  return;
}

// This is a macro now.
// static SCM
// scm_get_successor (SCM entry, SCM key)
// {
//   return scm_assq_unchecked_lax (key, GET_SUCCESSORS (entry));
// }

#define USV_REST(x)     (x + 1)

// Return the entry for /keys/ or SCM_UNDEFINED if it isn't found.
// RC lax (or should we set the return value flag???)
SCM
scm_i_get_entry_by_usv (SCM branch, usv_t *keys, size_t length)
{
  // if (SCM_UNDEFINED_P (branch)) return (SCM_UNDEFINED);
  if (scm_is_false (branch)) return (SCM_UNDEFINED);
  if (!scm_is_pair (branch)) { // XXX debugging
    fprintf (stderr, " * YOU damaged everything!!!\n");
    abort ();
  }

  if (length == 0)              return (branch);

  return ( \
    scm_i_get_entry_by_usv ( \
      scm_assq_unchecked_lax (SCM_MAKE_UC_CHAR (*keys), GET_SUCCESSORS (branch)),
      USV_REST (keys), --length));
}

// Return the entry for /keys/ or create one if necessary.
// RC lax (referencing is done within the data graph)
SCM
scm_i_create_entry_by_usv_x (SCM parent, usv_t *keys, size_t length)
{
  SCM key  =  SCM_MAKE_UC_CHAR (*keys);
  SCM entry = scm_get_successor (parent, key);

  if (scm_is_false (entry)) {
    entry = MAKE_ENTRY (key, SCM_UNDEFINED, SCM_EOL);
    scm_add_successor_x (parent, entry);
  }

  if (length == 1) return entry;
  return (scm_i_create_entry_by_usv_x (entry, USV_REST (keys),
                                       --length));
}

/* end of internal unchecked dangerous lax and direct dinosaur claws for entries */


/*
 * Dictionary
 */

/* SCM_DEFINE (scm_make_dictionary, "make-dictionary", 0, 0, 0, */
/*             (), */
/*             "Creates a new dictionary data type object.") */
/* #define FUNC_NAME s_scm_make_dictionary */
/* { */
/*   SCM root   = scm_make_dict_entry (SCM_BOOL_F, SCM_UNDEFINED, SCM_EOL); */
/*   /\* SCM answer = scm_cell (scm_tc7_dictionary, *\/ */
/*   /\*                        (scm_t_bits) 0, *\/ */
/*   /\*                        (scm_t_bits) root); *\/ */
/*   SCM answer = SCM_BOOL_F; */
/*   reference (answer, root); */

/*   return (answer); */
/* } */
/* #undef FUNC_NAME */


/* SCM_DEFINE (scm_dictionary_p, "dictionary?", 1, 0, 0, */
/*             (SCM object), */
/*             "Dictionary data type predicate.") */
/* #define FUNC_NAME s_scm_make_dictionary */
/* { */
/*   reference_parameter (object); */
/*   SCM answer = SCM_BOOL_F; */

/*   if (SCM_IS_DICTIONARY (object)) answer = SCM_BOOL_T; */

/*   dereference_parameter (object); */
/*   return (answer); */
/* } */
/* #undef FUNC_NAME */


/* /\* XXX that is bleeding internal data structures | transitional *\/ */
/* SCM_DEFINE (scm_dictionary_root, "dictionary-root", 1, 0, 0, */
/*             (SCM dictionary), */
/*             "") */
/* #define FUNC_NAME s_scm_dictionary_root */
/* { */
/*   reference_parameter (dictionary); */
/*   SCM answer = SCM_UNSPECIFIED; */

/*   if (SCM_IS_DICTIONARY (dictionary)) { */
/*     answer = SCM_DICTIONARY_ROOT (dictionary); */
/*   } else { */
/*     SCM_WRONG_TYPE_ARG (1, dictionary); */
/*   } */

/*   reference_count_return (answer); */
/*   dereference_parameter (dictionary); */
/*   return (answer); */
/* } */
/* #undef FUNC_NAME */


/* default SCM interface
 * TODO printer
 * TODO read/write serialization
 */

/* inline void */
/* trim_release_dictionary (SCM dictionary) */
/* { */
/*   reference_count_minus1 (SCM_DICTIONARY_ROOT (dictionary)); */
/*   return; */
/* } */

void
x_inclusion_dictionary ()
{
  #include "include/dictionary.x"
}
