#include "include/config.h"

#include "include/_scm.h"
#include "include/eq.h"

#include "include/validate.h"
#include "include/list.h"
#include "include/eval.h"
#include "include/srfi-1.h" // equal_trampoline

#include <stdarg.h>

/* This file is too long and list-processing.c too short.
 * Please take some time and move some definitions.
 */

/* creating lists */

/*
 * List Clock   D R A F T    Abstracted Cyclic List Detection
 * (No Reference Counting / see make-list-clock/no-count in the MCO)
 * [no user manual, no examples, no nothing]
 */
struct scm_list_clock {
  SCM graph;
  SCM hour;
  SCM minute;
  SCM second;
  unsigned short pendulum; // ??? there are no short registers right now
};

void
initialize_clock (SCM graph, struct scm_list_clock *clock_loc)
{
  clock_loc->graph  = graph;
  clock_loc->hour   = graph;
  clock_loc->minute = graph;
  clock_loc->second = TMP_CDR (graph);
  clock_loc->pendulum = 0;

  return;
}

SCM clock_next (struct scm_list_clock *clock_loc);

inline SCM
clock_next (struct scm_list_clock *clock_loc)
{
  if (clock_loc->pendulum)
    clock_loc->minute = TMP_CDR (clock_loc->minute);

  if (!scm_is_pair (clock_loc->second)) {
    return clock_loc->second;

  } else if (clock_loc->second == clock_loc->minute) {
    // XXX FUNC_NAME and SCM_MISC_ERROR
    scm_misc_error (NULL, "Circular structure in position 1: ~S",
                    scm_list_1 (clock_loc->graph));

  } else {
    clock_loc->pendulum = !(clock_loc->pendulum);

    clock_loc->hour = clock_loc->second;
    clock_loc->second = TMP_CDR (clock_loc->second);
    return (clock_loc->hour);
  }
}



/* Fuel Generators */

inline SCM
scm_list_1 (SCM e1)
{
  SCM c1 = scm_i_cons (e1, SCM_EOL);
  return c1;
}

inline SCM
scm_list_2 (SCM e1, SCM e2)
{
  SCM c1, c2;
  c2 = scm_i_cons (e2, SCM_EOL);
  c1 = scm_i_cons (e1, c2);
  return c1;
}

inline SCM
scm_list_3 (SCM e1, SCM e2, SCM e3)
{
  SCM c1, c2, c3;
  c3 = scm_i_cons (e3, SCM_EOL);
  c2 = scm_i_cons (e2, c3);
  c1 = scm_i_cons (e1, c2);
  return c1;
}

SCM
scm_list_4 (SCM e1, SCM e2, SCM e3, SCM e4)
{
  return scm_cons2 (e1, e2, scm_list_2 (e3, e4));
}

SCM
scm_list_5 (SCM e1, SCM e2, SCM e3, SCM e4, SCM e5)
{
  return scm_cons2 (e1, e2, scm_list_3 (e3, e4, e5));
}

// XXX goes to header
void scm_i_copy_list (SCM list, SCM *head, SCM *tail);

/* MCO->SCM scm-i-copy-list
 * Deviations from the MCO:
 * - improper list extension: The head return value contains that
 *                            non list element while the tail return
 *                            value is SCM_EOL.
 * - return value flags: As the tail return value is already list
 *                       referenced and because it is tedious to
 *                       dereference its return value flag every
 *                       time the tail return value does not have
 *                       that flag anymore but is only list refe-
 *                       renced (if it doesn't happen to be eq to
 *                       head.)
 * XXX copy this into the MCO (SCM->MCO please)
 */
inline void
scm_i_copy_list (SCM list, SCM *head, SCM *tail)
{
  reference_parameter (list);

  if (!scm_is_pair (list)) {
    reference_count_return (list);  // XXX improper list extension
    dereference_parameter (list);
    *head = list;  // XXX improper list extension
    *tail = SCM_EOL;
    return;
  }

  *head = scm_i_cons (TMP_CAR (list), SCM_EOL);
  *tail = *head;
  reference_parameter (*head);
  reference_parameter (*tail);

  SCM current = TMP_CDR (list);
  while (scm_is_pair (current)) {
    scm_i_set_cdr_x (*tail, scm_i_cons (TMP_CAR (current), SCM_EOL));
    *tail = fly_cdr (*tail);
    current = TMP_CDR (current);
  }
  if (!scm_is_null (current))  // XXX improper list extension
    scm_i_set_cdr_x (*tail, current);

  rc_return_stack_reference (*head);
  dereference_parameter (*tail);  // list referenced
  dereference_parameter (list);
}

/* MCO->SCM scm-list-n/counting-iterative */
SCM scm_list_n (SCM first_element, ...) {
  va_list arguments;
  SCM answer, last, element;

  answer = scm_i_cons (first_element, SCM_EOL);
  reference_parameter (answer);

  va_start (arguments, first_element);
  last = answer;
  element = va_arg (arguments, SCM);
  while (!SCM_UNBNDP (element))
    {
#if (SCM_DEBUG_CELL_ACCESSES == 1)
      if (SCM_NIMP (elt))
	SCM_VALIDATE_CELL(elt, 0);
#endif      

      scm_i_set_cdr_x (last, scm_i_cons (element, SCM_EOL));

      last = scm_i_cdr (last);
      element = va_arg (arguments, SCM);
    }
  va_end (arguments);

  dereference_return_value (last);
  return rc_return_stack_reference (answer);
}


SCM_DEFINE (scm_make_list, "make-list", 1, 1, 0,
            (SCM n, SCM init),
	    "Create a list containing of @var{n} elements, where each\n"
	    "element is initialized to @var{init}.  @var{init} defaults to\n"
	    "the empty list @code{()} if not given.")
#define FUNC_NAME s_scm_make_list
{
  unsigned nn = scm_to_uint (n);
  unsigned i;
  SCM ret = SCM_EOL;

  if (SCM_UNBNDP (init))
    init = SCM_EOL;

  for (i = 0; i < nn; i++)
    ret = scm_cons (init, ret);
  return ret;
}
#undef FUNC_NAME


/* general questions about lists --- null?, list?, length, etc.  */

SCM_DEFINE (scm_null_p, "null?", 1, 0, 0, 
           (SCM x),
	    "Return @code{#t} iff @var{x} is the empty list, else @code{#f}.")
#define FUNC_NAME s_scm_null_p
{
  SCM answer = scm_from_bool (scm_is_null (x));
  dereference_return_value (x);
  return answer;
}
#undef FUNC_NAME

int
scm_i_is_list (SCM list)
{
  return (scm_is_null (list) ||
          (scm_is_pair (list) && (scm_is_pair (TMP_CDR (list))
                                  || scm_is_null (TMP_CDR (list)) )));
}

SCM_DEFINE (scm_list_p, "list?", 1, 0, 0, 
           (SCM x),
	    "Return @code{#t} if @var{x} is a list, else @code{#f}.")
#define FUNC_NAME s_scm_list_p
{
  SCM answer = scm_from_bool (scm_i_is_list (x));
  dereference_return_value (x);
  return answer;
}
#undef FUNC_NAME


/* Return the length of SX, or -1 if it's not a proper list.
   This uses the "tortoise and hare" algorithm to detect "infinitely
   long" lists (i.e. lists with cycles in their cdrs), and returns -1
   if it does find one.  */
long
scm_ilength(SCM sx)
{
  long i = 0;
  SCM tortoise = sx;
  SCM hare = sx;

  do {
    if (scm_is_null(hare)) return i;
    if (!scm_is_pair (hare)) return -1;
    hare = SCM_CELL_OBJECT_1(hare);
    i++;
    if (scm_is_null(hare)) return i;
    if (!scm_is_pair (hare)) return -1;
    hare = SCM_CELL_OBJECT_1(hare);
    i++;
    /* For every two steps the hare takes, the tortoise takes one.  */
    tortoise = SCM_CELL_OBJECT_1(tortoise);
  }
  while (!scm_is_eq (hare, tortoise));

  /* If the tortoise ever catches the hare, then the list must contain
     a cycle.  */
  return -1;
}

void
debug_trim_print_list (SCM sx)
{
  fprintf (stderr, "%p: ", SCM2PTR (sx));

  long i = 0;
  SCM tortoise = sx;
  SCM hare = sx;

  do {
    float count;

    // Step 1
    if (scm_is_null(hare)) {
      fprintf (stderr, "EOL\n");
      return;
    }
    if (!scm_is_pair (hare)) {
      if (FREE_CELL_P (hare)) {
        count = (float) TRIM_REFCOUNT (hare) / 2.0;
        fprintf (stderr, "(%.1f . free cell . ||\n", count);
      } else {
        fprintf (stderr, " non list object\n");
      }
      return;
    }
    count = (float) TRIM_REFCOUNT (hare) / 2.0;
    fprintf (stderr, "(%.1f . . ", count);
    hare = SCM_CELL_OBJECT_1(hare);
    i++;

    // Step 2
    if (scm_is_null(hare)) {
      fprintf (stderr, "EOL\n");
      return;
    }
    if (!scm_is_pair (hare)) {
      if (FREE_CELL_P (hare)) {
        count = (float) TRIM_REFCOUNT (hare) / 2.0;
        fprintf (stderr, "(%.1f . free cell . ", count);
      } else {
        fprintf (stderr, " non list object\n");
      }
      return;
    }
    count = (float) TRIM_REFCOUNT (hare) / 2.0;
    fprintf (stderr, "(%.1f . . ", count);
    hare = SCM_CELL_OBJECT_1(hare);
    i++;

    /* For every two steps the hare takes, the tortoise takes one.  */
    tortoise = SCM_CELL_OBJECT_1(tortoise);
  }
  while (!scm_is_eq (hare, tortoise));

  fprintf (stderr, " cyclic list detected\n");
  return;
}

void
debug_trim_assert_list (SCM sx)
{
  SCM tortoise = sx;
  SCM hare = sx;
  long j = 0;

  do {
    for (j = 0; j < 2; j++) {

      if (scm_is_null(hare)) { return; }

      if (!scm_is_pair (hare)) {
        if (FREE_CELL_P (hare)) {
          fprintf (stderr, " TRIM Assertion: Broken List References: ");
          debug_trim_print_list (sx);
          abort ();
        } else {
          fprintf (stderr, " TRIM Assertion:  non list object\n");
          return;
        }
      }
      if (hare != sx && TRIM_REFCOUNT (hare) < 2) {
        fprintf (stderr, " TRIM Assertion: Broken List References: ");
        debug_trim_print_list (sx);
        abort ();
      }
      hare = SCM_CELL_OBJECT_1(hare);
    }

    /* For every two steps the hare takes, the tortoise takes one.  */
    tortoise = SCM_CELL_OBJECT_1(tortoise);
  }
  while (!scm_is_eq (hare, tortoise));

  fprintf (stderr, " TRIM Assertion: Cyclic List: ");
  debug_trim_print_list (sx);

  return;
}


SCM_DEFINE (scm_length, "length", 1, 0, 0, 
           (SCM lst),
	    "Return the number of elements in list @var{lst}.")
#define FUNC_NAME s_scm_length
{
  long i;
  reference_parameter (lst);

  SCM_VALIDATE_LIST_COPYLEN (1, lst, i);

  dereference_parameter (lst);
  return scm_from_long (i);
}
#undef FUNC_NAME



/* appending lists */
// MCO: scm-append/list-count
SCM_DEFINE (scm_append, "append", 0, 0, 1, 
            (SCM args),
	    "Return a list consisting of the elements the lists passed as\n"
	    "arguments.\n"
	    "@lisp\n"
	    "(append '(x) '(y))          @result{}  (x y)\n"
	    "(append '(a) '(b c d))      @result{}  (a b c d)\n"
	    "(append '(a (b)) '((c)))    @result{}  (a (b) (c))\n"
	    "@end lisp\n"
	    "The resulting list is always newly allocated, except that it\n"
	    "shares structure with the last list argument.  The last\n"
	    "argument may actually be any object; an improper list results\n"
	    "if the last argument is not a proper list.\n"
	    "@lisp\n"
	    "(append '(a b) '(c . d))    @result{}  (a b c . d)\n"
	    "(append '() 'a)             @result{}  a\n"
	    "@end lisp")
#define FUNC_NAME s_scm_append
{
  SCM_VALIDATE_REST_ARGUMENT (args);
  int argnum = 1;

  reference_parameter (args);

  if (scm_is_null (args)) {
    dereference_parameter (args);
    return SCM_EOL;
  }

  /* Initialization */
  SCM answer = SCM_EOL;
  SCM last   = SCM_EOL;
  SCM tmp    = SCM_EOL;
  SCM current = TMP_CAR (args);
  if (scm_is_null (current)) {
    return scm_append (scm_i_cdr (rc_Fly (args)));
  } else if (scm_is_null (TMP_CDR (args))) {
    answer = current;
    reference_count_return (answer);
    dereference_parameter (args);
    return answer;
  } else {
    SCM head, tail;
    scm_i_copy_list (current, &head, &tail);
    if (!scm_is_pair (head) && !scm_is_null (head))
      scm_misc_error (FUNC_NAME, "Improper list: ~S", args);
    answer = head; reference_parameter (answer);
    last = tail;
  }

  args = fly_cdr (args);
  if (scm_is_null (args)) {
    dereference_parameter (args);
    return rc_return_stack_reference (answer);
  }
  /* Main Loop */
  current = args;
  while (!scm_is_null (TMP_CDR (current))) {
    if (scm_is_null (TMP_CAR (current))) {
      current = TMP_CDR (current);
      continue;
    }
    SCM head, tail;
    scm_i_copy_list (TMP_CAR (current), &head, &tail);
    if (!scm_is_pair (head) && !scm_is_null (head))
      scm_misc_error (FUNC_NAME, "Improper list: ~S", args);
    scm_i_set_cdr_x (last, head);
    last = tail;
    current = TMP_CDR (current);
  }
  if (!scm_is_pair (current) && !scm_is_null (current))
    scm_misc_error (FUNC_NAME, "Improper list: ~S", args);
  scm_i_set_cdr_x (last, TMP_CAR (current));
  dereference_parameter (args);
  return rc_return_stack_reference (answer);
}
#undef FUNC_NAME



// MCO: scm-append!/list-count (copy&paste work)
SCM_DEFINE (scm_append_x, "append!", 0, 0, 1, 
            (SCM lists),
	    "A destructive version of @code{append} (@pxref{Pairs and\n"
	    "Lists,,,r5rs, The Revised^5 Report on Scheme}).  The cdr field\n"
	    "of each list's final pair is changed to point to the head of\n"
	    "the next list, so no consing is performed.  Return\n"
	    "the mutated list.")
#define FUNC_NAME s_scm_append_x
{ // XXX TODO needs to check for improper list (dotted rest)
  SCM_VALIDATE_REST_ARGUMENT (lists);

  SCM args = lists;
  reference_parameter (args);

  if (scm_is_null (args)) {
    dereference_parameter (args);
    return SCM_EOL;
  }

  /* Initialization */
  SCM answer = SCM_EOL;
  SCM last   = SCM_EOL;
  SCM tmp    = SCM_EOL;
  SCM current = TMP_CAR (args);
  if (scm_is_null (current)) {
    return scm_append_x (scm_i_cdr (rc_Fly (args)));
  } else if (scm_is_null (TMP_CDR (args))) {
    answer = current;
    reference_count_return (answer);
    dereference_parameter (args);
    return answer;
  } else {
    answer = current; reference_parameter (answer);
    last = scm_last_pair (current);
  }

  args = fly_cdr (args);
  if (scm_is_null (args)) {
    dereference_parameter (args);
    return rc_return_stack_reference (answer);
  }

  /* Main Loop */
  current = args;
  while (!scm_is_null (TMP_CDR (current))) {
    if (scm_is_null (TMP_CAR (current))) {
      current = TMP_CDR (current);
      continue;
    }
    SCM head, tail;
    head = TMP_CAR (current);
    tail = scm_last_pair (head);
    scm_i_set_cdr_x (last, head);
    last = tail;
    current = TMP_CDR (current);
  }
  scm_i_set_cdr_x (last, TMP_CAR (current));
  dereference_parameter (args);
  return rc_return_stack_reference (answer);
}
#undef FUNC_NAME


SCM_DEFINE (scm_last_pair, "last-pair", 1, 0, 0, 
           (SCM lst),
	    "Return the last pair in @var{lst}, signalling an error if\n"
	    "@var{lst} is circular.")
#define FUNC_NAME s_scm_last_pair
{
  reference_parameter (lst);

  SCM tortoise = lst;
  SCM hare = lst;

  if (scm_is_null (lst))
    return lst;

  SCM_VALIDATE_CONS (SCM_ARG1, lst);
  do {
    SCM ahead = TMP_CDR(hare);
    if (!scm_is_pair (ahead)) {
      reference_count_return (hare);
      dereference_parameter (lst);
      return hare;
    }
    hare = ahead;

    ahead = TMP_CDR(hare);
    if (!scm_is_pair (ahead)) {
      reference_count_return (hare);
      dereference_parameter (lst);
      return hare;
    }
    hare = ahead;

    tortoise = TMP_CDR(tortoise);
  }
  while (!scm_is_eq (hare, tortoise));
  SCM_MISC_ERROR ("Circular structure in position 1: ~S", scm_list_1 (lst));
}
#undef FUNC_NAME


/* reversing lists */

/* MCO->SCM scm-reverse/list-count */
SCM_DEFINE (scm_reverse, "reverse", 1, 0, 0,
            (SCM list),
	    "Return a new list that contains the elements of @var{lst} but\n"
	    "in reverse order.")
#define FUNC_NAME s_scm_reverse
{
  reference_parameter (list);

  if (scm_is_null (list)) {
    dereference_parameter (list);  // bogus
    return SCM_EOL;
  } else if (!scm_is_pair (list)) {
    SCM_ASSERT(scm_is_pair(list), list, 1, FUNC_NAME);
  }

  SCM last = scm_i_cons (TMP_CAR (list), SCM_EOL);
  struct scm_list_clock clock;
  initialize_clock (list, &clock);

  SCM element = clock_next (&clock);
  while (!scm_is_null (element)) {
    SCM_ASSERT(scm_is_pair(element), list, 1, FUNC_NAME);
    last = scm_i_cons (TMP_CAR (element), last);
    element = clock_next (&clock);
  }

  reference_count_return (last);
  dereference_parameter (list);
  return last;
}
#undef FUNC_NAME


SCM_DEFINE (scm_reverse_x, "reverse!", 1, 1, 0,
            (SCM list, SCM new_tail),
	    "A destructive version of @code{reverse} (@pxref{Pairs and Lists,,,r5rs,\n"
	    "The Revised^5 Report on Scheme}).  The cdr of each cell in @var{lst} is\n"
	    "modified to point to the previous list element.  Return the\n"
	    "reversed list.\n\n"
	    "Caveat: because the list is modified in place, the tail of the original\n"
	    "list now becomes its head, and the head of the original list now becomes\n"
	    "the tail.  Therefore, the @var{lst} symbol to which the head of the\n"
	    "original list was bound now points to the tail.  To ensure that the head\n"
	    "of the modified list is not lost, it is wise to save the return value of\n"
	    "@code{reverse!}")
#define FUNC_NAME s_scm_reverse_x
{
  SCM_VALIDATE_LIST (1, list);
  if (SCM_UNBNDP (new_tail))
    new_tail = SCM_EOL;
  else
    SCM_VALIDATE_LIST (2, new_tail);

  if (scm_is_null (list)) {
    return new_tail;
  }

  // MCO->SCM scm-reverse! + check MCO please
  reference_parameter (list);

  SCM element = list;
  SCM rest    = TMP_CDR (list);
  SCM second  = new_tail;
  reference (list, new_tail); dereference_return_value (new_tail);
  do {
    SCM_SET_CELL_OBJECT_1 (element, second);

    if (scm_is_null (rest)) break;

    second  = element;
    element = rest;
    rest    = TMP_CDR (rest);
  } while (1);

  // "Calculated" Count
  reference (element, list);
  reference_count_return (element);
  dereference (list, element);
  dereference_parameter (list);
  return element;
}
#undef FUNC_NAME






/* indexing lists by element number */

SCM_DEFINE (scm_list_ref, "list-ref", 2, 0, 0,
	    (SCM list, SCM k),
	    "Return the @var{k}th element from @var{list}.")
#define FUNC_NAME s_scm_list_ref
{
  reference_parameter (list);
  reference_parameter (k);

  SCM lst = list;
  unsigned long int i;
  i = scm_to_ulong (k);
  while (scm_is_pair (lst)) {
    if (i == 0) {
      reference_count_return (lst);
      dereference_parameter (list);
      dereference_parameter (k);
      return SCM_CAR (lst);
    } else {
      --i;
      lst = TMP_CDR (lst);
    }
  };
  if (scm_is_null (lst))
    SCM_OUT_OF_RANGE (2, k);
  else
    SCM_WRONG_TYPE_ARG (1, list);
}
#undef FUNC_NAME


SCM_DEFINE (scm_list_set_x, "list-set!", 3, 0, 0,
	    (SCM list, SCM k, SCM val),
	    "Set the @var{k}th element of @var{list} to @var{val}.")
#define FUNC_NAME s_scm_list_set_x
{
  reference_parameter (list);
  reference_parameter (k);
  reference_parameter (val);

  SCM lst = list;
  unsigned long int i = scm_to_ulong (k);
  while (scm_is_pair (lst)) {
    if (i == 0) {
      SCM_SETCAR (lst, val);
      reference_count_return (val);
      dereference_parameter (list);
      dereference_parameter (k);
      dereference_parameter (val);
      return val;
    } else {
      --i;
      lst = TMP_CDR (lst);
    }
  };

  if (scm_is_null (lst)) {
    SCM_OUT_OF_RANGE (2, k);
  } else {
    SCM_WRONG_TYPE_ARG (1, list);
  }
}
#undef FUNC_NAME


SCM_REGISTER_PROC(s_list_cdr_ref, "list-cdr-ref", 2, 0, 0, scm_list_tail);

SCM_DEFINE (scm_list_tail, "list-tail", 2, 0, 0,
           (SCM k, SCM list),  // Viper Argument Order
	    "@deffnx {Scheme Procedure} list-cdr-ref lst k\n"
	    "Return the \"tail\" of @var{lst} beginning with its @var{k}th element.\n"
	    "The first element of the list is considered to be element 0.\n\n"
	    "@code{list-tail} and @code{list-cdr-ref} are identical.")
#define FUNC_NAME s_scm_list_tail
{
  reference_parameter (list);
  reference_parameter (k);

  SCM lst = list;
  size_t i = scm_to_size_t (k);
  while (i-- > 0) {
    SCM_VALIDATE_CONS (1, lst);
    lst = TMP_CDR(lst);
  }
  reference_count_return (lst);
  dereference_parameter (list);
  dereference_parameter (k);
  return lst;
}
#undef FUNC_NAME


SCM_DEFINE (scm_list_cdr_set_x, "list-cdr-set!", 3, 0, 0,
           (SCM list, SCM k, SCM val),
	    "Set the @var{k}th cdr of @var{list} to @var{val}.")
#define FUNC_NAME s_scm_list_cdr_set_x
{
  reference_parameter (list);
  reference_parameter (val);

  SCM lst = list;
  size_t i = scm_to_size_t (k);
  while (scm_is_pair (lst)) {
    if (i == 0) {
      SCM_SETCDR (lst, val);
      dereference_parameter (list);
      dereference_parameter (val);
      return val;
    } else {
      --i;
      lst = TMP_CDR (lst);
    }
  };
  if (scm_is_null (lst))
    SCM_OUT_OF_RANGE (2, k);
  else
    SCM_WRONG_TYPE_ARG (1, list);
}
#undef FUNC_NAME



SCM_DEFINE (scm_list_head, "list-head", 2, 0, 0,
           (SCM k, SCM list), // Viper Argument Order
	    "Copy the first @var{k} elements from @var{lst} into a new list, and\n"
	    "return it.")
#define FUNC_NAME s_scm_list_head
{
  size_t i = scm_to_size_t (k);

  reference_parameter (list);
  SCM current, last;
  SCM answer_box;

  answer_box = scm_cons (SCM_BOOL_F, SCM_EOL); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = list; reference_parameter (current);
  dereference_parameter (list);  // Release it on the fly.

  while (i-- > 0) {
    if (scm_is_null (current)) {
      SCM_SETCDR (last, SCM_EOL);
      break;
    }

    if (scm_is_pair (current)) {
      SCM_SETCDR (last, scm_cons (TMP_CAR (current), SCM_EOL));
      last = TMP_CDR (last);
      current = FLY_CDR (current);
    } else {
      SCM_VALIDATE_CONS (1, current);
    }
  }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  return (answer_box);
}
#undef FUNC_NAME


/* Copy a list which is known to be finite.  The last pair may or may not have
 * a '() in its cdr.  That is, improper lists are accepted.  */
SCM
scm_i_finite_list_copy (SCM list)
{
  reference_parameter (list);
  if (!scm_is_pair (list))
    {
      return rc_return_stack_reference (list);
    }
  else
    {
      SCM head = SCM_EOL;
      SCM tail = SCM_EOL;

      scm_i_copy_list (list, &head, &tail);
      reference_parameter (head);
      reference_parameter (tail);
      
      reference_count_return (head);
      dereference_parameter (head);
      dereference_parameter (tail);
      dereference_parameter (list);
      return head;
    }
}


SCM_DEFINE (scm_list_copy, "list-copy", 1, 0, 0, 
            (SCM lst),
	    "Return a (newly-created) copy of @var{lst}.")
#define FUNC_NAME s_scm_list_copy
{
  reference_parameter (lst);

  SCM head = SCM_EOL;
  SCM tail = SCM_EOL;

  SCM_VALIDATE_LIST (1, lst);

  scm_i_copy_list (lst, &head, &tail);
  reference_parameter (head);

  dereference_parameter (lst);
  debug_trim_assert_list (head);
  return (rc_return_stack_reference (head));
}
#undef FUNC_NAME


SCM_PROC (s_list, "list", 0, 0, 1, scm_list_copy);
SCM_SNARF_DOCS (primitive, scm_list_copy, "list", (SCM objs), 0, 0, 1,
                "Return a list containing @var{objs}, the arguments to\n"
                "@code{list}.")

/* This used to be the code for "list", but it's wrong when used via apply
   (it should copy the list).  */
/*
   SCM
   scm_list (SCM objs)
   {
     return objs;
   }
*/


/* membership tests (memq, memv, etc.) */ 

/* The function scm_c_memq returns the first sublist of list whose car is
 * 'eq?' obj, where the sublists of list are the non-empty lists returned by
 * (list-tail list k) for k less than the length of list.  If obj does not
 * occur in list, then #f (not the empty list) is returned.
 * List must be a proper list, otherwise scm_c_memq may crash or loop
 * endlessly.
 */
SCM
scm_c_memq (SCM x, SCM list)
{
  reference_parameter (x);
  reference_parameter (list);

  SCM lst = list;
  for (; !scm_is_null (lst); lst = TMP_CDR (lst))
    {
      if (scm_is_eq (TMP_CAR (lst), x)) {
        reference_count_return (lst);
        dereference_parameter (x);
        dereference_parameter (list);
	return lst;
      }
    }
  dereference_parameter (x);
  dereference_parameter (list);
  return SCM_BOOL_F;
}

SCM_DEFINE (scm_memq, "memq", 2, 0, 0,
           (SCM x, SCM list),
	    "Return the first sublist of @var{lst} whose car is @code{eq?}\n"
	    "to @var{x} where the sublists of @var{lst} are the non-empty\n"
	    "lists returned by @code{(list-tail @var{lst} @var{k})} for\n"
	    "@var{k} less than the length of @var{lst}.  If @var{x} does not\n"
	    "occur in @var{lst}, then @code{#f} (not the empty list) is\n"
	    "returned.")
#define FUNC_NAME s_scm_memq
{
  SCM_VALIDATE_LIST (2, list);
  return scm_c_memq (x, list);
}
#undef FUNC_NAME


SCM_DEFINE (scm_memv, "memv", 2, 0, 0,
           (SCM x, SCM list),
	    "Return the first sublist of @var{lst} whose car is @code{eqv?}\n"
	    "to @var{x} where the sublists of @var{lst} are the non-empty\n"
	    "lists returned by @code{(list-tail @var{lst} @var{k})} for\n"
	    "@var{k} less than the length of @var{lst}.  If @var{x} does not\n"
	    "occur in @var{lst}, then @code{#f} (not the empty list) is\n"
	    "returned.")
#define FUNC_NAME s_scm_memv
{
  reference_parameter (x);
  reference_parameter (list);

  SCM lst = list;
  SCM_VALIDATE_LIST (2, lst);
  for (; !scm_is_null (lst); lst = TMP_CDR (lst))
    {
      if (! scm_is_false (scm_eqv_p (TMP_CAR (lst), x))) {
        reference_count_return (lst);
        dereference_parameter (x);
        dereference_parameter (list);
	return lst;
      }
    }
  dereference_parameter (x);
  dereference_parameter (list);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


SCM_DEFINE (scm_member, "member", 2, 0, 0,
           (SCM x, SCM list),
	    "Return the first sublist of @var{lst} whose car is\n"
	    "@code{equal?} to @var{x} where the sublists of @var{lst} are\n"
	    "the non-empty lists returned by @code{(list-tail @var{lst}\n"
	    "@var{k})} for @var{k} less than the length of @var{lst}.  If\n"
	    "@var{x} does not occur in @var{lst}, then @code{#f} (not the\n"
	    "empty list) is returned.")
#define FUNC_NAME s_scm_member
{
  reference_parameter (x);
  reference_parameter (list);

  SCM lst = list;
  SCM_VALIDATE_LIST (2, lst);
  for (; !scm_is_null (lst); lst = TMP_CDR (lst))
    {
      if (! scm_is_false (scm_equal_p (TMP_CAR (lst), x))) {
        reference_count_return (lst);
        dereference_parameter (x);
        dereference_parameter (list);
	return lst;
      }
    }
  dereference_parameter (x);
  dereference_parameter (list);
  return SCM_BOOL_F;
}
#undef FUNC_NAME


/* deleting elements from a list (delq, etc.) */

SCM_DEFINE (scm_delq_x, "delq!", 2, 0, 0,
           (SCM item, SCM lst),
	    "@deffnx {Scheme Procedure} delv! item lst\n"
	    "@deffnx {Scheme Procedure} delete! item lst\n"
	    "These procedures are destructive versions of @code{delq}, @code{delv}\n"
	    "and @code{delete}: they modify the existing @var{lst}\n"
	    "rather than creating a new list.")
#define FUNC_NAME s_scm_delq_x
{
  reference_parameter (item);
  reference_parameter (lst);

  SCM current, last;
  SCM answer_box;

  answer_box = scm_cons (SCM_BOOL_F, lst); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = lst; reference_parameter (current);
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      if (scm_is_eq (TMP_CAR (current), item)) {
        SCM_SETCDR (last, TMP_CDR (current));
      } else {
	last = current;
      }
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (item);
  dereference_parameter (lst);
  return (answer_box);
}
#undef FUNC_NAME


SCM_DEFINE (scm_delv_x, "delv!", 2, 0, 0,
	    (SCM item, SCM lst),
	    "Destructively remove all elements from @var{lst} that are\n"
	    "@code{eqv?} to @var{item}.")
#define FUNC_NAME s_scm_delv_x
{
  reference_parameter (item);
  reference_parameter (lst);

  SCM current, last;
  SCM answer_box;

  answer_box = scm_cons (SCM_BOOL_F, lst); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = lst; reference_parameter (current);
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      if (! scm_is_false (scm_eqv_p (TMP_CAR (current), item))) {
        SCM_SETCDR (last, TMP_CDR (current));
      } else {
	last = current;
      }
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (item);
  dereference_parameter (lst);
  return (answer_box);
}
#undef FUNC_NAME



SCM_DEFINE (scm_delete_x, "delete!", 2, 1, 0,
	    (SCM item, SCM lst, SCM pred),
	    "Destructively remove all elements from @var{lst} that are\n"
	    "@code{equal?} to @var{item}.")
#define FUNC_NAME s_scm_delete_x
{
  reference_parameter (item);
  reference_parameter (lst);
  reference_parameter (pred);

  scm_t_trampoline_2 equal_p;
  if (!scm_i_is_list (lst)) {
    scm_wrong_type_arg (FUNC_NAME, 2, lst);
  }
  if (SCM_UNBNDP (pred)) {
    equal_p = equal_trampoline;
  } else {
    equal_p = scm_trampoline_2 (pred);
    SCM_ASSERT (equal_p, pred, 3, FUNC_NAME);
  }

  SCM current, last;
  SCM answer_box;

  answer_box = scm_cons (SCM_BOOL_F, lst); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = lst; reference_parameter (current);
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      if (!scm_is_false (equal_p (pred, TMP_CAR (current), item))) {
        SCM_SETCDR (last, TMP_CDR (current));
      } else {
	last = current;
      }
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (item);
  dereference_parameter (lst);
  dereference_parameter (pred);
  return (answer_box);
}
#undef FUNC_NAME




SCM_DEFINE (scm_delq, "delq", 2, 0, 0,
            (SCM item, SCM lst),
	    "Return a newly-created copy of @var{lst} with elements\n"
	    "@code{eq?} to @var{item} removed.  This procedure mirrors\n"
	    "@code{memq}: @code{delq} compares elements of @var{lst} against\n"
	    "@var{item} with @code{eq?}.")
#define FUNC_NAME s_scm_delq
{
  reference_parameter (lst);
  SCM copy = scm_list_copy (lst);
  dereference_parameter (lst);
  return scm_delq_x (item, copy);
}
#undef FUNC_NAME

SCM_DEFINE (scm_delv, "delv", 2, 0, 0,
            (SCM item, SCM lst),
	    "Return a newly-created copy of @var{lst} with elements\n"
	    "@code{eqv?}  to @var{item} removed.  This procedure mirrors\n"
	    "@code{memv}: @code{delv} compares elements of @var{lst} against\n"
	    "@var{item} with @code{eqv?}.")
#define FUNC_NAME s_scm_delv
{
  reference_parameter (lst);
  SCM copy = scm_list_copy (lst);
  dereference_parameter (lst);
  return scm_delv_x (item, copy);
}
#undef FUNC_NAME

SCM_DEFINE (scm_delete, "delete", 2, 1, 0,
            (SCM item, SCM lst, SCM pred),
	    "Return a newly-created copy of @var{lst} with elements\n"
	    "@code{equal?}  to @var{item} removed.  This procedure mirrors\n"
	    "@code{member}: @code{delete} compares elements of @var{lst}\n"
	    "against @var{item} with @code{equal?}.")
#define FUNC_NAME s_scm_delete
{
  reference_parameter (lst);
  SCM copy = scm_list_copy (lst);
  dereference_parameter (lst);
  return scm_delete_x (item, copy, pred);
}
#undef FUNC_NAME


SCM_DEFINE (scm_delq1_x, "delq1!", 2, 0, 0,
           (SCM item, SCM lst),
	    "Like @code{delq!}, but only deletes the first occurrence of\n"
	    "@var{item} from @var{lst}.  Tests for equality using\n"
	    "@code{eq?}.  See also @code{delv1!} and @code{delete1!}.")
#define FUNC_NAME s_scm_delq1_x
{
  reference_parameter (item);
  reference_parameter (lst);

  SCM current, last;
  SCM answer_box;

  answer_box = scm_cons (SCM_BOOL_F, lst); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = lst; reference_parameter (current); // bogus reference???
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      if (scm_is_eq (TMP_CAR (current), item))
	{
	  SCM_SETCDR (last, TMP_CDR (current));
	  break;
	}
      else
	last = current;
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (item);
  dereference_parameter (lst);
  return (answer_box);
}
#undef FUNC_NAME


SCM_DEFINE (scm_delv1_x, "delv1!", 2, 0, 0,
            (SCM item, SCM lst),
	    "Like @code{delv!}, but only deletes the first occurrence of\n"
	    "@var{item} from @var{lst}.  Tests for equality using\n"
	    "@code{eqv?}.  See also @code{delq1!} and @code{delete1!}.")
#define FUNC_NAME s_scm_delv1_x
{
  reference_parameter (item);
  reference_parameter (lst);

  SCM current, last;
  SCM answer_box;

  answer_box = scm_cons (SCM_BOOL_F, lst); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = lst; reference_parameter (current);
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      if (! scm_is_false (scm_eqv_p (TMP_CAR (current), item)))
	{
	  SCM_SETCDR (last, TMP_CDR (current));
	  break;
	}
      else
	last = current;
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (item);
  dereference_parameter (lst);
  return (answer_box);
}
#undef FUNC_NAME


SCM_DEFINE (scm_delete1_x, "delete1!", 2, 0, 0,
            (SCM item, SCM lst),
	    "Like @code{delete!}, but only deletes the first occurrence of\n"
	    "@var{item} from @var{lst}.  Tests for equality using\n"
	    "@code{equal?}.  See also @code{delq1!} and @code{delv1!}.")
#define FUNC_NAME s_scm_delete1_x
{
  reference_parameter (item);
  reference_parameter (lst);

  SCM current, last;
  SCM answer_box;

  answer_box = scm_cons (SCM_BOOL_F, lst); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = lst; reference_parameter (current);
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      if (! scm_is_false (scm_equal_p (TMP_CAR (current), item)))
	{
	  SCM_SETCDR (last, TMP_CDR (current));
	  break;
	}
      else
	last = current;
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (item);
  dereference_parameter (lst);
  return (answer_box);
}
#undef FUNC_NAME

SCM_DEFINE (scm_filter, "filter", 2, 0, 0,
	    (SCM pred, SCM list),
	    "Return all the elements of 2nd arg @var{list} that satisfy predicate @var{pred}.\n"
	    "The list is not disordered -- elements that appear in the result list occur\n"
	    "in the same order as they occur in the argument list. The returned list may\n"
	    "share a common tail with the argument list. The dynamic order in which the\n"
	    "various applications of pred are made is not specified.\n\n"
	    "@lisp\n"
	    "(filter even? '(0 7 8 8 43 -4)) => (0 8 8 -4)\n"
	    "@end lisp")
#define FUNC_NAME s_scm_filter
{
  reference_parameter (pred);
  reference_parameter (list);  // not strictly needed

  scm_t_trampoline_1 call = scm_trampoline_1 (pred);
  SCM current, last;
  SCM answer_box;
  SCM_ASSERT (call, pred, 1, FUNC_NAME);
  SCM_VALIDATE_LIST (2, list);
  
  answer_box = scm_cons (SCM_BOOL_F, SCM_EOL); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = list; reference_parameter (current);
  dereference_parameter (list);  // Release it on the fly.
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      SCM result = call (pred, TMP_CAR (current));
      if (scm_is_true (result))
	{
	  SCM_SETCDR (last, scm_cons (TMP_CAR (current), SCM_EOL));
          last = TMP_CDR (last);
	}
      dereference_return_value (result);
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (pred);
  return (answer_box);
}
#undef FUNC_NAME

SCM_DEFINE (scm_filter_x, "filter!", 2, 0, 0,
	    (SCM pred, SCM list),
	    "Linear-update variant of @code{filter}.")
#define FUNC_NAME s_scm_filter_x
{
  reference_parameter (pred);
  reference_parameter (list);

  scm_t_trampoline_1 call = scm_trampoline_1 (pred);
  SCM current, last;
  SCM answer_box;
  SCM_ASSERT (call, pred, 1, FUNC_NAME);
  SCM_VALIDATE_LIST (2, list);
  
  answer_box = scm_cons (SCM_BOOL_F, list); reference_parameter (answer_box);
  last = answer_box;  // temporary / list referenced
  current = list; reference_parameter (current);
  for (;
       scm_is_pair (current);
       current = FLY_CDR (current))
    {
      SCM result = call (pred, TMP_CAR (current));
      if (scm_is_true (result)) {
        last = current;
      } else {
	SCM_SETCDR (last, TMP_CDR (current));
      }
      dereference_return_value (result);
    }

  answer_box = FLY_CDR (answer_box);
  reference_count_return (answer_box);
  dereference_parameter (answer_box);
  dereference_parameter (current);
  dereference_parameter (pred);
  dereference_parameter (list);
  return (answer_box);
}
#undef FUNC_NAME


void
scm_init_list ()
{
#include "include/list.x"
}
