/*
 * VSI TRIM Double Cell Allocation (Copy & Paste from TRIM Allocation)
 */
#include "include/ihub.h"     // dependency loop on SCM at tags.h:10
#include "include/tags.h"     // SCM_EOL, scm_tc_free_cell
#include "include/trim.h"     // scm_t_cell

unsigned long global_double_cell_segments  = 0;

static SCM trim_global_double_free_list = TRIM_EOL;

/* To mix things up we define our own scm_cell_t not to be confused
   with the old scm_t_cell or the new scm_t. */
typedef struct scm_double_cell_t // XXX goes into trim.h
{
  scm_t_tag      cell_word_m1a;
  scm_t_refcount cell_word_m1b;
  SCM  cell_word_0;
  SCM  cell_word_1;
  SCM  cell_word_2;
  SCM  cell_word_3;
} scm_double_cell_t;

#define TRIM_DOUBLE_CELL_SIZE sizeof (scm_double_cell_t)

static void
drop_free_list ()
{
  trim_global_double_free_list   =    TRIM_EOL;
}

static void
add_free_list_segment_x (SCM head, SCM tail)
{
  TRIM_SET_CELL_WORD_1 (tail, trim_global_double_free_list);
  trim_global_double_free_list = head;
}


// Segment List
static void *trim_global_double_segment_list = (void*) TRIM_EOL;

#define default_double_segment_size 6000

#define NEXT_SEGMENT(x)          (*(void**)x)
#define SEGMENT_DOUBLE_CELLS(x)  (scm_double_cell_t*)(((char*)x)+sizeof(void*))
#define SET_NEXT_SEGMENT_X(x, y) (*(void**)x = y)
#define SEGMENT_EOL_P(x)         ((void*)x == (void*)TRIM_EOL)

static void add_segment (void *segment); // Inlining Prototype

static void
add_segment (void *segment) // this could be shared
{
  SET_NEXT_SEGMENT_X (segment, trim_global_double_segment_list);
  trim_global_double_segment_list = segment;
}

static size_t
trim_double_segment_list_length ()
{
  size_t answer = 0;
  void *segment = trim_global_double_segment_list;
  while (!SEGMENT_EOL_P (segment)) {
    segment = NEXT_SEGMENT (segment);
    answer++;
  }
  return answer;
}

// Returns 0 on zero saturation else non zero. This procedure is on
// the edge between SCM, cells and arbitray bits.
static int
add_freelist_segment (void *segment)
{
  scm_double_cell_t *cells = \
    (scm_double_cell_t*) SEGMENT_DOUBLE_CELLS (segment);
  SCM free_list_segment_tail = TRIM_EOL;
  register SCM free_list_segment_head = TRIM_EOL;
  register int zero_saturation = 0;
  register int n = default_double_segment_size - 1;

  // Scan the segment to initialize the tail of the free list segment.
  while (n >= 0) {
    if (FREE_CELL_P (PTR2SCM(&(cells[n])))) break;
    zero_saturation = 1;
    n--;
  }
  if (n >= 0) {
    free_list_segment_tail = PTR2SCM(&(cells[n]));
    free_list_segment_head = free_list_segment_tail;
    TRIM_SET_CELL_WORD_1 (free_list_segment_tail, TRIM_EOL);
    n--;
  }

  // Scan the segment constructing the free list segment and keeping
  // track of the zero saturation state.
  while (n >= 0) {
    if (!FREE_CELL_P (PTR2SCM(&(cells[n])))) {
      zero_saturation = 1;
      n--; continue;
    }
    TRIM_SET_CELL_WORD_1 (PTR2SCM (&(cells[n])), free_list_segment_head);
    free_list_segment_head = PTR2SCM (&(cells[n]));
    n--;
  }

  // Deliver the result depending on wether we found a segment with
  // zero saturation or not.
  if (zero_saturation != 0) {
    if (!TRIM_EOL_P (free_list_segment_head)) {
      add_free_list_segment_x (free_list_segment_head, free_list_segment_tail);
    }
    return 1;
  } else {
    return 0;
  }
}

void
trim_trim_double_allocation ()
{
  drop_free_list ();

  void *segment = trim_global_double_segment_list;
  void *last    = (void*) TRIM_EOL;

  while (!SEGMENT_EOL_P (segment)) {

    int zero_saturation  = add_freelist_segment (segment); 
    void *next = NEXT_SEGMENT (segment);

    if (zero_saturation == 0) {

      if (!SEGMENT_EOL_P (last))
        SET_NEXT_SEGMENT_X (last, next);

      if (segment == trim_global_double_segment_list)
        trim_global_double_segment_list = next;
      
      free (segment);
      global_double_cell_segments--;
      segment = next;

    } else {
      last = segment;
      segment = next;
    }
  }
  return;
}

static int
saturation (scm_double_cell_t *cells)
{
  int n = 0;
  int c = 0;
  while (n != default_double_segment_size) {
    if (!FREE_CELL_P (PTR2SCM(&(cells[n])))) {
      c++;
    }
    n++;
  }
  return c;
}

// Returns a saturation map as a list in allocation order.
SCM
trim_double_saturation_map ()
{
  void * segment  = trim_global_double_segment_list;
  size_t segments = trim_double_segment_list_length ();
  size_t index    = 0;

  int *counts     = malloc (segments * sizeof (int));
  if (!counts) return SCM_BOOL_F;

  while (index < segments) {
    if (!SEGMENT_EOL_P (segment)) {
      counts[index] = saturation (SEGMENT_DOUBLE_CELLS (segment));
    } else {
      counts[index] = -1;
    }
    segment = NEXT_SEGMENT (segment);
    index++;
  }

  SCM answer = SCM_EOL; // RC: consumed & pass through
  SCM segment_size_scm = scm_from_int (default_double_segment_size);
  reference_parameter (segment_size_scm);

  for (index = 0; index < segments; index++) {
    answer = scm_cons (scm_divide (scm_from_int (counts[index]),
                                   segment_size_scm),
                       answer);
  }
  // RC: answer pass through

  free (counts);
  dereference_parameter (segment_size_scm);
  return answer;
}



static void
s_double_alloc (int number_of_double_cells)
{
  void *cell_segment = \
    malloc ((TRIM_DOUBLE_CELL_SIZE * number_of_double_cells) + sizeof(void));
  if (cell_segment == NULL) {
    scm_i_memory_backup_ignition ();
    fprintf (stderr, " * VSI Fatal Error: memory allocation failed * \n");
    abort ();
  }
  scm_double_cell_t *segment = SEGMENT_DOUBLE_CELLS (cell_segment);
  int n = 0;


  /* Initialize the new segment: cell words 2 and 3 remain uninitialized */
  segment[0].cell_word_m1a = scm_tc_free_cell;
  segment[0].cell_word_m1b = 0;
  segment[0].cell_word_0  = 0;
  segment[0].cell_word_1  = TRIM_EOL;
  scm_double_cell_t *previous = &(segment[0]);
  for (n = 1; n < number_of_double_cells; n++) {
    segment[n].cell_word_m1a = scm_tc_free_cell;
    segment[n].cell_word_m1b = 0;
    segment[n].cell_word_0  = 0;
    segment[n].cell_word_1  = (SCM) previous;
    previous = &(segment[n]);
  }

  /* Install the new free list. */
  if (!(TRIM_EOL_P (trim_global_double_free_list))) {
    fprintf (stderr, "s_alloc: assertion failed: double free list not empty\n");
    exit (253); /* ??? */
  }
  trim_global_double_free_list = (SCM) &(segment[number_of_double_cells - 1]);

  /* Update the segment list */
  add_segment (cell_segment);
  return;
}

static inline void
assert_free_cell (SCM cell)
{
  if (TRIM_EOL_P (cell) || \
      !((SCM_CELL_TYPE (cell) == scm_tc_free_cell) \
        && (TRIM_REFCOUNT (cell) == 0))) {
    fprintf (stderr, " * VSI Fatal Error: assert_free_cell failed * ");
    abort ();
  }
}

static SCM inline
one_double_cell_please ()
{
  global_double_cells_requested++; // include this in the model ???

  if (TRIM_EOL_P (trim_global_double_free_list)) {
    s_double_alloc (default_double_segment_size);
    global_double_cell_segments++;
  }

  SCM cell = trim_global_double_free_list;
  // if (global_debug_interpreter) assert_free_cell (cell);
  assert_free_cell (cell);
  trim_global_double_free_list = SCM_FREE_CELL_CDR (cell);

  TRIM_SET_CELL_WORD_3 (cell, SCM_UNSPECIFIED);
  TRIM_SET_CELL_WORD_2 (cell, SCM_UNSPECIFIED);
  TRIM_SET_CELL_WORD_1 (cell, SCM_UNSPECIFIED);
  TRIM_SET_CELL_WORD_0 (cell, SCM_UNSPECIFIED);

  RC_RETURN (cell);
}

SCM inline
scm_double_cell (scm_t_tag tag,
                 scm_t_bits car_value, scm_t_bits cdr_value,
                 scm_t_bits cbr_value, scm_t_bits ccr_value)
{
  SCM cell = one_double_cell_please ();

  TRIM_SET_CELL_WORD_3 (cell, ccr_value);
  TRIM_SET_CELL_WORD_2 (cell, cbr_value);
  TRIM_SET_CELL_WORD_1 (cell, cdr_value);
  TRIM_SET_CELL_WORD_0 (cell, car_value);
  SCM_SET_CELL_TYPE (cell, tag);

  RC_RETURN (cell);
}


inline void
rc_release_reinsert_double (SCM double_cell)
{
  TRIM_SET_CELL_WORD_1 (double_cell, trim_global_double_free_list);
  trim_global_double_free_list = double_cell;
  return;
}


/*
 * Debugging Routines
 */
void
debug_trim_reset_double_allocation ()
{
  if (global_debug_interpreter)
    fprintf (stderr, "VSI-C: Explicit Reset of the Double Allocation Freelist\n");
  trim_global_double_free_list      = TRIM_EOL;
}
