#include "include/config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>

#include "include/_scm.h"
#include "include/eval.h"
#include "include/stime.h"
#include "include/smob.h"
#include "include/ports.h"
#include "include/root.h"
#include "include/byte-strings.h"
#include "include/brass.h"
#include "include/vectors.h"
#include "include/tags.h"

#include "include/validate.h"
#include "include/trim.h"


#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif


/* Function for non-cell memory management.
 */

inline void *
scm_realloc (void *mem, size_t size)
{

  void *ptr;
  ptr = realloc (mem, size);
  if (ptr == NULL) {
    scm_i_memory_backup_ignition ();
    fprintf (stderr, "VSI: scm_realloc: no reallocation now no no\n");
    abort ();
  }
  
  return ptr;
}

inline void *
scm_malloc (size_t sz)
{
  void *segment = NULL;

  segment = malloc (sz);
  if (segment == NULL) {
    scm_i_memory_backup_ignition ();
    fprintf (stderr, " * VSI: TRIM (TRIple M): Memory Manager Mi* ");
    abort ();
  }

  return segment;
}

void *
scm_calloc (size_t sz)
{
  void *segment = NULL;

  segment = calloc (sz, 1);
  if (segment == NULL) {
    scm_i_memory_backup_ignition ();
    fprintf (stderr, " * VSI: TRIM (TRIple M): Memory Manager M0* ");
    abort ();
  }

  return segment;
}


char *
scm_strndup (const char *str, size_t n)
{
  char *dst = scm_malloc (n + 1);
  memcpy (dst, str, n);
  dst[n] = 0;
  return dst;
}

char *
scm_strdup (const char *str)
{
  return scm_strndup (str, strlen (str));
}

char *
scm_gc_strndup (const char *str, size_t n, const char *what)
{
  char *dst = scm_gc_malloc (n+1, what);
  memcpy (dst, str, n);
  dst[n] = 0;
  return dst;
}

char *
scm_gc_strdup (const char *str, const char *what)
{
  return scm_gc_strndup (str, strlen (str), what);
}

