#include <stdio.h>
#include "include/trim.h"

#define ODD_P(x)  ((x) & 1)
#define EVEN_P(x) (!ODD_P(x))

/*
 * Simple Counting plus minus one
 */
inline void
reference_count_plus1 (SCM object)
{
  if (SCM_NIMP (object)) {
    TRIM_SET_REFCOUNT (object, (2 + TRIM_REFCOUNT (object)));
  }
  return;
}

inline void
reference_count_minus1 (SCM object)
{
  if (SCM_NIMP (object)) {
    TRIM_SET_REFCOUNT (object, (TRIM_REFCOUNT (object) - 2));

    if (TRIM_REFCOUNT (object) == 0)
      {
        rc_release (object);
      }
    else if (TRIM_REFCOUNT (object) < 0L)
      {
        if (global_debug_interpreter) {
          fprintf (stderr, "VSI TRIM/RC: miscount in minus1\n");
          abort ();
        }
      }
  }
  return;
}

static inline void
reference_count_plus_n (long n, SCM object)
{
  if (SCM_NIMP (object)) {
    TRIM_SET_REFCOUNT (object, ((2*n) + TRIM_REFCOUNT (object)));
  }
  return;
}

/*
 *
 */
void
rc_declare_data_graph_root (SCM object)
{
  reference_count_plus1 (object);
  dereference_return_value (object);

  return;
}

/*
 * Inter-Collectable Reference Declarations
 */
inline void
reference (SCM referrent, SCM referred)
{
  if (referrent != referred) {
    reference_count_plus1 (referred);
  }
  /* else { */
  /*   if (global_debug_interpreter) */
  /*     fprintf (stderr, "VSI TRIM/RC: reference to self ignored\n"); */
  /* } */

  return;
}

inline void
dereference (SCM referrent, SCM referred)
{
  if (referrent != referred) {
    reference_count_minus1 (referred);
  }
  /*  else { */
  /*   if (global_debug_interpreter) */
  /*     fprintf (stderr, "VSI TRIM/RC: not dereferencing inner circle\n"); */
  /* } */

  return;
}

inline void
multiple_reference (SCM referrent, long n, SCM referred)
{
  if (referrent != referred) {
    reference_count_plus_n (n, referred);
  }

  return;
}


/*
 * Stack Reference Declarations
 */

inline SCM
reference_count_return (SCM object)
{
  /* if (SCM_NIMP (object) && EVEN_P (TRIM_REFCOUNT (object))) */
  /*   TRIM_SET_REFCOUNT (object, (1 + TRIM_REFCOUNT (object))); */

  if (SCM_NIMP (object))
    TRIM_SET_REFCOUNT (object, (1 | TRIM_REFCOUNT (object)));

  return object;
}

inline void
reference_count_parameter_up (SCM object)
{
  if (!SCM_NIMP (object))
    return;

  if (ODD_P (TRIM_REFCOUNT (object))) {
    /* return value -> stack reference */
    TRIM_SET_REFCOUNT (object, (1 + TRIM_REFCOUNT (object)));

  } else { // count is EVEN_P
    /* one-plus stack reference */
    TRIM_SET_REFCOUNT (object, (2 + TRIM_REFCOUNT (object)));
  }

  return;
}

/* stack reference -> return reference XXX rename please */
inline SCM
rc_return_stack_reference (SCM object)
{
  reference_count_return ((object));
  reference_count_minus1 ((object));

  return (object);
}

inline void
dereference_return_value (SCM object) // aka normalize down
{
  if (!SCM_NIMP (object))
    return;

  if (ODD_P (TRIM_REFCOUNT (object))) {
    TRIM_SET_REFCOUNT (object, (TRIM_REFCOUNT (object) - 1));
    if (TRIM_REFCOUNT (object) == 0) {
      rc_release (object);
    }
  }
}
  

inline SCM
rc_let_parameter_fly (SCM object)
{
  reference_count_return (object);
  dereference_parameter  (object);
  return object;
}

/* It would probably have been better to wrap the
 * non passing functions as macros around those
 * which pass the their parameter back as their
 * return value.
 */
inline SCM
rc_count_parameter_up_and_pass (SCM object)
{
  reference_parameter (object);
  return object;
}
