#include "include/config.h"

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <assert.h>

#include "include/_scm.h"
#include "include/init.h"
#include "include/eval.h"
#include "include/stime.h"
#include "include/smob.h"
#include "include/ports.h"
#include "include/root.h"
#include "include/byte-strings.h"
#include "include/brass.h"
#include "include/vectors.h"
#include "include/tags.h"

#include "include/validate.h"
#include "include/trim.h"

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

void *global_memory_backup_segment = NULL;

/*
 * Statistics
 */
unsigned long global_cells_requested      = 0;
unsigned long global_last_cells_requested = 0;

SCM_DEFINE (scm_global_cells_requested, "%global-cells-requested", 0, 0, 0,
            (),
            "Returns the total number of cells requested.")
{
  return scm_from_ulong (global_cells_requested);
}

SCM_DEFINE (scm_global_cells_requested_delta,
            "%global-cells-requested-delta", 0, 0, 0,
            (),
            "Returns the number of cells requested since the last call to this function.")
{
  unsigned long delta;

  delta = global_cells_requested - global_last_cells_requested;
  global_last_cells_requested = global_cells_requested;

  return scm_from_ulong (delta);
}

unsigned long global_double_cells_requested      = 0;
unsigned long global_last_double_cells_requested = 0;

SCM_DEFINE (scm_global_double_cells_requested, "%global-double-cells-requested", 0, 0, 0,
            (),
            "Returns the total number of double cells requested.")
{
  return scm_from_ulong (global_double_cells_requested);
}

SCM_DEFINE (scm_global_double_cells_requested_delta,
            "%global-double-cells-requested-delta", 0, 0, 0,
            (),
            "Returns the number of double cells requested since the last call to this function.")
{
  unsigned long delta;

  delta = global_double_cells_requested - global_last_double_cells_requested;
  global_last_double_cells_requested = global_double_cells_requested;

  return scm_from_ulong (delta);
}



SCM_SYMBOL (sym_protected_objects, "protected-objects");


SCM
scm_permanent_object (SCM obj)
{
  heap_reference (&scm_permanent_object, obj);
  RC_RETURN (obj);
}


int scm_i_terminating;


void
scm_i_memory_backup_ignition ()
{
  if (global_memory_backup_segment) {
    free (global_memory_backup_segment);
    global_memory_backup_segment = NULL;
    scm_out_of_memory ();
  }
  return;
}

SCM_DEFINE (scm_trim, "trim", 0, 0, 0, (), "")
{
  trim_trim_phalanx ();
  trim_trim_allocation ();
  trim_trim_double_allocation ();
  malloc_trim (0); // 0 means minimum else some space
  return SCM_UNSPECIFIED;
}

SCM_DEFINE (trim_debug_purple, "debug-purple", 2, 0, 0,
            (SCM object, SCM ticks), "")
#define FUNC_NAME s_trim_debug_purple
{ // RC lax (sic!)
  int nticks; SCM_VALIDATE_INT_COPY (1, ticks, nticks);
  if (!SCM_NIMP (object)) {
    scm_wrong_type_arg_msg (FUNC_NAME, 1, object,
                            "non-immediate-memory-object");
  }

  // trim_i_debug_purple (object, nticks); // depends on accessors in trim.h

  double m1 = TRIM_REFCOUNT (object) / 2.0;
  return scm_from_double (m1);
}
#undef FUNC_NAME

SCM_DEFINE (scm_trim_saturation_map, "trim-saturation-map", 0, 0, 0, (), "")
{
  return (scm_list_2 (trim_saturation_map (),
                      trim_double_saturation_map ()));
}

SCM_DEFINE (scm_phalanx_maps, "trim-phalanx-maps", 0, 0, 0, (), "")
{
  return (phalanx_saturation_maps());
}

void
scm_storage_prehistory ()
{
}


void
scm_init_storage ()
{
  size_t j;

  if (global_debug_interpreter) {
    fprintf (stderr, "VSI initialization trace: scm_init_storage\n");
  }

  j = SCM_NUM_PROTECTS;
  while (j)
    scm_sys_protects[--j] = SCM_BOOL_F;

  init_phalanx ();

  global_memory_backup_segment = calloc (1000000, sizeof (SCM));
  if (!global_memory_backup_segment && global_debug_interpreter) {
    fprintf (stderr, "VSI: TRIM warning: failed to allocate backup memory\n");
  }
}


void
x_inclusion_storage ()
{
#include "include/trim.x"
}
