// This file of VSI-C had originally been written by a wingding and it
// kept it under the wings ... and then we just copied and pasted. LOL
// Others had found evidence which suggested this file documents the
// revenge of the evil parity bit. Just because they bumped it out
// completely ...

#include "include/_scm.h"
#include "include/chars.h"
#include "include/brass.h"
#include "include/byte-strings.h"
#include "include/inline.h" // getc / read-byte

#define utf8_2byte_mask   (uint8_t)0b11100000
#define utf8_2byte_tag    (uint8_t)0b11000000

#define utf8_3byte_mask   (uint8_t)0b11110000
#define utf8_3byte_tag    (uint8_t)0b11100000

#define utf8_4byte_mask   (uint8_t)0b11111000
#define utf8_4byte_tag    (uint8_t)0b11110000

#define utf8_sbyte_mask   (uint8_t)0b11000000
#define utf8_sbyte_tag    (uint8_t)0b10000000


/* UTF-8 Encoding */

#define utf8_boff_first   0
#define utf8_boff_second  6
#define utf8_boff_third   (6 + utf8_boff_second)
#define utf8_boff_fourth  (4 + utf8_boff_third)

#define utf8_mask_first   (uint32_t)(0b00111111U)
#define utf8_mask_second  (uint32_t)(0b00111111U << utf8_boff_second)
#define utf8_mask_third   (uint32_t)(0b00001111U << utf8_boff_third)
#define utf8_mask_fourth  (uint32_t)(0b00011111U << utf8_boff_fourth)

#define utf8_4byte_lim    2097151UL
#define utf8_3byte_lim      65535UL
#define utf8_2byte_lim       2047UL
#define utf8_1byte_lim        127UL

SCM
utf8_usv_to_bs (uint32_t usv)
{
  uint32_t usv1 = usv & utf8_mask_first;
  uint32_t usv2 = usv & utf8_mask_second;
  uint32_t usv3 = usv & utf8_mask_third;
  uint32_t usv4 = usv & utf8_mask_fourth;

  if (usv > utf8_4byte_lim) {
    return SCM_BOOL_F;

  } else if (usv > utf8_3byte_lim) {
    // uint8_t utf[4]; utf[0] = 0; utf[1] = 0; utf[2] = 0; utf[3] = 0;
    SCM ans = scm_c_make_byte_string (4);
    uint8_t *utf = BYTE_STRING_DATA (ans);
    utf[0] = utf8_4byte_tag | (uint8_t)(usv4 >> (utf8_boff_fourth + 2));
    utf[1] = utf8_sbyte_tag | (uint8_t)((usv3 >> utf8_boff_third) \
                                        | ( 48  \
                                           & (usv4 >> (utf8_boff_fourth - 4))));
    utf[2] = utf8_sbyte_tag | (uint8_t)(usv2 >> utf8_boff_second);
    utf[3] = utf8_sbyte_tag | (uint8_t)(usv1 >> utf8_boff_first);
    return ans;

  } else if (usv > utf8_2byte_lim) {
    // uint8_t utf[3]; utf[0] = 0; utf[1] = 0; utf[2] = 0;
    SCM ans = scm_c_make_byte_string (3);
    uint8_t *utf = BYTE_STRING_DATA (ans);
    utf[0] = utf8_3byte_tag | (uint8_t)(usv3 >> utf8_boff_third);
    utf[1] = utf8_sbyte_tag | (uint8_t)(usv2 >> utf8_boff_second);
    utf[2] = utf8_sbyte_tag | (uint8_t)(usv1 >> utf8_boff_first);
    return ans;

  } else if (usv > utf8_1byte_lim) {
    // uint8_t utf[2]; utf[0] = 0; utf[1] = 0;
    SCM ans = scm_c_make_byte_string (2);
    uint8_t *utf = BYTE_STRING_DATA (ans);
    utf[0] = utf8_2byte_tag | (uint8_t)(usv2 >> utf8_boff_second);
    utf[1] = utf8_sbyte_tag | (uint8_t)(usv1 >> utf8_boff_first);
    return ans;

  } else if (usv <= utf8_1byte_lim) {
    // uint8_t utf = (uint8_t) usv;
    SCM ans = scm_c_make_byte_string (1);
    uint8_t *utf = BYTE_STRING_DATA (ans);
    utf[0] = (uint8_t) usv;
    return ans;
  }

  fprintf (stderr, " * VSI-C: utf8_usv_to_bs: "
                   "Control may reach end of non void function? *\n ");
  abort (); return SCM_UNSPECIFIED;
}

// Array encoding

inline size_t
utf8_usv_to_bytes (uint32_t usv, uint8_t *utf)
{
  uint32_t usv1 = usv & utf8_mask_first;
  uint32_t usv2 = usv & utf8_mask_second;
  uint32_t usv3 = usv & utf8_mask_third;
  uint32_t usv4 = usv & utf8_mask_fourth;

  if (usv > utf8_4byte_lim) {
    return 0;

  } else if (usv > utf8_3byte_lim) {
    // uint8_t utf[4]; utf[0] = 0; utf[1] = 0; utf[2] = 0; utf[3] = 0;
    *utf = utf8_4byte_tag | (uint8_t)(usv4 >> (utf8_boff_fourth + 2));
    utf++;
    *utf = utf8_sbyte_tag | (uint8_t)((usv3 >> utf8_boff_third) \
                                        | ( 48  \
                                           & (usv4 >> (utf8_boff_fourth - 4))));
    utf++;
    *utf = utf8_sbyte_tag | (uint8_t)(usv2 >> utf8_boff_second);
    utf++;
    *utf = utf8_sbyte_tag | (uint8_t)(usv1 >> utf8_boff_first);
    return 4;

  } else if (usv > utf8_2byte_lim) {
    // uint8_t utf[3]; utf[0] = 0; utf[1] = 0; utf[2] = 0;
    *utf = utf8_3byte_tag | (uint8_t)(usv3 >> utf8_boff_third);
    utf++;
    *utf = utf8_sbyte_tag | (uint8_t)(usv2 >> utf8_boff_second);
    utf++;
    *utf = utf8_sbyte_tag | (uint8_t)(usv1 >> utf8_boff_first);
    return 3;

  } else if (usv > utf8_1byte_lim) {
    // uint8_t utf[2]; utf[0] = 0; utf[1] = 0;
    *utf = utf8_2byte_tag | (uint8_t)(usv2 >> utf8_boff_second);
    utf++;
    *utf = utf8_sbyte_tag | (uint8_t)(usv1 >> utf8_boff_first);
    return 2;

  } else if (usv <= utf8_1byte_lim) {
    // uint8_t utf = (uint8_t) usv;
    *utf = (uint8_t) usv;
    return 1;
  }

  fprintf (stderr, " * VSI-C: utf8_usv_to_bytes: "
                   "Control may reach end of non void function? *\n ");
  abort (); return SCM_UNSPECIFIED;
}

inline size_t
utf8_usv_num_bytes (uint32_t usv)
{
  if (usv < (utf8_1byte_lim + 1)) return 1;
  if (usv < (utf8_2byte_lim + 1)) return 2;
  if (usv < (utf8_3byte_lim + 1)) return 3;
  if (usv < (utf8_4byte_lim + 1)) return 4;

  return 0;
}


/* UTF-8 Decoding */

static int      sequence_byteP (uint8_t byte);
static int      utf8_2byteP    (uint8_t byte);
static int      utf8_3byteP    (uint8_t byte);
static int      utf8_4byteP    (uint8_t byte);
static uint32_t utf8_sequence_bits (uint8_t byte);

static inline int
utf8_sbyteP (uint8_t byte) {return (utf8_sbyte_mask & byte) == utf8_sbyte_tag;}
static inline int
utf8_2byteP (uint8_t byte) {return (utf8_2byte_mask & byte) == utf8_2byte_tag;}
static inline int
utf8_3byteP (uint8_t byte) {return (utf8_3byte_mask & byte) == utf8_3byte_tag;}
static inline int
utf8_4byteP (uint8_t byte) {return (utf8_4byte_mask & byte) == utf8_4byte_tag;}

static inline uint32_t
utf8_sequence_bits (uint8_t byte)
{
  if (utf8_sbyteP (byte)) {
    return (~utf8_sbyte_mask) & byte;
  } else {
    // shouldn't happen but happend in the model while testing the
    // encoding ???
    scm_encoding_error ("decoding utf8_read_char");
    return 0; // never reached
  }
}

static SCM utf8_scm_from_bytes (uint32_t usv1, uint32_t usv2,
                                uint32_t usv3, uint32_t usv4);

static inline SCM
utf8_scm_from_bytes (uint32_t usv1, uint32_t usv2,
                     uint32_t usv3, uint32_t usv4)
{
  return SCM_MAKE_UC_CHAR (usv1 | (usv2 >> 2) | (usv3 >> 4) | (usv4 >> 6));
}

/* UTF-8 Stream Decoding */

static SCM utf8_deco_four_byte (SCM port, int sel,
                                uint32_t usv1, uint32_t usv2,
                                uint32_t usv3, uint32_t usv4);
static inline SCM 
utf8_deco_four_byte (SCM port, int sel,
                     uint32_t usv1, uint32_t usv2,
                     uint32_t usv3, uint32_t usv4)
{
  usv4 = ((uint32_t)((~utf8_4byte_mask) & (uint8_t)sel)) << 24;
  sel = scm_i_read_byte (port);

  if (EOF == sel) scm_encoding_error ("decoding utf8_read_char underflow");

  if (utf8_sbyteP (sel)) {
    usv3 = ((uint32_t)(utf8_sequence_bits (sel))) << 16;
    sel = scm_i_read_byte (port);
    if (EOF == sel) scm_encoding_error ("decoding utf8_read_char underflow");

    if (utf8_sbyteP (sel)) {
      usv2 = ((uint32_t)(utf8_sequence_bits (sel))) << 8;
      sel = scm_i_read_byte (port);
      if (EOF == sel) scm_encoding_error ("decoding utf8_read_char underflow");

      if (utf8_sbyteP (sel)) {
        usv1 = ((uint32_t)(utf8_sequence_bits (sel))) << 0;
        return utf8_scm_from_bytes (usv1, usv2, usv3, usv4);
      }
    }
  }
  scm_encoding_error ("decoding utf8_read_char bad sequence");
}

static SCM 
utf8_deco_three_byte (SCM port, int sel,
                      uint32_t usv1, uint32_t usv2, uint32_t usv3);
static inline SCM 
utf8_deco_three_byte (SCM port, int sel,
                      uint32_t usv1, uint32_t usv2, uint32_t usv3)
{
  usv3 = ((uint32_t)((~utf8_3byte_mask) & (uint8_t)sel)) << 16;
  sel = scm_i_read_byte (port);

  if (EOF == sel) scm_encoding_error ("decoding utf8_read_char underflow");

  if (utf8_sbyteP (sel)) {
    usv2 = ((uint32_t)(utf8_sequence_bits (sel))) << 8;
    sel = scm_i_read_byte (port);
    if (EOF == sel) scm_encoding_error ("decoding utf8_read_char underflow");

    if (utf8_sbyteP (sel)) {
      usv1 = ((uint32_t)(utf8_sequence_bits (sel))) << 0;
      return utf8_scm_from_bytes (usv1, usv2, usv3, 0);
    }
  }
  scm_encoding_error ("decoding utf8_read_char bad sequence");
}

static SCM 
utf8_deco_two_byte (SCM port, int sel, uint32_t usv1, uint32_t usv2);
static inline SCM 
utf8_deco_two_byte (SCM port, int sel, uint32_t usv1, uint32_t usv2)
{
  usv2 = ((uint32_t)((~utf8_2byte_mask) & (uint8_t)sel)) << 8;
  sel = scm_i_read_byte (port);

  if (EOF == sel) scm_encoding_error ("decoding utf8_read_char underflow");

  if (utf8_sbyteP (sel)) {
    usv1 = ((uint32_t)(utf8_sequence_bits (sel))) << 0;
    return utf8_scm_from_bytes (usv1, usv2, 0, 0);
  }
  scm_encoding_error ("decoding utf8_read_char bad sequence");
}

SCM
utf8_read_char (SCM port)
{
  uint32_t usv1 = 0;
  uint32_t usv2 = 0;
  uint32_t usv3 = 0;
  uint32_t usv4 = 0;
  int sel = scm_i_read_byte (port);

  if (EOF == sel) {
    return SCM_EOF_VAL;

  } else if (utf8_4byteP (sel)) {
    return utf8_deco_four_byte (port, sel, usv1, usv2, usv3, usv4);

  } else if (utf8_3byteP (sel)) {
    return utf8_deco_three_byte (port, sel, usv1, usv2, usv3);

  } else if (utf8_2byteP (sel)) {
    return utf8_deco_two_byte (port, sel, usv1, usv2);

  } else if (sel < 128) {
    return SCM_MAKE_CHAR (sel);

  } else {
    scm_encoding_error ("decoding utf8_read_char bad terminator");
  }
}

/* UTF-8 Array Decoding */
static inline uint32_t
utf8_usv_from_bytes (uint32_t usv1, uint32_t usv2,
                     uint32_t usv3, uint32_t usv4)
{
  return (usv1 | (usv2 >> 2) | (usv3 >> 4) | (usv4 >> 6));
}

static uint32_t utf8_adeco_four_byte (char **string, int sel,
                                 uint32_t usv1, uint32_t usv2,
                                 uint32_t usv3, uint32_t usv4);
static inline uint32_t 
utf8_adeco_four_byte (char **string, int sel,
                      uint32_t usv1, uint32_t usv2,
                      uint32_t usv3, uint32_t usv4)
{
  usv4 = ((uint32_t)((~utf8_4byte_mask) & (uint8_t)sel)) << 24;
  sel = **string; *string = *string + 1;

  if (utf8_sbyteP (sel)) {
    usv3 = ((uint32_t)(utf8_sequence_bits (sel))) << 16;
    sel = **string; *string = *string + 1;

    if (utf8_sbyteP (sel)) {
      usv2 = ((uint32_t)(utf8_sequence_bits (sel))) << 8;
      sel = **string; *string = *string + 1;

      if (utf8_sbyteP (sel)) {
        usv1 = ((uint32_t)(utf8_sequence_bits (sel))) << 0;
        return utf8_usv_from_bytes (usv1, usv2, usv3, usv4);
      }
    }
  }
  scm_encoding_error ("decoding utf8_next_char bad sequence");
}

static uint32_t 
utf8_adeco_three_byte (char **string, int sel,
                       uint32_t usv1, uint32_t usv2, uint32_t usv3);
static inline uint32_t 
utf8_adeco_three_byte (char **string, int sel,
                       uint32_t usv1, uint32_t usv2, uint32_t usv3)
{
  usv3 = ((uint32_t)((~utf8_3byte_mask) & (uint8_t)sel)) << 16;
  sel = **string; *string = *string + 1;

  if (utf8_sbyteP (sel)) {
    usv2 = ((uint32_t)(utf8_sequence_bits (sel))) << 8;
    sel = **string; *string = *string + 1;

    if (utf8_sbyteP (sel)) {
      usv1 = ((uint32_t)(utf8_sequence_bits (sel))) << 0;
      return utf8_usv_from_bytes (usv1, usv2, usv3, 0);
    }
  }
  scm_encoding_error ("decoding utf8_next_char bad sequence");
}

static uint32_t 
utf8_adeco_two_byte (char **string, int sel, uint32_t usv1, uint32_t usv2);
static inline uint32_t 
utf8_adeco_two_byte (char **string, int sel, uint32_t usv1, uint32_t usv2)
{
  usv2 = ((uint32_t)((~utf8_2byte_mask) & (uint8_t)sel)) << 8;
  sel = **string;
  *string = *string + 1;

  if (utf8_sbyteP (sel)) {
    usv1 = ((uint32_t)(utf8_sequence_bits (sel))) << 0;
    return utf8_usv_from_bytes (usv1, usv2, 0, 0);
  }
  scm_encoding_error ("decoding utf8_next_char bad sequence");
}

static usv_t
utf8_next_char (char **string, size_t *num_bytes)
{
  uint32_t usv1 = 0;
  uint32_t usv2 = 0;
  uint32_t usv3 = 0;
  uint32_t usv4 = 0;
  int sel = **string; *string = *string + 1;

  if (utf8_4byteP (sel)) {
    *num_bytes = *num_bytes + 4;
    return utf8_adeco_four_byte (string, sel, usv1, usv2, usv3, usv4);

  } else if (utf8_3byteP (sel)) {
    *num_bytes = *num_bytes + 3;
    return utf8_adeco_three_byte (string, sel, usv1, usv2, usv3);

  } else if (utf8_2byteP (sel)) {
    *num_bytes = *num_bytes + 2;
    return utf8_adeco_two_byte (string, sel, usv1, usv2);

  } else if (sel < 128) {
    *num_bytes = *num_bytes + 1;
    return sel;

  } else {
    scm_encoding_error ("decoding utf8_next_char bad terminator");
  }
}

static size_t
utf8_length (char *string, size_t num_bytes)
{
  unsigned char *p = (unsigned char*) string;
  unsigned char *f = (unsigned char*) string + num_bytes;
  size_t res = 0;

  while (p < f) {
         if ((*p) < 128)       p++;
    else if (utf8_2byteP (*p)) p += 2;
    else if (utf8_3byteP (*p)) p += 3;
    else if (utf8_4byteP (*p)) p += 4;
    else scm_encoding_error ("decoding utf8_length");

    res++;
  }

  return res;
}

SCM
utf8_to_brass (const char *str, size_t num_bytes)
{
  char *string = str;
  size_t len   = utf8_length (string, num_bytes);
  int zero_pad = 0;

  if (len > 0) {
    zero_pad = (string[num_bytes-1] == '\0');
         len = (zero_pad ? len - 1 : len);
  }
  SCM br     = scm_c_make_brass (len);
  usv_t *cp  = BR_CHARS (br);
  size_t pb  = 0;

  while (len != 0) {
    *cp = utf8_next_char (&string, &pb);
    cp++; len--;
  }
  // if (!zero_pad) *cp = utf8_next_char (&string, &pb);

  if (!((zero_pad && num_bytes == pb+1) || num_bytes == pb))
    scm_encoding_error ("decoding utf8_to_brass");

  return br;
}
