#include "include/config.h"

#include "include/_scm.h"
#include "include/eq.h"
#include "include/ports.h"
#include "include/root.h"
#include "include/smob.h"

#include "include/validate.h"
#include "include/variable.h"

/*
 * Variable Printer
 */
void
scm_i_variable_print (SCM exp, SCM port, scm_print_state *pstate)
{
  scm_puts ("#<variable ", port);
  scm_uintprint (SCM_UNPACK (exp), 16, port);
  scm_puts (" value: ", port);
  scm_iprin1 (TMP_VARIABLE_REF (exp), port, pstate);
  scm_putc('>', port);
}

/*
 * Variable Statistics
 */
long global_number_of_variables = 0;
long global_last_number_of_variables    = 0;

SCM_DEFINE (scm_global_variable_count, "global-variable-count", 0, 0, 0,
            (),
            "Returns the total number of variables.")
{
  return scm_from_long (global_number_of_variables);
}

SCM_DEFINE (scm_global_variable_count_delta,
            "global-variable-count-delta", 0, 0, 0, (), "")
{
  long delta;

  delta = global_number_of_variables - global_last_number_of_variables;
  global_last_number_of_variables = global_number_of_variables;

  return scm_from_long (delta);
}


/*
 * Variables
 */
static SCM
make_variable (SCM init)
{
  reference_parameter (init);

  SCM var = scm_cell (scm_tc7_variable, (scm_t_bits) 0, SCM_UNPACK (init));
  reference (var, init);
  /* SCM port = scm_current_error_port (); */
  /* if (port != NULL && !scm_is_false (port)) /\* would happen in init *\/ */
  /*   scm_display (init, scm_current_error_port ()); */
  /* else */
  /*   fprintf (stderr, "£"); */
  global_number_of_variables++; 

  dereference_parameter (init);
  return var;
}

SCM_DEFINE (scm_make_variable, "make-variable", 1, 0, 0, 
            (SCM init),
            "Return a variable initialized to value @var{init}.")
#define FUNC_NAME s_scm_make_variable
{
  return make_variable (init);
}
#undef FUNC_NAME


SCM_DEFINE (scm_make_undefined_variable, "make-undefined-variable", 0, 0, 0, 
            (),
            "Return a variable that is initially unbound.")
#define FUNC_NAME s_scm_make_undefined_variable
{
  return make_variable (SCM_UNDEFINED);
}
#undef FUNC_NAME


SCM_DEFINE (scm_variable_p, "variable?", 1, 0, 0, 
            (SCM obj),
            "Return @code{#t} iff @var{obj} is a variable object, else\n"
	    "return @code{#f}.")
#define FUNC_NAME s_scm_variable_p
{
  return scm_from_bool (SCM_VARIABLEP (obj));
}
#undef FUNC_NAME


SCM_DEFINE (scm_variable_ref, "variable-ref", 1, 0, 0, (SCM var), "")
#define FUNC_NAME s_scm_variable_ref
{
  reference_parameter (var);

  SCM val;
  SCM_VALIDATE_VARIABLE (1, var);
  val = SCM_VARIABLE_REF (var);
  if (val == SCM_UNDEFINED)
    SCM_MISC_ERROR ("variable is unbound: ~S", scm_list_1 (var));

  dereference_parameter (var);
  return val;
}
#undef FUNC_NAME

SCM_DEFINE (scm_variable_set_x, "variable-set!", 2, 0, 0,
            (SCM var, SCM val),
            "Set the value of the variable @var{var} to @var{val}.\n"
            "@var{var} must be a variable object, @var{val} can be any\n"
	    "value. Return an unspecified value.")
#define FUNC_NAME s_scm_variable_set_x
{
  SCM_VALIDATE_VARIABLE (1, var);
  SCM_VARIABLE_SET (var, val);
  return SCM_UNSPECIFIED;
}
#undef FUNC_NAME

inline void
scm_i_variable_set_x (SCM variable, SCM value)
{
  reference_parameter (variable);
  reference_parameter (value);

  reference (variable, value);
  dereference (variable, SCM_CELL_OBJECT_1 (variable));
  SCM_SET_CELL_OBJECT_1 (variable, value);

  dereference_parameter (variable);
  dereference_parameter (value);
  return;
}

inline SCM
scm_i_variable_ref (SCM variable)
{
  SCM answer = SCM_CELL_OBJECT_1 (variable);
  reference_parameter (answer);

  dereference_return_value (variable);
  return rc_return_stack_reference (answer);
}

SCM_DEFINE (scm_variable_bound_p, "variable-bound?", 1, 0, 0, 
            (SCM var),
            "Return @code{#t} iff @var{var} is bound to a value.\n"
            "Throws an error if @var{var} is not a variable object.")
#define FUNC_NAME s_scm_variable_bound_p
{
  reference_parameter (var);
  SCM_VALIDATE_VARIABLE (1, var);

  int answer = (TMP_VARIABLE_REF (var) != SCM_UNDEFINED);

  dereference_parameter (var);
  return (scm_from_bool (answer));
}
#undef FUNC_NAME

void
trim_release_variable (SCM variable)
{
  global_number_of_variables--;

  reference_count_minus1 (TMP_VARIABLE_REF (variable));
  return;
}

void
scm_init_variable ()
{
#include "include/variable.x"
}
