(define (byte-string . bytes) (list->byte-string bytes))

(define (byte-reverse byte-string)
  (byte-reverse! (byte-string-copy byte-string)))

(define (bit-not byte-string)
  (bit-not! (byte-string-copy byte-string)))

(define (bit-reverse byte-string)
  (bit-reverse! (byte-string-copy byte-string)))

(define (bit-shift-left  byte-string offset)
  (bit-shift-left!  (byte-string-copy byte-string) offset))

(define (bit-shift-right byte-string offset)
  (bit-shift-right! (byte-string-copy byte-string) offset))

(define (signed-right-shift! n r) ; see SCM_SRS in numbers.h
 (if (< (byte-string->number+sign n) 0)
  (bit-not! (bit-shift-right! (bit-not! n) r))
  (bit-shift-right! n r)))
(define (signed-right-shift n r)
 (signed-right-shift! (byte-string-copy n) r))

;;
;; Domain Boundary: Boolean

(define (byte-string-false? byte-string)
  (= 0 (byte-string->number byte-string)))

(define (byte-string-true? byte-string)
  (not (byte-string-false? byte-string)))

(define byte-string->boolean byte-string-true?)
  
(define (bit-and? bits mask) ; promote?
  (byte-string->boolean (bit-and bits mask)))


;;
;; Domain Boundary: Numbers

(define (number->byte-string n l)
  (case l
   ((8) (number->byte-string-8 n))
   ((4) (number->byte-string-4 n))
   ((2) (number->byte-string-2 n))
   ((1) (byte->byte-string n))
   (else
    (error "length out of range" l))))

(define (number+sign->byte-string n l)
  (case l
   ((8) (number+sign->byte-string-8 n))
   ((4) (number+sign->byte-string-4 n))
   ((2) (number+sign->byte-string-2 n))
   ((1) (byte+sign->byte-string n))
   (else
    (error "length out of range" l))))

;;
;; Domain Boundary: Lists

(define (byte-string->list byte-string)
 (let ((answer-box (cons #f '())))
  (let loop ((n 0) (last answer-box))
   (if (= n (byte-string-length byte-string))
     (cdr answer-box)
     (begin
       (set-cdr! last (cons (byte-string->number
                             (byte-string-select byte-string n 1))
                            '()))
       (loop (+ n 1) (cdr last)) )) )))

(define (list->byte-string list)
  (apply byte-string-adjoin (map byte->byte-string list)))

;;
;; Run Time System Specifics

(define (big-endian-system?) (once
  (= 1 (byte-string->number #"0 0 0 1")) ))

(define system-is-big-endian (big-endian-system?))

;;
;; TODO network byte order conversion should be part of VSI-C
(if system-is-big-endian
  (begin (define (network-byte-order! byte-string)
           byte-string)
         (define (network-byte-order byte-string)
           (byte-string-copy byte-string)))

  (begin (define network-byte-order! byte-reverse!)
         (define network-byte-order  byte-reverse)))

(if system-is-big-endian
  (begin (define (from-network-byte-order! byte-string)
           byte-string)
         (define (from-network-byte-order byte-string)
           (byte-string-copy byte-string)))
  (begin (define from-network-byte-order! byte-reverse!)
         (define from-network-byte-order  byte-reverse)))


;;
;; Utilities & Compatibility
;;

;;; Byte Bit Printer
;;
;; XXX characters.scm has another definition of /byte?/
;; These should be unified and all relevant accessors
;; should be updated.
(define (byte-bit-printer:byte? object)
 (or ; yet to come (and (character? byte)
     ;                  (is-unicode-c0?
     ; depricated (and (char? byte)
     (and (number? byte) (exact? byte) (integer? byte)
          (>= byte 0) (<= byte 255))
     (and (byte-string? object)
          (= 1 (byte-string-length object)) )))

(define (byte->byte byte)
 (cond ((and (byte-string? byte) (= 1 (byte-string-length byte)))
          (byte-string-copy byte))
       ((and (number? byte) (>= byte 0) (<= byte 255))
          (byte->byte-string byte))
       (else #f)))

;; byte-bit-printer
;; Prints one byte as two groups of bits in regular binary
;; notation with leading digits having a higher value than
;; lower digits.
(define (one-byte-bit-printer byte)
 (let ((bits (byte->byte byte))
       (test-mask  #"128") ; XXX TODO read-only byte string literal constants
       (break-mask #"16"))
  (let loop ((test-mask (byte-string-copy test-mask))) ;; XXX constant and !
   (display (if (bit-and? bits test-mask) "1" "0"))
   (when (bit-and? break-mask test-mask) (display " "))
   (bit-shift-right! test-mask 1)
   (when (byte-string-true? test-mask)
     (loop test-mask)))
  *unspecified*))
;;; End of Byte Bit Printer

(define (byte-bit-printer:multi-byte-preambel bytes)
 ;(display "Multi Byte Bit Printer: ")
 (display "MSB")      ; Most Significant Byte first
 (display "(#bytes ") (display (length bytes)) (display ")")
 (display " "))

(define (byte-bit-printer object)
 (cond
  ((byte? object) (one-byte-bit-printer object))
  ((byte-string? object)
    (let ((bytes (byte-string->list object)))
     (byte-bit-printer:multi-byte-preambel bytes)
     (for-each (lambda (byte)
                  (one-byte-bit-printer byte)
                  (display "  "))
               (if system-is-big-endian
                 bytes
                 (reverse! bytes)))
     ))
  (else
   (error "byte-bit-printer: wrong type argument:" object)) ))


(define (brass->c-string string)
  (list->byte-string (append (map char->integer (brass->list string)) '(0))))
(define (brass->c-string/raw string)
  (list->byte-string (append (map char->integer (brass->list string)) )))

(define (c-string->brass/raw byte-string)
  (list->brass (map integer->char (byte-string->list byte-string))))

(define (c-string:null-terminated list-of-usv)
 (cond
  ((or (null? list-of-usv)
       (= (first list-of-usv) 0))
       '())
  (else
   (cons (car list-of-usv)
         (c-string:null-terminated (cdr list-of-usv))) )))
(define (c-string->brass byte-string)
  (list->brass
   (map integer->char
        (c-string:null-terminated
         (byte-string->list byte-string)))))
