(define (char->byte char)
  (char->integer char))

;; ;; SEE TEXT-LINE.CLASS FOR DEVELOPMENT
;; ;; Unicode C0 | Traditional Character Code and Name Mapping
;; ;; TODO Compare with C0 code chart | base analyzer reader on this table
;; (define c0-character-name-vector
;; #( ("nul" "null")       ; 0
;;    ("soh")
;;    ("stx")
;;    ("etx")

;;    ("eot")
;;    ("enq")
;;    ("ack")
;;    ("bel" "bell")

;;    ("bs" "backspace")   ; 8
;;    ("ht" "tab")
;;    ("newline" "lf" "nl")
;;    ("vt")

;;    ("np" "page")
;;    ("cr" "return")
;;    ("so")
;;    ("si")

;;    ("dle")              ; 16
;;    ("dc1")
;;    ("dc2")
;;    ("dc3")

;;    ("dc4")
;;    ("nak")
;;    ("syn")
;;    ("etb")

;;    ("can")              ; 24
;;    ("em")
;;    ("sb")
;;    ("esc")

;;    ("fs")
;;    ("gs")
;;    ("rs")
;;    ("us")

;;    ("space" "sp")      ; 32
;;    ("!")
;;    ("\"")
;;    ("#")
;;    ("$")
;;    ("%")
;;    ("&")
;;    ("'")

;;    ("(")               ; 40
;;    (")")
;;    ("*")
;;    ("+")
;;    (",")
;;    ("-")
;;    (".")
;;    ("/")

;;    ("0")               ; 48
;;    ("1")
;;    ("2")
;;    ("3")
;;    ("4")
;;    ("5")
;;    ("6")
;;    ("7")

;;    ("8")               ; 56
;;    ("9")
;;    (":")
;;    (";")
;;    ("<")
;;    ("=")
;;    (">")
;;    ("?")

;;    ("@")               ; 64
;;    ("A")
;;    ("B")
;;    ("C")

;;    ("D")
;;    ("E")
;;    ("F")
;;    ("G")

;;    ("H")               ; 72
;;    ("I")
;;    ("J")
;;    ("K")

;;    ("L")
;;    ("M")
;;    ("N")
;;    ("O")

;;    ("P")               ; 80
;;    ("Q")
;;    ("R")
;;    ("S")

;;    ("T")
;;    ("U")
;;    ("V")
;;    ("W")

;;    ("X")               ; 88
;;    ("Y")
;;    ("Z")
;;    ("[")

;;    ("\")
;;    ("]")
;;    ("^")
;;    ("_")
;;    ("`")
;;    ("a")
;;    ("b")
;;    ("c")
;;    ("d")
;;    ("e")
;;    ("f")
;;    ("g")
;;    ("h")
;;    ("i")
;;    ("j")
;;    ("k")
;;    ("l")
;;    ("m")
;;    ("n")
;;    ("o")
;;    ("p")
;;    ("q")
;;    ("r")
;;    ("s")
;;    ("t")
;;    ("u")
;;    ("v")
;;    ("w")
;;    ("x")
;;    ("y")
;;    ("z")
;;    ("{")
;;    ("|")
;;    ("}")
;;    ("~")
;;    ("del")
;; ))
;; (vector-set! c0-character-name-vector 92 (list (list->string (list #\\))))


;;
;; Unicode C0 Classes
(define (byte? number)
  "Return true if the parameter number is a valid byte value."
  (and (number?  number)
       (exact?   number)
       (integer? number)
       (and (>= number 0) (<= number 255))) ) ; beware of the evil minus
(define (is-unicode-c0? byte)
  "Return true if the byte value is part of Unicode C0."
  (and (byte? byte) (>= byte 0) (<= byte 127)))
(define (is-unicode-c0-printable? byte)
  "Return true if the byte value is a printable character in Unicode C0."
  (and (byte? byte) (>= byte 33) (<= byte 126)))
(define (is-unicode-c0-letter? byte)
  (and (byte? byte)
    (or (and (>= byte #x0041) (<= byte #x005A))
        (and (>= byte #x0061) (<= byte #x007A)) )))

(define (letter? character)
  (is-unicode-c0-letter? (char->byte character)))
(define (printable? character) ; convenience abbreviation
  (and (char? character)
       (is-unicode-c0-printable? (char->byte character))))
(define (control-code? character)
  (not (printable? character)))
(define (white-space? character)
  (and (char? character) (control-code? character)
    (case (char->byte character)
      ((#x0020 #x0009) #t)
      (else #f)) ))

(define (space? character)
  (eq? character #\space))

(define (digit? character)
  (if (char? character)
    (let ((byte (char->byte character)))
      (and (>= byte #x0030)
           (<= byte #x0039)) )
    #f))

;;
;; Compatibility
(define char-numeric? digit?)

;;
;; Character Debugging
(define (represent-char-safely char)
  (cond ((printable? char) char)
        ((char? char)      (string-append "#x" (number->string (char->integer char) 16)))
        (else              (brass-append "represent-char-safely: not a character"))))

;; Import from format compatibility
;; Compatibility: Unicode C0 character case conversion & class
(define (char-locase/c0? char) (<= #x0061 (char->integer char) #x007A))
(define (char-upcase/c0? char) (<= #x0041 (char->integer char) #x005A))

(define (char-upcase/c0 char)
 (let ((n (char->byte char)))
  (cond ((<= #x0061 n #x007A) (integer->char (- n #x20)))
        (else                 char)) ))
(define (char-downcase/c0 char)
 (let ((n (char->byte char)))
  (cond ((<= #x0041 n #x005A) (integer->char (+ n #x20)))
        (else                 char)) ))
