;;;; "format.scm" traditional derived text output formatter
;; TODO test & document
;; TODO repair ~S it uses brass-length on a byte string
;;
;; Please note:
;; Because of the missing CLR4 autocut for inner defines changes
;; to the inner procedures need to be reflected in the cutting
;; routines: one at the end of /format:format-work/ and one at
;; the end of /format/.

;; The current prototype of namespaces collides with
;; m_expand_body when it comes to inner defines? Fixed.
;(define format:namespace (namespace

;; Compatibility: Unicode C0 character case conversion & class
(define char-upcase   char-upcase/c0)
(define char-downcase char-downcase/c0)
(define char-whitespace? white-space?)

;;; Configuration ------------------------------------------------------------

(define format:symbol-case-conv #f)
;; Symbols are converted by symbol->string so the case of the printed
;; symbols is implementation dependent. format:symbol-case-conv is a
;; one arg closure which is either #f (no conversion), string-upcase!,
;; string-downcase! or string-capitalize!.

(define format:iobj-case-conv #f)
;; As format:symbol-case-conv but applies for the representation of
;; implementation internal objects.

(define format:expch #\E)
;; The character prefixing the exponent value in ~e printing.

(define format:floats #t)
(define format:complex-numbers #t)

(define format:radix-pref (char=? #\# (string-ref (number->string 8 8) 0)))
;; Detects if number->string adds a radix prefix.

(define format:ascii-non-printable-charnames ;; XXX see encoding for the "new" charnames ;-)
  '#("nul" "soh" "stx" "etx" "eot" "enq" "ack" "bel"
     "bs"  "ht"  "nl"  "vt"  "np"  "cr"  "so"  "si"
     "dle" "dc1" "dc2" "dc3" "dc4" "nak" "syn" "etb"
     "can" "em"  "sub" "esc" "fs"  "gs"  "rs"  "us" "space"))

;;; End of configuration ----------------------------------------------------

;; Roman Number Format
(define format:roman-alist
        '((1000 #\M) (500 #\D) (100 #\C) (50 #\L)
          (10 #\X) (5 #\V) (1 #\I)))

(define format:roman-boundary-values
        '(100 100 10 10 1 1 #f))

(define format:num->roman (lambda (n)
  (if (and (integer? n) (> n 0))
      (let loop ((n n)
                 (romans format:roman-alist)
                 (boundaries format:roman-boundary-values)
                 (s '()))
        (if (null? romans)
            (list->string (reverse s))
            (let ((roman-val (caar romans))
                  (roman-dgt (cadar romans))
                  (bdry (car boundaries)))
              (let loop2 ((q (quotient n roman-val))
                          (r (remainder n roman-val))
                          (s s))
                (if (= q 0)
                    (if (and bdry (>= r (- roman-val bdry)))
                        (loop (remainder r bdry) (cdr romans)
                              (cdr boundaries)
                              (cons roman-dgt
                                    (append
                                     (cdr (assv bdry romans))
                                     s)))
                        (loop r (cdr romans) (cdr boundaries) s))
                    (loop2 (- q 1) r (cons roman-dgt s)))))))
      (error "only positive integers can be romanized"))))

(define format:version "IV")
(define (format . args) ; open until the end | this is one stack machine

(define format:port #f)                 ; curr. format output port
(define format:output-col 0)            ; curr. format output tty column
(define format:flush-output #f)         ; flush output at end of formatting
(define format:case-conversion #f)
(define format:args #f)
(define format:pos 0)                   ; curr. format string parsing position
(define format:arg-pos 0)               ; curr. format argument position
                                        ; this is global for error presentation
       
;; format string and char output routines on format:port

(define format:out-str
        (lambda (str)
          (if format:case-conversion
              (display (format:case-conversion str) format:port)
              (display str format:port))
          (set! format:output-col
                (+ format:output-col (string-length str)))))

(define format:out-char
        (lambda (ch)
          (if format:case-conversion
              (display (format:case-conversion (string ch))
                       format:port)
              (write-char ch format:port))
          (set! format:output-col
                (if (char=? ch #\newline)
                    0
                    (+ format:output-col 1)))))
       
       ;;(define (format:out-substr str i n)
       ;;  (display (substring str i n) format:port)
       ;;  (set! format:output-col (+ format:output-col n)))

(define format:out-substr
        (lambda (str i n)
          (let loop ((k i))
           (when (not (= k n))
            (write-char (string-ref str k) format:port)
            (loop (+ k 1))))
          (set! format:output-col (+ format:output-col (- n i)))))

       ;;(define (format:out-fill n ch)
       ;;  (format:out-str (make-string n ch)))

(define format:out-fill
        (lambda (n ch)
          (let loop ((i 0))
           (when (not (= i n))
            (write-char ch format:port)
            (loop (+ i 1))))
          (set! format:output-col (+ format:output-col n))))

       ;; format's user error handler

(define format:error
        (lambda args                ; never returns!
          (let ((format-args format:args)
                (port (current-error-port)))
            (set! format:error format:intern-error)
            (if (and (>= (length format:args) 2)
                     (string? (cadr format:args)))
                (let ((format-string (cadr format-args)))
                  (if (not (zero? format:arg-pos))
                      (set! format:arg-pos (- format:arg-pos 1)))
                  (format port
                          (once(string-append
                            "~%FORMAT error: ( format ~a"
                            "~%  string split: \"~a\" <- !!! \"~a\""
                            "~%argument split: ~a !!! -> ~a )~%        "))
                          (car format:args)
                          (substring format-string 0 format:pos)
                          (substring format-string format:pos
                                     (string-length format-string))
                          (list-head format:arg-pos (cddr format:args))
                          (list-tail format:arg-pos (cddr format:args))))
                (format port 
                        "~%FORMAT: error with call: (format ~a)~%        "
                        format:args))

            (apply format port args)
            (newline port)
            (set! format:error format:error-save)
            (format:abort))))

(define format:intern-error
        (lambda args
          ;;if something goes wrong in format:error
          (display "FORMAT: INTERNAL ERROR IN FORMAT:ERROR!") (newline)
          (display "        format args: ") (write format:args) (newline)
          (display "        error args:  ") (write args) (newline)
          (set! format:error format:error-save)
          (format:abort)))
              
(define format:error-save #f)

(define format:format
        (lambda args                ; the formatter entry
          (set! format:args args)
          (set! format:arg-pos 0)
          (set! format:pos 0)
          (if (< (length args) 1)
              (format:error "not enough arguments"))

          ;; If the first argument is a string, then that's the format string.
          ;; (Scheme->C)
          ;; In this case, put the argument list in canonical form.
          (let ((args (if (string? (car args))
                          (cons #f args)
                          args)))
            ;; Use this canonicalized version when reporting errors.
            (set! format:args args)

            (let ((destination (car args))
                  (arglist (cdr args)))
              (cond
               ((or (and (boolean? destination)        ; port output
                         destination)
                    (output-port? destination)
                    (number? destination))
                (format:out (cond
                             ((boolean? destination) (current-output-port))
                             ((output-port? destination) destination)
                             ((number? destination) (current-error-port)))
                            (car arglist) (cdr arglist)))
               ((and (boolean? destination)        ; string output
                     (not destination))
                (byte-string->brass (call-with-output-string
                 (lambda (port) (format:out port (car arglist) (cdr arglist))))))
               (else
                (format:error "illegal destination `~a'" destination)))))))

(define format:out                              ; the output handler for a port
        (lambda (port fmt args)        
          (set! format:port port)                ; global port for
                                                 ; output routines
          (set! format:case-conversion #f)       ; modifier case
                                                 ; conversion procedure
          (set! format:flush-output #f)                ; ~! reset
          (let ((col (port-column port)))        ; get current column from port
            (and col (set! format:output-col col)))
          (let ((arg-pos (format:format-work fmt args))
                (arg-len (length args)))
            (cond
             ((> arg-pos arg-len)
              (set! format:arg-pos (+ arg-len 1))
              (display format:arg-pos)
              (format:error "~a missing argument~:p" (- arg-pos arg-len)))
             (else
              (if format:flush-output (force-output port))
              #t)))))

(define format:parameter-characters
        '(#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\- #\+ #\v #\# #\'))

(define format:format-work ; does the formatting work
        (lambda (format-string arglist)
          (letrec
              ((format-string-len (string-length format-string))
               (arg-pos 0)                        ; argument position in arglist
               (arg-len (length arglist))        ; number of arguments
               (modifier #f)                        ; 'colon | 'at | 'colon-at | #f
               (params '())                        ; directive parameter list
               (param-value-found #f)                ; a directive
                                        ; parameter value
                                        ; found
               (conditional-nest 0)                ; conditional nesting level
               (clause-pos 0)                        ; last cond. clause
                                        ; beginning char pos
               (clause-default #f)                ; conditional default
                                        ; clause string
               (clauses '())                        ; conditional clause
                                        ; string list
               (conditional-type #f)                ; reflects the
                                        ; contional modifiers
               (conditional-arg #f)                ; argument to apply the conditional
               (iteration-nest 0)                ; iteration nesting level
               (iteration-pos 0)                ; iteration string
                                        ; beginning char pos
               (iteration-type #f)                ; reflects the
                                        ; iteration modifiers
               (max-iterations #f)                ; maximum number of
                                        ; iterations
               (recursive-pos-save format:pos)
               
               (next-char                        ; gets the next char
                                        ; from format-string
                (lambda ()
                  (let ((ch (peek-next-char)))
                    (set! format:pos (+ 1 format:pos))
                    ch)))
               
               (peek-next-char
                (lambda ()
                  (if (>= format:pos format-string-len)
                      (format:error "illegal format string")
                      (string-ref format-string format:pos))))
               
               (one-positive-integer?
                (lambda (params)
                  (cond
                   ((null? params) #f)
                   ((and (integer? (car params))
                         (>= (car params) 0)
                         (= (length params) 1)) #t)
                   (else
                    (format:error
                     "one positive integer parameter expected")))))
               
               (next-arg
                (lambda ()
                  (if (>= arg-pos arg-len)
                      (begin
                        (set! format:arg-pos (+ arg-len 1))
                        (format:error "missing argument(s)")))
                  (add-arg-pos 1)
                  (list-ref arglist (- arg-pos 1))))
               
               (prev-arg
                (lambda ()
                  (add-arg-pos -1)
                  (if (negative? arg-pos)
                      (format:error "missing backward argument(s)"))
                  (list-ref arglist arg-pos)))
               
               (rest-args
                (lambda ()
                  (let loop ((l arglist) (k arg-pos)) ; list-tail definition
                    (if (= k 0) l (loop (cdr l) (- k 1))))))
               
               (add-arg-pos
                (lambda (n) 
                  (set! arg-pos (+ n arg-pos))
                  (set! format:arg-pos arg-pos)))
               
               (anychar-dispatch                ; dispatches the format-string
                (lambda ()
                  (if (>= format:pos format-string-len)
                      arg-pos                        ; used for ~? continuance
                      (let ((char (next-char)))
                        (cond
                         ((char=? char #\~)
                          (set! modifier #f)
                          (set! params '())
                          (set! param-value-found #f)
                          (tilde-dispatch))
                         (else
                          (if (and (zero? conditional-nest)
                                   (zero? iteration-nest))
                              (format:out-char char))
                          (anychar-dispatch)))))))
               
               (tilde-dispatch
                (lambda ()
                  (cond
                   ((>= format:pos format-string-len)
                    (format:out-str "~")        ; tilde at end of
                                        ; string is just
                                        ; output
                    arg-pos)                        ; used for ~?
                                        ; continuance
                   ((and (or (zero? conditional-nest)
                             (memv (peek-next-char) ; find conditional
                                        ; directives
                                   (append '(#\[ #\] #\; #\: #\@ #\^)
                                           format:parameter-characters)))
                         (or (zero? iteration-nest)
                             (memv (peek-next-char) ; find iteration
                                        ; directives
                                   (append '(#\{ #\} #\: #\@ #\^)
                                           format:parameter-characters))))
                    (case (char-upcase (next-char))
                      
                      ;; format directives
                      
                      ((#\A)                        ; Any -- for humans
                       (set! format:read-proof
                             (memq modifier '(colon colon-at)))
                       (format:out-obj-padded (memq modifier '(at colon-at))
                                              (next-arg) #f params)
                       (anychar-dispatch))
                      ((#\S)                        ; Slashified -- for parsers
                       (set! format:read-proof
                             (memq modifier '(colon colon-at)))
                       (format:out-obj-padded (memq modifier '(at colon-at))
                                              (next-arg) #t params)
                       (anychar-dispatch))
                      ((#\D)                        ; Decimal
                       (format:out-num-padded modifier (next-arg) params 10)
                       (anychar-dispatch))
                      ((#\X)                        ; Hexadecimal
                       (format:out-num-padded modifier (next-arg) params 16)
                       (anychar-dispatch))
                      ((#\O)                        ; Octal
                       (format:out-num-padded modifier (next-arg) params 8)
                       (anychar-dispatch))
                      ((#\B)                        ; Binary
                       (format:out-num-padded modifier (next-arg) params 2)
                       (anychar-dispatch))
                      ((#\R)
                       (if (null? params)
                           (format:out-obj-padded ; Roman, cardinal, ordinal numerals
                            #f
                            ((case modifier
                               ((at)    format:num->roman)
                               ((colon) format:num->ordinal)
                               (else    format:num->cardinal))
                             (next-arg))
                            #f params)
                           (format:out-num-padded ; any Radix
                            modifier (next-arg) (cdr params) (car params)))
                       (anychar-dispatch))
                      ((#\F)                        ; Fixed-format floating-point
                       (if format:floats
                           (format:out-fixed modifier (next-arg) params)
                           (format:out-str (number->string (next-arg))))
                       (anychar-dispatch))
                      ((#\E)                        ; Exponential floating-point
                       (if format:floats
                           (format:out-expon modifier (next-arg) params)
                           (format:out-str (number->string (next-arg))))
                       (anychar-dispatch))
                      ((#\G)                        ; General floating-point
                       (if format:floats
                           (format:out-general modifier (next-arg) params)
                           (format:out-str (number->string (next-arg))))
                       (anychar-dispatch))
                      ((#\$)                        ; Dollars floating-point
                       (if format:floats
                           (format:out-dollar modifier (next-arg) params)
                           (format:out-str (number->string (next-arg))))
                       (anychar-dispatch))
                      ((#\I)                        ; Complex numbers
                       (if (not format:complex-numbers)
                           (format:error
                            "complex numbers not supported by this scheme system"))
                       (let ((z (next-arg)))
                         (if (not (complex? z))
                             (format:error "argument not a complex number"))
                         (format:out-fixed modifier (real-part z) params)
                         (format:out-fixed 'at (imag-part z) params)
                         (format:out-char #\i))
                       (anychar-dispatch))
                      ((#\C)                        ; Character
                       (let ((ch (if (one-positive-integer? params)
                                     (integer->char (car params))
                                     (next-arg))))
                         (if (not (char? ch))
                             (format:error "~~c expects a character"))
                         (case modifier
                           ((at)
                            (format:out-str (format:char->str ch)))
                           ((colon)
                            (let ((c (char->integer ch)))
                              (if (< c 0)
                                  (set! c (+ c 256))) ; compensate
                                        ; complement
                                        ; impl.
                              (cond
                               ((< c #x20)        ; assumes that control
                                        ; chars are < #x20
                                (format:out-char #\^)
                                (format:out-char
                                 (integer->char (+ c #x40))))
                               ((>= c #x7f)
                                (format:out-str "#\\")
                                (format:out-str
                                 (if format:radix-pref
                                     (let ((s (number->string c 8)))
                                       (substring s 2 (string-length s)))
                                     (number->string c 8))))
                               (else
                                (format:out-char ch)))))
                           (else (format:out-char ch))))
                       (anychar-dispatch))
                      ((#\P)                        ; Plural
                       (if (memq modifier '(colon colon-at))
                           (prev-arg))
                       (let ((arg (next-arg)))
                         (if (not (number? arg))
                             (format:error "~~p expects a number argument"))
                         (if (= arg 1)
                             (if (memq modifier '(at colon-at))
                                 (format:out-char #\y))
                             (if (memq modifier '(at colon-at))
                                 (format:out-str "ies")
                                 (format:out-char #\s))))
                       (anychar-dispatch))
                      ((#\~)                        ; Tilde
                       (if (one-positive-integer? params)
                           (format:out-fill (car params) #\~)
                           (format:out-char #\~))
                       (anychar-dispatch))
                      ((#\%)                        ; Newline
                       (if (one-positive-integer? params)
                           (format:out-fill (car params) #\newline)
                           (format:out-char #\newline))
                       (set! format:output-col 0)
                       (anychar-dispatch))
                      ((#\&)                        ; Fresh line
                       (if (one-positive-integer? params)
                           (begin
                             (if (> (car params) 0)
                                 (format:out-fill (- (car params)
                                                     (if (>
                                                          format:output-col
                                                          0) 0 1))
                                                  #\newline))
                             (set! format:output-col 0))
                           (if (> format:output-col 0)
                               (format:out-char #\newline)))
                       (anychar-dispatch))
                      ((#\_)                        ; Space character
                       (if (one-positive-integer? params)
                           (format:out-fill (car params) #\space)
                           (format:out-char #\space))
                       (anychar-dispatch))
                      ((#\/)                        ; Tabulator character
                       (if (one-positive-integer? params)
                           (format:out-fill (car params) #\tab)
                           (format:out-char #\tab))
                       (anychar-dispatch))
                      ((#\|)                        ; Page seperator
                       (if (one-positive-integer? params)
                           (format:out-fill (car params) #\page)
                           (format:out-char #\page))
                       (set! format:output-col 0)
                       (anychar-dispatch))
                      ((#\T)                        ; Tabulate
                       (format:tabulate modifier params)
                       (anychar-dispatch))
                      ((#\newline)                ; Continuation lines
                       (if (eq? modifier 'at)
                           (format:out-char #\newline))
                       (if (< format:pos format-string-len)
                           (let loop ((ch (peek-next-char) ))
                            (when (not (or (not (char-whitespace? ch))
                                           (= format:pos (- format-string-len 1))))
                             (if (eq? modifier 'colon)
                                 (format:out-char (next-char))
                                 (next-char))
                             (loop (peek-next-char)) )))
                       (anychar-dispatch))
                      
                      ;; format directive modifiers and parameters

                      ((#\@)                        ; `@' modifier
                       (if (memq modifier '(at colon-at))
                           (format:error "double `@' modifier"))
                       (set! modifier (if (eq? modifier 'colon) 'colon-at 'at))
                       (tilde-dispatch))
                      ((#\:)                        ; `:' modifier
                       (if (memq modifier '(colon colon-at))
                           (format:error "double `:' modifier"))
                       (set! modifier (if (eq? modifier 'at) 'colon-at 'colon))
                       (tilde-dispatch))
                      ((#\')                        ; Character parameter ???
                       (if modifier (format:error "misplaced modifier"))
                       (set! params (append params (list (char->integer (next-char)))))
                       (set! param-value-found #t)
                       (tilde-dispatch))
                      ((#\0 #\1 #\2 #\3 #\4 #\5 #\6 #\7 #\8 #\9 #\- #\+) ; num. paramtr
                       (if modifier (format:error "misplaced modifier"))
                       (let ((num-str-beg (- format:pos 1))
                             (num-str-end format:pos))
                         (let loop ((ch (peek-next-char)))
                           (when (char-numeric? ch)
                             (next-char)
                             (set! num-str-end (+ 1 num-str-end))
                             (loop (peek-next-char)) ))
                         (set! params
                               (append params
                                       (list (string->number
                                              (substring format-string
                                                         num-str-beg
                                                         num-str-end))))))
                       (set! param-value-found #t)
                       (tilde-dispatch))
                      ((#\,)                        ; Parameter separators
                       (if modifier (format:error "misplaced modifier"))
                       (if (not param-value-found)
                           (set! params (append params '(#f)))) ; append empty paramtr
                       (set! param-value-found #f)
                       (tilde-dispatch))
                      ((#\Q)                        ; Inquiry messages
                       (if (eq? modifier 'colon)
                           (format:out-str format:version)
                           (let ((nl (string #\newline)))
                             (format:out-str
                              (string-append
                               "Derived SLIB Common LISP format version " format:version nl
                               "  it's just your own copy of it, I guess" nl
                               ))))
                       (anychar-dispatch))
                      (else                        ; Unknown tilde directive
                       (format:error "unknown control character `~c'"
                                     (string-ref format-string (- format:pos 1))))))
                   (else (anychar-dispatch)))))) ; in case of conditional

            (set! format:pos 0)
            (set! format:arg-pos 0)
            (anychar-dispatch)                        ; start the formatting
            (set! format:pos recursive-pos-save)
            (begin ; XXX missing CLR4 cuts
              (set!  next-char             #f)
              (set!  peek-next-char        #f)
              (set!  one-positive-integer? #f)
              (set!  next-arg              #f)
              (set!  prev-arg              #f)
              (set!  rest-args             #f)
              (set!  add-arg-pos           #f)
              (set!  anychar-dispatch      #f)
              (set!  tilde-dispatch        #f) )
            arg-pos)))                                ; return the position in the arg. list

       ;; when format:read-proof is true, format:obj->str will wrap
       ;; result strings starting with "#<" in an extra pair of double
       ;; quotes.
       
(define format:read-proof #f)

       ;; format:obj->str returns a R4RS representation as a string of
       ;; an arbitrary scheme object.

(define format:obj->str
        (lambda (obj slashify)
          (let ((res (if slashify
                         (object->string obj)
                         (with-output-to-string (lambda () (display obj))))))
            (if (and format:read-proof (string-prefix? "#<" res))
                (object->string res)
                res))))

       ;; format:char->str converts a character into a slashified string as
       ;; done by `write'. The procedure is dependent on the integer
       ;; representation of characters and assumes a character number according to
       ;; the ASCII character set.

(define format:char->str
        (lambda (ch)
          (let ((int-rep (char->integer ch)))
            (if (< int-rep 0)                        ; if chars are [-128...+127]
                (set! int-rep (+ int-rep 256)))
            (string-append
             "#\\"
             (cond
              ((char=? ch #\newline) "newline")
              ((and (>= int-rep 0) (<= int-rep 32))
               (vector-ref format:ascii-non-printable-charnames int-rep))
              ((= int-rep 127) "del")
              ((>= int-rep 128)                ; octal representation
               (if format:radix-pref
                   (let ((s (number->string int-rep 8)))
                     (substring s 2 (string-length s)))
                   (number->string int-rep 8)))
              (else (string ch)))))))

(define format:space-ch (char->integer #\space))
(define format:zero-ch (char->integer #\0))

(define format:par
        (lambda (pars length index default name)
          (if (> length index)
              (let ((par (list-ref pars index)))
                (if par
                    (if name
                        (if (< par 0)
                            (format:error 
                             "~s parameter must be a positive integer" name)
                            par)
                        par)
                    default))
              default)))

(define format:out-obj-padded
        (lambda (pad-left obj slashify pars)
               (if (null? pars)
                   (format:out-str (format:obj->str obj slashify))
                   (let ((l (length pars)))
                     (let ((mincol (format:par pars l 0 0 "mincol"))
                           (colinc (format:par pars l 1 1 "colinc"))
                           (minpad (format:par pars l 2 0 "minpad"))
                           (padchar (integer->char
                                     (format:par pars l 3 format:space-ch #f)))
                           (objstr (format:obj->str obj slashify)))
                       (if (not pad-left)
                           (format:out-str objstr))
                       (let ((objstr-len (string-length objstr)))
                         (let loop ((i minpad))
                           (if (not (>= (+ objstr-len i) mincol))
                             (format:out-fill i padchar)
                             (loop objstr-len (+ i colinc))) ))
                       (if pad-left
                           (format:out-str objstr)))))))

(define format:out-num-padded
        (lambda (modifier number pars radix)
          (if (not (integer? number)) (format:error "argument not an integer"))
          (let ((numstr (number->string number radix)))
            (if (and format:radix-pref (not (= radix 10)))
                (set! numstr (substring numstr 2 (string-length numstr))))
            (if (and (null? pars) (not modifier))
                (format:out-str numstr)
                (let ((l (length pars))
                      (numstr-len (string-length numstr)))
                  (let ((mincol (format:par pars l 0 #f "mincol"))
                        (padchar (integer->char
                                  (format:par pars l 1 format:space-ch #f)))
                        (commachar (integer->char
                                    (format:par pars l 2 (char->integer #\,) #f)))
                        (commawidth (format:par pars l 3 3 "commawidth")))
                    (if mincol
                        (let ((numlen numstr-len)) ; calc. the output len of number
                          (if (and (memq modifier '(at colon-at)) (>= number 0))
                              (set! numlen (+ numlen 1)))
                          (if (memq modifier '(colon colon-at))
                              (set! numlen (+ (quotient (- numstr-len 
                                                           (if (< number 0) 2 1))
                                                        commawidth)
                                              numlen)))
                          (if (> mincol numlen)
                              (format:out-fill (- mincol numlen) padchar))))
                    (if (and (memq modifier '(at colon-at))
                             (>= number 0))
                        (format:out-char #\+))
                    (if (memq modifier '(colon colon-at)) ; insert comma character
                        (let ((start (remainder numstr-len commawidth))
                              (ns (if (< number 0) 1 0)))
                          (format:out-substr numstr 0 start)
                          (let loop ((i start))
                            (if (not (>= i numstr-len))
                                (format:out-char commachar))
                            (format:out-substr numstr i (+ i commawidth))
                            (loop (+ i commawidth))))
                        (format:out-str numstr))))))))

(define format:tabulate
        (lambda (modifier pars)
          (let ((l (length pars)))
            (let ((colnum (format:par pars l 0 1 "colnum"))
                  (colinc (format:par pars l 1 1 "colinc"))
                  (padch (integer->char (format:par pars l 2 format:space-ch #f))))
              (case modifier
                ((colon colon-at)
                 (format:error "unsupported modifier for ~~t"))
                ((at)                                ; relative tabulation
                 (format:out-fill
                  (if (= colinc 0)
                      colnum                        ; colnum = colrel
                      (let loop ((c 0))
                        (if (>= c col)
                            (- c format:output-col)
                            (loop (+ c colinc))) ))
                  padch))
                (else                                ; absolute tabulation
                 (format:out-fill
                  (cond
                   ((< format:output-col colnum)
                    (- colnum format:output-col))
                   ((= colinc 0)
                    0)
                   (else
                    (let loop ((c 0))
                      (if (>= c col)
                          (- c format:output-col)
                          (loop (+ c colinc))) )))
                  padch)))))))



(define format:cardinal-ones-list
        '(#f "one" "two" "three" "four" "five"
             "six" "seven" "eight" "nine" "ten" "eleven" "twelve" "thirteen"
             "fourteen" "fifteen" "sixteen" "seventeen" "eighteen"
             "nineteen"))

(define format:cardinal-tens-list
        '(#f #f "twenty" "thirty" "forty" "fifty" "sixty" "seventy" "eighty"
             "ninety"))

(define format:num->cardinal999
        (lambda (n)
                                        ;this procedure is inspired by the Bruno Haible's CLisp
                                        ;function format-small-cardinal, which converts numbers
                                        ;in the range 1 to 999, and is used for converting each
                                        ;thousand-block in a larger number
          (let* ((hundreds (quotient n 100))
                 (tens+ones (remainder n 100))
                 (tens (quotient tens+ones 10))
                 (ones (remainder tens+ones 10)))
            (append
             (if (> hundreds 0)
                 (append
                  (string->list
                   (list-ref format:cardinal-ones-list hundreds))
                  (string->list" hundred")
                  (if (> tens+ones 0) '(#\space) '()))
                 '())
             (if (< tens+ones 20)
                 (if (> tens+ones 0)
                     (string->list
                      (list-ref format:cardinal-ones-list tens+ones))
                     '())
                 (append
                  (string->list
                   (list-ref format:cardinal-tens-list tens))
                  (if (> ones 0)
                      (cons #\-
                            (string->list
                             (list-ref format:cardinal-ones-list ones)))
                      '())))))))

(define format:cardinal-thousand-block-list
        '("" " thousand" " million" " billion" " trillion" " quadrillion"
          " quintillion" " sextillion" " septillion" " octillion" " nonillion"
          " decillion" " undecillion" " duodecillion" " tredecillion"
          " quattuordecillion" " quindecillion" " sexdecillion" " septendecillion"
          " octodecillion" " novemdecillion" " vigintillion"))

(define format:num->cardinal
        (lambda (n)
          (cond ((not (integer? n))
                 (format:error
                  "only integers can be converted to English cardinals"))
                ((= n 0) "zero")
                ((< n 0) (string-append "minus " (format:num->cardinal (- n))))
                (else
                 (let ((power3-word-limit
                        (length format:cardinal-thousand-block-list)))
                   (let loop ((n n)
                              (power3 0)
                              (s '()))
                     (if (= n 0)
                         (list->string s)
                         (let ((n-before-block (quotient n 1000))
                               (n-after-block (remainder n 1000)))
                           (loop n-before-block
                                 (+ power3 1)
                                 (if (> n-after-block 0)
                                     (append
                                      (if (> n-before-block 0)
                                          (string->list ", ") '())
                                      (format:num->cardinal999 n-after-block)
                                      (if (< power3 power3-word-limit)
                                          (string->list
                                           (list-ref
                                            format:cardinal-thousand-block-list
                                            power3))
                                          (append
                                           (string->list " times ten to the ")
                                           (string->list
                                            (format:num->ordinal
                                             (* power3 3)))
                                           (string->list " power")))
                                      s)
                                     s))))))))))

(define format:ordinal-ones-list
        '(#f "first" "second" "third" "fourth" "fifth"
             "sixth" "seventh" "eighth" "ninth" "tenth" "eleventh" "twelfth"
             "thirteenth" "fourteenth" "fifteenth" "sixteenth" "seventeenth"
             "eighteenth" "nineteenth"))

(define format:ordinal-tens-list
        '(#f #f "twentieth" "thirtieth" "fortieth" "fiftieth" "sixtieth"
             "seventieth" "eightieth" "ninetieth"))

(define format:num->ordinal
        (lambda (n)
          (cond ((not (integer? n))
                 (format:error
                  "only integers can be converted to English ordinals"))
                ((= n 0) "zeroth")
                ((< n 0) (string-append "minus " (format:num->ordinal (- n))))
                (else
                 (let ((hundreds (quotient n 100))
                       (tens+ones (remainder n 100)))
                   (string-append
                    (if (> hundreds 0)
                        (string-append
                         (format:num->cardinal (* hundreds 100))
                         (if (= tens+ones 0) "th" " "))
                        "")
                    (if (= tens+ones 0) ""
                        (if (< tens+ones 20)
                            (list-ref format:ordinal-ones-list tens+ones)
                            (let ((tens (quotient tens+ones 10))
                                  (ones (remainder tens+ones 10)))
                              (if (= ones 0)
                                  (list-ref format:ordinal-tens-list tens)
                                  (string-append
                                   (list-ref format:cardinal-tens-list tens)
                                   "-"
                                   (list-ref format:ordinal-ones-list ones))))
                            ))))))))

       ;; format inf and nan.

(define format:out-inf-nan
        (lambda (number width digits edigits overch padch)
          ;; inf and nan are always printed exactly as "+inf.0", "-inf.0" or
          ;; "+nan.0", suitably justified in their field.  We insist on
          ;; printing this exact form so that the numbers can be read back in.
          ;; (Those to insist were the same ones who "forgot" what really needed
          ;;  to be "read proof"?: /write/ which didn't pass the serialization
          ;; tests.)

          (let* ((str (number->string number))
                 (len (string-length str))
                 (dot (string-index str #\.))
                 (digits (+ (or digits 0)
                            (if edigits (+ edigits 2) 0))))
            (if (and width overch (< width len))
                (format:out-fill width (integer->char overch))
                (let* ((leftpad (if width
                                    (max (- width (max len (+ dot 1 digits))) 0)
                                    0))
                       (rightpad (if width
                                     (max (- width leftpad len) 0)
                                     0))
                       (padch (integer->char (or padch format:space-ch)))) 
                  (format:out-fill leftpad padch)
                  (format:out-str str)
                  (format:out-fill rightpad padch))))))

       ;; format fixed flonums (~F)

(define format:out-fixed
        (lambda (modifier number pars)
          (if (not (or (number? number) (string? number)))
              (format:error "argument is not a number or a number string"))

          (let ((l (length pars)))
            (let ((width (format:par pars l 0 #f "width"))
                  (digits (format:par pars l 1 #f "digits"))
                  (scale (format:par pars l 2 0 #f))
                  (overch (format:par pars l 3 #f #f))
                  (padch (format:par pars l 4 format:space-ch #f)))

              (cond
               ((or (inf? number) (nan? number))
                (format:out-inf-nan number width digits #f overch padch))

               (digits
                (format:parse-float 
                 (if (string? number) number (number->string number)) #t scale)
                (if (<= (- format:fn-len format:fn-dot) digits)
                    (format:fn-zfill #f (- digits (- format:fn-len format:fn-dot)))
                    (format:fn-round digits))
                (if width
                    (let ((numlen (+ format:fn-len 1)))
                      (if (or (not format:fn-pos?) (eq? modifier 'at))
                          (set! numlen (+ numlen 1)))
                      (if (and (= format:fn-dot 0) (> width (+ digits 1)))
                          (set! numlen (+ numlen 1)))
                      (if (< numlen width)
                          (format:out-fill (- width numlen) (integer->char padch)))
                      (if (and overch (> numlen width))
                          (format:out-fill width (integer->char overch))
                          (format:fn-out modifier (> width (+ digits 1)))))
                    (format:fn-out modifier #t)))

               (else
                (format:parse-float
                 (if (string? number) number (number->string number)) #t scale)
                (format:fn-strip)
                (if width
                    (let ((numlen (+ format:fn-len 1)))
                      (if (or (not format:fn-pos?) (eq? modifier 'at))
                          (set! numlen (+ numlen 1)))
                      (if (= format:fn-dot 0)
                          (set! numlen (+ numlen 1)))
                      (if (< numlen width)
                          (format:out-fill (- width numlen) (integer->char padch)))
                      (if (> numlen width)        ; adjust precision if possible
                          (let ((dot-index (- numlen
                                              (- format:fn-len format:fn-dot))))
                            (if (> dot-index width)
                                (if overch        ; numstr too big for required width
                                    (format:out-fill width (integer->char overch))
                                    (format:fn-out modifier #t))
                                (begin
                                  (format:fn-round (- width dot-index))
                                  (format:fn-out modifier #t))))
                          (format:fn-out modifier #t)))
                    (format:fn-out modifier #t))))))))

       ;; format exponential flonums (~E)

(define format:out-expon
        (lambda (modifier number pars)
          (if (not (or (number? number) (string? number)))
              (format:error "argument is not a number"))

          (let ((l (length pars)))
            (let ((width   (format:par pars l 0 #f "width"))
                  (digits  (format:par pars l 1 #f "digits"))
                  (edigits (format:par pars l 2 #f "exponent digits"))
                  (scale   (format:par pars l 3  1  #f))
                  (overch  (format:par pars l 4 #f  #f))
                  (padch   (format:par pars l 5 format:space-ch #f))
                  (expch   (format:par pars l 6 #f  #f)))
              
              (cond
               ((or (inf? number) (nan? number))
                (format:out-inf-nan number width digits edigits overch padch))

               (digits                                ; fixed precision

                (let ((digits (if (> scale 0)
                                  (if (< scale (+ digits 2))
                                      (+ (- digits scale) 1)
                                      0)
                                  digits)))
                  (format:parse-float 
                   (if (string? number) number (number->string number)) #f scale)
                  (if (<= (- format:fn-len format:fn-dot) digits)
                      (format:fn-zfill #f (- digits (- format:fn-len format:fn-dot)))
                      (format:fn-round digits))
                  (if width
                      (if (and edigits overch (> format:en-len edigits))
                          (format:out-fill width (integer->char overch))
                          (let ((numlen (+ format:fn-len 3))) ; .E+
                            (if (or (not format:fn-pos?) (eq? modifier 'at))
                                (set! numlen (+ numlen 1)))
                            (if (and (= format:fn-dot 0) (> width (+ digits 1)))
                                (set! numlen (+ numlen 1)))        
                            (set! numlen
                                  (+ numlen 
                                     (if (and edigits (>= edigits format:en-len))
                                         edigits 
                                        format:en-len)))
                            (if (< numlen width)
                                (format:out-fill (- width numlen)
                                                 (integer->char padch)))
                            (if (and overch (> numlen width))
                                (format:out-fill width (integer->char overch))
                                (begin
                                  (format:fn-out modifier (> width (- numlen 1)))
                                  (format:en-out edigits expch)))))
                      (begin
                        (format:fn-out modifier #t)
                        (format:en-out edigits expch)))))

               (else
                (format:parse-float
                 (if (string? number) number (number->string number)) #f scale)

                (format:fn-strip)

                (if width
                    (if (and edigits overch (> format:en-len edigits))
                        (format:out-fill width (integer->char overch))
                        (let ((numlen (+ format:fn-len 3))) ; .E+
                          (if (or (not format:fn-pos?) (eq? modifier 'at))
                              (set! numlen (+ numlen 1)))
                          (if (= format:fn-dot 0)
                              (set! numlen (+ numlen 1)))
                          (set! numlen
                                (+ numlen
                                   (if (and edigits (>= edigits format:en-len))
                                       edigits 
                                       format:en-len)))
                          (if (< numlen width)
                              (format:out-fill (- width numlen)
                                               (integer->char padch)))
                          (if (> numlen width) ; adjust precision if possible
                              (let ((f (- format:fn-len format:fn-dot))) ; fract len
                                (if (> (- numlen f) width)
                                    (if overch ; numstr too big for required width
                                        (format:out-fill width 
                                                         (integer->char overch))
                                        (begin
                                          (format:fn-out modifier #t)
                                          (format:en-out edigits expch)))
                                    (begin
                                      (format:fn-round (+ (- f numlen) width))
                                      (format:fn-out modifier #t)
                                      (format:en-out edigits expch))))
                              (begin
                                (format:fn-out modifier #t)
                                (format:en-out edigits expch)))))
                    (begin
                      (format:fn-out modifier #t)
                      (format:en-out edigits expch)))))))))
       
       ;; format general flonums (~G)

(define format:out-general
        (lambda (modifier number pars)
          (if (not (or (number? number) (string? number)))
              (format:error "argument is not a number or a number string"))

          (let ((l (length pars)))
            (let ((width (if (> l 0) (list-ref pars 0) #f))
                  (digits (if (> l 1) (list-ref pars 1) #f))
                  (edigits (if (> l 2) (list-ref pars 2) #f))
                  (overch (if (> l 4) (list-ref pars 4) #f))
                  (padch (if (> l 5) (list-ref pars 5) #f)))
              (cond
               ((or (inf? number) (nan? number))
                ;; FIXME: this isn't right.
                (format:out-inf-nan number width digits edigits overch padch))
               (else
                (format:parse-float
                 (if (string? number) number (number->string number)) #t 0)
                (format:fn-strip)
                (let* ((ee (if edigits (+ edigits 2) 4)) ; for the following algorithm
                       (ww (if width (- width ee) #f))   ; see Steele's CL book p.395
                       (n (if (= format:fn-dot 0)        ; number less than (abs 1.0) ?
                              (- (format:fn-zlead))
                              format:fn-dot))
                       (d (if digits
                              digits
                              (max format:fn-len (min n 7)))) ; q = format:fn-len
                       (dd (- d n)))
                  (if (<= 0 dd d)
                      (begin
                        (format:out-fixed modifier number (list ww dd #f overch padch))
                        (format:out-fill ee #\space)) ;~@T not implemented yet
                      (format:out-expon modifier number pars)))))))))

       ;; format dollar flonums (~$)

(define format:out-dollar
        (lambda (modifier number pars)
          (if (not (or (number? number) (string? number)))
              (format:error "argument is not a number or a number string"))

          (let ((l (length pars)))
            (let ((digits (format:par pars l 0 2 "digits"))
                  (mindig (format:par pars l 1 1 "mindig"))
                  (width (format:par pars l 2 0 "width"))
                  (padch (format:par pars l 3 format:space-ch #f)))

              (cond
               ((or (inf? number) (nan? number))
                (format:out-inf-nan number width digits #f #f padch))

               (else
                (format:parse-float
                 (if (string? number) number (number->string number)) #t 0)
                (if (<= (- format:fn-len format:fn-dot) digits)
                    (format:fn-zfill #f (- digits (- format:fn-len format:fn-dot)))
                    (format:fn-round digits))
                (let ((numlen (+ format:fn-len 1)))
                  (if (or (not format:fn-pos?) (memq modifier '(at colon-at)))
                      (set! numlen (+ numlen 1)))
                  (if (and mindig (> mindig format:fn-dot))
                      (set! numlen (+ numlen (- mindig format:fn-dot))))
                  (if (and (= format:fn-dot 0) (not mindig))
                      (set! numlen (+ numlen 1)))
                  (if (< numlen width)
                      (case modifier
                        ((colon)
                         (if (not format:fn-pos?)
                             (format:out-char #\-))
                         (format:out-fill (- width numlen) (integer->char padch)))
                        ((at)
                         (format:out-fill (- width numlen) (integer->char padch))
                         (format:out-char (if format:fn-pos? #\+ #\-)))
                        ((colon-at)
                         (format:out-char (if format:fn-pos? #\+ #\-))
                         (format:out-fill (- width numlen) (integer->char padch)))
                        (else
                         (format:out-fill (- width numlen) (integer->char padch))
                         (if (not format:fn-pos?)
                             (format:out-char #\-))))
                      (if format:fn-pos?
                          (if (memq modifier '(at colon-at)) (format:out-char #\+))
                          (format:out-char #\-))))
                (if (and mindig (> mindig format:fn-dot))
                    (format:out-fill (- mindig format:fn-dot) #\0))
                (if (and (= format:fn-dot 0) (not mindig))
                    (format:out-char #\0))
                (format:out-substr format:fn-str 0 format:fn-dot)
                (format:out-char #\.)
                (format:out-substr format:fn-str format:fn-dot format:fn-len)))))))

                                        ; the flonum buffers

(define format:fn-max 400)     ; max. number of number digits
(define format:fn-str #f)      ; number buffer
(define format:fn-len 0)       ; digit length of number
(define format:fn-dot #f)      ; dot position of number
(define format:fn-pos? #t)     ; number positive?
(define format:en-max 10)      ; max. number of exponent digits
(define format:en-str #f)      ; exponent buffer
(define format:en-len 0)       ; digit length of exponent
(define format:en-pos? #t)     ; exponent positive?

(define format:parse-float
        (lambda (num-str fixed? scale)
          (let ((left-zeros 0)
                (mantissa? #t)
                (all-zeros? #t)
                (c #f)                        ; current exam. character in num-str
                (num-len (string-length num-str)))
            (set! format:fn-pos? #t)
            (set! format:fn-len 0)
            (set! format:fn-dot #f)
            (set! format:en-pos? #t)
            (set! format:en-len 0)

            (let loop  ((i 0))
              (if (= i num-len)
                   (begin
                      (if (not format:fn-dot)
                          (set! format:fn-dot format:fn-len))

                      (if all-zeros?
                          (begin
                            (set! left-zeros 0)
                            (set! format:fn-dot 0)
                            (set! format:fn-len 1)))

                      ;; now format the parsed values according to format's need

                      (if fixed?

                          (begin                        ; fixed format m.nnn or .nnn
                            (if (and (> left-zeros 0) (> format:fn-dot 0))
                                (if (> format:fn-dot left-zeros) 
                                    (begin                ; norm 0{0}nn.mm to nn.mm
                                      (format:fn-shiftleft left-zeros)
                                      (set! format:fn-dot (- format:fn-dot left-zeros))
                                      (set! left-zeros 0))
                                    (begin                ; normalize 0{0}.nnn to .nnn
                                      (format:fn-shiftleft format:fn-dot)
                                      (set! left-zeros (- left-zeros format:fn-dot))
                                      (set! format:fn-dot 0))))
                            (if (or (not (= scale 0)) (> format:en-len 0))
                                (let ((shift (+ scale (format:en-int))))
                                  (cond
                                   (all-zeros? #t)
                                   ((> (+ format:fn-dot shift) format:fn-len)
                                    (format:fn-zfill
                                     #f (- shift (- format:fn-len format:fn-dot)))
                                    (set! format:fn-dot format:fn-len))
                                   ((< (+ format:fn-dot shift) 0)
                                    (format:fn-zfill #t (- (- shift) format:fn-dot))
                                    (set! format:fn-dot 0))
                                   (else
                                    (if (> left-zeros 0)
                                        (if (<= left-zeros shift) ; shift always > 0 here
                                            (format:fn-shiftleft shift) ; shift out 0s
                                            (begin
                                              (format:fn-shiftleft left-zeros)
                                              (set! format:fn-dot (- shift left-zeros))))
                                        (set! format:fn-dot (+ format:fn-dot shift))))))))

                          (let ((negexp                ; expon format m.nnnEee
                                 (if (> left-zeros 0)
                                     (- left-zeros format:fn-dot -1)
                                     (if (= format:fn-dot 0) 1 0))))
                            (if (> left-zeros 0)
                                (begin                        ; normalize 0{0}.nnn to n.nn
                                  (format:fn-shiftleft left-zeros)
                                  (set! format:fn-dot 1))
                                (if (= format:fn-dot 0)
                                    (set! format:fn-dot 1)))
                            (format:en-set (- (+ (- format:fn-dot scale) (format:en-int))
                                              negexp))
                            (cond 
                             (all-zeros?
                              (format:en-set 0)
                              (set! format:fn-dot 1))
                             ((< scale 0)                ; leading zero
                              (format:fn-zfill #t (- scale))
                              (set! format:fn-dot 0))
                             ((> scale format:fn-dot)
                              (format:fn-zfill #f (- scale format:fn-dot))
                              (set! format:fn-dot scale))
                             (else
                              (set! format:fn-dot scale)))))
                      #t)

                  (begin ;; else 
                   (set! c (string-ref num-str i))        ; parse the output of number->string
                   (cond                                  ; which can be any valid number
                    ((char-numeric? c)                    ; representation of R4RS except 
                     (if mantissa?                        ; complex numbers
                         (begin
                           (if (char=? c #\0)
                               (if all-zeros?
                                   (set! left-zeros (+ left-zeros 1)))
                               (begin
                                 (set! all-zeros? #f)))
                           (string-set! format:fn-str format:fn-len c)
                           (set! format:fn-len (+ format:fn-len 1)))
                         (begin
                           (string-set! format:en-str format:en-len c)
                           (set! format:en-len (+ format:en-len 1)))))
                    ((or (char=? c #\-) (char=? c #\+))
                     (if mantissa?
                         (set! format:fn-pos? (char=? c #\+))
                         (set! format:en-pos? (char=? c #\+))))
                    ((char=? c #\.)
                     (set! format:fn-dot format:fn-len))
                    ((char=? c #\e)
                     (set! mantissa? #f))
                    ((char=? c #\E)
                     (set! mantissa? #f))
                    ((char-whitespace? c) #t)
                    ((char=? c #\d) #t)                ; decimal radix prefix
                    ((char=? c #\#) #t)
                    (else
                     (format:error "illegal character `~c' in number->string" c)))
(loop (+ i 1))
))) )))

(define format:en-int
        (lambda ()                        ; convert exponent string to integer
          (if (= format:en-len 0)
              0
               (let loop ((i 0) (n 0))
                (if (= i format:en-len) 
                   (if format:en-pos? n (- n))
                   (loop (+ i 1)
                         (+ (* n 10) (- (char->integer (string-ref format:en-str i))
                                     format:zero-ch))) )) )))

(define format:en-set          ; set exponent string number
        (lambda (en)                
          (set! format:en-len 0)
          (set! format:en-pos? (>= en 0))
          (let* ((en-str (number->string en))
                 (en-len (string-length en-str)))
           (let loop ((i 0) (c (string-ref en-str 0)))
            (when (not (= i en-len))
              (when (char-numeric? c)
                (string-set! format:en-str format:en-len c)
                (set! format:en-len (+ format:en-len 1)))
              (loop (+ i 1) (string-ref en-str i)))) 
)))

(define format:fn-zfill        ; fill current number string with 0s
        (lambda (left? n)
          (if (> (+ n format:fn-len) format:fn-max) ; from the left or right
              (format:error "number is too long to format (enlarge format:fn-max)"))
          (set! format:fn-len (+ format:fn-len n))
          (if left?
              (let loop ((i format:fn-len))        ; fill n 0s to left
                (when (not (< i 0)) 
                  (string-set! format:fn-str i
                               (if (< i n)
                                   #\0
                                   (string-ref format:fn-str (- i n)) ))
                  (loop (- i 1))))
              (let loop ((i (- format:fn-len n))) ; fill n 0s to the right
                (when (not (= i format:fn-len))
                  (string-set! format:fn-str i #\0)
                  (loop (+ i 1)) )) )))

(define format:fn-shiftleft                ; shift left current number n positions
        (lambda (n)
          (if (> n format:fn-len)
              (format:error "internal error in format:fn-shiftleft (~d,~d)"
                            n format:fn-len))
          (let loop ((i n))
            (if (= i format:fn-len)
                (set! format:fn-len (- format:fn-len n))
                (begin
                  (string-set! format:fn-str (- i n)
                               (string-ref format:fn-str i))
                  (loop (+ i 1)) ))) ))

(define format:fn-round        ; round format:fn-str
        (lambda (digits)
          (set! digits (+ digits format:fn-dot))
          (let loop ((i digits)                ; "099",2 -> "10"
                     (c 5))                ; "023",2 -> "02"
            (if (not (or (= c 0) (< i 0)))
                (begin
            (set! c (+ (- (char->integer (string-ref format:fn-str i))
                          format:zero-ch) c))
            (string-set! format:fn-str i (integer->char
                                          (if (< c 10) 
                                              (+ c format:zero-ch)
                                              (+ (- c 10) format:zero-ch))))
            (set! c (if (< c 10) 0 1))
            (loop (- i 1) c))

                (begin
                  (if (= c 1)                        ; "005",2 -> "01"
                   (begin                        ; carry overflow
                     (set! format:fn-len digits)
                     (format:fn-zfill #t 1)        ; add a 1 before fn-str
                     (string-set! format:fn-str 0 #\1)
                     (set! format:fn-dot (+ format:fn-dot 1)))
                   (set! format:fn-len digits))) )) ))

(define format:fn-out
        (lambda (modifier add-leading-zero?)
          (if format:fn-pos?
              (if (eq? modifier 'at) 
                  (format:out-char #\+))
              (format:out-char #\-))
          (if (= format:fn-dot 0)
              (if add-leading-zero?
                  (format:out-char #\0))
              (format:out-substr format:fn-str 0 format:fn-dot))
          (format:out-char #\.)
          (format:out-substr format:fn-str format:fn-dot format:fn-len)))

(define format:en-out
        (lambda (edigits expch)
          (format:out-char (if expch (integer->char expch) format:expch))
          (format:out-char (if format:en-pos? #\+ #\-))
          (if edigits 
              (if (< format:en-len edigits)
                  (format:out-fill (- edigits format:en-len) #\0)))
          (format:out-substr format:en-str 0 format:en-len)))

(define format:fn-strip                ; strip trailing zeros but one
        (lambda ()
          (string-set! format:fn-str format:fn-len #\0)
          (let loop ((i format:fn-len))
           (if (or (not (char=? (string-ref format:fn-str i) #\0))
                   (<= i format:fn-dot))
               (set! format:fn-len (+ i 1))
               (loop (- i 1)))) ))

(define format:fn-zlead                ; count leading zeros
        (lambda ()
          (let loop ((i 0))
            (when (not (or (= i format:fn-len)
                           (not (char=? (string-ref format:fn-str i) #\0))))
              (if (= i format:fn-len)                ; found a real zero
                   0
                   (loop (+ i 1)))))))


;;; some global functions not found in SLIB

(define string-capitalize-first        ; "hello" -> "Hello"
        (lambda (str)
          (let ((cap-str (string-copy str))        ; "hELLO" -> "Hello"
                (non-first-alpha #f)                ; "*hello" -> "*Hello"
                (str-len (string-length str)))        ; "hello you" -> "Hello you"
            (let loop ((i 0))
              (if (= i str-len) cap-str
                  (let ((c (string-ref str i)))
                    (if (char-alphabetic? c)
                        (if non-first-alpha
                            (string-set! cap-str i (char-downcase c))
                            (begin
                              (set! non-first-alpha #t)
                              (string-set! cap-str i (char-upcase c)))))
                    (loop (+ i 1))) )) )))


       ;; Aborts the program when a formatting error occures. This is a null
       ;; argument closure to jump to the interpreters toplevel continuation.

(define format:abort (lambda () (error "error in format")))

;; How about some cutting let* called let:cut (or kall the other let-keep)?
;; BTW What was the difference between letrec and let*?
;; HS doesn't know about /letrec/ (as /let*/ is procounced like that but
;; we have all the three of them) but guess what /let/ operates in parallel
;; but evaluates the form step by step? [It's about their binding routines.
;; Is that old /let/ parallel binding thing still needed or do we just want
;; to change /let/ to be /let*/ with top down sequence again? [you can do
;; this easily in the core but the monster has its protocols of inner
;; dialects, doesn't it.]
;; There was a keyword /special/ (lexical vs dynamic variable and sth wrt
;; references) and if things keep going like that ANS soon looks like SCA
;; which itself looks like a poor man's CL?!!!
(define format:rc-cut (lambda (val)
  ;; XXX missing CLR4 autocut for inner defines
  (set! format:out-str        #f)
  (set! format:out-char       #f)
  (set! format:out-substr     #f)
  (set! format:out-fill       #f)
  (set! format:error          #f)
  (set! format:intern-error   #f)
  (set! format:error-save     #f)
  (set! format:format         #f)
  (set! format:out            #f)
  (set! format:parameter-characters #f)
  (set! format:format-work    #f) ;
  (set! format:obj->str       #f)
  (set! format:char->str      #f)
  (set! format:par            #f)
  (set! format:out-obj-padded #f)
  (set! format:out-num-padded #f)
  (set! format:tabulate       #f)
  (set! format:num->cardinal999  #f)
  (set! format:num->cardinal  #f)
  (set! format:num->ordinal   #f)
  (set! format:out-inf-nan    #f)
  (set! format:out-fixed      #f)
  (set! format:out-expon      #f)
  (set! format:out-general    #f)
  (set! format:out-dollar     #f)
  (set! format:parse-float    #f)
  (set! format:en-int         #f)
  (set! format:en-set         #f)
  (set! format:fn-zfill       #f)
  (set! format:fn-shiftleft   #f)
  (set! format:fn-round       #f)
  (set! format:fn-out         #f)
  (set! format:en-out         #f)
  (set! format:fn-strip       #f)
  (set! format:fn-zlead       #f)
  (set! string-capitalize-first #f)
  (set! format:abort          #f)
  (set! format:rc-cut     #f)

(set! format:port #f)                     ; curr. format output port
(set! format:output-col 0)                ; curr. format output tty column
(set! format:flush-output #f)             ; flush output at end of formatting
(set! format:case-conversion #f)
(set! format:args #f)
(set! format:pos 0)                       ; curr. format string parsing position
(set! format:arg-pos 0)                   ; curr. format argument position
  val))

  ; begin
  (set! format:error-save format:error)
  (set! format:fn-str (make-string format:fn-max #\0)) ; number buffer
  (set! format:en-str (make-string format:en-max #\0)) ; exponent buffer
  ;(apply format:format args))
  (format:rc-cut (apply format:format args))
) ; closes format
;)) ; closes namespace
