;; 
;; XXX check this please
;; Derived from "Transcen.scm", Complex trancendental functions for SCM.
;;
;; EXPERIMENTAL:
;; Rounding error harmonization for trigonometric functions by keeping
;; the arguments to sine and cosine within the period of 2 pi.
;; TODO Try this for the tangens, too. (See also the corresponding
;; issue in the development notes.)
;;
;; The previous definitions included the handling of complex
;; numbers which seemed bogus. They are left here commented
;; until they definitions of sine and cosine for complex numbers
;; will have been reviewed.
;;
;; NB /inexact->exact/ works for inexact numbers, too.
;;

(define (expt z1 z2)  ;; XXX original variant to be reviewed
  (cond ((and (exact? z2) (integer? z2))
           (integer-expt z1 z2))
        ((and (real? z2) (real? z1) (>= z1 0))
           ($expt z1 z2))
        (else
           (exp (* z2 (log z1))) )))

(define (delta x y) ;; ??? rename to numerical-distance? 
 (cond ((>= x y) (- x y))
       (else     (- y x))))


(define scalar? (lambda (obj) (and (number? obj)
                                   (not (nan? obj))
                                   (= 0 (imag-part obj)) )))

;(define (scalar? obj) (and (number? obj) (not (complex? obj)))) ; <- depends on math pred revisionative change


(define (sin ro) ($sin (exact->inexact (remainder ro (* 2 pi)))))
(define (cos ro) ($cos (exact->inexact (remainder ro (* 2 pi)))))

(define (tan z)
  (if (real? z) ($tan z)
      (let* ((x (* 2 (real-part z)))
	     (y (* 2 (imag-part z)))
	     (w (+ ($cos x) ($cosh y))))
	(make-rectangular (/ ($sin x) w) (/ ($sinh y) w)))))

;; (define (asin z) XXX 
;;   (if (and (real? z) (>= z -1) (<= z 1))
;;       ($asin z)
;;       (* -i (asinh (* +i z)))))

(define (asin z) ($asin (exact->inexact z)))
(define (acos z) ($acos (exact->inexact z)))

;; (define (acos z)
;;   (if (and (real? z) (>= z -1) (<= z 1))
;;       ($acos z)
;;       (+ (/ (angle -1) 2) (* +i (asinh (* +i z))))))

(define (atan z . y)
  (if (null? y)
      (if (real? z) ($atan z)
	  (/ (log (/ (- +i z) (+ +i z))) +2i))
      ($atan2 z (car y))))


(define integer-modulo modulo)
(define (quotient  a b) (floor (/ a b)))
(define (remainder a b) (- a (* b (quotient a b))))


;;
;; Some Constants
;;
(define inexact-pi     3.14159265358979323846)
(define inexact-e      2.71828182845904523536)
(define inexact-log10E 0.43429448190325182765)  ; (/ 1 (log 10))


;; Numerical Abbreviations
(define pi inexact-pi)
(define 1k 1000)
(define 1M (* 1k 1k))


;;
;; Additional Inexact Number Routines
;;

(define (inexact= accuracy . operands)
 (or (null? operands)
     (null? (cdr operands))
     (reduce-predicate (lambda (a b) (<= (abs (- a b)) accuracy))
                       operands)))

; (define (accuracy accuracy x) (- x (remainder x accuracy))) <- depends on /quotient/ /remainder/ /modulo/ issue XXX
;          ^^^^^^^^ some other name here?


;;
;; Original Hyperbolisms ;-)
(define (sinh z)
  (if (real? z) ($sinh z)
      (let ((x (real-part z)) (y (imag-part z)))
	(make-rectangular (* ($sinh x) ($cos y))
			  (* ($cosh x) ($sin y))))))
(define (cosh z)
  (if (real? z) ($cosh z)
      (let ((x (real-part z)) (y (imag-part z)))
	(make-rectangular (* ($cosh x) ($cos y))
			  (* ($sinh x) ($sin y))))))
(define (tanh z)
  (if (real? z) ($tanh z)
      (let* ((x (* 2 (real-part z)))
	     (y (* 2 (imag-part z)))
	     (w (+ ($cosh x) ($cos y))))
	(make-rectangular (/ ($sinh x) w) (/ ($sin y) w)))))

(define (asinh z)
  (if (real? z) ($asinh z)
      (log (+ z (sqrt (+ (* z z) 1))))))

(define (acosh z)
  (if (and (real? z) (>= z 1))
      ($acosh z)
      (log (+ z (sqrt (- (* z z) 1))))))

(define (atanh z)
  (if (and (real? z) (> z -1) (< z 1))
      ($atanh z)
      (/ (log (/ (+ 1 z) (- 1 z))) 2)))
