;;
;; The traditional Read Eval Print Loop (REPL) is here
;;
;; implemented as a Prompt Read Echo Eval Print Loop (PREPL)
;;
;; with error handling and error handling and colors.
;;
(sys-load "VSI-core/terminal.scm")

;; XXX experimental namespace testspace
;; In fact the new /load/ variants should remain in bounds and
;; load into the current namespace respectively the current
;; environment. Where the term "environment" only makes sense
;; for the outer (textual) form which is then "transposed" onto
;; expressions and further implementation details. With the new
;; /load/ variants the user can choose wether to plainly load a
;; file into the system environment of his program or if he
;; prefers to import into a namespace object to then selectively
;; import or use individual definitions (or bindings). The current
;; setup in this file serves to test the new namespaces and their
;; prototype implementation.
;; (sys-load "VSI-core/namespace.scm")
;; (define repl-namespace (namespace
;;  (namespace-import (namespace-load "VSI-core/terminal.scm"))

;; The following parameters can't be set from outside our namespace.
;; We should provide some accessors.
(define repl-echoing  #t)
(define repl-coloring #t)

;; One could try to detect the current terminal size with /get-dimensions/
;; and offer limits based on percentage or line numbers. But the REPL is
;; better known for its simplicity.
(define repl-printer:limit    (* 74 23)) ; use +inf.0 to turn the limit off
(define repl-printer:ellipsis " ... ")

(define leave-this-repl #f)
(when (not (defined? 'repl:level)) (define repl:level 0))

(define (quit-repl) (set! leave-this-repl #t)) ;; ??? drain?

(define (sys-REPL expression environment)
 (when (> (length expression) 1) (error "mischiefed REPL statement" expression))
 (repl-drain-input (current-input-port))

 (set! repl:level (+ repl:level 1))

 (catch #t (lambda () (repl-PREPL environment)) repl-error-handler)

 (set! repl:level (- repl:level 1))

 (if leave-this-repl
   (set! leave-this-repl #f)
   (sys-REPL expression environment)))

(define REPL (procedure->syntax sys-REPL))

(define (repl-PREPL environment)
 (repl-prompt)
 (repl-print (repl-eval (repl-echo (read)) environment ))
 (newline)
 (when (not leave-this-repl)
   (repl-PREPL environment)))

(define (repl-eval expression environment)
 ; EXPERIMENTAL Quit on EOF
 (if (eof-object? expression)
   (quit-repl)
   (let ((evaluated (eval! expression environment)))
     (trim)  ; EXPERIMENTAL trim inclusion
     evaluated)))

(define (repl-prompt)
  (when (isatty? (current-input-port))
    (%terminal-cooked-mode (port->fdes (current-input-port)))
    ;XXX (repl-drain-input      (current-input-port))
    (repl:color 'default)
    (when (isatty? (current-output-port))
      (erase-line (current-output-port))
      (if (colored-repl?)
        (repl-terminal-prompt)
        (display "VSI> "))
      (cursor-show (current-output-port))))
  (force-output)
)

(define (colored-repl?)
  (and repl-coloring (isatty? (current-output-port))))

(define (repl:white-on-black what)
 (when (colored-repl?)
  (let ((terminal (current-output-port)))
   (tty-color terminal 'background 'black)
   (case what
    ((default)
       (tty-color terminal 'foreground 'white))
    ((evaluation)
       ;(tty-color terminal 'background 'black)
       (tty-color terminal 'foreground 'green))
    ((unspecified)
       (tty-color terminal 'foreground 'blue))
    ((ellipsis)
       (tty-color terminal 'background 'yellow)
       (tty-color terminal 'foreground 'white))
    ((error)
       (tty-color terminal 'foreground 'red))
 )) ))
(define (repl:blue-on-white what)
 (when (colored-repl?)
  (let ((terminal (current-output-port)))
   (tty-color terminal 'background 'white)
   (case what
    ((default)
       (tty-color terminal 'foreground 'blue))
    ((evaluation)
       (tty-color terminal 'foreground 'green))
    ((unspecified)
       (tty-color terminal 'foreground 'green))
    ((ellipsis)
       (tty-color terminal 'background 'blue)
       (tty-color terminal 'foreground 'white))
    ((error)
       (tty-color terminal 'foreground 'red))
 )) ))
(define (repl:white-on-blue what)
 (when (colored-repl?)
  (let ((terminal (current-output-port)))
   (tty-color terminal 'background 'blue)
   (case what
    ((default)
       (tty-color terminal 'foreground 'white))
    ((evaluation)
       (tty-color terminal 'foreground 'green))
    ((unspecified)
       (tty-color terminal 'foreground 'black))
    ((ellipsis)
       (tty-color terminal 'background 'white)
       (tty-color terminal 'foreground 'black))
    ((error)
       (tty-color terminal 'foreground 'red))
 )) ))

;; White on Black is currently the only one compatible
;; with virtual terminal color profiles where white
;; and black may be mapped to user defined colors.
;(define repl:color repl:white-on-blue)
(define repl:color repl:blue-on-white)
;(define repl:color repl:white-on-black)

(define (repl-terminal-prompt)
 (let ((terminal (current-output-port)))
  (repl:color 'default)
  (tty-color terminal 'foreground 'cyan)    (display "V")
  (tty-color terminal 'foreground 'red)     (display "S")
  (tty-color terminal 'foreground 'green)   (display "I")
  (repl:color 'default)                 (display "> ")))

(define (repl-echo expression)
  (when repl-echoing
    (repl:color 'default)    (write expression)
    (repl:color 'evaluation) (display " => ")
    (repl:color 'default)
    (force-output))
  (set! repl-read-done #t)  ; Error Loop Check
  expression)

(define (repl-print:truncate string port)
 (let ((n (string-length string))
       (half (quotient (- repl-printer:limit
                          (string-length repl-printer:ellipsis))
                       2)))
   (repl:color 'default)  (display (substring string 0 half) port)
                          (newline port)
   (repl:color 'ellipsis) (display repl-printer:ellipsis port)
                          (newline port)
   (repl:color 'default)  (display (substring string (- n half) n) port) ))

(define (repl-print value . port-argument)
 (let ((port (or (and (pair? port-argument) (car port-argument))
                 (current-output-port))))
  (cond
   ((eq? value *unspecified*)
     (repl:color 'unspecified)
     (write value))
   (else
    (let ((output (with-output-to-string (lambda () (write value)) )))
     (if (> (string-length output) repl-printer:limit)
       (repl-print:truncate output port)
       (begin (repl:color 'default) (display output port))) )))
 (repl:color 'default)))


(define repl-read-done #t)
(define (repl-error-handler key . args)
  ;; pre error handling throw key dispatch
  (case key
    ((quit)
       ;; ??? w/o the following input drain other command interpreters
       ;; start to interpret our rests: "(quit)ls" would show you a
       ;; directory listing.
       (repl-drain-input (current-input-port))
       (exit 0))
    ((read-error)
       ;; When /read/ throws an error this means that a read was done.
       (set! repl-read-done #t)))

  (repl:color 'error) (display "throw catched: ")
  (repl:color 'default)

  (cond
   ;; Regular error
   ((and (pair? args) (pair? (cdr args)) (string? (cadr args)))
      (when (not (false-if-exception
                (repl-regular-error-printer key args)))
      ;
      ; XXX  Most of those conditions that make simple format
      ;  fail output their own error messages but forget about
      ;  adding a line feed.
      (newline (current-error-port))
      ; XXX There have been to many missing error messages. We try
      ; to catch the error reporting protocol violations by using
      ; the fallback error handler.
      (display "error fallback handler: " (current-error-port))
      (repl-fallback-error-printer key args)) )

   ;; Fallback error reporting
   (else
    (repl-fallback-error-printer key args)) )

  ;; Error Post Processing
  ; Check for an error loop. This can happen if one of the variables
  ; or procedures used by the REPL is redefined in an incompatible
  ; way.
  (if (not repl-read-done)
    (begin (display "VSI REPL: internal error" (current-error-port))
           (newline (current-error-port))
           (exit 1))
    (set! repl-read-done #f))
  ; Clear input output
  (force-output (current-error-port))
  ;(repl-drain-input (current-input-port))
  )

(define (repl-fallback-error-printer key args)
  (display key  (current-error-port))
  (when (not (null? args))
    (display    ": "   (current-error-port))
    (repl-print args   (current-error-port)))
  (newline      (current-error-port))
  #t)

(define (repl-regular-error-printer key args)
  (repl-print key (current-error-port))
  (display ": "   (current-error-port))
  (when (car args)
      (repl-print (car args) (current-error-port))
      (display    ": "       (current-error-port)))
  (let ((buffer #f)
        (format-string     (cadr args))
        (format-arguments (caddr args))) 
    (when (not format-arguments)
      (set! format-arguments '())) ; XXX lisp error format protocol rest 
    (set! buffer (with-output-to-string (lambda ()
                   (apply simple-format ; some manual nconc2last here 
                                        (append (list #t    ; port argument 
                                                      format-string)
                                                format-arguments)) )))
    (repl-print buffer (current-error-port)))
  (newline (current-error-port))
  #t)

(define (repl-drain-input port)
 (when (char-ready? port)
  (when (not (eof-object? (read-char port)))
   (repl-drain-input port))))

;;
;; REPL auxiliary
;;

;;
;; Clear the output screen if it is a terminal.
(define (repl-clear)
 (if (isatty? (current-output-port))
   (begin (erase-display (current-output-port))
          (cursor-home (current-output-port))
          #t)
   (begin #f)))

(define (repl-level)
 (let ((terminal (current-output-port))
       (M (- repl:level 1)))
  (repl:color 'default)
  (when (= M 3) (display (format:num->roman (- M 2))))
  (when (< M 8)
  (if (<= 3 M 4)
    (repl:color 'default)
    (tty-color terminal 'foreground 'cyan)) (display "V")
  (when (< M 5)
  (tty-color terminal 'foreground 'red)     (display "S"))
  (tty-color terminal 'foreground 'green)   (display "I")
  (when (<= 1 M 2)                      (display (format:num->roman M)))
  (when (<= 6 M 7)                      (display (format:num->roman (- M 5)))))
  (when (>= M 8) (tty-color terminal 'foreground
  (list-ref '(cyan red green) (modulo (quotient (+ M 1) 5) 3)))
  (display (format:num->roman (+ M 1))))
  (repl:color 'default)                 (display "> ")))
  (set! repl-terminal-prompt repl-level)

;;
;; (transcript expressions)
;;
;; Experimental form to produce transcripts for
;; given expressions. This is meant to be used
;; in files to "simulate" a REPL session. It
;; doesn't save any transcript into a file.
;;
;; - Memoized expressions are shown as such but shouldn't be.
;; - Redirections in effect of the standard I/O ports may give
;;   surprising results.
;;
(define (repl-transcript:loop expressions environment)
  (when (not (null? expressions))
   (let ((expression (first expressions)))
    ;; unmemoize_expression in interpreter.eval.c should do the job but
    ;; needs to be exported to Scheme level.
    ;; (repl-echo (if (and (pair? expression) (memoized? (car expression)))
    ;;              (unmemoize-expr expression)
    ;;              expression))
    (repl-echo expression)
    ;; ??? Maybe this should use /repl-eval/ to call /trim/?
    ;; Maybe we should /trim/ only when interactive, i.e. in the prompt?
    (repl-print (eval expression environment))
    (newline)
    (repl-transcript:loop (cdr expressions) environment) )))

(define (repl-transcript:transformer expression environment)
 (catch #t (lambda () (repl-transcript:loop (cdr expression) environment))
        repl-error-handler))

(define repl-transcript (procedure->syntax repl-transcript:transformer))

;;
;; (REPL-HERE)
;;
;; Starts a REPL at that point in that environment. [The REPL can do that, too now.]
;; (Experimental | it looks like a debugger and should become breakpoints)
;;
;; TODO provide means to specify a return value
;; XXX Port indirection lead to unexpected results: some new IO path
;; should provide information about a controlling terminal and maybe
;; default ports should be replaced with mandatory port arguments?
;;
;; TODO incorporate a little something like this please:
;; (let ((port (current-input-port)))
;;   (with-input-from-file "/dev/tty" (lambda () (REPL-HERE))) ...
;;
(define no-repl-here #f)
(define (no-more-repl-here)
  (set! no-repl-here    #t)
  (set! leave-this-repl #t))
(define (repl-here:transformer expressions environment)
 (when (not (null? (cdr expressions)))
   (error "repl-here: syntax error"))
 (when (not no-repl-here)
  (when (colored-repl?)
    (tty-color (current-output-port) 'foreground 'yellow))
  (display " * REPL HERE * ")
  (newline)
  (display " Use (quit-repl) or (no-more-repl-here) to continue. ")
  (newline)
  (display "Local environment: (the-environment)")
  (newline)

  (repl-drain-input (current-input-port))
  (catch #t (lambda () (repl-here:PREPL environment))
         repl-error-handler)
  (if leave-this-repl
    (set! leave-this-repl #f)
    (repl-here:PREPL environment))))

(define (repl-here:PREPL environment)
 (repl-prompt)
 (repl-print (eval! (repl-echo (read)) environment ))
 (newline)
 (when (not leave-this-repl)
   (repl-here:PREPL environment)))
(define REPL-HERE (procedure->syntax repl-here:transformer))


;; )) ; end of repl-namespace

;; (define REPL            (from-namespace repl-namespace 'REPL))
;; (define repl-transcript (from-namespace repl-namespace 'repl-transcript))
;; (define REPL-HERE       (from-namespace repl-namespace 'REPL-HERE))
; XXX quit-repl is missing
;
; perhaps we should use some selective import here to achieve the
; export
; XXX experimental namespace repl
