;; XXX Resolve Dependencies
;; TODO prepend prefix terminal or deliver exporting scopes
;; (load "pretext.scm")
;; (load "characters.scm")

;;   :export (<terminal> full-reset
;;            tty-raw-mode tty-sane-mode
;;            SRM-echo-off SRM-echo-on
;;            CSI-do-send-meta-escape CSI-dont-send-meta-escape
;;            CSI-do-send-alt-escape CSI-dont-send-alt-escape
;;            save-cursor restore-cursor
;;            position-cursor cursor-home
;;            cursor-up cursor-down
;;            cursor-forward cursor-backward
;;            beep color
;;            erase-line erase-display
;;            backspace
;;            get-dimensions
;;            start-event-loop))

;; XXX Some terminals feed back unrecognized sequences (Gnome
;; Terminal) as input. It might be a good a idea to drain input
;; occasionally???
;; XXX We could also check the argument port with /isatty?/ but right
;; now error reporting isn't there yet.
(define (control port control-characters)
  (display (list->string control-characters) port))

(define (full-reset port)
  (control port '(#\esc #\c)))

;;
;; Control Sequence Introduction
(define CSI '(#\esc #\[))

;;
;; SRM: Send Receive Mode
;; turn terminal echo on or off
(define (SRM-echo-off port)
  (control port (append CSI '(#\1 #\2 #\h))))
(define (SRM-echo-on port)
  (control port (append CSI '(#\1 #\2 #\l))))

;;
;; CSI: Cursor Control
(define (position-cursor terminal x y)
  (let ((column (string->list (number->string x)))
        (line   (string->list (number->string y))))
    (control terminal
             (append CSI line '(#\;) column '(#\H)))))
(define (save-cursor terminal)
  (control terminal '(#\esc #\7)))
(define (restore-cursor terminal)
  (control terminal '(#\esc #\8)))

(define (cursor-home terminal)
  (control terminal (append CSI '(#\0 #\; #\0 #\H))))
(define (cursor-up terminal)
  (control terminal (append CSI '(#\A))))
(define (cursor-down terminal)
  (control terminal (append CSI '(#\B))))
(define (cursor-forward terminal)
  (control terminal (append CSI '(#\C))))
(define (cursor-backward terminal)
  (control terminal (append CSI '(#\D))))

(define (cursor-show terminal)
  (control terminal (append CSI '(#\? #\2 #\5 #\h))))
(define (cursor-hide terminal)
  (control terminal (append CSI '(#\? #\2 #\5 #\l))))


;;
;; CSI: Erase
(define (erase-line terminal)
  (control terminal (append CSI '(#\2 #\K))))
;; XXX pushes lines into vt scroll-buffer
(define (erase-display terminal)
  (control terminal (append CSI '(#\2 #\J))))

;;
;; CSI: Character Attributes / Colors
(define (inverse terminal)
  (control terminal (append CSI '(#\7 #\m))))

(define (tty-color terminal foreground/background color-name)
 (control terminal
  (append CSI
    (case foreground/background
          ((foreground) '(#\3))
          ((background) '(#\4))
          (else (error "terminal: color foreground/background: domain error")))
    (case color-name
          ((black)   '(#\0))
          ((red)     '(#\1))
          ((green)   '(#\2))
          ((yellow)  '(#\3))
          ((blue)    '(#\4))
          ((magenta) '(#\5))
          ((cyan)    '(#\6))
          ((white)   '(#\7))
          ((default original) '(#\8))
          (else (error "terminal: color-name: domain error" color-name)))
    '(#\m) )))

(define (reset-colors terminal)
  (tty-color terminal 'background 'default)
  (tty-color terminal 'foreground 'original))

;;
;; Editing

;; (define-method (backspace (terminal <terminal>)) ;; XXX viperism
;;   (cursor-backward terminal)
;;   (display #\space (get-port terminal))
;;   (cursor-backward terminal) )

;;
;; CSI: DEC Private / Meta Escape, Alt Escape
(define (CSI-do-send-meta-escape terminal)
  (control terminal (append CSI '(#\? #\1 #\0 #\3 #\6 #\h))))
(define (CSI-do-send-alt-escape terminal)
  (control terminal (append CSI '(#\? #\1 #\0 #\3 #\9 #\h))))
(define (CSI-dont-send-meta-escape terminal)
  (control terminal (append CSI '(#\? #\1 #\0 #\3 #\6 #\l))))
(define (CSI-dont-send-alt-escape terminal)
  (control terminal (append CSI '(#\? #\1 #\0 #\3 #\9 #\l))))

;; ;;
;; ;; DEFUNCT CSI: Window Control (dtterm&extensions, XTerm)
;; (define (CSI-window-control string)
;;   (append CSI (string->list string)))
;; (define (get-dimensions/CSI terminal)
;;   "Deprecated routine of how not to do it."
;;   (let ((default-values (values #f 80 24)) ; detected? X Y
;;         (response-introduction (append CSI '(#\8 #\;)))
;;         (port (get-port terminal))
;;         (x '()) (y '()))
;;   (define (x-value char)
;;     (cond ((digit? char)
;;             (set! x (append x (list char)))
;;             (x-value (read-char port)))
;;           ((and (eq? char #\t) (not (null? x)))
;;                 #t)
;;           ((and (eq? char #\;) (null? x))
;;                 #f)
;;           (else #f)))
;;   (define (y-value char)
;;     (cond ((digit? char) 
;;              (set! y (append y (list char)))
;;              (y-value (read-char port)))
;;           ((and (eq? char #\;) (not (null? y)))
;;                 #t)
;;           ((and (eq? char #\;) (null? y))
;;                 #f)
;;           (else
;;            ;(my-debug " y-value: not-a-digit " (print-safely char))
;;            #f)))
;;    (control terminal (append CSI (string->list "18t")))
;;    (if (and (begin (sleep 1) #t) ; should we rely on this time interval ???
;;             (char-ready? port)          ; some VTs do not respond to this and
;;             (eat!-char-sequence? response-introduction port) ; we rely on CSI.
;;             (y-value (read-char port))  ; y comes first
;;             (x-value (read-char port))) ; and x is second.
;;        (values #t ; yes, it's true: we detected these values.
;;                (string->number (list->string x))  ; all checks done
;;                (string->number (list->string y))) ;   => this can't fail
;;        default-values))) ; read out failed => default values
;;
(define (get-port port) port) ; XXX
(define (get-dimensions/VSI-C terminal)
  (let ((default-values (values #f 80 24)) ; (not) detected width height 
        (file-descriptor (port->fdes (get-port terminal))))
    (let ((answer (%terminal-get-win-size file-descriptor)))
      (if answer
          (receive answer (x y pixel-x pixel-y)
                   (values #!detected:!#  #t
                           #!   width:!#  x
                           #!  height:!#  y ))
          default-values)) ))

;; XXX need buffered and unbuffered versions (update-dimensions)
(define (get-dimensions terminal)
  (get-dimensions/VSI-C terminal))
(define (get-width terminal)
  (call-with-values (lambda () (get-dimensions terminal))
                    (lambda (detected x y) x)))
(define (get-height terminal)
  (call-with-values (lambda () (get-dimensions terminal))
                    (lambda (detected x y) y)))


;;
;; Raw Mode / Sane Mode
(define (private/tty-raw-mode tty)
  (system (string-append "stty --file=" (ttyname tty) " "
                         "raw" " "
                         "-echo")))
(define (private/tty-sane-mode tty)
  (system (string-append "stty --file=" (ttyname tty) " "
                         "sane")))
(define (tty-raw-mode terminal)
  (%terminal-raw-mode (port->fdes (get-port terminal)))) ; XXX local echo still on
(define (tty-cooked-mode terminal)
  (%terminal-cooked-mode (port->fdes (get-port terminal)))) ; XXX tested. works

;;
;; Beep (n/a or output the control character)
(define (beep terminal)
  (if #f 'unspecified))

(define (bell terminal) (display #\bel terminal))

;;
;; Printing on the Virtual Terminal
;; (define-method (display (argument <object>) (terminal <terminal>))
;;   (display argument (get-port terminal)))
;(define-slot <terminal> 'display ...)

;;
;; Input & Event Loop
(define (eat!-char? char port)
  (eq? char (read-char port)))
(define (eat!-char-sequence? char-list port)
  (if (null? char-list)
      #t
      (if (eat!-char? (car char-list) port)
          (eat!-char-sequence? (cdr char-list) port)
          #f)))

(define (debug-drain port)
  (define (inner-drain)
    (if (not (char-ready? port))
        '()
        (cons (represent-char-safely (read-char port))
              (inner-drain))))
  (reverse! (inner-drain)))
(define (fast-drain port)
  (if (not (char-ready? port))
      #t
      (begin (read-char port) (fast-drain))))
;; (define (drain port) ;; XXX does not drain reliably
;;   (map represent-char-safely
;;        (string->list (drain-input port))))
(define unrecognized-input #f)
(define (convert-csi-input port)
  (let ((char (read-char port)))
    (case char
      ((#\A) 'up)
      ((#\B) 'down)
      ((#\C) 'right)
      ((#\D) 'left)
      ((#\H) 'home)
      ((#\F) 'end)
      ((#\1) (if (eat!-char? #\~ port) 'home #f)) ; vt220 find / linux home
      ((#\2) (if (eat!-char? #\~ port) 'insert-here  #f)) ; numpad enter???
      ((#\3) (if (eat!-char? #\~ port)  #\del        #f))
      ((#\4) (if (eat!-char? #\~ port) 'end #f)) ; vt220 select / linux end
      ((#\5) (if (eat!-char? #\~ port) 'page-up      #f))
      ((#\6) (if (eat!-char? #\~ port) 'page-down    #f))
      (else
       (my-debug "terminal: unrecognized input:" ; XXX drain&dump input here
                 " CSI " (cons (represent-char-safely char)
                               (debug-drain port)))
       unrecognized-input) )))
(define (convert-escape-sequence port)
  (let ((char (read-char port)))
    (case char
      ((#\[) (convert-csi-input port))
      (else
       (my-debug "terminal: unrecognized input:" ; XXX drain&dump input here
                 " ESC-SEQ " (cons (represent-char-safely char)
                                   (debug-drain port)))
          unrecognized-input) )))
(define (dispatch-on-escape port)
  (if (not (char-ready? port))
      #\esc ; user typed escape goes to event handler
      (convert-escape-sequence port) ))

(define (start-event-loop terminal event-handler)
 (let ((port (get-port terminal)))
  (let event-loop ((char (read-char port)))
    ;(my-debug (char->byte char))
    (and (event-handler (cond
           ((eof-object? char) 'end-of-file)
           ((is-unicode-c0? (char->byte char))
              (case char
                ((#\esc) (dispatch-on-escape port))
                ((#\del) #\backspace) ; delete / backspace
                ((#\cr)  'enter) ; XXX check numpad enter input code
                (else    char)) )
           (else #f)))
    (event-loop (read-char port)) )) ))


;;
;; Key Event Protocol 
;;
;; event-handler <key event>
;;
;; A key event is either
;;  - false (#f) for unrecognized and malformed input,
;;  - a character in Unicode C0 including controls or
;;  - a symbol corresponding to the respective keys or 'end-of-file
;;    '(enter up down right left home end insert-here
;;      end-of-file ;page-up page-down)
;;
;; Event Loop Control: if the handler returns falsity the loop is
;; left.
