(define (vector-index vec val) (list-index (vector->list vec) val))

(define (vector-project! v i elts) 
 (let ((end (+ i (vector-length elts)) )) 
  (when (not (<= 0 i (- end 1) (- (vector-length v) 1)))
    (error "vector-project!: out of range")) 
  (let loop ((i i) (idx 0))
   (cond ((< i end) (vector-set! v i (vector-ref elts idx))
                    (loop (+ i 1) (+ idx 1)) ))) 
  v)) 

(define (vector-append . vectors)
 (let* ((lengths (map vector-length vectors)) 
        (length  (apply + lengths))
        (res     (make-vector length))
        (rdx     0))
   (for-each (lambda (v l) (let loop ((i 0))
              (cond ((< i l) (vector-set! res (+ rdx i) (vector-ref v i)) 
                             (loop (+ i 1)))
              (else          (set! rdx (+ rdx i)) )) )) 
    vectors lengths) 
   res)) 

(define (vector-select v start length)
 (let ((res (make-vector length)))
  (let loop ((idx 0))
   (cond ((< idx length) (vector-set! res idx (vector-ref v (+ start idx)))
                         (loop (+ idx 1)) )))
  res)) 

(define (vector-for-each closure . vectors)
 (when (not (null? vectors))
  (let ((length (vector-length (car vectors))))
   (let loop ((idx 0))
    (cond ((< idx length) (apply closure (map (lambda (v) (vector-ref v idx))
                                              vectors))
                          (loop (+ idx 1)) )) )) )) 

(define (vector-map closure . vectors)
 (let* ((length (if (null? vectors) 0 (vector-length (car vectors)) )) 
        (res    (make-vector length)))
  (let loop ((idx 0))                                                                     
   (cond ((< idx length) (vector-set! res idx
                           (apply closure (map (lambda (v) (vector-ref v idx))
                                               vectors))) 
                         (loop (+ idx 1)) )))
  res))

(when #f 
(sys-load "VSI-core/repl.scm")
(repl-transcript
(define v1 #(a b c)) 
(define v2 #(def))
(vector-append v1 v2) 
(vector-append #() #()) 
(vector-append) 
(vector-select #(1 2 3 4) 2 2) 
(vector-for-each (lambda (e1 e2) (display e1) (display e2) (newline)) v1 v1) 
(vector-map symbol->brass v1) 
(vector-map symbol->brass) 
(vector-project! #(1 2 3 4) 2 #(5 6)) 
)
) 
   
  
