;;
;; Run Time System Statistics Support
;;

(define sysinfo:object-counter-labels
     '(cell-segments double-cell-segments
       pairs vectors brass variables symbols
       closures byte-strings byte-string-mappings ports
       real-numbers complex-numbers big-numbers fractions
       smobs))

;; The SMOB counters need to be the last element in the list.
(define sysinfo:smob-index (- (length sysinfo:object-counter-labels)
                              1))

(define (global-object-count->alist)
  (receive (%global-object-counts)
    ( . counts)
   (map cons sysinfo:object-counter-labels
             counts)))

(define sysinfo:call-counter-labels
        '(
          cell-requests double-cell-requests reinsertions drops
          green blu red violet yellow
          non-phalanx objprop procprop debug
          srcprop dynwind hashtab ceval
         ))

(define (global-call-count->alist)
  (receive (%global-call-counters)
    ( . counts)
   (map cons sysinfo:call-counter-labels
             counts)))

(define sysinfo:last-smob-labels #f) ;; XXX 
(define (global-object-count->vector)
 (let* ((counts (%global-object-counts))
        (smob-location (last-pair counts))
        (smobs         (car smob-location))
        (smob-labels   (map car smobs))
        (smob-values   (map cdr smobs)))
   (set! sysinfo:last-smob-labels smob-labels)
   (set-car! smob-location (car smob-values))
   (set-cdr! smob-location (cdr smob-values))
   (list->vector counts)))

(define (global-call-count->vector)
  (receive (%global-call-counters)
    ( . counts)
   (list->vector counts)))

;; ??? map minus would have done the job
(define sysinfo:last-object-count #(0))
(define (global-object-delta)
 (let* ((current (global-object-count->vector))
        (delta   (vector-minus current sysinfo:last-object-count)))
   (set! sysinfo:last-object-count current)
   (map cons (append (drop-right 1 sysinfo:object-counter-labels)
                     sysinfo:last-smob-labels)
             (vector->list delta))))

(define sysinfo:last-call-count #(0))
(define (global-call-delta)
 (let* ((current (global-call-count->vector))
        (delta   (vector-minus current sysinfo:last-call-count)))
   (set! sysinfo:last-call-count current)
   (map cons sysinfo:call-counter-labels
             (vector->list delta))))


;; PHALANX   PRELIMINARY statistical support
(define (ph1) ; XXX
  (filter (lambda (e) (not (null? (fifth e)))) ; non empty saturation map
          (vector->list (trim-phalanx-maps))))

(define (ph2) ; XXX
  (filter (lambda (e) (not (= (third e) 0))) ; number of requests neq zero 
          (vector->list (trim-phalanx-maps))))


;; END OF THE STATISTICAL ***

(define (sys-report)
 (list (cons 'sys-debug-interpreter sys-debug-interpreter)
       (cons 'sys-send-bytes:incomplete-writes sys-send-bytes:incomplete-writes)
       (cons 'trim:drops (assq-ref (global-call-count->alist) 'drops))
       ;; TODO minimum and average segment saturation someone?
       ;;      ^^^^^^^ should be number of (very) low saturation segments
))
