;; NB This file is also used directly by XI. 
;; 
;; TODO ??? Shouldn't the append position signal trailing text by displaying
;; the respective character? Currently we're one off here: (has this always
;; been like that or is that due to the new display model?) The append
;; position appears after len characters and needs an additional position.
;; The diagram suggests this is wanted and has always been like that. The
;; current viper initialization routine doesn't compensate for that ...

;; ITL: Input Text Line               (Viper Developments)
;; abc UE user experience
;;                         ITL Layout  (tested & approved)
;;
;; left border |--------col
;;                       |
;; display line length   |----------------------len-1
;;                       |                       |
;; display offset dof----|                       |
;;                 +     |                       |
;;                dof   col                     len-1
;;                 +-----|-----------------------|
;;  visible:       +     ext in a line some text i0
;;                 +-----|---------^-------------|0
;;                 +     |---------^             |0
;;   cursor:       +     |        cpos           |0 append position
;;                 +     |         ^             |        0
;; buffered:      (some text in a line some text in a line0)
;;            head ^               ^                      ^ tail
;;                                 ^ current
;;
;; The final null element (possible input future continuation location)
;; has been chosen to be a #\newline character. Have fun!
;;
;;\ Replace /dof/ with /doe/: that's not as simple ...:
;;       - the implicit cursor of dlists conflicts with another pointer
;;         onto an element -> /current/ should be replaced with /length/
;;         in D-Lists and the accessors should take two arguments: list
;;         handle and element??? Or we just access the slots of the
;;         Display Offset Element directly? Or there will be accessors
;;         for D-List elements?
;;       - the end key function needs to access the last elements of a
;;         D-List in a fast way: or change /dlist-ref/ to allow for
;;         negative offsets into the end of the list (with -1 being the
;;         the index of the last element) or provide a specialized
;;         accessor
;;       - maybe /doe/ isn't worth it - it should have replaced /dlist-ref/
;;         which slows down the widget for very long lines at the end
;;         of the line
;;\ Insert a maximum buffer length slot as well as a buffer length slot.
;; DONE Try to reduce /dof/ (/doe/) where possible so the text sticks to
;;      to the right margin no matter what when it's longer than the
;;      display length of the ITL. No matter what: only if there isn't
;;      anymore text hidden to the right. (blen-dof)<len
;;       - add buffer length, please. Is that with or without the zero
;;         element? [w/o]

(sys-load "VOS/vos.scm")
(sys-load "Viper/vchars.scm")


(define itl-class (class (extension viper-object-class)
 (slots terminal row col len buf buf-len dof cpos sdraw)

 (method initialize (row column display-length) ; TODO change to x,y coords
;; Extending the call routine to allow for multiple messages
;; was out of reach on the first day but is considered possible
;; not only on the current syntactical level.
   (this (set! 'row    row))
   (this (set! 'col column))
   (this (set! 'len display-length))
   (this (set! 'sdraw    0))
   (this (set! 'terminal sys-terminal))
   (this (set! 'event-listeners  '()))
   (this (set! 'buf #f))
   (this (set-string ""))
 )

 (method resize (x y length)
  (this (set! 'col x))
  (this (set! 'row y))
  (this (set! 'len length))
  (when (>= (this cpos) length)
   (this (set! 'dof (+ (this dof) 1 (- (this cpos) length)) ))
   (this (set! 'cpos (- length 1))))
 )

 (method set-string (string)
  (when (this buf) (dlist-destroy! (this buf)))
  (let* ((buf (dlist-from-values (string->list string))))
   (dlist-tail buf) (dlist-add! buf #\newline)
   (dlist-head buf)
   (this (set! 'buf    buf))
   (this (set! 'buf-len (string-length string)))
   (this (set! 'dof      0))
   (this (set! 'cpos     0))
 ))

 (method destroy ()
   (dlist-destroy! (this buf))
   (this (set! 'buf 'released)))


 (method cursor-home ()
   ((this terminal) (position-cursor (this col) (this row))))

 (method draw ()
   (let ((terminal (this terminal)))
     (if (char-ready? (terminal port))
       (this (set! 'sdraw (+ 1 (this sdraw))))
       (this (force-draw)))))

 (method force-draw () ;; TODO split this into the terminal class
   (let* ((terminal (this terminal))
          ;; (iobuf (cons #t '() ))
          (buf   (this buf))
          (ocur  (slot-ref buf 'current))
          (w     (this len))
          (atoms (values-from-dlist buf))
          (dof   (this dof))
          (iobuf (cons #t
                  (list-head w
                   (list-tail dof
                    (apply append (map display-form atoms))))))
          (dlen  (- (length iobuf) 1))
          (opos  (this cpos)))
    ;; (begin (terminal (cursor-home)) (terminal (erase-line))
    ;;        (display (format "dof ~A dlen ~A w ~A" dof dlen w)))
    (terminal (hide-cursor)) (this (cursor-home))
    (when (<= dlen w)
     (set-cdr! (last-pair iobuf) (make-list (- w dlen) #\space)))
    (terminal (play-char-list (cdr iobuf)))
    (slot-set! buf 'current ocur)
    (terminal (position-cursor (+ (this col) opos)
                               (this row)))
    (terminal (show-cursor))))

 (method append-pos? () (dlist-end? (this buf)))
 (method empty-buf?  ()
    (eq? (dlist-head (this buf))
         (dlist-tail (this buf)) ))

 (method buf-len () (this buf-len))
    ;; (- (length (values-from-dlist (this buf))) 1))

 ;(method extends-right? () (<= (+ (this dof) (- (this len) 1)) (this buf-len)))
 ;(method extends-left?  () (not (eq? (this dle) (slot-ref (slot-ref (this buf)
 ;                                                                   'tail)
 ;                                                         'prev)))
 ;(method extends-left?  () (not (= (this dof) 0)))
 ;(method extends-left?  () (not (eq? (this doe) (slot-ref (this buf) 'head))))

 Some comment on literate programming here? (Don't put things into parenthesis
 all the time!)

 (slots event-listeners)
 (method add-event-listener (listener)
   ; has-method? ...
   (this (set! 'event-listeners (cons listener (this event-listeners)))))

 (method create-event-handler ()
  (let ((terminal (this terminal)))
  (lambda (key-event)
     ;; !!! Specify a return value! See event loop protocol. !!!
    (case key-event
      ;; Perhaps some ASCII diagram parsing code generator
      ;; would be more readable! ;-)
      ((enter) (let* ((string (list->string (values-from-dlist (this buf))))
                      ; now trim the trailing newline
                      (brass  (subrass string 0 (- (brass-length string) 1))))
                 (for-each (lambda (listener) (listener (input-ready brass)))
                           (this event-listeners)))
               #t)

      ((home) (this (set! 'dof  0))
              (this (set! 'cpos 0))
              (let ((buf (this buf)))
               ;; (if (<= (- (this len) 1) (this buf-len))
               ;;   (this (set! 'doe (dlist-ref buf (- (this len) 1))))
               ;;   (this (set! 'doe (slot-ref (slot-ref buf 'tail) 'prev))))
               (slot-set! buf 'current (slot-ref buf 'head)))
              (this (draw))
              #t)
      ((end)  (let ((bl  (this (buf-len)))
                    (dll (buffer-display-length (this buf)))
                    (w   (this len)))
               (cond
                ((< dll w)  (this (set! 'dof 0))
                            (this (set! 'cpos dll)))
                (else       (this (set! 'dof (- dll w)))
                            (this (set! 'cpos w))) ))
              ; (log-debug 'bl bl 'dll dll 'w w 'cpos (this cpos) 'dof (this dof)))
              (let ((buf (this buf)))
               (slot-set! buf 'current (slot-ref buf 'tail)))
              (this (draw))
              #t)

      ((left) (let* ((cpos (this cpos))
                     (dof  (this dof ))
                     (prev (dlist-prev (this buf)))
                     (dl   (if (dnull? prev)
                               0
                               (display-length prev))))
                ;(log-debug 'cpos cpos 'dof dof (cons 'prev prev) 'dl dl)
                (cond
                 ((>= cpos dl)          (this (set! 'cpos (- cpos dl))))
                 ((< 0 cpos dl)         (this (set! 'dof (- dof (- dl cpos))))
                                        (this (set! 'cpos 0)))
                 ((and (= cpos 0)
                       (not (= dof 0))) (this (set! 'dof  (max 0 (- dof dl)))))

                 (else                  (terminal (bell)))))
              (this (draw))
              #t)
      ((right) (let* ((cpos (this cpos))
                      (w    (this len))
                      (dof  (this dof))
                      (cur  (slot-ref (slot-ref (this buf) 'current) 'value))
                      (dl   (display-length cur))
                      (tail (dlist-end? (this buf))))
                ; (log-debug 'cpos cpos 'dof dof 'w w 'dl dl)
                ; XXX tail would be append position but we can't reach there
                ;     visibly. Same for end of line.
                (dlist-next (this buf))
                (cond
                ((and (not tail) (<= (+ dl cpos) w))(this (set! 'cpos (+ cpos dl))))
                ((and (not tail) (= cpos w))        (this (set! 'dof (+ dof dl))))
                ((and (not tail) (> (+ dl cpos) w)) (this (set! 'cpos w))
                                 (this (set! 'dof (+ dof (- dl (- w cpos))) )))
                (else (terminal (bell)) ))
               (this (draw)))
              #t)

      ((up)
       (let* ((dim    ((this terminal) (get-dimensions)))
              (height (third dim))
              (row    (this row)))
         (when (> row 1)
           (this            (set! 'row (- row 1)))
           ((this terminal) (erase-line))
           (this            (draw)))
         #t))
      ((down)
       (let* ((dim    ((this terminal) (get-dimensions)))
              (height (third dim))
              (row    (this row)))
         (when (< row height)
           (this            (set! 'row (+ row 1)))
           ((this terminal) (erase-line))
           (this            (draw)))
         #t))

      ((page-up)
       (let* ((dim    ((this terminal) (get-dimensions)))
              (width  (second dim))
              (col    (this col)))
         (when (< (+ col (this len)) width)
           (this            (set! 'col (+ col 1)))
           ((this terminal) (erase-line))
           (this            (draw)))
         #t))
      ((page-down)
       (let* ((dim    ((this terminal) (get-dimensions)))
              (width  (second dim))
              (col    (this col)))
         (when (> col 1)
           (this            (set! 'col (- col 1)))
           ((this terminal) (erase-line))
           (this            (draw)))
         #t))
      ((#\del) (cond
               ((not (dlist-end? (this buf))) (dlist-remove! (this buf)))
               (else                          (terminal (bell)) ))
              (this (draw)))
      ((#\backspace)
              (let* ((cpos (this cpos))                                              
                     (dof  (this dof ))                                              
                     (prev (dlist-prev (this buf)))                                  
                     (dl   (if (dnull? prev)                                         
                               0                                                     
                               (display-length prev))))                              
                ;(log-debug 'cpos cpos 'dof dof (cons 'prev prev) 'dl dl)            
               (if (dnull? prev)
                (begin (terminal (bell)))
                (begin (dlist-remove! (this buf))
                (cond                                             ; just like /left/                                                               
                 ((>= cpos dl)          (this (set! 'cpos (- cpos dl))))             
                 ((< 0 cpos dl)         (this (set! 'dof (- dof (- dl cpos))))       
                                        (this (set! 'cpos 0)))                       
                 ((and (= cpos 0)                                                    
                       (not (= dof 0))) (this (set! 'dof  (max 0 (- dof dl)))))      
                                                                                     
                 (else                  (terminal (bell)) ))) ))                         
              (this (draw)) 
       #t)

      (else
       ;; (when (and wingding-mode (< (char->integer key-event) 128))
       ;;  ;(set! key-event (integer->char (+ (char->integer key-event) #x25a1))))
       ;;  (set! key-event (integer->char (+ (char->integer key-event) #x263a))))
       (cond
        ((and (char? key-event) (or (printable? key-event) (space? key-event)))
         (let ((dl (display-length key-event)))
          (dlist-insert! (this buf) key-event)
          (this (set! 'buf-len (+ (this buf-len) 1))) ; atom length
          (dlist-next    (this buf))
          (if (> (+ dl (this cpos)) (this len))
           (begin (this (set! 'dof  (+ (- dl (- (this len) (this cpos)))
                                       (this dof))))
                  (this (set! 'cpos (this len))))
           (this (set! 'cpos (+ dl (this cpos)))))
         (this (draw))
         #t))
        (else #t))
       #t
  ))) ))

 (method attach ()
   (this         (force-draw))
   (sys-terminal (set-event-handler (this (create-event-handler)) )) )

;; This should be kept as state of the ITL.
 (define (buffer-display-length buf)
  (let ((chars (values-from-dlist buf)))
   (apply + (map display-length chars))))

)) ; closes define itl class

(when #f
(sys-load "VSI-core/terminal.scm")
(sys-load "VSI-core/repl.scm")
(define wingding-mode #f)
(define dbg-label (make-object itl-class 1 1 50))
(define (log-debug . args)
 (when #f (dbg-label (set-string (format "~A" args)))
          (dbg-label (force-draw)))
)

(repl-transcript
(repl-clear)
 ; UE TODO some maximum buffer length
(define itl (make-object itl-class 4 10 20))
(begin (sys-terminal (tty-raw-mode))
       (itl (attach))
       (sys-control (roll)))
)
(repl-transcript
(sys-terminal (tty-cooked-mode))
(define b (itl buf))
(list->string (values-from-dlist b))
(itl sdraw)
;; (itl (show))
;; (define x (make-object itl-class 15 15 39))
;; (x sdraw)
;; (x (show))
)
(REPL)
)


;; TRIM/RC VO Stress Test
;; ~ 8.7 sec for 10k (no memoization)
;; ~ 7.6 sec for 10k (with memoization)
(when #f
(repl-transcript
(let loop ((i (* 10 1k)))
  ((make-object itl-class i 2 3) (destroy))
  (when (> i 0) (loop (- i 1))))
)
(REPL)
)




