;;; The Viper TEXTISSUE surface form patterns
;;; (Strings & Brass) 

;;; TODO
;;; - every mode should be a class

(define viper-class (class (extension viper-object-class)
 (slots command-line current-buffer)
 (slots mode)
 (slots pundo)

(method initialize ()
 (let* ((dim    (sys-terminal (get-dimensions)))
        (width  (second dim))
        (height (third  dim)))

  (this (set! 'pundo #f))

  (this (set! 'command-line
    (make-object itl-class 1 2 (- width 2)))) ; trailing append pos

  (this (set! 'current-buffer
    (make-object vtb-class sys-terminal 1 2 width (- height 1))))

  (this (set-mode (this (command-mode))))

  ((this command-line) (add-event-listener this))
  (sys-terminal (add-resize-listener this))
))

(method terminal-resize (detected x y)
 ;; Some sort of focus on text position might meat user expectations better.
  ((this command-line)   (resize 2 1 (- x 1)))
  ((this current-buffer) (resize 1 2 x (- y 1)))
  (this (draw))
)

(method start (arguments)
 (sys-terminal (tty-raw-mode))
 (sys-terminal (erase-display))
 (Viper (attach))
 (Viper (force-draw))
 (let ((arguments (cdr arguments)))
  (when (not (null? arguments))
    (Viper (load-file (first arguments)))
    (Viper (force-draw)) ))
 (sys-control (roll))
)


(method quit ()
 ;; TODO release all and deattach
 (sys-terminal (tty-cooked-mode))
 (sys-terminal (erase-display))
 (sys-terminal (cursor-home))
 (primitive-exit 0)
)


(method draw ()
 ;; TODO check for pending resize events
 (this (force-draw)))

(method force-draw ()
 ; TODO Clear c1r1 please
 ((this command-line)   (force-draw)) ; XXX this repositions the cursor
 ((this current-buffer) (force-draw)))


(method set-mode (mode)
 ;; This mode system looks like two snakes which ...
  ((this command-line) (set-string ""))
  ; ((this command-line) (draw)) ; XXX this repositions the cursor
  (this (set! 'mode mode)))

(method debug ()
  (sys-terminal (tty-cooked-mode))
  (REPL-HERE)               ; XXX debugging
  (sys-terminal (tty-raw-mode))
  (this (force-draw)))


(method command-mode ()
 (let ((vtb (this current-buffer)))
   ; (vtb (cursor-home))
   (vtb (force-draw))
 (lambda (key-event)
  (this (assert-buffer))
  (case key-event
   ((#\bel)       (this (debug)))
   ;((#\backspace) (vtb (backspace)))
   ;((enter)       (vtb (newline)))
   ;((#\esc) (sys-control (remove-read-handler (sys-terminal port))))

   ((#\! #\^ #\0) (vtb (start-of-line)))
   ((#\$)         (vtb (end-of-line)))
   ((page-up)     (vtb (page-up)))
   ((page-down)   (vtb (page-down)))
   ((down  #\j)   (vtb (down)))
   ((up    #\k)   (vtb (up)))
   ((right #\l)   (vtb (right)))
   ((left  #\h)   (vtb (left)))

   ((enter)       (vtb (down))
                  (vtb (start-of-line)))

   ((#\:)         (this (set-mode (this (cmd-line-mode)))))
   ((#\i)         (this (set-mode (this (input-mode)))))
   ((#\a)         (vtb (right))
                  (this (set-mode (this (input-mode)))))

   ((#\x)         (vtb (delete)))
   
   ((#\J)         (vtb (end-of-line))
                  ; We should insert a space if something is not empty but
                  ; maybe /join-lines/ could know how to do that.
                  (vtb (join-lines)))

   ((#\I)       ; insert at start of line
                  (vtb (start-of-line))
                  (this (set-mode (this (input-mode)))))
   ((#\A)       ; append at end of line
                  (vtb (end-of-line))
                  (this (set-mode (this (input-mode)))))

   ((#\O)       ; insert line above
                  (vtb (start-of-line))
                  (vtb (newline))
                  (vtb (up))
                  (this (set-mode (this (input-mode)))))
   ((#\o)       ; insert line below
                  (vtb (end-of-line))
                  (vtb (newline))
                  (this (set-mode (this (input-mode)))))

   ((#\page)      (this (force-draw))) ; CTRL+L

 ;; XXX development of the infamous UNDO
   ;; that full copy is cozy but costly
   ((#\s)         (when (this pundo) ((this pundo) (destroy)))
                  (this (set! 'pundo ((this current-buffer) (copy)))))
   ((#\u)         (when (this pundo)
                   (let ((x (this pundo)))
                     (this (set! 'pundo (this current-buffer)))
                     (this (set! 'current-buffer x))
                     (set! vtb x) )))

   (else          (sys-terminal (bell)) ))

  (vtb (draw))
  #t))
)



(method cmd-line-mode ()
 ((this command-line) (force-draw))
 (let ((handler ((this command-line) (create-event-handler))))
  (lambda (key-event)
   (case key-event
    ((#\esc) (this (set-mode (this (command-mode)))))
    (else    (handler key-event))))))


(method assert-buffer ()
 (when #f ; TODO wire this to some debugging switch please
 (let* ((lbuf  ((this current-buffer) lines))
        (lines (values-from-dlist lbuf))
        (n     (length lines))
        (d     (delete-duplicates lines eq?)))
  (when (not (= n (length d)))
   (error "broken VTB contains duplicate line objects"))
  (when (not (equal? "\n" ((slot-ref (slot-ref lbuf 'tail) 'value)
                             atom-sequence)))
   (error "broken VTB misses its null element")) )) )
     
(method input-mode ()
 (let ((vtb (this current-buffer)))
   ;(vtb (cursor-home))
   (vtb (force-draw))
 (lambda (key-event)
  (this (assert-buffer))
  (case key-event
   ((#f)          (sys-terminal (bell)))
   ((#\bel)       (REPL-HERE)) ; XXX debugging
   ((#\backspace) (vtb (backspace)))
   ((#\del)       (vtb (delete)))
   ((enter)       (vtb (newline)))
   ((page-up)     (vtb (page-up)))
   ((page-down)   (vtb (page-down)))
   ((down)        (vtb (down)))
   ((up)          (vtb (up)))
   ((right)       (vtb (right)))
   ((left)        (vtb (left)))
   ((#\esc)       (this (set-mode (this (command-mode)))))
   (else          (if (char? key-event) (vtb (insert key-event))
                                        (sys-terminal (bell))) ))
  (vtb (draw))
  #t))
)


(method attach ()
 (sys-terminal (set-event-handler (lambda (key) ((Viper mode) key)))) 
)


(method load-file (path)
 (trim)
 ; we really forget about the name of the file here XXX
 ((this current-buffer) (set-lines (load-file path)))
 (trim))

;(method create-cmd-listener ()

; (lambda (brass)

(method input-ready (brass) ; command line listener interface XXX
 (and (viper-oops  ;; and cludge to see the error message before ...
  (cond
   ((= (brass-length brass) 0) '...)
   (else (let ((cmd (brass-ref brass 0)))
    (case cmd
     ((#\e) (this (load-file (subrass brass 1 (brass-length brass)))))
     ((#\w) (this (save-file (subrass brass 1 (brass-length brass)))))
     ((#\q) (this (quit)))
     (else  '...) )) ));)
  (this (set-mode (this (command-mode)))))))


(method save-file (path)
 (trim)
 (with-output-to-file path (lambda ()
  (let ((lines (values-from-dlist ((this current-buffer) lines))))
   (for-each display (map (lambda (tl) (tl atom-sequence))
                          (list-head (- (length lines) 1) lines )) )) ))
 (trim))

(define (load-file path)
 (let ((lbuf  (cons #f '()))
       (lines (cons #f '())))
  (with-input-from-file path (lambda ()
   (let loop ((char (read-char)) (ll lines) (l lbuf))
    (cond
    ((eof-object? char)   (when (not (null? (cdr lbuf)))
                            ;; ??? Does this handle a missing newline correctly?
                            (set-cdr! ll (cons (list->text-line (cdr lbuf)) '()))
                            (set! ll (cdr ll)))
                          (set-cdr! ll (list (list->text-line
                                              (list #\newline))))

                          (dlist-from-values (cdr lines)))

    ((eq? char #\newline) (set-cdr! l (cons #\newline '()))
                          (set-cdr! ll (cons (list->text-line (cdr lbuf)) '()))
                          (set! lbuf (cons #f '()))
                          (loop (read-char) (cdr ll) lbuf))

    (else                 (set-cdr! l (cons char '()))
                          (loop (read-char) ll (cdr l))) ))) )))



(define viper-oops (macro expressions
 `(catch #t (lambda () (begin ,@expressions) #t)
   (lambda (key . args)
    ((this command-line) (set-string (format "voops: ~A  ~A" key args)))
    ((this command-line) (force-draw))
    #f))))

)) ; closes viper class









