; abc
;  There is no sense if not aesthetical sense.
;
;      Viper Text Buffer Class
;
(sys-load "VOS/vos.scm")
(sys-load "VSI-core/structures.scm")
(sys-load "Viper/text-line.class.scm")

(define vtb-class (class (extension viper-object-class)
 (slots terminal row col width height lines)
 (slots dof   ; Display OFfset
        crow  ; Current ROW    (zero based offset)
        ccol  ; Current COLumn (zero based offset)
        fl)   ; dlist *element* of the First Line on screen
 (slots sdraw)

 (method initialize (terminal col row width height)
  (this (set! 'terminal terminal)) ;XXX
  (this (set! 'row      row))
  (this (set! 'col      col))
  (this (set! 'height   height))
  (this (set! 'width    width))
  (this (set! 'sdraw    0))
  (let ((etl (make-object text-line-class (list #\newline))))
    (this (set-lines (dlist-from-values (list etl)))))
 )

(method resize (col row width height)
  (this (set! 'row      row))
  (this (set! 'col      col))
  (this (set! 'height   height))
  (this (set! 'width    width))
  (when (>= (this ccol) (- width 1))
    (this (set! 'dof  (+ (this dof) 1 (- (this ccol) width))))
    (this (set! 'ccol (- width 1)))
  )
  (when (> (this crow) (- height 1))
    (let loop ((le (this fl)) (dif (- (this crow) (- height 1))))
      (cond
      ((or (dnull? (slot-ref le 'next))
           (= dif 0))   (this (set! 'fl le)))

      (else             (loop (slot-ref le 'next) (- dif 1))) ))
    (this (set! 'crow (+ 1 (- height row))))
  )
)    

(method copy ()
 (let ((cpy (make-object vtb-class (this terminal)
                                   (this col) (this row)
                                   (this width) (this height))))
  (cpy (set-lines  (copy-lines (this lines))))
 ;; XXX pay attention to dlist state and fl ref !!!
 ;; we revert to the initial buffer start position for now
  ;; (cpy (set! 'dof  (this dof)))
  ;; (cpy (set! 'crow (this crow)))
  ;; (cpy (set! 'ccol (this ccol)))

  ;; (cpy (set! 'fl   (slot-ref (cpy lines) 'head))) ;; XXX developing exploration
  ;; ??? maybe it would have been better to adapt crow and ...

  (cpy (set! 'sdraw (this sdraw)))
  cpy))

(define (copy-lines lines)
 (dlist-from-values (map (lambda (e) (e (copy))) (values-from-dlist lines))))

(method destroy ()
 (when (dlist? (this lines))
   (for-each (lambda (e) (e (destroy))) (values-from-dlist (this lines)))
   (dlist-destroy! (this lines))))

(method set-lines (lines)
  (when (dlist? (this lines))
    (for-each (lambda (e) (e (destroy))) (values-from-dlist (this lines)))
    (dlist-destroy! (this lines)))
  (this (set! 'lines lines))
  (this (set! 'fl    (slot-ref (this lines) 'head)))
  (dlist-head (this lines))
  (this (set! 'dof   0))
  (this (set! 'crow  0))
  (this (set! 'ccol  0)))

(method cursor-home ()
 (slot-set! (this lines) 'current (this fl))
 ((this terminal) (position-cursor (this col) (this row))))

(method draw ()
 (let ((terminal (this terminal)))
  (if (char-ready? (terminal port))
    (this (set! 'sdraw (+ 1 (this sdraw))))
    (this (force-draw)))))

(method force-draw ()
 (let* ((cr       0)
        (cc       0)
        (dof      (this dof))
        (iobuf    (cons #f '()))
        (terminal (this terminal))
        (lbuf     (this lines))
        (w        (this width))
        (h        (this height))
        (fl       (this fl))
        (ocur     (slot-ref lbuf 'current)))
  (terminal (hide-cursor))
  (this (cursor-home))
  (let loop ((le fl) (last iobuf))
   (when (not (= cr h))
    (set-cdr! last (terminal (seq-position-cursor (this col)
                                                  (+ (this row) cr))))
    (set! last (last-pair last))
    (cond
     ((dnull? le)
             (set-cdr! last (make-list w #\space))
             (set! cr (+ cr 1))
             (loop dnull (last-pair last)))
     (else (let* ((l  (slot-ref le 'value))
                  (sn (max 0 (- (l (display-length)) dof))))
             (when (> sn w) (set! sn w))
             (when (> sn 0)
               (set-cdr! last (string->list (l (display-seq dof (+ dof sn)))))
               (set! last (last-pair last)))
             (when (< sn w)
               (set-cdr! last (make-list (- w sn) #\space))
               (set! last (last-pair last)))
             (set! cr (+ cr 1)))
           (loop (slot-ref le 'next) last))) ))
  (terminal (play-char-list (cdr iobuf)))

  (slot-set! (this lines) 'current ocur)
  (terminal (position-cursor (+ (this col) (this ccol))
                             (+ (this row) (this crow))))
  (terminal (show-cursor))
))

(define (adjust-col! this)
 (let* ((curl (slot-ref (slot-ref (this lines) 'current) 'value))
        (ccol (this ccol))
        (dof  (this dof))
        (w    (this width))
        (ndc  (curl (adjust-dpos (+ dof ccol)))))
  (cond
  ((<= ndc dof)      (this (set! 'ccol 0))
                     (this (set! 'dof  ndc)))

  ((>= ndc (+ dof w)) (this (set! 'ccol (- w 1)))
                      (this (set! 'dof  (+ 1 (- ndc w)))))

  (else              (this (set! 'ccol (- ndc dof))) )) ))
;; Editing

(method join-lines ()
(let* ((lbuf (this lines))
       (cure (slot-ref lbuf 'current))
       (curl (slot-ref cure 'value)))
  (cond
  ((or (dlist-end? lbuf) (dnull? (slot-ref cure 'next)))
          (terminal (bell)))

  (else (let ((next (dlist-next lbuf)))
          (curl (append next))
          (dlist-remove! lbuf)
          (dlist-prev lbuf)
))) ))

(method delete ()
 (let* ((terminal (this terminal))
        (ccol (this ccol))
        (dof  (this dof))
        (lbuf (this lines))
        (cure (slot-ref lbuf 'current))
        (curl (slot-ref cure 'value))
        (dlen (curl (display-length))) )
        ;(dl   (curl (display-length (+ dof ccol)))))
  (cond
  ((< (+ ccol dof) dlen) (curl (remove (+ dof ccol))))
  (else                  (terminal (bell)))
)))

(method backspace () ;; XXX this should left + delete
 (let* ((terminal (this terminal))
        (lbuf (this lines))
        (cure (slot-ref lbuf 'current))
        (curl (slot-ref cure 'value))
        (len  (curl (display-length)))
        (ccol (this ccol))
        (dof  (this dof))
        (is-fl (eq? (this fl) cure)))
  (cond
  ((> (+ ccol dof) len)     (this (left)))         ; free space

  ((and (dlist-start? lbuf)                        ; start of buffer
        (= dof  0)
        (= ccol 0))         (terminal (bell)))

  ((and (dlist-end? lbuf)                          ; start of last line
        (= dof  0)
        (= ccol 0))         (let* ((pres (dlist-prev lbuf))
                                   (crow (this crow))
                                   (dlen (pres (display-length)))
      (when is-fl       (this (set! 'fl (slot-ref lbuf 'current))))
      (when (> crow 0)  (this (set! 'crow (- crow 1))))
      (this (set! 'dof  (max 0 (- dlen (this width)))))
      (this (set! 'ccol (- dlen (this dof)))) )))

  ((< 0 ccol (+ 1 (- len dof))) (let ((dl (curl (left-length (+ dof ccol)) )))
                              (curl (remove (- (+ dof ccol) dl))) 
                              (if (< ccol dl)
                                  (begin (this (set! 'dof (- dof (- dl ccol))))
                                         (this (set! 'ccol 0)))
                                  (begin (this (set! 'ccol (- ccol dl))) )) ))

  ((and (= ccol 0)
        (not (= dof 0)))    (let ((dl (curl (left-length (+ dof ccol)) )))
                             (curl (remove (- (+ dof ccol) dl)))
                             (this (set! 'dof (- dof dl))) ))
                                         

  ((and (= ccol 0)                                 ; join lines
        (= dof  0))         (dlist-remove! lbuf)
                            (let* ((pres (dlist-prev lbuf))
                                   (olen (pres (display-length)))
                                   (crow (this crow))
                                   (news (pres (append curl))))
      ; bogus ? (slot-set! (slot-ref lbuf 'current) 'value news)
      (when is-fl       (this (set! 'fl (slot-ref lbuf 'current))))
      (cond ((> crow 0) (this (set! 'crow (- crow 1))))
      )     ;((= crow 0) (this (set! 'fl (slot-ref (this fl) 'prev)))
            ;             (this (set! 'crow (+ crow 1)))))
            ; Scroll cursor action into scope but for now we
            ; stick to the behavior of the ITL.
      (this (set! 'dof  (max 0 (+ 1 (- olen (this width))))))
      (this (set! 'ccol (- olen (this dof))))
      (curl (destroy)) ))
               
  (else                     (terminal (bell)) ))
)) ; closes method backspace

(method newline ()
 (let* ((terminal (this terminal))
        (lbuf  (this lines))
        (cure  (slot-ref lbuf 'current))
        (curl  (slot-ref cure 'value))
        (len   (curl (display-length)))
        (dof   (this dof))
        (parts (if (>= (+ (this ccol) dof) len)      ; free space
                 (values curl (make-object text-line-class (brass-copy "\n")))
                 (curl (split (+ (this dof) (this ccol))) )) )
        (one   (first  parts))
        (two   (second parts))
        (is-fl (eq? (this fl) cure))
        (is-end (dlist-end? lbuf))) ; This predicate doesn't work well after
                                    ; the /remove/ operation. Why? [... leads
                                    ; to strange behavior after the /remove/]
                                    ; It's called: second to last. 
   (dlist-remove! lbuf)
   (cond
   (is-end (dlist-add! lbuf one)
           (dlist-add! lbuf two)
           (dlist-prev lbuf))
   (else   (dlist-insert! lbuf two)
           (dlist-insert! lbuf one)))
   ;; current is on one
   (when is-fl (this (set! 'fl (slot-ref lbuf 'current)) ))
   (begin         ; start of line (carriage return)
     (this (set! 'dof  0))
     (this (set! 'ccol 0))
   )
   (this (down))  ; line feed
)) ; closes method newline

(method down ()
 (let ((terminal (this terminal))
       (crow (this crow))
       (eh   (- (this height) 1)))
  (when (not (dlist-end? (this lines))) ;; XXX missing bell
    (cond
    ((< crow eh) (this (set! 'crow (+ crow 1)))
                 (dlist-next (this lines))) 
    ((= crow eh) (let ((next (slot-ref (this fl) 'next)))
                   (if (not (dnull? next))
                     (begin (this (set! 'fl next))
                            (dlist-next (this lines)))
                     (terminal (bell))))))
    (adjust-col! this)
))) ; closes method down

(method up ()
 (let ((terminal (this terminal))
       (crow (this crow))
       (eh   (- (this height) 1)))
  ;; (when (not (dlist-start? (this lines)))
  (cond
  ((> crow 0)  (this (set! 'crow (- crow 1)))
               (dlist-prev (this lines))) 
  ((= crow 0)  (let ((prev (slot-ref (this fl) 'prev)))
                 (if (not (dnull? prev))
                   (begin (this (set! 'fl prev))
                          (dlist-prev (this lines)))
                   (terminal (bell))))))
  (adjust-col! this)
)) ; closes method up

(method left () ; display-form-length proof
 (let* ((terminal (this terminal))
        (ccol (this ccol))
        (dof  (this dof))
        (w    (this width))
        (cure (slot-ref (this lines) 'current))
        (curl (slot-ref cure 'value)))
  (cond
  ((and (> (+ dof ccol) (curl (display-length)))) ; free space step
                      (if (= ccol 0) (this (set! 'dof (- dof 1)))
                                     (this (set! 'ccol (- ccol 1)))))

  ((> ccol 0)         (let ((dl (curl (left-length (+ dof ccol)))))
                        (if (< ccol dl)
                            (begin (this (set! 'dof (- dof (- dl ccol)) ))
                                   (this (set! 'ccol 0)))
                            (begin (this (set! 'ccol (- ccol dl)) ))) ))

  ((and (not (dlist-start? (this lines))) ; line before + goto end of line
        (= (this dof) 0)
        (= ccol 0))   (when (eq? (this fl) cure)
                        (this (set! 'fl (slot-ref cure 'prev))))
                      (dlist-prev (this lines))
                      ; the following should equal a goto end of line
                      (let* ((curl (slot-ref (slot-ref (this lines) 'current) 'value))
                             (len  (curl (display-length)))
                             (dof  (this dof))
                             (crow (this crow))
                             (dl   (if (= len 0)
                                     0  ; empty line
                                     (curl (left-length len)))))
                        ; cfr with backspace - functionalize this seq
                        (this (set! 'dof (max 0 (+ 1 (- len (this width))))))
                        (this (set! 'ccol (- len (this dof))))
                        (when (> crow 0) (this (set! 'crow (- crow 1)))) ))

  ((and (> (this dof) 0)) ; non free space dof reduction (cond top down seq)
   (= ccol 0)         (let ((dl (curl (left-length (+ dof ccol)))))
                        (this (set! 'dof (- (this dof) dl)))))

  (else               (terminal (bell))))
))

(method right ()
 (let* ((terminal (this terminal))
        (ccol (this ccol))
        (dof  (this dof))
        (w    (- (this width) 1))
        (curl (slot-ref (slot-ref (this lines) 'current) 'value))
        (dlen (curl (display-length)))
        (tail (>= (+ dof ccol) dlen))
        (dl   (if tail 1 ; free space step width
                       (curl (display-length (+ ccol dof))))))
  (cond
  ((<= (+ dl ccol) w) (this (set! 'ccol (+ ccol dl))))
  ((= ccol w)         (this (set! 'dof (+ dof dl))))
  ((> (+ dl ccol) w)  (this (set! 'ccol w))
                      (this (set! 'dof (+ dof 1 (- dl (- w ccol))) )))
  (else               (terminal (bell)) ))
))

(method start-of-line ()
 (this (set! 'dof 0))
 (this (set! 'ccol 0)))

(method end-of-line ()
 (let* ((curl (slot-ref (slot-ref (this lines) 'current) 'value))
        (dlen (curl (display-length))))
   (this (set! 'dof (max 0 (+ 1 (- dlen (this width)) )) ))
   (this (set! 'ccol (- dlen (this dof)) ))
))

(method page-up ()
  (slot-set! (this lines) 'current (this fl))
  (let loop ((le (this fl)) (h (this height)))
    (if (not (or (dnull? (slot-ref le 'prev)) (= h 1))) 
      (begin (dlist-prev (this lines))
             (loop (slot-ref le 'prev) (- h 1)))
      (this (set! 'fl le))))
  (this (set! 'crow 0))
  (adjust-col! this)
)

(method page-down ()
  (slot-set! (this lines) 'current (this fl))
  (let loop ((le (this fl)) (h (this height)))
    (if (not (or (dnull? (slot-ref le 'next)) (= h 1))) 
      (begin (dlist-next (this lines))
             (loop (slot-ref le 'next) (- h 1)))
      (this (set! 'fl le))))
  (this (set! 'crow 0))
  (adjust-col! this)
)

(method insert (char) ; or unhandled key event
 (let ((terminal (this terminal)))
  (cond
  ((and (char? char)
        (or (printable? char) (space? char)))
    (let* ((lbuf (this lines))
           (cur  (slot-ref lbuf 'current))
           (curl (slot-ref cur 'value))
           (dlen (curl (display-length)))
           (dof  (this dof))
           (ccol (this ccol))
           (dl   (display-length char)))
      (when (>= ccol (- dlen dof)) ; free white space
        (curl (append-atoms (brass-append
                             (make-brass (+ 1 (- ccol (- dlen dof)))
                                         #\space)
                             (brass-copy "\n") )) ))
      (curl (insert (+ dof ccol) char))
      (when (dlist-end? lbuf)
        (dlist-add! lbuf (make-object text-line-class (list #\newline)))
        (dlist-prev lbuf))
      (cond ((< (+ ccol dl) (this width)) (this (set! 'ccol (+ ccol dl))))
            (else                         (this (set! 'dof (+ (this dof) dl)))))
       ))
  (else
      (terminal (bell)) ))
))

)) ; closes vtb class



;; (define (display-len-tab ccol width)
;;  (define tabulation-step 8)
;;  (define next-tab (- tabulation-step
;;                      (modulo (+ tabulation-step ccol)
;;                              tabulation-step)))
;;  (if (> (+ ccol next-tab) width)
;;    (- width ccol)
;;    next-tab))

(define (display-len-tab ccol width)
 (let* ((tabulation-step 8)
        (next-tab (- tabulation-step
                     (modulo (+ tabulation-step ccol)
                             tabulation-step))))
   (if (> (+ ccol next-tab) width)
     (- width ccol)
     next-tab)))

