;;; Unicode Database
;;; (about 29.000 entries; continuos indices up until #887; max: 1.114.109)
;;; We somehow need some efficient way to "compile" this in. (Or at least
;;; the case mappings. Maybe we just /write/ and /read/ the vectors for now.
;;; But then: didn't they reach 100.000 code points at some point? Maybe
;;; we'll just "grab" the Latin blocks plus the Common blocks. Perhaps the
;;; actual implementation of /brass/ and /characters/ should use the shared
;;; resources if available.

;;
;; Auxiliary I: Character Database (UnicodeData.txt)
(define (uc-data:read-field)
 (let loop ((byte (read-char))
            (token '()))
  (cond
   ((eof-object? byte)   token)
   ((eq? byte #\;)       token)
   ((eq? byte #\newline) (unread-char byte) token)
   (else
    (loop (read-char) (cons byte token))) )))
(define (uc-data:read-entry)
 (let ((entry (make-vector 15 #f))) ; XXX fix reader wrt unspecified serialized
  (let loop ((n 0))
   (let ((datum (reverse! (uc-data:read-field))))
    (vector-set! entry n
     (case n
      ((0) (string->number (list->string (append (list #\# #\x) datum))))
      (else
       (list->string datum)))))
   (cond
    ((eq? (peek-char) #\newline) (read-char) entry)
    ((eof-object? (peek-char))   entry)
    ((< n 14)           (loop (+ n 1)))
    (else                        entry)) )))
  
(define (uc:read-data file-name)
 (let ((answer-box (cons #f '()) ))
  (with-input-from-file file-name
   (lambda ()
    (let loop ((n 0) (byte (peek-char)) (last answer-box))
     (when (= 0 (modulo n 1000)) (display "."))
     (cond
      ((eof-object? byte) (cdr answer-box))
      ((> n 50000)
         (display "warning: data limit reached" (current-error-port))
         (cdr answer-box))
      (else
       (set-cdr! last (cons (uc-data:read-entry) '()))
       (loop (+ n 1) (peek-char) (cdr last))) ))) )))

;; Unicode Data Access
(define (uc-lookup-codepoint db cp)
 (let ((length (vector-length db)))
  (let loop ((n 0))
   (and (< n length)
        (let ((entry (vector-ref db n)))
          (if (= cp (vector-ref entry 0))
            entry
            (loop (+ n 1))) )) )))


;;
;; Auxiliary II: Case Mappings (CaseFolding.txt)
;;
;; NB The standard makes a big fuzz about some german special
;; symbols which simply don't have any upper case correspondence
;; where italian has other orthography rules or conventions
;; especially regarding upper case accented letters. These
;; language specifics don't need to be reflected by an encoding
;; but the focus on some languages in the standards remains XXX-T73
;;
(define (uc-case:skip-comment)
 (let ((byte (read-char)))
  (if (eq? byte #\newline)
    byte
    (uc-case:skip-comment))))
(define (uc-case:read-char)
 (let ((byte (read-char)))
  (cond
   ((eq? byte #\#)
     (uc-case:skip-comment)) ; returns the trailing newline
   ((eq? byte #\space)
     (uc-case:read-char)) ; XXX
   (else
     byte))))

;; Some fields contain more than one code point, i.e. a sequence
;; of code points, where their description reads "multiple characters
;; are seperated by spaces". Only entries of type "S" (where type is
;; called status) are guaranteed to map to one character. That is
;; nice and next to useless as such an entry only "occurs" "where
;; different from F". Let's pronounce it like that: "These years
;; haven't been driven by aristocracy."
;; This implementation reflects the current "standards" in "business".
(define (uc-case:field-split datum) ; XXX that's not worth it
 (let* ((answers (cons #f '() ))
        (last-answer answers))
  (let loop ((ll '()) (last datum))
  (cond
   ((null? last)
      (when (not (null? ll)) (set-cdr! last-answer (cons (reverse! ll) '())))
      (cdr answers))
   ((white-space? (car last))
      (when (not (null? ll)) (set-cdr! last-answer (cons (reverse! ll) '())))
      (loop '() (cdr last)))
   (else
      (loop (cons (car last) last) (cdr last)) ))) ))
(define (uc:list->number datum)
 (string->number (list->string (append (list #\# #\x) datum))))

(define (uc-case:read-field) ; the same uc-data:read-field
 (let loop ((byte (uc-case:read-char))
            (token '()))
  (cond
   ((eof-object? byte)   token)
   ((eq? byte #\;)       token)
   ((eq? byte #\newline) (unread-char byte) token)
   (else
    (loop (uc-case:read-char) (cons byte token))) )))
(define (uc-case:read-entry)
 (let ((entry (make-vector 3 #f))) ; XXX fix reader wrt unspecified serialized
  (let loop ((n 0))
   (let ((datum (reverse! (uc-case:read-field)) ))
      ;(display n) (display ": ") (write datum) (newline)
    (when (< n 3)
     (vector-set! entry n
      (case n
       ((0) (uc:list->number datum))
       ((2) (uc:list->number datum))
       ;((2) (map uc:list->number (uc-case:field-split datum)))
       (else
        (list->string datum))) )))
   (cond
    ((eq? (peek-char) #\newline) (uc-case:read-char) entry)
    ((eof-object? (peek-char))   entry)
    ((< n 3)            (loop (+ n 1)))
    (else                        entry)) )))
  
(define (uc:read-case file-name)
 (let ((answer-box (cons #f '()) ))
  (with-input-from-file file-name
   (lambda ()
    (let loop ((n 0) (byte (peek-char)) (last answer-box))
     ; (when (= 0 (modulo n 1000)) (display "."))
     (cond
      ((eof-object? byte) (cdr answer-box))
      ((> n 50000)
         (display "warning: data limit reached" (current-error-port))
         (cdr answer-box))
      (else
       (set-cdr! last (cons (uc-case:read-entry) '()))
       (loop (+ n 1) (peek-char) (cdr last))) ))) )))

(define (uc-case-mapping mappings cp)
 (let ((answer-box (cons #f '())))
  (let loop ((current mappings) (last answer-box))
   (if (null? current)
     (cdr answer-box)
     (let ((entry (car current)))
       (cond
        ((and (vector-ref entry 0) (= cp (vector-ref entry 0)))
          (set-cdr! last (cons (vector-ref entry 2) '()))
          (loop (cdr current) (cdr last)))
        ((and (vector-ref entry 2)
              (not (list? (vector-ref entry 2)))
              (= cp (vector-ref entry 2)))
          (set-cdr! last (cons (vector-ref entry 0) '() ))
          (loop (cdr current) (cdr last)))
        (else
          (loop (cdr current) last)) )) )) ))


(define (uc-db-init)
 (let ((uc-data
  (if (access? "/opt/VSI/VSI-core/uc.data.scm" R_OK)
   (with-input-from-file "/opt/VSI/VSI-core/uc.data.scm" (lambda () (read)))
   (if (access? "UnicodeData.txt" R_OK)
     (begin (newline) (display "Importing Unicode Character Data ")
            (list->vector (uc:read-data "UnicodeData.txt")))
     (begin (newline) (newline)
      (display "Sorry: Missing Unicode Character Data - see sources please"
               (current-error-port))
      (newline (current-error-port))
      (error "missing data") )))))

 (when (and (not (access? "/opt/VSI/VSI-core/uc.data.scm" F_OK)))
            ;(access? "VSI-core/uc.data.scm" W_OK))
   (newline) (display "Storing Unicode Character Data")
;;
 ; Serializer: use this to create the data file VSI-core/uc.data.scm from
 ; UnicodeData.txt which can be downloaded from www.unicode.org
;;
   (with-output-to-file "/opt/VSI/VSI-core/uc.data.scm" (lambda () (write uc-data))))
 uc-data
))
