;; Test Selection

(define yes #t) (define no #f)
 (set! no #t) ; drive all tests

(define do-level-1   no)
(define do-level-2   no)
(define do-level-3   no)
(define do-level-D   no)
(define do-level-V/a no)
(define do-level-Vb  yes)
(define do-level-B   no)


;; Symbolic Numbers
(define infinity      +inf.0)
(define anti-infinity -inf.0)
(define zero+0         0.0)
(define zero-0        -0.0)

(define symbolic-numbers (list infinity anti-infinity zero+0 zero-0))
(define equality-operations (list eq? eqv? equal?))

(define (test-operator operation operand-1 operand-2)
 (display "test: ")
 (display operation) (display " ")
 (display operand-1) (display " ") (display operand-2)
 (display " result: ")
 (display (operation operand-1 operand-2))
 (newline))

(define (test-controller operation operand-1 test-set)
 (cond
  ((not (null? test-set))
   (test-operator operation operand-1 (car test-set))
   (test-controller operation operand-1 (cdr test-set)))))

(define (test-manager operation test-set)
 (cond
  ((not (null? test-set))
   (test-controller operation (car test-set) (cdr test-set))
   (test-manager operation (cdr test-set))))) 

(define (raise-test-M1 operations test-set)
 (cond
  ((null? operations) (display " FINE ") (newline))
  (else
   (test-manager (car operations) test-set)
   (raise-test-M1 (cdr operations) test-set) )))

(define (level-1)
  (define M1-operations (append (list =) equality-operations))
  (define M1-set        (append (list 0) symbolic-numbers))
  (display "LEVEL 1") (newline)
  (display " M1-operations: ") (display M1-operations) (newline)
  (display "        M1-set: ") (display M1-set)        (newline)
  (raise-test-M1 M1-operations M1-set))

(when do-level-1
(level-1)
)

(define mixed-set-positive
  (list 3 10000000000000000000000000000000000000
        8/13 1.89 122.0+55.3i))

(define primary-operations (list + * / -))
;(define primary-operations (list *))

(define (level-2)
 (define M2-operations primary-operations)
 (define M2-set        (append mixed-set-positive mixed-set-positive))
 (display "LEVEL 2") (newline)
 (raise-test-M1 M2-operations M2-set))

(when do-level-2
(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)

(level-2)

(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)
)


(define M3-operations (list->vector primary-operations))
(define M3-set        (list->vector (append mixed-set-positive
                                            (map - mixed-set-positive))))

;; XXX inner defines still create "evil" closure rings => global helpers
(define (random-ref v) (vector-ref v (random (vector-length v))))
(define (M3-op) (random-ref M3-operations))
(define (M3-num) (random-ref M3-set))

(define silent-M3-op M3-op)
(define silent-M3-num M3-num)

(define (verbose-M3-op)
  (let ((op (random-ref M3-operations)))
    (newline)
    (display op)
    op))

(define (verbose-M3-num)
  (let ((num (random-ref M3-set)))
    (display " # ")
    (display num)
    num))

(define (M3-formulator)
  ((M3-op) ((M3-op) ((M3-op) (M3-num) (M3-num) (M3-num))
                    ((M3-op) (M3-num) (M3-num) (M3-num))
                    ((M3-op) (M3-num) (M3-num) (M3-num)))
           ((M3-op) ((M3-op) (M3-num) (M3-num) (M3-num))
                    ((M3-op) (M3-num) (M3-num) (M3-num))
                    ((M3-op) (M3-num) (M3-num) (M3-num)))
           ((M3-op) ((M3-op) (M3-num) (M3-num) (M3-num))
                    ((M3-op) (M3-num) (M3-num) (M3-num))
                    ((M3-op) (M3-num) (M3-num) (M3-num))) ))
(define 1k 1000)
(define M3-repetitions (* 300 1k))
(define (level-3)
 (display "LEVEL 3 repetitions: ") (display M3-repetitions) (newline)
 (let loop ((n 1))
    (if (<= n M3-repetitions)
      (begin
        (if (= 0 (modulo n (* 10 1k)))  ;; XXX /when/ is missing before core
         (begin (display ".")  ))
        (M3-formulator)
        (loop (+ n 1)) )))
 (newline)
 (display " End of Level 3 ") (newline) )

(when do-level-3
(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)

(level-3)

(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)
)

(define d-one-ary
  (vector round truncate floor ceiling
          $abs $exp $log
          $sin $cos $tan $asin $acos
          $sqrt))

(define d-two-ary
  (vector $expt $atan2
          rationalize))
        
(define d-set-0 (list 0.0 0.1 2.0 2.3))
(define d-set d-set-0)
(set! d-set-0 (list->vector (append d-set-0 (map - d-set-0))))

;(define d-set (list 0.001 0.1 2.0 2.3))
(set! d-set (list->vector (append d-set (map - d-set))))

;; (define old-random-ref random-ref)
;; (define (random-ref v)
;;   (let ((element (old-random-ref v)))
;;     (display element) (newline)
;;     element))

(define (d-formulator)
  ((random-ref d-one-ary)
       ((random-ref d-two-ary) (random-ref d-set) (random-ref d-set))))

(define D-repetitions (* 10 1k 1k))

(define (level-D)
  (display "Level D repetitions: ") (display D-repetitions) (newline)
  (let loop ((n 1))
    (if (<= n D-repetitions)
      (begin (if (= 0 (modulo n (* 100 1k))) (display "."))
             (d-formulator)
             (loop (+ n 1)) )))
  (newline)
  (display "End of Level D")
  (newline))

(when do-level-D
(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)

(level-D)

(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)
)


;;
;; LEVEL V - we need regular initialization to test vector calculus
;;
(define V-set (list->vector
               (append mixed-set-positive (map - mixed-set-positive))))
(define V/a-operations (vector vector-plus vector-minus))
(define V-maximum (* 1k))
(define (random-vector) (make-vector (+ 1 (random V-maximum)) 0))
(define (mixed-vector v n)
  (vector-set! v (- n 1) (random-ref V-set))
  (if (<= n 1)
      v
      (mixed-vector v (- n 1))))

(define V-horizontal (* 10 1k))
(define (vector-working-set)
 (map (lambda (el) (mixed-vector el (vector-length el)))
   (map (lambda (el) (random-vector))
     (make-list (+ 2 (random (- V-horizontal 1)))) )))

(define (vector-work set)
  (apply (random-ref V/a-operations) set))

(define V/a-repetitions (* 1 3))
(define (n-V/a)
  (display "Level V/a - Vector Calculus") (newline)
  (let loop ((n 1))
    (when (= 1 (modulo n (* 2 1)))
      (display ".") (usleep (* 100 1k)))
    (vector-work (vector-working-set))
    (when (< n V/a-repetitions)
      (loop (+ n 1))))
  (newline)(display " V/a end ")(newline))
    
;;
;; Please note this test puts a little stress on memory management.
;; Reported memory usage by the operating system might rise and fall.
;; Stick to the statistics or include the statistics of malloc, too.

(when do-level-V/a
 (sys-load "lib/vector-geometry.scm") ; cross product
;(trim) (trim) (trim)
(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)

;(repeat 5 n-V/a)
(n-V/a)

(display (%global-object-counts)) (newline)
(display (%global-call-counters)) (newline)
;; (n-V/a)

;; (display (%global-object-counts)) (newline)
;; (display (%global-call-counters)) (newline)
)

(define (dim-random) (/ (random 1000) 100))
(define (vb-create-set n l)
 (if (= n 0)
   l
   (begin
    (vb-create-set (- n 1) (cons (vector (dim-random) (dim-random) (dim-random))
                                 l)))))

(define vb-set-size 100000)
(define (vector-exact->inexact! v)
  (let ((length (vector-length v)))
    (let loop ((n 0))
      (when (< n length)
        (vector-set! v n (exact->inexact (vector-ref v n)))
        (loop (+ n 1)) )) ))

(define (level-Vb)
  (define vb-set-1 '())
  (define vb-set-2 '())
  (define vb-res-1 '())
  (define vb-res-2 '())
  (display "level-Vb")
  (set! vb-set-1 (vb-create-set vb-set-size '()))
  (set! vb-set-2 (vb-create-set vb-set-size '()))
  (display " exact cross product")
;(set! vb-res-1 (map (lambda (idem) (pause)) vb-set-2))
;(set! vb-res-1 (map (lambda (idem) 1) vb-set-2))
;(set! vb-res-1 (map (lambda (idem idem2) 1) vb-set-1 vb-set-2))
;(set! vb-res-1 (map (lambda (idem idem2) idem) vb-set-1 vb-set-2))
(set! vb-res-1 (map vector-cross vb-set-1 vb-set-2))
  (display " inexact")
 ; (display vb-set-1)(newline)
  (map vector-exact->inexact! vb-set-1)
  (map vector-exact->inexact! vb-set-2)
  (display " cross")
(set! vb-res-2 (map vector-cross vb-set-1 vb-set-2))
  (display " done.") (newline)
;; (display (global-object-count->alist)) (newline)
;; (display (%global-call-counters)) (newline)
;;   (display "Inner Release:") (define sum 0)
;;   (set! sum (apply + (map length (list vb-set-1 vb-set-2 vb-res-1 vb-res-2))))
;;   (display sum)
  (set! vb-set-1 #f) (set! vb-set-2 #f) (set! vb-res-1 #f) (set! vb-res-2 #f)
  (sleep 1)
)
;;  (display " done.") (newline))

(when do-level-Vb
;(trim) (trim) (trim)
(display (global-object-count->alist)) (newline)
(display (%global-call-counters)) (newline)

(repeat 2 level-Vb)
;(level-Vb)

(display (global-object-count->alist)) (newline)
(display (%global-call-counters)) (newline)

;; (display "Level V release")
;; (define level-Vb #f)

;; (display (global-object-count->alist)) (newline)
;; (display (%global-call-counters)) (newline)
)

;;
;; Level B
;;

(define mixed-set-positive
  (list 3 10000000000000000000000000000000000000
        ;8/13 1.89 122.0+55.3i)) ;; complex
        8/13 1.89))

(define primary-operations (list + * / -))
(define comparisions (list >= < > <= =))
;(define comparisions (list =))  ;; complex

(define B-set           (list->vector mixed-set-positive))
(define B-shuffles      (list->vector primary-operations))
(define B-comparisions  (list->vector comparisions))

(define (random-ref v)  (vector-ref v (random (vector-length v))))
(define (B-num)         (random-ref B-set))
(define (B-op)          (random-ref B-shuffles))
(define (B-comparision) (random-ref B-comparisions))

(define (B-test)
  ((B-comparision)
   ((B-op) (B-num) (B-num) (B-num))
   ((B-op) (B-num) (B-num) (B-num))
   ((B-op) (B-num) (B-num) (B-num)) ))

(define B-repetitions (* 1k 100))
(when do-level-B
;(trim) (trim) (trim)
(display (global-object-count->alist)) (newline)
(display "Level B repetitions: ") (display B-repetitions) (newline)
(repeat B-repetitions B-test)
(display (global-object-count->alist)) (newline)
(display (global-call-count->alist))   (newline)
)
