(sys-load "VSI-core/repl.scm")

(define yes #t) (define no  #f)
;(define no #t) ; drive all tests 

(define test-intro     no)
(define test-destroy  yes)
(define test-equality  no)

(when test-intro (repl-transcript
 (define d (make-dlist))
 (dlist-add! d 12)
 d
 (dlist-add! d 11)
 (dlist-add! d 10)
))

;; Repetitions of the destruction stress test slowed it
;; down: 1: 16 sec, 2: 33 sec, 3: 1m14s, 4: 2m42
;; Overall performance is better than expected and
;; should meet requirements: one thousand "operations"
;; on a double linked list (vector access with memoized
;; accessors): (/ 33.841 (* 2 1M)) => 1.69205e-5
;; Adaption: 1k op on a set of 1k:
;;   (/ 33.841 (* 2 1k)) => 0.0169205
;; That's already on the edge when we want to arrive
;; at a typing frame rate with about 40ms per frame.
;; (And that isn't even suited to the task of typing.)
;;                                 Cfr. Aftermath/v.scm
(when test-destroy (repl-transcript 
 (global-object-delta) ; (ph1) (ph2) 
 (define l (iota- 1M))
 (define dl (dlist-from-values l))
 (global-object-delta) ; (ph1) (ph2) 
 (equal? (values-from-dlist dl) l)
 (dlist-destroy! dl) 
 (global-object-delta) ; (ph1) (ph2) 
))

(when test-equality (repl-transcript
 (define a (dlist-from-values (values 1 2 3)))
 (define b (dlist-from-values (values 1 2 3)))
 (define c (dlist-from-values (values 3 4 5)))
 (dlist-equal? a b)
 (dlist-equal? b c)
 (dlist-equal? a c)
 (dlist-equal? a a)
 (dlist-equal? b b)
 (dlist-equal? c c)
 (eqv? a b) ;; XXX see VSI-core/structures.scm
 (eqv? b c)
 (eqv? a c)
 (eqv? a a)
 (eqv? b b)
 (eqv? c c)
 (eq? a b)
 (eq? b c)
 (eq? a c)
 (eq? a a)
 (eq? b b)
 (eq? c c)
 (define xx (make-dlist))
 (begin (slot-set! xx 'head    (slot-ref a 'head))
        (slot-set! xx 'current (slot-ref a 'current))
        (slot-set! xx 'tail    (slot-ref a 'tail)))
 (eqv? xx a) ;; XXX
 (eq?  xx a)
 (dlist-equal? xx a)
))

;(REPL)
