(sys-load "VSI-core/repl.scm")

(sys-load "VOS/vos.scm")
(sys-load "Viper/vchars.scm")

(sys-load "GIO/gio.scm")
(sys-load "Examples/label.class.scm")


;; ITL: Input Text Line 
;;                         ITL Layout
;;
;;                       x
;;                       |
;; display width         |----------------------width-1
;;                       |                       |
;; display offset dof----|                       |
;;                 +     |                       |
;;                dof    x                      width-1
;;                 +-----|-----------------------|
;;  visible:       +     ext in a line some text i0
;;                 +-----|---------^-------------|0
;;                 +     |---------^             |0
;;   cursor:       +     |        cpos           |0 append position
;;                 +     |         ^             |        0
;; buffered:      (some text in a line some text in a line0)
;;            head ^               ^                      ^ tail
;;                                 ^ current
;;
;; The final null element (possible input future continuation location)
;; has been chosen to be a #\newline character. Have fun!
;;


;; Quirks
;;
;; - Layout Issues due to terminal cell width calculation for /dof/ et sim


(define itl-font (make-object font-class sys-display
                              "-misc-fixed-medium-r-normal--20-140-100-100-c-100-iso8859-1"))
                              ;"-bitstream-bitstream charter-bold-r-normal--0-0-0-0-p-0-iso8859-1"))
                              ;"-urw-urw gothic l-semibold-r-normal--0-0-0-0-p-0-iso8859-1"))
(gio-font (destroy))
(define gio-font itl-font)

(define itl-class (class                        (extension label-class)
 (slots buf buf-len dof cpos sdraw)
 (slots len)
 (slots font)

 (method bell () #t)

 (method initialize (parent x y len)
  (this (set! 'font itl-font)) ;; this should set the font of the label??
  (let* ((dim (itl-font (line-metrics "M")))
         (height (second dim))
         (width  (* len (first dim))))
   (super " " parent x y width 1)
   (this (set! 'sdraw    0))
   (this (set! 'event-listeners  '()))
   (this (set! 'buf #f))
   (this (set! 'len len))
   (this (set-string " ")))
  (let ((k ((this display) keyboard))
        (p ((this display) pointer-device)))
    (k (add-key-listener this))
    (p (add-button-listener this)))
)

 (method resize (x y length) ;; XXX
  (this (set! 'col x))
  (this (set! 'row y))
  (this (set! 'len length))
  (when (>= (this cpos) length)
   (this (set! 'dof (+ (this dof) 1 (- (this cpos) length)) ))
   (this (set! 'cpos (- length 1))))
 )

 (method set-string (string)
  (when (this buf) (dlist-destroy! (this buf)))
  (let* ((buf (dlist-from-values (string->list string))))
   (dlist-tail buf) (dlist-add! buf #\newline)
   (dlist-head buf)
   (this (set! 'buf    buf))
   (this (set! 'buf-len (string-length string)))
   (this (set! 'dof      0))
   (this (set! 'cpos     0))
 ))

 (method destroy ()
   (dlist-destroy! (this buf))
   (this (set! 'buf 'released)))


 (method expose-event (event)
  ; (X-log-debug "itl expose")
  ; X Quirk: expose event redraws don't need to clear or they flicker.
  (when (= (event count) 0) (this (draw 'dont-clear)))
 )

 (method draw ( . dont-clear)
   (let* (
          ;; (iobuf (cons #t '() ))
          (buf   (this buf))
          (w     (this len))
          (atoms (values-from-dlist buf))
          (dof   (this dof))
          (iobuf (cons #t
                  (list-head w
                   (list-tail dof
                    (apply append (map display-form atoms))))))
          (cpos  (first (itl-font (line-metrics (list->brass (list-head (this cpos) (cdr iobuf)))))))
          (dlen  (- (length iobuf) 1))
          (opos  (this cpos)))
     (this (set! 'label (list->brass (cdr iobuf))))
     ;(this (set! 'label (list->brass atoms)))
     ;(X-log-debug "itl draw atoms " atoms " w " w " dof " dof " label " (this label))
     (when (null? dont-clear) (this (clear)))
     (super)
     (let* ((em  (itl-font (line-metrics "M")))
            (emw (first em))
            (emh (second em))
            (w   (quotient emw 2))
            (w   50)
            (y   (- (this baseline) (quotient w 2)))
            (x   (+ (this indention) (- cpos (quotient w 2)))))
       ;(X-log-debug "itl cursor cpos " cpos " coord " x "," y)
       ((this highlight-gc) (polifill-arc x y 50 50 (* 255 64) (* 30 64))))
))

 (method append-pos? () (dlist-end? (this buf)))
 (method empty-buf?  ()
    (eq? (dlist-head (this buf))
         (dlist-tail (this buf)) ))

 (method buf-len () (this buf-len))
    ;; (- (length (values-from-dlist (this buf))) 1))


 (slots event-listeners)
 (method add-event-listener (listener)
   ; has-method? ...
   (this (set! 'event-listeners (cons listener (this event-listeners)))))


 (method button-event (event)
  (when (and (= 1 (event button))
             (eq? 'down (event type)))
  (X-log-debug "itl button-event " event)
   (this (set! 'cpos (let* (
          ;; (iobuf (cons #t '() ))
          (buf   (this buf))
          (w     (this len))
          (atoms (values-from-dlist buf))
          (dof   (this dof))
          (buf   (list-head w
                   (list-tail dof
                    (apply append (map display-form atoms)))) ))
      (max 0 ( - (itl-font (char-index-for-x buf (event event-x))) 1 )) )))
   (dlist-ref (this buf) (+ (this dof) (this cpos)))
   (this (draw))
))


 (method key-event (event)
  (X-log-debug "itl key-event " event)
  (when (eq? 'down (event type))
  (let ((key-event (event value)))
  ;(lambda (key-event)
     ;; !!! Specify a return value! See event loop protocol. !!!
    (case key-event
      ;; Perhaps some ASCII diagram parsing code generator
      ;; would be more readable! ;-)
      ((enter
        Return) (let* ((string (list->string (values-from-dlist (this buf))))
                      ; now trim the trailing newline
                      (brass  (subrass string 0 (- (brass-length string) 1))))
                 (for-each (lambda (listener) (listener (input-ready brass)))
                           (this event-listeners)))
               #t)

      ((home Home) (this (set! 'dof  0))
              (this (set! 'cpos 0))
              (let ((buf (this buf)))
               ;; (if (<= (- (this len) 1) (this buf-len))
               ;;   (this (set! 'doe (dlist-ref buf (- (this len) 1))))
               ;;   (this (set! 'doe (slot-ref (slot-ref buf 'tail) 'prev))))
               (slot-set! buf 'current (slot-ref buf 'head)))
              (this (draw))
              #t)
      ((end End)  (let ((bl  (this (buf-len)))
                    (dll (buffer-display-length (this buf)))
                    (w   (this len)))
               (cond
                ((< dll w)  (this (set! 'dof 0))
                            (this (set! 'cpos dll)))
                (else       (this (set! 'dof (- dll w)))
                            (this (set! 'cpos w))) ))
              ; (log-debug 'bl bl 'dll dll 'w w 'cpos (this cpos) 'dof (this dof)))
              (let ((buf (this buf)))
               (slot-set! buf 'current (slot-ref buf 'tail)))
              (this (draw))
              #t)

      ((left Left) (let* ((cpos (this cpos))
                     (dof  (this dof ))
                     (prev (dlist-prev (this buf)))
                     (dl   (if (dnull? prev)
                               0
                               (display-length prev))))
                ;(log-debug 'cpos cpos 'dof dof (cons 'prev prev) 'dl dl)
                (cond
                 ((>= cpos dl)          (this (set! 'cpos (- cpos dl))))
                 ((< 0 cpos dl)         (this (set! 'dof (- dof (- dl cpos))))
                                        (this (set! 'cpos 0)))
                 ((and (= cpos 0)
                       (not (= dof 0))) (this (set! 'dof  (max 0 (- dof dl)))))

                 (else                  (this (bell)))))
              (this (draw))
              #t)
      ((right Right) (let* ((cpos (this cpos))
                      (w    (this len))
                      (dof  (this dof))
                      (cur  (slot-ref (slot-ref (this buf) 'current) 'value))
                      (dl   (display-length cur))
                      (tail (dlist-end? (this buf))))
                ; (log-debug 'cpos cpos 'dof dof 'w w 'dl dl)
                ; XXX tail would be append position but we can't reach there
                ;     visibly. Same for end of line.
                (dlist-next (this buf))
                (cond
                ((and (not tail) (<= (+ dl cpos) w))(this (set! 'cpos (+ cpos dl))))
                ((and (not tail) (= cpos w))        (this (set! 'dof (+ dof dl))))
                ((and (not tail) (> (+ dl cpos) w)) (this (set! 'cpos w))
                                 (this (set! 'dof (+ dof (- dl (- w cpos))) )))
                (else (this (bell)) ))
               (this (draw)))
              #t)

      ((#\del Delete)
             (cond
               ((not (dlist-end? (this buf))) (dlist-remove! (this buf)))
               (else                          (this (bell)) ))
              (this (draw)))
      ((#\backspace BackSpace)
              (let* ((cpos (this cpos))                                              
                     (dof  (this dof ))                                              
                     (prev (dlist-prev (this buf)))                                  
                     (dl   (if (dnull? prev)                                         
                               0                                                     
                               (display-length prev))))                              
                ;(log-debug 'cpos cpos 'dof dof (cons 'prev prev) 'dl dl)            
               (if (dnull? prev)
                (begin (this (bell)))
                (begin (dlist-remove! (this buf))
                (cond                                             ; just like /left/                                                               
                 ((>= cpos dl)          (this (set! 'cpos (- cpos dl))))             
                 ((< 0 cpos dl)         (this (set! 'dof (- dof (- dl cpos))))       
                                        (this (set! 'cpos 0)))                       
                 ((and (= cpos 0)                                                    
                       (not (= dof 0))) (this (set! 'dof  (max 0 (- dof dl)))))      
                                                                                     
                 (else                  (this (bell)) ))) ))                         
              (this (draw)) 
       #t)

      (else
       ;; (when (and wingding-mode (< (char->integer key-event) 128))
       ;;  ;(set! key-event (integer->char (+ (char->integer key-event) #x25a1))))
       ;;  (set! key-event (integer->char (+ (char->integer key-event) #x263a))))
       (cond
        ((or ; #t  ;; XXX high code point hack
             (printable? key-event)
             (space? key-event))
         (let ((dl (display-length key-event)))
          (dlist-insert! (this buf) key-event)
          (this (set! 'buf-len (+ (this buf-len) 1))) ; atom length
          (dlist-next    (this buf))
          (if (> (+ dl (this cpos)) (this len))
           (begin (this (set! 'dof  (+ (- dl (- (this len) (this cpos)))
                                       (this dof))))
                  (this (set! 'cpos (this len))))
           (this (set! 'cpos (+ dl (this cpos)))))
         (this (draw))
         #t))
        (else #t))
       #t
  )) )))

;; This should be kept as state of the ITL.
 (define (buffer-display-length buf)
  (let ((chars (values-from-dlist buf)))
   (apply + (map display-length chars))))

)) ; closes define itl class

(define waclass (class (extension itl-class)
 (method input-ready (str)
  (X-log-debug "input ready: " str)
  (eval (list 'repl-transcript (with-input-from-string str read)) '())
  (this (set-string ""))
  (this (draw)))
))

(when #t
(repl-transcript
(define w (make-object waclass sys-display 400 300 50 ))
;(define i (make-object itl-class sys-display 5 5 20))
(w (add-event-listener w))
(w (show))
(sys-control (roll)) ; maybe this should take an initialization lambda
)
)
