(sys-load "VSI-core/repl.scm")
(sys-load "GIO/gio.scm")

(define fonts (sys-display (list-fonts "*")))
(define exts  (sys-display (X-extensions)))
(define xkbe  (sys-display (X-query-extension "XKEYBOARD")))

 (define (wait ms) (let ((res (usleep (* ms 1k))))
                     (when (not (= res 0)) (wait (- ms res)))))

;;; The following f (its rid) will be reported as the child
;;; window of events even if we don't use it. Maybe some sort
;;; of silent warning state???
;; (define f (make-object font-class sys-display
;;  "-misc-fixed-medium-r-normal--20-140-100-100-c-100-iso8859-1"))

;(REPL)

(define gio-font (make-object font-class sys-display
                              ;"-misc-fixed-medium-r-normal--20-140-100-100-c-100-iso8859-1"))
                              ;"-bitstream-bitstream charter-bold-r-normal--0-0-0-0-p-0-iso8859-1"))
                              "-urw-urw gothic l-semibold-r-normal--0-0-0-0-p-0-iso8859-1"))
;; (define gio-font (make-object font-class sys-display "-urw-urw gotic l-semibold-r-normal--0-0-0-0-p-0-iso8859-1"))
;;                                                                ^^^^^
;;  => throw catched: misc-error: "Window Sytem Error xrr query fonts 0"
;; Error recovery? false-if-exception might not be suited to recognize 0 as an undefined error code?
;;
(define menu-class (class  (extension window-class)
 ;; DONE handle grabs to allow for: click, move&select, confirm
 ;; ALMOST: this flat menu class window (surrounding container of the menu items) shouldn't
 ;;         grab or just request an ungrab; see also the method button-event of the menu item class
 ;;   just deselect button events from this event mask <- prerequisite: ChangeWindowAttributes
 (slots items)
 (slots last-y)

 (method initialize (parent x y width height) (super parent x y width height 4)
   (this (set! 'items '()))
   (this (set! 'last-y 6))
 )

 (method add-item (menu-item)
  (when (not (is-a? menu-item menu-item-class)) (error "GIO menu add-item wrong-typ-arg"))
  ;; (when (not (eq? this (menu-item parent)))     (error "GIO menu add-item mischiefed item: wrong parent"))
  (this (set! 'items (append! (this items) (list menu-item)))) ;; doubles add-child
  (this (set! 'last-y (+ (this last-y)
                         (menu-item height)
                         0
                         ;(menu-item border-width) )))
                         (* 2 (menu-item border-width)) )))
 )

 ;;; When the pointer enters a child window that provokes a leave event.
 ;;; Grabbing makes the server send enter/leave events ??? which can
 ;;; provoke a feedback loop. Moving the pointer onto another menu item
 ;;; stops drains that feedback loop. Maybe a regrab (or ChangeGrab???)
 ;;; with handling of implicit enter/leave events ...
 ;;;
 ;;; => ungrab pointer in the menu items
 ;; (method enter-event (event) (X-log-debug (event (description))))
 ;;  (this             (grab-pointer (event time))))
 ;; (method leave-event (event) (X-log-debug (event (description))))
 ;;  ((this display) (ungrab-pointer (event time))))

 (method expose-event (event)
  ; (X-log-debug "menu expose")
  ;; !!! actually every window should handle exposures on their own.
  ;;     The events arrive here only for regions which aren't obscured
  ;;     by the items. The count always seems to be zero (when the
  ;;     regions aren't disjoint).
  (when (= (event count) 0) (this (draw)))
 )

 (method draw ()
  (this (clear))
  ; (for-each (lambda (m) (m (draw))) (this children))
 )

 (method show () (super)
  (for-each (lambda (m) (m (show))) (this children))
 )
))


(define menu-item-class (class                        (extension window-class)
 (slots action label)
 (slots baseline indention reg-gc highlight-gc)
 (slots is-selected)

 (method initialize (menu label action)
  (when (not (is-a? menu menu-class)) (error "GIO menu-item initialize wrong-typ-arg"))
  ; (this (set! 'exposures 0)) ; XXX debugging; see below
  (this (set! 'action action))
  (this (set! 'label label))
  (let* ((border-width 0)
         (dim (gio-font (line-metrics label)))
         (h-space (round (/ (- (menu width) (first dim)) 2)))
         ;(dim (gio-font (line-metrics "M")))
         ;(h-space (floor (* 6 (* 1/13 (first dim)))))
         (ascent  (gio-font ascent))
         (v-space (* 1/2 ascent 1/2))
         (v-space-b (floor   v-space))
         (v-space-u (ceiling v-space))
         (v-space (+ v-space-u v-space-b)))
    (super menu 0 (menu last-y)
           (- (menu width) (* 2 border-width)) (+ ascent v-space)
           border-width)
    (this (set! 'indention h-space))
    (this (set! 'baseline (+ ascent v-space-u))))
  (let ((gc1 (make-object graphics-context-class this))
        (gc2 (make-object graphics-context-class this)))
    (gc1 (set-font gio-font))
    (gc1 (set-color 'foreground (from-namespace sca/x-common 'white-pixel)))
    (gc1 (set-color 'background (from-namespace sca/x-common 'vsiblue-pixel)))
    (this (set! 'reg-gc gc1))
    (gc2 (set-font gio-font))
    (gc2 (set-color 'foreground (from-namespace sca/x-common 'white-pixel)))
    (gc2 (set-color 'background (from-namespace sca/x-common 'black-pixel)))
    ;(gc2 (set-color 'foreground (from-namespace sca/x-common 'white-pixel)))
    ;(gc2 (set-color 'background (from-namespace sca/x-common 'black-pixel)))
    (this (set! 'highlight-gc gc2))
  )
  (this (set! 'is-selected #f))
  (menu (add-item this))
 )

 (method enter-event  (event)  (this (set! 'is-selected #t)) (this (draw)))
 (method leave-event  (event)  (this (set! 'is-selected #f)) (this (draw)))
 (method button-event (event)
  (cond
  ((and (this is-selected)
        (= 1 (event button))
        (eq? 'up (event type))) ((this action)))
  ;; The following flat menu support kills the pop-up interaction model if
  ;; there isn't any notion of the difference somewhere: if the parent is 
  ;; a pop-up the pointer grab - but then ...
  ;; Every window grabs by default? Isn't there any method against that?
  ;; The following also is redundant with respect to is-selected but it's
  ;; part of the X quirks where is-selected is part of our interaction model.
  ((eq? 'down (event type))     ((this display) (ungrab-pointer (event time))))
 ))

 ;; (slots exposures) ; this should become standard? current simple classes don't handle tiled redraws and
                      ; the delivered event count seems to be different from zero only for those regional
                      ; redraws.
 (method expose-event (event)
  ;(this (set! 'exposures (+ (this exposures) 1)))
  (when (= (event count) 0) ;(X-log-debug "menu-item-exposure #" (this exposures)) (wait 20) (this (set! 'exposures 0))
    (this (draw))))

 (method draw ()
  ; (this (clear))
  (let ((gc (if (this is-selected) (this highlight-gc) (this reg-gc))))  ;; put the gc into the menu???
   (gc (image-text-8 (this indention) (this baseline) (this label)))
 ))
)) ; closes menu item class
(define font-menu-item-class (class (extension menu-item-class)
 (method initialize (font . args) (apply super args)
   ((this reg-gc)       (set-font font))
   ((this highlight-gc) (set-font font))
 )
))


(define myw-class (class (extension window-class) (slots font gc menu sdraw)
 (method initialize () (super sys-display 50 50 300 400 0)
                       (this (set! 'sdraw 0))
                       (this (set! 'font (make-object font-class sys-display
                                           ;"-misc-fixed-medium-r-normal--20-140-100-100-c-100-iso8859-1")))
                                           "-bitstream-bitstream charter-bold-r-normal--0-0-0-0-p-0-iso8859-1")))
                       (this (set! 'gc (make-object graphics-context-class this)))
                       (let ((gc (this gc)))
                        (gc (set-font (this font)))
                        (gc (set-color 'foreground (from-namespace sca/x-common 'black-pixel)))
                        (gc (set-color 'background (from-namespace sca/x-common 'vsiblue-pixel)) ))
                       (let ((menu (make-object menu-class this 10 10 200 200)))
                         (this (set! 'menu menu))
                         (make-object menu-item-class menu "Dagobert Duck I"  (lambda () (X-log-debug "DD sdraw " (this sdraw))))
                         (make-object menu-item-class menu "GIO font"         (lambda () ((this gc) (set-font gio-font))))
                         (make-object menu-item-class menu "Apple Juice III"  (lambda () (let ((x (this font)))
                                                                               ((this gc) (set-font gio-font))
                                                                               (this (set! 'font gio-font))
                                                                               (set! gio-font x))))
                         (make-object font-menu-item-class (this font)
                                      menu "myw font"
                                      (lambda () ((this gc) (set-font (this font)))))
                         (make-object menu-item-class menu "quit"    (lambda () (xit)))
                       )
 )
 (method show () ((this menu) (show)) (super))
 (method draw () #t)
 (define wrapper-event-class (class (extension display-object-class) (slots event-x event-y)
                              (method destroy () #t) ; goes to viper object class please?
                              (method initialize (expose-event)
                               (this (set! 'event-x (expose-event x)))
                               (this (set! 'event-y (expose-event y))))))
 (method expose-event (event)
  (cond
  ((= 0 (event count)) (let ((ev (make-object wrapper-event-class event)))
                         (this (clear))
                         (draw-logo this ev)
                         (ev (destroy))))
  (else                (this (set! 'sdraw (+ 1 (this sdraw)))))
 ))

 (define (draw-event-arc this event)
   (let* ((width 100) (height 100)
          (x (- (event event-x) (quotient width  2)))
          (y (- (event event-y) (quotient height 2))))
    ((this gc) (polifill-arc x y width height (* 60 64) (* 180 64))) ))
 (define (draw-logo this event)
   ;; Actually such a complex drawing class private routine is non sense:
   ;;  even the most simple cases should be a drawable or window and keep
   ;;  their state. Only relying on the events to create the illusion of
   ;;  a mutable position of the logo doesn't work out well.
    (draw-event-arc this event)
    ((this gc) (set-color 'foreground (from-namespace sca/x-common 'white-pixel)))
    (this (set! 'last-x (event event-x)))
    (this (set! 'last-y (event event-y)))
    ((this gc) (image-text-8 (event event-x) (event event-y) "VSI/GIO"))
    ((this gc) (set-color 'foreground (from-namespace sca/x-common 'black-pixel)))
)
 (method motion-event (event) ; probably (should be) turned off in the event mask
  (draw-event-arc this event)
  (wait 20)
  (super event)
  (begin #t) ; empty overridden method leads to some strange memoization error message from let*
 )
  ;(super event))
 (slots last-x last-y)
 (method button-event (event)
  (cond
  ((and (eq? (event type) 'down))
  (case (event button)
  ((1) (this (clear)) (draw-logo this event) )
  ((4)
    (this (set! 'last-y (+ (this last-y) scroll-step)))
    (event (set! 'event-x (this last-x)))
    (event (set! 'event-y (this last-y)))
    (this (clear))
    (draw-logo this event))
  ((5)
    (this (set! 'last-y (- (this last-y) scroll-step)))
    (event (set! 'event-x (this last-x)))
    (event (set! 'event-y (this last-y)))
    (this (clear))
    (draw-logo this event))
  ((6)
    (this (set! 'last-x (+ (this last-x) scroll-step)))
    (event (set! 'event-x (this last-x)))
    (event (set! 'event-y (this last-y)))
    (this (clear))
    (draw-logo this event))
  ((7)
    (this (set! 'last-x (- (this last-x) scroll-step)))
    (event (set! 'event-x (this last-x)))
    (event (set! 'event-y (this last-y)))
    (this (clear))
    (draw-logo this event))
  (else                     (super event))
  )))
  ;(super event)
 )
(define scroll-step 8)
 (method key-event (event) (super event)
   (this (clear))
   ((this gc) (set-color 'foreground (from-namespace sca/x-common 'white-pixel)))
   ((this gc) (image-text-8 20 20 "VSI/GIO"))
   ((this gc) (set-color 'foreground (from-namespace sca/x-common 'black-pixel)))
   (this (draw))
 )
))

(repl-clear)
(repl-transcript
(define w (make-object myw-class))
(w (show))
(sys-load "VOS/xi.class.scm")
)
;(REPL)
